/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "cover.h"
#include "toolbox.h"
#include "options/preferences.h"
#include "util/loghook.h"

#include <kio/job.h>
#include <klocale.h>
#include <kurl.h>

#include <QIcon>

#include "aid.h"

namespace Texel {

namespace CoverSearch {

Cover::Cover (QObject* parent)
    :QObject(parent),
     m_SuggestedSizeSet(false)
{
}

Cover::Cover (QObject* parent, const QUrl& source, const QUrl& preview)
    :QObject(parent),
      m_Preview(preview),
      m_SuggestedSizeSet(false)
{
    if (preview.isEmpty()) {
        startLoad (source, false);
    }
    else {
        m_Source=source;
        startLoad (preview, true);
    }
}

Cover::~Cover ()
{
}

void Cover::startLoadCover (bool ok, const QByteArray& url)
{
    if (ok) {
        QByteArray copy=url;
        std::cout << "Copy: " << copy.data() << std::endl;
        std::cout << "fromUser: " << QUrl::fromUserInput(copy).toString() << std::endl;
        startLoad(QUrl::fromUserInput(copy),false);
    }
}

void Cover::startLoad (const QUrl& url, bool preview)
{
    KIO::Job* loadPreview = KIO::storedGet(url,KIO::NoReload,KIO::HideProgressInfo);
    connect (loadPreview, SIGNAL(result(KJob*)), this, SLOT(onImageReceive(KJob*)));
    loadPreview->setProperty("Preview",QVariant(preview));
    loadPreview->start();
    emit startLoading();
}

void Cover::onImageReceive (KJob* job)
{
    KIO::StoredTransferJob* transferJob = dynamic_cast<KIO::StoredTransferJob*>(job);
    disconnect(transferJob, SIGNAL(result(KJob*)), this, SLOT(onImageReceive(KJob*)));
    QImage img;
    if (job->error()==0) {
        img=makeImage(transferJob->data());
    }
    else {
        img=KIconLoader::global()->loadIcon("dialog-warning",KIconLoader::Panel,KIconLoader::SizeLarge).toImage();
        m_Error=job->errorString();
        LogHook::instance()->warning(i18n("Could not load cover from (%2): %1",m_Error,transferJob->url().prettyUrl()),0l);
    }
    if (job->property("Preview").toBool()) {
        if (img.width()>128 && img.width()>img.height()) {
            m_PreviewImage=img.scaledToWidth(128);
        } else if (img.height()>128 ){
            m_PreviewImage=img.scaledToHeight(128);
        } else {
           m_PreviewImage=img;
        }
    }
    else {
        m_SourceImage=img;
        emit coverReceived();
    }
    emit imageReceived();
}

QImage Cover::getPreview() const
{
    if (m_PreviewImage.isNull()) {
        return getCover();
    }
    return m_PreviewImage;
}

QImage Cover::getCover() const
{
    if (m_SourceImage.isNull()) {
        return KIconLoader::global()->loadIcon("image-loading",KIconLoader::Panel,KIconLoader::SizeLarge).toImage();
    }
    return m_SourceImage;
}

void Cover::clear ()
{
    m_SourceImage=m_PreviewImage=QImage();
    m_Source=m_Preview=QUrl();
    m_SuggestedSizeSet=false;
    m_Description=QString::null;
    emit imageReceived();
}

QString Cover::imageFormat (const QByteArray& firstBytes)
{
    static const char magic_jpeg[] = { (const char)0xff, (const char)0xd8, (const char)0 };
    static const char magic_gif[] = "GIF8";
    static const char magic_png[] = { (const char)0x89, (const char)0x50, (const char)0x4E, (const char)0x47,
                                      (const char)0x0D, (const char)0x0A, (const char)0x1A, (const char)0x0A,
                                      (const char)0 };

    if (firstBytes.startsWith(magic_jpeg)) return "JPG";
    if (firstBytes.startsWith(magic_gif)) return "GIF";
    if (firstBytes.startsWith(magic_png)) return "PNG";

    return "other";
}

QImage Cover::makeImage(const QByteArray &data)
{
    QImage cover=QImage::fromData(data);
    cover.setText("Format",imageFormat(data));
    return scale (cover);
}

QImage Cover::scale (const QImage& cover)
{

    if (Preferences::instance()->coverimages.m_ShrinkImage &&
        (Preferences::instance()->coverimages.m_ShrinkX<cover.width() ||
         Preferences::instance()->coverimages.m_ShrinkY<cover.height())) {

        return cover.scaled(Preferences::instance()->coverimages.m_ShrinkX,
                            Preferences::instance()->coverimages.m_ShrinkY,
                            Qt::KeepAspectRatio);
    }
    return cover;
}

QImage Cover::loadImage (const KUrl& url)
{
    if (url.isLocalFile()) {
        QString filename=url.toLocalFile();
        QFileInfo info(filename);
        QImage cover;
        if (info.exists() && info.isReadable()) {
            cover.load(filename);
            QFile file(filename);
            file.open(QIODevice::ReadOnly);
            file.seek(0);
            QByteArray firstBytes = file.read(8);
            cover.setText("Format",imageFormat(firstBytes));
            file.close();
        }
        if (Preferences::instance()->coverimages.m_SaveFileNameInDescription) {
            cover.setText("Desc",url.prettyUrl());
        }

        return scale(cover);
    }

    KIO::StoredTransferJob* job = KIO::storedGet(url);
    // To do: Risky (see KDE docs)
    job->exec();
    return makeImage(job->data());
}

}
}
