/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#ifndef COVER_H
#define COVER_H

#include <QObject>
#include <QImage>
#include <QUrl>
#include <QSize>
#include <QVariant>

class KJob;
class KUrl;

namespace Texel {

namespace CoverSearch {

class XQueryCoverSearcher;

class Cover: public QObject
{
    Q_OBJECT
public:
    Cover (QObject* parent);
    Cover (QObject* parent, const QUrl& source, const QUrl& preview=QUrl());
    virtual ~Cover();

    void setDescription (const QString& desc);
    void setSuggestedSize (const QSize& sugSize);
    void setSourceDomain (const QString& source);
    void setSizeInBytes (int bytes);

    bool hasSize () const;
    QSize getSize () const;
    bool hasBytes () const;
    int getBytes () const;
    QString getSourceDomain () const;

    QString description () const;

    QImage getPreview() const;
    QImage getCover() const;

    QUrl getSourceUrl () const;

    void clear ();

    static QImage makeImage (const QByteArray& data);
    static QImage loadImage (const KUrl& url);

    void setAlbum (const QString& album);
    void setArtist (const QString& artist);

    QString getArtist () const;
    QString getAlbum () const;

    void  setSearcher (XQueryCoverSearcher* searcher);
    XQueryCoverSearcher* getSearcher ();

    void setReference ( const QVariant& ref);
    QVariant getReference () const;

signals:
    void imageReceived();
    void startLoading ();
    void coverReceived ();

public slots:
    void startLoad (const QUrl& url, bool preview=false);
    void startLoadCover (bool ok, const QByteArray& url);

private slots:
    void onImageReceive (KJob* job);

private:
    static QImage scale (const QImage& cover);
    static QString imageFormat (const QByteArray& firstBytes);

    QString m_Album;
    QString m_Artist;

    QImage m_PreviewImage;
    QImage m_SourceImage;
    QUrl m_Source;
    QUrl m_Preview;
    QString m_Description;
    QString m_Error;

    // Reference to the item to which this cover belongs to
    QVariant m_Reference;

    // Contains the size of the image if found before loading. If size if valid the flag
    // m_SuggestedSizeSet is true
    bool m_SuggestedSizeSet;
    QSize m_SuggestedSize;

    // contains the domain from whcih the cover is supposed to loaded. This is not necessary the
    // domain of the url, as many sites use subdomains to host static images.
    QString m_SourceDomain;

    // Size of the resulting image in bytes. If set to -1 no information is available.
    int m_Bytes;

    // Searcher which created this cover
    XQueryCoverSearcher* m_Searcher;
};

inline void Cover::setDescription (const QString &desc)
{
    m_Description=desc;
}

inline QString Cover::description () const
{
    return m_Description;
}

inline QUrl Cover::getSourceUrl () const
{
    return m_Source;
}

inline bool Cover::hasSize () const
{
    return m_SuggestedSizeSet;
}

inline QSize Cover::getSize () const
{
    return m_SuggestedSize;
}

inline bool Cover::hasBytes () const
{
    return m_Bytes>0;
}

inline QString Cover::getSourceDomain () const
{
    return m_SourceDomain;
}

inline int Cover::getBytes () const
{
    return m_Bytes;
}

inline void Cover::setSourceDomain (const QString& source)
{
    m_SourceDomain=source;
}

inline void Cover::setSizeInBytes(int bytes)
{
    m_Bytes=bytes;
}

inline void Cover::setSuggestedSize (const QSize& sugSize)
{
    m_SuggestedSize=sugSize;
    m_SuggestedSizeSet=true;
}

inline void Cover::setAlbum (const QString& album)
{
    m_Album=album;
}

inline void Cover::setArtist (const QString& artist)
{
    m_Artist=artist;
}

inline QString Cover::getArtist () const
{
    return m_Artist;
}

inline QString Cover::getAlbum () const
{
    return m_Album;
}

inline void Cover::setSearcher (XQueryCoverSearcher* searcher)
{
    m_Searcher=searcher;
}

inline XQueryCoverSearcher* Cover::getSearcher ()
{
    return m_Searcher;
}

inline void Cover::setReference ( const QVariant& ref)
{
    m_Reference=ref;
}

inline QVariant Cover::getReference () const
{
    return m_Reference;
}

}
}
#endif // COVER_H
