/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "coversearchwizard.h"
#include "ui_coversearchwizard.h"

#include "xquerycoversearcher.h"
#include "util/xmlhighlighter.h"
#include "util/filedialog.h"
#include "simpleselector.h"

#include <QMovie>
#include <QPixmap>

#include <kstandarddirs.h>
#include <kfiledialog.h>
#include <ktoolinvocation.h>

#include <fcntl.h>

#include "aid.h"

namespace Texel {

namespace CoverSearch {

CoverSearchWizard::CoverSearchWizard(QWidget *parent) :
    QWizard(parent),
    ui(new Ui::CoverSearchWizard),
    m_CoverModel(new CoverSelectorModel(0l,this)),
    m_CurrentCover(0),
    m_CoverImage(new Texel::CoverSearch::Cover(this))
{
    ui->setupUi(this);
    setButtonText (QWizard::CustomButton1,"&Load");
    setButtonText (QWizard::CustomButton2,"&Save");
    connect (this, SIGNAL(customButtonClicked(int)), this, SLOT(customButtonClicked(int)));
    connect (this, SIGNAL(helpRequested()), this, SLOT(onHelpRequested()));

    // Setup of wizard page 0

    // Setup of wizard page 1
    ui->xmlDocument->setTextInteractionFlags(ui->xmlDocument->textInteractionFlags() | Qt::TextSelectableByKeyboard);

    connect (ui->urlInput, SIGNAL(returnPressed(QString)), this, SLOT(load(QString)));
    connect (ui->albumEdit, SIGNAL(returnPressed()), this, SLOT(load()));
    connect (ui->artistEdit, SIGNAL(returnPressed()), this, SLOT(load()));
    connect (ui->webView->page(), SIGNAL(loadProgress(int)), ui->loadProgress, SLOT(setValue(int)));
    connect (ui->evaluateButton, SIGNAL(clicked()), this, SLOT(onPreviewQueryChanged()));
    connect (ui->testPreviewButton, SIGNAL(clicked()), this, SLOT(testPreviewButtonClicked()));
    connect (ui->throbbler, SIGNAL(clicked()), this, SLOT(load()));
    connect (ui->updateUrlButton, SIGNAL(clicked()), this, SLOT(onUpdateUrl()));
    connect (&searcher, SIGNAL(previewXMLResult(QByteArray)), this, SLOT(onPreviewXML(QByteArray)));

    ui->throbbler->setIcons("process-working-kde");
    connect (ui->webView, SIGNAL(loadStarted()), ui->throbbler, SLOT(start()));
    connect (ui->webView, SIGNAL(loadFinished(bool)), ui->throbbler, SLOT(stop()));

    new XmlHighlighter(ui->resultTextBox);
    new XmlHighlighter(ui->xmlDocument);
    new XmlHighlighter(ui->xqueryEdit);

    // Setup of wizardpage 2
    ui->xmlDocument_2->setTextInteractionFlags(ui->xmlDocument_2->textInteractionFlags() | Qt::TextSelectableByKeyboard);

    connect (ui->copyUrlButton, SIGNAL(clicked()), this, SLOT(onUpdateCoverUrl()));
    connect (ui->webView_2->page(), SIGNAL(loadProgress(int)), ui->loadProgress_2, SLOT(setValue(int)));
    connect (ui->directLinkCheck, SIGNAL(clicked(bool)), ui->urlContent, SLOT(setDisabled(bool)));
    connect (ui->evaluateButton_2, SIGNAL(clicked()), this, SLOT(onExtractQueryChanged()));
    connect (m_CoverImage, SIGNAL(imageReceived()), this, SLOT(onCoverImageReceived()));
    connect (m_CoverImage, SIGNAL(startLoading()), this, SLOT(onStartLoadingCover()));

    connect (&searcher, SIGNAL(extractXMLResult(QByteArray)), this, SLOT(onCoverXML(QByteArray)));

    ui->throbbler_2->setIcons("process-working-kde");
    connect (ui->webView_2, SIGNAL(loadStarted()), ui->throbbler_2, SLOT(start()));
    connect (ui->webView_2, SIGNAL(loadFinished(bool)), ui->throbbler_2, SLOT(stop()));

    new XmlHighlighter(ui->resultTextBox_2);
    new XmlHighlighter(ui->xmlDocument_2);
    new XmlHighlighter(ui->xqueryCoverEdit);
}

CoverSearchWizard::~CoverSearchWizard()
{
    //delete ui->throbbler->movie();
    delete ui;
}

void CoverSearchWizard::onCoverImageReceived ()
{
    QImage cover = m_CoverImage->getCover();
    const qreal wRatio = (qreal)cover.width()/(qreal)ui->coverView->width();
    const qreal hRatio = (qreal)cover.height()/(qreal)ui->coverView->height();
    const qreal ratio = std::max(1.0,std::max(wRatio,hRatio));
    ui->coverView->setPixmap(QPixmap::fromImage(cover).scaled((qreal)cover.width()/ratio,(qreal)cover.height()/ratio));
}

void CoverSearchWizard::load ()
{
    QString url=ui->urlInput->text();
    if (!url.isEmpty()) {
        load(url);
    }
}

void CoverSearchWizard::load (const QString& url)
{
    std::cout << "Load: " << url << std::endl;
    searcher.setWebPage(ui->webView->page());
    searcher.setPreviewUrlTemplate(url);
    searcher.search(ui->artistEdit->text(),ui->albumEdit->text());
}

void CoverSearchWizard::onPreviewXML (const QByteArray& xml)
{
    ui->xmlDocument->setPlainText(xml);
}

void CoverSearchWizard::onCoverXML (const QByteArray& xml)
{
    ui->xmlDocument_2->setPlainText(xml);
}

void CoverSearchWizard::onUpdateUrl ()
{
    ui->urlInput->setText(ui->webView->url().toString());
}

void CoverSearchWizard::onPreviewQueryChanged ()
{
    searcher.setQueryPreview(ui->xqueryEdit->toPlainText());
    SimpleXMLFromPage* query = new SimpleXMLFromPage(this);
    query->addVar("squareOnly",QVariant(true));
    query->addVar("maxpreviews",QVariant(20));
    QPair<bool,QByteArray> result=query->executeQuery(ui->xmlDocument->toPlainText().toUtf8(), ui->xqueryEdit->toPlainText());
    searcher.setPreviewXQueryResult(result.first,result.second);
    delete query;

    if (result.first==true) {
        ui->resultTextBox->setPlainText(result.second);

        try {
            m_CoverModel->removeRows(0,m_CoverModel->rowCount());
            m_CurrentCover=0;
            searcher.XQuerySearcher::parsePreviewResultXML(result.second, m_CoverModel);
            ui->checkResult->setPixmap(KIconLoader::global()->loadIcon("task-complete",KIconLoader::Toolbar));
            ui->checkResult->setToolTip(i18n("Result parseable"));
            ui->copyUrlButton->setEnabled(true);
        }
        catch (XQueryCoverSearcher::xml_parse_error e) {
            ui->checkResult->setPixmap(KIconLoader::global()->loadIcon("task-attention",KIconLoader::Toolbar));
            ui->checkResult->setToolTip(e.errormsg);
            ui->copyUrlButton->setEnabled(false);
        }
    }
    else {
        ui->checkResult->setPixmap(KIconLoader::global()->loadIcon("task-rejected",KIconLoader::Toolbar));
        ui->checkResult->setToolTip(i18n("Error in xquery"));
        ui->resultTextBox->setHtml(result.second);
        ui->copyUrlButton->setEnabled(false);
    }
}

void CoverSearchWizard::onExtractQueryChanged ()
{
    searcher.setQueryExtraction(ui->xqueryCoverEdit->toPlainText());
    SimpleXMLFromPage* query = new SimpleXMLFromPage(this);
    QPair<bool,QByteArray> result=query->executeQuery(ui->xmlDocument_2->toPlainText().toUtf8(), ui->xqueryCoverEdit->toPlainText());
    delete query;

    if (result.first==true) {
        ui->resultTextBox_2->setPlainText(result.second);
        if (result.second.count('\n')<=1 && result.second.count('\r')<=1) {
            // single line: assume URL and load it
            QUrl url=QUrl::fromUserInput(result.second);
            m_CoverImage->startLoad(url,false);
        }
    }
    else {
        ui->resultTextBox_2->setText(result.second);
        ui->coverView->setPixmap(KIconLoader::global()->loadIcon("dialog-warning",KIconLoader::Panel,KIconLoader::SizeLarge));
    }
}

void CoverSearchWizard::customButtonClicked (int which)
{
    if (which==QWizard::CustomButton1) loadConfig();
    else if (which==QWizard::CustomButton2) saveConfig();
}

void CoverSearchWizard::loadConfig ()
{
    QString startDir=KGlobal::dirs()->resourceDirs ("grabber").first();
    if (!KGlobal::dirs()->exists(startDir)) {
        KGlobal::dirs()->makeDir(startDir);
    }
    QString filename=KFileDialog::getOpenFileName(startDir,"*.xml",0l,i18n("Load cover search config"));

    if (filename!="") {
        searcher.loadConfig(filename);
        ui->urlInput->setText(searcher.previewUrlTemplate());
        ui->xqueryEdit->setPlainText(searcher.queryPreview());
        ui->titleEdit->setText(searcher.title());
        ui->descEdit->setPlainText(searcher.description());
        ui->xqueryCoverEdit->setText(searcher.queryExtraction());
        ui->directLinkCheck->setChecked(searcher.resultIsUrl());
        if (KGlobal::dirs()->checkAccess(filename,W_OK)) {
            m_Filename=filename;
        }
    }
}

void CoverSearchWizard::saveConfig ()
{
    QString startDir=m_Filename;

    if (m_Filename.isEmpty()) {
        startDir=KGlobal::dirs()->resourceDirs ("grabber").first();
        if (!KGlobal::dirs()->exists(startDir)) {
            KGlobal::dirs()->makeDir(startDir);
        }
    }
    QString filename=KFileDialog::getSaveFileName(startDir,"*.xml",0l,i18n("Save cover search config"));

    if (filename!="") {
        searcher.setQueryPreview(ui->xqueryEdit->toPlainText());
        searcher.setPreviewUrlTemplate(ui->urlInput->text());
        searcher.setDescription(ui->descEdit->toPlainText());
        searcher.setTitle(ui->titleEdit->text());
        searcher.setQueryExtraction(ui->xqueryCoverEdit->toPlainText());
        searcher.setResultIsUrl(ui->directLinkCheck->isChecked());
        searcher.saveConfig(filename);
    }
}

void CoverSearchWizard::testPreviewButtonClicked()
{
    QList<XQueryCoverSearcher*> searchers;
    QList<SimpleSelector::SearchArgs> args;

    searchers << &searcher;
    args << SimpleSelector::SearchArgs(ui->artistEdit->text(),ui->albumEdit->text());

    SimpleSelector* dialog = new SimpleSelector(searchers,args);
    dialog->show();
}

void CoverSearchWizard::onUpdateCoverUrl ()
{
    m_CoverImage->clear();
    if (m_CoverModel->rowCount()>0) {
        ++m_CurrentCover;
        if (m_CurrentCover>=m_CoverModel->rowCount()) {
            m_CurrentCover=0;
        }
        QUrl source=m_CoverModel->getCover(m_CoverModel->index(m_CurrentCover,0))->getSourceUrl();
        ui->mainUrlInput->setText(source.toString());

        if (ui->directLinkCheck->isChecked()) {
            m_CoverImage->startLoad(source.toString(),false);
        }
        else {
            loadCoverPage ();
        }
    }
}

void CoverSearchWizard::loadCoverPage()
{
    searcher.setWebPage(ui->webView_2->page());
    searcher.XQuerySearcher::startExtraction(QUrl::fromUserInput(ui->mainUrlInput->text()));
}

void CoverSearchWizard::onHelpRequested()
{
    KToolInvocation::invokeHelp("newcoversource");
}

void CoverSearchWizard::onStartLoadingCover ()
{
}

}
}
