/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


*************************************************************************/

#include "simpleselector.h"
#include "ui_simpleselector.h"
#include "toolbox.h"
#include "util/widgets/htmldelegate.h"
#include "coverselectormodel.h"

#include <QStandardItemModel>
#include <QListView>
#include <QLabel>
#include <QMovie>
#include <QResizeEvent>

#include <ktoolinvocation.h>

#include <set>

#include "aid.h"

using namespace std;

namespace Texel {
    namespace CoverSearch {

        SimpleSelector::SimpleSelector(QList<XQueryCoverSearcher*> searchers,const QList<SearchArgs>& args) :
                QWidget(0l),
                ui(new Ui::SimpleSelector),
                m_Searchers(),
                m_Args(args),
                m_OpenSearches(0)
        {
            ui->setupUi(this);
            setWindowIcon(KIconLoader::global()->loadIcon("search-cover",KIconLoader::Toolbar));

            ui->errorLabel->hide();
            ui->errorIcon->hide();
            while (ui->tabbedList->count()>0) ui->tabbedList->removeTab(0);

            m_Searchers=searchers;

            // Create all tab
            m_Master=createListView(i18n("All"),0l,0l);
            m_MasterController = new Internal::ListController(0,dynamic_cast<QListView*>(ui->tabbedList->widget(0)),m_Master);

            foreach (XQueryCoverSearcher* searcher, m_Searchers) {
                createListView(searcher->title(), m_Master, searcher);
                // Wire tab
                connect (searcher, SIGNAL(startSearch()), this, SLOT(onSearchStart()));
                connect (searcher, SIGNAL(endSearch()), this, SLOT(onSearchEnd()));
            }

            connect (ui->albumList, SIGNAL(clicked(QModelIndex)), this, SLOT(onAlbumSelection(QModelIndex)));
            connect (ui->previousButton, SIGNAL(clicked()), this, SLOT(onPreviousAlbum()));
            connect (ui->nextButton, SIGNAL(clicked()), this, SLOT(onNextAlbum()));
            connect (ui->researchButton, SIGNAL(clicked()), this, SLOT(onResearchClicked()));

            connect (ui->closeButton, SIGNAL(clicked(bool)), this, SLOT(close()));
            connect (ui->errorLabel, SIGNAL(linkActivated(QString)), this, SLOT(onErrorLink(QString)));

            setAttribute(Qt::WA_DeleteOnClose);
            setWindowTitle(i18n("Search for cover"));

            // Fill album list
            QStandardItemModel* albumListModel=new QStandardItemModel(this);
            ui->albumList->setModel (albumListModel);
            for (int i=0;i<m_Args.size();i++) {
                QStandardItem* item=new QStandardItem();
                item->setText(m_Args[i].artist+" - "+m_Args[i].album);
                item->setIcon(QIcon(QPixmap::fromImage(m_Args[i].cover)));
                item->setData(m_Args[i].artist,Qt::UserRole+1);
                item->setData(m_Args[i].album,Qt::UserRole+2);
                albumListModel->appendRow(item);
            }

            if (args.size()>0) {
                onAlbumSelection(albumListModel->index(0,0));
            }
        }

        CoverSelectorModel* SimpleSelector::createListView (const QString& name, CoverSelectorModel* master,
                                                            XQueryCoverSearcher* searcher)
        {
            QListView* list=new QListView(ui->tabbedList);
            list->setIconSize(QSize(128,128));
            list->setGridSize(QSize((3*128),135));
            list->setUniformItemSizes(true);
            list->setFlow(QListView::LeftToRight);
            list->setWrapping(true);
            list->setResizeMode(QListView::Adjust);
            HtmlDelegate* delegate = new HtmlDelegate(list);
            delegate->setGridSize(QSize(3*128,128));
            // Give 4 points more to leave space to the text
            delegate->setIconSize(QSize(132,128));
            list->setItemDelegate(delegate);

            try {
                CoverSelectorModel* model=new CoverSelectorModel(master,this);

                if (searcher!=0l) {
                    searcher->setModel(model);
                    Internal::ListController* controller = new Internal::ListController(searcher,list,model);
                    connect (searcher, SIGNAL(startSearch()), controller, SLOT(onSearchStart()));
                    connect (searcher, SIGNAL(progress(int)), controller, SLOT(onProgress(int)));
                    connect (searcher, SIGNAL(endSearch()), controller, SLOT(onSearchEnd()));
                }
                connect (list, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(onDoubleClick(QModelIndex)));

                list->setModel (model);
                ui->tabbedList->addTab (list, name);

                return model;
            }
            catch (XQueryCoverSearcher::xml_parse_error e) {
            }
            return 0l;
        }

        void SimpleSelector::onDoubleClick (const QModelIndex& index)
        {
            const CoverSelectorModel* m = dynamic_cast<const CoverSelectorModel*>(index.model());
            Cover* cover=const_cast<Cover*>(m->getCover(index));
            connect (cover, SIGNAL(coverReceived()), this, SLOT(onSelectedCover()));
            cover->getSearcher()->startExtraction (cover);
        }

        void SimpleSelector::onSelectedCover ()
        {
            Cover* cover = dynamic_cast<Cover*>(sender());

            if (cover!=0l) {
                emit newCover(cover->getArtist(),cover->getAlbum(),cover->getCover());
                QStandardItemModel* model = dynamic_cast<QStandardItemModel*>(ui->albumList->model());
                const int row = cover->getReference().toInt();
                model->item(row)->setIcon(QIcon(QPixmap::fromImage(cover->getCover())));
            }
        }

        SimpleSelector::~SimpleSelector()
        {
            delete ui;
        }

        void SimpleSelector::onError (const QString& message)
        {
            ui->errorLabel->show();
            ui->errorIcon->show();
            ui->errorLabel->setText(message);
            ui->errorIcon->setPixmap(KIconLoader::global()->loadIcon("dialog-warning",KIconLoader::Panel,KIconLoader::SizeLarge));
        }

        void SimpleSelector::changeEvent(QEvent *e)
        {
            QWidget::changeEvent(e);
            switch (e->type()) {
            case QEvent::LanguageChange:
                ui->retranslateUi(this);
                break;
            default:
                break;
            }
        }

        void SimpleSelector::onAlbumSelection (const QModelIndex& i)
        {
            m_SelectedRow=i.row();
            ui->albumList->setCurrentIndex(i);
            dynamic_cast<CoverSelectorModel*>(m_Master)->setReference(i.row());
            foreach (XQueryCoverSearcher* searcher, m_Searchers) {
                dynamic_cast<CoverSelectorModel*>(searcher->model())->setReference(i.row());
            }
            onAlbumSelection(m_Args[i.row()].artist,m_Args[i.row()].album);
        }

        void SimpleSelector::onAlbumSelection (const QString& artist, const QString& album)
        {
            m_Master->removeRows(0,m_Master->rowCount());
            foreach (XQueryCoverSearcher* searcher, m_Searchers) {
                searcher->model()->removeRows(0,searcher->model()->rowCount());
                searcher->search(artist,album);
            }
            ui->previousButton->setDisabled(m_SelectedRow==0);
            ui->nextButton->setDisabled(m_SelectedRow==ui->albumList->model()->rowCount()-1);
            ui->title->setText (artist+"\n"+album);
            ui->albumEdit->setText(album);
            ui->artistEdit->setText(artist);
        }

        void SimpleSelector::onResearchClicked ()
        {
            onAlbumSelection(ui->artistEdit->text(), ui->albumEdit->text());
        }

        void SimpleSelector::onSearchStart ()
        {
            if (m_OpenSearches==0) {
                m_MasterController->onSearchStart ();
                // Hide error, will be shown again when error occures
                ui->errorLabel->hide();
                ui->errorIcon->hide();
            }
            ++m_OpenSearches;
        }

        void SimpleSelector::onSearchEnd ()
        {
            // Can be called multiple times as label is onyl deleted once
            m_MasterController->onSearchEnd ();

            --m_OpenSearches;
            if (m_OpenSearches==0) {
                if (m_Master->rowCount()==0) {
                    onError (i18n("No results found"));
                }
            }
        }

        void SimpleSelector::onNextAlbum ()
        {
            if (m_SelectedRow<ui->albumList->model()->rowCount()-1) {
                onAlbumSelection(ui->albumList->model()->index(m_SelectedRow+1,0));
            }
        }

        void SimpleSelector::onPreviousAlbum ()
        {
            if (m_SelectedRow>0) {
                onAlbumSelection(ui->albumList->model()->index(m_SelectedRow-1,0));
            }
        }

        void SimpleSelector::onErrorLink (const QString& url)
        {
            KToolInvocation::invokeBrowser(url);
        }

        namespace Internal
        {
            ListController::ListController (XQueryCoverSearcher* searcher, QListView* view,
                                            CoverSelectorModel* model)
            : QObject (view),
                m_Model(model),
                m_Searcher(searcher),
                m_View(view),
                m_OverlayLabel(0l),
                m_ProgressLabel(0l)
            {
            }

            ListController::~ListController ()
            {
            }

            void ListController::onSearchStart ()
            {
                m_OverlayLabel=createOverlay();
            }

            void ListController::onProgress (int p)
            {
                if (m_ProgressLabel!=0l) {
                    m_ProgressLabel->setText (QString::number(p)+"%");
                }
            }

            void ListController::onSearchEnd ()
            {
                if (m_OverlayLabel!=0l) {
                    delete m_OverlayLabel;
                    m_OverlayLabel=0l;
                    m_ProgressLabel=0l;
                }
            }

            bool ListController::eventFilter(QObject *obj, QEvent *event)
            {
                if (obj==m_View && event->type()==QEvent::Resize && m_OverlayLabel!=0l) {
                    QResizeEvent* resizeEvent = dynamic_cast<QResizeEvent*>(event);
                    centerContainer(m_OverlayLabel, resizeEvent->size());
                }
                return QObject::eventFilter(obj,event);
            }

            void ListController::centerContainer(QWidget* container, const QSize&  viewSize)
            {
                const QSize size=container->sizeHint();

                container->setGeometry((viewSize.width()-size.width())/2,
                                       (viewSize.height()-size.height())/2,
                                       size.width(),size.height());
            }

            QWidget* ListController::createOverlay()
            {
                QWidget* container= new QWidget(m_View);
                QLabel* throbblerLabel = new QLabel(container);
                m_View->installEventFilter(this);
                QMovie* throbbler = KIconLoader::global()->loadMovie("throbbler",KIconLoader::User);
                if (throbbler!=0l) {
                    throbblerLabel->setMovie(throbbler);
                    throbbler->start();
                }

                m_ProgressLabel = new QLabel(container);
                m_ProgressLabel->setAlignment(Qt::AlignCenter);

                QLayout* layout=new QVBoxLayout(container);
                container->setLayout(layout);
                layout->addWidget(throbblerLabel);
                layout->addWidget(m_ProgressLabel);

                centerContainer(container, m_View->size());
                container->show();
                return container;
            }
        }

        void SimpleSelector::search (const TagItem::List& items, QObject* receiver, const char* slot)
        {
            // std::set to avoid hash function
            std::set<CoverSearch::SimpleSelector::SearchArgs> searchArgs;
            foreach (TagItem* item, items) {
                CoverSearch::SimpleSelector::SearchArgs arg;
                arg.artist = item->data(ColumnManager::Artist,Columns::Column::RoleDisplay);
                arg.album  = item->data(ColumnManager::Album,Columns::Column::RoleDisplay);
                arg.cover  = item->getDefaultCoverImage();

                std::set<CoverSearch::SimpleSelector::SearchArgs>::iterator ref=searchArgs.find(arg);
                if (ref!=searchArgs.end()) {
                    if ((*ref).cover!=arg.cover) {
                        // const_cast necessary? its a iterator, not a const_iterator!
                        const_cast<CoverSearch::SimpleSelector::SearchArgs&>(*ref).cover=QImage();
                    }
                }
                else {
                    searchArgs.insert(arg);
                }
            }

            QList<CoverSearch::SimpleSelector::SearchArgs> list;
            foreach (CoverSearch::SimpleSelector::SearchArgs a, searchArgs) {
                list << a;
            }

            QList<XQueryCoverSearcher*> searchers = XQueryCoverSearcher::getConfiguredSearchers(true,0l);
            SimpleSelector* selector=new SimpleSelector(searchers,list);
            foreach (XQueryCoverSearcher* s, searchers) s->setParent(selector);
            QObject::connect (selector, SIGNAL(newCover(QString,QString,QImage)), receiver, slot);
            selector->show();
        }
    }
}

#include "simpleselector.moc"
