/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "xquerycoversearcher.h"
#include "coverselectormodel.h"
#include "options/preferences.h"
#include "util/loghook.h"

#include <QBuffer>
#include <QXmlStreamWriter>
#include <QProgressBar>
#include <QStandardItemModel>
#include <QBuffer>
#include <QWebPage>

#include <kglobal.h>
#include <kstandarddirs.h>
#include <klocale.h>

#include "aid.h"

namespace Texel {

namespace CoverSearch {

XQueryCoverSearcher::XQueryCoverSearcher(QObject *parent) :
    XQuerySearcher(parent),
    m_MaxPreviews(20)
{
    setWebPage(new QWebPage(this));
}

QUrl XQueryCoverSearcher::getSpecificUrl (const QString& artist, const QString& album) const
{
    QString replacer=m_PreviewUrlTemplate;
    QUrl specificUrl(replacer.replace("${Artist}",artist).replace("${Album}",album));
    if (specificUrl.scheme().isEmpty()) {
        specificUrl.setScheme("http");
    }
    return specificUrl;
}

void XQueryCoverSearcher::search (const QString& artist, const QString& album)
{
    m_Artist=artist;
    m_Album=album;
    QUrl specificUrl = getSpecificUrl(artist,album);
    XQuerySearcher::search (specificUrl);
}

void XQueryCoverSearcher::doAddVariablesToQuery (SimpleXMLFromPage *loader)
{
    loader->addVar("squareOnly",QVariant(m_SquareImagesOnly));
    loader->addVar("maxpreviews",QVariant(m_MaxPreviews));
}

void XQueryCoverSearcher::startExtraction (Cover* cover)
{
    if (resultIsUrl()) {
        cover->startLoad(cover->getSourceUrl(),false);
    }
    else {
        connect (this, SIGNAL(extractQueryResult(bool,QByteArray)), cover, SLOT(startLoadCover(bool,QByteArray)));
        XQuerySearcher::startExtraction(cover->getSourceUrl());
    }
}

void XQueryCoverSearcher::setPreviewUrlTemplate (const QString& url)
{
    m_PreviewUrlTemplate=url;
}

void XQueryCoverSearcher::parseCoverSource_result (QXmlStreamReader* reader, QAbstractItemModel* model) throw (xml_parse_error)
{
    while (!reader->atEnd()) {
        if (reader->readNextStartElement()) {
            if (reader->name()=="cover") {
                parseCoverSource_cover (reader,model);
            }
            else {
                xml_parse_error err;
                err.errornum=QXmlStreamReader::CustomError;
                err.errormsg="Unexpected tag: \""+reader->name().toString()+"\" (allowed: <cover>)";
                err.line=reader->lineNumber();
                err.col=reader->columnNumber();
                throw err;
            }
        }
    }
}

void XQueryCoverSearcher::parseCoverSource_cover (QXmlStreamReader* reader, QAbstractItemModel* model) throw (xml_parse_error)
{
    QByteArray previewurl;
    QByteArray sourceurl;
    QString desc;
    QSize sugSize;
    QString sourcedomain;
    bool sugSizeSet=false;
    int bytes=-1;

    while (!reader->atEnd()) {
        switch (reader->readNext()) {
        case QXmlStreamReader::StartElement:
            if (reader->name()=="preview") {
                previewurl=reader->attributes().value("src").toAscii();
            } else if (reader->name()=="size") {
                if (reader->attributes().hasAttribute("height") && reader->attributes().hasAttribute("width")) {
                    bool ok1,ok2;
                    sugSize=QSize (
                                reader->attributes().value("height").toString().toInt(&ok1),
                                reader->attributes().value("width").toString().toInt(&ok2)
                                );
                    sugSizeSet=ok1&&ok2;
                }
                if (reader->attributes().hasAttribute("bytes")) {
                    bool ok;
                    bytes = reader->attributes().value("bytes").toString().toInt(&ok);
                    if (!ok) bytes=-1;
                }
            } else if (reader->name()=="sourcedomain") {
                while (reader->readNext()!=QXmlStreamReader::Characters);
                sourcedomain=reader->text().toString();
            } else if (reader->name()=="destpage") {
                sourceurl=reader->attributes().value("href").toAscii();
            } else if (reader->name()=="description") {
                while (reader->readNext()!=QXmlStreamReader::Characters);
                desc=reader->text().toString();
            } else {
                xml_parse_error err;
                err.errornum=QXmlStreamReader::CustomError;
                err.errormsg="Unexpected tag: "+reader->name().toString();
                err.line=reader->lineNumber();
                err.col=reader->columnNumber();
                throw err;
            }
            break;
        case QXmlStreamReader::EndElement:
            if (reader->name()=="cover") {
                if (model!=0l) {
                    Texel::CoverSearch::Cover* cover = new Texel::CoverSearch::Cover(model, QUrl::fromEncoded(sourceurl), QUrl::fromEncoded(previewurl));
                    cover->setDescription(desc);
                    cover->setSourceDomain(sourcedomain);
                    cover->setSizeInBytes(bytes);
                    cover->setSearcher(this);
                    if (sugSizeSet) {
                        cover->setSuggestedSize(sugSize);
                    }
                    model->insertRow(0);
                    model->setData(model->index(0,0),QVariant::fromValue(dynamic_cast<QObject*>(cover)));
                    CoverSelectorModel* m =dynamic_cast<CoverSelectorModel*>(model);
                    if (m!=0l) {
                        if (m->master()!=0l) {
                            m->master()->insertRow(0);
                            m->master()->setData(model->index(0,0),QVariant::fromValue(dynamic_cast<QObject*>(cover)));
                            connect (cover, SIGNAL(imageReceived()), m, SLOT(updatedCover()));
                            connect (cover, SIGNAL(imageReceived()), m->master(), SLOT(updatedCover()));
                        }
                    }

                }
                return;
            }
            break;
         case QXmlStreamReader::Invalid:
            {
                xml_parse_error err;
                err.errornum=reader->error();
                err.errormsg=reader->errorString();
                err.line=reader->lineNumber();
                err.col=reader->columnNumber();
                throw err;
            }
            break;
        default:
            ; // Ignore
        }
    }
}

void XQueryCoverSearcher::parsePreviewResultXML (QXmlStreamReader* reader, QAbstractItemModel* model) throw (xml_parse_error)
{
    CoverSelectorModel* m = dynamic_cast<CoverSelectorModel*>(model);
    int leftToAdd = Preferences::instance()->coverimages.m_MaximumItemsToQuery;
    if (m!=0l) {
        m->setArtistAndAlbum(m_Artist,m_Album);
        if (m->master()!=0l) m->master()->setArtistAndAlbum(m_Artist,m_Album);
    }
    if (reader->readNextStartElement()) {
        if (reader->name()=="result") {
            if (leftToAdd>0) {
                parseCoverSource_result (reader,model);
                --leftToAdd;
            }
            else {
                LogHook::instance()->warning (i18n("More results than expected (%1), ignoring left overs.",
                                              Preferences::instance()->coverimages.m_MaximumItemsToQuery),0l);
            }
        }
        else {
            xml_parse_error err;
            err.errornum=QXmlStreamReader::CustomError;
            err.errormsg="Unexpected tag: \""+reader->name().toString()+"\" (allowed: <result>)";
            err.line=reader->lineNumber();
            err.col=reader->columnNumber();
            throw err;
        }
    }
}

void XQueryCoverSearcher::doSaveLocalConfig (QXmlStreamWriter& writer)
{
    writer.writeStartElement("page");
    writer.writeAttribute("url",m_PreviewUrlTemplate);
    writer.writeEndElement();
}

void XQueryCoverSearcher::doReadLocalConfig (QXmlStreamReader& reader)
{
    if (reader.name()=="page") {
        m_PreviewUrlTemplate=reader.attributes().value("url").toString();
    }
}

QList<XQueryCoverSearcher*> XQueryCoverSearcher::getConfiguredSearchers (bool activeOnly, QObject *parent)
{
    QList<XQueryCoverSearcher*> searchers;

    foreach (QString filename, KGlobal::dirs()->findAllResources("grabber","*.xml",KStandardDirs::NoDuplicates)) {
        XQueryCoverSearcher* searcher = new XQueryCoverSearcher(parent);
        searcher->loadConfig(filename);
        if ((activeOnly && Preferences::instance()->htmlgrabber.isCoverGrabberActive(searcher->title())) || !activeOnly) {
            searcher->setObjectName(filename);
            searcher->setSquareImagesOnly (Preferences::instance()->coverimages.m_SquareImagesOnly);
            searcher->setMaxPreviews(Preferences::instance()->coverimages.m_MaximumItemsToQuery);
            searchers << searcher;
        }
    }

    return searchers;
}


}

}

#include "xquerycoversearcher.moc"

