#include "xquerysearcher.h"
#include "util/loghook.h"

#include <QWebElement>
#include <QWebFrame>
#include <QWebPage>
#include <qxmlquery.h>
#include <QXmlFormatter>
#include <QBuffer>
#include <QFile>

#include <klocale.h>

#include "aid.h"

namespace Texel {

namespace CoverSearch {

SimpleXMLFromPage::SimpleXMLFromPage (QObject* parent)
    : QAbstractMessageHandler(parent),
      m_Page(0l),
      m_AutoDelete(false)
{
}

SimpleXMLFromPage::~SimpleXMLFromPage()
{
}

void SimpleXMLFromPage::load (const QUrl& url)
{
    if (m_Page==0l) {
        m_Page=new QWebPage(this);
    }
    connect (m_Page, SIGNAL(loadFinished(bool)), this, SLOT(onPageLoaded(bool)));
    connect (m_Page, SIGNAL(loadProgress(int)), this, SIGNAL(progress(int)));
    m_Page->mainFrame()->load(url);
}

bool isAttributeValid (const QString& att)
{
    for (int i=0;i<att.size();i++) {
        if (!att[i].isLetterOrNumber() && att[i]!='_') {
            return false;
        }
    }
    return true;
}

void SimpleXMLFromPage::addVar(const QString &name, const QVariant &value)
{
    m_QueryBindVariables.insert(name, value);
}

void SimpleXMLFromPage::createSimpleStructure (const QWebElement& elem, QXmlStreamWriter* writer)
{
    if (elem.isNull()) return;

    writer->writeStartElement(elem.tagName());
    foreach (QString name, elem.attributeNames()) {
        if (!name.startsWith("on") && isAttributeValid(name)) {
            writer->writeAttribute(name,elem.attribute(name));
        }
    }

    if (elem.tagName()!="SCRIPT" && elem.tagName()!="STYLE") {
        writer->writeCharacters(elem.toPlainText());
    }

    for (QWebElement c=elem.firstChild();!c.isNull();c=c.nextSibling()) {
        createSimpleStructure(c,writer);
    }

    writer->writeEndElement();
}

QPair<bool,QByteArray> SimpleXMLFromPage::executeQuery (const QByteArray& inputXML, const QString& queryString)
{
    QByteArray xml(inputXML);
    QBuffer buffer(&xml);
    QXmlQuery query;
    QByteArray queryarray;
    QBuffer queryBuffer(&queryarray);

    buffer.open(QIODevice::ReadOnly);
    m_QueryError=QString::null;
    query.setMessageHandler(this);
    query.bindVariable("xml",&buffer);
    foreach (QString name, m_QueryBindVariables.keys()) {
        query.bindVariable(name,m_QueryBindVariables[name]);
    }
    query.setQuery(queryString);

    QXmlFormatter formatter(query, &queryBuffer);
    queryBuffer.open(QIODevice::WriteOnly);
    if (!query.evaluateTo(&formatter)) {
        QPair<bool,QByteArray> (false, QByteArray());
    }
    queryBuffer.close();

    if (!m_QueryError.isEmpty()) {
        return QPair<bool,QByteArray>(false,m_QueryError.toUtf8());
    }

    return QPair<bool,QByteArray>(true,queryarray);
}

void SimpleXMLFromPage::query (const QByteArray& inputXML, const QString& queryString)
{
    QPair<bool,QByteArray> result=executeQuery(inputXML, queryString);

    if (m_AutoDelete) {
        deleteLater();
        m_AutoDelete=false;
    }

    emit queryResult(result.first,result.second);
}

void SimpleXMLFromPage::onPageLoaded(bool)
{
    QByteArray out;
    QXmlStreamWriter writer(&out);
    writer.setAutoFormatting(true);
    createSimpleStructure(m_Page->mainFrame()->documentElement(), &writer);

    if (m_Page->parent()==this) {
        m_Page->deleteLater();
        m_Page=0L;
    }

    emit simpleXMLLoaded(out);
    if (!m_Query.isEmpty()) {
        query (out, m_Query);
    }

    if (m_AutoDelete) {
        deleteLater();
    }
}

void SimpleXMLFromPage::handleMessage (QtMsgType, const QString& description, const QUrl&, const QSourceLocation& source)
{
    m_QueryError=description+
                 "["+source.uri().toString()+" Line:"+QString::number(source.line())+"/ Col:"+QString::number(source.column())+"]";
    LogHook::instance()->warning (i18n("Query execution failed: %1",m_QueryError),0l);
}

XQuerySearcher::XQuerySearcher(QObject *parent) :
    QObject(parent),
    m_Model(0l),
    m_ResultIsUrl(false),
    m_WorkInProgress(false)
{
}

void XQuerySearcher::search (const QUrl& specificUrl)
{
    if (!m_WorkInProgress) {
        m_WorkInProgress=true;
        SimpleXMLFromPage* loader = new SimpleXMLFromPage(this);
        connect (loader, SIGNAL(queryResult(bool,QByteArray)), this, SIGNAL(previewQueryResult(bool,QByteArray)));
        connect (loader, SIGNAL(queryResult(bool,QByteArray)), this, SLOT(onPreviewQueryResult(bool,QByteArray)));
        connect (loader, SIGNAL(simpleXMLLoaded(QByteArray)), this, SIGNAL(previewXMLResult(QByteArray)));
        connect (loader, SIGNAL(destroyed()), this, SLOT(loadFinished()));
        connect (loader, SIGNAL(progress(int)), this, SIGNAL(progress(int)));
        loader->setWebPage(m_Page);
        loader->setQuery(m_QueryPreview);
        doAddVariablesToQuery(loader);
        loader->setAutoDelete(true);
        emit startSearch ();
        loader->load(specificUrl);
        LogHook::instance()->message(i18n("Retrieve URL: %1",specificUrl.toString()),0l);
    }
}

void XQuerySearcher::loadFinished()
{
    m_WorkInProgress=false;
}

void XQuerySearcher::onPreviewQueryResult (bool ok, const QByteArray& result)
{
    if (m_Model!=0l && ok) {
        parsePreviewResultXML(result,m_Model);
    }
    emit endSearch();
}

void XQuerySearcher::parsePreviewResultXML (const QByteArray &buf, QAbstractItemModel* model) throw (xml_parse_error)
{
    QBuffer stream;
    stream.setData(buf);
    stream.open(QIODevice::ReadOnly);
    QXmlStreamReader reader(&stream);
    parsePreviewResultXML(&reader,model);
}

void XQuerySearcher::parsePreviewResultXML (QAbstractItemModel* model) throw (xml_parse_error)
{
    parsePreviewResultXML(m_PreviewXQueryResult,model);
}

void XQuerySearcher::setPreviewXQueryResult (bool ok, const QByteArray& result)
{
    if (ok) {
        m_PreviewXQueryResult=result;
    }
}

void XQuerySearcher::startExtraction (const QUrl& url)
{
    SimpleXMLFromPage* loader = new SimpleXMLFromPage(this);
    connect (loader, SIGNAL(queryResult(bool,QByteArray)), this, SIGNAL(extractQueryResult(bool,QByteArray)));
    connect (loader, SIGNAL(simpleXMLLoaded(QByteArray)), this, SIGNAL(extractXMLResult(QByteArray)));
    connect (loader, SIGNAL(progress(int)), this, SIGNAL(progress(int)));
    loader->setWebPage(m_Page);
    loader->setQuery(m_QueryExtraction);
    loader->setAutoDelete(true);
    loader->load(url);
}

void XQuerySearcher::doAddVariablesToQuery (SimpleXMLFromPage*)
{
}

void XQuerySearcher::loadConfig (const QString &filename)
{
    QFile file(filename);
    file.open(QIODevice::ReadOnly);
    QXmlStreamReader reader(&file);

    try {
        bool root=false;
        while (!reader.atEnd()) {
            switch (reader.readNext()) {
            case QXmlStreamReader::StartElement:
                if (reader.name()=="coversearcher") {
                    if (reader.attributes().value("version")!="1.0") {
                        xml_parse_error e;
                        e.line=reader.lineNumber();
                        e.col=reader.columnNumber();
                        e.errormsg=i18n("Version of file is not supported. Only Version 1.0 is valid.");
                        throw e;
                    }
                    root=true;
                }
                else if (root) {
                    if (reader.name()=="title") {
                        while (reader.readNext()!=QXmlStreamReader::Characters);
                        setTitle(reader.text().toString());
                    }
                    else if (reader.name()=="description") {
                        while (reader.readNext()!=QXmlStreamReader::Characters);
                        setDescription(reader.text().toString());
                    }
                    else if (reader.name()=="preview-query") {
                        QString booleanResult=reader.attributes().value("isurl").toString();
                        setResultIsUrl(booleanResult.toLower()=="true");
                        while (reader.readNext()!=QXmlStreamReader::Characters);
                        setQueryPreview(reader.text().toString());
                    }
                    else if (reader.name()=="extraction-query") {
                        while (reader.readNext()!=QXmlStreamReader::Characters);
                        setQueryExtraction(reader.text().toString());
                    }
                    else doReadLocalConfig(reader);
                }
                else {
                    xml_parse_error e;
                    e.line=reader.lineNumber();
                    e.col=reader.columnNumber();
                    e.errormsg=i18n("Unsupported tag element: %1",reader.name().toString());
                    throw e;
                }
                break;
            case QXmlStreamReader::Invalid:
            {
                xml_parse_error e;
                e.line=reader.lineNumber();
                e.col=reader.columnNumber();
                e.errormsg=i18n("Parsing failed. Unknown error. Probably no valid XML.");
                throw e;
            }
            default:
                ; // Ignore
            }
        }
    }
    catch (xml_parse_error e) {
        LogHook::instance()->error(i18n("Could not read file %1. XML Parsing failed: %2 [Line: %3, Column %4]",filename,e.errormsg, e.line, e.col),0l);
    }

    file.close();
}

void XQuerySearcher::saveConfig (const QString &filename)
{
    QFile file(filename);
    file.open(QIODevice::WriteOnly);
    QXmlStreamWriter writer(&file);
    writer.setAutoFormatting(true);

    writer.writeStartDocument();
    writer.writeStartElement("coversearcher");
    writer.writeAttribute("version","1.0");

    writer.writeStartElement("title");
    writer.writeCharacters(title());
    writer.writeEndElement();

    writer.writeStartElement("description");
    writer.writeCharacters(description());
    writer.writeEndElement();

    writer.writeStartElement("preview-query");
    if (resultIsUrl()) {
        writer.writeAttribute("isurl","true");
    }
    writer.writeCDATA(queryPreview());
    writer.writeEndElement();

    writer.writeStartElement("extraction-query");
    writer.writeCDATA(queryExtraction());
    writer.writeEndElement();

    doSaveLocalConfig(writer);

    writer.writeEndDocument();

    file.close();
}

void XQuerySearcher::doSaveLocalConfig (QXmlStreamWriter&)
{
}

void XQuerySearcher::doReadLocalConfig (QXmlStreamReader&)
{
}

}
}
