#ifndef XQUERYSEARCHER_H
#define XQUERYSEARCHER_H

#include <QObject>
#include <QAbstractMessageHandler>
#include <QXmlStreamReader>
#include <QMap>
#include <QPair>

class QWebPage;
class QWebElement;
class QAbstractItemModel;

namespace Texel {

namespace CoverSearch {

class SimpleXMLFromPage: public QAbstractMessageHandler
{
    Q_OBJECT

public:
    SimpleXMLFromPage (QObject* parent);
    virtual ~SimpleXMLFromPage();

    void setQuery (const QString& query);
    void setAutoDelete (bool autoDelete);
    void setWebPage (QWebPage* page);
    void addVar (const QString& name, const QVariant& value);

    virtual void handleMessage (QtMsgType type, const QString& description, const QUrl& identifier, const QSourceLocation& sourceLocation);

    QPair<bool,QByteArray> executeQuery (const QByteArray& inputXML, const QString& queryString);

public slots:
    void load (const QUrl& url);
    void query (const QByteArray& inputXML, const QString& query);

signals:
    void simpleXMLLoaded (const QByteArray& xml);
    void queryResult (bool ok, const QByteArray& result);
    void progress (int i);

private slots:
    void onPageLoaded (bool ok);

private:
    void createSimpleStructure (const QWebElement&, QXmlStreamWriter* writer);

    QWebPage* m_Page;
    QString m_Query;
    QMap<QString, QVariant> m_QueryBindVariables;
    bool m_AutoDelete;

    // Temporary
    QString m_QueryError;
};

class XQuerySearcher : public QObject
{
    Q_OBJECT
public:
    struct xml_parse_error
    {
        QXmlStreamReader::Error errornum;
        QString errormsg;
        int col;
        int line;
    };

    explicit XQuerySearcher(QObject *parent = 0);
    
    void search (const QUrl& specificUrl);

    void startExtraction (const QUrl &url);

    void setWebPage (QWebPage* page);

    void setTitle (const QString& title);
    QString title () const;
    void setDescription (const QString& desc);
    QString description () const;

    void setQueryPreview (const QString& queryPreview);
    QString queryPreview () const;
    void setQueryExtraction (const QString& queryPreview);
    QString queryExtraction () const;

    void setResultIsUrl (bool r);
    bool resultIsUrl () const;

    void parsePreviewResultXML (QAbstractItemModel* model) throw (xml_parse_error);
    void parsePreviewResultXML (const QByteArray &buf, QAbstractItemModel* model) throw (xml_parse_error);

    void setModel (QAbstractItemModel* model);
    QAbstractItemModel* model() const;

    void loadConfig (const QString& filename);
    void saveConfig (const QString& filename);

signals:
    void startSearch ();
    void progress (int p);
    void endSearch ();

    void previewXMLResult (const QByteArray& xml);
    void previewQueryResult (bool ok, const QByteArray& result);

    void extractXMLResult (const QByteArray& xml);
    void extractQueryResult (bool ok, const QByteArray& result);

public slots:
    void setPreviewXQueryResult (bool ok, const QByteArray& result);
    void onPreviewQueryResult (bool ok, const QByteArray& result);

private slots:
    void loadFinished ();

protected:
    virtual void doAddVariablesToQuery (SimpleXMLFromPage* loader);
    virtual void parsePreviewResultXML (QXmlStreamReader* reader, QAbstractItemModel* model) throw (xml_parse_error)=0;

    virtual void doSaveLocalConfig (QXmlStreamWriter& writer);
    virtual void doReadLocalConfig (QXmlStreamReader& reader);

private:
    // General
    QString m_Title;
    QString m_Description;

    QAbstractItemModel* m_Model;

    // Cover retrieving
    bool m_ResultIsUrl;

    QString m_QueryPreview;
    QString m_QueryExtraction;

    bool m_WorkInProgress;
    QWebPage* m_Page;
    QByteArray m_PreviewXQueryResult;

};

inline void SimpleXMLFromPage::setAutoDelete (bool autoDelete)
{
    m_AutoDelete=autoDelete;
}

inline void SimpleXMLFromPage::setWebPage (QWebPage* page)
{
    m_Page=page;
}

inline void SimpleXMLFromPage::setQuery(const QString& query)
{
    m_Query=query;
}

inline void XQuerySearcher::setTitle (const QString& title)
{
    m_Title=title;
}

inline QString XQuerySearcher::title () const
{
    return m_Title;
}

inline void XQuerySearcher::setDescription (const QString& desc)
{
    m_Description=desc;
}

inline QString XQuerySearcher::description () const
{
    return m_Description;
}

inline void XQuerySearcher::setModel (QAbstractItemModel* model)
{
    m_Model=model;
}

inline QAbstractItemModel* XQuerySearcher::model() const
{
    return m_Model;
}

inline QString XQuerySearcher::queryPreview() const
{
    return m_QueryPreview;
}

inline void XQuerySearcher::setQueryPreview (const QString& query)
{
    m_QueryPreview=query;
}

inline QString XQuerySearcher::queryExtraction() const
{
    return m_QueryExtraction;
}

inline void XQuerySearcher::setQueryExtraction (const QString& query)
{
    m_QueryExtraction=query;
}

inline void XQuerySearcher::setWebPage(QWebPage *page)
{
    m_Page=page;
}

inline void XQuerySearcher::setResultIsUrl (bool r)
{
    m_ResultIsUrl=r;
}

inline bool XQuerySearcher::resultIsUrl () const
{
    return m_ResultIsUrl;
}


}
}
#endif // XQUERYSEARCHER_H
