/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

// Necessary to enable progress counter also in QT 4.6
#undef QT_NO_STL

#include <QtConcurrentFilter>
#include <QFutureWatcher>
#include <QList>
#include <QSharedPointer>
#include <QVector>
#include <QTime>
#include <QMutex>

#include <KMessageBox>

#include "duplicatedetection.h"
#include "util/qtlevenshtein.h"
#include "model/tagitem.h"
#include "model/columnmanager.h"
#include "func/textfiltersingleton.h"
#include "ui_duplicatedetection.h"
#include "options/preferences.h"
#include "model/tagmodelproxy.h"

#include "aid.h"

using namespace Texel;
using namespace std;

namespace {

    typedef QPair<int,int> DuplicateIndex;
    typedef QList<DuplicateIndex> DuplicateIndexList;

    /**
  Simplified Tag structure which does only store important attributes for a duplicate
  detection. Provides a difference function between two @class Items.
 */
    struct Item
    {
        Item (const QString& title, const QString& album, const QString& artist)
            : item(title,album,artist)
        {
        }

        int nr;
        LevenstheinDistanceTree::Item item;
    };

    inline ostream& operator<< (ostream& out, const Item& item)
    {
        out << item.nr << ":" << item.item;
        return out;
    }

    /**
  The @class MapFunctor is a functor to map a @class TagItem pointer to a simplified
  @class Item defined specially for duplicate detection. The item does only contain
  title, artist and albm - these are all string which are used for duplicate detection.
  It does also contain a reference back into the tag model list (by refering the linenumber).

  As this class does aslo simplify all copied strings it does maintain a finder element.
  The finder is a performant possibiliy to replace certain patterns at the beginning and within
  the string. As the functor is copied around before its executed the finder pointer is a
  @class QSharedPointer to manage a singe object.
  */
    struct MapFunctor
    {
        typedef Item* result_type;

        MapFunctor (bool useFilter, bool alphaOnly, bool noAlbum, bool swapped)
            : finder(0L),
            flags((alphaOnly?AlphaOnly:0) | (noAlbum?0:Album) | (swapped?Swapped:0))
        {
            if (useFilter) finder=QSharedPointer<CompiledFinder>(TextFilterSingleton::instance()->textFilter()->compile());
        }

        MapFunctor (const MapFunctor& other)
            : finder(other.finder),
            flags(other.flags)
        {
        }

        Item* operator() (TagItem* tag)
        {
            // In case of a header (ignoring it is impossible as just maping is allowed
            // in the current pattern and not reduction
            if (tag==0L) return 0L;

            QString title;
            QString album;
            QString artist;

            if (finder) {
                title=finder->replace(tag->data(ColumnManager::Title,Columns::Column::RoleDisplay));
                if (flags&Album) album=finder->replace(tag->data(ColumnManager::Album,Columns::Column::RoleDisplay));
                artist=finder->replace(tag->data(ColumnManager::Artist,Columns::Column::RoleDisplay));
            }
            else {
                title=tag->data(ColumnManager::Title,Columns::Column::RoleDisplay);
                if (flags&Album) album=tag->data(ColumnManager::Album,Columns::Column::RoleDisplay);
                artist=tag->data(ColumnManager::Artist,Columns::Column::RoleDisplay);
            }

            if (flags&AlphaOnly) {
                title=filterToLetterAndNumber(title);
                if (flags&Album) album=filterToLetterAndNumber(album);
                artist=filterToLetterAndNumber(artist);
            }

            Item* item=new Item(title.simplified(),album.simplified(),artist.simplified());
            item->nr=tag->getRow();

            return item;
        }

        MapFunctor& operator= (const MapFunctor& other)
        {
            finder=other.finder;
            flags=other.flags;
            return *this;
        }

    private:
        enum { Album=1, AlphaOnly=2, Swapped=4 };

        static QString filterToLetterAndNumber (const QString& s)
        {
            QString result;
            for (int i=0;i<s.size();i++) {
                if (s[i].isLetterOrNumber() || s[i].isSpace()) result+=s[i];
            }
            return result;
        }

        QSharedPointer<CompiledFinder> finder;
        int flags;
    };

    void append (DuplicateIndexList& list, const DuplicateIndexList& i)
    {
        list.append(i);
    }

    struct SearchFunctor
    {
        typedef DuplicateIndexList result_type;

        SearchFunctor(const LevenstheinDistanceTree& tree_, int max_, bool swapped_)
            : tree(tree_),
            max(max_),
            swapped(swapped_)
        {
        }

        SearchFunctor (const SearchFunctor& other)
            : tree(other.tree),
            max(other.max),
            swapped(other.swapped)
        {
        }

        result_type operator() (Item* item) const
        {
            DuplicateIndexList result;
            if (item!=0L) {
                const int size = item->item.size();
                // Ignore trivial entries (as two separator chars are used, you'll need more than
                // max characters. Otherwise every equally long string will match)
                if (size>2+max) {
                    {
                        QList<int> intermediate=tree.findMinimalDistanceItems(item->nr, item->item, max, Preferences::instance()->m_DD_maxErrorInARow);
                        foreach (int i, intermediate) result << DuplicateIndex(item->nr, i);
                    }
                    if (swapped) {
                        LevenstheinDistanceTree::Item swappedItem(item->item.artist,item->item.album,item->item.title);
                        QList<int> intermediate=tree.findMinimalDistanceItems(item->nr, swappedItem, max, Preferences::instance()->m_DD_maxErrorInARow);
                        foreach (int i, intermediate) result << DuplicateIndex(item->nr, i);
                    }
                }
            }
            return result;
        }

    private:
        const SearchFunctor& operator= (const SearchFunctor& other);

        const LevenstheinDistanceTree& tree;
        const int max;
        const bool swapped;
    };

}

struct DuplicateDetection::Private
{
    Private(TagModelProxy& model_,Ui::DuplicateDetection* dd_)
        : model(model_),
        dd(dd_),
        mainWidget(new QWidget())
    {
    }

    ~Private ()
    {
        delete mainWidget;
        delete dd;
    }

    TagModelProxy& model;
    Ui::DuplicateDetection* dd;
    QWidget* mainWidget;
    QFutureWatcher<QList<QPair<int,int> > > detectionwatcher;
    QList<Item*> compareAgainstItems;
    QList<Item*> items;
    QList<TagItem*> selection;
};

DuplicateDetection::DuplicateDetection(TagModelProxy* model_, QList<TagItem*> toCheck, QWidget *parent) :
        KDialog(parent),
        p(new Private(*model_,new Ui::DuplicateDetection()))
{
    setCaption (i18n("Duplicate detection"));
    setHelp("duplicatedetection");
    enableLinkedHelp(Preferences::instance()->appearance.m_ShowHelpLinks);
    setWindowIcon(QIcon(KIconLoader::global()->loadIcon("detect-duplicates",KIconLoader::Toolbar)));
    p->dd->setupUi(p->mainWidget);
    setMainWidget(p->mainWidget);

    if (p->model.hasFilter() || !toCheck.isEmpty()) {
        p->dd->compareAgainstAllCheckBox->setEnabled(true);
        p->dd->compareAgainstAllCheckBox->setChecked(true);
        p->selection=toCheck;
    }

    connect (this, SIGNAL(okClicked()), this, SLOT(process()));
    connect (this, SIGNAL(cancelClicked()), this, SLOT(smartCancel()));

    connect (&p->detectionwatcher, SIGNAL(progressRangeChanged(int,int)), p->dd->progressBar, SLOT(setRange(int,int)));
    connect (&p->detectionwatcher, SIGNAL(progressValueChanged(int)), p->dd->progressBar, SLOT(setValue(int)));
    connect (&p->detectionwatcher, SIGNAL(started()), this, SLOT(asyncStateChanged()));
    connect (&p->detectionwatcher, SIGNAL(finished()), this, SLOT(finished()));
}

DuplicateDetection::~DuplicateDetection ()
{
    delete p;
}

void DuplicateDetection::changeEvent(QEvent *e)
{
    KDialog::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        p->dd->retranslateUi(this);
        break;
    default:
        break;
    }
}

void DuplicateDetection::process ()
{
    // When splitting up the UI class and the functionality: these are the only parameters
    // which are retrieved from the UI part.
    const int max = p->dd->exactnessSlider->value();
    const bool filter = p->dd->filterCheckBox->isChecked();
    const bool alpha = p->dd->alphaOnlyCheckBox->isChecked();
    const bool noAlbum = p->dd->noAlbumCheckBox->isChecked();
    const bool swapped = p->dd->swappedCheckBox->isChecked();
    const bool useAllItems = p->dd->compareAgainstAllCheckBox->isChecked();

    enableButtonOk(false);
    // UI Part end

    TagModel* m = dynamic_cast<TagModel*>(p->model.sourceModel());

    // Create searchable list by mapping the current tag set with the mapping functor
    // paralized to simplified item set.
    QFutureWatcher<Item*> watcher;
    connect (&watcher, SIGNAL(progressRangeChanged(int,int)), p->dd->progressBar, SLOT(setRange(int,int)));
    connect (&watcher, SIGNAL(progressValueChanged(int)), p->dd->progressBar, SLOT(setValue(int)));
    p->dd->stepCounter->setText(i18n("Step 1/3: prepare tags"));

    {
        QFuture<Item*> result;
        TagItem::List allItems;

        if (useAllItems) {
            allItems = m->getTagItems();
        }
        else {
            allItems = p->model.getTagItemList();
        }

        result = QtConcurrent::mapped<TagItem::List::const_iterator,MapFunctor>
                (allItems.begin(), allItems.end(), MapFunctor(filter,alpha,noAlbum,false));
        watcher.setFuture(result);

        while (!result.isFinished() && !result.isCanceled()) {
            QCoreApplication::processEvents(QEventLoop::WaitForMoreEvents);
        }

        p->compareAgainstItems = result.results();
        if (useAllItems) {
            if (p->selection.isEmpty()) {
                for (int i=0;i<p->model.rowCount();i++) {
                    p->items << p->compareAgainstItems[p->model.getRow(i)->getRow()];
                }
            }
            else {
                // Take items from selection
                for (int i=0;i<p->selection.size();i++) {
                    p->items << p->compareAgainstItems[p->selection[i]->getRow()];
                }
            }
        }
        else {
            p->items = p->compareAgainstItems;
        }
    }

    // The resulting list is sorted by length of the title. Idea is a performance optimization:
    // As soon as the length difference between to compared items exceeds max comparison can be
    // stopped for all following items as the string size will only increase. Comparison can
    // continue with the next item pairs.
    // not used for parallel comparisons

    p->dd->stepCounter->setText(i18n("Step 2/3: setup search structure"));

    QTime clock;
    clock.start();
    LevenstheinDistanceTree searchtree;
    int counter=0;
    p->dd->progressBar->setMinimum(0);
    p->dd->progressBar->setMaximum(p->compareAgainstItems.size());
    foreach (Item* item, p->compareAgainstItems)
    {
        if (item!=0L && item->item.size()>2+max) {
            searchtree.insert(item->nr, item->item);
        }
        if (((++counter)%500)==0) {
            p->dd->progressBar->setValue(counter);
            QCoreApplication::processEvents(QEventLoop::WaitForMoreEvents);
        }
    }

    p->dd->stepCounter->setText(i18n("Step 3/3: search"));
    QFuture<DuplicateIndexList> result=QtConcurrent::mappedReduced                            // Reduce Functor
    (p->items.begin(),p->items.end(),
     SearchFunctor(searchtree,max,swapped),
     append);
    p->detectionwatcher.setFuture(result);

    // Do regulary call the process events loop. To avoid busy waiting
    // waiting for events is enabled when the clearRelations()'s are all done.
    while (!result.isFinished() && !result.isCanceled()) {
        QCoreApplication::processEvents(QEventLoop::WaitForMoreEvents);
    }

    if (result.isFinished() && !p->detectionwatcher.isCanceled()) {
        for (int i=0;i<p->model.rowCount();++i) {
            if (p->model.getRow(i)!=0L && p->model.getRow(i)->hasRelations()) {
                p->model.getRow(i)->getRelations()->clearDuplicates();
            }
        }

        foreach (DuplicateIndex potentialDuplicate, (DuplicateIndexList)result)
        {
            const int i=potentialDuplicate.first;
            const int j=potentialDuplicate.second;
            m->getRow(i)->getRelations()->addDuplicate(m->getRow(j));
        }

        static_cast<TagModel*>(p->model.sourceModel())->setNewDuplicateList();

        cout << "Used time: " << clock.elapsed()/60/1000 << " ms" << endl;
    }
}

void DuplicateDetection::smartCancel ()
{
    if (!p->detectionwatcher.isRunning()) return;

    // First calculate current progress (in percent)
    const int progress = 100*p->detectionwatcher.progressValue()/
                         (p->detectionwatcher.progressMaximum()-p->detectionwatcher.progressMinimum());
    bool breakup=false;

    if (progress<20) breakup=true;
    else if (progress<80) {
        const int result = KMessageBox::questionYesNo(this,
                                                      i18n("Do you really want to break up?"),
                                                      i18n("Stop duplicate processing"));
        if (result==KMessageBox::Yes) breakup=true;
    }
    else if (progress>=80) {
        const int result = KMessageBox::warningContinueCancel(this,
                                                              i18n("You nearly made it! Do you really want to break up?"),
                                                              i18n("Stop duplicate processing"),
                                                              KStandardGuiItem::cont(),
                                                              KGuiItem(i18nc("Button for stop processing","Stop")));
        if (result==KMessageBox::Cancel) breakup=true;
    }
    if (breakup) {
        p->detectionwatcher.cancel();
        p->detectionwatcher.waitForFinished();
    }
}

void DuplicateDetection::asyncStateChanged ()
{
    enableButtonCancel(p->detectionwatcher.isRunning());
    enableButtonOk(!p->detectionwatcher.isRunning());
}

void DuplicateDetection::finished ()
{
}

#include "duplicatedetection.moc"
