/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "m3ulist.h"

#include "model/tagitem.h"
#include "model/texeltagmodel.h"

#include <QFile>
#include <QDir>
#include <QRegExp>
#include <QSharedData>

#include <kmessagebox.h>
#include <kfiledialog.h>

#include "aid.h"

using namespace std;

namespace Texel {

    namespace {
        QString getCommonPrefix (const QStringList& list)
        {
            QString prefix;

            int minLength=list[0].size();
            for (int i=1;i<list.size();i++) {
                minLength=qMin(minLength,list[i].size());
            }

            for (int i=0;i<minLength;i++) {
                const QChar c = list[0][i];
                for (int j=1;j<list.size();j++) {
                    if (list[j][i]!=c) {
                        return prefix;
                    }
                }
                prefix+=c;
            }

            return prefix;
        }
    }

    namespace M3U {

        struct List::ListData: public QSharedData
        {
            QString m_Filename;
            QString m_LoadedFromFile;
            QList<List::Track> m_Tracks;
            TagModel* m_Model;
            enum Format { Stripped, Extended };
            Format m_Format;
            unsigned int m_AlteredTracks;
            bool m_Dirty;
            bool m_PlaylistInLocalDir;
            bool m_SavedOnDisk;
            quint32 m_UniqId;

            static quint32 createUniqId ();
            static void adjustUniqId (quint32 id);

            static quint32 s_UniqIdSequence;
        };

        quint32 List::ListData::s_UniqIdSequence=1;

        quint32 List::ListData::createUniqId ()
        {
            return ++s_UniqIdSequence;
        }

        void List::ListData::adjustUniqId (quint32 id)
        {
            if (id>s_UniqIdSequence) {
                s_UniqIdSequence=id+1;
            }
        }

        List::List(TagModel* model)
            : p(new ListData())
        {
            p->m_Model=model;
            p->m_AlteredTracks=0;
            p->m_Dirty=false;
            p->m_PlaylistInLocalDir=false;
            p->m_UniqId=ListData::createUniqId();
        }

        List::List (const List& other)
            : p(other.p)
        {
        }

        List::~List ()
        {
        }

        void List::removeFromTags ()
        {
            for (int i=0;i<p->m_Tracks.size();i++) {
                removeFromItem (p->m_Tracks[i].item);
            }
        }

        List::Track List::createTrack (TagItem* item)
        {
            Track track;
            track.item=item;
            track.length=track.item->getIntValue(ColumnManager::Length);
            track.title=track.item->data(ColumnManager::Title,Columns::Column::RoleDisplay);
            return track;
        }

        void List::setTracks (const QList<TagItem*>& tracks)
        {
            p->m_Tracks.clear();
            for (int i=0;i<tracks.size();i++) {
                p->m_Tracks << createTrack(tracks[i]);
            }
            p->m_Dirty=true;
        }

        void List::setTracks (const QList<Track>& tracks)
        {
            p->m_Tracks=tracks;
            for (int i=0;i<tracks.size();i++) {
                p->m_Dirty|=tracks[i].dirty;
            }
        }

        void List::insertTrack (int pos, TagItem *track)
        {
            p->m_Tracks.insert(pos,createTrack(track));
            p->m_Tracks[pos].dirty=true;
            insertIntoItem (track);
            p->m_Dirty=true;
        }

        void List::removeTrack (TagItem* track)
        {
            QList<Track>::iterator iter = p->m_Tracks.begin();
            while (iter!=p->m_Tracks.end()) {
                if ((*iter).item==track) {
                    // May appear several times, therefore continue
                    iter=p->m_Tracks.erase(iter);
                }
                else {
                    ++iter;
                }
            }
            removeFromItem(track);
            p->m_Dirty=true;
        }

        void List::removeTrack (int pos)
        {
            if (pos>=0 && pos<p->m_Tracks.size()) {
                TagItem* toRemove=p->m_Tracks[pos].item;
                p->m_Tracks.removeAt(pos);

                // Search if item is contained a second time in the playlist. If not
                // the playlist is removed from the track
                bool found=false;
                QList<Track>::iterator iter = p->m_Tracks.begin();
                while (iter!=p->m_Tracks.end()) {
                    if ((*iter).item==toRemove) {
                        // May appear several times, therefore continue
                        found=true;
                        break;
                    }
                    ++iter;
                }

                if (!found) {
                    removeFromItem(toRemove);
                }
                p->m_Dirty=true;
            }
        }

        void List::removeFromItem (TagItem* track)
        {
            if (track!=0l && track->hasRelations()) {
                track->getRelations()->removePlaylist(*this);
                if (track->getRelations()->numberOfPlaylists()==0) {
                    track->getModel()->emitChangedHeader (track->getRow());
                }
            }
        }

        void List::insertIntoItem (TagItem* track)
        {
            if (track!=0l) {
                track->getRelations()->addPlaylist(*this,track);
            }
        }

        QString List::getName (int index) const
        {
            if (index>=0 && index<size()) {
                if (p->m_Tracks[index].item==0l){
                    if (p->m_Tracks[index].title.trimmed()=="") {
                        return p->m_Tracks[index].filename;
                    }
                    return p->m_Tracks[index].title;
                }
                return p->m_Tracks[index].item->data (ColumnManager::Title,Columns::Column::RoleDisplay);
            }
            return QString();
        }

        TagItem* List::getItem (int index) const
        {
            if (index>=0 && index<size()) {
                return p->m_Tracks[index].item;
            }
            return 0l;
        }

        QString List::getName () const
        {
            QFileInfo info (p->m_Filename);
            return info.fileName();
        }

        QList<TagItem*> List::getItems () const
        {
            QList<TagItem*> tracks;
            for (int i=0;i<p->m_Tracks.size();i++) {
                if (p->m_Tracks[i].item!=0l) {
                    tracks << p->m_Tracks[i].item;
                }
            }
            return tracks;
        }

        void List::load (const QString& filename)
        {
            QFile file(filename);
            QFileInfo info(filename);

            if (!file.open(QIODevice::ReadOnly)) {
                return;
            }

            const QDir currentWorkingDir = QDir::current();
            const QString workingDir=info.absolutePath();
            QDir::setCurrent(workingDir);

            p->m_Filename=filename;

            load (&file,workingDir,true);

            file.close();
            QDir::setCurrent(currentWorkingDir.absolutePath());
            p->m_Dirty=false;
            p->m_LoadedFromFile=filename;
        }

        void List::load (QIODevice* device, const QString& workingDir, bool expandFileNames)
        {
            bool firstLine=true;
            p->m_Format=ListData::Stripped;
            bool extInfoRead=false;
            QRegExp extinfoLine ("^#EXTINF:?\\s*([0-9]*)\\s*,\\s*(.*)$",Qt::CaseInsensitive);
            p->m_Tracks.clear();

            // Check if all files are in the local directory. If so, mark list. This enables
            // to check uponsave if maybe all files are moved to another directory and move
            // the playlist accordingly.
            p->m_PlaylistInLocalDir=true;

            while (!device->atEnd()) {
                const QString line = QString::fromLocal8Bit(device->readLine()).trimmed();

                // Skip empty lines
                if (line.isEmpty()) {
                    continue;
                }

                if (firstLine==true) {
                    firstLine=false;
                    if (line.startsWith("#EXTM3U",Qt::CaseInsensitive)) {
                        p->m_Format=ListData::Extended;
                        continue;
                    }
                }

                if (p->m_Format==ListData::Stripped) {
                    QFileInfo info(line);
                    Track track;
                    track.item = p->m_Model->getRowByOrgFilename(info.canonicalFilePath());
                    if (track.item!=0l) {
                        track.item->getRelations()->addPlaylist(*this,track.item);
                    }
                    else {
                        if (expandFileNames) {
                            track.filename=info.canonicalFilePath();
                        }
                        else {
                            track.filename=line;
                        }
                    }
                    p->m_PlaylistInLocalDir&=track.getFilename().startsWith(workingDir);
                    p->m_Tracks << track;
                }
                else {
                    if (extinfoLine.indexIn(line)!=-1) {
                        if (extInfoRead==false) {
                            p->m_Tracks << Track();
                        }

                        Track& track=p->m_Tracks.last();
                        bool ok;
                        track.length = extinfoLine.cap(1).toInt(&ok);
                        if (!ok) track.length=-1;
                        track.title = extinfoLine.cap(2);
                        extInfoRead=true;
                    }
                    else {
                        if (extInfoRead==false) {
                            p->m_Tracks << Track();
                        }
                        QString fName = line;
                        if (expandFileNames) {
                            QFileInfo info(line);
                            fName=info.canonicalFilePath();
                        }
                        TagItem* item = p->m_Model->getRowByOrgFilename(fName);
                        p->m_Tracks.last().item=item;
                        if (item!=0l) {
                            item->getRelations()->addPlaylist(*this,item);
                            p->m_PlaylistInLocalDir&=p->m_Tracks.last().item->getAbsolutFilename().startsWith(workingDir);
                        }
                        else {
                            p->m_Tracks.last().filename=fName;
                            p->m_PlaylistInLocalDir&=p->m_Tracks.last().filename.startsWith(workingDir);
                        }
                        extInfoRead=false;
                    }
                }
            }
        }

        void List::save (QIODevice* device, bool enforceAbsolutePathes) const
        {
            QFileInfo info(p->m_Filename);
            const QString root=info.absolutePath();

            if (p->m_Format==ListData::Extended) {
                writeLine (device,"#EXTM3U");
            }

            for (int i=0;i<p->m_Tracks.size();i++) {
                const Track& track = p->m_Tracks[i];

                // Write extended information when format is desired and information is available.
                // The necessary information is available when either a pointer to data is available
                // or both length and title of the loaded list was set.
                if (p->m_Format==ListData::Extended &&
                    (track.item!=0l || (track.length!=-1 && !track.title.isEmpty()))) {
                    const int len = (track.length==-1)?track.item->getIntValue(ColumnManager::Length):track.length;
                    const QString title = (track.title.isEmpty())?(
                            track.item->data(ColumnManager::Artist,Columns::Column::RoleDisplay)+QString(" - ")
                            +track.item->data(ColumnManager::Title,Columns::Column::RoleDisplay)
                            ):track.title;
                    writeLine (device,QString("#EXTM3U:%1,%2").arg(QString::number(len),title).toUtf8());
                }

                QString name=track.getFilename();
                // Make filename relative
                if (!enforceAbsolutePathes && name.startsWith(root)) {
                    name = name.mid(root.size()+1);
                }

                p->m_Tracks[i].dirty=false;
                writeLine (device,name);
            }
        }

        void List::save (const QString& filename)
        {
            if (filename!="") {
                setFileName(filename);
            }

            QFile file(p->m_Filename);

            if (p->m_Tracks.empty()) {
                // Playlist is empty. Offer via dialog to delete the playlist instead of saving it
                // (would be empty anyway)
                int result= KMessageBox::questionYesNoCancel (0l,
                                                              i18n("Playlist %1 is empty. Shall I delete it instead of saving an emptylist?", p->m_Filename),
                                                              i18n("Empty playlist"),
                                                              KStandardGuiItem::yes(),
                                                              KStandardGuiItem::no(),
                                                              KStandardGuiItem::cancel(),
                                                              QString("OnEmptyPlaylist")
                                                              );
                if (result==KMessageBox::Yes) {
                    file.remove();
                }
                else if (result==KMessageBox::Cancel) {
                    return; // On cancel do neither save nor delete
                }
            }

            if (!file.open(QIODevice::WriteOnly|QIODevice::Truncate)) {
                return;
            }

            save (&file,false);
            p->m_Dirty=false;

            // Rename when name was changed
            if (!p->m_LoadedFromFile.isEmpty() && p->m_LoadedFromFile!=p->m_Filename) {
                if (QFile::exists(p->m_LoadedFromFile)) {
                    QFile::remove(p->m_LoadedFromFile);
                }
            }
            p->m_LoadedFromFile=p->m_Filename;
        }

        void List::writeLine (QIODevice* file, const QString& line) const
        {
            file->write(line.toUtf8());
            file->write("\n");
        }

        M3U::List List::createPlaylist (TagModel* model, QIODevice* device, const QString& filename)
        {
            M3U::List playlist(model);
            playlist.setFileName(filename);
            QFileInfo info(filename);
            playlist.load(device,info.absolutePath(),false);
            return playlist;
        }

        M3U::List List::createPlaylist (TagModel* model, quint32 id, const QString& filename, const QString& loadedFilename,
                                        int alteredTracks, bool extended, bool dirty, bool local, bool saved)
        {
            List playlist(model);
            playlist.p->m_Filename=filename;
            playlist.p->m_LoadedFromFile=loadedFilename;
            playlist.p->m_Format=extended?ListData::Extended:ListData::Stripped;
            playlist.p->m_Dirty=dirty;
            playlist.p->m_AlteredTracks=alteredTracks;
            playlist.p->m_PlaylistInLocalDir=local;
            playlist.p->m_SavedOnDisk=saved;
            playlist.p->m_UniqId=id;
            return playlist;
        }

        M3U::List List::createPlaylist (const QList<TagItem*>& tracks)
        {
            M3U::List playlist(tracks.first()->getModel());
            DirTree m3uTree;

            foreach (TagItem* item, tracks) {
                QString file=item->getAbsolutFilename();
                QFileInfo info(file);
                m3uTree.store(info.absolutePath(),0l);
            }
            playlist.setTracks(tracks);

            QDir dir(m3uTree.getCommonPath());
            const QString proposal=dir.dirName()+".m3u";
            const QString prefix=m3uTree.getCommonPath();
            const QString filename = KFileDialog::getSaveFileName(prefix+QDir::separator()+proposal,"*.m3u",0L,i18n("Playlist name"));

            playlist.setFileName(filename);
            return playlist;
        }

        void List::clear()
        {
            p->m_Tracks.clear();
        }

        QString List::Track::getFilename () const
        {
            if (item!=0l) {
                return item->getAbsolutFilename();
            }
            return filename;
        }

        QString List::getFileName (int index) const
        {
            return p->m_Tracks[index].getFilename();
        }

        void List::setFileName (const QString& filename)
        {
            p->m_Filename=filename;
        }

        QString List::getFileName () const
        {
            return p->m_Filename;
        }

        QString List::getLoadFileName () const
        {
            return p->m_LoadedFromFile;
        }

        void List::incAlteredTracks ()
        {
            ++p->m_AlteredTracks;
        }

        void List::decAlteredTracks ()
        {
            if (p->m_AlteredTracks>0) --p->m_AlteredTracks;
        }

        int List::alteredTracks () const
        {
            return p->m_AlteredTracks;
        }

        int List::size () const
        {
            return p->m_Tracks.size();
        }

        bool List::isExtended () const
        {
            return p->m_Format==ListData::Extended;
        }

        bool List::wasSaved () const
        {
            return p->m_SavedOnDisk;
        }

        bool List::isLocal () const
        {
            return p->m_PlaylistInLocalDir;
        }

        bool List::isDirty () const
        {
            return p->m_Dirty || p->m_AlteredTracks>0;
        }

        bool List::isDirty (int pos) const
        {
            return get(pos).dirty;
        }

        List::Track List::get (int index) const
        {
            return p->m_Tracks[index];
        }

        void List::move (int from, int to)
        {
            p->m_Tracks[from].dirty=true;
            if (to<size()) {
                p->m_Tracks.move(from,to);
            }
            else {
                p->m_Tracks.append(p->m_Tracks.takeAt(from));
            }
            p->m_Dirty=true;
        }

        quint32 List::getUniqId () const
        {
            return p->m_UniqId;
        }

        void List::remove (int pos)
        {
            p->m_Tracks.removeAt(pos);
            p->m_Dirty=true;
        }

        void List::set (int pos, const Track& track)
        {
            p->m_Tracks[pos]=track;
            p->m_Tracks[pos].dirty=true;
            p->m_Dirty=true;
        }

        QList<List::Track> List::getTracks () const
        {
            return p->m_Tracks;
        }


        bool List::operator== (const List& other) const
        {
            return p==other.p;
        }

        const M3U::List& List::operator= (const List& other)
        {
            p=other.p;
            return *this;
        }

    }
}
