/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "playlistmodel.h"

#include "model/columnmanager.h"
#include "model/tagitem.h"
#include "model/tagmodelproxy.h"
#include "model/mimesupport.h"

#include "options/preferences.h"

#include <QApplication>
#include <QPalette>
#include <QBuffer>

#include <kurl.h>

#include "aid.h"

using namespace std;

namespace Texel {

    static const char* s_DragMimeTypeItem = "playlistindexes";

    PlaylistModel::PlaylistModel (TagModel* model, QObject *parent) :
            QAbstractItemModel(parent),
            m_Model(model),
            m_UntitledCounter(0)
    {
    }

    QModelIndex PlaylistModel::index(int row, int column, const QModelIndex& parent) const
    {
        if (parent.isValid()) {
            return createIndex(row,column,parent.row());
        }
        return createIndex(row,column,(int)-1);
    }

    QModelIndex PlaylistModel::parent (const QModelIndex& child) const
    {
        if (child.internalId()!=-1) {
            return index(child.internalId(),0,QModelIndex());
        }
        return QModelIndex();
    }

    int PlaylistModel::rowCount (const QModelIndex& parent) const
    {
        if (parent.isValid()) {
            if (parent.internalId()==-1) {
                return m_Playlists[parent.row()].size();
            }
            return 0;
        }
        return m_Playlists.size();
    }

    int PlaylistModel::columnCount (const QModelIndex&) const
    {
        return 1;
    }

    QVariant PlaylistModel::data (const QModelIndex& index, int role) const
    {
        if (role==Qt::DisplayRole) {
            if (index.internalId()==-1) {
                return m_Playlists[index.row()].getFileName();
            }
            else {
                return m_Playlists[index.internalId()].getName(index.row());
            }
        }
        else if (role==Qt::ForegroundRole) {
            if (index.internalId()!=-1 && m_Playlists[index.internalId()].getItem(index.row())==0) {
                return QApplication::palette().color(QPalette::Mid);
            }
            return QApplication::palette().color(QPalette::Text);
        }
        else if (role==Qt::ToolTipRole) {
            QString tip;
            if (index.internalId()==-1) {
                tip=m_Playlists[index.row()].getFileName()+"\n";
                tip+="("+QString::number(m_Playlists[index.row()].size())+" tracks)\n";
                if (m_Playlists[index.row()].isExtended()) {
                    tip+="("+i18n("Extened format")+")";
                }
                return tip;
            }
            else {
                tip+="Playlist title: "+m_Playlists[index.internalId()].getName(index.row());
                if (m_Playlists[index.internalId()].getItem(index.row())!=0l) {
                    tip+="\n"+i18n("Correlated to loaded track");
                }
                return tip;
            }
        }
        else if (role==Qt::DecorationRole) {
            if (index.internalId()==-1) {
                if (!m_Playlists[index.row()].isDirty()) {
                    //return IconProvider::instance()->icon(IconProvider::Check_Green);
                }
                else {
                    return KIconLoader::global()->loadIcon("document-save",KIconLoader::Small);
                }
            }
        }
        else if (role==Qt::FontRole) {
            if (index.internalId()!=-1) {
                if (getPlaylist(index.internalId()).isDirty(index.row())) {
                    QFont f;
                    f.setBold(true);
                    return f;
                }
            }
        }

        return QVariant();
    }

    bool PlaylistModel::setData (const QModelIndex &index, const QVariant &value, int role)
    {
        if (role==Qt::EditRole && index.internalId()==-1) {
            getPlaylist(index.row()).setFileName(value.toString());
            return true;
        }

        return false;
    }

    void PlaylistModel::addPlaylist (M3U::List playlist)
    {
        // Check if playlist is already contained
        for (int i=0;i<m_Playlists.size();i++) {
            if (m_Playlists[i].getFileName()==playlist.getFileName()) {
                const QModelIndex idx=index(i,0);

                const int delta = m_Playlists[i].size()-playlist.size();
                if (delta>0) {
                    beginRemoveRows(idx,0,delta);
                }
                else {
                    beginInsertRows(idx,0,-delta);
                }

                m_Playlists[i].clear();
                m_Playlists[i].setTracks(playlist.getItems());

                if (delta>0) {
                    endRemoveRows();
                }
                else {
                    endInsertRows();
                }
                return;
            }
        }

        beginInsertRows(QModelIndex(),m_Playlists.size(),m_Playlists.size());
        PlaylistGroup::addPlaylist(playlist);
        endInsertRows();
    }

    TagItem::List PlaylistModel::getTracks (QModelIndex idx) const
    {
        if (idx.internalId()==-1) {
            return m_Playlists[idx.row()].getItems();
        }
        return QList<TagItem*>();
    }

    void PlaylistModel::removePlaylist (int pos, bool deleteOnSave)
    {
        if (deleteOnSave) {
            m_PlaylistsToDelete << getPlaylist(pos).getFileName();
        }

        beginRemoveRows(QModelIndex(),pos,pos);
        m_Playlists.removeAt(pos);
        endRemoveRows();
    }

    void PlaylistModel::deleteAll ()
    {
        beginRemoveRows(QModelIndex(),0,rowCount());
        PlaylistGroup::deleteAll();
        endRemoveRows();
    }

    void PlaylistModel::removeTrack (QModelIndex idx)
    {
        QModelIndex parent = idx.parent();
        if (parent.isValid()) {
            beginRemoveRows(parent,idx.row(),idx.row());
            m_Playlists[parent.row()].removeTrack(idx.row());
            endRemoveRows();
        }
    }

    void PlaylistModel::updateChanged()
    {
        foreach (M3U::List playlist, m_Playlists) {
            if (playlist.isDirty()) {
                playlist.save();
            }
        }

        // Delete removed playlists
        foreach (QString file, m_PlaylistsToDelete) {
            if (QFile::exists(file)) {
                QFile::remove(file);
            }
        }
        m_PlaylistsToDelete.clear();
    }

    void PlaylistModel::decDirtyCounter (const PlaylistGroup& playlists)
    {
        foreach (M3U::List playlist, playlists.getPlaylists()) {
            playlist.decAlteredTracks();
            if (!playlist.isDirty()) {
                int pos = m_Playlists.indexOf(playlist);
                if (pos!=-1) {
                    QModelIndex idx = index(pos,0);
                    emit dataChanged(idx,idx);
                }
            }
        }
    }

    void PlaylistModel::incDirtyCounter (const PlaylistGroup& playlists)
    {
        foreach (M3U::List playlist, playlists.getPlaylists()) {
            const bool dirty = playlist.isDirty();
            playlist.incAlteredTracks();
            if (!dirty) {
                int pos = m_Playlists.indexOf(playlist);
                if (pos!=-1) {
                    QModelIndex idx = index(pos,0);
                    emit dataChanged(idx,idx);
                }
            }
        }
    }

    Qt::ItemFlags PlaylistModel::flags (const QModelIndex& idx) const
    {
        Qt::ItemFlags result=Qt::ItemFlags(Qt::ItemIsDragEnabled|Qt::ItemIsDropEnabled|Qt::ItemIsEnabled|Qt::ItemIsSelectable);
        if (idx.internalId()==-1) {
            result|=Qt::ItemIsEditable;
        }
        return result;
    }

    bool PlaylistModel::dropMimeData (const QMimeData* data, Qt::DropAction action, int row, int, const QModelIndex& parent)
    {   
        if (action==Qt::MoveAction || action==Qt::CopyAction) {
            int ret=false;
            int pos=0; // Position where to insert new element(s)
            bool droppedOntoParent = false;
            bool overWriteItem=false; // True, when item shall be overwritten
            bool setFilename=false;
            QModelIndex p=parent;
            QModelIndex changedPlaylist;

            // Check weather drop occured on parent or between rows. QT threads this differently
            if (row==-1) {
                // dropped onto the parent
                droppedOntoParent=true;
                // When dropped onto the parent take the parent of it as pointer
                row=parent.row();
                p=parent.parent();
            }
            // p does not references the playlist or the root element

            // Check type of parent
            if (p.isValid()) {
                // There is another parent? than we have an track
                changedPlaylist=p;
                pos=row;
                if (droppedOntoParent) overWriteItem=true;
            }
            else {
                if (row>=0 && row<m_Playlists.size() && droppedOntoParent) {
                    changedPlaylist=index(row,0);
                    pos=m_Playlists[row].size();
                }
                else {
                    // A new empty playlist must be created and the items inserted there
                    M3U::List playlist=M3U::List(m_Model);
                    setFilename=true;
                    addPlaylist(playlist);

                    pos=0;
                    changedPlaylist=index(m_Playlists.size()-1,0);
                }
            }

            if (MimeSupport::hasSubFormat(data,s_DragMimeTypeItem)) {
                QList<MimeSupport::Index> indexes = MimeSupport::getIndexList(data,s_DragMimeTypeItem);

                int adjustment=0;
                // Run through the indexes and copy all tracks into a single list and remove items from source if necessary
                TagItem::List tracks;
                foreach (MimeSupport::Index idx, indexes) {
                    int playlist=idx.first;
                    int sourceRow=idx.second;

                    if (sourceRow!=-1) {
                        tracks << getPlaylist(playlist).getItem(sourceRow);
                        if (action==Qt::MoveAction || playlist==changedPlaylist.row()) {
                            removeTrack(index(sourceRow,0,index(playlist,0)));
                            if (playlist==changedPlaylist.row() && sourceRow<pos) {
                                ++adjustment;
                            }
                        }
                    }
                    else {
                        tracks << getPlaylist(playlist).getItems();
                        if (action==Qt::MoveAction) {
                            removePlaylist(playlist,true);
                        }
                    }
                }

                // now insert tracks at destination position
                insertTracks (index(pos-adjustment,0,changedPlaylist),tracks);

                ret=true;
            }

            if (!ret && MimeSupport::hasItems(data)) {
                TagItem::List items = MimeSupport::getItemList(data);
                if (overWriteItem) {
                    removeTrack(index(pos,0,changedPlaylist));
                }
                insertTracks (index(pos,0,changedPlaylist),items);

                ret=true;
            }

            if (!ret && data->hasUrls()) {
                QList<QUrl> urls = data->urls();
                QStringList filenames;
                TagItem::List items;
                TagItem::List newItems;

                // Expand urls to filenames
                foreach (QUrl url, urls) {
                    QString filename = url.toLocalFile();
                    QFileInfo finfo(filename);
                    if (finfo.isDir()) {
                        QDir dir(filename);
                        foreach (const QFileInfo& info, dir.entryInfoList())
                        {
                            if (info.fileName()!=QString(".") && info.fileName()!=QString("..")) {
                                if (!info.isDir()) {
                                    filenames << info.canonicalFilePath();
                                }
                            }
                        }
                    }
                    else {
                        filenames << finfo.canonicalFilePath();
                    }
                }

                foreach (QString filename, filenames) {
                    TagItem* item=m_Model->getRowByOrgFilename(filename);
                    if (item==0l) {
                        if (filename.endsWith(".m3u")) {
                            M3U::List playlist(m_Model);
                            playlist.load(filename);
                            addPlaylist(playlist);
                        }
                        else {
                            try {
                                item=TagItem::createTagItem(m_Model,filename);
                                newItems << item;
                                items << item;
                            }
                            catch (TagItem::InvalidFormat e) {
                                // ignore entry
                            }
                        }
                    }
                    else {
                        items << item;
                    }
                }

                if (items.isEmpty()) {
                    return false;
                }

                m_Model->appendList(newItems);
                insertTracks (index(pos,0,changedPlaylist),items);
                ret=true;
            }

            // if necessary set filename
            if (setFilename) {
                M3U::List* playlist=&m_Playlists[changedPlaylist.row()];
                playlist->setFileName(playlist->getItem(0)->getPathPointer(Columns::Column::RoleDisplay)->getName()+QDir::separator()+
                                      Preferences::instance()->filelocs.m_DefaultPlaylistName+
                                      QString::number(m_UntitledCounter++));
            }
            return ret;
        }

        return false;
    }

    void PlaylistModel::insertTracks (const QModelIndex& idx, TagItem::List items)
    {
        if (idx.isValid() && idx.parent().isValid()) {
            M3U::List* playlist=&m_Playlists[idx.parent().row()];
            int pos = idx.row();
            // Add Indexes in desired playlist / position
            beginInsertRows(idx.parent(),pos,pos+items.size()-1);
            foreach (TagItem* item, items) {
                playlist->insertTrack(pos++,item);
            }
            endInsertRows();
            // emit changed playlist as here the "unsaved" symbol must appear
            emit dataChanged (idx.parent(),idx.parent());
        }
    }

    namespace
    {
        struct LessFunctor
        {
            bool operator() (const QModelIndex& i1, const QModelIndex& i2) const
            {
                if ((!i1.parent().isValid() && !i2.parent().isValid()) ||
                    i1.parent()==i2.parent()) {
                    return i1.row()>i2.row();
                }

                if (i1.parent().isValid() && i2.parent().isValid()) {
                    return i1.parent().row()>i2.parent().row();
                }

                return i1.parent().isValid();
            }
        };
    }

    void PlaylistModel::sortIndexList (QModelIndexList& list)
    {
        qSort (list.begin(),list.end(),LessFunctor());
    }

    QMimeData* PlaylistModel::mimeData (const QModelIndexList& unsortedIndexes) const
    {
        QModelIndexList indexes=unsortedIndexes;

        // Sort indexes in reverse order - this allows the moves (namely: deletions) without disturbing
        // the indexes as deletins starts with the last deleted row in the list
        sortIndexList (indexes);

        MimeSupport* mime = new MimeSupport();
        TagItem::List items;
        QList<MimeSupport::Index> indexList;

        foreach (QModelIndex i,indexes) {
            if (i.parent().isValid()) {
                items << m_Playlists[i.parent().row()].getItem(i.row());
                indexList << MimeSupport::Index(i.parent().row(), i.row());
            }
            else {
                items << m_Playlists[i.row()].getItems();
                indexList << MimeSupport::Index(i.row(), -1);
            }
        }

        mime->setUrls(items);
        mime->setItems(items);
        mime->setIndexList(indexList,s_DragMimeTypeItem);

        if (indexes.size()==1 && !indexes.first().parent().isValid()) {
            // Single playlist. Create and export it
            QBuffer buffer;
            buffer.open(QIODevice::WriteOnly);
            getPlaylist(indexes.first().row()).save(&buffer,true);
            buffer.close();
            mime->setText(buffer.data().data());
            mime->setData("audio/x-mpegurl",buffer.data());
        }

        return mime;
    }

    Qt::DropActions PlaylistModel::supportedDropActions() const
    {
        return Qt::DropActions(Qt::MoveAction|Qt::CopyAction);
    }

    QStringList PlaylistModel::mimeTypes () const
    {
        QStringList l;
        l << MimeSupport::getMimeTypeForSubType(s_DragMimeTypeItem)
          << MimeSupport::getMimeTypeForSubType(MimeSupport::s_TagItemSubFormat)
          << "text/uri-list";
        return l;
    }

    void PlaylistModel::removeRow (int row)
    {
        beginRemoveRows (QModelIndex(), row,row);
        m_Playlists.removeAt(row);
        endRemoveRows();
    }

 }
