/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "checkserviceartist.h"

#include "aid.h"
#include "util/persistentcache.h"

#include <QTimer>

namespace Texel
{

namespace CheckService {

CheckArtist::CheckArtist(ArtistQuery* query)
    : QThread(query),
      m_Query(query)
{
    m_Query->setParent(0l);
    m_Query->moveToThread(this);
}

CheckArtist::~CheckArtist ()
{
    exit();
    wait();
    m_Query->deleteLater();
}

void CheckArtist::attach (QObject* obj, const char* lookupSignal, const char* resultSlot)
{
    QObject::connect (obj, lookupSignal, m_Query, SLOT(lookup (QVariant, QString)),Qt::QueuedConnection);
    QObject::connect (m_Query, SIGNAL(result(QVariant,QStringList)), obj, resultSlot,Qt::QueuedConnection);
}

void CheckArtist::run ()
{
    QTimer::singleShot(10,m_Query,SLOT(execLookup()));
    exec();
}

ArtistQuery::ArtistQuery (QObject* parent)
    : QObject(parent)
{
}

ArtistQuery::~ArtistQuery ()
{
}

// Does just queue the lookup request (its safe as its in the same thread)
void ArtistQuery::lookup(const QVariant &ref, const QString &artist)
{
    m_Queue << QPair<QVariant,QString>(ref,artist);
}

void ArtistQuery::execLookup ()
{
    if (m_Queue.isEmpty()) {
        QTimer::singleShot(500,this,SLOT(execLookup()));
        return;
    }

    const QString& artist = m_Queue.first().second;
    const QVariant& ref = m_Queue.first().first;

    bool loaded=false;
    QStringList _result;
    try {
        QByteArray content = PersistentCache::instance()->lookup(1,artist);
        if (!content.isEmpty()) {
            QDataStream stream(&content,QIODevice::ReadOnly);
            stream >> _result;
            loaded=true;
        }
    }
    catch (SqLite::Exception e) {
        std::cout << "Cache access error (lookup): " << e.msg << std::endl;
        std::cout << e.details << std::endl;
    }

    if (!loaded) {
        _result=doLookup(artist);

        if (!_result.isEmpty()) {
            try {
                QByteArray content;
                QDataStream stream(&content,QIODevice::WriteOnly);
                stream << _result;
                PersistentCache::instance()->insert(1,artist,content);
            }
            catch (SqLite::Exception e) {
                std::cout << "Cache access error (insert): " << e.msg << std::endl;
                std::cout << e.details << std::endl;
            }
        }
    }

    emit result (ref,_result);

    m_Queue.takeFirst();
    // Restart immediately after return to the event loop
    QTimer::singleShot(0,this,SLOT(execLookup()));
}

QPair<QStringList,QStringList> CheckArtist::getCacheInformation () const
{
    if (m_Query!=0l) {
        return m_Query->getCacheInformation();
    }
    return QPair<QStringList,QStringList>();
}

QPair<QStringList,QStringList> ArtistQuery::getCacheInformation () const
{
    //return m_Cache->getCacheInformation();
    return QPair<QStringList,QStringList>();
}

}

}

#include "checkserviceartist.moc"
