/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "cdentry.h"
#include "model/columnmanager.h"

#include <kglobal.h>

#include "aid.h"

using namespace std;

namespace Texel {

    QList<QString> CDReference::categories;
    bool CDReference::_init=false;

    void CDReference::init ()
    {
        if (!_init) {
            categories << "" << "blues" << "classical" << "data" << "country"
                    << "folk" << "jazz" << "newage" << "reggae"
                    << "rock" << "soundtrack" << "misc";
            _init=true;
        }
    }

    const QString& CDEntry::Track::col (int column) const
    {
        switch (column) {
        case ColumnManager::Artist: return artist;
        case ColumnManager::Title: return title;
        }
        return KGlobal::staticQString("");
    }

    CDReference::CDReference (const QString& g, quint32 id, const QString& t)
        :discid(id),title(t)
    {
        init();
        const int pos = categories.indexOf(g);
        genre=(pos==-1)?0:pos;
    }

    const QString& CDReference::getGenre () const
    {
        init();
        return categories[genre];
    }

    CDEntry* CDEntry::parse (QIODevice *buffer)
    {
        CDEntry* cd = new CDEntry();
        QString artist;
        bool inParsingOffsets=false;
        int cumulated=0;

        while (!buffer->atEnd()) {
            QString line = QString::fromUtf8(buffer->readLine()).simplified();
            if (line.isEmpty()) continue;

            if (line[0]!='#') {
                QStringList parts=line.split('=');
                if (parts[0]=="DTITLE") {
                    artist=parts[1].section("/",0,0).trimmed();
                    cd->title=parts[1].section("/",1).trimmed();
                }
                else if (parts[0]=="DYEAR") {
                    cd->year=parts[1].toInt();
                }
                else if (parts[0]=="DGENRE") {
                    cd->genre=parts[1].trimmed();
                }
                else if (parts[0].startsWith("TTITLE")) {
                    const int num=parts[0].right(parts[0].length()-6).toInt();
                    Q_ASSERT(num>=0 && num<cd->tracks.size());
                    cd->tracks[num].title=parts[1];
                }
            }
            else {
                if (line.contains("Track frame offsets:")) {
                    inParsingOffsets=true;
                    cumulated=0;
                }
                else if (inParsingOffsets) {
                    line[0]=' ';
                    line=line.simplified();
                    if (line.contains("Disc length")) {
                        QString sSeconds = line;
                        sSeconds.remove("Disc length:").remove("seconds");
                        cd->tracks << CDEntry::Track((sSeconds.toLong()*75-cumulated)/75);
                        inParsingOffsets=false;
                    }
                    else if (!line.isEmpty()) {
                        bool success;
                        const int offset=line.toInt(&success,10);
                        if (success && offset!=0) {
                            if (cumulated>0) {
                                cd->tracks << CDEntry::Track((offset-cumulated)/75);
                            }
                            cumulated=offset;
                        }
                    }
                }
            }
        }

        // Check if a sampler was found
        QStringList samplerseparators;
        samplerseparators << " / " << " - " << "/" << "-" << "_";
        bool found=true;
        foreach (QString separator, samplerseparators) {
            for (int i=0;i<cd->tracks.size();i++) {
                if (cd->tracks[i].title.count(separator)!=1) {
                    found=false;
                    break;
                }
            }
            if (found) {
                for (int i=0;i<cd->tracks.size();i++) {
                    QStringList parts = cd->tracks[i].title.split(separator);
                    cd->tracks[i].title = parts[1];
                    cd->tracks[i].artist= parts[0];
                }
                cd->sampler=true;
                break;
            }
        }

        if (!found) {
            for (int i=0;i<cd->tracks.size();i++) {
                cd->tracks[i].artist=artist;
            }
        }

        return cd;
    }

    void CDEntry::mapToItem (TagItem *const item, int track, bool unsetOnly)
    {
        if (track<=0 || track>tracks.size()) {
            return;
        }

        if (!unsetOnly || item->data(ColumnManager::Title).isEmpty()) {
            item->setData(ColumnManager::Title,tracks[track-1].title);
        }
        if (!unsetOnly || item->data(ColumnManager::Artist).isEmpty()) {
            item->setData(ColumnManager::Artist,tracks[track-1].artist);
        }
        if (!unsetOnly || item->data(ColumnManager::Track).isEmpty()) {
            item->setData(ColumnManager::Track,track);
        }
        if (!unsetOnly || item->data(ColumnManager::Album).isEmpty()) {
            item->setData(ColumnManager::Album,title);
        }
        if (!unsetOnly || item->data(ColumnManager::Genre).isEmpty()) {
            item->setData(ColumnManager::Genre,genre);
        }
    }

    void CDEntry::mapContent (const TagItem::List& items, bool unsetOnly)
    {
        int enumeratedTrack=0;

        foreach (TagItem* item, items) {
            enumeratedTrack++;
            if (item->hasColumn(ColumnManager::Track)) {
                const int track=item->getIntValue(ColumnManager::Track);
                if (track!=0) {
                    mapToItem(item,track,unsetOnly);
                    continue;
                }
            }

            if (item->data(ColumnManager::Title)!="")
            {
                QRegExp extractNumber("\\D*(\\d+)\\D*");
                extractNumber.indexIn(item->data(ColumnManager::Title));
                const int track=extractNumber.cap().toInt();
                if (track!=0) {
                    mapToItem(item,track,unsetOnly);
                    continue;
                }
            }

            // Last resort: Map by order
            mapToItem(item,enumeratedTrack,unsetOnly);
        }
    }
}

QDataStream& operator<< (QDataStream& stream, const Texel::CDEntry& ref)
{
    stream << ref.title << ref.genre << ref.year << ref.tracks << ref.sampler;
    return stream;
}

QDataStream& operator<< (QDataStream& stream, const Texel::CDEntry::Track& track)
{
    stream << track.title << track.artist << track.runtime;
    return stream;
}

QDataStream& operator<< (QDataStream& stream, const Texel::CDReference& ref)
{
    stream << ref.genre << ref.discid << ref.title;
    return stream;
}

QDataStream& operator>> (QDataStream& stream, Texel::CDEntry& ref)
{
    stream >> ref.title >> ref.genre >> ref.year >> ref.tracks >> ref.sampler;
    return stream;
}

QDataStream& operator>> (QDataStream& stream, Texel::CDEntry::Track& track)
{
    stream >> track.title >> track.artist >> track.runtime;
    return stream;
}

QDataStream& operator>> (QDataStream& stream, Texel::CDReference& ref)
{
    stream >> ref.genre >> ref.discid >> ref.title;
    return stream;
}
