/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "detailsmenu.h"
#include "querymenu.h"
#include "model/tagmodelproxy.h"

#include <QWidgetAction>
#include <QLabel>
#include <QProgressBar>
#include <QVBoxLayout>
#include <QApplication>
#include <QTextDocument>
#include <QTextCursor>
#include <QTextTable>
#include <QTextTableCell>
#include <QTextBlockFormat>
#include <QTextDocumentFragment>
#include <QTextBrowser>

#include <klocale.h>
#include <kdebug.h>

#include "aid.h"

using namespace std;

namespace Texel {

    namespace FreeDB {

        DetailsMenu::DetailsMenu(FreeDBAccess* access, const CDReference& ref, QueryMenu* parent)
            : QMenu(parent),
            m_Accessor(access),
            m_Reference(ref),
            status(0)
        {
            connect (this, SIGNAL(aboutToShow()), this, SLOT(onShow()));
            setTitle (ref.title);
        }

        void DetailsMenu::onShow ()
        {
            if (status==0) {
                connect (m_Accessor, SIGNAL(detailsMatch(CDEntry)), this, SLOT(onResult(CDEntry)));
                connect (m_Accessor, SIGNAL(detailsError(QString,QString)), this, SLOT(onError(QString)));
                m_Accessor->readDiscContent (m_Reference);

                //
                //  Install a widget with a progressbar and a describing text
                //
                if (status==0) {
                    QWidgetAction* progress = new QWidgetAction(this);
                    QWidget* container = new QWidget(this);
                    delete container->layout();
                    container->setLayout(new QVBoxLayout(container));
                    QLabel* label = new QLabel(container);
                    label->setText("Wait for lookup");
                    container->layout()->addWidget(label);
                    QProgressBar* bar = new QProgressBar(container);
                    bar->setRange(0,0);
                    container->layout()->addWidget(bar);
                    progress->setDefaultWidget(container);
                    addAction(progress);
                    status=1;
                }
            }
        }

        void DetailsMenu::onError(const QString& description)
        {
            if (status==1) {
                status=2;
                clear();
                QWidgetAction* error = new QWidgetAction(this);
                QLabel* message = new QLabel(this);
                message->setMargin(3);
                message->setWordWrap(true);
                message->setText (description);
                error->setDefaultWidget(message);
                addAction(error);
            }
            disconnect (m_Accessor, SIGNAL(detailsMatch(CDEntry)), this, SLOT(onResult(CDEntry)));
            disconnect (m_Accessor, SIGNAL(detailsError(QString,QString)), this, SLOT(onError(QString)));
        }

        void DetailsMenu::onResult (const CDEntry& entry)
        {
            QueryMenu* qm = dynamic_cast<QueryMenu*>(parent());
            clear();

            QAction* takeAllAction = addAction (i18n("Take all data"),this, SLOT(takeAll()));
            QAction* unfilledAction= addAction (i18n("Take just unfill fields"), this, SLOT(takeUnfilled()));
            if (!qm->isSorted()) {
                addAction (i18n("Take track numbers only"), this, SLOT(takeTrackNumbers()));
                connect (takeAllAction, SIGNAL(triggered()), this, SLOT(takeTrackNumbers()));
                connect (unfilledAction, SIGNAL(triggered()), this, SLOT(takeTrackNumbers()));
            }

            m_MappedRows = qm->reorderByCDEntry(entry);

            QWidgetAction* cd = new QWidgetAction(this);
            QTextBrowser* label=renderTable(entry);
            if (label!=0l) {
                label->setAutoFillBackground(true);
                label->setBackgroundRole(QPalette::Light);
                connect (label, SIGNAL(anchorClicked(QUrl)), this, SLOT(takeFromLink(QUrl)));
                connect (label, SIGNAL(highlighted(QUrl)), this, SLOT(setToolTip(QUrl)));
            }
            cd->setDefaultWidget(label);
            addAction(cd);

            m_Entry=entry;
            status=2;

            disconnect (m_Accessor, SIGNAL(detailsMatch(CDEntry)), this, SLOT(onResult(CDEntry)));
            disconnect (m_Accessor, SIGNAL(detailsError(QString,QString)), this, SLOT(onError(QString)));
        }

        void DetailsMenu::insertCell (QTextTable* table, const QTextBlockFormat& blockformat, const QString& text, bool asLink, int row, int col, int kind)
        {
            QTextCursor cursor=table->cellAt(row,col).firstCursorPosition();
            cursor.setBlockFormat(blockformat);

            if (asLink) {
                const QUrl link = createUrl (kind, row);
                cursor.insertHtml("<a href=\""+link.toString()+"\">"+text+"</a>");
            }
            else {
                cursor.insertText(text);
            }
        }

        QTextBrowser* DetailsMenu::renderTable (const CDEntry& entry)
        {
            QueryMenu* qm = dynamic_cast<QueryMenu*>(parent());
            TagModelProxy* _m = qm->model();
            TagModel* model = dynamic_cast<TagModel*>(_m->sourceModel());

            if (m_MappedRows.size()!=entry.tracks.size()) {
                kDebug() << "Requested " << m_MappedRows.size() << ", but found " << entry.tracks.size() << "!" << endl;
                return 0L;
            }

            QTextDocument* doc=new QTextDocument();
            doc->setIndentWidth(10.0);

            QTextCursor cursor(doc);
            QTextCharFormat format;
            format.setFont(QApplication::font());
            format.setFontPointSize(10.0);
            cursor.setCharFormat(format);

            QTextTableFormat tableFormat;
            tableFormat.setCellSpacing(0.0);
            tableFormat.setCellPadding(2.0);
            tableFormat.setMargin(0.0);
            tableFormat.setBorderStyle(QTextFrameFormat::BorderStyle_None);
            cursor.insertTable(entry.tracks.size()+1,2,tableFormat);
            m_Table = cursor.currentTable();
            QTextCharFormat cellformat=m_Table->cellAt(0,0).format();
            m_Table->mergeCells(0,0,1,2);

            //
            // Create header (album title)
            //
            QTextTableCell header=m_Table->cellAt(0,0);
            QTextCursor c=header.firstCursorPosition();
            QTextBlockFormat blockFormat=c.blockFormat();
            QTextBlockFormat cellFormat=c.blockFormat();
            blockFormat.setAlignment (Qt::AlignCenter);
            cellFormat.setAlignment(Qt::AlignVCenter);
            c.setBlockFormat(blockFormat);
            QTextCharFormat  headerFormat = cellformat;
            headerFormat.setFontPointSize(10.0);
            headerFormat.setFontItalic(true);
            header.setFormat(headerFormat);
            bool asLink = false;
            for (int i=0;i<m_MappedRows.size();i++) {
                asLink|=model->data(model->index(m_MappedRows[i]->getRow(),ColumnManager::Album)).toString()!=entry.title;
            }
            insertCell (m_Table, blockFormat, entry.title, asLink, 0, 0, ColumnManager::Album);

            //
            // Create table entries (title and artist)
            //
            for (int i=0;i<entry.tracks.size();i++) {
                asLink = model->data(model->index(m_MappedRows[i]->getRow(),ColumnManager::Title)).toString()!=entry.tracks[i].title;
                insertCell (m_Table, cellFormat, entry.tracks[i].title, asLink, i+1, 0, ColumnManager::Title);

                asLink = model->data(model->index(m_MappedRows[i]->getRow(),ColumnManager::Artist)).toString()!=entry.tracks[i].artist;
                insertCell (m_Table, cellFormat, entry.tracks[i].artist, asLink, i+1, 1, ColumnManager::Artist);
            }

            QTextBrowser* label=new QTextBrowser();
            label->setDocument (doc);
            label->setReadOnly(true);
            label->setOpenLinks(false);
            label->setOpenExternalLinks(false);
            return label;
        }

        void DetailsMenu::takeAll ()
        {
            for (int i=0;i<m_MappedRows.size();i++) {
                m_MappedRows[i]->setData(ColumnManager::Artist,QVariant(m_Entry.tracks[i].artist));
                m_MappedRows[i]->setData(ColumnManager::Title ,QVariant(m_Entry.tracks[i].title));
                m_MappedRows[i]->setData(ColumnManager::Year  ,QVariant(m_Entry.year));
                m_MappedRows[i]->setData(ColumnManager::Genre ,QVariant(m_Entry.genre));
            }
        }

        void DetailsMenu::takeUnfilled ()
        {
            QueryMenu* qm = dynamic_cast<QueryMenu*>(parent());
            TagItem::List rows = qm->getSelectedRows();

            for (int i=0;i<rows.size();i++) {
                if (rows[i]->data(ColumnManager::Artist).isEmpty()) {
                    rows[i]->setData(ColumnManager::Artist,m_Entry.tracks[i].artist);
                }
                if (rows[i]->data(ColumnManager::Title).isEmpty()) {
                    rows[i]->setData(ColumnManager::Title,m_Entry.tracks[i].title);
                }
                if (rows[i]->data(ColumnManager::Year).isEmpty()) {
                    rows[i]->setData(ColumnManager::Year,m_Entry.year);
                }
                if (rows[i]->data(ColumnManager::Genre).isEmpty()) {
                    rows[i]->setData(ColumnManager::Genre,m_Entry.genre);
                }
            }
        }

        QUrl DetailsMenu::createUrl (int col, int x)
        {
            QUrl link("txlref:set");
            QList<QPair<QString, QString> > values;
            values << QPair<QString,QString>("col",ColumnManager::instance()->getName(col));
            values << QPair<QString,QString>("x",QString::number(x));
            link.setQueryItems(values);
            return link;
        }

        void DetailsMenu::parseUrl (const QUrl& url, int& col, int& x)
        {
            const QList<QPair<QString,QString> > items = url.queryItems();
            for (int i=0;i<items.size();i++) {
                if (items[i].first=="col") {
                    col = ColumnManager::instance()->indexByName(items[i].second);
                }
                else if (items[i].first=="x") {
                    x= items[i].second.toInt();
                }
            }
        }

        void DetailsMenu::takeFromLink (const QUrl& link)
        {
            int column=0;
            int x=0;
            parseUrl (link, column, x);

            QTextBrowser* browser = dynamic_cast<QTextBrowser*>(sender());
            if (column==ColumnManager::Artist && !m_Entry.sampler) {
                // When its not a sampler and the artist was selected update all rows as artist
                // should be the same. Please note that the artist flag is self determined and
                // set always when more than one artist did the album.
                for (int i=0;i<m_Entry.tracks.size();i++) {
                    updateCell (browser, m_Entry.tracks[i].artist, m_MappedRows[i], ColumnManager::Artist, i);
                }
            }
            else if (column==ColumnManager::Album) {
                for (int i=0;i<m_MappedRows.size();i++) {
                    updateCell (browser, m_Entry.title, m_MappedRows[i], column, 0);
               }
            }
            else {
                updateCell (browser, m_Entry.tracks[x-1].col(column), m_MappedRows[x-1], column, x);
            }
        }

        void DetailsMenu::updateCell (QTextBrowser* browser, const QString& value, TagItem* item, int col, int x)
        {
            item->setData(col,QVariant(value));

            if (browser!=0L) {
                QTextTableCell cell = m_Table->cellAt(x,(col==ColumnManager::Artist)?1:0);
                QTextCursor cursor(browser->document());
                cursor.setPosition(cell.firstCursorPosition().position());
                cursor.setPosition(cell.lastCursorPosition().position(),QTextCursor::KeepAnchor);
                cursor.removeSelectedText();
                cell.firstCursorPosition().insertText(value);
            }
        }

        void DetailsMenu::setToolTip (const QUrl& link)
        {
            if (link.isEmpty()) {
                dynamic_cast<QWidget*>(sender())->setToolTip("");
                return;
            }

            int column=0;
            int x=0;
            parseUrl (link, column, x);

            QString tagValue,suggestionValue;

            if (column==ColumnManager::Album) {
                tagValue=m_MappedRows[0]->data(ColumnManager::Album);
                suggestionValue=m_Entry.title;
            }
            else {
                tagValue=m_MappedRows[x-1]->data(column);
                suggestionValue = m_Entry.tracks[x-1].col(column);
            }
            QString tooltip=i18n("Tag:\t")+tagValue+"\n";
            tooltip+=i18n("FreeDB:\t")+suggestionValue;

            dynamic_cast<QWidget*>(sender())->setToolTip(tooltip);
        }

        void DetailsMenu::takeTrackNumbers ()
        {
            QueryMenu* qm = dynamic_cast<QueryMenu*>(parent());
            TagModelProxy* _m = qm->model();
            TagModel* model = dynamic_cast<TagModel*>(_m->sourceModel());

            TagItem::List rows = qm->getSelectedRows();
            for (int i=0;i<rows.size();i++) {
                const QModelIndex index = model->index(rows[i]->getRow(),ColumnManager::Track);
                model->setData(index,QVariant(i+1),Qt::EditRole);
            }
        }
    }
}

#include "detailsmenu.moc"
