/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    $Id$

*************************************************************************/

#include "freedbaccess.h"
#include "indexline.h"
#include "aid.h"
#include "options/preferences.h"
#include "model/columnmanager.h"
#include "util/persistentcache.h"
#include "util/loghook.h"
#include "texelversion.h"

#include <QFile>
#include <QFileInfo>
#include <QBuffer>
#include <QHttp>
#include <QUrl>
#include <QApplication>

#include <KUser>
#include <klocale.h>
#include <kio/jobclasses.h>
#include <kio/job.h>

#include <iostream>
#include <unistd.h>

using namespace std;

namespace {

    int cddb_sum(int n)
    {
        int	ret=0;

        while (n > 0) {
            ret += (n % 10);
            n = n / 10;
        }

        return ret;
    }

    quint32 cddb_discid (const QList<int>& cdtoc)
    {
        int	i=0, n = 0;

        const int tot_trks = cdtoc.size()-1;

        while (i < tot_trks) {
            n += cddb_sum(cdtoc[i]);
            i++;
        }

        const int t = cdtoc[tot_trks] - cdtoc[0];
        return ((n % 0xff) << 24 | t << 8 | tot_trks);
    }
}

namespace Texel {

    namespace FreeDB {

        const QChar FreeDBAccess::s_CddbSeparator = QChar(' ');
        bool FreeDBAccess::s_RegisteredMetaType = false;

        /*

        This functionality is not taken from libkcddb as the library required the discid
        to be calculated from the toc. As the toc is unknown the library retrieves a fitting
        discid from the index. This is than used for further lookup.

      */

        FreeDBAccess::FreeDBAccess()
            :  m_QueryCounter(0)
        {
            if (s_RegisteredMetaType==false) {
                qRegisterMetaType<CDReference::List>("FreeDBAccess::CDReference::List");
                s_RegisteredMetaType=true;
            }
        }

        FreeDBAccess::~FreeDBAccess()
        {
        }

        bool FreeDBAccess::existIndex()
        {
            const QString& name=Preferences::instance()->freedb.m_IndexFile;
            const QFileInfo info(name);
            if (info.exists() && info.isReadable() && info.size()>0) {
                return true;
            }
            return false;
        }

        QString FreeDBAccess::hello () const
        {
            KUser user;
            char buffer[256];
            QString name;
            ::gethostname(buffer,255);
            name+=buffer;
            if (::getdomainname(buffer,255)==0) {
                if (*buffer!=0) name+=QString(".")+buffer;
            }
            return user.loginName()+" "+name+" "+QApplication::applicationName()+" "+TEXEL_VERSION;
        }

        QString FreeDBAccess::proto () const
        {
            return "6";
        }

        QString FreeDBAccess::queryString (quint32 trackid, const QList<int>& toc) const
        {
            QString query="cddb query "+QString::number(trackid,16)+" "+QString::number(toc.size())+" ";
            int sum=150;
            for (int i=0;i<toc.size();i++) {
                query+=QString::number(sum)+" ";
                sum+=75*toc[i];
            }
            query+=QString::number(quint16((trackid>>8)&0xffff));
            return query;
        }

        QString FreeDBAccess::readString (const CDReference& reference) const
        {
            return QString("cddb read ")+reference.getGenre()+" "+QString::number(reference.discid,16);
        }

        KJob* FreeDBAccess::executeHttpCommand (const QString& cmd)
        {
            KUrl url(Preferences::instance()->freedb.m_Url);
            url.setPath("/~cddb/cddb.cgi/");
            url.addQueryItem("cmd",cmd);
            url.addQueryItem("hello",hello());
            url.addQueryItem("proto",proto());
            KIO::Job* job=KIO::storedGet(url,KIO::NoReload,KIO::HideProgressInfo);
            job->setAutoDelete(true);
            return job;
        }

        void FreeDBAccess::onReadDiscContentResponse (KJob* job)
        {
            if (job->error()!=0) {
                emit detailsError("Could not read the requested disc entry from freedb: "+job->errorText(),"");
                return;
            }

            QBuffer buffer;
            buffer.setData(dynamic_cast<KIO::StoredTransferJob*>(job)->data());
            buffer.open(QIODevice::ReadOnly);

            const QString firstLine= QString::fromUtf8(buffer.readLine());
            const int code = firstLine.section(s_CddbSeparator,0,0,QString::SectionSkipEmpty).toInt();

            if (code==210) {
                CDEntry* cd = CDEntry::parse(&buffer);
                emit detailsMatch(*cd);
                try {
                    QByteArray content;
                    QDataStream stream(&content,QIODevice::WriteOnly);
                    stream << *cd;
                    PersistentCache::instance()->insert(2,QString::number(m_CurrentDiscIdForEntry.discid,16),content);
                }
                catch (SqLite::Exception e) {
                    cout << e.msg << endl;
                    cout << e.details << endl;
                }
            }
            else {
                emit detailsError("Could not read the requested disc entry from freedb.","Freedb error code: "+QString::number(code));
            }

            buffer.close();
        }

        void FreeDBAccess::removeDiscidFromList (quint32 discid)
        {
            for (int i=m_LookedUpCDReferences.size()-1;i>=0;i--) {
                if (m_LookedUpCDReferences[i].discid==discid) {
                    m_LookedUpCDReferences.removeAt(i);
                }
            }
        }

        void FreeDBAccess::onQueryDiscIdResponse (KJob* job)
        {
            if (job->error()==0) {
                QBuffer buffer;
                buffer.setData(dynamic_cast<KIO::StoredTransferJob*>(job)->data());
                buffer.open(QIODevice::ReadOnly);
                buffer.seek(0);
                QString firstLine= QString::fromUtf8(buffer.readLine());
                bool success;
                int code = firstLine.section(s_CddbSeparator,0,0,QString::SectionSkipEmpty).toInt(&success);
                CDReference::List references;
                if (!success) {
                    emit queryError(i18n("Couldn't understand response from freedb. Ignoring result."),
                                    "Expected number code in first word of response: "+firstLine);
                }
                else switch (code) {
                case 200:
                    // exact match:
                    references.append(CDReference(firstLine.section(s_CddbSeparator,1,1),
                                                  firstLine.section(s_CddbSeparator,2,2).toUInt(NULL,16),
                                                  firstLine.section(s_CddbSeparator,3)));
                    break;
                case 210:
                    // Several exact matches
                case 211:
                    // several matches
                    while (!buffer.atEnd()) {
                        const QString line = QString::fromUtf8(buffer.readLine()).trimmed();
                        if (line==".") break;
                        CDReference ref(line.section(s_CddbSeparator,0,0),
                                        line.section(s_CddbSeparator,1,1).toUInt(NULL,16),
                                        line.section(s_CddbSeparator,2));
                        references.append(ref);
                        removeDiscidFromList (ref.discid);
                    }
                    break;
            case 202:
                    // no match: In this case t might work just reading the cd as is. But this is true for
                    // all genres found for this discid. Therefore all of these are copied to the return
                    // values. List is set up from the end to allow easy removal of discids from
                    // m_LookedUpCDReferences list.
                    for (int i=m_LookedUpCDReferences.size()-1;i>=0;--i) {
                        if (m_LookedUpCDReferences[i].discid==m_CurrentDiscId.discid &&
                            m_LookedUpCDReferences[i].genre!=0) {
                            m_LookedUpCDReferences[i].title="Give it a try";
                            references.prepend(m_LookedUpCDReferences[i]);
                            m_LookedUpCDReferences.removeAt(i);
                        }
                    }
                    if (m_CurrentDiscId.genre!=0) {
                        m_CurrentDiscId.title="Give it a try";
                        references.prepend(m_CurrentDiscId);
                    }
                    break;
            case 403:
            case 409:
                    // cases 403,409: errors
                    emit queryError(i18n("Freedb told me that the db entry is corrupt and could not be read. "
                                         "Maybe try another database?"),firstLine);
                    break;
            default:
                    emit queryError(i18n("Unknown error (%1) when looking up freedb entry. "
                                         "Connection to freedb was established! [Received: %2]",code,firstLine),"");
                }

                try {
                    QByteArray content;
                    QDataStream stream(&content,QIODevice::WriteOnly);
                    stream << references;
                    PersistentCache::instance()->insert(3,QString::number(m_CurrentDiscId.discid,16),content);
                }
                catch (SqLite::Exception e) {
                    LogHook::instance()->warning (i18n("Could not write data to cache: %1",e.msg),0l);
                }

                if (references.size()>0) {
                    emit potentialMatches(references);
                }

                buffer.close();

                queryDiscid();
            }
            else {
                emit queryError("Error while establishing connection to freedb.",job->errorText());
            }
        }

        void FreeDBAccess::queryDiscid ()
        {
            if (m_LookedUpCDReferences.isEmpty() || m_QueryCounter>2) {
                finishedQuery();
                return;
            }

            const CDReference discid = m_LookedUpCDReferences.takeFirst();

            ++m_QueryCounter;
            try {
                QByteArray content= PersistentCache::instance()->lookup(3, QString::number(discid.discid));

                if (!content.isNull()) {
                    CDReference::List entries;
                    QDataStream stream(&content,QIODevice::ReadOnly);
                    stream >> entries;

                    foreach (CDReference r, entries) {
                        m_LookedUpCDReferences.removeOne(r);
                    }

                    emit potentialMatches(entries);
                    queryDiscid();
                    return;
                }
            }
            catch (SqLite::Exception e) {
                cout << e.msg << endl;
                cout << e.details << endl;
            }

            const QString query = queryString(discid.discid,m_TOC);
            m_CurrentDiscId=discid;
            KJob* job=executeHttpCommand (query);
            connect (job, SIGNAL(finished(KJob*)), this, SLOT(onQueryDiscIdResponse(KJob*)));
        }

        void FreeDBAccess::readDiscContent (const CDReference& reference)
        {
            try {
                QByteArray content=PersistentCache::instance()->lookup(2,QString::number(reference.discid,16));
                if (!content.isNull()) {
                    CDEntry entry;
                    QDataStream stream(&content,QIODevice::ReadOnly);
                    stream >> entry;
                    emit detailsMatch(entry);
                    return;
                }
            }
            catch (SqLite::Exception e) {
                cout << e.msg << endl;
                cout << e.details << endl;
            }

            m_CurrentDiscIdForEntry=reference;
            const QString query=readString(reference);
            KJob* job=executeHttpCommand (query);
            connect (job, SIGNAL(finished(KJob*)), this, SLOT(onReadDiscContentResponse(KJob*)));
        }

        quint32 FreeDBAccess::calcDiscId (const QList<int>& toc)
        {
            QList<int> recalcedToc;
            recalcedToc << 2; // Typical value for CD start
            int sum=2;
            for (int i=0;i<toc.size();i++) {
                sum+=toc[i];
                recalcedToc << sum;
            }
            return cddb_discid(recalcedToc);
        }

        CDReference::List FreeDBAccess::searchDiscId (quint64 footprint)
        {
            CDReference::List discids;

            const int width=sizeof(quint64)+sizeof(quint32);

            //
            // Binary search for the best fitting index
            //
            QFile index(Preferences::instance()->freedb.m_IndexFile);
            index.open(QIODevice::ReadOnly);
            int lower=0;
            int upper=index.size()/width;
            IndexLine value;
            while (value.getFootprint()!=footprint && lower<upper) {
                int current = lower+(upper-lower)/2;
                index.seek(current*width);
                value.read(index);
                if (value.getFootprint()<footprint) {
                    lower=current+1;
                }
                else if (value.getFootprint()>footprint) {
                    upper=current-1;
                }
                else {
                    break;
                }
            }

            // No success? return immediately
            if (value.getFootprint()!=footprint) return discids;

            // Go back to lower treshold of footprint
            int i=0;
            qint64 pos = index.pos()-width;
            while (value.getFootprint()==footprint) {
                pos-=width;
                index.seek(pos);
                value.read(index);
                i++;
            }

            // Read all values up to upper treshhold
            int j=0;
            value.read(index);
            do {
                j++;
                discids << CDReference(0 /*item.getCategory()*/,value.getDiscid());
                value.read(index);
            } while (value.getFootprint()==footprint);

            index.close();

            return discids;
        }

        void FreeDBAccess::lookup (const QList<int>& toc, bool sorted)
        {
            if (toc.size()==0) return;

            if (existIndex()) {
                // Calc footprint and do lookup
                quint32 length=0;
                QList<int> mytoc;
                for (int i=0;i<toc.size();i++) {
                    mytoc << toc[i]*75;
                    length+=toc[i]*75;
                }
                const quint64 footprint = IndexLine::calcFootprint(length, mytoc);

                // Do not only collect the discids but also the distance to the requested value. As the
                // data gathering takes place (due to IO restrictions) once upwards the index and once
                // downwards the discids are sorted afterwards by relevance
                m_LookedUpCDReferences=searchDiscId(footprint);
                if (m_LookedUpCDReferences.isEmpty()) {
                    m_LookedUpCDReferences=searchDiscId(footprint+1)+searchDiscId(footprint-1);
                }
            }
            else {
                CDReference discid=CDReference("",calcDiscId(toc));
                if (!m_LookedUpCDReferences.contains(discid)) {
                    m_LookedUpCDReferences << discid;
                }
            }

            m_TOC=toc;
            queryDiscid();
        }
    }
}

#include "freedbaccess.moc"
