/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    $Id$

*************************************************************************/

#include "indexdownloader.h"
#include "options/preferences.h"
#include "util/loghook.h"

#include <QFile>
#include <QFileInfo>
#include <QDir>
#include <QTemporaryFile>
#include <QProcess>
#include <QApplication>
#include <QWidget>

#include <kio/jobclasses.h>
#include <kio/job.h>
#include <kurl.h>
#include <kstatusbarjobtracker.h>
#include <klocale.h>
#include <kmessagebox.h>

#include "aid.h"

using namespace KIO;
using namespace std;

namespace Texel {
namespace FreeDB {

    struct TrackerWithWarning: public KStatusBarJobTracker
    {
        TrackerWithWarning (): KStatusBarJobTracker() {}
        void _warning (KJob *job, const QString &plain, const QString &rich)
        {
            warning(job,plain,rich);
        }
    };

IndexDownloader::IndexDownloader(QObject* parent)
: QObject(parent),
  m_Dest(0l),
  m_TemporaryFile(false)
{
    KUrl url=Preferences::instance()->freedb.m_IndexDownloadURL;

    KIO::TransferJob* job = KIO::get (url,KIO::Reload);
    // Useragent wget starts download from sourceforge without an intermediate download page
    job->addMetaData("UserAgent","wget");

    connect (job, SIGNAL(data(KIO::Job*,QByteArray)), this, SLOT(onData(KIO::Job*,QByteArray)));
    connect (job, SIGNAL(finished(KJob*)), this, SLOT(onFinished(KJob*)));

    const QString destination = Preferences::instance()->freedb.m_IndexFile;
    m_Dest = new QFile(destination);
    bool canWrite=m_Dest->open (QIODevice::WriteOnly|QIODevice::Truncate);
    if (!canWrite) {
        // When write fails try to correct (check if directory exists, remove write protection)
        // if this does not help write to tmp file and offer sudo copy.
        if (m_Dest->exists() && !m_Dest->isWritable()) {
            // file exists is it write procteted can can that be removed?
            if (m_Dest->setPermissions(QFile::WriteUser|m_Dest->permissions())) {
                // retry open. When it fails again the rights are not sufficent
                canWrite=m_Dest->open (QIODevice::WriteOnly|QIODevice::Truncate);
            }
        }
        else if (!m_Dest->exists()) {
            // Check directory and create when necessary
            QFileInfo info(destination);
            if (!info.absoluteDir().exists()) {
                info.absoluteDir().mkpath(info.absoluteFilePath());
                canWrite=m_Dest->open (QIODevice::WriteOnly|QIODevice::Truncate);
            }
        }
    }

    if (!canWrite) {
        // check if kdesudo is accessable
        if (QProcess::execute("which kdesudo")>0) {
            KMessageBox::error(0l,i18n("Can't find command kdesudo. File destination is only accessible with admin "
                                       "privileges therefore kdesudo is required. Please make it accessible for texel "
                                       "or choose another destination."),i18n("Download freedb index"));
            deleteLater();
            return;
        }
        delete m_Dest;
        m_Dest= new QTemporaryFile(this);
        canWrite=m_Dest->open (QIODevice::WriteOnly|QIODevice::Truncate);
        m_TemporaryFile=true;
        LogHook::instance()->warning(i18n("FreeDB reverse lookup index: Destination file not writeable. Write to temporary file and copy later with admin priviledges."),0l);
    }

    m_Tracker = new TrackerWithWarning();
    m_Tracker->registerJob(job);
    m_Tracker->setAutoDelete(job,true);
    m_Tracker->description(job,i18n("Download freedb reverse index"), QPair<QString,QString>(), QPair<QString,QString>());

    connect (m_Tracker, SIGNAL(stopped(KJob*)), this, SLOT(onCancel(KJob*)));

    strm = new z_stream();
    strm->zalloc = Z_NULL;
    strm->zfree = Z_NULL;
    strm->opaque = Z_NULL;
    strm->avail_in = 0;
    strm->next_in = Z_NULL;
    const int ret = inflateInit2(strm,32|15);
    if (ret!=Z_OK) {
        LogHook::instance()->error(i18n("Error initalizing zlib: Download failed"),0l);
        return;
    }
    else {
        LogHook::instance()->message(i18n("FreeDB reverse lookup index: Download started"),0l);
    }
}

IndexDownloader::~IndexDownloader ()
{
    delete strm;
    if (m_Dest!=0l) {
        m_Dest->close();
        delete m_Dest;
    }
    delete m_Tracker;
}

void IndexDownloader::onData (KIO::Job *job, const QByteArray& buffer)
{
    if (dynamic_cast<KIO::TransferJob*>(job)->mimetype()=="application/x-gzip") {
        strm->avail_in=buffer.length();
        strm->next_in=(unsigned char*)(buffer.data());

        char out[0x4000];
        do {
            strm->avail_out = 0x4000;
            strm->next_out = (unsigned char*)(out);

            const int ret = inflate(strm, Z_NO_FLUSH);
            switch (ret) {
            case Z_NEED_DICT:
            case Z_DATA_ERROR:
            case Z_MEM_ERROR:
                QString msg=i18n("Error occured while downlading: Can't decompress file. Error: %1",strm->msg);
                m_Tracker->_warning(job,msg,QString());
                LogHook::instance()->error (msg,0l);
                inflateEnd(strm);
                job->kill();
                return;
            }

            const int have = 0x4000 - strm->avail_out;
            m_Dest->write(out,have);
        } while (strm->avail_out == 0);
    }
}

void IndexDownloader::onFinished (KJob* job)
{
    bool success=true;
    if (m_TemporaryFile) {
        m_Dest->close();
        LogHook::instance()->message(i18n("Move temporary index file %1 to %2",
                                          m_Dest->fileName(),
                                          Preferences::instance()->freedb.m_IndexFile)
                                     ,0l);

        QString attachString;
        if (QApplication::activeWindow()!=0l) {
            attachString=" --attach "+QString::number(QApplication::activeWindow()->winId());
        }

        QString command = "kdesudo -d -u root --comment \""+
                           i18n("Copying of freedb lookup index needs admin priviledges")+"\""
                           +attachString
                           +"-c mv "+m_Dest->fileName()+" "+Preferences::instance()->freedb.m_IndexFile;
        if (QProcess::execute(command)!=0) {
            LogHook::instance()->error(i18n("Command failed: %1",command),0l);
            KMessageBox::error(0l,i18n("File copying failed"),i18n("Download freedb index"));
            success=false;
        }
    }

    if (success) {
        LogHook::instance()->message(i18n("FreeDB reverse lookup index: Download completed"),0l);
    }

    inflateEnd(strm);
    job->deleteLater();
    deleteLater();
}

void IndexDownloader::onCancel (KJob* job)
{
    job->kill();
    job->deleteLater();
    deleteLater();
    inflateEnd(strm);
    // Delete file ats its not complete
    m_Dest->remove();
}

}
}

#include "indexdownloader.moc"
