/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include <QTextStream>
#include <QList>
#include <QStringList>
#include <QFile>
#include <QByteArray>
#include <QRegExp>

#include <KAboutData>
#include <KCmdLineArgs>
#include <klocale.h>

#include <set>

#include "indexline.h"

static QTextStream qout(stdout);
static QTextStream qerr(stderr);

void dump (const QString& filename, quint32 searchDiscid)
{
    QFile f(filename);
    f.open(QIODevice::ReadOnly);
    while (!f.atEnd()) {
        Texel::FreeDB::IndexLine line;
        line.read(f);
        if (searchDiscid==0 || line.getDiscid()==searchDiscid) {
            qout << " " << hex << line.getFootprint() << " "
                  << bin << line.getFootprint() << " "
                  << hex << line.getDiscid()
                 << " " << dec << line.getSize() << " " << line.getLength()<<2;
            qout << endl;
        }
    }
    f.close();
}

int main (int argc, char** argv)
{
    // Commandline handling
    KAboutData aboutData(
      // The program name used internally.
      "indexer",
      // The message catalog name
      // If null, program name is used instead.
      0,
      // A displayable program name string.
      ki18n("Texel indexer"),
      // The program version string.
      "0.1",
      // Short description of what the app does.
      ki18n("Creates a reverse lookup index for freedb entries for the mp3 tagger texel"),
      // The license this code is released under
      KAboutData::License_GPL,
      // Copyright Statement
      ki18n("(c) 2010"),
      // Optional text shown in the About box.
      // Can contain any information desired.
      ki18n("This indexer creates a small index for use with texel to reverse lookup the freedb data "
            "based on existing music files. Input is the complete freedb database as text stream. "
            "Additional helping functionality is provided as well, e.g. reformating for a better "
            "compression."),
      // The program homepage string.
      "http://texel.sf.net< dec << line.getCategory()",
      // The bug report email address
      "mb.jk@t-online.de");

    KCmdLineOptions opts;
    opts.add ("i").add("createindex <indexfile>",ki18n("Reads index data from stdin a creates (binary) indexfile"));
    opts.add ("d").add("dump <file>",ki18n("dump the given index file into a human readable format. Result is written to stdout"));
    opts.add ("discid <id>",ki18n("Discid to dump"));
    opts.add ("decode <id>",ki18n("footprint to decode"));
    opts.add ("",ki18n("Result is always written to <indexfile>"));

    KCmdLineArgs::addCmdLineOptions(opts);
    KCmdLineArgs::init(argc, argv, &aboutData);
    KCmdLineArgs* args = KCmdLineArgs::parsedArgs();

    if (args->getOption("dump")!="") {
        quint32 discid=0;
        if (args->getOption("discid")!=0) {
            discid = args->getOption("discid").toUInt(NULL,16);
        }
        dump(args->getOption("dump"),discid);
        return 2;
    }

    if (args->getOption("decode")!="") {
        quint64 footprint = args->getOption("decode").toULong(NULL,16);
        Texel::FreeDB::IndexLine::decode(footprint);
        return 4;
    }

    QString filename = args->getOption("createindex");
    if (filename.isEmpty()) {
        return 3;
    }

    {
        QFile temp(filename);
        if (!temp.open(QIODevice::WriteOnly)) {
            qerr << i18n("Can't write indexfile %1. Please check.",filename) << endl;
            return 4;
        }
        temp.close();
    }


    QMap<QString,quint32> categories;
    categories.insert("blues",1);
    categories.insert("classical",2);
    categories.insert("data",3);
    categories.insert("country",4);
    categories.insert("folk",5);
    categories.insert("jazz",6);
    categories.insert("newage",7);
    categories.insert("reggae",8);
    categories.insert("rock",9);
    categories.insert("soundtrack",10);
    categories.insert("misc",11);

    QTextStream in(stdin);
    int discs=0;
    int ignored=0;

    std::set<Texel::FreeDB::IndexLine> hashes;
    QRegExp filepattern("^[a-z]*/[a-f0-9]{8,8}$");

    QList<int> lengthes;
    quint32 discLength=0;
    bool inParsingOffsets=false;
    int last_offset=0;

    while (!in.atEnd()) {
        QString line = in.readLine().simplified();
        if (line[0]=='#' && line.contains("Track frame offsets:")) {
            inParsingOffsets=true;
            lengthes.clear();
            last_offset=0;
        }
        else if (inParsingOffsets) {
            if (line[0]=='#') {
                line[0]=' ';
                line=line.simplified();
                if (line.contains("Disc length")) {
                    QString sSeconds = line;
                    sSeconds.remove("Disc length:").remove("seconds");
                    discLength = sSeconds.toLong()*75;
                    lengthes << discLength-last_offset;
                    inParsingOffsets=false;
                }
                else if (!line.isEmpty()) {
                    bool success;
                    const int offset=line.toInt(&success,10);
                    if (success && offset!=0) {
                        if (last_offset>0) {
                            lengthes << offset-last_offset;
                        }
                        last_offset=offset;
                    }
                }
            }
        }
        else if (line.startsWith("DISCID=")) {
            bool success;
            QString sId = line;
            sId.remove("DISCID=");
            sId=sId.split(",")[0];
            const quint32 discid=sId.toULong(&success,16);
            if (!success) {
                qerr << "Warning: could not parse discid from: " << line
                     << "(" << sId << ")" << endl;
                ignored++;
            }

            quint64 footprint = Texel::FreeDB::IndexLine::calcFootprint(discLength,lengthes);
            hashes.insert(Texel::FreeDB::IndexLine(footprint,discid));

            ++discs;
            if (discs%1000==0) {
                qerr << "Processed " << dec << discs << " entries\r" << flush;
            }
        }
    }

    QFile file (filename);
    file.open (QIODevice::WriteOnly|QIODevice::Truncate);

    for (std::set<Texel::FreeDB::IndexLine>::iterator iter=hashes.begin();iter!=hashes.end();iter++) {
        (*iter).write(file);
    }

    file.close();
}
