/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "indexline.h"

#include <QSet>
#include <QTime>

#include "aid.h"
#include <iostream>

using namespace std;

namespace Texel {

    namespace FreeDB {

        IndexLine::IndexLine ()
            : footprint(0), discid(0)
        {
        }

        IndexLine::IndexLine (quint64 f)
            : footprint(f), discid(0)
        {
        }

        IndexLine::IndexLine (quint64 f, quint32 c)
            : footprint(f), discid(c)
        {
        }

        void IndexLine::read (QIODevice& dev)
        {
            dev.read((char*)&footprint,sizeof(footprint));
            dev.read((char*)&discid,sizeof(discid));
        }

        void IndexLine::write (QIODevice& dev) const
        {
            dev.write((const char*)&footprint,sizeof(footprint));
            dev.write((const char*)&discid,sizeof(discid));
        }

        bool IndexLine::operator> (const IndexLine& other) const
        {
            return other.operator<(*this);
        }

        bool IndexLine::operator< (const IndexLine& other) const
        {
            if (footprint==other.footprint) {
                return discid<other.discid;
            }
            return footprint<other.footprint;
        }

        bool IndexLine::operator!= (const IndexLine& other) const
        {
            return footprint!=other.footprint;
        }

        QString decode_time (quint16 seconds, quint16 delta=3)
        {
            QTime t(0,0,0);
            return t.addSecs(seconds).toString("mm:ss") + " - " + t.addSecs(seconds+delta).toString("mm:ss");
        }

        void IndexLine::decode (quint64 footprint)
        {
            const quint16 no_tracks = ((footprint>>(64-6))&0b111111);
            quint16 tracks[6];
            tracks[0]= 4*((footprint>>(64-6-8  ))&0b11111111);
            tracks[1]= 4*((footprint>>(64-6-8*2))&0b11111111);
            tracks[2]= 4*((footprint>>(64-6-8*3))&0b11111111);
            tracks[3]= 4*((footprint>>(64-6-8*4))&0b11111111);
            tracks[4]= 4*((footprint>>(64-6-8*5))&0b11111111);
            tracks[5]= 4*((footprint>>(64-6-8*6))&0b11111111);
            const quint16 length = ((footprint>>(0))&(quint64)(0b1111111111))*8;

            cout << "Print    : " << hex << footprint << dec << endl
                 << "Tracks   : " << no_tracks << endl
                 << "CD Length: " << decode_time(length,7) << endl
                 << "Track0   : " << decode_time(tracks[0]) << endl
                 << "Track1   : " << decode_time(tracks[1]) << endl
                 << "Track2   : " << decode_time(tracks[2]) << endl
                 << "TrackN-2 : " << decode_time(tracks[3]) << endl
                 << "TrackN-1 : " << decode_time(tracks[4]) << endl
                 << "TrackN   : " << decode_time(tracks[5]) << endl;
        }

        inline int set(quint64& bits,int startbit, quint16 source, int bits2copy)
        {
            const quint64 cutted_source = ((quint64)source)&((((quint64)2)<<(bits2copy))-1);
            const quint64 move = startbit-bits2copy;
#if 0
            cout << "Source: " << source << "," << cutted_source << endl;
            cout << "Move: " << move << "," << (cutted_source<<move) << endl;
            cout << hex << bits << endl;
#endif
            bits|=(cutted_source<<move);
#if 0
            cout << hex << bits << dec << endl;
#endif
            return move;
        }

        /*
         ** The 64 bits of the footprint are calculated as follows:
         **
         **    6 bits: number of tracks (if more than 63 tracks are given 0 is stored)
         **            Taken fromt he number of elements in the toc
         **   11 bits: length of cd in seconds (last two bits are skipped, this gives a
         **            range of 2^13 = 4096 seconds = 136 min. As the
         **  5x8 bits: Bits 8-3 from the three longest songs and the three shortest
         ** +1x7 bits  songs of the cd. The shortest song gets just 5 bits. 6 bits offer
         **            1024 seconds = 17 minutes range (as the least two bits are skipped)
         **            7 bits 8 minutes.
         **
         ** The least two bits are always skipped to allow the a second difference between
         ** found cds and the coded tracks.
         **
         ** length: Length of the CD in 1/75th seconds.
         ** toc: Length of the tracks in 1/75th seconds.
         ** sorted: if true the to is not sorted ascending by length but assumed to be
         */

        quint64 IndexLine::calcFootprint (quint32 _length, QList<int> toc)
        {
            quint64 footprint=0;
            int unsetbits=64;

            qSort(toc);

            // Bits 64-58: number of tracks
            quint8 size= (toc.size()>63)?0:toc.size();
            unsetbits=set (footprint, unsetbits, size, 6);

            while (toc.size()<6) toc<<0;    // extend to six entries
            const int last=toc.size()-1;

            unsetbits=set (footprint, unsetbits, (toc[last]/75)>>2, 8);
            unsetbits=set (footprint, unsetbits, (toc[last-1]/75)>>2, 8);
            unsetbits=set (footprint, unsetbits, (toc[last-2]/75)>>2, 8);
            unsetbits=set (footprint, unsetbits, (toc[2]/75)>>2, 8);
            unsetbits=set (footprint, unsetbits, (toc[1]/75)>>2, 8);
            unsetbits=set (footprint, unsetbits, (toc[0]/75)>>2, 8);

            // Bits 10-1: length without the three last bits
            quint16 length=(_length/75)>>3;
            unsetbits=set (footprint, unsetbits, length, 10);

            //decode(footprint);

            return footprint;
        }

#if 0
        bool IndexLine::fit (quint16 _length, const QList<int>& toc, bool sorted) const
        {
            if (toc.size() != size()) return false;
            if (diff(_length,cdlength())>toc.size()) return false;

            QSet<int> used;
            quint32 toShift=chksum;
            for (int i=3;i>=0;i--) {
                if (sorted) {
                    const int n = (toc.size()>i)?(toc[i]&0xff):0;
                    if (diff(n,int(toShift&0xff))>1) return false;
                }
                else {
                    int minIndex=-1;
                    for (int j=0;j<toc.size();j++) {
                        const int n = toc[j]&0xff;
                        if (diff(n,int(toShift&0xff))<3 && !used.contains(j) &&  // value j fits and is unused
                            (minIndex==-1 || toc[minIndex]>toc[j]))             // value j as also a smaller distance than everything already found
                        {
                            minIndex=j;
                        }32 length;
                    }
                    if (minIndex==-1) return false;
                    used << minIndex;
                }
                toShift>>=8;
            }

            return true;
        }

        IndexLine IndexLine::deltaTo (const IndexLine& smaller) const
        {
            IndexLine result;
            result.length=length-smaller.length;
            result.chksum=chksum;
            const quint8 s = size()-smaller.size();
            quint16 l=(discid>>8)&0xffff;
            if (s==0) {
                l-=((smaller.discid>>8)&0xffff);
            }
            result.discid= (discid&0xff000000)|(l<<8)|s;
            return result;
        }

        IndexLine IndexLine::addDelta (const IndexLine& delta) const
        {
            IndexLine result;
            result.length=length+delta.length;
            result.chksum=delta.chksum;
            const quint8 s = size()+delta.size();
            quint16 l;
            if (delta.size()==0) {
                l=((discid>>8)&0xffff)+((delta.discid>>8)&0xffff);
            }
            else l=((delta.discid>>8)&0xffff);
            result.discid=(delta.discid&0xff000000)|(l<<8)|s;
            return result;
        }
#endif

        std::ostream& operator<< (std::ostream& out, const IndexLine& l)
        {
            out << l.getLength() << "/" << (int)l.getSize();
            return out;
        }
    }
}
