/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#ifndef INDEXLINE_H
#define INDEXLINE_H

#include <iostream>
#include <QIODevice>
#include <QVector>

namespace Texel {
    namespace FreeDB {

        /**
      Represents a index line of the index file. The class provides functionality to read, write
      and interpret a cache line.
      */
        class IndexLine
        {
        private:
            quint64 footprint;
            quint32 discid;

        public:
            IndexLine ();
            explicit IndexLine (quint64 f);
            explicit IndexLine (quint64 f, quint32 c);

            quint32 getLength() const;
            quint32 getSize() const;
            quint32 getDiscid() const;
            quint64 getFootprint() const;

            /**
          Reads an entry from the current position at the device to the current instance. The data
          is not read cpu architecture independent but does use the architecture byte order.
          */
            void read (QIODevice& dev);

            /**
          Writes an entry at the current position to the given device. The data is not written cpu
          architecture independent but does use the architecture byte order.
          */
            void write (QIODevice& dev) const;

            /** Returns the number of tracks encoded in the discid
          */
            quint8 size () const;

            /** Returns the cdlength encoded in the discid.
          */
            quint16 cdlength () const;

            /**
          Tests if an given IndexLine is smaller than this one. There is a complete order
          defined on the cache lines to allow binary search.
          */
            bool operator> (const IndexLine& other) const;

            /**
          Tests if an given IndexLine is larger than this one. There is a complete order
          defined on the cache lines to allow binary search.
          */
            bool operator< (const IndexLine& other) const;

            /**
          Tests of a given index is different from another one.
          @param other other IndexLine to test against.
          @return true when both instances contain different data.
          */
            bool operator!= (const IndexLine& other) const;

            /**
          Checks if an index entry does really fit to the searched toc. In case the toc
          is unsorted it is only checked if the toc contains similar lengthes as this
          indexline, otherwise the correct position is checked.
          @param _length ength of the cd
          @param toc Table of contents of the CD searched for
          @param sorted If true, strict checking is done, other relaxed, position
                 independent checking.
          */

            static quint64 calcFootprint (quint32 _length, QList<int> toc);
            static void decode (quint64 footprint);
        };

        inline quint32 IndexLine::getLength() const
        {
            return (quint32)(footprint>>46)&((1<<12)-1);
        }

        inline quint32 IndexLine::getSize() const
        {
            return (quint32)(footprint>>58);
        }

        inline quint32 IndexLine::getDiscid() const
        {
            return discid;
        }

        inline quint64 IndexLine::getFootprint() const
        {
            return footprint;
        }
    }
}

std::ostream& operator<< (std::ostream& out, const Texel::FreeDB::IndexLine& l);

#endif // INDEXLINE_H
