#include "querymenu.h"
#include "detailsmenu.h"
#include "model/tagmodelproxy.h"
#include "func/metatagservices/freedb/indexdownloader.h"
#include "util/loghook.h"

#include <QWidgetAction>
#include <QVBoxLayout>
#include <QLabel>
#include <QProgressBar>
#include <QStyle>
#include <QIcon>
#include <QHBoxLayout>

#include <klocale.h>
#include <kdebug.h>

#include <boost/tuple/tuple.hpp>

#include "aid.h"

using namespace std;

namespace Texel {
    namespace FreeDB {

        QueryMenu::QueryMenu(TagModelProxy* model, const TagItem::List& selectedRows, QWidget *parent) :
                QMenu(parent),
                m_SelectedRows(selectedRows),
                m_Accessor(0L),
                m_Model(model),
                status(0),
                m_ResultsAvailable(false)
        {
            connect (this, SIGNAL(aboutToShow()), this, SLOT(onShow()));
            setTitle(i18n("Lookup in FreeDB"));
        }

        QueryMenu::~QueryMenu ()
        {
            delete m_Accessor;
        }

        namespace {
            typedef boost::tuple<int,int,TagItem*> sortitem;

            bool tuple_sort (const sortitem& p1, const sortitem& p2)
            {
                return p1.get<1>()<p2.get<1>();
            }

            typedef boost::tuple<int,int,QString> matchitem;
            typedef boost::tuple<TagItem*,int,QString> matchitem_p;

            bool matchitem_sort (const matchitem& p1, const matchitem& p2)
            {
                if (p1.get<1>()==p2.get<1>()) {
                    return p1.get<2>()<p2.get<2>();
                }
                return p1.get<1>()<p2.get<1>();
            }

            bool matchitem_p_sort (const matchitem_p& p1, const matchitem_p& p2)
            {
                if (p1.get<1>()==p2.get<1>()) {
                    return p1.get<2>()<p2.get<2>();
                }
                return p1.get<1>()<p2.get<1>();
            }

        }

        void QueryMenu::onShow()
        {
            if (status==0) {
                QList<int> toc;
                TagItem::List sortedSelectedRows;

                // Sort toc first by track information. When its sorted it can be determined if a complete
                // order was giben or no. If no complete order was given the tracklist is considered as
                // unsorted and a relaxed search will be done.
                QList<sortitem> positionToc;
                for (int i=0;i<m_SelectedRows.size();i++) {
                    positionToc << sortitem(m_SelectedRows[i]->getIntValue(ColumnManager::Length),
                                            m_SelectedRows[i]->getIntValue(ColumnManager::Track),
                                            m_SelectedRows[i]);
                }

                qSort (positionToc.begin(),positionToc.end(),tuple_sort);

                // Copy and examine if tracks are increasing without a gap
                int lastIndex=0;
                m_Sorted=true;
                for (int i=0;i<positionToc.size();i++) {
                    if (lastIndex+1!=positionToc[i].get<1>() && m_Sorted==true) m_Sorted=false;
                    toc << positionToc[i].get<0>();
                    lastIndex=positionToc[i].get<1>();
                    sortedSelectedRows << positionToc[i].get<2>();
                }

                m_SelectedRows=sortedSelectedRows;

                m_Accessor = new FreeDBAccess ();
                connect (m_Accessor, SIGNAL(potentialMatches (CDReference::List)), this, SLOT(onResults(CDReference::List)));
                connect (m_Accessor, SIGNAL(queryError(QString,QString)), this, SLOT(onError(QString)));
                connect (m_Accessor, SIGNAL(finishedQuery()), this, SLOT(onFinished()));

                //
                //  Install a widget with a progressbar and a describing text
                //
                QWidgetAction* progress = new QWidgetAction(this);
                QWidget* container = new QWidget(this);
                delete container->layout();
                container->setLayout(new QVBoxLayout(container));
                QLabel* label = new QLabel(container);
                label->setText("Wait for lookup");
                container->layout()->addWidget(label);
                QProgressBar* bar = new QProgressBar(container);
                bar->setRange(0,0);
                container->layout()->addWidget(bar);
                progress->setDefaultWidget(container);
                addAction(progress);
                status=1;

                m_Accessor->lookup(toc,m_Sorted);
            }
        }

        void QueryMenu::onError(const QString& description)
        {
            if (status==1) {
                status=2;
                clear();
                errorToLabel(description);
            }
            LogHook::instance()->error(description,0l);
        }

        void QueryMenu::errorToLabel(const QString &description)
        {
            QWidgetAction* error = new QWidgetAction(this);
            QWidget* container = new QWidget(this);
            QHBoxLayout* layout = new QHBoxLayout();
            QLabel* message = new QLabel(this);
            message->setMargin(3);
            message->setWordWrap(true);
            message->setText (description);

            QLabel* icon = new QLabel(this);
            icon->setPixmap(message->style()->standardIcon(QStyle::SP_MessageBoxWarning).pixmap(24,24));

            container->setLayout(layout);
            layout->addWidget(icon);
            layout->addWidget(message);

            error->setDefaultWidget(container);
            addAction(error);
        }

        void QueryMenu::onResults (const CDReference::List& references)
        {
            if (status!=2) {
                clear();
            }

            if (!references.isEmpty()) {
                foreach (CDReference ref, references) {
                    if (!m_Displayed.contains(ref)) {
                        addMenu(new DetailsMenu(m_Accessor, ref, this));
                        m_Displayed.insert(ref);
                    }
                }
                m_ResultsAvailable=true;
            }

            status=2;
        }

        void QueryMenu::onFinished ()
        {
            if (!m_ResultsAvailable) {
                clear();
                QString text(i18n("No match found!"));
                if (!FreeDBAccess::existIndex()) {
                    text+="\n\n";
                    text+=i18n("You may improve the result by downloading the reverse lookup index for freedb "
                               "from the texel homepage. Please consult the manual.");
                    addAction(i18n("Download index"), this, SLOT(downloadIndex()));
                    errorToLabel(text);
                }
                else {
                    errorToLabel(text);
                }
            }
        }

        TagItem::List QueryMenu::reorderByCDEntry (const CDEntry& entry)
        {
            if (m_Sorted) {
                return m_SelectedRows;
            }

            QList<matchitem> tags;
            QList<matchitem> proposal;

            if (entry.tracks.size()!=m_SelectedRows.size()) {
                kDebug() << "Different sizes, can't match proposal to selected rows" << endl;
                return TagItem::List();
            }

            for (int i=0;i<m_SelectedRows.size();i++) {
                tags << matchitem(i,
                                  m_SelectedRows[i]->getIntValue(ColumnManager::Length),
                                  m_SelectedRows[i]->data(ColumnManager::Title,Columns::Column::RoleDisplay));
                proposal << matchitem(i,entry.tracks[i].runtime,entry.tracks[i].title);
            }

            qStableSort (tags.begin(),tags.end(),matchitem_sort);
            qStableSort (proposal.begin(),proposal.end(),matchitem_sort);

            TagItem::List mappedRows;
            for (int i=0;i<m_SelectedRows.size();i++) mappedRows << 0;
            for (int i=0;i<m_SelectedRows.size();i++) {
                mappedRows[proposal[i].get<0>()]=m_SelectedRows[tags[i].get<0>()];
            }

            return mappedRows;
        }

        void QueryMenu::downloadIndex ()
        {
            // Class instance does its work and deletes itself
            new IndexDownloader(this);
        }
    }
}

#include "querymenu.moc"
