/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "puid.h"
#include <sox.h>
#include <ofa1/ofa.h>

#include <algorithm>

#include <QString>
#include <QHttp>
#include <QBuffer>
#include <QByteArray>
#include <QFile>
#include <QXmlQuery>

#include "aid.h"

#include <kio/job.h>

#include "model/tagitem.h"
#include "func/metatagservices/musicbrainz/musicbrainztagsuggestion.h"

static bool m_Init=false;

namespace Texel {

PUID::PUID()
{
    if (!m_Init) { sox_init(); m_Init=true; }
}

PUID::~PUID ()
{
    //sox_quit();
}

QByteArray PUID::readSamplesWithSOX (const QString& filename)
{
    sox_format_t* in;
    in=sox_open_read(filename.toLocal8Bit(),0l,0l,0l);
    const int channels = in->signal.channels;
    const int rate = in->signal.rate;

    // Read first 135 seconds
    const int samplesToRead = 135*channels*rate;
    QByteArray buffer;
    buffer.resize(samplesToRead*sizeof(short));

    const int toRead=4096;
    sox_sample_t buf[toRead];
    int p=0;

    while (p<samplesToRead) {
        const int read=sox_read(in, buf, toRead);
        if (read==SOX_EOF) break;

        short* tmp=(short*)buffer.data();
        SOX_SAMPLE_LOCALS;
        for (int i=0;i<read && p<samplesToRead;i++){
            tmp[p++] = SOX_SAMPLE_TO_SIGNED_16BIT(buf[i],in->clips);
        }
    }

    sox_close(in);
    return  buffer;
}

QString PUID::getPUID (TagItem* tag)
{
    const QString filename=tag->getAbsolutFilename();
    QByteArray samples = readSamplesWithSOX(filename);

    const char* fp=::ofa_create_print((unsigned char*)samples.constData(),
                                      OFA_LITTLE_ENDIAN,
                                      samples.size()/sizeof(short),
                                      tag->getIntValue(ColumnManager::SampleRate),
                                      tag->getIntValue(ColumnManager::Channels)==2?1:0);

    // Retrive puid
    QByteArray data;

    QString params
      = "cid=03a1344bad8caaea06299360c9530c8f&" // cid=57aae6071e74345f69143baa210bda87&"       // Client ID
            "cvr=1.0&"                  // Client Version
            "fpt=%1&"                   // Fingerprint
            "rmd=0&"                    // m = 1: return metadata; m = 0: only return id
            "brt=%2&"                   // bitrate (kbps)
            "fmt=%3&"                   // File extension (e.g. mp3, ogg, flac)
            "dur=%4&"                   // Length of track (milliseconds)
            "art=unknown&"              // Artist name. If there is none, send "unknown"
            "ttl=unknown&"              // Track title. If there is none, send "unknown"
            "alb=unknown&"              // Album name. If there is none, send "unknown"
            "tnm=0&"                    // Track number in album. If there is none, send "0"
            "gnr=unknown&"              // Genre. If there is none, send "unknown"
            "yrr=0&"                    // Year. If there is none, send "0"
            "enc=e"                     // Encoding. e = true: ISO-8859-15; e = false: UTF-8 (default). Optional.
            "\r\n";

    data.append(params.arg(QString(fp),
               QString::number(tag->getIntValue(ColumnManager::Bitrate)),
               tag->typeName(),
               QString::number(tag->getIntValue(ColumnManager::Length)*1000)).toLatin1());

    std::cout << data.data() << std::endl;

    KIO::TransferJob* job=KIO::http_post(KUrl("http://ofa.musicdns.org/ofa/1/track"), data, KIO::HideProgressInfo);
    job->addMetaData("content-type", "application/x-www-form-urlencoded" );
    job->setProperty("tagitem",QVariant::fromValue<TagItem*>(tag));
    connect (job,SIGNAL(data(KIO::Job*,QByteArray)), this, SLOT(onData(KIO::Job*,QByteArray)));
    connect (job,SIGNAL(result(KJob*)),this,SLOT(onFinish(KJob*)));

    job->start();

    return "";
}

void PUID::onData(KIO::Job*, const QByteArray &d)
{
    m_Response+=d;
}

void PUID::onFinish(KJob *job)
{
    std::cout << "Data: " << m_Response.data() << std::endl;

    TagItem* referer = job->property("tagitem").value<TagItem*>();

    QXmlQuery d;
    QStringList puids;
    d.setFocus(m_Response);
    d.setQuery("declare default element namespace \"http://musicbrainz.org/ns/mmd-1.0#\";\
                /metadata/track/puid-list/puid/@id/string()");
    if (!d.evaluateTo(&puids)) { std::cout << "Error" << std::endl; }

    if (puids.size()==0) {
        std::cout << "Kein Ergebnis" << std::endl;
        return;
    }

    foreach (QString puid, puids) {
        MusicBrainzTagSuggestion::createTagItemFromPUID(referer,puid);
    }
}

}

#include "puid.moc"

