/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "parser.h"
#include "searchpatternparser.h"
#include "model/tagitem.h"

#include <boost/spirit/include/classic.hpp>
#include <boost/spirit/include/classic_position_iterator.hpp>

#include <QStack>
#include <QStringList>
#include <QMap>
#include <QtAlgorithms>

#include <klocalizedstring.h>
#include <kstringhandler.h>

#include "func/parserproblem.h"
#include "util/qtlevenshtein.h"
#include "textfiltersingleton.h"
#include "aid.h"
#include "auto_ptr.h"

using namespace std;
using namespace BOOST_SPIRIT_CLASSIC_NS;

typedef position_iterator<QString::const_iterator> posIterator;

////////////////////////////////////////////////////////////////////////////
//
// Additional methods needed to use a QString directly instead of the
// default char.
//
////////////////////////////////////////////////////////////////////////////

static int isdigit (QChar c)
{
    return c.isDigit();
}

#if 0
static int isprint (QChar c)
{
    return c.isPrint();
}

static int isspace (QChar c)
{
    return c.isSpace();
}
#endif

static int isalnum (QChar c)
{
    return c.isLetterOrNumber();
}

static int isalpha (QChar c)
{
    return c.isLetter();
}

static char operator- (QChar c, char n)
{
    return c.toAscii()-n;
}

static char operator+ (QChar c, char n)
{
    return c.toAscii()+n;
}

namespace Texel {

    namespace {
        static void copy (QString& dest, posIterator start, posIterator end)
        {
            while (start!=end) dest.append(*(start++));
        }

     /******************************************************************************************
     *
     * Define the structures used in the stack
     *
     *****************************************************************************************/

        struct Expression
        {
            QString result;
            virtual QString name () const { return "Expression"; }
            virtual ~Expression() {}
        };

        /******************************************************************************************
     *
     * Defines the context used while parsing. All temporay results are stored here.
     * Main parsing results are stored in the stack, the rest is runrime management.
     *
     *****************************************************************************************/

        struct AggregatedContext
        {
            AggregatedContext (ParserContext& _context, const QVariant& _localContext)
                :context(_context),
                localContext(_localContext)
            {
            }

            ~AggregatedContext ()
            {
                while (!stack.isEmpty()) delete stack.pop();
            }

            mutable QStack<Expression*> stack;
            ParserContext& context;
            const QVariant& localContext;
            QString m_Input;
            position_iterator<QString::const_iterator> position;

            template <class T> T top_cast (posIterator pos_=posIterator()) const
            {
                return safe_cast<T>(stack.top(),pos_);
            }

            QString getMarkedLine (int marker) const
            {
                const int rightFraction = m_Input.size()-marker;
                QString result=m_Input.left(marker)+"<-- ";
                if (rightFraction>0) result+=m_Input.right(rightFraction);
                return result;
            }

        private:
            template <class T> T safe_cast (Expression* exp, posIterator pos_) const
            {
                const T t = dynamic_cast<T>(exp);
                if (t==0l) {
                    int pos=0;
                    if (pos_!=posIterator()) pos=pos_.get_position().column;

                    UnknownProblem* p=new UnknownProblem(i18n("Parse error at [%1]:\n"
                                                              "Its probably a bug in the application logic."
                                                              "Please file a bug report.\n",QString::number(pos))
                                                         + getStackTrace());

                    throw_ (pos_, (Texel::InputProblem*)p);
                }
                return t;
            }

            QString getStackTrace () const
            {
                QString result;
                for (int i=0;i<stack.size();i++) {
                    result+="ST -- " +stack[i]->name()+": "+stack[i]->result+"\n";
                }
                return result;
            }
        };

        struct Variable: public Expression
        {
            bool org;
            virtual QString name () const { return "Variable"; }
            virtual ~Variable () {}
        };

        class Function: public Expression
        {
            void assertArgs (posIterator start, int nargs, const QString funcName) const
            {
                if (args.size()!=nargs) {
                    InputProblem* ip=new UnknownProblem(i18n("Wrong number of arguments for function %1:exepected %2, found %3",funcName,nargs,args.size()));
                    throw_ (start,ip);
                }
            }

            QString trim (const AggregatedContext&, posIterator start) const
            {
                assertArgs (start, 0, "trim");
                return result.simplified();
            }

            QString filter (const AggregatedContext&, posIterator start) const
            {
                assertArgs(start, 1, "filter");
                QString r;
                for (int i=0;i<result.size();i++) {
                    if (args[0].indexOf(result[i])!=-1) {
                        r+=result[i];
                    }
                }
                return r;
            }

            QString eval (const AggregatedContext& agg, posIterator start) const
            {
                assertArgs(start, 0, "eval");
                return agg.context.eval("="+result,QVariant());
            }

            QString replace (const AggregatedContext&, posIterator start) const
            {
                assertArgs (start,2,"replace");
                QString text = result;
                QRegExp rx (args[0]);
                return text.replace(rx,args[1]);
            }

            QString lower (const AggregatedContext&, posIterator start) const
            {
                assertArgs (start,0,"lower");
                return result.toLower();
            }

            QString upper (const AggregatedContext&, posIterator start) const
            {
                assertArgs (start,0,"upper");
                return result.toUpper();
            }

            QString simplify (const AggregatedContext&, posIterator start) const
            {
                assertArgs (start,0,"simplify");
                return TextFilterSingleton::instance()->filter(result);
            }

            QString prefix (const AggregatedContext&, posIterator start) const
            {
                assertArgs (start,1,"prefix");
                if (result!="") return args[0]+result;
                return "";
            }

            QString prefix0 (const AggregatedContext&, posIterator start) const
            {
                assertArgs (start,1,"prefix0");
                if (result!="0") return args[0]+result;
                return "";
            }

            QString postfix (const AggregatedContext&, posIterator start) const
            {
                assertArgs (start,1,"postfix");
                if (result!="") return result+args[0];
                return result;
            }

            QString remove (const AggregatedContext&, posIterator) const
            {
                QString r;
                QString filter("/:;*?\"<>|");
                if (args.size()==1) {
                    filter=args[0];
                }

                for (int i=0;i<result.size();i++) {
                    const QChar& c=result[i];
                    if (c.isLetterOrNumber() || filter.contains(c)) {
                        r+=c;
                    }
                }

                return r;
            }

            QString capwords (const AggregatedContext&, posIterator start) const
            {
                assertArgs(start,0,"capwords");
                return KStringHandler::capwords(result.toLower());
            }

            QString caseCorrection (const AggregatedContext&, posIterator start) const
            {
                assertArgs(start,0,"casec");

                bool hasLower=false;
                bool hasUpper=false;
                foreach (QChar u, result) {
                    hasLower|=u.isLower();
                    hasUpper|=u.isUpper();
                }

                if (!hasLower || !hasUpper) {
                    return KStringHandler::capwords(result.toLower());

                }

                return result;
            }

            QString substitute (const AggregatedContext&, posIterator start) const
            {
                assertArgs (start,2,"substitute");

                const QString& org=args[0];
                const QString& rep=args[1];

                if (org.size()!=rep.size()) {
                    InputProblem* ip=new UnknownProblem(i18nc("substitute must not be translated, its a function name!",
                                                              "substitute argument error: the two strings must have the same length."));
                    throw_ (start,ip);
                }

                QString result;
                for (int i=0;i<result.size();i++) {
                    const QChar& c = result[i];
                    const int pos=org.indexOf(c);
                    if (pos!=-1) result+=rep[pos];
                    else result+=c;
                }

                return result;
            }

            QString extract (const AggregatedContext&, posIterator start) const
            {
                assertArgs(start,2,"extract");

                const QString& regex=args[0];
                const QString& replacement=args[1];

                QRegExp rx(regex);
                if (rx.isValid()) {
                    QString copy(result); // as result is const
                    return copy.replace(rx,replacement);
                }

                // Invalid reular expression, throw an error
                InputProblem* ip=new UnknownProblem(i18n("Error in the regular expression \"%1\". Can't parse.",regex));
                throw_ (start,ip);
                return ""; // Never reached
            }

            typedef QString(Function::*FuncPointer)(const AggregatedContext&, posIterator) const;
            static QMap<QString, FuncPointer> funcs;

        public:
            static QStringList availableFunctions ()
            {
                return funcs.keys();
            }

            void exec (const AggregatedContext& agg, posIterator start)
            {
                FuncPointer function = funcs.value(funcName,0l);
                if (function!=0l) {
                    result=(this->*function)(agg,start);
                }
                else {
                    QString suggestedFuncName;
                    int dist = max(funcName.size()-2,2);
                    foreach (QString name, funcs.keys())
                    {
                        const int d=Texel::distance(funcName,name,dist);
                        if (d<dist) {
                            suggestedFuncName=name;
                        }
                    }

                    if (suggestedFuncName!="") {
                        const QString problem=i18n("Found unknown function <%1>, did you mean <%2>?",funcName,suggestedFuncName);
                        ParserProblem* ip=new ParserProblem(problem);
                        ip->setSuggestion(agg.m_Input, suggestedFuncName, m_Start.get_position().column-1,
                                          m_End.get_position().column-1);
                        throw_ (m_End,(Texel::InputProblem*)ip);
                    }
                    else {
                        InputProblem* ip=new UnknownProblem(i18n("Found unknown function <%1>.",funcName));
                        throw_ (m_End,ip);
                    }
                }
            }

            static void init ()
            {
                if (funcs.size()==0) {
                    funcs.insert("remove",&Function::remove);
                    funcs.insert("prefix",&Function::prefix);
                    funcs.insert("prefix0",&Function::prefix0);
                    funcs.insert("postfix",&Function::postfix);
                    funcs.insert("upper",&Function::upper);
                    funcs.insert("lower",&Function::lower);
                    funcs.insert("simplify",&Function::simplify);
                    funcs.insert("replace",&Function::replace);
                    funcs.insert("trim",&Function::trim);
                    funcs.insert("substitute",&Function::substitute);
                    funcs.insert("extract",&Function::extract);
                    funcs.insert("capwords",&Function::capwords);
                    funcs.insert("eval",&Function::eval);
                    funcs.insert("casec",&Function::caseCorrection);
                }
            }

            virtual ~Function() {}
            virtual QString name () const { return "Function"; }

            QString funcName;
            posIterator m_Start;
            posIterator m_End;
            QStringList args;
        };

        QMap<QString, Function::FuncPointer> Function::funcs;

        struct Range: public Expression
        {
            virtual QString name () const { return "Range"; }

            virtual ~Range() {}
            int start;
            int end;
            bool interval;
            QString org;

            void reset ()
            {
                start=-1;
                end=-1;
                interval=false;
            }
        };

     /******************************************************************************************
     *
     * Definition of all used functors. All functors are linked to the context.
     *
     *****************************************************************************************/

        struct do_StartExpression
        {
            do_StartExpression(const AggregatedContext& a)
                : agg(a)
            {
            }

            void operator() (posIterator, posIterator) const
            {
                start();
            }

            void operator() (const QChar) const
            {
                start();
            }

        private:
            void start() const
            {
                agg.stack.push(new Expression());
            }

            const AggregatedContext& agg;
        };

        struct do_StartVar
        {
            do_StartVar (const AggregatedContext& a)
                : agg(a)
            {
            }

            void operator() (posIterator, posIterator) const
            {
                agg.stack.push(new Variable());
            }

        private:
            const AggregatedContext& agg;
        };

        struct do_SetVarType
        {
            do_SetVarType (const AggregatedContext& a)
                : agg(a)
            {
            }

            void operator() (posIterator start, posIterator) const
            {
                const bool org=(*start!='$');
                agg.top_cast<Variable*>(start)->org=org;
            }

        private:
            const AggregatedContext& agg;
        };

        struct do_SetVarName
        {
            do_SetVarName (const AggregatedContext& a)
                : agg(a)
            {
            }

            void operator() (posIterator start, posIterator end) const
            {
                QString varname;
                copy (varname,start,end);
                Variable* var=agg.top_cast<Variable*>(start);

               if (agg.context.setDependency(varname, var->org, agg.localContext)) {
                    var->result=agg.context.getVar(varname, var->org, agg.localContext);
                }
                else {
                    const QString suggestion=agg.context.suggestVarName(varname);
                    QString problem;
                    if (suggestion.isEmpty()) {
                        problem=i18n("Variable \"%1\" is not defined.",varname);
                    }
                    else {
                        problem=i18n("Variable \"%1\" is not defined. Did you mean \"%2\"?",varname,suggestion);
                    }
                    ParserProblem* p=new ParserProblem(problem);
                    p->setSuggestion(agg.m_Input, suggestion, start.get_position().column-1,
                                     end.get_position().column-1);
                    throw_ (end, (Texel::InputProblem*)p);
                }
            }

        private:
            const AggregatedContext& agg;
        };

        struct do_Pop
        {
            do_Pop  (const AggregatedContext& a)
                : agg(a)
            {
            }


            void operator() (posIterator, posIterator) const
            {
                pop();
            }

            void operator() (const QChar) const
            {
                pop();
            }

        private:
            void pop() const
            {
                Expression* var = agg.stack.pop();
                agg.stack.top()->result+=var->result;
                delete var;
            }

            const AggregatedContext& agg;
        };

        struct do_Literal
        {
            do_Literal (const AggregatedContext& a)
                : agg(a)
            {
            }

            void operator() (posIterator start, posIterator end) const
            {
                copy (agg.stack.top()->result,start,end);
            }

        private:
            const AggregatedContext& agg;
        };

        struct do_StartFunc
        {
            do_StartFunc (const AggregatedContext& a)
                : agg(a)
            {
            }

            void operator() (posIterator, posIterator) const
            {
                agg.stack.push(new Function());
            }

            void operator()(const QChar)  const
            {
                agg.stack.push(new Function());
            }

        private:
            const AggregatedContext& agg;
        };

        struct do_SetFuncName
        {
            do_SetFuncName (const AggregatedContext& a)
                : agg(a)
            {
            }

            void operator() (posIterator start, posIterator end) const
            {
                Function* f=agg.top_cast<Function*>(start);
                copy (f->funcName, ++start,end);
                f->m_Start=start;
                f->m_End=end;
            }

        private:
            const AggregatedContext& agg;
        };

        struct do_SetLetters
        {
            do_SetLetters (const AggregatedContext& a)
                : agg(a)
            {
            }

            void operator() (posIterator start, posIterator end) const
            {
                copy (agg.stack.top()->result,start,end);
            }

        private:
            const AggregatedContext& agg;
        };

        struct do_AddArg
        {
            do_AddArg (const AggregatedContext& a)
                : agg(a)
            {
            }

            void operator() (posIterator start, posIterator) const
            {
                auto_ptr<Expression> ex(agg.stack.pop());
                agg.top_cast<Function*>(start)->args.append(ex->result);
            }

        private:
            const AggregatedContext& agg;
        };

        struct do_EvalFunc
        {
            do_EvalFunc (const AggregatedContext& a)
                : agg(a)
            {
            }

            void operator() (posIterator start, posIterator) const
            {
                Function* func=agg.top_cast<Function*>(start);

                if (func->funcName!="") {
                    func->exec(agg,start);
                }
            }

        private:
            const AggregatedContext& agg;
        };

        struct do_StartRange
        {
            do_StartRange (const AggregatedContext& a)
                : agg(a)
            {
            }

            void operator() (const QChar) const
            {
                QString org = agg.stack.top()->result;
                Range* range= new Range();
                range->reset();
                range->org=org;
                agg.stack.push(range);
            }

        private:
            const AggregatedContext& agg;
        };

        struct do_SetRange
        {
            do_SetRange  (const AggregatedContext& a)
                : agg(a)
            {
            }

            void operator() (int i) const
            {
                Range* range = agg.top_cast<Range*>();
                if (range->start==-1) range->start=i-1;
                range->end=i-1;
            }

        private:
            const AggregatedContext& agg;
        };

        struct do_SetRangeInterval
        {
            do_SetRangeInterval  (const AggregatedContext& a)
                : agg(a)
            {
            }

            void operator() (const QChar) const
            {
                agg.top_cast<Range*>()->interval=true;
                agg.top_cast<Range*>()->end=-1;
            }

        private:
            const AggregatedContext& agg;
        };

        struct do_EvalRange
        {
            do_EvalRange (const AggregatedContext& a)
                : agg(a)
            {
            }


            void operator() (posIterator start, posIterator) const
            {
                Range* range = agg.top_cast<Range*>(start);
                if (range->start<0) {
                    range->start=range->org.length()+1+range->start;
                }
                if (range->end<0) {
                    range->end=range->org.length()+1+range->end;
                }
                if (range->start<range->org.size()) {
                    if (range->interval && range->start!=range->end)
                    {
                        if (range->start<range->end) {
                            range->result+=range->org.mid(range->start,range->end-range->start+1);
                        }
                        else {
                            QString toAdd=range->org.mid(range->end,range->start-range->end+1);
                            const int s=toAdd.size();
                            for (int i=0;i<s;i++) {
                                QChar tmp = toAdd[i];
                                toAdd[i]=toAdd[s-i];
                                toAdd[s-i]=tmp;
                            }
                            range->result+=toAdd;
                        }
                    }
                    else
                    {
                        range->result+=range->org[range->start];
                    }
                }
                range->reset();
            }

        private:
            const AggregatedContext& agg;
        };

        struct do_PopRange
        {
            do_PopRange  (const AggregatedContext& a)
                : agg(a)
            {
            }

            void operator() (const QChar) const
            {
                Expression* var = agg.stack.pop();
                agg.stack.top()->result=var->result;
                delete var;
            }

        private:
            const AggregatedContext& agg;
        };

        class ExpressionParser: public grammar<ExpressionParser>
        {
        private:

        public:
            AggregatedContext agg;

            explicit ExpressionParser (ParserContext* _context, const QVariant& _localContext)
                : agg(*_context,_localContext)
            {
            }

            /**
        Defines a parser which simply stops parsing with an error message.
        */
            struct error_report_parser {
                const QString errormsg;
                const QString eol_errormsg;

                error_report_parser (const QString& errormsg_, const QString& eol_errormsg_=""):
                        errormsg(errormsg_),
                        eol_errormsg(eol_errormsg_)
                {}

                typedef nil_t result_t;

                template <typename ScannerT>
                        int operator()(ScannerT const& scan, result_t& /*result*/) const
                {
                    QString msg=errormsg;
                    if (scan.at_end() && !eol_errormsg.isEmpty()) {
                        msg = eol_errormsg;
                    }
                    UnknownProblem* p = new UnknownProblem(msg);
                    throw_ (scan.first, (InputProblem*)p);
                    return -1;
                }
            };
            typedef functor_parser<error_report_parser> error_report_p;

            template <typename ScannerT>
                    struct definition
            {
                definition(ExpressionParser const& self)
                {
                    // An expression is always evaluated on an own stack element as an following
                    // range might reduce the output. As a new expression shall only start when
                    // really input is available "eps_p(anychar_p)" does enforce that the following
                    // function is just called when input is available.
                    formular = ch_p('=') >> expression;

                    expression = +( eps_p(anychar_p-ch_p(')')-',')[do_StartExpression(self.agg)]
                                    >> (string | func) >> !range
                                    >> eps_p[do_Pop(self.agg)]
                                    );

                    error_report_p missing_closing_exp (i18n("Expression started but not closed with a ')'"));

                    func = (funcname[do_StartFunc(self.agg)][do_SetFuncName(self.agg)] >> '('
                            | ch_p('(')[do_StartFunc(self.agg)])
                           >> expression
                           >> *( ch_p(',')[do_StartExpression(self.agg)] >> expression[do_AddArg(self.agg)] )
                           >> ( ch_p(')') | missing_closing_exp)
                           >> eps_p[do_EvalFunc(self.agg)][do_Pop(self.agg)];

                    string =     literal
                                 | variable
                                 | letters;

                    chset<QChar> letter_p("0-9a-zA-Z_/+* ");
                    // Workaround to add a '-' character, is otherwise threaded as a range
                    letter_p|=chset<QChar>('-');
                    letter_p|=chset<QChar>('.');

                    letters = (+letter_p)[do_SetLetters(self.agg)];

                    /* Defines a literal. A literal begins with an " and ends also with one.
                       In the middle might be arbitrary text. Escaping is not possible, the first
                       " will close the literal */
                    error_report_p missing_eos (i18n("Missing a closing quotation (\")"));
                    literal = '"' >> lexeme_d[*(anychar_p-'"')][do_Literal(self.agg)]
                              >> (ch_p('"') | missing_eos);

                    variable = (ch_p('$')| ch_p('%'))[do_StartVar(self.agg)][do_SetVarType(self.agg)]
                               >> ((ch_p('{')
                                    >> (*(alnum_p | '_' | ' ' | '#' | '$' | '%' | '&'))[do_SetVarName(self.agg)][do_Pop(self.agg)]
                                    >> ch_p('}')
                                    ) |
                                   (*(alnum_p | '_' ))[do_SetVarName(self.agg)][do_Pop(self.agg)]);

                    funcname = ch_p('&') >> alpha_p >> *alnum_p;

                    error_report_p missing_range_bracket(i18n("Expected a closing bracket (']'): "));

                    range = ch_p('[')[do_StartRange(self.agg)]
                            >> range_item[do_EvalRange(self.agg)] % ','
                            >> (ch_p(']')[do_PopRange(self.agg)] | missing_range_bracket);

                    range_item = int_p[do_SetRange(self.agg)]
                                 >> !((ch_p('-'))[do_SetRangeInterval(self.agg)]
                                      >> !int_p[do_SetRange(self.agg)]);

                }

                rule<ScannerT> formular, expression, string, literal, variable, func, funcname, range, range_item, letters;

                rule<ScannerT> const& start() const { return formular; }
            };

            QString eval (const QString& input)
            {
                // Clear stack;
                while (!agg.stack.empty()) delete agg.stack.pop();
                agg.stack.push(new Expression());
                agg.m_Input=input;
                agg.position = posIterator(input.begin(), input.end());
                try {
                    parse_info<posIterator> info =
                            boost::spirit::classic::parse(agg.position,
                                                          posIterator(),
                                                          *this/*, space_p*/);

                    if (info.full) {
                        Expression* result = agg.stack.pop();
                        QString r=result->result;
                        delete result;
                        return r;
                    }
                    else {
                        InputProblem* problem = new UnknownProblem(i18n("Parse error at character %1",info.stop.get_position().column+1));
                        throw_ (info.stop,problem);
                        return "="+input;
                    }
                }
                catch (parser_error<Texel::InputProblem*, posIterator> e)
                {
                    InputProblem* ip=e.descriptor;
                    // remove one from column as array start counting at zero!
                    ip->setLocator (agg.getMarkedLine(e.where.get_position().column-1));
                    throw ip;
                }
            }
        };
    }

    QStringList ParserContext::availableFunctions ()
    {
        return Function::availableFunctions();
    }

    QString ParserContext::eval (const QString& input, const QVariant& context)
    {
        Function::init();
        ExpressionParser p(this, context);
        return p.eval(input);
    }

    ParserContext::~ParserContext ()
    {
    }

 /********************************************************************************************************
  *
  * Searchpattern parser
  *
  * This parser is needed for the first replacement step to evaluate the searchpatterns. It does replace
  * the variable names by QRegExp regular expressions. The next step will then utilze the QRegExp class
  * to segment the necessary information. However, as variable names might get complex due to possible
  * spaces in them and as the source regular expression might already contain captions these must be
  * detected as well. The result is quite complex, therefore build a parser.
  *
  *******************************************************************************************************/

    namespace SearchPatternReplacementPrivate
    {

        struct SearchPatternReplacementParserContext
        {
            void clear ()
            {
                vartype=None;
                captionDepth=0;
                variables.clear();
                result.clear();
            }

            SearchPatternReplacementParserContext ()
                : vartype (None),
                captionDepth (0)
            {
            }

            enum VarType { None, String, Number };

            VarType vartype;
            int captionDepth;
            QList<QPair<int, QString> > variables;
            QString result;
            position_iterator<QString::const_iterator> position;

            QList<int> source;
        };

        struct do_EnterCap
        {
            do_EnterCap (SearchPatternReplacementParserContext& context)
                : m_Context(context)
            {
            }

            void operator() (posIterator, posIterator) const
            {
                ++m_Context.captionDepth;
            }

            void operator() (QChar) const
            {
                ++m_Context.captionDepth;
            }

            SearchPatternReplacementParserContext& m_Context;
        };

        struct do_CopyEscapedChar
        {
            do_CopyEscapedChar (SearchPatternReplacementParserContext& context)
                : m_Context(context)
            {
            }

            void operator() (posIterator start, posIterator end) const
            {
                copy (m_Context.result,++start,end);
            }

            SearchPatternReplacementParserContext& m_Context;
        };

        struct do_Copy
        {
            do_Copy (SearchPatternReplacementParserContext& context)
                : m_Context(context)
            {
            }

            void operator() (posIterator start, posIterator end) const
            {
                copy (m_Context.result,start,end);
            }

            void operator() (QChar c) const
            {
                m_Context.result+=c;
            }

            SearchPatternReplacementParserContext& m_Context;
        };

        struct do_SetVarType
        {
            do_SetVarType (SearchPatternReplacementParserContext& context, SearchPatternReplacementParserContext::VarType type)
                : m_Context(context), m_VarType(type)
            {
            }

            void operator() (QChar) const
            {
                m_Context.vartype=m_VarType;
            }

            SearchPatternReplacementParserContext& m_Context;
            SearchPatternReplacementParserContext::VarType m_VarType;
        };

        struct do_AddVar
        {
            do_AddVar (SearchPatternReplacementParserContext& context)
                : m_Context(context)
            {
            }

            void operator() (posIterator start, posIterator end) const
            {
                QString varname;
                copy (varname,start,end);
                ++m_Context.captionDepth;
                m_Context.variables << QPair<int,QString>(m_Context.captionDepth,varname);
                if (m_Context.vartype==SearchPatternReplacementParserContext::String) {
                    m_Context.result+="(.+)";
                }
                else if (m_Context.vartype==SearchPatternReplacementParserContext::Number) {
                    m_Context.result+="(\\d+)";
                }
                m_Context.vartype=SearchPatternReplacementParserContext::None;
            }

            SearchPatternReplacementParserContext& m_Context;
        };


        struct do_AddSource
        {
            do_AddSource (SearchPatternReplacementParserContext& context)
                : m_Context(context)
            {
            }

            void operator() (posIterator start, posIterator end) const
            {
                QString varname;
                copy (varname,start,end);

                cout << "Source: " << varname << endl;

                const int idx=ColumnManager::instance()->indexByName(varname);
                if (idx!=-1) {
                    m_Context.source << idx;
                }
                else {
                    InputProblem* problem = new UnknownProblem(i18n("Unknown column name %1",varname));
                    throw_ (start,problem);
                }
            }

            SearchPatternReplacementParserContext& m_Context;
        };

        class SearchPatternReplacementParser: public grammar<SearchPatternReplacementParser>
        {
            mutable SearchPatternReplacementParserContext context;

        public:
            template <typename ScannerT>
                    struct definition
            {
                definition(SearchPatternReplacementParser const& self)
                {
                    searchpattern = eps_p(~ch_p('{')) >> pattern |
                           ( str_p("{{") >> *blank_p >> options >> *blank_p >> str_p("}}") >> *blank_p >> pattern);

                    pattern = +(variable | caps | code);

                    caps = ch_p('(')[SearchPatternReplacementPrivate::do_Copy(self.context)]
                           >> ((ch_p('?') >> ch_p(':')) | eps_p[SearchPatternReplacementPrivate::do_EnterCap(self.context)])
                           >> searchpattern
                           >> ch_p(')')[SearchPatternReplacementPrivate::do_Copy(self.context)];

                    variable = (ch_p('$')[SearchPatternReplacementPrivate::do_SetVarType(self.context,SearchPatternReplacementParserContext::String)]
                                | ch_p('#')[SearchPatternReplacementPrivate::do_SetVarType(self.context,SearchPatternReplacementParserContext::Number)])
                               >> ((ch_p('{')
                                    >> (*(alnum_p | '_' | ' ' | '#' | '$' | '%' | '&'))[SearchPatternReplacementPrivate::do_AddVar(self.context)]
                                    >> ch_p('}')
                                    ) |
                                   (+(alnum_p | '_' ))[SearchPatternReplacementPrivate::do_AddVar(self.context)]);

                    code = (str_p("\\$") | str_p("\\#"))[SearchPatternReplacementPrivate::do_CopyEscapedChar(self.context)]
                            | +(anychar_p-ch_p('(')-ch_p(')')-ch_p('$')-ch_p('#'))[SearchPatternReplacementPrivate::do_Copy(self.context)];

                    options=source;

                    source=str_p("source") >> *blank_p >> ch_p('=') >> *blank_p >> (source_var % '+');

                    source_var = (+(alnum_p | '_' | ' ' | '#' | '$' | '%' | '&'))[SearchPatternReplacementPrivate::do_AddSource(self.context)];
                }

                rule<ScannerT> searchpattern, pattern, variable, caps, code, options, source, source_var;

                rule<ScannerT> const& start() const { return searchpattern; }
            };

            QString eval (const QString& input, QList<QPair<int,QString> >& variables, QList<int>& source)
            {
                context.clear();
                context.position = posIterator(input.begin(), input.end());
                try {
                    parse_info<posIterator> info =
                            boost::spirit::classic::parse(context.position,
                                                          posIterator(),
                                                          *this/*, space_p*/);

                    if (info.full) {
                        variables=context.variables;
                        source=context.source;
                        return context.result;
                    }
                    else {
                        InputProblem* problem = new UnknownProblem(i18n("Parse error at character %1",info.stop.get_position().column+1));
                        throw_ (info.stop,problem);
                        return input;
                    }
                }
                catch (parser_error<Texel::InputProblem*, posIterator> e)
                {
                    InputProblem* ip=e.descriptor;
                    // remove one from column as array start counting at zero!
                    //ip->setLocator (agg.getMarkedLine(e.where.get_position().column-1));
                    throw ip;
                }
            }

        };

    }

    void SearchPatternReplacement::process (const QString& text_in, const QString& searchpattern,TagItem* item)
    {
        variables.clear();
        source.clear();

        QString text=text_in; // copy to remove constness

        // Create a regexp were the #<varname> and $<varname> are replaced by regular expressions
        // and store the names in a list
        QList<QPair<int, QString> > vars;
        SearchPatternReplacementPrivate::SearchPatternReplacementParser p;
        const QString escaped=p.eval(searchpattern, vars, source);
        // If a source is set within the parser remove the default and replace
        if (!source.isEmpty()) {
            text="";
            foreach (int i, source) {
                text+=item->data(i);
            }
        }

        // Match resulting regular expression agains the pattern string and build up an variable map
        QRegExp rx("^"+escaped+"$",Qt::CaseSensitive,QRegExp::RegExp2);
        int pos=0;
        while ((pos=rx.indexIn(text,pos))!=-1) {
            for (int i=0;i<vars.size();i++) {
               const QString value = rx.cap(vars[i].first);
               variables[vars[i].second]=value;
               const int col=ColumnManager::instance()->indexByName(vars[i].second);
               if (col!=-1) {
                  item->setDataDirect(col,value,Columns::Column::RoleReevaluate);
               }
            }
            pos+=rx.matchedLength();
        }
    }

}
