/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "phrasedelegate.h"

#include <QKeySequence>
#include <QKeyEvent>
#include <QLabel>

#include <kkeysequencewidget.h>

#include "aid.h"

using namespace Texel;
using namespace std;

namespace {


/** Helper class to record a key sequence for hotkey usage. The class KKeySequenceWidget
    does similar, however is interweaves UI (a QPushButton) with the logic to record and
    compare with existing shortcuts. Furthermore its impossible to add additional logic
    when recording - e.g. as recording is done in a tabewidget cell pressing a cursor key
    will also stop recording.

    Therefore this is a new implementation with the hope to join the efforts lateron.

    This class is also an extension of an QWidget (in this case a QLabel) as just a QWidget
    allows to grab the keyboard and record all key presses. For future extension either choose
    a QWidget itself and make the visible part the only child for this QWidget or use a template
    and extend some QWidget based class to the keyrecorder.
  */

class KeyRecorder: public QLabel
{
public:
    KeyRecorder (QWidget* parent)
    : QLabel(parent),
      m_Recording(false)
    {
        setAutoFillBackground(true);
    }

    virtual ~KeyRecorder()
    {
    }

    void startRecording ()
    {
        grabKeyboard();
        m_Recording=true;
        m_BackupSeq=m_Seq;
        m_Removed=false;
        setFrameShape(QFrame::Box);
        setLineWidth(3);
        updateText();
    }

    void endRecording ()
    {
        m_Recording=false;
        // When nothing was recorded recover old
        if (m_Seq.isEmpty() && !m_Removed) {
            m_Seq=m_BackupSeq;
        }
        releaseKeyboard();
        setFrameShape(QFrame::NoFrame);
        updateText();
    }

    QString getData () const
    {
        return m_Seq.toString(QKeySequence::PortableText);
    }

protected:
    virtual void focusInEvent (QFocusEvent* event)
    {
        startRecording();
        QLabel::focusInEvent(event);
    }

    virtual void focusOutEvent (QFocusEvent* event)
    {
        endRecording();
        QLabel::focusOutEvent(event);
    }

    virtual void keyPressEvent (QKeyEvent* event)
    {
        if (!m_Recording) QLabel::keyPressEvent (event);

        const int key = event->key();
        switch (key) {
            // Keys to ignore
            case Qt::Key_AltGr: //or else we get unicode salad
                return;

            case Qt::Key_Delete:
                m_Seq=QKeySequence();
                m_Removed=true;
                break;

            // Modifiers, indicate a start of a new sequence
            case Qt::Key_Shift:
            case Qt::Key_Control:
            case Qt::Key_Alt:
            case Qt::Key_Meta:
            case Qt::Key_Menu:
                break;

            // Keys which will lead to a end of recording
            case Qt::Key_Escape:
                m_Seq=m_BackupSeq;
            case Qt::Key_Left:
            case Qt::Key_Right:
            case Qt::Key_Up:
            case Qt::Key_Down:
                endRecording();
                QLabel::keyPressEvent (event);
                return;

            // Keys to record
            default:
                m_Seq = QKeySequence(event->modifiers()+event->key());
        }

        updateText();
    }

    virtual void keyReleaseEvent (QKeyEvent* event)
    {
        if (!m_Recording) QLabel::keyReleaseEvent (event);
    }

private:
    void updateText ()
    {
        if (m_Seq.isEmpty()) {
            setAlignment(Qt::AlignHCenter|Qt::AlignVCenter);
            setText("...");
            return;
        }

        setAlignment(Qt::AlignLeft|Qt::AlignVCenter);
        setText (m_Seq.toString());
    }

    QKeySequence m_Seq;
    QKeySequence m_BackupSeq;
    bool m_Recording;
    bool m_Removed;
};
}

PhraseDelegate::PhraseDelegate(QWidget* parent)
: ProblemAwareDelegate(parent)
{
}

PhraseDelegate::~PhraseDelegate()
{
}

QWidget* PhraseDelegate::createEditor (QWidget* parent, const QStyleOptionViewItem&, const QModelIndex&) const
{
    KeyRecorder* editor=new KeyRecorder(parent);
    return editor;
}

#include "phrasedelegate.moc"
