/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "phrasemodel.h"
#include "model/columnmanager.h"
#include "model/texeltagmodel.h"
#include "func/phrases/phraseitem.h"
#include "options/preferences.h"
#include "options/enums.h"
#include "aid.h"
#include "model/mimesupport.h"

#include <klocalizedstring.h>

#include <QXmlStreamReader>
#include <QXmlStreamWriter>
#include <QFile>

using namespace std;

namespace Texel {

    static const char* s_DragMimeType = "phrase::row";

    PhraseModel::PhraseModel(QObject* parent)
        : QAbstractTableModel(parent)
    {
        const QString filename = Preferences::instance()->getReadNamePhrasesFile();
        read(filename);
    }

    PhraseModel::~PhraseModel()
    {
        for (int i=0;i<m_Items.size();i++) delete m_Items[i];
    }

    int PhraseModel::rowCount (const QModelIndex&) const
    {
        return m_Items.size();
    }

    int PhraseModel::columnCount (const QModelIndex&) const
    {
        return PhraseItem::Offset+ColumnManager::instance()->getEditableColumns();
    }

    QVariant PhraseModel::headerData (int section, Qt::Orientation orientation, int role) const
    {
        if (role==Qt::DisplayRole) {
            if (orientation==Qt::Horizontal) {
                switch (section) {
                case 0: return QVariant(i18n("Name"));
                case 1: return QVariant(i18n("Hotkey"));
                case 2: return QVariant(i18n("Comment"));
                default: return ColumnManager::instance()->getName(ColumnManager::instance()->mapFromEditableColumn(section-PhraseItem::Offset));
                }
            }
        }
        return QAbstractItemModel::headerData(section,orientation,role);
    }

    QVariant PhraseModel::data (const QModelIndex& index, int role) const
    {
        const int row = index.row();

        if (index.isValid() && row>=0 && row<m_Items.size()) {
            if (role==Qt::DisplayRole || role==Qt::EditRole) {
                switch (index.column()) {
                case 0: return m_Items[row]->getName();
                case 1: return m_Items[row]->getHotKeyAsString();
                case 2: return m_Items[row]->getDesc();
                default: return m_Items[row]->data(index.column()-PhraseItem::Offset);
                }
            }
            else if (role==Enums::RoleProblem) {
                return m_Items[row]->hasProblem(index.column())?Enums::ProblemError:Enums::ProblemUnknown;
            }
            else if (role==Qt::ToolTipRole) {
                if (m_Items[row]->hasProblem(index.column())) {
                    return m_Items[row]->getProblem(index.column())->getText();
                }
            }
        }

        return QVariant();
    }

    namespace {
        struct NotifyFunctor
        {
            void operator() (PhraseItem* item, bool duplicate, int)
            {
                if (duplicate) {
                    item->setProblem(0,new UnknownProblem(i18n("A phrase with the same name exists already")));
                }
                else item->setProblem(0,0l);
            }
        };
    }

    bool PhraseModel::setData (const QModelIndex& index, const QVariant& value, int role)
    {
        const QString string = value.toString();
        const int row = index.row();

        if (index.isValid() && row>=0 && row<m_Items.size()) {
            if (role==Qt::EditRole) {
                switch (index.column()) {
                case 0:
                    {
                        m_UniqueNames.alter(m_Items[row], string, NotifyFunctor());
                        m_Items[row]->setName(string);
                        break;
                    }
                case 1: m_Items[row]->setHotKeyFromString(string); break;
                case 2: m_Items[row]->setDesc(string);
                default: m_Items[row]->setData(index.column()-PhraseItem::Offset,string); break;
                }
                return true;
            }
        }

        return false;
    }

    void PhraseModel::triggered ()
    {
        QAction* s = dynamic_cast<QAction*>(sender());
        if (s!=0l) {
            emit phraseTriggered (s->data().value<PhraseItem*>());
        }
    }

    Qt::ItemFlags PhraseModel::flags (const QModelIndex& index) const
    {
        if (index.isValid()) {
            return Qt::ItemFlags(Qt::ItemIsSelectable|Qt::ItemIsEditable|Qt::ItemIsEnabled|Qt::ItemIsDragEnabled|Qt::ItemIsDropEnabled);
        }
        return Qt::ItemIsDropEnabled;
    }

    Qt::DropActions PhraseModel::supportedDropActions() const
    {
        return Qt::MoveAction;
    }

    bool PhraseModel::dropMimeData (const QMimeData* data, Qt::DropAction action, int row, int, const QModelIndex& parent)
    {
        if (!data->hasFormat(MimeSupport::getMimeTypeForSubType(s_DragMimeType)) || action!=Qt::MoveAction) return false;

        QList<int> rows = MimeSupport::getRowList(data,s_DragMimeType);

        if (parent.isValid()) {
            if (row==-1) row=parent.row();
            int alreadyMoved=0;
            foreach (int r, rows) {
                // when a row before the current row is removed it is removed first from the list
                // and then inserted at a later stage. Correct the insertion index by one to
                // honour the removed index.
                const int offset = (r<row)?1:0;
                m_Items.move (r-alreadyMoved,row-offset);
                if (r<row) {
                    ++alreadyMoved;
                }
                else {
                    ++row;
                }
            }
        }
        else {
            int alreadyMoved=0;
            foreach (int r, rows) {
                m_Items.append (m_Items.takeAt(r-alreadyMoved));
                ++alreadyMoved;
            }
        }

        return true;
    }

    QMimeData* PhraseModel::mimeData (const QModelIndexList& indexes) const
    {
        MimeSupport* mime = new MimeSupport();
        mime->setRowList (indexes,s_DragMimeType);
        return mime;
    }

    QStringList PhraseModel::mimeTypes () const
    {
        QStringList l;
        l << s_DragMimeType;
        return l;
    }

    PhraseItem* PhraseModel::addRow()
    {
        PhraseItem* newItem=new PhraseItem();
        beginInsertRows(QModelIndex(),m_Items.size(),m_Items.size());
        m_Items << newItem;
        endInsertRows();
        return newItem;
    }

    int PhraseModel::addRow (QAbstractItemModel* model, const QModelIndexList& indexes)
    {
        PhraseItem* newItem = addRow();
        if (indexes.size()>0) {
            for (int i=0;i<indexes.size();i++) {
                const int col=indexes[i].column();
                newItem->setData(ColumnManager::instance()->mapToEditableColumn(col),
                                 model->data(indexes[i],TagModel::RoleFormular).toString());
            }
        }
        return m_Items.size()-1;
    }

    void PhraseModel::rebuildMenu(QMenu* menu)
    {
        foreach (PhraseItem* item, m_Items) item->addToMenu(menu,this,SLOT(triggered()));
    }

    void PhraseModel::write (const QString& filename)
    {
        QFile file(filename);
        file.open(QIODevice::WriteOnly|QIODevice::Truncate);
        QXmlStreamWriter writer(&file);
        writer.setAutoFormatting(true);
        writer.writeStartDocument("1.0");
        write(writer);
        writer.writeEndDocument();
        file.close();
    }

    void PhraseModel::write (QXmlStreamWriter& writer)
    {
        writer.writeStartElement("phrasecollection");
        for (int i=0; i<m_Items.size(); i++) {
            writer.writeStartElement("phrase");
            writeItem (writer, i);
            writer.writeEndElement();
        }
        writer.writeEndElement();
    }

    void PhraseModel::read (const QString& filename)
    {
        QFile file(filename);
        if (file.exists()) {
            file.open(QIODevice::ReadOnly);
            QXmlStreamReader reader(&file);

            reader.readNext(); // startDocument
            read (reader);
            file.close();
        }
    }

    void PhraseModel::read (QXmlStreamReader& reader)
    {
        reader.readNext();
        if (reader.isStartElement() && reader.name()=="phrasecollection") {
            while (!reader.atEnd()) {
                reader.readNext();
                if (reader.isStartElement() && reader.name()=="phrase") {
                    PhraseItem* item = new PhraseItem();
                    m_Items.append(item);
                    readItem (reader, m_Items.size()-1);
                }
            }
        }
    }

    void PhraseModel::readItem (QXmlStreamReader& reader, int row)
    {
        QStringRef current;
        while (!reader.atEnd()) {
            reader.readNext();
            if (reader.isStartElement()) {
                current=reader.name();
            }
            else if (reader.isEndElement()) {
                if (current.isNull()) {
                    return;
                }
                current=QStringRef();
            }
            else if (reader.isCharacters()) {
                if (current=="name") {
                    setData(index(row,0),reader.text().toString(),Qt::EditRole);
                }
                else if (current=="hotkey") {
                    setData(index(row,1),reader.text().toString(),Qt::EditRole);
                }
                else if (current=="comment") {
                    setData(index(row,2),reader.text().toString(),Qt::EditRole);
                }
                else {
                    const int column = ColumnManager::instance()->indexByXmlName(current.toString());
                    if (column!=-1) {
                        const int mapped = ColumnManager::instance()->mapToEditableColumn(column);
                        if (mapped!=-1) setData(index(row, mapped+PhraseItem::Offset),reader.text().toString(),Qt::EditRole);
                    }
                }
            }
        }
    }

    void PhraseModel::writeItem (QXmlStreamWriter& writer, int row) const
    {
        writer.writeTextElement ("name",data(index(row, 0),Qt::DisplayRole).toString());
        writer.writeTextElement ("hotkey",data(index(row, 1),Qt::DisplayRole).toString());
        writer.writeTextElement ("comment",data(index(row, 2),Qt::DisplayRole).toString());
        for(int i=0;i<ColumnManager::instance()->getEditableColumns();i++) {
            const int column = ColumnManager::instance()->mapFromEditableColumn(i);
            const QString tag= ColumnManager::instance()->getXmlName(column);
            writer.writeTextElement (tag, data(index(row, i+PhraseItem::Offset),Qt::DisplayRole).toString());
        }
    }

    bool PhraseModel::removeRows (int row, int count, const QModelIndex& parent)
    {
        beginRemoveRows(parent,row,row+count-1);
        for (int i=0;i<count;i++) {
            m_UniqueNames.remove(m_Items[row],NotifyFunctor());
            //m_UniqueHotKeys.remove(m_Items[row],...)
            delete m_Items[row];
            m_Items.removeAt(row);
        }
        endRemoveRows();
        return true;
    }

}

#include "phrasemodel.moc"
