/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "func/player/enqueuemodel.h"
#include "model/texeltagmodel.h"
#include "model/mimesupport.h"

#include <QApplication>

#include <kfiledialog.h>

namespace Texel {
    namespace Player {

    static const char* s_DragMimeType = "player::rowindex";

    TagItem* EnqueueModel::getRowByFilename (const QString& name) const
    {
        return m_Queue.front()->getModel()->getRowByOrgFilename(name);
    }

    EnqueueModel::EnqueueModel (QObject* parent)
        : QAbstractListModel(parent),
          m_PlayedIndex(-1)
    {
    }

    EnqueueModel::~EnqueueModel ()
    {
    }

    void EnqueueModel::append (TagItem* item)
    {
        beginInsertRows (QModelIndex(), m_Queue.size(), m_Queue.size());
        m_Queue.append(item);
        endInsertRows();
    }

    void EnqueueModel::prepend (TagItem* item)
    {
        beginInsertRows (QModelIndex(), 0, 0);
        m_Queue.prepend(item);
        endInsertRows();
    }

    int EnqueueModel::getNext () const
    {
        if (m_PlayedIndex+1==m_Queue.size()) return -1;
        return m_PlayedIndex+1;
    }

    int EnqueueModel::getPrev () const
    {
        if (m_PlayedIndex<=0) return -1;
        return m_PlayedIndex-1;
    }

    QString EnqueueModel::setSong (int i)
    {
        static const QString empty;
        if (i<0 || i>=m_Queue.size()) return empty;

        const int old = m_PlayedIndex;
        m_PlayedIndex=i;
        const QModelIndex oi=createIndex(old,0);
        const QModelIndex ni=createIndex(i,0);
        dataChanged (oi,oi);
        dataChanged (ni,ni);
        return m_Queue[m_PlayedIndex]->getAbsolutFilename();
    }

    void EnqueueModel::songRequested (const QModelIndex& index)
    {
        if (index.isValid()) {
            emit songRequested (index.row(),true);
        }
    }

    int EnqueueModel::rowCount (const QModelIndex& ) const
    {
        return m_Queue.size();
    }

    QVariant EnqueueModel::data (const QModelIndex& index, int role) const
    {
        using namespace Columns;
        TagItem* item = m_Queue[index.row()];

        if (role==Qt::DisplayRole) {
            return "<html><h2>"+item->data(ColumnManager::Artist,Column::RoleDisplay)+"</h2>"
                    "<table width=\"100%\"><tr>"
                    "<td><i>"+item->data(ColumnManager::Title,Column::RoleDisplay)+"</i></td>"
                    "<td align=\"right\">["+item->data(ColumnManager::Length,Column::RoleDisplay)+"]</td>"
                    "</tr></table></html>";
        }
        if (role==Qt::DecorationRole) {
            QPixmap px;

            if (item->supportsCoverImages()) {
                px=QPixmap::fromImage(item->getDefaultCoverImage());
            }
            if (px.isNull()) {
                px=KIconLoader::global()->loadIcon("music-file",KIconLoader::Panel,KIconLoader::SizeMedium);
            }
            return px.scaledToWidth(32);
        }
        if (role==Qt::ToolTipRole) {
            return item->getAbsolutFilename();
        }
        if (role==Qt::BackgroundRole && index.row()==m_PlayedIndex) {
            return QApplication::palette().midlight();
        }
        return QVariant();
    }

    Qt::ItemFlags EnqueueModel::flags (const QModelIndex& ) const
    {
        return Qt::ItemFlags(Qt::ItemIsDragEnabled|Qt::ItemIsDropEnabled|Qt::ItemIsEnabled|Qt::ItemIsSelectable);
    }

    void EnqueueModel::clear ()
    {
        beginRemoveRows (QModelIndex(), 0, m_Queue.size());
        m_Queue.clear();
        m_PlayedIndex=-1;
        endRemoveRows();
    }

    bool EnqueueModel::dropMimeData (const QMimeData* data, Qt::DropAction action, int row, int, const QModelIndex& parent)
    {
        if (MimeSupport::hasItems(data)) {
            row=(parent.isValid())?parent.row():m_Queue.size();
            const TagItem::List items = MimeSupport::getItemList(data);
            beginInsertRows(parent,row,row+items.size()-1);
            foreach (TagItem* item, items) {
                m_Queue.insert(row++,item);
            }
            endInsertRows();
            return true;
        }

        if (data->hasFormat(MimeSupport::getMimeTypeForSubType(s_DragMimeType)) && action==Qt::MoveAction)  {
            QList<int> rows = MimeSupport::getRowList(data,s_DragMimeType);
            if (parent.isValid()) {
                if (row==-1) row=parent.row();
                int alreadyMoved=0;
                foreach (int r, rows) {
                    // when a row before the current row is removed it is removed first from the list
                    // and then inserted at a later stage. Correct the insertion index by one to
                    // honour the removed index.
                    const int offset = (r<row)?1:0;
                    m_Queue.move (r-alreadyMoved,row-offset);
                    if (r<row) {
                        ++alreadyMoved;
                    }
                    else {
                        ++row;
                    }
                }
            }
            else {
                int alreadyMoved=0;
                foreach (int r, rows) {
                    m_Queue.append (m_Queue.takeAt(r-alreadyMoved));
                    ++alreadyMoved;
                }
            }
            return true;

        }
        return false;
    }

    QMimeData* EnqueueModel::mimeData (const QModelIndexList& indexes) const
    {
        MimeSupport* mime = new MimeSupport();
        mime->setRowList(indexes,s_DragMimeType);
        TagItem::List items;
        foreach (QModelIndex i, indexes) {
            items << m_Queue[i.row()];
        }
        mime->setItems (items);

        return mime;
    }

    Qt::DropActions EnqueueModel::supportedDropActions() const
    {
        return Qt::DropActions(Qt::MoveAction|Qt::CopyAction);
    }

    QStringList EnqueueModel::mimeTypes () const
    {
        QStringList l;
        l << MimeSupport::getMimeTypeForSubType(s_DragMimeType)
          << MimeSupport::getMimeTypeForSubType(MimeSupport::s_TagItemSubFormat);
        return l;
    }

    void EnqueueModel::removeRow (int row)
    {
        beginRemoveRows (QModelIndex(), row,row);
        m_Queue.removeAt(row);
        endRemoveRows();
    }

    void EnqueueModel::createPlaylist()
    {
        if (m_Queue.size()>0) {
            TagModel* _m = dynamic_cast<TagModel*>(m_Queue.front()->getModel());
            if (_m!=0l) {
                // Calc common path
                DirTree tree;
                foreach (TagItem* item, m_Queue) {
                    tree.store(item->getAbsolutFilename(),0l);
                }
                KUrl dir = tree.getCommonPath();
                QString filename = KFileDialog::getSaveFileName(dir,"*.m3u",0l,i18n("Playlist filename"));
                if (filename!="") {
                    M3U::List playlist(_m);
                    playlist.setTracks(m_Queue);
                    playlist.save(filename);

                    _m->addPlaylist(playlist);
                }
            }
        }
    }

}
}

#include "func/player/enqueuemodel.moc"
