/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "func/player/player.h"
#include "func/player/enqueuelist.h"

#include "util/slowclicktoolbutton.h"
#include "model/texeltagmodel.h"
#include "view/texeltablewidget.h"

#include <Phonon/VolumeSlider>
#include <Phonon/SeekSlider>
#include <Phonon/AudioOutput>

#include <QAction>
#include <QAbstractProxyModel>
#include <QAbstractItemView>
#include <QWidgetAction>
#include <QMenu>

#include <knotification.h>
#include <kstatusbar.h>

using namespace Phonon;

namespace Texel {
    namespace Player {

    Player::Player ()
        : QObject (0l),
          m_Queue(this),
          m_DisplayElapsedTime(true),
          m_TimeLabel(0l),
          m_player (new MediaObject (this)),
          m_ItemView (0l),
          m_Menu(0l),
          m_Notification(0l)
    {
        connect (m_player, SIGNAL(aboutToFinish()), this, SLOT(enqueueNextSong()));
        connect (m_player, SIGNAL(currentSourceChanged (const Phonon::MediaSource &)),
                 this, SLOT(songChanged(const Phonon::MediaSource &)));
        connect (m_player, SIGNAL(stateChanged(Phonon::State, Phonon::State)),
                 this, SLOT(onPhononStateChange(Phonon::State,Phonon::State)));

        // Creation of actions
        actionPlay = new QAction(this);
        actionPlay->setText (i18n("Play song"));
        actionPlay->setObjectName(QString::fromUtf8("actionPlay"));
        actionPlay->setCheckable(true);
        actionPlay->setToolTip(i18n("Play or pause the playlist. When the playlist is empty the\n"
                                    "current selection is enqueued and played. Long pressing\n"
                                    "does enqueue the current selection and start playing there."));
        actionPlay->setIcon(KIconLoader::global()->loadIcon("media-playback-start",KIconLoader::Toolbar));
        actionPlay->setShortcut(QKeySequence("Shift+F10"));

        actionAddAndPlay = new QAction(this);
        actionAddAndPlay->setText (i18n("Enqueue and play song"));
        actionAddAndPlay->setObjectName(QString::fromUtf8("actionEnqueuePlay"));
        actionAddAndPlay->setCheckable(false);
        actionAddAndPlay->setToolTip(i18n("Enqueue in playlist and play this song immediately."));
        actionAddAndPlay->setShortcut(QKeySequence("F10"));

        actionStop = new QAction(this);
        actionStop->setObjectName(QString::fromUtf8("actionStop"));
        actionStop->setToolTip (i18n("Stops current replay. A long press stops replay and clears the playlist"));
        actionStop->setIcon(KIconLoader::global()->loadIcon("media-playback-stop",KIconLoader::Toolbar));
        actionStop->setText(i18n("Stop replay"));
        actionStop->setShortcut(QKeySequence("ALT+F10"));
        actionStop->setDisabled(true);

        actionPause = new QAction(this);
        actionPause->setObjectName(QString::fromUtf8("actionPause"));
        actionPause->setToolTip (i18n("Pauses current replay."));
        actionPause->setIcon(KIconLoader::global()->loadIcon("media-playback-pause",KIconLoader::Toolbar));
        actionPause->setText(i18n("Pause replay"));
        actionPause->setShortcut(QKeySequence("F11"));
        actionPause->setDisabled(true);

        actionNext = new QAction(this);
        actionNext->setObjectName(QString::fromUtf8("actionPrev"));
        actionNext->setToolTip (i18n("Start playing next song"));
        actionNext->setIcon(KIconLoader::global()->loadIcon("media-skip-forward",KIconLoader::Toolbar));
        actionNext->setText(i18n("Play next"));
        actionNext->setShortcut(QKeySequence("F9"));
        actionNext->setDisabled(true);

        actionPrev = new QAction(this);
        actionPrev->setObjectName(QString::fromUtf8("actionPrev"));
        actionPrev->setToolTip (i18n("Start playing previous song"));
        actionPrev->setIcon(KIconLoader::global()->loadIcon("media-skip-backward",KIconLoader::Toolbar));
        actionPrev->setText(i18n("Play previous"));
        actionPrev->setShortcut(QKeySequence("F8"));
        actionPrev->setDisabled(true);

        actionEnqueue = new QAction(this);
        actionEnqueue->setObjectName(QString::fromUtf8("actionEnqueue"));
        actionEnqueue->setToolTip(i18n("Enqueues the current selection without disturbing the current replay.\n"
                                       "Long pressing opens a play list dialog"));
        actionEnqueue->setIcon(KIconLoader::global()->loadIcon("list-add",KIconLoader::Toolbar));
        actionEnqueue->setText(i18n("Enqueue selection"));
        actionEnqueue->setShortcut(QKeySequence("CTRL+F10"));

        actionClearList = new QAction(this);
        actionClearList->setText ("Clear playlist");
        connect (actionClearList, SIGNAL(triggered()), &m_Queue, SLOT(clear()));
        connect (actionClearList, SIGNAL(triggered()), m_player, SLOT(stop()));

        connect (actionEnqueue, SIGNAL(triggered()), this, SLOT(enqueueSong()));
        connect (actionAddAndPlay, SIGNAL(triggered()), this, SLOT (addAndPlaySong()));
        connect (actionPlay, SIGNAL(triggered()), this, SLOT (onPlayPressed()));
        connect (actionClearList, SIGNAL(triggered()), actionClearList, SLOT (trigger()));
        connect (actionStop, SIGNAL(triggered()), m_player, SLOT(stop()));
        connect (actionPause, SIGNAL(triggered()), this, SLOT(pause()));
        connect (actionNext, SIGNAL(triggered()), this, SLOT(nextSong()));
        connect (actionPrev, SIGNAL(triggered()), this, SLOT(prevSong()));

        out=new AudioOutput(Phonon::MusicCategory);
        createPath (m_player, out);
    }

    void Player::setItemView (QAbstractItemView* itemview)
    {
        m_ItemView=itemview;
        QAbstractProxyModel* proxy=dynamic_cast<QAbstractProxyModel*>(itemview->model());
        TagModel* model = dynamic_cast<TagModel*>(proxy->sourceModel());
        connect (model, SIGNAL(isEmpty(bool)), actionPlay, SLOT(setDisabled(bool)));
        connect (model, SIGNAL(isEmpty(bool)), actionEnqueue, SLOT(setDisabled(bool)));
        connect (model, SIGNAL(isEmpty(bool)), actionClearList, SLOT(setDisabled(bool)));
        connect (model, SIGNAL(isEmpty(bool)), actionAddAndPlay, SLOT(setDisabled(bool)));
    }

    Player::~Player ()
    {
        delete out;
        delete m_Menu;
    }

    void Player::onPhononStateChange (Phonon::State newstate, Phonon::State oldstate)
    {
        switch (newstate) {
        case Phonon::PlayingState:
            actionPlay->setChecked(true);
            if (oldstate==Phonon::PausedState) {
                m_PlayButton->setIcon(KIconLoader::global()->loadIcon("media-playback-start",KIconLoader::Toolbar));
            }
            actionStop->setDisabled(false);
            actionPause->setDisabled(false);
            break;
        case Phonon::PausedState:
            actionPlay->setChecked(true);
            m_PlayButton->setIcon(KIconLoader::global()->loadIcon("media-playback-pause",KIconLoader::Toolbar));
            actionStop->setDisabled(false);
            actionPause->setDisabled(true);
            break;
        case Phonon::StoppedState:
            actionPlay->setChecked(false);
            if (oldstate==Phonon::PausedState) {
                m_PlayButton->setIcon(KIconLoader::global()->loadIcon("media-playback-start",KIconLoader::Toolbar));
            }
            actionStop->setDisabled(true);
            actionPause->setDisabled(true);

            if (m_TimeLabel!=0l) {
                m_TimeLabel->setToolTip("");
                emit timeElapsed("");
            }
            if (m_Notification!=0l) {
                m_Notification->close();
            }
            emit playing (0l);
            break;
        case Phonon::LoadingState:
        case Phonon::BufferingState:
        case Phonon::ErrorState:
            break;
        }
    }

    void Player::initToolBar (KToolBar* toolbar)
    {
        m_PlayButton = new SlowClickToolButton(toolbar);
        m_PlayButton->setDefaultAction (actionPlay);
        m_PlayButton->setSlowClickAction (actionAddAndPlay);
        toolbar->addWidget (m_PlayButton);

        m_StopButton = new SlowClickToolButton(toolbar);
        m_StopButton->setDefaultAction(actionStop);
        m_StopButton->setSlowClickAction (actionClearList);
        toolbar->addWidget (m_StopButton);

        m_EnqueueButton=new QToolButton(toolbar);
        m_EnqueueButton->setDefaultAction(actionEnqueue);
        toolbar->addWidget (m_EnqueueButton);
        QMenu* tmpmenu = new QMenu(toolbar);
        QWidgetAction* actionwidget = new QWidgetAction(tmpmenu);
        EnqueueList* list = new EnqueueList(m_ItemView);
        actionwidget->setDefaultWidget(list);
        connect (list, SIGNAL(close()), tmpmenu, SLOT(hide()));
        connect (list, SIGNAL(openPlayDock(bool)), this, SIGNAL(openPlayDock(bool)));
        tmpmenu->addAction(actionwidget);
        m_EnqueueButton->setMenu(tmpmenu);
        connect (&m_Queue, SIGNAL(songRequested (int,bool)), this, SLOT(playSong (int,bool)));

        toolbar->addWidget (new VolumeSlider(out,toolbar));
        toolbar->addWidget (new SeekSlider(m_player, toolbar));
    }

    void Player::initCompleteToolBar(KToolBar *parent)
    {
        QToolButton* prev = new QToolButton(parent);
        prev->setDefaultAction (actionPrev);
        parent->addWidget (prev);
        parent->addSeparator();

        QToolButton* play = new QToolButton(parent);
        play->setDefaultAction (actionPlay);
        parent->addWidget (play);

        QToolButton* pause = new QToolButton(parent);
        pause->setDefaultAction (actionPause);
        parent->addWidget (pause);

        QToolButton* stop = new QToolButton(parent);
        stop->setDefaultAction (actionStop);
        parent->addWidget (stop);
        parent->addSeparator();

        QToolButton* next = new QToolButton(parent);
        next->setDefaultAction (actionNext);
        parent->addWidget (next);

        parent->addWidget (new VolumeSlider(out,parent));
    }

    QMenu* Player::createMenu ()
    {
        m_Menu = new QMenu("Player");
        m_Menu->setIcon(KIconLoader::global()->loadIcon("multimedia-player",KIconLoader::Toolbar));
        m_Menu->addAction (actionPlay);
        m_Menu->addAction (actionAddAndPlay);
        m_Menu->addAction (actionStop);
        m_Menu->addAction (actionEnqueue);
        m_Menu->addAction (actionClearList);

        return m_Menu;
    }

    void Player::initStatusBar (KStatusBar* statusbar)
    {
        // Create a text icon in the status bar to display current elapsed time
        m_TimeLabel=new ClickableLabel (statusbar);
        m_TimeLabel->setFixedWidth(QFontMetrics(statusbar->font()).width("<a href='test'>00:00</a>"));
        m_TimeLabel->setToolTip(i18n("Displays elapsed time of current song. Clicking toggles to remaining time."));
        statusbar->addPermanentWidget(m_TimeLabel);
        m_player->setTickInterval(1000);
        m_DisplayElapsedTime=true;
        connect (m_player, SIGNAL(tick(qint64)), this, SLOT(displayElapsedTime(qint64)));
        connect (m_TimeLabel, SIGNAL(pressed()), this, SLOT(toggleTimeDisplay()));
        connect (this, SIGNAL(timeElapsed(QString)), m_TimeLabel, SLOT(setText(QString)));
    }

    void Player::displayElapsedTime (qint64 time)
    {
        const Phonon::State state=m_player->state();
        if (state!=Phonon::StoppedState && state!=Phonon::ErrorState && state!=Phonon::LoadingState) {
            QTime t(0,0);
            if (m_DisplayElapsedTime) {
                t=t.addMSecs (time);
            }
            else {
                t=t.addMSecs(m_player->totalTime()-time);
            }
            const QString str = t.toString("m:ss");
            emit timeElapsed(str);
        }
    }

    void Player::nextSong()
    {
        playSong(m_Queue.getNext(),true);
    }

    void Player::prevSong()
    {
        playSong(m_Queue.getPrev(),true);
    }

    void Player::pause ()
    {
        m_player->pause();
    }

    void Player::toggleTimeDisplay ()
    {
        m_DisplayElapsedTime=!m_DisplayElapsedTime;
        displayElapsedTime(m_player->currentTime());
    }

    void Player::addAndPlaySong ()
    {
        int last = m_Queue.rowCount();
        enqueueSong();
        playSong(last,true);
    }

    void Player::onPlayPressed ()
    {
        const Phonon::State state=m_player->state();

        if (state==Phonon::PlayingState) {
            m_player->pause();
        }
        else if (state==Phonon::PausedState) {
            // continue paused play
            m_player->play();
        }
        else if (state==Phonon::StoppedState) {
            if (m_Queue.isEmpty()) {
                enqueueSong();
            }
            int nextSong = m_Queue.getNext();
            // when last song is reached and queue is not empty start again!
            if (nextSong==-1 && !m_Queue.isEmpty()) nextSong=0;
            playSong(nextSong,true);
        }
    }

    void Player::enqueueNextSong ()
    {
        if (!m_Queue.isEmpty()) {
            const int nextSong = m_Queue.getNext();
            if (nextSong!=-1) {
                playSong (nextSong, false);
            }
        }
    }

    void Player::playSong (int i, bool replace)
    {
        const QString& song=m_Queue.setSong(i);
        QFileInfo test(song);
        if (!test.isReadable()) {
            emit error (i18n("File %1 is not readable - song is not played",song));
        }
        else if (!song.isEmpty()) {
            if (m_player->state()==Phonon::PlayingState && !replace) {
                m_player->enqueue (song);
            }
            else {
                m_player->setCurrentSource(song);
                m_player->play();
            }
            actionPrev->setEnabled(i>0);
            actionNext->setEnabled(m_Queue.rowCount()-1>i);
        }
    }

    void Player::songChanged (const Phonon::MediaSource& source)
    {
        const QString song = source.fileName();
        const QString text=i18n("Currently playing: \n%1",song);
        QString label=text;

        m_TimeLabel->setToolTip (text);
        QPixmap icon;
        if (m_Queue.rowCount()>0) {
            TagItem* item = m_Queue.getRowByFilename(song);
            emit playing (item);
            if (item!=0l) {
                if (item->supportsCoverImages()) {
                    icon = QPixmap::fromImage(item->getDefaultCoverImage());
                }
                using namespace Columns;
                label = "<h2>"+item->data(ColumnManager::Title,Column::RoleDisplay)+"</h2><p>"
                        +item->data(ColumnManager::Artist,Column::RoleDisplay)+"</p><p>"
                        +item->data(ColumnManager::Album,Column::RoleDisplay)+"</p>";
            }
        }
        if (icon.isNull()) {
            icon=KIconLoader::global()->loadIcon("music-file",KIconLoader::Panel,KIconLoader::SizeMedium);
        }

        if (m_Notification!=0l) {
            m_Notification->close();
        }

        m_Notification = new KNotification("playTrack");
        connect( m_Notification, SIGNAL(closed()), this, SLOT(notificationClosed()) );
        m_Notification->setPixmap(icon.scaled(64,64,Qt::KeepAspectRatio));
        m_Notification->setTitle(i18n("Playing:"));
        m_Notification->setText(label);
        m_Notification->sendEvent();
    }

    void Player::notificationClosed ()
    {
        if (sender()==m_Notification) {
            m_Notification=0l;
        }
    }

    void Player::enqueueSong ()
    {
        Texel::TableWidget* view = dynamic_cast<Texel::TableWidget*>(m_ItemView);
        TagItem::List items = view->getSelectedRowsSorted(false);

        if (items.isEmpty()) {
            const Texel::TagModelProxy* model=dynamic_cast<Texel::TagModelProxy*>(m_ItemView->model());
            if (m_ItemView->currentIndex().isValid()) {
                const int row = m_ItemView->currentIndex().row();
                m_Queue.append(model->getRow(row));
            }
        }
        else {
            foreach (TagItem* item, items) {
                m_Queue.append(item);
            }
        }
    }

}
}

#include "player.moc"
