/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "searchwidget.h"
#include "ui_searchwidget.h"
#include "options/preferences.h"

#include <QTableWidget>
#include <QMenu>
#include <QTimer>
#include <QtConcurrentMap>

#include <kstandardaction.h>
#include <kaction.h>
#include <kcompletionbox.h>

#include "aid.h"

using namespace std;

namespace Texel {

    namespace Internal
    {
        struct Comparator
        {
            Comparator (): model(0l) {}

            virtual bool test (const QString& text, const QModelIndex& idx) const=0;
            virtual QString replace (const QString& text, const QString& replacement) const=0;
            virtual ~Comparator() {}

            static Comparator* createComparator (const QString& pattern,
                                                 Qt::CaseSensitivity cs, bool wholeword, bool regexp);

            void setModel (TagModelProxy* model_)
            {
                model=model_;
            }

            virtual void setPattern (const QString& pattern)=0;

            TagModelProxy* model;
        };

        struct ContainsComparator: public Comparator
        {
            ContainsComparator (const QString& pattern, Qt::CaseSensitivity cs_)
                : m_Pattern (pattern),
                cs (cs_)
            {
            }

            virtual ~ContainsComparator() {}

            bool test (const QString& text, const QModelIndex& idx) const
            {
                bool found=false;
                int pos=0;

                while ((pos = text.indexOf(m_Pattern,pos,cs))!=-1) {
                    found=true;
                    if (model!=0l) {
                        model->setMark(model->getRow(idx.row()),idx.column(),pos,pos+m_Pattern.size());
                    }
                    else break; // when no mark is set first match is sufficent
                    ++pos;
                }

                return found;
            }

            QString replace (const QString& text, const QString& replacement) const
            {
                QString ret=text;
                return ret.replace(m_Pattern,replacement,cs);
            }

            virtual void setPattern (const QString& pattern)
            {
                m_Pattern=pattern;
            }

        private:
            QString m_Pattern;
            Qt::CaseSensitivity cs;
        };

        struct RegExpComparator: public Comparator
        {
            RegExpComparator (const QRegExp& pattern)
                : m_Pattern (pattern)
            {
            }

            virtual ~RegExpComparator() {}

            bool test (const QString& text, const QModelIndex& idx) const
            {
                bool found=false;
                int pos = 0;
                while ((pos = m_Pattern.indexIn(text, pos)) != -1)  {
                    found=true;
                    if (model!=0) {
                        model->setMark(model->getRow(idx.row()),idx.column(),pos,pos+m_Pattern.matchedLength());
                    }
                    else break;
                }
                return found;
            }

            QString replace (const QString& text, const QString& replacement) const
            {
                QString ret=text;
                return ret.replace(m_Pattern,replacement);
            }

            virtual void setPattern (const QString& pattern)
            {
                m_Pattern=QRegExp(pattern);
            }

        private:
            QRegExp m_Pattern;
        };

        Comparator* Comparator::createComparator (const QString& pattern,
                                                  Qt::CaseSensitivity cs, bool wholeword, bool regexp)
        {
            Comparator* comp;

            if (!wholeword && !regexp) {
                comp=new ContainsComparator(pattern,cs);
            }
            else if (wholeword && !regexp) {
                comp=new RegExpComparator(QRegExp("\b"+QRegExp::escape(pattern)+"\b",cs));
            }
            else if (!wholeword && regexp) {
                comp=new RegExpComparator(QRegExp(pattern,cs));
            }
            else {
                comp=new RegExpComparator(QRegExp("\b"+pattern+"\b",cs));
            }

            return comp;
        }

    }

    SearchWidget::SearchWidget(QWidget *parent) :
            QWidget(parent),
            ui(new Ui::SearchWidget),
            m_Model(0l)
    {
        ui->setupUi(this);
        ui->closeButton->setIcon(KIconLoader::global()->loadIcon("window-close",KIconLoader::Small,KIconLoader::SizeSmall));
        ui->toolButton->setIcon(KIconLoader::global()->loadIcon("configure",KIconLoader::Small,KIconLoader::SizeSmall));

        KAction* a=KStandardAction::replace(this, SLOT(replaceAll()),ui->replaceAllButton);
        ui->replaceAllButton->setDefaultAction (a);
        ui->replaceAllButton->setToolButtonStyle(Qt::ToolButtonIconOnly);

        connect (ui->searchText, SIGNAL(textEdited(QString)), this, SLOT(searchCurrent()));
        connect (ui->forwardButton, SIGNAL(clicked()), this, SLOT(searchForward()));
        connect (ui->backwardButton, SIGNAL(clicked()), this, SLOT(searchBackward()));
        connect (ui->forwardButton_2, SIGNAL(clicked()), this, SLOT(replaceForward()));
        connect (ui->backwardButton_2, SIGNAL(clicked()), this, SLOT(replaceBackward()));
        connect (ui->closeButton, SIGNAL(clicked()), this, SLOT(hide()));

        QMenu* toolMenu = new QMenu (ui->toolButton);
        m_ActionWholeWords=toolMenu->addAction(i18n("Whole words only"));
        m_ActionWholeWords->setCheckable(true);
        m_ActionWholeWords->setChecked(Preferences::instance()->search.m_WholeWords);
        m_ActionSelection=toolMenu->addAction(i18n("Selection only"));
        m_ActionSelection->setCheckable(true);
        m_ActionSelection->setChecked(Preferences::instance()->search.m_Selection);
        m_ActionRegExp=toolMenu->addAction(i18n("Regular expression"));
        m_ActionRegExp->setCheckable(true);
        m_ActionRegExp->setChecked(Preferences::instance()->search.m_RegExp);
        m_ActionCaseSensitive=toolMenu->addAction(i18n("Case sensitive"));
        m_ActionCaseSensitive->setCheckable(true);
        m_ActionCaseSensitive->setChecked(Preferences::instance()->search.m_CaseSensitive);
        ui->toolButton->setMenu(toolMenu);
        connect (m_ActionWholeWords, SIGNAL(triggered()), this, SLOT(resetComparator()));
        connect (m_ActionSelection, SIGNAL(triggered()), this, SLOT(resetComparator()));
        connect (m_ActionRegExp, SIGNAL(triggered()), this, SLOT(resetComparator()));
        connect (m_ActionCaseSensitive, SIGNAL(triggered()), this, SLOT(resetComparator()));

        ui->searchText->setCompletionMode (KGlobalSettings::CompletionAuto);
        ui->searchText->completionObject()->setOrder (KCompletion::Insertion);
        ui->replacementText->setCompletionMode (KGlobalSettings::CompletionAuto);
        ui->replacementText->completionObject()->setOrder (KCompletion::Insertion);

        KConfigGroup config(KGlobal::config(),"Search & Replace");
        if (Preferences::instance()->search.m_Persist) {
            ui->searchText->completionObject()->setItems(config.readEntry("findlist",QStringList()));
            ui->replacementText->completionObject()->setItems(config.readEntry("replacelist",QStringList()));
        }

        m_Current.comp=0l;
    }

    void SearchWidget::setup ()
    {
        connect (m_TableWidget, SIGNAL(clicked(QModelIndex)), this, SLOT(setStartIterator(QModelIndex)));
        setStartIterator(m_TableWidget->currentIndex());
    }

    SearchWidget::~SearchWidget()
    {
        if (Preferences::instance()->search.m_Persist) {
            KConfigGroup config(KGlobal::config(),"Search & Replace");
            config.writeEntry("findlist",ui->searchText->completionObject()->items());
            config.writeEntry("replaceList",ui->replacementText->completionObject()->items());
        }

        delete ui;
        delete m_Current.comp;
    }

    void SearchWidget::changeEvent(QEvent *e)
    {
        QWidget::changeEvent(e);
        switch (e->type()) {
        case QEvent::LanguageChange:
            ui->retranslateUi(this);
            break;
        default:
            break;
        }
    }

    void SearchWidget::setStartIterator (const QModelIndex& start)
    {
        m_Current.searchPosition = m_Model->begin(start);
        m_Current.searchPosition.setSearchableOnly(true);
        m_Current.searchPosition.setWrapOver(true);
        if (m_ActionSelection->isChecked()) {
            m_Current.searchPosition.setSelectionOnly(m_TableWidget->selectionModel());
        }
    }

    void SearchWidget::setComparator ()
    {
        delete m_Current.comp;
        m_Current.comp=Internal::Comparator::createComparator(
                        ui->searchText->text(),
                        m_ActionCaseSensitive->isChecked()?Qt::CaseSensitive:Qt::CaseInsensitive,
                        m_ActionWholeWords->isChecked(),
                        m_ActionRegExp->isChecked());
    }

    void SearchWidget::resetComparator ()
    {
        delete m_Current.comp;
        m_Current.comp=0l;
    }

    void SearchWidget::addCurrentText (KLineEdit* editor)
    {
        editor->completionObject()->addItem(editor->text());
        QStringList items = editor->completionObject()->items();
        if (items.size()>Preferences::instance()->search.m_Historysize) {
            editor->completionObject()->removeItem(items.front());
        }
    }

    void SearchWidget::searchCurrent ()
    {
        if (!ui->searchText->text().isEmpty() && !m_ActionRegExp->isChecked()
            && !m_ActionWholeWords->isChecked()) {
            m_Model->clearMarks();
            if (m_Current.comp==0l) setComparator();
            m_Current.comp->setModel(m_Model);
            m_Current.forward=true;
            m_Current.comp->setPattern(ui->searchText->text());            
            m_Current.searchPosition.pin();
            searchPortion ();
        } else if (ui->searchText->text().isEmpty()) {
            m_Model->clearMarks();
        }
    }

    void SearchWidget::searchForward ()
    {
        if (!ui->searchText->text().isEmpty()) {
            m_Model->clearMarks();
            addCurrentText(ui->searchText);
            m_Current.searchPosition.pin();
            m_Current.searchPosition++;
            if (m_Current.comp==0l) setComparator();
            m_Current.forward=true;
            m_Current.comp->setModel(m_Model);
            searchPortion ();
        }
    }

    void SearchWidget::searchBackward ()
    {
        if (!ui->searchText->text().isEmpty()) {
            m_Model->clearMarks();
            addCurrentText(ui->searchText);
            m_Current.searchPosition.pin();
            m_Current.searchPosition--;
            if (m_Current.comp==0l) setComparator();
            m_Current.forward=false;
            m_Current.comp->setModel(m_Model);
            searchPortion ();
        }
    }

    void SearchWidget::replaceForward ()
    {
        if (!ui->searchText->text().isEmpty()) {
            addCurrentText(ui->replacementText);
            if (m_Current.comp==0l) setComparator();
            replace (m_Current.searchPosition.index(),m_Current.comp);
            m_Current.searchPosition.pin();
            m_Current.searchPosition++;
            m_Current.forward=true;
            m_Current.comp->setModel(0l);
            searchPortion ();
        }
    }

    void SearchWidget::replaceBackward ()
    {
        if (!ui->searchText->text().isEmpty()) {
            addCurrentText(ui->replacementText);
            if (m_Current.comp==0l) setComparator();
            replace (m_Current.searchPosition.index(),m_Current.comp);
            m_Current.searchPosition.pin();
            m_Current.searchPosition--;
            m_Current.forward=false;
            m_Current.comp->setModel(0l);
            searchPortion ();
        }
    }

    void SearchWidget::searchPortion ()
    {
        int processed=0;

        while (m_Current.searchPosition!=m_Model->end()) {
            if (m_Current.comp->test((*m_Current.searchPosition).toString(),m_Current.searchPosition.index())) {
                m_TableWidget->setCurrentIndex(m_Current.searchPosition.index());
                return;
            }
            ++processed;
            if (processed>300) {
                QTimer::singleShot(0,this,SLOT(searchPortion()));
                return;
            }
            if (m_Current.forward) m_Current.searchPosition++;
            else m_Current.searchPosition--;
        }
    }

    void SearchWidget::replace (const QModelIndex& idx, Internal::Comparator* comp)
    {
        m_Model->setData(idx,QVariant(comp->replace (m_Model->data(idx).toString(), ui->replacementText->text())));
    }

    void SearchWidget::setVisible (bool visible)
    {
        QWidget::setVisible(visible);
        ui->searchText->setFocus();
    }

    namespace {
        struct Change
        {
            Change (TagItem* tag_, int col_, const QString& v)
                : tag(tag_),
                  col(col_),
                  newValue(v)
            {
            }

            TagItem* tag;
            int col;
            QString newValue;
        };

        struct MapFunctor {
            typedef QList<Change> result_type;

            MapFunctor (Internal::Comparator* comp_, const QString repl_)
                : comp(comp_),
                  repl(repl_)
            {
            }

            result_type operator() (TagItem* tag)
            {
                QList<Change> result;

                for (int i=0;i<ColumnManager::instance()->getColumnNumber();i++) {
                    if (ColumnManager::instance()->isSearchable(i)) {
                        const QString text=tag->data(i);
                        if (comp->test(text, QModelIndex())) {
                            result << Change (tag,i,comp->replace(text,repl));
                        }
                    }
                }

                return result;
            }

            Internal::Comparator* comp;
            QString repl;
        };

        struct ReduceFunctor {
            void operator() (int& /* dummy */, const QList<Change>& changes)
            {
                foreach (Change t, changes) {
                    TagModel* m = t.tag->getModel();
                    m->setData(m->index(t.tag->getRow(),t.col),QVariant(t.newValue));
                }
            }
        };
    }

    void SearchWidget::replaceAll ()
    {
        if (!ui->searchText->text().isEmpty()) {
            // When no selection was done do a replace on all rows. In this case an
            // efficient parallel search and replace is used. Based on the map and reduced
            // pattern the mappingstep (executed parallel) search and prepare replacements
            // in a single row. The reduction step does not return any value (uses just
            // a dummy value int to keep compiler happy) but instead tothe setData calls
            // in the cells where a replacemet was done. setData can only be called
            // single threaded.
            QModelIndexList indexes = m_TableWidget->selectionModel()->selectedIndexes();
            Internal::Comparator* comp = Internal::Comparator::createComparator (
                    ui->searchText->text(),
                    m_ActionCaseSensitive->isChecked()?Qt::CaseSensitive:Qt::CaseInsensitive,
                    m_ActionWholeWords->isChecked(),
                    m_ActionRegExp->isChecked());
            m_Model->clearMarks();

            if (!m_ActionSelection->isChecked() || indexes.size()<2) {
                ui->replacementText->completionObject()->addItem(ui->replacementText->text());
                MapFunctor mapper(comp,ui->replacementText->text());
                m_Model->mapAndReduce<int> (mapper,ReduceFunctor());
            }
            else {
                // When a selection was done to simple and stupi iterate over all selected
                // cells and check for replacements. Preparing a parallel step is very
                // time-consuming as the indexes must be reformatted so that bunches can be
                // processed at once.
                foreach (QModelIndex idx, indexes) {
                    if (ColumnManager::instance()->isSearchable(idx.column())) {
                        const QString text=m_Model->data(idx).toString();
                        if (comp->test(text, idx)) {
                            m_Model->setData(idx,comp->replace(text,ui->replacementText->text()));
                        }
                    }
                }
            }
        }
    }
}
