/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "textfiltermodel.h"

#include "aid.h"

#include <QFile>
#include <QXmlStreamReader>
#include <QXmlStreamWriter>
#include <QMimeData>
#include <QStringList>
#include <QStringRef>

using namespace std;
using namespace Texel;

TextFilterModel::TextFilterModel(QObject* parent)
   : QAbstractTableModel(parent),
     m_ModelChanged(false)
{
}

int TextFilterModel::columnCount (const QModelIndex& ) const
{
   return 5;
}

int TextFilterModel::rowCount (const QModelIndex& ) const
{
   return items.size();
}

QVariant TextFilterModel::data (const QModelIndex & index, int role) const
{
   switch (role)
   {
   case Qt::EditRole:
   case Qt::DisplayRole:
      switch (index.column()) {
      case 0: return items[index.row()].pattern;
      case 1: return items[index.row()].replacement;
      }
      break;
   case Qt::CheckStateRole:
      switch (index.column()) {
      case 2: return items[index.row()].beginOnly?Qt::Checked:Qt::Unchecked;
      case 3: return items[index.row()].regex?Qt::Checked:Qt::Unchecked;
      case 4: return items[index.row()].wholeWord?Qt::Checked:Qt::Unchecked;
      }
      break;
   }

   return QVariant();
}

bool TextFilterModel::setData ( const QModelIndex & index, const QVariant& value, int role)
{
   if (role==Qt::CheckStateRole)
   {
      switch (index.column()) {
      case 2: items[index.row()].beginOnly=value.toBool(); break;
      case 3: items[index.row()].regex=value.toBool(); break;
      case 4: items[index.row()].wholeWord=value.toBool(); break;
      }
   }
   else if (role==Qt::EditRole) {
      switch (index.column()) {
      case 0: items[index.row()].pattern=value.toString(); break;
      case 1: items[index.row()].replacement=value.toString(); break;
      }
   }

   emit dataChanged(index,index);
   emit modelChanged();
   m_ModelChanged=true;
   return true;
}

QVariant TextFilterModel::headerData ( int section, Qt::Orientation orientation, int role) const
{
   if (role==Qt::DisplayRole && orientation==Qt::Horizontal) {
      switch (section) {
      case 0: return QVariant("Pattern");
      case 1: return QVariant("Replacement");
      case 2: return QVariant("Begin only");
      case 3: return QVariant("Regex");
      case 4: return QVariant("Whole word");
      }
   }
   return QAbstractTableModel::headerData(section,orientation,role);
}

Qt::ItemFlags TextFilterModel::flags ( const QModelIndex & index) const
{
   Qt::ItemFlags flags=Qt::ItemIsEnabled|Qt::ItemIsDropEnabled|Qt::ItemIsSelectable|Qt::ItemIsDragEnabled;
   if (index.column()>1) flags|=Qt::ItemIsUserCheckable;
   if (index.column()<2) flags|=(Qt::ItemIsEditable);
   return flags;
}

bool TextFilterModel::insertRows ( int row, int count, const QModelIndex & parent)
{
   if (row>rowCount()) {
      beginInsertRows(parent,rowCount(),rowCount()+count-1);
      for (int i=0;i<count;i++) items.append(Item());
      endInsertRows();
      return true;
   }
   return false;
}

bool TextFilterModel::removeRows (int row, int count, const QModelIndex& parent)
{
    beginRemoveRows(parent,row,row+count);
    for (int i=0;i<count;i++) items.removeAt(row);
    endRemoveRows();
    return true;
}

bool TextFilterModel::dropMimeData(const QMimeData *mime, Qt::DropAction action, int row, int column, const QModelIndex &parent)
{
   if (action==Qt::CopyAction)
   {
      QModelIndex ix=index(row,column,parent);
      QString text=data(parent).toString()+mime->text();
      setData(parent,text);
   }

   return true;
}

Qt::DropActions TextFilterModel::supportedDropActions () const
{
   return Qt::CopyAction;
}

QStringList TextFilterModel::mimeTypes () const
{
   QStringList mimes;
   mimes << "text/plain";
   mimes+=QAbstractTableModel::mimeTypes();
   return mimes;
}

void TextFilterModel::load (const QString& filename)
{
   QFile file(filename);
   file.open(QIODevice::ReadOnly);
   QXmlStreamReader reader(&file);

   Item item;
   QList<Item> tempItems;
   while (!reader.atEnd())
   {
      QXmlStreamReader::TokenType token=reader.readNext();
      switch (token)
      {
         case QXmlStreamReader::StartDocument:
            break;
         case QXmlStreamReader::StartElement:
            if (reader.name()=="pattern") {
               QXmlStreamAttributes attribs=reader.attributes();
               if (attribs.hasAttribute("replacement")) {
                  item.replacement=attribs.value("replacement").toString();
               }
               else item.replacement="";

               if (attribs.hasAttribute("beginOnly")) {
                  QStringRef v=attribs.value("beginOnly");
                  item.beginOnly=(v=="true")?true:false;
               }
               else item.beginOnly=false;

               if (attribs.hasAttribute("regex")) {
                  QStringRef v=attribs.value("regex");
                  item.regex=(v=="true")?true:false;
               }
               else item.regex=false;

               if (attribs.hasAttribute("wholeword")) {
                  QStringRef v=attribs.value("wholeword");
                  item.wholeWord=(v=="true")?true:false;
               }
               else item.wholeWord=false;

               item.pattern="";
            }
            break;
         case QXmlStreamReader::Characters:
            item.pattern=reader.text().toString();
            break;
         case QXmlStreamReader::EndElement:
            if (reader.name()=="pattern")
            {
                tempItems.append(item);
            }
            break;
         default:
            break;
      }
   }
   file.close();

   beginInsertRows(QModelIndex(),rowCount(),rowCount()+tempItems.size()-1);
   items+=tempItems;
   endInsertRows();
   m_ModelChanged=false;
}

void TextFilterModel::save (const QString& filename)
{
   QFile file(filename);
   file.open(QIODevice::WriteOnly);
   QXmlStreamWriter writer(&file);
   writer.setAutoFormatting(true);

   writer.writeStartDocument("1.0");
   writer.writeStartElement ("list");

   foreach (Item item,items)
   {
      writer.writeStartElement("pattern");
      if(item.replacement!="") {
          writer.writeAttribute(QXmlStreamAttribute("replacement",item.replacement));
      }
      if (item.beginOnly==true) {
        writer.writeAttribute(QXmlStreamAttribute("beginonly","true"));
      }
      if (item.regex==true) {
          writer.writeAttribute(QXmlStreamAttribute("regex","true"));
      }
      if (item.wholeWord==true) {
          writer.writeAttribute(QXmlStreamAttribute("wholeword","true"));
      }
      writer.writeCharacters(item.pattern);
      writer.writeEndElement();
   }

   writer.writeEndElement();
   writer.writeEndDocument();
   file.close();
   m_ModelChanged=false;
}

CompiledFinder::~CompiledFinder ()
{
}

namespace {

class CompiledFinderSet: public CompiledFinder
{
    virtual ~CompiledFinderSet ()
    {
    }

    struct FragmentsByLength
    {
        FragmentsByLength (int size_): size(size_) {}

        int size;
        QMap<QString,QString> fragments;
        QMap<QString,QString> words;
    };

    QList<FragmentsByLength> replaceableFragments;
    QList<FragmentsByLength> replaceableFragmentsAtBegin;

    public:
        virtual void insertFragment (const QString& pattern, const QString& replacement, bool beginOnly, bool word)
        {
            insertFragment(replaceableFragmentsAtBegin,pattern.toLower(),replacement.toLower(),word);
            if (!beginOnly) {
                insertFragment(replaceableFragments,pattern.toLower(),replacement.toLower(),word);
            }
        }

        virtual QString replace (const QString& _source) const
        {
            const QString source=_source.toLower();
            int i=0;
            bool beginOfWord=true;
            QString result=replace(replaceableFragmentsAtBegin,source,i,beginOfWord);
            while (i<source.size()) {
                if (source[i].isSpace()) {
                    // just add a space if last character was not an whitespace
                    if (!beginOfWord) result+=" ";
                    beginOfWord=true;
                    ++i;
                }
                else {
                    result+=replace(replaceableFragments,source,i,beginOfWord);
                    beginOfWord=false;
                }
            }
            return result;
        }

     private:
        QString replace (const QList<FragmentsByLength>& fragments, const QString& source, int& pos, bool bow) const
        {
            if (source.isEmpty()) return "";

            foreach (FragmentsByLength frags, fragments)
            {
                if (pos+frags.size<source.size()) {
                    const QString ref=source.mid(pos,frags.size);
                    // test if word bondary is there
                    const bool word = bow && (source.size()<pos+frags.size || source[pos+frags.size].isSpace());

                    if (word && frags.words.contains(ref)) {
                        const QString& r=frags.words[ref];
                        pos+=frags.size;
                        return r;
                    }

                    // test if a fragment fits
                    if (frags.fragments.contains(ref)) {
                        const QString& r=frags.fragments[ref];
                        pos+=frags.size;
                        return r;
                    }
                }
            }

            return source[pos++];
        }

        void insertFragment (QList<FragmentsByLength>& list, const QString& pattern, const QString& replacement, bool word)
        {
            const int size = pattern.size();
            for (int i=0;i<=list.size();i++) {
                if (i==list.size()) {
                    list.append(FragmentsByLength(size));
                }
                else if (list[i].size<size) {
                    list.insert(i,FragmentsByLength(size));
                }

                if (list[i].size==size) {
                    if (word) {
                        list[i].words.insert(pattern.toLower(),replacement);
                    }
                    else {
                        list[i].fragments.insert(pattern.toLower(),replacement);
                    }
                    break;
                }
            }
        }
};

}

CompiledFinder* TextFilterModel::compile () const
{
    CompiledFinderSet* compiled=new CompiledFinderSet();

    foreach (Item item, items)
    {
        if (item.regex==false) {
            compiled->insertFragment(item.pattern,item.replacement,item.beginOnly,item.wholeWord);
        }
    }

    return compiled;
}

#include "textfiltermodel.moc"
