/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "undocommands.h"

#include "model/columnsbase.h"
#include "model/texeltagmodel.h"
#include "model/tagextension.h"

#include <klocale.h>

#include <set>

namespace {
    template <typename T>
    QList<T> mergeLists (const QList<T>& a, const QList<T>& b)
    {
        std::set<T> merge;
        foreach (T e,a) {
            merge.insert(e);
        }
        foreach (T e,b) {
            merge.insert (e);
        }

        QList<T> result;
        for (typename std::set<T>::iterator i=merge.begin();i!=merge.end();i++) {
            result << *i;
        }

        return result;
    }
}

namespace Texel {

const int UpdateList::_id = 6;
const int RemoveRows::_id = 7;

UpdateList::UpdateList  (TagModel* model, const QList<Entry>& entries, Columns::Column::Role role, const QString& text)
: m_Entries(entries),
  m_Role(role),
  m_Model(model)
{
    if (text!="") {
        setText(text);
    }
    else {
        setText(i18n("Bulk update"));
    }
}

UpdateList::UpdateList (TagModel* model, const SlimIndex& index, const QVariant& oldValue, const QVariant& updValue, Columns::Column::Role role)
: m_Role(role),
  m_Model(model)
{
    m_Entries << Entry (index,oldValue,updValue);
    setText(i18n("Cell update"));
}

int UpdateList::id () const
{
    return _id;
}

bool UpdateList::mergeWith (const QUndoCommand* command)
{
    if (id()==command->id() &&
        m_Role==static_cast<const UpdateList*>(command)->m_Role &&
        m_Model==static_cast<const UpdateList*>(command)->m_Model) {
        const UpdateList* other = static_cast<const UpdateList*>(command);

        const int sRow = singleRow();
        const int sCol = singleCol();

        const bool equalRow = sRow!=-1 && sRow==other->singleRow();
        const bool equalCol = sCol!=-1 && sCol==other->singleCol();

        if (equalRow || equalCol) {
            // Merge both lists; replace duplicates
            m_Entries = mergeLists(m_Entries, other->m_Entries);
            other->m_Entries.clear();
            if (equalRow) setText(i18n("Update row"));
            else setText(i18n("Update column"));
            return true;
        }
    }
    return false;
}

int UpdateList::singleRow () const
{
    int singleRow=-1;
    foreach (Entry e, m_Entries) {
        if (singleRow==-1) singleRow=e.m_Index.row();
        else {
            if (singleRow!=e.m_Index.row()) {
                return -1;
            }
        }
    }
    return singleRow;
}

int UpdateList::singleCol () const
{
    int singleCol=-1;
    foreach (Entry e, m_Entries) {
        if (singleCol==-1) singleCol=e.m_Index.column();
        else {
            if (singleCol!=e.m_Index.column()) {
                return -1;
            }
        }
    }
    return singleCol;
}

void UpdateList::undo()
{
    m_Model->setDataListRedo(m_Entries, Entry::OriginalValue, m_Role);
}

void UpdateList::redo()
{
    m_Model->setDataListRedo(m_Entries, Entry::UpdatedValue, m_Role);
}

RemoveRows::RemoveRows (const TagItem::List& rows)
: m_Rows(rows),
  toDelete(true)
{
    setText (i18n("Remove %1 rows",rows.size()));
}

RemoveRows::~RemoveRows ()
{
    if (toDelete) {
        foreach (TagItem* item, m_Rows) delete item;
    }
}

int RemoveRows::id () const
{
    return _id;
}

bool RemoveRows::mergeWith (const QUndoCommand* command)
{
    const RemoveRows* rr = dynamic_cast<const RemoveRows*>(command);
    if (rr!=0l) {
        if (id()==command->id() && toDelete==rr->toDelete &&
                !m_Rows.isEmpty() && !rr->m_Rows.isEmpty() &&
                m_Rows.first()->getModel()==rr->m_Rows.first()->getModel()) {
            m_Rows=mergeLists(m_Rows,rr->m_Rows);
            rr->m_Rows.clear();
            setText (i18n("Remove %1 rows",m_Rows.size()));
            return true;
        }
    }
    return false;
}

void RemoveRows::undo ()
{
    if (m_Rows.size()>0) {
        m_Rows[0]->getModel()->insertRowsRedo(m_Rows[0]->getRow(),m_Rows);
    }
    toDelete=false;
}

void RemoveRows::redo ()
{
    if (m_Rows.size()>0) {
        m_Rows[0]->getModel()->removeRowsRedo(m_Rows.first()->getRow(),m_Rows.last()->getRow());
    }
    foreach (TagItem* tag, m_Rows) {
        if (tag->getRelations()!=0l) {
            tag->getRelations()->onRemovalFromList(tag);
        }
    }

    toDelete=true;
}

Clear::Clear (const TagItem::List& rows)
: m_Rows(rows),
  toDelete(true)
{
    setText (i18n("Clear all"));
}

Clear::~Clear ()
{
    if (toDelete) {
        foreach (TagItem* item, m_Rows) delete item;
    }    
}

void Clear::undo ()
{
    if (m_Rows.size()>0) {
        m_Rows[0]->getModel()->insertRowsRedo(m_Rows[0]->getRow(),m_Rows);
    }
    toDelete=false;
}

void Clear::redo ()
{
    if (m_Rows.size()>0) {
        m_Rows[0]->getModel()->clearRedo();
    }
    toDelete=true;
}

ToggleBool::ToggleBool (const QModelIndex& index, bool value)
:  m_Index(index),
   m_Value(value)
{
    setText(i18n("Toggle boolean"));
}

ToggleBool::~ToggleBool ()
{
}

void ToggleBool::undo ()
{
    const_cast<TagModel*>(dynamic_cast<const TagModel*>(m_Index.model()))->setDataRedo (m_Index,QVariant(!m_Value),Columns::Column::RoleDisplay);
}

void ToggleBool::redo ()
{
    const_cast<TagModel*>(dynamic_cast<const TagModel*>(m_Index.model()))->setDataRedo (m_Index,QVariant(m_Value),Columns::Column::RoleDisplay);
}

UpdateExtensionAttrib::UpdateExtensionAttrib (const TagItem* row, const QString& attrib, const QString& old, const QString& upd)
    : m_Row(row),
      m_Attrib(attrib),
      m_OldValue(old),
      m_UpdValue(upd)
{
    setText(i18n("Update extended attribute"));
}

UpdateExtensionAttrib::~UpdateExtensionAttrib ()
{
}

void UpdateExtensionAttrib::undo ()
{
    m_Row->getExtensions()->setAttribute(m_Attrib,m_OldValue);
}

void UpdateExtensionAttrib::redo ()
{
    m_Row->getExtensions()->setAttribute(m_Attrib,m_UpdValue);
}

InsertExtensionAttrib::InsertExtensionAttrib (const TagItem* row, const QString& attrib, const QString& value, bool removed)
    : m_Row(row),
      m_Attrib(attrib),
      m_Value(value),
      m_Removed(removed)
{
}

InsertExtensionAttrib::~InsertExtensionAttrib ()
{
    setText(i18n("Remove extended attribute"));
}

void InsertExtensionAttrib::undo ()
{
    if (m_Removed) {
        m_Row->getExtensions()->setAttribute(m_Attrib,m_Value);
    }
    else {
        m_Row->getExtensions()->rmAttribute(m_Attrib);
    }
}

void InsertExtensionAttrib::redo ()
{
    if (m_Removed) {
        m_Row->getExtensions()->rmAttribute(m_Attrib);
    }
    else {
        m_Row->getExtensions()->setAttribute(m_Attrib,m_Value);
    }
}

}
