/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#ifndef UNDOCOMMANDS_H
#define UNDOCOMMANDS_H

#include <QUndoCommand>
#include <QModelIndex>

#include "model/columnsbase.h"
#include "model/tagitem.h"
#include "util/slimindex.h"

namespace Texel {

class TagModel;

// ----------------------------------------------------------------------------------------------
//
// Undo classes for main table
//
// ----------------------------------------------------------------------------------------------

class UpdateList: public QUndoCommand
{
public:
    struct Entry
    {
        Entry (const SlimIndex& idx, const QVariant& old, const QVariant& upd)
            : m_Index(idx)
        {
            m_Values[0]=old;
            m_Values[1]=upd;
        }

        enum ValueTypes { OriginalValue=0, UpdatedValue=1 };

        void setValue (ValueTypes t, const QVariant& v)
        {
            m_Values[(int)t]=v;
        }

        const QVariant& value (ValueTypes t) const
        {
            return m_Values[(int)t];
        }

        bool operator< (const Entry& other) const
        {
            return m_Index<other.m_Index;
        }

        SlimIndex m_Index;
        QVariant m_Values[2];
    };

    UpdateList (TagModel* model, const QList<Entry>& entries, Columns::Column::Role role, const QString& text);
    UpdateList (TagModel* model, const SlimIndex& index, const QVariant& oldValue, const QVariant& updValue, Columns::Column::Role role);

    virtual void undo ();
    virtual void redo ();
    virtual int id() const;
    virtual bool mergeWith (const QUndoCommand* command);

private:
    static const int _id;

    int singleRow () const;
    int singleCol () const;

    mutable QList<Entry> m_Entries;
    Columns::Column::Role m_Role;
    TagModel* m_Model;
};

class RemoveRows: public QUndoCommand
{
public:
    RemoveRows (const TagItem::List& rows);
    virtual ~RemoveRows();

    virtual void undo ();
    virtual void redo ();

    virtual int id() const;
    virtual bool mergeWith (const QUndoCommand* command);

private:
    static const int _id;

    mutable TagItem::List m_Rows;
    // When the items are deleted from the model (after redo) this model must take care
    // of deletion
    bool toDelete;
};

class Clear: public QUndoCommand
{
public:
    Clear (const TagItem::List& rows);
    virtual ~Clear ();

    virtual void undo ();
    virtual void redo ();

private:
    TagItem::List m_Rows;
    // When the items are deleted from the model (after redo) this model must take care
    // of deletion
    bool toDelete;
};

class ToggleBool: public QUndoCommand
{
public:
    ToggleBool (const QModelIndex& index, bool value);
    virtual ~ToggleBool ();

    virtual void undo ();
    virtual void redo ();

private:
    QModelIndex m_Index;
    bool m_Value;
};

// ----------------------------------------------------------------------------------------------
//
// Undo classes for detailed dock
//
// ----------------------------------------------------------------------------------------------

class UpdateExtensionAttrib: public QUndoCommand
{
public:
    UpdateExtensionAttrib (const TagItem* row, const QString& attrib, const QString& old, const QString& upd);
    virtual ~UpdateExtensionAttrib ();

    virtual void undo ();
    virtual void redo ();

private:
    const TagItem* m_Row;
    QString m_Attrib;
    QString m_OldValue;
    QString m_UpdValue;
};

class InsertExtensionAttrib: public QUndoCommand
{
public:
    InsertExtensionAttrib (const TagItem* row, const QString& attrib, const QString& value, bool removed);
    virtual ~InsertExtensionAttrib ();

    virtual void undo ();
    virtual void redo ();

private:
    const TagItem* m_Row;
    QString m_Attrib;
    QString m_Value;
    bool m_Removed;
};

}

#endif // UNDOCOMMANDS_H
