/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "columnmanager.h"
#include "global_exceptions.h"

#include <QRegExp>
#include <QXmlStreamReader>
#include <QFile>

#include <kstandarddirs.h>
#include <kglobal.h>
#include <klocale.h>
#include "klocalizedstring.h"

#include <aid.h>

namespace Texel {

    ColumnManager::Data::Data () :
      searchable(true),
      hidden(false),
      editType(QMetaType::Void),
      sortType(QMetaType::Void),
      streamToList(true)
    {
    }

    namespace {
        bool texelLessThan (const QPair<QString,int>& p1, const QPair<QString,int>& p2)
        {
            return p1.first<p2.first;
        }

        inline bool parseBool (const QString& s, bool default_)
        {
            return s.isEmpty()?default_:s=="true" || s=="1" || s=="on";
        }
    }

    void ColumnManager::initFromXML (QFile* file,const QString& selection)
    {
        struct ParseError
        {
            ParseError (const QString& r): reason(r) {}
            QString reason;
        };

        enum Pieces { None=0, Attributes=1, Type=2, Description=4 };

        try {
            if (!file->exists()) {
                throw ParseError(i18n("File column.xml does not exists. Can't create columns.",file->fileName()));
            }

            file->open(QIODevice::ReadOnly);
            QXmlStreamReader reader(file);

            int depth=0;
            Data* currentItem=0l;
            unsigned currentOrder=0;
            int currentPieces=None;

            while (!reader.atEnd()) {
                reader.readNext();
                if (reader.tokenType()==QXmlStreamReader::StartElement) {
                    if (depth==0) {
                        if (reader.name()=="columns") {
                            ++depth;
                        }
                        else throw ParseError(i18n("Invalid document: Root node must be <columns>"));
                    }
                    else if (depth==1) {
                        if (reader.name()!=selection) {
                            reader.skipCurrentElement();
                        }
                        else {
                            ++depth;
                        }
                    }
                    else if (depth==2) {
                        if (reader.name()=="column") {
                            ++depth;
                            currentItem = new Data();
                            if (reader.attributes().hasAttribute("id")) {
                                currentOrder = reader.attributes().value("id").toString().toInt();
                            }
                            else ++currentOrder;

                            currentItem->commonName = reader.attributes().value("name").toString();
                            currentItem->name = ki18n(currentItem->commonName.toLocal8Bit());
                            currentItem->xmlName = asXml(currentItem->commonName);
                        }
                        else throw ParseError(i18n("Invalid document: expected <column> tag, found <%1>",reader.name().toString()));
                    }
                    else if (depth==3) {
                        ++depth;
                        if (reader.name()=="attribs") {
                            currentItem->hidden=parseBool(reader.attributes().value("hidden").toString(),false);
                            currentItem->searchable=parseBool(reader.attributes().value("searchable").toString(),true);
                            currentItem->streamToList=parseBool(reader.attributes().value("stream").toString(),true);
                            currentPieces|=Attributes;
                        }
                        else if (reader.name()=="description") {
                            currentItem->description=ki18n(reader.text().toLocal8Bit());
                            currentPieces|=Description;
                        }
                        else if (reader.name()=="type") {
                            if (reader.attributes().hasAttribute("sort")) {
                                currentItem->sortType=static_cast<QMetaType::Type>(QMetaType::type(reader.attributes().value("sort").toLocal8Bit()));
                                                        }
                            else throw ParseError(i18n("Missing sort type in type declaration of %1. Must be defined.",currentItem->commonName));
                            if (reader.attributes().hasAttribute("edit")) {
                                currentItem->editType=static_cast<QMetaType::Type>(QMetaType::type(reader.attributes().value("edit").toLocal8Bit()));
                            }
                            else {
                                currentItem->editType=currentItem->sortType;
                            }
                            currentPieces|=(int)Type;
                        }
                    }
                }
                else if (reader.tokenType()==QXmlStreamReader::EndElement) {
                    switch (depth) {
                    case 3:
                        add((ColumnNames)currentOrder,currentItem);
                        break;
                    }
                    --depth;
                }
            }
        }
        catch (ParseError p) {
            file->close();
            throw application_initialization_termination(-10,p.reason.toLocal8Bit());
        }
        file->close();
    }

    ColumnManager::ColumnManager ()
    {
        QFile file(KStandardDirs::locate("config","columns.xml"));
        initFromXML(&file,"musictags");

        // The i18nc function has no effect as this class is created before the main window.
        // However, this allows it to place the strings in the localization. The translation
        // is done via an ki18n call.
        /*
        add(Pattern, 	new Data(I18N_NOOP2("Column name","Searchpattern"), true, false,
                                 QMetaType::QString, QMetaType::QString));
        add(DestinationPath, new Data(I18N_NOOP2("Column name","Destination Path"), true, true,
                                      QMetaType::QString, QMetaType::QString));
        add(Path,    	new Data(I18N_NOOP2("Column name","Path"), true, false,
                                 QMetaType::QString, QMetaType::QString));
        add(Filename,	new Data(I18N_NOOP2("Column name","Filename"), true, false,
                                 QMetaType::QString, QMetaType::QString));
        add(Artist,  	new Data(I18N_NOOP2("Column name","Artist"), true, false,
                                 QMetaType::QString, QMetaType::QString));
        add(Title,   	new Data(I18N_NOOP2("Column name","Title"), true, false,
                                 QMetaType::QString, QMetaType::QString));
        add(Album,   	new Data(I18N_NOOP2("Column name","Album"), true, false,
                                 QMetaType::QString, QMetaType::QString));
        add(Length,  	new Data(I18N_NOOP2("Column name","Length"), true, false,
                                 QMetaType::Void, QMetaType::Int));
        add(Bitrate, 	new Data(I18N_NOOP2("Column name","Bitrate"), true, false,
                                 QMetaType::Void, QMetaType::Int));
        add(Track,   	new Data(I18N_NOOP2("Column name","Track"), true, false,
                                 QMetaType::Int, QMetaType::Int));
        add (CD,        new Data(I18N_NOOP2("Column name","CD Nr"), false, false,
                                  QMetaType::Int, QMetaType::Int));
        add (Rating,    new Data(I18N_NOOP2("Column name","Rating"), false, false,
                                  QMetaType::Int, QMetaType::Int));
        add(SampleRate, new Data(I18N_NOOP2("Column name","SampleRate"), true, false,
                                  QMetaType::Void, QMetaType::Int));
        add(Comment,   	new Data(I18N_NOOP2("Column name","Comment"), true, false,
                                 QMetaType::QString, QMetaType::QString));
        add(Genre,   	new Data(I18N_NOOP2("Column name","Genre"), true, false,
                                 QMetaType::QString, QMetaType::QString));
        add(Year,   		new Data(I18N_NOOP2("Column name","Year"), true, false,
                                     QMetaType::Int, QMetaType::Int));
        add(Channels,   	new Data(I18N_NOOP2("Column name","Channels"), true, false,
                                     QMetaType::Void, QMetaType::Int));
        add(FileSize,        new Data(I18N_NOOP2("Column name","Size"), false, false,
                                      QMetaType::Void, QMetaType::Int));
        // Must always be the last as it has no attached column
        add(MarkForDeletion, new Data(I18N_NOOP2("Column name","Mark for deletion"), false, false,
            QMetaType::Void, QMetaType::Void));

        m_Columns[Pattern]->description = ki18n("Pattern to parse path and filename into variables or meta tags");
        m_Columns[Artist]->description = ki18n("Artist of song");
        m_Columns[Title]->description = ki18n("Title of song");
        m_Columns[Album]->description = ki18n("Album on which the song was released");
        m_Columns[DestinationPath]->description = ki18n("When a path is entered here the file is copied to this directory upon save");
        m_Columns[Year]->description = ki18n("Year of the songs release");
        m_Columns[Channels]->description = ki18n("Number of channels in the file. Can be used to eleminate the correct duplicated");
        m_Columns[FileSize]->description = ki18n("Filesize of the source file");
        m_Columns[CD]->description = ki18n("Just used in MPEG ID3V2 and MP4 tag format to store the current cd number in case of a mlti CD album");
        m_Columns[Rating]->description = ki18n("Rating of the track. Is a number between 1 (worst) and 5 (best) or 0 for unknown");
        m_Columns[Genre]->description = ki18n("Genre of the song. Autocompletion uses ID3V1 tags.");
        m_Columns[Comment]->description = ki18n("Optional comment for the track");
        m_Columns[SampleRate]->description = ki18n("Sample rate of the file. The higher the better the quality. Typical CD quality is 44.100. Can be used to compare duplicate files.");
        m_Columns[Path]->description = ki18n("Path where the track is stored. The common prefix of the path is not display to shorten the displayed characters");
        m_Columns[Filename]->description = ki18n("Filename of the track");
        m_Columns[Length]->description = ki18n("Length of the track in minutes and seconds. Its rounded to the complete seconds.");
        m_Columns[Bitrate]->description = ki18n("Bitrate of the track. Indicates in case of lossy compression the quality of the file.");
        m_Columns[Track]->description = ki18n("Tracknumber on the album. Must be between 0 and 99");
        m_Columns[MarkForDeletion]->description = ki18n("When checked the item will be deleted upon the next save");
        m_Columns[MarkForDeletion]->streamToList=false;
        */

        calcEditables();

        m_LeastPrios << Channels << SampleRate << Year << Comment;
    }

    ColumnManager::~ColumnManager ()
    {
    }

    QString ColumnManager::getDescription (int col) const
    {
        if (col>=0 && col<getColumnNumber()) {
            return m_Columns[col]->description.toString();
        }
        return KGlobal::staticQString("");
    }

    void ColumnManager::add (ColumnNames pos, Data* data)
    {
        if (pos>=m_Columns.size()) {
            m_Columns.resize(pos+1);
        }
        m_Columns[pos].reset(data);
        if (data->xmlName!="") {
            m_IndexXmlMapper.insert(data->xmlName,pos);
        }
    }

    void ColumnManager::calcEditables()
    {
        for (int i=0;i<m_Columns.size();i++) {
            if (isEditable(i)) {
                m_EditableColumns << i;
            }
        }
    }

    int ColumnManager::mapFromEditableColumn(int editableColumn) const
    {
        return m_EditableColumns[editableColumn];
    }

    int ColumnManager::mapToEditableColumn(int column) const
    {
        return m_EditableColumns.indexOf(column);
    }

    int ColumnManager::indexByName (const QString& name) const
    {
        for (int i=0;i<m_Columns.size();i++) {
            if (m_Columns[i]->name.toString()==name || m_Columns[i]->commonName==name) return i;
        }
        return -1;
    }

    int ColumnManager::indexByXmlName (const QString& name) const
    {
        QHash<QString,int>::const_iterator iter=m_IndexXmlMapper.find(name);
        if (iter==m_IndexXmlMapper.end()) return -1;
        return (*iter);
    }

    ColumnManager::ColumnNames ColumnManager::getLeastPriority (int n) const
    {
        if (n>0 && n<m_LeastPrios.size()) return m_LeastPrios[n];
        return Invalid;
    }

    QString ColumnManager::asXml (const QString& name)
    {
        return name.toLower().replace(QRegExp("[ #_]"),"");
    }

}
