/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "model/tagitem.h"
#include "model/columns.h"
#include "model/problem.h"
#include "model/texeltagmodel.h"
#include "model/completermodel.h"

#include "util/qtlevenshtein.h"
#include "func/textfiltersingleton.h"
#include "func/searchpatternparser.h"
#include "func/m3u/playlistmodel.h"

#include "options/preferences.h"

#include <QString>
#include <QCache>
#include <QFileInfo>
#include <QDir>
#include <QCompleter>
#include <QListView>

#include <klocale.h>
#include <kdebug.h>

#include <taglib/taglib.h>
#include <taglib/id3v1genres.h>

#include <boost/multi_index_container.hpp>
#include <boost/multi_index/member.hpp>
#include <boost/multi_index/ordered_index.hpp>
#include <boost/multi_index/composite_key.hpp>
#include <boost/multi_index/sequenced_index.hpp>

#include "aid.h"

using namespace Texel::Columns;
using namespace boost::multi_index;
using namespace Texel;

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Caching for autocompletion
//
//////////////////////////////////////////////////////////////////////////////////////////////

namespace {
class BasicCache
{
private:
    struct CacheEntry
    {
        unsigned timestamp;
        int rowid;
        QString formular;

        CacheEntry (int t, int r, const QString& f)
            : timestamp(t), rowid(r), formular(f)
        {
        }
    };

    typedef multi_index_container< CacheEntry,
    indexed_by<
    sequenced<>,
    ordered_unique <
    composite_key < CacheEntry,
    member<CacheEntry, int, &CacheEntry::rowid>,
    member<CacheEntry, QString, &CacheEntry::formular>
    >
    >
    >
    > Cache;

    Cache cache;

    unsigned max;
    unsigned order;

public:
    BasicCache(int maxsize)
        : cache(),
          max(maxsize),
          order(0)
    {
    }

    /**
          Inserts a new formular for a certain column. If this formular does already exist
          than it is simply renewed as oldest entries are removed when the cache is full.
          */
    void insert (int id, const QString& formular)
    {
        std::pair<Cache::iterator,bool> p=cache.push_back (CacheEntry(order,id,formular));

        if (!p.second) {
            cache.relocate (cache.end(),p.first);
        }
        else {
            // If the forumar is not already in the cache for this column than
            // the completer must be removed to be recreated.
            invalidateModel(id);

            if (cache.size()>max) {
                invalidateModel (cache.front().rowid);
                cache.pop_front();
            }
        }
    }

private:
    QStringList entries (int id) const
    {
        QStringList e;

        std::pair<nth_index<Cache,1>::type::iterator,nth_index<Cache,1>::type::iterator> p=get<1>(cache).equal_range(id);
        while (p.first!=p.second)
        {
            e << (*p.first).formular;
            ++p.first;
        }
        qStableSort (e);

        return e;
    }

private:
    struct ModelSet
    {
        int set;
        bool current;
        CompleterModelIF* model;

        ModelSet(int s, CompleterModelIF* m)
            : set(s), current(false), model(m)
        {
        }
    };
    QMap<int, ModelSet> models;

    void invalidateModel (int id)
    {
        QMap<int, ModelSet>::iterator m = models.find(id);
        if (m!=models.end()) {
            (*m).current=false;
        }
    }

public:
    void registerModel (int id, int set, CompleterModelIF* model)
    {
        models.insert(id,ModelSet(set,model));
    }

    void refreshModel (int id)
    {
        QMap<int, ModelSet>::iterator m = models.find(id);
        if (m!=models.end() && (*m).current==false) {
            (*m).model->setData ((*m).set, entries(id));
            (*m).current=true;
        }
    }

};

struct FormularCache: public BasicCache, public Singleton<FormularCache>
{
    friend class Singleton<FormularCache>;

    FormularCache()
        : BasicCache(100)
    {
    }
};

struct ValueCache: public BasicCache, public Singleton<ValueCache>
{
    friend class Singleton<ValueCache>;

    ValueCache()
        : BasicCache(100)
    {
    }
};

struct Completers: public Singleton<Completers>
{
    friend class Singleton<Completers>;

    QVector<QCompleter*> completers;

    Completers()
        : completers(ColumnManager::instance()->getColumnNumber())
    {
        for (int i=0;i<completers.size();i++) completers[i]=NULL;
    }

    ~Completers()
    {
        for (int i=0;i<completers.size();i++) delete completers[i];
    }

    QCompleter* createCompleter (QAbstractItemModel* model)
    {
        QCompleter* completer=new QCompleter();
        model->setParent(completer);
        completer->setModel(model);
        completer->setModelSorting(QCompleter::CaseSensitivelySortedModel);
        completer->setCompletionMode(QCompleter::InlineCompletion);
        return completer;
    }

    void createCompleterBoth (int id)
    {
        if (completers[id]==0l) {
            CompleterModel* model=new CompleterModel();
            ValueCache::instance()->registerModel(id,0,model);
            FormularCache::instance()->registerModel(id,1,model);
            completers[id]=createCompleter(model);
        }
    }

    void createCompleterFormular (int id)
    {
        if (completers[id]==0l) {
            CompleterModel* model=new CompleterModel();
            FormularCache::instance()->registerModel(id,1,model);
            completers[id]=createCompleter(model);
        }
    }

    void createCompleterValue (int id)
    {
        if (completers[id]==0l) {
            CompleterModel* model=new CompleterModel();
            ValueCache::instance()->registerModel(id,0,model);
            completers[id]=createCompleter(model);
        }
    }
};
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Base class for all columns
//
//////////////////////////////////////////////////////////////////////////////////////////////

Column::Column ()
{
}

Column::~Column()
{
}

inline DependencyMap& Column::getDependencyMap(TagItem* item)
{
    return item->m_Dependencies;
}

inline TagModel* Column::model(TagItem* item)
{
    return item->model;
}

QCompleter* Column::getCompleter () const
{
    return 0l;
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Base class for all static (uneditable) columns
//
//////////////////////////////////////////////////////////////////////////////////////////////

Static::Static ()
{
}

Static::~Static()
{
}

bool Static::isEditable () const
{
    return false;
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Base class for all static (uneditable) columns
//
//////////////////////////////////////////////////////////////////////////////////////////////

StaticInt::StaticInt ()
{
}

StaticInt::~StaticInt()
{
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Base class for all editable columns
//
//////////////////////////////////////////////////////////////////////////////////////////////

Editable::Editable ()
    : problem(0l),
      flags()
{
}

Editable::~Editable ()
{
}

bool Editable::isEditable () const
{
    return true;
}

QString Editable::data (Role role) const
{
    if (role==RoleFormular ||
            (hasFormular() && (role==RolePersistence || role==RoleDisplayFormularFirst)) ||
            (problem!=0l && problem->isFormularProblem() && role==RoleDisplay)
            ) {
        if (problem!=0l) {
            return formular;
        }
        return formular;
    }
    else if (role==RoleSorting) {
        return TextFilterSingleton::instance()->filter(doData(!isUpdated()));
    }
    else if (role==RoleAsRawVariable || !isUpdated()) {
        return doData (true);
    }
    else {
        return doData(false);
    }
}

bool Editable::setData(const QString& value, Role role, TagItem* item)
{
    // Just to speed up loading of large XML lists
    if (role==RoleInitialization) return doSetData(value,role,item);

    InputProblem* oldProblem=problem;
    problem=0l;

    bool result=false;
    if (!value.isEmpty() && value[0]=='=') {
        formular=value;
        FormularCache::instance()->insert(column(),formular);
        if (role!=RoleReInitalization) {
            QString replaceableValue=evaluateFormula(role==RoleReevaluate, item);
            if (problem==0l) { // otherwise formular error
                result=doSetData(replaceableValue,RoleFormular,item);
            }
        }
    }
    else {
        formular="";
        result=doSetData(value,role,item);
    }

    if (result && oldProblem!=0l) {
        delete oldProblem;
    }

    return result;
}

void Editable::evaluate (TagItem* item)
{
    if (hasFormular()) {
        setData(formular,RoleDisplay,item);
    }
}

QString Editable::evaluateFormula (bool reevaluate, TagItem* item)
{
    if (!reevaluate) {
        getDependencyMap(item).clearDependency(column());
    }

    try {
        return item->eval(formular,QVariant(column()));
        problem=0;
    }
    catch (InputProblem* ip)
    {
        ip->setIndex(model(item)->index(item->getRow(),column()));
        problem=ip;
        return formular;
    }
}

void Editable::setUpdated (bool updated, TagItem*)
{
    if (flags.testFlag(FlagIsUpdated)!=updated) {
        flags^=FlagIsUpdated;
    }
}

void Editable::reset (TagItem*)
{
    flags&=~FlagIsUpdated;
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Base class for all editable Integer Columns
//
//////////////////////////////////////////////////////////////////////////////////////////////

EditableInt::EditableInt ()
    : orgValue(-1),
      updValue(-1)
{
}

EditableInt::~EditableInt ()
{
}

QString EditableInt::doData (bool original) const
{
    if (original) return QString::number(orgValue);
    return QString::number(updValue);
}

bool EditableInt::setIntValue (int value, Role role, TagItem* item)
{
    if (role==RoleInitialization && orgValue!=value) {
        orgValue=value;
        return true;
    }
    else if (value==orgValue) {
        if (isUpdated()) {
            updValue=value;
            setUpdated(false, item);
            return true;
        }
        else return false;
    }
    else if (!isUpdated() || value!=updValue || role==RoleFormular) {
        updValue=value;
        setUpdated(true, item);
        return true;
    }
    return false;
}

bool EditableInt::doSetData(const QString& value, Role role, TagItem* item)
{
    bool converted;
    const int v=value.toInt(&converted);
    if (converted && (v!=getIntValue(RoleAsRawVariable) || isUpdated())) {
        setIntValue(v,role,item);
    }
    return converted;
}

void EditableInt::setUpdated (bool updated, TagItem* item)
{
    if (updated==false && isUpdated()==true) {
        orgValue=updValue;
    }
    Editable::setUpdated(updated, item);
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Base class for all editable String Columns
//
//////////////////////////////////////////////////////////////////////////////////////////////

EditableString::EditableString ()
{
}

EditableString::~EditableString()
{
}

QString EditableString::doData (bool original) const
{
    if (original) return orgValue;
    return updValue;
}

bool EditableString::doSetData(const QString& value, Role role, TagItem* item)
{
    if (role==RoleInitialization && orgValue!=value) {
        orgValue=value;
        setUpdated(false, item);
        return true;
    }
    else if (value==orgValue) {
        if (isUpdated()) {
            updValue=value;
            setUpdated(false, item);
            return true;
        }
        else return false;
    }
    else if (!isUpdated() || value!=updValue || role==RoleFormular) {
        updValue=value;
        setUpdated(true, item);
        return true;
    }
    return false;
}

void EditableString::setUpdated (bool updated, TagItem* item)
{
    if (updated==false && isUpdated()==true) {
        orgValue=updValue;
        updValue.clear();
    }
    Editable::setUpdated(updated,item);
}

void EditableString::reset (TagItem* item)
{
    updValue.clear();
    Editable::reset(item);
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Base class for filename related columns
//
//////////////////////////////////////////////////////////////////////////////////////////////

FileRelated::FileRelated ()
{
}

FileRelated::~FileRelated ()
{
}

void FileRelated::changeFilename (const QString& value, Role role, TagItem* item)
{
    // To indicate if a playlist contains altered items increase dirty counter upon first
    // change and switch it back when filename is resetted to original value
    if (item->hasRelations())
    {
        PlaylistModel* model = dynamic_cast<PlaylistModel*>(item->getModel()->getPlayListModel());
        if (role==RoleInitialization) {
            model->decDirtyCounter(item->getRelations()->getPlaylistGroup());
        }
        else {
            if (!item->isAltered(ColumnManager::Filename) && !item->isAltered(ColumnManager::Path)) {
                model->incDirtyCounter(item->getRelations()->getPlaylistGroup());
            }
        }
    }

    setFileRelatedData (value,role,item);
    DirTree::Name filename=DirTree::Name(item->getFilenameId(RoleAsVariable));
    model(item)->alteredPersistentFilename(item,filename);
}

bool FileRelated::doSetData(const QString& value, Role role, TagItem* item)
{
    if (value==data(RolePersistence)) {
        return false;
    }

    if (role==RoleInitialization) {
        setFileRelatedData (value,role,item);
    }
    else if (value==data(RoleAsRawVariable)) {
        // Value is equal to the original stored value. Rewind to original or leave as is
        if (isUpdated()) {
            changeFilename (value,RoleInitialization,item);
            return true;
        }
        else return false;
    }
    else if (!isUpdated() || value!=data(RoleDisplay) || role==RoleFormular) {
        changeFilename (value,role,item);
        return true;
    }

    return false;
}

/*********************************************************************************************
 *********************************************************************************************
 **
 ** Concrete column classes
 **
 *********************************************************************************************
 ********************************************************************************************/

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Searchpattern
//
//////////////////////////////////////////////////////////////////////////////////////////////

Searchpattern::Searchpattern()
    : problem(0l)
{
}

Searchpattern::~Searchpattern()
{
}

bool Searchpattern::isEditable () const
{
    return false;
}

void Searchpattern::parse (bool reevaluate, TagItem* item)
{
    // Reset the dependencies before parsing the string
    if (!reevaluate) getDependencyMap(item).clearDependency(0);

    if (searchpattern.isEmpty()) {
        return;
    }

    //
    // Create a pattern which contains as many directory parts as slashes are in the search expression
    //
    QFileInfo info(item->getAbsolutFilename(Columns::Column::RoleDisplay));
    QDir dir=info.dir();
    QString pattern=info.completeBaseName();
    int noOfSeparators = searchpattern.count('/');
    while (noOfSeparators>0) {
        pattern=dir.dirName()+QDir::separator()+pattern;
        if (!dir.cdUp()) break;
        --noOfSeparators;
    }

    try {
        SearchPatternReplacement spr;
        spr.process(pattern, searchpattern, item);

        variables=spr.getVariables();
        QList<int> source = spr.getSource();
        if (!reevaluate) {
            if (source.isEmpty()) {
                source << ColumnManager::Filename << ColumnManager::Path;
            }
            foreach (int j,source) {
                getDependencyMap(item).setDependency(0,ColumnManager::instance()->mapToEditableColumn(j));
            }
        }
        problem=0l;
    }
    catch (InputProblem* ip)
    {
        ip->setIndex(model(item)->index(item->getRow(),column()));
        problem=ip;
    }
}

QString Searchpattern::suggestVarName (const QString& name) const
{
    QString suggestion;
    int distance=-1;

    QStringList suggestions;
    suggestions+=variables.keys();
    suggestions+=Preferences::instance()->m_GlobalVars.keys();
    suggestions << "g_COMMON_PATH";
    for (int i=0;i<ColumnManager::instance()->getColumnNumber();i++) {
        suggestions << ColumnManager::instance()->getName(i);
    }

    foreach (QString n, suggestions) {
        const int d = Texel::distance(name,n);
        if (distance==-1 || d<distance) {
            suggestion=n;
            distance=d;
        }
    }

    if (distance>name.size() || distance>5) return QString();

    return suggestion;
}

QString Searchpattern::data (Role) const
{
    return searchpattern;
}

bool Searchpattern::setData(const QString& value, Role role, TagItem* item)
{
    if (value!=searchpattern || role==RoleReevaluate) {
        searchpattern=value;
        if (role!=RoleInitialization) {
            parse(role==RoleReevaluate,item);
            ValueCache::instance()->insert(column(),value);
        }
        return true;
    }
    return false;
}

QCompleter* Searchpattern::getCompleter () const
{
    Completers::instance()->createCompleterValue(column());
    ValueCache::instance()->refreshModel(column());
    return Completers::instance()->completers[column()];
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// DestinationPath
//
//////////////////////////////////////////////////////////////////////////////////////////////

DestinationPath::DestinationPath()
{
}

DestinationPath::~DestinationPath()
{
}

bool DestinationPath::setData(const QString& value, Role role, TagItem* item)
{
    const bool result=EditableString::setData(value,role,item);
    if (result) {
        ValueCache::instance()->insert(column(),value);
    }
    return result;
}

QCompleter* DestinationPath::getCompleter () const
{
    Completers::instance()->createCompleterBoth(column());
    FormularCache::instance()->refreshModel(column());
    ValueCache::instance()->refreshModel(column());
    return Completers::instance()->completers[column()];
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Path
//
//////////////////////////////////////////////////////////////////////////////////////////////

Path::Path ()
    : orgValue(0l),
      updValue(0l)
{
}

Path::~Path ()
{
}

QString Path::data (Role role) const
{
    DirTree::Dir* act=(isUpdated())?updValue:orgValue;

    if (role==RoleFormular || (hasFormular() && (role==RolePersistence || role==RoleDisplayFormularFirst))) {
        return Editable::data(role);
    }
    else if (role==RoleDisplay || role==RoleSorting || role==RoleDisplayFormularFirst) {
        if (act==0l) return QString();
        return act->getDistinctName();
    }
    else if (role==RolePersistence) {
        if (act==0l) return QString();
        return act->getName();
    }
    else if (role==RoleAsRawVariable) {
        if (orgValue==0l) return QString();
        return orgValue->getName();
    }
    else if (role==RoleAsVariable) {
        if (act==0l) return QString();
        return act->getName();
    }

    return QString();
}

QString Path::doData (bool) const
{
    return QString();
}

bool Path::setFileRelatedData (const QString& v_, Role role, TagItem* item)
{
    QString value=v_;

    // Check the value first -> is it a relative path? Then add the commonPath to it
    // An relative path is also assumed when just slashes are contained
    if (QDir::isRelativePath(value) || value.count('/')==value.size()) {
        value=model(item)->getDirCache()->getCommonPath()+QDir::separator()+v_;
    }
    value=QDir::cleanPath(value);

    DirTree::Dir* newDir = model(item)->getDirCache()->store(value,item);

    if (role==RoleInitialization) {
        reset(item);
        if (orgValue!=0l && orgValue!=newDir) orgValue->remove(item);
        orgValue=newDir;
        setUpdated(false,item);
        return true;
    }
    if (orgValue==newDir) {
        reset(item);
        setUpdated(false,item);
        return true;
    }
    if (updValue!=newDir) {
        if (updValue!=0l) updValue->remove(item);
        updValue=newDir;
        setUpdated(true,item);
        return true;
    }
    return false;
}

QCompleter* Path::getCompleter() const
{
    if (Completers::instance()->completers[column()]==NULL) {
        DirCompleterModel* model=new DirCompleterModel();
        FormularCache::instance()->registerModel(column(),1,model);
        Completers::instance()->completers[column()]=Completers::instance()->createCompleter(model);
    }
    FormularCache::instance()->refreshModel(column());
    return Completers::instance()->completers[column()];
}

void Path::removeFromDirTree (TagItem* item)
{
    if (updValue!=0l) {
        updValue->remove(item);
    }
    if (orgValue!=0l) orgValue->remove(item);
}

void Path::reset (TagItem* item)
{
    if (updValue!=0l) {
        updValue->remove(item);
        updValue=0l;
    }
    FileRelated::reset(item);
}

void Path::setUpdated (bool updated, TagItem* item)
{
    if (updated==false && isUpdated()==true && updValue!=0l) {
        orgValue->remove(item);
        orgValue=updValue;
        updValue=0l;
    }
    FileRelated::setUpdated(updated,item);
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Filename
//
//////////////////////////////////////////////////////////////////////////////////////////////

Filename::Filename ()
{
}

Filename::~Filename ()
{
}

QString Filename::data (Role role) const
{
    // The filename should not be simplified in case of sorting
    if (role==RoleSorting) return doData(!isUpdated());
    return FileRelated::data(role);
}

QString Filename::doData (bool original) const
{
    if (original) return orgValue;
    return updValue;
}

namespace {
QString getExtension (const QString& filename)
{
    const int pos=filename.lastIndexOf('.');
    if (pos==-1) {
        return QString();
    }
    return filename.right(filename.size()-pos);
}
}

bool Filename::setFileRelatedData (const QString& v_, Role role, TagItem* item)
{
    QString value=v_;

    // Check if an extension is given. If not take the extension fro the original string
    if (value.endsWith('.')) {
        value.chop(1);
    }
    if (getExtension(value).isEmpty()) {
        value+=getExtension(orgValue);
    }

    if (role==RoleInitialization) {
        orgValue=value;
        updValue.clear();
        setUpdated(false,item);
    }
    else {
        updValue=value;
        setUpdated(true,item);
    }
    return true;
}

void Filename::setUpdated (bool updated, TagItem* item)
{
    if (updated==false && isUpdated()==true) {
        orgValue=updValue;
        updValue="";
    }
    FileRelated::setUpdated(updated,item);
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Artist
//
//////////////////////////////////////////////////////////////////////////////////////////////

Artist::Artist ()
    : orgValue(0l),
      updValue(0l)
{
}

Artist::~Artist ()
{
}

QString Artist::doData (bool original) const
{
    if (original && orgValue!=0l) return orgValue->getName();
    if (!original && updValue!=0l) return updValue->getName();
    return QString();
}

bool Artist::doSetData(const QString& value, Role role, TagItem* item)
{
    const ArtistManager::Artist* artist=MasterManager::instance()->getArtistManager()->storeName(value,item);

    if (role==RoleInitialization && orgValue!=artist) {
        orgValue=artist;
        return true;
    }
    else if (artist==orgValue) {
        if (isUpdated()) {
            updValue=artist;
            setUpdated(false,item);
            return true;
        }
        else return false;
    }
    else if (artist!=updValue || role==RoleFormular) {
        updValue=artist;
        setUpdated(true, item);
        return true;
    }
    return false;
}

void Artist::setUpdated (bool updated, TagItem* item)
{
    if (updated==false && isUpdated()==true) {
        orgValue=updValue;
    }
    Editable::setUpdated(updated,item);
}

const QStringList& Artist::getProposals (Role) const
{
    const static QStringList empty;
    const ArtistManager::Artist* artist=(updValue==0l)?orgValue:updValue;
    if (artist!=0l) return artist->getProposals();
    return empty;
}

QCompleter* Artist::getCompleter() const
{
    if (Completers::instance()->completers[column()]==NULL) {
        CompleterModel* model=MasterManager::instance()->getArtistManager()->getModel();
        FormularCache::instance()->registerModel(column(),1,model);
        Completers::instance()->completers[column()]=Completers::instance()->createCompleter(model);
    }
    FormularCache::instance()->refreshModel(column());
    return Completers::instance()->completers[column()];
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Title
//
//////////////////////////////////////////////////////////////////////////////////////////////

Title::Title ()
{
}

Title::~Title()
{
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Album
//
//////////////////////////////////////////////////////////////////////////////////////////////

Album::Album ()
{
}

Album::~Album ()
{
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Comment
//
//////////////////////////////////////////////////////////////////////////////////////////////

Comment::Comment ()
{
}

Comment::~Comment ()
{
}


bool Comment::setData(const QString& value, Role role, TagItem* item)
{
    const bool result=EditableString::setData(value,role,item);
    if (result) {
        ValueCache::instance()->insert(column(),value);
    }
    return result;
}

QCompleter* Comment::getCompleter () const
{
    Completers::instance()->createCompleterBoth (column());
    FormularCache::instance()->refreshModel(column());
    ValueCache::instance()->refreshModel(column());
    return Completers::instance()->completers[column()];
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Genre
//
//////////////////////////////////////////////////////////////////////////////////////////////

Genre::Genre()
{
}

Genre::~Genre ()
{
}

QCompleter* Genre::getCompleter() const
{
    if (Completers::instance()->completers[column()]==NULL) {
        Completers::instance()->createCompleterFormular (column());
        QCompleter* c=Completers::instance()->completers[column()];
        CompleterModel* model=dynamic_cast<CompleterModel*>(c->model());

        QStringList rows;
        const TagLib::StringList genres=TagLib::ID3v1::genreList();
        for (TagLib::StringList::ConstIterator iter=genres.begin();iter!=genres.end();iter++) {
            rows << QString((*iter).toCString());
        }
        qStableSort (rows);

        model->setData(0,rows);
    }
    FormularCache::instance()->refreshModel(column());
    ValueCache::instance()->refreshModel(column());
    return Completers::instance()->completers[column()];
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Track
//
//////////////////////////////////////////////////////////////////////////////////////////////

Track::Track ()
{
}

Track::~Track()
{
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// CDNumber
//
//////////////////////////////////////////////////////////////////////////////////////////////


CDNumber::CDNumber ()
{
    setIntValue(0,RoleInitialization,0l);
}

CDNumber::~CDNumber ()
{
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Rating
//
//////////////////////////////////////////////////////////////////////////////////////////////


Rating::Rating ()
{
    setIntValue(0,RoleInitialization,0l);
}

Rating::~Rating ()
{
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Year
//
//////////////////////////////////////////////////////////////////////////////////////////////

Year::Year ()
{
}

Year::~Year ()
{
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Length
//
//////////////////////////////////////////////////////////////////////////////////////////////

Length::Length ()
{
}

Length::~Length ()
{
}

QString Length::data (Role role) const
{
    if (role==RolePersistence) {
        return time.toString("H:mm:ss");
    }

    return  KGlobal::locale()->formatLocaleTime (time,KLocale::TimeFoldHours|KLocale::TimeDuration);
}

bool Length::setData(const QString& value, Role role, TagItem* item)
{
    if (role==Column::RoleInitialization) {
        return setIntValue(value.toInt(),role,item);
    }
    return false;
}

bool Length::setIntValue (int value, Role role, TagItem*)
{
    if (role==RoleInitialization) {
        time=QTime(0,0,0,0).addSecs(value);
        return true;
    }
    return false;
}

int Length::getIntValue (Role) const
{
    return time.hour()*3600+time.minute()*60+time.second();
}


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Bitrate
//
//////////////////////////////////////////////////////////////////////////////////////////////

Bitrate::Bitrate ()
{
}

Bitrate::~Bitrate ()
{
}

QString Bitrate::data (Role) const
{
    return QString::number(bitrate);
}

bool Bitrate::setData(const QString& value, Role role, TagItem*)
{
    if (role==RoleInitialization) {
        bitrate=value.toInt();
        return true;
    }
    return false;
}

bool Bitrate::setIntValue (int value, Role role, TagItem*)
{
    if (role==RoleInitialization) {
        bitrate=value;
        return true;
    }
    return false;
}

int Bitrate::getIntValue (Role) const
{
    return bitrate;
}


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Channels
//
//////////////////////////////////////////////////////////////////////////////////////////////

Channels::Channels ()
{
}

Channels::~Channels ()
{
}

QString Channels::data (Role role) const
{
    if (role==Column::RolePersistence) {
        return QString::number(channels);
    }
    else {
        if (channels==1) return "mono";
        if (channels==2) return "stereo";
    }
    return QString::number(channels);
}

bool Channels::setData(const QString& value, Role role, TagItem*)
{
    if (role==RoleInitialization) {
        channels=value.toInt();
        return true;
    }
    return false;
}

bool Channels::setIntValue (int value, Role role, TagItem*)
{
    if (role==RoleInitialization) {
        channels=(unsigned char)value;
        return true;
    }
    return false;
}

int Channels::getIntValue (Role) const
{
    return channels;
}


///////////////////////////////////////////////////////////////////////////////////////////////
//
// SampleRate
//
//////////////////////////////////////////////////////////////////////////////////////////////

SampleRate::SampleRate ()
{
}

SampleRate::~SampleRate ()
{
}

QString SampleRate::data (Role) const
{
    return QString::number(sampleRate);
}

bool SampleRate::setData(const QString& value, Role role, TagItem*)
{
    if (role==RoleInitialization) {
        sampleRate=value.toInt();
        return true;
    }
    return false;
}

bool SampleRate::setIntValue (int value, Role role, TagItem*)
{
    if (role==RoleInitialization) {
        sampleRate=value;
        return true;
    }
    return false;
}

int SampleRate::getIntValue (Role) const
{
    return sampleRate;
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// FileSize
//
//////////////////////////////////////////////////////////////////////////////////////////////

FileSize::FileSize()
{
}

FileSize::~FileSize ()
{
}

QString FileSize::data (Role role) const
{
    if (role==RolePersistence) {
        return QString::number(filesize);
    }
    return KGlobal::locale()->formatByteSize(filesize);
}

bool FileSize::setData(const QString& value, Role role, TagItem*)
{
    if (role==RoleInitialization) {
        filesize=value.toInt();
        return true;
    }
    return false;
}

bool FileSize::setIntValue (int value, Role role, TagItem*)
{
    if (role==RoleInitialization) {
        filesize=value;
        return true;
    }
    return false;
}

int FileSize::getIntValue (Role) const
{
    return filesize;
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class completions
//
//////////////////////////////////////////////////////////////////////////////////////////////

#define COMPLETE_CLASS(CLAZZ,NAME,NUMBER) \
    QString CLAZZ ::name () const { return NAME; }\
    int CLAZZ ::column() const { return NUMBER; }

COMPLETE_CLASS(Searchpattern,"Searchpattern",ColumnManager::Pattern)
COMPLETE_CLASS(DestinationPath,"Destination Path",ColumnManager::DestinationPath)
COMPLETE_CLASS(Path,"Path",ColumnManager::Path)
COMPLETE_CLASS(Filename,"Filename",ColumnManager::Filename)
COMPLETE_CLASS(Artist,"Artist",ColumnManager::Artist)
COMPLETE_CLASS(Title,"Title",ColumnManager::Title)
COMPLETE_CLASS(Album,"Album",ColumnManager::Album)
COMPLETE_CLASS(Comment,"Comment",ColumnManager::Comment)
COMPLETE_CLASS(Genre,"Genre",ColumnManager::Genre)
COMPLETE_CLASS(Track,"Track",ColumnManager::Track)
COMPLETE_CLASS(CDNumber,"CD#",ColumnManager::CD)
COMPLETE_CLASS(Rating,"Rating",ColumnManager::Rating)
COMPLETE_CLASS(Year,"Year",ColumnManager::Year)
COMPLETE_CLASS(Length,"Length",ColumnManager::Length)
COMPLETE_CLASS(Bitrate,"Bitrate",ColumnManager::Bitrate)
COMPLETE_CLASS(Channels,"Channels",ColumnManager::Channels)
COMPLETE_CLASS(SampleRate,"Samplerate",ColumnManager::SampleRate)
COMPLETE_CLASS(FileSize,"FileSize",ColumnManager::FileSize)

#undef COMPLETE_CLASS
