/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#ifndef COLUMNS_H
#define COLUMNS_H

#include <QTime>
#include <QMap>
#include <QFlags>

#include "model/columnsbase.h"
#include "model/tagitem.h"
#include "model/problem.h"

#include "model/structure/artistmanager.h"

namespace Texel
{

namespace Columns
{

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Base class for all static (uneditable) columns
//
//////////////////////////////////////////////////////////////////////////////////////////////

class Static:public Column
{
public:
    Static ();
    virtual ~Static();

    virtual bool isEditable () const;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Base class for all static integer (uneditable) columns
//
//////////////////////////////////////////////////////////////////////////////////////////////

class StaticInt:public Static
{
public:
    StaticInt ();
    virtual ~StaticInt();

    virtual int getIntValue (Role role) const=0;
    virtual bool setIntValue (int value, Role role, TagItem* item)=0;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Base class for all editable columns
//
//////////////////////////////////////////////////////////////////////////////////////////////

class Editable:public Column
{
public:
    Editable ();
    virtual ~Editable();

    bool setData(const QString& value_, Role role, TagItem* item);
    virtual QString data (Role role) const;

    virtual bool isEditable () const;

    bool isUpdated() const;
    virtual void setUpdated (bool updated, TagItem* item);
    virtual void reset (TagItem* item);

    bool isCorrect() const;
    void setCorrect (bool correct);
    bool isChecked() const;
    void setChecked (bool checked);
    bool hasProposals () const;
    void setHasProposals (bool hasProposals);

    bool hasFormular() const;

    InputProblem* getProblem() const;
    void setProblem (InputProblem* p);

    void evaluate(TagItem* item);

protected:
    QString evaluateFormula (bool reevaluate, TagItem* item);

    virtual bool doSetData(const QString& value_, Role role, TagItem* item)=0;
    virtual QString doData (bool original) const=0;

private:
    QString formular;
    InputProblem* problem;
    enum Flags { FlagIsUpdated=1, FlagIsCorrect=2, FlagIsChecked=4, FlagHasProposals=8 };
    QFlags<Flags> flags;
};

inline InputProblem* Editable::getProblem() const
{
    return problem;
}

inline void Editable::setProblem(InputProblem* p)
{
    delete problem;
    problem=p;
}

inline bool Editable::isUpdated() const
{
    return flags.testFlag(FlagIsUpdated);
}

inline bool Editable::hasFormular() const
{
    return formular!="";
}

inline bool Editable::isCorrect() const
{
    return flags.testFlag(FlagIsCorrect);
}

inline void Editable::setCorrect (bool correct)
{
    if (flags.testFlag(FlagIsCorrect)!=correct) {
        flags^=FlagIsCorrect;
    }
}

inline bool Editable::isChecked() const
{
    return flags.testFlag(FlagIsChecked);
}

inline void Editable::setChecked (bool checked)
{
    if (flags.testFlag(FlagIsChecked)!=checked) {
        flags^=FlagIsChecked;
    }
}

inline bool Editable::hasProposals () const
{
    return flags.testFlag(FlagHasProposals);
}

inline void Editable::setHasProposals (bool hasProposals)
{
    if (flags.testFlag(FlagHasProposals)!=hasProposals) {
        flags^=FlagHasProposals;
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Base class for all editable Integer Columns
//
//////////////////////////////////////////////////////////////////////////////////////////////

class EditableInt: public Editable
{
public:
    EditableInt();
    virtual ~EditableInt();

    int getIntValue (Role role) const;
    bool setIntValue (int value, Role role, TagItem* item);
    virtual void setUpdated (bool updated, TagItem* item);

    QString doData (bool original) const;

protected:
    virtual bool doSetData(const QString& value, Role role, TagItem* item);

private:
    int orgValue;
    int updValue;
};

inline int EditableInt::getIntValue(Role role) const
{
    if (role==Column::RoleAsRawVariable || !isUpdated()) {
       return orgValue;
    }
    return updValue;
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Base class for all editable String Columns
//
//////////////////////////////////////////////////////////////////////////////////////////////

class EditableString: public Editable
{
public:
    EditableString();
    virtual ~EditableString();

    QString doData (bool original) const;
    virtual void setUpdated (bool updated, TagItem* item);
    virtual void reset (TagItem* item);

 protected:
    virtual bool doSetData(const QString& value, Role role, TagItem* item);

private:
    QString orgValue;
    QString updValue;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Base class for filename related columns
//
//////////////////////////////////////////////////////////////////////////////////////////////

class FileRelated: public Editable
{
public:
    FileRelated();
    virtual ~FileRelated();

    virtual bool doSetData(const QString& value, Role role, TagItem* item);

protected:
    virtual bool setFileRelatedData (const QString& value, Role role, TagItem* item)=0;

private:
    void changeFilename (const QString& value, Role role, TagItem* item);
};

/*********************************************************************************************
 *********************************************************************************************
 **
 ** Concrete column classes
 **
 *********************************************************************************************
 ********************************************************************************************/

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Searchpattern
//
//////////////////////////////////////////////////////////////////////////////////////////////

class Searchpattern:public Column
{
public:
    Searchpattern();
    virtual ~Searchpattern();

    virtual QString data (Role role) const;
    bool setData(const QString& value, Role role, TagItem* item);

    virtual QString name () const;
    virtual int column() const;

    bool containsVar (const QString& name) const;
    QString getVar (const QString& name) const;
    QString suggestVarName (const QString& name) const;
    QStringList varNames () const;

    virtual bool isEditable () const;
    void parse (bool reevaluate, TagItem* item);

    QCompleter* getCompleter () const;

    InputProblem* getProblem() const;

private:
    InputProblem* problem;
    QString searchpattern;
    QMap<QString,QString> variables;
};

inline QStringList Searchpattern::varNames () const
{
    return variables.keys();
}

inline bool Searchpattern::containsVar (const QString& name) const
{
    return variables.contains(name);
}

inline QString Searchpattern::getVar (const QString& name) const
{
    return variables.value(name,"");
}

inline InputProblem* Searchpattern::getProblem() const
{
    return problem;
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// DestinationPath
//
//////////////////////////////////////////////////////////////////////////////////////////////

class DestinationPath: public EditableString
{
public:
    DestinationPath();
    virtual ~DestinationPath();

    virtual bool setData(const QString& value, Role role, TagItem* item);

    virtual QString name () const;
    virtual int column() const;

    QCompleter* getCompleter () const;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Path
//
//////////////////////////////////////////////////////////////////////////////////////////////

class Path:public FileRelated
{
public:
    Path();
    virtual ~Path();

    virtual QString data (Role role) const;

    virtual QString name () const;
    virtual int column() const;

    DirTree::Dir* getPathPointer (Role role) const;
    QCompleter* getCompleter() const;
    virtual void reset (TagItem* item);

    void setUpdated (bool updated, TagItem* item);

    void removeFromDirTree (TagItem* item);

protected:
    virtual bool setFileRelatedData (const QString& value, Role role, TagItem* item);
    virtual QString doData (bool original) const;

private:
    DirTree::Dir* orgValue;
    DirTree::Dir* updValue;
};

inline DirTree::Dir* Path::getPathPointer (Role role) const
{
     DirTree::Dir* const ret=(role==RolePersistence || role==RoleAsRawVariable || updValue==0l)?orgValue:updValue;
     if (ret==0) {
         qFatal ("Undefined path pointer in item, can't be.");
     }
     return ret;
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Filename
//
//////////////////////////////////////////////////////////////////////////////////////////////

class Filename: public FileRelated
{
public:
    Filename();
    virtual ~Filename();

    virtual QString data (Role role) const;
    virtual QString doData (bool original) const;

    virtual QString name () const;
    virtual int column() const;

    void setUpdated (bool updated, TagItem* item);

protected:
    virtual bool setFileRelatedData (const QString& value, Role role, TagItem* item);

private:
    QString orgValue;
    QString updValue;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Artist
//
//////////////////////////////////////////////////////////////////////////////////////////////

class Artist: public Editable
{
public:
    Artist();
    virtual ~Artist();

    virtual QString name () const;
    virtual int column() const;

    virtual bool doSetData(const QString& value, Role role, TagItem* item);
    QString doData (bool original) const;

    const QStringList& getProposals (Role) const;
    void setUpdated (bool updated, TagItem* item);

    virtual QCompleter* getCompleter() const;

private:
    const ArtistManager::Artist* orgValue;
    const ArtistManager::Artist* updValue;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Title
//
//////////////////////////////////////////////////////////////////////////////////////////////

class Title: public EditableString
{
public:
    Title();
    virtual ~Title();

    virtual QString name () const;
    virtual int column() const;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Album
//
//////////////////////////////////////////////////////////////////////////////////////////////

class Album:public EditableString
{
public:
    Album();
    virtual ~Album();

    virtual QString name () const;
    virtual int column() const;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Comment
//
//////////////////////////////////////////////////////////////////////////////////////////////

class Comment:public EditableString
{
public:
    Comment();
    virtual ~Comment();

    virtual bool setData(const QString& value, Role role, TagItem* item);

    virtual QString name () const;
    virtual int column() const;

    QCompleter* getCompleter () const;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Genre
//
//////////////////////////////////////////////////////////////////////////////////////////////

class Genre:public EditableString
{
public:
    Genre();
    virtual ~Genre();

    virtual QString name () const;
    virtual int column() const;

    virtual QCompleter* getCompleter() const;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Track
//
//////////////////////////////////////////////////////////////////////////////////////////////

class Track:public EditableInt
{
public:
    Track();
    virtual ~Track();

    virtual QString name () const;
    virtual int column() const;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// CDNumber
//
//////////////////////////////////////////////////////////////////////////////////////////////

class CDNumber:public EditableInt
{
public:
    CDNumber();
    virtual ~CDNumber();

    virtual QString name () const;
    virtual int column() const;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Rating
//
//////////////////////////////////////////////////////////////////////////////////////////////

class Rating:public EditableInt
{
public:
    Rating();
    virtual ~Rating();

    virtual QString name () const;
    virtual int column() const;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Year
//
//////////////////////////////////////////////////////////////////////////////////////////////

class Year:public EditableInt
{
public:
    Year();
    virtual ~Year();

    virtual QString name () const;
    virtual int column() const;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Length
//
//////////////////////////////////////////////////////////////////////////////////////////////

class Length:public StaticInt
{
public:
    Length();
    virtual ~Length();

    virtual QString data (Role role) const;
    virtual QString name () const;

    virtual int column() const;
    virtual bool setData(const QString& value, Role role, TagItem* item);
    virtual bool setIntValue (int value, Role role, TagItem* item);
    virtual int getIntValue (Role role) const;

private:
    QTime time;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Bitrate
//
//////////////////////////////////////////////////////////////////////////////////////////////

class Bitrate:public StaticInt
{
public:
    Bitrate ();
    virtual ~Bitrate();

    virtual QString data (Role role) const;
    virtual bool setData(const QString& value, Role role, TagItem* item);
    virtual bool setIntValue (int value, Role role, TagItem* item);
    virtual int getIntValue (Role role) const;

    virtual QString name () const;
    virtual int column() const;

private:
    int bitrate;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Channels
//
//////////////////////////////////////////////////////////////////////////////////////////////

class Channels:public StaticInt
{
public:
    Channels ();
    virtual ~Channels();

    virtual QString data (Role role) const;
    virtual bool setData(const QString& value, Role role, TagItem* item);
    virtual bool setIntValue (int value, Role role, TagItem* item);
    virtual int getIntValue (Role role) const;

    virtual QString name () const;
    virtual int column() const;

private:
    unsigned char channels;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// SampleRate
//
//////////////////////////////////////////////////////////////////////////////////////////////

class SampleRate:public StaticInt
{
public:
    SampleRate();
    virtual ~SampleRate();

    virtual QString data (Role role) const;
    virtual bool setData(const QString& value, Role role, TagItem* item);
    virtual bool setIntValue (int value, Role role, TagItem* item);
    virtual int getIntValue (Role role) const;

    virtual QString name () const;
    virtual int column() const;

private:
    int sampleRate;
};

///////////////////////////////////////////////////////////////////////////////////////////////
//
// FileSize
//
//////////////////////////////////////////////////////////////////////////////////////////////

class FileSize:public StaticInt
{
public:
    FileSize ();
    virtual ~FileSize();

    virtual QString data (Role role) const;
    virtual bool setData(const QString& value, Role role, TagItem* item);
    virtual bool setIntValue (int value, Role role, TagItem* item);
    virtual int getIntValue (Role role) const;

    virtual QString name () const;
    virtual int column() const;

private:
    unsigned long filesize;
};

}

}

#endif // COLUMNS_H
