/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "dirtree.h"
#include "aid.h"
#include "tagitem.h"

#include <QStringList>
#include <QReadLocker>
#include <QWriteLocker>
#include <QFileInfo>

using namespace std;

namespace Texel {

    quint32 DirTree::Dir::s_UniqIdSequence=1;

    DirTree::DirTree ()
        : QObject(),
        m_Locker(),
        m_CommonPart(0l),
        m_Root(new Dir(this,0l,""))
    {
    }

    DirTree::~DirTree ()
    {
        delete m_Root;
    }

    DirTree::Dir* DirTree::addAdditionalInformationFile (const QString &file, AdditionalFileFormats format)
    {
        QFileInfo info(file);
        Dir* dir=getPath(info.absolutePath());
        if (dir!=0L) {
            dir->m_AdditionalFiles.append(Dir::MetaData(format,info.fileName()));
        }
        return dir;
    }

    DirTree::Dir* DirTree::store (const QString& name, TagItem* item)
    {
        if (name=="" || name=="/") {
            cout << "Stop" << endl;
        }

        if (name==m_CachedDir) {
            if (item!=0l) {
                QWriteLocker locker(&m_Locker);
                m_CachedDirPointer->m_Items.insert(item);
            }
            return m_CachedDirPointer;
        }

        Dir* current=0l;
        {
            QWriteLocker locker(&m_Locker);
            QStringList parts=QDir::cleanPath(name).split(QDir::separator(),QString::SkipEmptyParts);
            current=m_Root;
            foreach (QString part, parts) {
                current = current->child(part);
            }
            if (item!=0l) {
                current->m_Items.insert(item);
            }

            m_CommonPart=0l;
        }
        emit_Delayed();

        if (current==0l) {
            qFatal ("Could not store path in dirtree.");
        }

        // Update cache information (return same pointer when next call gives the same name)
        m_CachedDir=name;
        m_CachedDirPointer=current;

        return current;
    }

    DirTree::Name DirTree::getFileName (const QString& fileName) const
    {
        QReadLocker locker(&m_Locker);
        const QStringList parts=fileName.split(QDir::separator(),QString::SkipEmptyParts);
        if (parts.size()==0) return Name(m_Root,"");

        Dir* current=m_Root;
        for (int i=0;i<parts.size()-1;i++) {
            current = current->child(parts[i]);
            if (current==0l) return Name(0l,parts.last());
        }
        return Name(current,parts.last());
    }

    DirTree::Dir* DirTree::getPath (const QString& path) const
    {
        const QStringList parts=path.split(QDir::separator(),QString::SkipEmptyParts);
        QReadLocker locker(&m_Locker);
        Dir* current=m_Root;
        for (int i=0;i<parts.size();i++) {
            current = current->getChild(parts[i]);
            if (current==0l) return 0l;
        }
        return current;
    }

    bool DirTree::contains(const QString& path) const
    {
        return getPath(path)!=0l;
    }

    void DirTree::remove (const QString& name, TagItem* item)
    {
        QWriteLocker locker(&m_Locker);
        {
            Dir* current=m_Root;
            if (m_CachedDir==name) {
                current=m_CachedDirPointer;
            }
            else {
                QStringList parts=QDir::cleanPath(name).split(QDir::separator(),QString::SkipEmptyParts);
                foreach (QString part, parts) {
                    current = current->child(part);
                }
                m_CachedDir=name;
                m_CachedDirPointer=current;
            }

            if (item!=0l) {
                current->m_Items.remove(item);
                while (current->m_Items.empty()) {
                    Dir* toDelete = current;
                    current=current->m_Parent;
                    current->remove(toDelete);
                    if (toDelete==m_CachedDirPointer) {
                        m_CachedDir=QString();
                        m_CachedDirPointer=0l;
                    }
                }
            }
        }

        emit_Delayed();
    }

    DirTree::Dir::Dir (DirTree* _tree, Dir* parent, const QString& name)
        : m_Parent(parent),
        m_NameFragment(name),
        tree(_tree),
        m_UniqId(createUniqId())
    {
        if (parent!=0l) {
            tree->emit_hasTracksInList(getName(),true);
        }
    }

    DirTree::Dir::~Dir ()
    {
        if (m_Parent!=0l) {
            //tree->m_DirectMap.remove(getName());
            tree->emit_hasTracksInList(getName(),false);
        }
        foreach (Dir* child, m_Children) delete child;
    }

    void DirTree::Dir::getTagItems(TagItem::List& list)
    {
        foreach (Dir* child, m_Children) {
            child->getTagItems (list);
        }
        list.append (m_Items.toList());
    }

    TagItem::List DirTree::Dir::getTagItems () const
    {
        return m_Items.toList();
    }

    QString DirTree::Dir::getName () const
    {
        if (m_Parent==0) return QDir::rootPath();

        if (m_CompleteName.isEmpty()) {
            m_CompleteName=m_Parent->getName()+m_NameFragment+QDir::separator();
        }
        return m_CompleteName;
    }

    QString DirTree::Dir::getDistinctName () const
    {
        return m_CompleteName.mid(tree->getCommonPath().size());
    }

    DirTree::Dir* DirTree::Dir::getChild (const QString& name) const
    {
        QMap<QString,Dir*>::const_iterator iter=m_Children.find(name);
        if (iter!=m_Children.end()) {
            return *iter;
        }
        return 0l;
    }

    DirTree::Dir* DirTree::Dir::child (const QString& name)
    {
        QMap<QString,Dir*>::const_iterator iter=m_Children.find(name);
        if (iter==m_Children.end()) {
            return *m_Children.insert(name,new Dir(tree,this,name));
        }
        return *iter;
    }

    void DirTree::Dir::remove (Dir* child)
    {
        const QString name = child->m_NameFragment;
        delete child;
        m_Children.remove(name);
    }

    QString DirTree::getCommonPath ()
    {
        QReadLocker locker(&m_Locker);
        calcCommonPath();
        return m_CommonPart->getName();
    }

    void DirTree::calcCommonPath ()
    {
        if (m_CommonPart==0l) {
            m_CommonPart=m_Root;
            while (m_CommonPart->m_Children.size()==1 && m_CommonPart->m_Items.isEmpty()) m_CommonPart=*(m_CommonPart->m_Children.begin());
        }
    }

    TagItem::List DirTree::getTagItems (const QString& path,bool recursive) const
    {
        QReadLocker locker(&m_Locker);
        QList<TagItem*> items;

        Dir* dir = getPath(path);
        if (dir!=0l) {
            if (!recursive) {
                items = dir->m_Items.toList();
            }
            else {
                dir->getTagItems(items);
            }
        }

        return items;
    }

    void DirTree::Dir::remove (TagItem* item)
    {
        m_Items.remove(item);
        if (m_Items.empty() && m_Children.isEmpty()) {
            m_Parent->remove(this);
        }
    }

    void DirTree::emit_Delayed ()
    {
        foreach (ChangedDir pair, m_DelayedEmits) {
            emit hasTracksInList (pair.first,pair.second);
        }
        m_DelayedEmits.clear();
    }

    void DirTree::emit_hasTracksInList (const QString& dir, bool has)
    {
        m_DelayedEmits << ChangedDir(dir,has);
    }

}
