/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#ifndef texel_DIRTREE_H
#define texel_DIRTREE_H

#include <QMap>
#include <QSet>
#include <QString>
#include <QReadWriteLock>
#include <QObject>

#include "aid.h"

namespace Texel {

    class TagItem;

    class DirTree: public QObject
    {
        Q_OBJECT

    public:
        enum AdditionalFileFormats { CDDB, Image };

        class Dir
        {
            friend class DirTree;

        public:
            struct MetaData: public QPair<DirTree::AdditionalFileFormats, QString>
            {
                MetaData (DirTree::AdditionalFileFormats a, const QString& b);

                QString getType () const;
                void setType (const QString& type);
            };

            typedef QList<MetaData> MetaDataList;

            QString getName () const;
            QString getDistinctName () const;
            QList<TagItem*> getTagItems () const;

            void remove (TagItem* item);
            bool hasAdditionalInformationFiles () const;
            MetaDataList getMetaData () const;

            QList<Dir*> getChildren () const;
            quint32 getUniqId () const;
            void setUniqId (quint32 id);

        private:
            Dir* child (const QString& name);
            Dir* getChild (const QString& name) const;
            Dir (DirTree* tree, Dir* parent, const QString& name);
            ~Dir ();

            void remove (Dir* child);
            void getTagItems (QList<TagItem*>& list);

            Dir* m_Parent;
            // mutable because it is a calulated and than cached value
            mutable QString m_CompleteName;
            QString m_NameFragment;
            DirTree* tree;
            QMap<QString,Dir*> m_Children;
            QSet<TagItem*> m_Items;

            MetaDataList m_AdditionalFiles;
            quint32 m_UniqId;

            static quint32 s_UniqIdSequence;
            static quint32 createUniqId();
            static void adjustUniqId (quint32 id);
        };

        /** Little structure which defines a complete filename The directory is coded within
     * this DirTree structure, the filename itself is provided as a plain string.
     * To enable storing in a Map or Set a non aphanumerical sorting is provided.
     */
        struct Name
        {
            Name (const Dir* dir_=0l, const QString& name_="")
                : dir(dir_),
                name(name_)
            {
            }

            Name (const Name& other)
                : dir(other.dir),
                name(other.name)
            {
            }

            const Dir* dir;
            QString name;

            int operator< (const Name& other) const
            {
                return (dir==other.dir)?(name<other.name):(dir<other.dir);
            }

            bool operator!= (const Name& other) const
            {
                return dir!=other.dir || name!=other.name;
            }

            bool operator== (const Name& other) const
            {
                return dir==other.dir && name==other.name;
            }

            Name& operator= (const Name& other)
            {
                dir=other.dir;
                name=other.name;
                return *this;
            }

        };

        DirTree ();
        ~DirTree ();

        /** The common path is the path fragment which is equal for all pathes currently
          stored in this tree. */
        QString getCommonPath();

        Dir* store (const QString& dir, TagItem* item);
        void remove (const QString& dir, TagItem* item);

        // Returns the directory pointer to which the meta data file is added
        Dir* addAdditionalInformationFile (const QString& file, AdditionalFileFormats format);

        bool contains (const QString& name) const;
        DirTree::Name getFileName (const QString& path) const;
        DirTree::Dir* getPath (const QString& path) const;
        DirTree::Dir* getRoot () const;

        QList<TagItem*> getTagItems (const QString& path, bool recursive) const;

        void lock ();
        void unlock ();

    signals:
        void hasTracksInList (const QString& dir, bool has);

    private:
        void calcCommonPath();
        void emit_hasTracksInList (const QString& dir, bool has);

        /* Lock to synchronize a dirtree. Locks are necessary as the loading of
      objects takes place in a separate thread. However, due to the rare
      calls to this class the mutex is used for the whole structure when
      calling a function.
   */
        mutable QReadWriteLock m_Locker;
        QString m_CachedDir;
        Dir* m_CachedDirPointer;
        Dir* m_CommonPart;
        Dir* m_Root;

        typedef QPair<QString,bool> ChangedDir;
        QList<ChangedDir> m_DelayedEmits;
        void emit_Delayed ();

        friend class Dir;
    };

    inline std::ostream& operator<< (std::ostream& out, const DirTree::Name& name)
    {
        out << name.dir << ":" << name.name;
        return out;
    }

    inline bool DirTree::Dir::hasAdditionalInformationFiles () const
    {
        return !m_AdditionalFiles.isEmpty();
    }

    inline DirTree::Dir::MetaDataList DirTree::Dir::getMetaData () const
    {
        return m_AdditionalFiles;
    }

    inline QList<DirTree::Dir*> DirTree::Dir::getChildren () const
    {
        return m_Children.values();
    }

    inline quint32 DirTree::Dir::createUniqId()
    {
        return ++s_UniqIdSequence;
    }

    inline quint32 DirTree::Dir::getUniqId () const
    {
        return m_UniqId;
    }

    inline void DirTree::Dir::setUniqId (quint32 id)
    {
        m_UniqId=id;
        adjustUniqId(id);
    }

    inline void DirTree::Dir::adjustUniqId (quint32 id)
    {
        if (id>s_UniqIdSequence) {
            s_UniqIdSequence=id+1;
        }
    }

    inline DirTree::Dir* DirTree::getRoot() const
    {
        return m_Root;
    }

    inline void DirTree::lock ()
    {
        m_Locker.lockForRead();
    }

    inline void DirTree::unlock ()
    {
        m_Locker.unlock();
    }

    inline DirTree::Dir::MetaData::MetaData (DirTree::AdditionalFileFormats a, const QString& b)
        : QPair<DirTree::AdditionalFileFormats, QString>(a,b)
    {
    }

    inline QString DirTree::Dir::MetaData::getType () const
    {
        switch (first) {
        case DirTree::CDDB:
            return "CDDB";
        case DirTree::Image:
            return "Image";
        default:
            return "unknown";
        }
    }

    inline void DirTree::Dir::MetaData::setType (const QString& type)
    {
        if (type=="CDDB") {
            first=DirTree::CDDB;
        }
        else if (type=="Image") {
            first=DirTree::Image;
        }
    }

}

#endif
