/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "jobtracker.h"
#include "util/widgets/tooltip.h"

#include <QStyle>
#include <QPushButton>
#include <QProgressBar>
#include <QHelpEvent>
#include <QLabel>

#include <khbox.h>
#include <klocale.h>
#include <kiconloader.h>

#include "aid.h"

namespace Texel {

    JobTracker::JobInfo::JobInfo ()
        : percent(0),
          percentageUnset(true)
    {
    }

    JobTracker::JobTracker()
        : KAbstractWidgetJobTracker(),
        m_Tracker(0l),
        m_Tooltip(0l),
        m_UnsetJobs(0)
    {
        createTrackerWidget();
    }

    JobTracker::~JobTracker()
    {
        foreach (KJob* job, m_ControlledJobs.keys()) {
            job->kill();
        }

        delete m_Tracker;
    }

    void JobTracker::createTrackerWidget ()
    {
        KHBox* box = new KHBox();
        box->setSpacing(0);
        box->setMargin(0);

        m_ProgressBar = new QProgressBar(box);
        m_ProgressBar->setMinimum(0);
        m_ProgressBar->setMaximum(100);
        m_ProgressBar->setMinimumWidth(300);
        m_ProgressBar->setMaximumWidth(300);
        m_ProgressBar->setValue(0);
        m_ProgressBar->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Fixed);
        m_ProgressBar->installEventFilter(this);

        m_KillButton = new QPushButton(box);
        m_KillButton->setIcon(KIconLoader::global()->loadIcon("process-stop",KIconLoader::Toolbar));
        m_KillButton->setToolTip(i18n("Abort job"));
        m_KillButton->setFlat(true);
        m_KillButton->setEnabled(false);

        m_Tracker=box;
    }

    QWidget* JobTracker::widget (KJob*)
    {
        return m_Tracker;
    }

    void JobTracker::registerJob (KJob *job)
    {
        KAbstractWidgetJobTracker::registerJob(job);

        if (m_ControlledJobs.size()==0) {
            m_Tracker->show();
        }

        if (!m_ControlledJobs.contains(job)) {
            m_ControlledJobs.insert (job, JobInfo());
            ++m_UnsetJobs;
            m_ProgressBar->setMaximum(0);

            if (m_ControlledJobs.size()==1) {
                m_KillButton->setEnabled(true);
                connect (m_KillButton, SIGNAL(clicked()), job, SLOT(kill()));
            }

            setTooltipText ();
        }
    }

    void JobTracker::unregisterJob (KJob *job)
    {
        if (m_ControlledJobs.contains(job)) {
            m_ProgressBar->setValue(m_ProgressBar->value()-m_ControlledJobs[job].percent);
            if (m_ControlledJobs[job].percentageUnset) {
                --m_UnsetJobs;
            }
            m_ControlledJobs.remove(job);

            if (m_ControlledJobs.size()==0) {
                m_KillButton->setEnabled(false);
                m_ProgressBar->setMaximum(1);
                m_ProgressBar->reset();
                disconnect (m_KillButton, SIGNAL(clicked()), job, SLOT(kill()));
                emit done();
            }
            else if (m_UnsetJobs==0) {
                m_ProgressBar->setMaximum (100*m_ControlledJobs.size());
            }

            setTooltipText ();
        }

        KAbstractWidgetJobTracker::unregisterJob(job);
    }

    void JobTracker::description (KJob *job, const QString &title, const QPair<QString, QString>& p1, const QPair<QString, QString>& p2)
    {
        KAbstractWidgetJobTracker::description(job,title,p1,p2);
        if (m_ControlledJobs.contains(job)) {
            m_ControlledJobs[job].desc=title;
        }
    }

    void JobTracker::percent (KJob *job, unsigned long percent)
    {
        KAbstractWidgetJobTracker::percent(job,percent);

        if (m_ControlledJobs.contains(job)) {
            if (m_ControlledJobs[job].percentageUnset) {
                m_ControlledJobs[job].percentageUnset=false;
                --m_UnsetJobs;
                if (m_UnsetJobs==0) {
                    m_ProgressBar->setMaximum (100*m_ControlledJobs.size());
                }
            }

            const long delta = percent-m_ControlledJobs[job].percent;
            if (delta>0) {
                m_ControlledJobs[job].percent=percent;
                m_ProgressBar->setValue (m_ProgressBar->value()+delta);
                emit percentExecuted(m_ProgressBar->value()/m_ControlledJobs.size());
            }

            setTooltipText ();
        }
    }

    bool JobTracker::eventFilter (QObject* watched, QEvent* event)
    {
        if (watched==m_ProgressBar && event->type()==QEvent::ToolTip) {
            m_Tooltip=new QLabel(m_ProgressBar);
            setTooltipText();
            connect (m_Tooltip, SIGNAL(destroyed()), this, SLOT(closeTooltip()));
            Tooltip::show (m_Tooltip, static_cast<QHelpEvent*>(event)->globalPos(), QRect(), m_ProgressBar);
        }

        return false;
    }

    void JobTracker::setTooltipText ()
    {
        if (m_Tooltip!=0l) {
            QString text;
            if (m_ControlledJobs.empty()) {
                text=i18n("Currently no jobs in execution");
            }
            else {
                QList<JobInfo> values = m_ControlledJobs.values();
                for (int i=0;i<values.size();i++) {
                    if (i>0) text+="\n";
                    text+=QString::number(values[i].percent)+"% - "+values[i].desc;
                }
            }
            m_Tooltip->setText(text);
        }
    }

    void JobTracker::closeTooltip ()
    {
        if ((void*)sender()==(void*)m_Tooltip) {
            m_Tooltip=0l;
        }
    }

}
