/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "modelloader.h"
#include "model/texeltagmodel.h"
#include "model/dirtree.h"
#include "func/m3u/m3ulist.h"
#include "util/loghook.h"
#include "options/preferences.h"

#include <QFileInfo>
#include <QDir>
#include <QCoreApplication>
#include <QThread>
#include <QPair>

#include <klocale.h>

#include "aid.h"

using namespace std;

namespace Texel {

namespace {

struct FilesCollector
{
    QStringList tracks;
    QStringList playlists;
    DirTree::Dir::MetaDataList additionalFiles;
    DirTree tree;

    QStringList trackPatterns;
    QStringList playlistPatterns;

    FilesCollector ()
    : m_ItemsToLoad(0)
    {
    }

    unsigned long getItemsToLoad () const
    {
        return m_ItemsToLoad;
    }

    void expandPath (const QUrl& path)
    {
        QFileInfo info(path.toLocalFile());
        if (info.isDir()) {
            tree.store(info.canonicalFilePath(),0l);
            recGetAllFilesInDir (path);
        }
        else {
            addFile(info);
        }
    }

    bool addAdditionalFile (const QFileInfo& info)
    {
        if (info.suffix().toLower()=="jpg" || info.suffix().toLower()=="png") {
            additionalFiles << DirTree::Dir::MetaData(DirTree::Image,info.filePath());
            ++m_ItemsToLoad;
            return true;
        }

        QFile f(info.absoluteFilePath());
        f.open(QIODevice::ReadOnly);
        QByteArray b=f.read(10);
        // Check for CDDB Entry
        QRegExp cddbPattern("^\\s*#\\s*xmcd",Qt::CaseInsensitive);
        if (cddbPattern.indexIn(b)!=-1) {
            additionalFiles << DirTree::Dir::MetaData(DirTree::CDDB,info.filePath());
            ++m_ItemsToLoad;
            return true;
        }

        return false;
    }

private:
    unsigned long m_ItemsToLoad;

    void recGetAllFilesInDir (const QUrl& dir)
    {
        QDir d(dir.toLocalFile());

        foreach (const QFileInfo& info, d.entryInfoList())
        {
            if (info.fileName()!=QString(".") && info.fileName()!=QString("..")) {
                if (info.isDir()) {
                    recGetAllFilesInDir (info.filePath());
                }
                else {
                    addFile(info);
                }
            }
        }
    }

    void addFile (const QFileInfo& info)
    {
        if (info.isFile() && info.isReadable()) {
            if (playlistPatterns.contains(info.suffix())) {
                playlists << info.filePath();
                ++m_ItemsToLoad;
            }
            else if (trackPatterns.size()==0 || trackPatterns.contains(info.suffix())) {
                tracks << info.filePath();
                ++m_ItemsToLoad;
            }
            // Test for file content and add when known
            else {
                addAdditionalFile(info);
            }
        }
    }
};
}

class ModelLoader::Private: public QThread
{
    Q_OBJECT

public:
    Private (ModelLoader* job)
        : QThread(job),
          m_Stopped(false),
          m_Job(job)
    {
    }

    virtual ~Private()
    {
    }

    void setModel(TagModel *_model)
    {
        model=_model;
    }

    bool addFileNames (const QList<QUrl>& filelist)
    {
        const bool running = isRunning();
        if (!running) directories+=filelist;
        return !running;
    }

    volatile bool m_Stopped;

protected:
    virtual void run ()
    {
        if (model==0l) {
            Q_ASSERT_X(false, "Modelloader", "No model set, aborting");
            return;
        }
        m_Stopped=false;

        FilesCollector collector;
        collector.trackPatterns = Preferences::instance()->io.m_TrackPatterns.split(";",QString::SkipEmptyParts);
        collector.playlistPatterns = Preferences::instance()->io.m_PlaylistPatterns.split(";",QString::SkipEmptyParts);

        int inputDirectories=0;
        foreach (QUrl dir, directories)
        {
            emit m_Job->description(m_Job,i18n("Gather all files to process in directory %1",dir.toLocalFile()));
            collector.expandPath (dir);
            inputDirectories++;
            if (m_Stopped) return;
        }

        const QString commonPath=collector.tree.getCommonPath();
        emit m_Job->description(m_Job,i18n("Reading %1 entries from %2 directories [%3]", collector.tracks.size(), inputDirectories, commonPath));

        int itemsLoaded=0;

        //
        // Load tracks and signal them to the core model
        //

        TagItem::List extension;
        foreach (QString i, collector.tracks) {
            try {
                TagItem* item = TagItem::createTagItem(model,i);
                extension.append(item);
                if (extension.size()>=100) {
                    emit m_Job->loadedItems (extension);
                    extension.clear();
                }
            }
            catch (TagItem::InvalidFormat e) {
                const QFileInfo info(i);
                // File is not an known music file. Try again as additional file. Necessary
                // when no filename pattern for music files is defined.
                if (!collector.addAdditionalFile(info)) {
                    LogHook::instance()->error(e.what(), 0l);
                }
            }
            catch (std::runtime_error e) {
                LogHook::instance()->error(e.what(), 0l);
            }
            m_Job->emitPercent(++itemsLoaded,collector.getItemsToLoad());
            if (m_Stopped) break;
        }

        if (!extension.isEmpty()) {
            emit m_Job->loadedItems (extension);
        }

        directories.clear();

        //
        // Load playlists after all files are loaded
        //

        emit m_Job->description(m_Job,i18n("Reading %1 playlists from %2 directories [%3]",
                                           collector.playlists.size(),
                                           inputDirectories,
                                           commonPath));

        foreach (QString playlistName, collector.playlists) {
            if (m_Stopped) break;
            emit m_Job->loadPlaylist(playlistName);
            m_Job->emitPercent(++itemsLoaded,collector.getItemsToLoad());
        }

        //
        // All other files at last
        //

        emit m_Job->description(m_Job,i18n("Reading %1 additional information from %2 directories [%3]",
                                           collector.additionalFiles.size(),
                                           inputDirectories,
                                           commonPath));

        foreach (DirTree::Dir::MetaData f, collector.additionalFiles) {
            if (m_Stopped) break;
            emit m_Job->loadAdditionalFile(f.second,f.first);
            m_Job->emitPercent(++itemsLoaded,collector.getItemsToLoad());
        }

    }

private:
    TagModel* model;
    ModelLoader* m_Job;

    QList<QUrl> directories;
};

ModelLoader::ModelLoader (TagModel* parent)
    : KJob(parent),
      m_Thread(new Private(this))
{
    m_Thread->setModel(parent);
    setCapabilities(KJob::Killable);

    connect( m_Thread, SIGNAL(finished()), this, SLOT(onFinished()));
}

ModelLoader::~ModelLoader()
{
    if (m_Thread->isRunning()) {
        kill();
    }
    delete m_Thread;
}

void ModelLoader::onFinished ()
{
    emitResult();
}

void ModelLoader::start ()
{
    m_Thread->start();
}

bool ModelLoader::addFileNames (const QList<QUrl>& filelist)
{
    return m_Thread->addFileNames(filelist);
}

bool ModelLoader::doKill ()
{
    if ( m_Thread->isRunning() ) {
        m_Thread->m_Stopped = true;
        m_Thread->wait();
        return true;
    }
    else {
        return false;
    }
}

}

#include "modelloader.moc"
