/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "tagmodelxmlstreamer.h"
#include "model/tagitem.h"
#include "model/texeltagmodel.h"
#include "model/dirtree.h"
#include "util/loghook.h"
#include "func/m3u/playlistmodel.h"
#include "util/sqlitewrapper.h"

#include <QFile>

#include <auto_ptr.h>

#include "aid.h"

using namespace std;

namespace Texel {

namespace Internal {

    struct job_killed {};

class ArchiveSaveWorker::Private
{
    SqLite::Statement* m_InsertRelation;
    SqLite::Statement* m_InsertPlaylist;
    SqLite::Statement* m_InsertTracks;
    SqLite::Statement* m_InsertDirectory;
    SqLite::Statement* m_InsertMetaData;

    ArchiveSave* m_Job;
    TagModel& m_Model;

public:
    TagItem::List m_ItemsToSave;
    SqLite::Connection* m_db;
    QString m_FileName;

    Private (TagModel& model, const QString& filename, ArchiveSave* job)
        : m_InsertRelation(0l),
          m_InsertPlaylist(0l),
          m_InsertTracks(0l),
          m_InsertDirectory(0l),
          m_InsertMetaData(0l),
          m_Job(job),
          m_Model(model),
          m_FileName(filename)
    {
    }

    void clearContent () throw (SqLite::Exception)
    {
        LogHook::instance()->message("Remove old content from file",0l);
        m_db->prepare("DELETE FROM tagitems;")->executeOnce();
        m_db->prepare("DELETE FROM playlists;")->executeOnce();
        m_db->prepare("DELETE FROM directories;")->executeOnce();
    }

    bool createStructure () throw (SqLite::Exception)
    {
        std::auto_ptr<SqLite::Statement> header(m_db->prepare("SELECT 1 FROM sqlite_master WHERE type='table' AND name='header';")->execute());
        if (header->hasRow()) {
            // Header table exists, check for compliant version
            std::auto_ptr<SqLite::Statement> version(m_db->prepare("SELECT version FROM header;"));
            if (version->hasRow()) {
                QString ver = version->getString(1);
                if (ver!="1.0") {

                }
            }
            else {

            }
            clearContent ();
            return false;
        }
        else {
            // Create Structure as necessary
            try {
                // If the table exists already the version is not inserted again.
                m_db->prepare ("CREATE TABLE header ("
                               "version TEXT"
                               ");"
                               )->executeOnce();
                m_db->prepare("INSERT INTO header VALUES('1.0');")->executeOnce();
            }
            catch (SqLite::Exception) {
            }

            m_db->prepare("CREATE TABLE IF NOT EXISTS directories ("
                          "id INT PRIMARY KEY,"
                          "dirname TEXT"
                          ");"
                          )->executeOnce();

            m_db->prepare("CREATE TABLE IF NOT EXISTS metadata ("
                          "dirid INT REFERENCES directories(id) ON DELETE CASCADE,"
                          "filename TEXT,"
                          "type TINYINT"
                          ");"
                          )->executeOnce();

            m_db->prepare("CREATE TABLE IF NOT EXISTS tagitems ("
                           "id INTEGER PRIMARY KEY ASC,"
                          "type TINYINT,"
                          "pattern TEXT,"
                          "destinationpath TEXT,"
                           "path INT REFERENCES directories(id) ON DELETE CASCADE,"
                          "filename TEXT,"
                           "artist TEXT,"
                           "title TEXT,"
                           "album TEXT,"
                           "comment TEXT,"
                           "track TINYINT,"
                           "cd SMALLINT,"
                           "Rating TINYINT,"
                           "Year SMALLINT,"
                           "Genre TEXT,"
                           "Length SMALLINT,"
                           "Bitrate SMALLINT,"
                           "Channels TINYINT,"
                           "SampleRate SMALLINT,"
                           "FileSize INT,"
                           "MarkForDeletion TINYINT,"

                          // now the modify columns which are filled when an overwrite did take place
                          "mod_pattern TEXT,"
                          "mod_destinationpath TEXT,"
                           "mod_filename TEXT,"
                           "mod_path TEXT,"
                           "mod_artist TEXT,"
                           "mod_title TEXT,"
                           "mod_album TEXT,"
                           "mod_comment TEXT,"
                           "mod_track TINYINT,"
                           "mod_cd SMALLINT,"
                           "mod_rating TINYINT,"
                           "mod_year SMALLINT,"
                           "mod_genre TEXT"
                          ");"
                          )->executeOnce();

            m_db->prepare("CREATE TABLE IF NOT EXISTS relations ("
                          "rowid INTEGER PRIMARY KEY AUTOINCREMENT,"
                          "id_f INT REFERENCES tagitems(id) ON DELETE CASCADE,"
                          "id_t INT REFERENCES tagitems(id) ON DELETE CASCADE,"
                          "type TINYINT"
                          ");"
                          )->executeOnce();

            m_db->prepare("CREATE TABLE IF NOT EXISTS playlists ("
                          "ID INT4 PRIMARY KEY,"
                          "filename TEXT,"
                          "loaded_from_file TEXT,"
                          "dirty TINYINT,"
                          "saved_on_disc TINYINT,"
                          "playlist_in_local_dir TINYINT,"
                          "altered_tracks INT,"
                          "format TINYINT"
                          ");"
                          )->executeOnce();

            m_db->prepare("CREATE TABLE IF NOT EXISTS playlist2tracks ("
                          "rowid INTEGER PRIMARY KEY AUTOINCREMENT,"
                          "playlist INT4 REFERENCES playlists(id) ON DELETE CASCADE,"
                          "extended TINYINT,"
                          "dirty TINYINT,"
                          "title TEXT,"
                          "filename TEXT,"
                          "length SMALLINT,"
                          "track REFERENCES tagitems(id) ON DELETE CASCADE"
                          ");"
                          )->executeOnce();
        }
        return true;
    }

    void savePlaylist (M3U::List playlist) throw (SqLite::Exception)
    {
        m_InsertPlaylist->bind()(playlist.getUniqId())       (playlist.getFileName())
                                (playlist.getLoadFileName()) (playlist.isDirty())
                                (playlist.wasSaved())        (playlist.isLocal())
                                (playlist.alteredTracks())   (playlist.isExtended()).execute();

        foreach (M3U::List::Track t, playlist.getTracks()) {
            m_InsertTracks->bind(1);
            m_InsertTracks->bind(2, playlist.getUniqId());
            m_InsertTracks->bind(3, t.extended);
            m_InsertTracks->bind(4, t.dirty);
            m_InsertTracks->bind(5, t.title);
            m_InsertTracks->bind(6, t.filename);
            m_InsertTracks->bind(7, t.length);
            if (t.item!=0l) {
                m_InsertTracks->bind(8, t.item->getUniqId());
            }
            else m_InsertTracks->bind(8);
            m_InsertTracks->execute();
            m_InsertTracks->reset();
        }
    }

    void saveTag (SqLite::Statement* s, TagItem* item) throw (SqLite::Exception)
    {
        s->bind()(item->getUniqId())((int)item->type());

        ColumnManager::ColumnNames rows[] = { ColumnManager::Track, ColumnManager::CD, ColumnManager::Rating, ColumnManager::Year, ColumnManager::Length, ColumnManager::Bitrate, ColumnManager::Channels, ColumnManager::SampleRate, ColumnManager::FileSize, ColumnManager::Pattern, ColumnManager::DestinationPath, ColumnManager::Filename, ColumnManager::Artist, ColumnManager::Title, ColumnManager::Album, ColumnManager::Comment, ColumnManager::Genre, ColumnManager::Invalid};
        for (int i=0;rows[i]!=ColumnManager::Invalid;i++) {
            if (ColumnManager::instance()->getSortType(rows[i])==QMetaType::Int) {
                s->bind (rows[i]+3, item->getIntValue(rows[i]));
            }
            else {
                s->bind (rows[i]+3, item->data(rows[i],Columns::Column::RoleAsRawVariable));
            }
            // Modified values are always strings as these may contain a formular
            if (ColumnManager::instance()->isEditable(rows[i])) {
                if (item->isAltered(rows[i])) {
                    s->bind(rows[i]+22,item->data(rows[i],Columns::Column::RolePersistence));
                }
                else {
                    s->bind(rows[i]+22);
                }
            }
        }
        s->bind (21, item->isToDelete()?1:0);

        // Do not save the path as string but just the id of the dir element
        s->bind (2+3, item->getPathPointer(Columns::Column::RoleAsRawVariable)->getUniqId());
        if (item->isAltered(ColumnManager::Path)) {
            s->bind(2+22,item->data(ColumnManager::Path,Columns::Column::RolePersistence));
        }

        s->execute();
        s->reset();
    }

    void saveDuplicateRelations (TagItem* item)
    {
        foreach (TagItem* related, item->getRelations()->getDuplicates()) {
            m_InsertRelation->bind()(item->getUniqId())(related->getUniqId()).execute();
        }
    }

    void saveDirectories (DirTree::Dir* d)
    {
        foreach (DirTree::Dir* d1, d->getChildren()) {
            saveDirectories(d1);
        }

        m_InsertDirectory->bind()(d->getUniqId())(d->getName()).execute();

        foreach (DirTree::Dir::MetaData data, d->getMetaData()) {
            m_InsertMetaData->bind()(d->getUniqId())(data.second)((int)data.first).execute();
        }
    }

    void run ()
    {
        PlaylistModel* playlistModel=dynamic_cast<PlaylistModel*>(m_Model.getPlayListModel());
        m_Job->setRows(m_Model.rowCount()+playlistModel->rowCount());

        try {
            m_db=new SqLite::Connection (m_FileName);
            createStructure();
            m_db->prepare("BEGIN EXCLUSIVE TRANSACTION;")->executeOnce();

            // Create necessary statements
            m_InsertRelation=m_db->prepare("INSERT INTO relations (id_f,id_t,type) VALUES (?,?,1);");
            m_InsertDirectory=m_db->prepare("INSERT INTO directories VALUES (?,?);");
            m_InsertMetaData=m_db->prepare("INSERT INTO metadata VALUES (?,?,?);");

            // Save directories
            saveDirectories(m_Model.getDirCache()->getRoot());

            std::auto_ptr<SqLite::Statement> replace(m_db->prepare("INSERT OR REPLACE INTO tagitems VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);"));

            // Save tracks
            foreach (TagItem* item, m_ItemsToSave) {
                saveTag (replace.get(), item);
                saveDuplicateRelations(item);
                if (m_Job->m_Stopped==true) {
                    throw Internal::job_killed();
                }
                m_Job->incRows();
            }

            // Save playlists
            m_InsertPlaylist=m_db->prepare("INSERT OR REPLACE INTO playlists VALUES (?,?,?,?,?,?,?,?);");
            m_InsertTracks=m_db->prepare("INSERT INTO playlist2tracks VALUES (?,?,?,?,?,?,?,?);");
            for (int i=0;i<playlistModel->rowCount();i++) {
                savePlaylist (playlistModel->getPlaylist(i));
                if (m_Job->m_Stopped==true) {
                    throw Internal::job_killed();
                }
                m_Job->incRows();
            }

            m_db->commit();
            emit m_Job->saveListStatisticHint (m_FileName, true, 1 /*TODO*/);
        }
        catch (SqLite::Exception e) {
            m_db->rollback();
            emit m_Job->errorOccured (m_FileName,e.msg);
            LogHook::instance()->error (i18n("Error [%1] with sqlite db (%3): %2",e.errno,e.msg,m_FileName),0l);
        }
        catch (Internal::job_killed) {
            m_db->rollback();
        }

        delete m_db;
        m_db=0L;

        delete m_InsertRelation;
        delete m_InsertPlaylist;
        delete m_InsertTracks;
        delete m_InsertDirectory;
        delete m_InsertMetaData;
    }
};

class ArchiveLoadWorker::Private
{
public:
    Private (TagModel& model, const QString& archivename, ArchiveLoad* job)
        :  m_Model(model),
          m_Job(job),
          m_ArchiveName (archivename)
    {
    }

    TagItem* getItemFromRow (SqLite::Statement* stmt) throw (SqLite::Exception)
    {
        TagItemType::Type type = (TagItemType::Type)stmt->getInt(1);
        TagItem* item=TagItem::createTagItemByType(&m_Model,type);

        if (item!=0l) {
            item->setUniqId(stmt->getInt(0));
            item->setToDelete(stmt->getInt(20)==1);

            ColumnManager::ColumnNames rows[] = { ColumnManager::Track, ColumnManager::CD, ColumnManager::Rating, ColumnManager::Year, ColumnManager::Length, ColumnManager::Bitrate, ColumnManager::Channels, ColumnManager::SampleRate, ColumnManager::FileSize, ColumnManager::Pattern, ColumnManager::DestinationPath, ColumnManager::Filename, ColumnManager::Artist, ColumnManager::Title, ColumnManager::Album, ColumnManager::Comment, ColumnManager::Genre, ColumnManager::Invalid};
            for (int i=0;rows[i]!=ColumnManager::Invalid;i++) {
                item->setDataDirect(rows[i],stmt->getString(rows[i]+2),Columns::Column::RoleInitialization);
                if (ColumnManager::instance()->isEditable(rows[i]) && !stmt->isNull(rows[i]+21)) {
                    item->setDataDirect(rows[i],stmt->getString(rows[i]+21),Columns::Column::RoleReInitalization);
                }
            }

            // load Path from 35th element (d.dirname)
            item->setDataDirect(ColumnManager::Path,stmt->getString(34),Columns::Column::RoleInitialization);
            if (ColumnManager::instance()->isEditable(ColumnManager::Path) && !stmt->isNull(ColumnManager::Path+21)) {
                item->setDataDirect(ColumnManager::Path,stmt->getString(ColumnManager::Path+21),Columns::Column::RoleReInitalization);
            }

            item->reevaluate();
        }

        return item;
    }

    void loadPlaylists (const QMap<quint32, TagItem*>& byUniqId) throw (SqLite::Exception)
    {
        QMap<quint32, QList<M3U::List::Track> > tracks;
        std::auto_ptr<SqLite::Statement> select_tracks(m_db->prepare("SELECT * FROM playlist2tracks ORDER BY rowid;")->execute());
        while (select_tracks->hasRow()) {
            M3U::List::Track t;
            const quint32 id = select_tracks->getInt(0);
            t.extended=select_tracks->getInt(1)==1;
            t.dirty=select_tracks->getInt(2)==1;
            t.title=select_tracks->getString(3);
            t.filename=select_tracks->getString(4);
            t.length=select_tracks->getInt(5);
            quint32 tagitem=select_tracks->getInt(6);
            if (byUniqId.contains(tagitem)) {
                t.item=byUniqId[tagitem];
            }
            else {
                t.item=0l;
            }
            if (!tracks.contains(id)) {
                QList<M3U::List::Track> l;
                l << t;
                tracks.insert(id,l);
            }
            else {
                tracks[id] << t;
            }
            select_tracks->execute();
        }

        PlaylistModel* playlistModel=dynamic_cast<PlaylistModel*>(m_Model.getPlayListModel());
        std::auto_ptr<SqLite::Statement> select_playlists(m_db->prepare("SELECT * FROM playlists")->execute());
        while (select_playlists->hasRow()) {
            const quint32 id = select_playlists->getInt(0);
            M3U::List playlist = M3U::List::createPlaylist(&m_Model, id,
                                                           select_playlists->getString(1),
                                                           select_playlists->getString(2),
                                                           select_playlists->getInt(6),
                                                           select_playlists->getInt(7)==1,
                                                           select_playlists->getInt(3)==1,
                                                           select_playlists->getInt(5)==1,
                                                           select_playlists->getInt(4)==1);
            if (tracks.contains(id)) {
                playlist.setTracks(tracks[id]);
            }
            playlistModel->addPlaylist(playlist);
            select_playlists->execute();
        }
    }

    void loadMetadata ()
    {
        std::auto_ptr<SqLite::Statement> select_metadata(m_db->prepare("SELECT d.dirname, m.filename, m.type FROM metadata m JOIN directories d ON m.dirid=d.id")->execute());
        while (select_metadata->hasRow()) {
            QString filename = select_metadata->getString(0)+QDir::separator()+select_metadata->getString(1);
            DirTree::AdditionalFileFormats type = (select_metadata->getInt(2)==0)?DirTree::CDDB:DirTree::Image;
            m_Model.getDirCache()->addAdditionalInformationFile(filename,type);
            select_metadata->execute();
        }
    }

    void run ()
    {
        try {
            m_db = new SqLite::Connection (m_ArchiveName);

            QMap<quint32, TagItem*> byUniqId;

            // Detect size of items to load
            auto_ptr<SqLite::Statement> count(m_db->prepare("SELECT (SELECT count(*) FROM tagitems);")->execute());
            if (count->hasRow()) {
                m_Job->setRows(count->getInt(0));
            }

            // Load Tag items
            TagItem::List list;
            auto_ptr<SqLite::Statement> statement(m_db->prepare("SELECT t.*, d.dirname FROM tagitems t JOIN directories d ON t.path=d.id;")->execute());
            while (statement->hasRow()) {
                TagItem* item = getItemFromRow(statement.get());
                if (item!=0l) {
                    list << item;
                    byUniqId.insert(item->getUniqId(),item);
                    if (list.size()>=100) {
                         emit m_Job->newTags(list);
                         list.clear();
                    }
                }
                statement->execute();
                m_Job->incRows();
            }

            emit m_Job->newTags(list);

            // Add duplicate relations
            auto_ptr<SqLite::Statement> relations(m_db->prepare("SELECT * FROM relations;")->execute());
            while (relations->hasRow()) {
                if (relations->getInt(2)==1) {
                    const quint32 from = relations->getInt(0);
                    const quint32 to = relations->getInt(1);
                    if (byUniqId.contains(from) && byUniqId.contains(to)) {
                        byUniqId[from]->getRelations()->addDuplicate(byUniqId[to]);
                    }
                }
                relations->execute();
            }

            // Load Playlists
            loadPlaylists(byUniqId);

            loadMetadata();

            emit m_Job->saveListStatisticHint (m_ArchiveName, false, 1);
        }
        catch (SqLite::Exception e) {
            emit m_Job->errorOccured (m_ArchiveName,e.msg);
            LogHook::instance()->error (i18n("Error [%1] with sqlite db (%3): %2",e.errno,e.msg,m_ArchiveName),0l);
        }
        catch (Internal::job_killed) {
        }
    }

private:
    TagModel& m_Model;
    ArchiveLoad* m_Job;
    QString m_ArchiveName;
    SqLite::Connection* m_db;
};

}
    TagListArchive::TagListArchive (QThread* worker)
        : m_Thread (worker),
          m_Stopped(false),
          m_ProcessedRows(0),
          m_NumRows(0)
    {
        setCapabilities(KJob::Killable);
        connect( m_Thread, SIGNAL(finished()), this, SLOT(onFinished()));
    }

    TagListArchive::~TagListArchive ()
    {
        if (m_Thread->isRunning()) {
            kill();
        }
        delete m_Thread;
    }

    void TagListArchive::incRows ()
    {
        emitPercent (++m_ProcessedRows,m_NumRows);
    }

    void TagListArchive::setRows(int rows)
    {
        m_NumRows=rows;
    }

    bool TagListArchive::doKill ()
    {
        if ( m_Thread->isRunning() ) {
            m_Stopped = true;
            m_Thread->wait();
            return true;
        }
        else {
            return false;
        }
    }

    void TagListArchive::onFinished ()
    {
        emitResult();
    }

    ArchiveSave::ArchiveSave (const QString& filename, Texel::TagModel& model)
        : TagListArchive(new Internal::ArchiveSaveWorker(model,filename,this)),
          m_Model(model)
    {
    }

    ArchiveSave::~ArchiveSave ()
    {
    }

    void ArchiveSave::onFinished ()
    {
        foreach (TagItem* item, static_cast<Internal::ArchiveSaveWorker*>(m_Thread)->p->m_ItemsToSave) {
            item->setLockedForUpdate(false);
        }
        TagListArchive::onFinished();
    }

    void ArchiveSave::start ()
    {
        Internal::ArchiveSaveWorker* thread=static_cast<Internal::ArchiveSaveWorker*>(m_Thread);
        thread->p->m_ItemsToSave=m_Model.getTagItems();

        foreach (TagItem* item, thread->p->m_ItemsToSave) {
            item->setLockedForUpdate(true);
        }

        thread->start();
    }

    ArchiveLoad::ArchiveLoad (const QString& filename, Texel::TagModel& model)
        : TagListArchive(new Internal::ArchiveLoadWorker(model,filename,this)),
        m_Model(model)
    {
    }

    ArchiveLoad::~ArchiveLoad ()
    {
    }

    void ArchiveLoad::start ()
    {
        m_Model.removeRows(0,m_Model.rowCount());
        m_Thread->start();
    }

    namespace Internal {
        ArchiveSaveWorker::ArchiveSaveWorker (TagModel& model, const QString& filename, ArchiveSave* job)
            : QThread(),
              p(new Private(model,filename,job))
        {
            setObjectName ("Sqlite Save");
        }

        ArchiveSaveWorker::~ArchiveSaveWorker ()
        {
            delete p;
        }

        void ArchiveSaveWorker::run ()
        {
            p->run();
        }

        ArchiveLoadWorker::ArchiveLoadWorker (TagModel& model, const QString& filename, ArchiveLoad* job)
            : QThread(),
              p(new Private(model,filename,job))
        {
            setObjectName ("Sqlite Load");
        }

        ArchiveLoadWorker::~ArchiveLoadWorker ()
        {
            delete p;
        }

        void ArchiveLoadWorker::run ()
        {
            p->run();
        }
    }
} // namespace Texel

