#include "updatetagsjob.h"
#include "model/tagitem.h"
#include "util/loghook.h"

#include <QThread>

#include <kio/copyjob.h>
#include <klocale.h>

namespace Texel {

    struct UpdateTagsJob::Private: public QThread
    {
        Q_OBJECT

    public:
        Private (UpdateTagsJob* job)
            : QThread(job),
            m_Job(job)
        {
        }

        virtual void run ()
        {
            m_Stopped=false;
            DestinationMap m_CopySet;

            int updated=0;
            foreach (TagItem* item,m_ToUpdate) {
                if (m_Stopped) {
                    return;
                }
                try {
                    QString* emptyDirToDelete=item->update(m_EnforceRename,true);
                    if (emptyDirToDelete!=0l) {
                        emit m_Job->directoryToDelete(*emptyDirToDelete);
                        delete emptyDirToDelete;
                    }

                    if (item->isToDelete()) {
                        LogHook::instance()->message(i18n("Deleted file ",item->getAbsolutFilename(Columns::Column::RoleAsRawVariable)), 0l);
                        emit m_Job->removedTag(item);
                    }
                    else {
                        emit m_Job->updatedTag(item);
                    }
                    emit m_Job->changeUnsavedItems(-1);

                    //
                    // Check if copying to another path is desired and do it
                    //
                    QString copyDestinations = item->data(ColumnManager::DestinationPath,Columns::Column::RoleAsVariable);
                    if (!copyDestinations.isEmpty()) {
                        foreach (QString dest, copyDestinations.split(':')) {
                            const QString file = item->getAbsolutFilename(Columns::Column::RoleAsVariable);
                            DestinationMap::iterator found = m_CopySet.find(dest);
                            if (found==m_CopySet.end()) {
                                found = m_CopySet.insert(dest, FileSet());
                            }
                            (*found).insert(file);
                        }
                        item->setDataDirect(ColumnManager::DestinationPath,"",Columns::Column::RoleInitialization);
                    }
                }
                catch (std::runtime_error ex) {
                    LogHook::instance()->message(ex.what(), item);
                }
                catch (TagItem::IOError e) {
                    LogHook::instance()->error(e.message, e.tag);
                }
                catch (TagItem::RenameCollision e) {
                    emit m_Job->renameCollision(e.first, e.second);
                }

                m_Job->emitPercent(updated++,m_ToUpdate.size());
            }

            // Execute copy set
            if (m_CopySet.size()>0) {
                int allFiles=0;
                foreach (FileSet files, m_CopySet.values())
                {
                    allFiles+=files.size();
                }
                emit m_Job->emitPercent(0,allFiles);

                int copiedFiles=0;
                foreach (QString destDir, m_CopySet.keys())
                {
                    FileSet set = *m_CopySet.begin();
                    emit m_Job->description (m_Job, i18n("Copy %1 files to directory %2",set.size(),destDir));

                    KUrl::List urls;
                    foreach (QString s, set) {
                        urls << KUrl(s);
                    }

                    KIO::CopyJob* job = KIO::copy(urls,destDir,KIO::HideProgressInfo);
                    job->setAutoDelete(true);
                    job->setWriteIntoExistingDirectories(true);
                    job->setAutoSkip(true);
                    job->exec();

                    copiedFiles+=set.size();
                    emit m_Job->emitPercent(copiedFiles,allFiles);
                }
            }
        }

        volatile bool m_Stopped;

        UpdateTagsJob* m_Job;
        bool m_EnforceRename;

        TagItem::Set m_ToUpdate;

        typedef QSet<QString> FileSet;
        typedef QMap<QString, FileSet> DestinationMap;
    };

    UpdateTagsJob::UpdateTagsJob (QObject* parent)
        : KJob(parent),
        m_Thread(new Private(this))
    {
        setCapabilities(KJob::Killable);
        m_Thread->setObjectName("Update tags");

        connect (m_Thread, SIGNAL(finished()), this, SLOT(onFinished()));
    }

    UpdateTagsJob::~UpdateTagsJob ()
    {
        if (m_Thread->isRunning()) {
            kill();
        }
        delete m_Thread;
    }

    void UpdateTagsJob::onFinished ()
    {
        emitResult();
        deleteLater();
    }

    void UpdateTagsJob::start ()
    {
        foreach (TagItem* tag, m_Thread->m_ToUpdate) {
            tag->setLockedForUpdate(true);
        }

        m_Thread->start();
    }

    bool UpdateTagsJob::doKill ()
    {
        if ( m_Thread->isRunning() ) {
             m_Thread->m_Stopped = true;
             m_Thread->wait();
             return true;
         }
         else {
             return false;
         }
    }

    void UpdateTagsJob::setItemsForUpdate (const TagItem::Set& items)
    {
        m_Thread->m_ToUpdate=items;
    }

    void UpdateTagsJob::setEnforceRename (bool enforce)
    {
        m_Thread->m_EnforceRename=enforce;
    }
}

#include "updatetagsjob.moc"
