#include "mimesupport.h"

#include <model/texeltagmodel.h>

#include <QList>
#include <QSet>
#include <QUrl>
#include <QDataStream>
#include <QByteArray>

#include "aid.h"

namespace Texel {

QString MimeSupport::s_AppSpecificMimeTemplate="application/x-texel;value=\"%1\"";
QString MimeSupport::s_TagItemSubFormat = "ItemPointer";

MimeSupport::MimeSupport():
    QMimeData ()
{
}

QString MimeSupport::getMimeTypeForSubType(const QString &subformat)
{
    return s_AppSpecificMimeTemplate.arg(subformat);
}

bool MimeSupport::hasSubFormat (const QMimeData *data, const QString& subformat)
{
    return data->hasFormat(getMimeTypeForSubType(subformat));
}

bool MimeSupport::hasItems (const QMimeData *data)
{
    return hasSubFormat(data,s_TagItemSubFormat);
}

void MimeSupport::setUrls (TagItem::List items)
{
    QList<QUrl> urls;

    foreach (TagItem* item, items) {
        urls << QUrl(item->getAbsolutFilename().toLocal8Bit());
    }

    QMimeData::setUrls(urls);
}

void MimeSupport::setItems (TagItem::List items)
{
    TagItem** itemArray = new TagItem*[items.size()];
    int i=0;
    foreach (TagItem* item, items) {
        itemArray[i++]=item;
    }

    QByteArray buffer(reinterpret_cast<const char*>(itemArray),items.size()*sizeof(TagItem*));
    QMimeData::setData(getMimeTypeForSubType(s_TagItemSubFormat),buffer);
}

void MimeSupport::setItems (const QModelIndexList &indexes, const TagModel* const model)
{
    if (!indexes.isEmpty()) {
        QSet<int> rows = getRowsFromIndexes(indexes);
        TagItem::List items;
        foreach (int row, rows) {
            items << model->getRow(row);
        }
        setItems (items);
    }
}

void MimeSupport::setIndexList (QList<Index>& indexes, const QString& subtype)
{
    QByteArray internalBuffer;
    QDataStream stream(&internalBuffer, QIODevice::WriteOnly);

    foreach (Index i, indexes) {
        stream << i.first << i.second;
    }

    setData(getMimeTypeForSubType(subtype), internalBuffer);
}

QSet<int> MimeSupport::getRowsFromIndexes (const QModelIndexList& indexes)
{
    QSet<int> rows;

    // To order rows
    foreach (QModelIndex i,indexes) {
        rows.insert(i.row());
    }

    return rows;
}

void MimeSupport::setRowList (const QModelIndexList& indexes, const QString& subtype)
{
    QSet<int> rows = getRowsFromIndexes(indexes);

    int j=0;
    int* source=new int[rows.size()];
    foreach (int row, rows) {
        source[j++]=row;
    }

    QByteArray buffer(reinterpret_cast<const char*>(source),rows.size()*sizeof(int));
    setData(getMimeTypeForSubType(subtype),buffer);
    delete source;
}

QList<int> MimeSupport::getRowList (const QMimeData* data, const QString& subformat)
{
    QByteArray buffer = data->data(getMimeTypeForSubType(subformat));
    const int* const source = reinterpret_cast<const int* const>(buffer.data());

    QList<int> rows;
    for (unsigned i=0;i<buffer.size()/sizeof(int);i++) {
        rows << source[i];
    }
    return rows;
}

QList<MimeSupport::Index> MimeSupport::getIndexList(const QMimeData* data, const QString& subformat)
{
    QList<Index> indexes;
    QByteArray buffer = data->data(getMimeTypeForSubType(subformat));
    QDataStream stream(&buffer,QIODevice::ReadOnly);

    while (!stream.atEnd()) {
        int a,b;
        stream >> a >> b;
        indexes << Index(a,b);
    }

    return indexes;
}

TagItem::List MimeSupport::getItemList (const QMimeData* data)
{
    TagItem::List items;
    if (hasSubFormat(data, s_TagItemSubFormat)) {
        QByteArray buffer = data->data(getMimeTypeForSubType(s_TagItemSubFormat));
        TagItem** itemList = reinterpret_cast<TagItem**>(buffer.data());

        for (unsigned i=0;i<buffer.size()/sizeof(TagItem*);i++) {
            items << itemList[i];
        }
    }
    return items;
}

}

#include "mimesupport.moc"
