/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "mastermanager.h"
#include "artistmanager.h"
#include "model/completermodel.h"

#include "aid.h"

namespace Texel {

ArtistManager::Artist::Artist (const QString& _name)
:name (_name),
 state(0),
 proposals()
{
}

void ArtistManager::Artist::remove (ManagedObject* obj)
{
    const int pos = items.indexOf(obj);
    if (pos!=-1) items.remove(pos);
}

ArtistManager::ArtistManager ()
: model(new CompleterModel(this))
{
}

ArtistManager::~ArtistManager ()
{
    foreach (Artist* artist, storage.values()) delete artist;
}

void ArtistManager::deleteName (const QString& name)
{
   if (!storage.contains(name)) return;

   delete storage[name];
   storage.remove(name);
   model->remove(name);
}

void ArtistManager::changeName (const QString& oldName, const QString& newName)
{
   if (!storage.contains(oldName) || newName=="") return;

   Artist* oldArtist=storage[oldName];
   Artist* newArtist=storeName(newName, 0l);

   foreach (ManagedObject* obj, oldArtist->items) {
       if (oldArtist!=newArtist) newArtist->add(obj);
       obj->changeArtistName(newName);
   }

   deleteName(oldName);
}

void ArtistManager::removeObject (const QString& name, ManagedObject* obj)
{
    QMap<QString,Artist*>::iterator iter = storage.find(name);
    if (iter!=storage.end()) {
        Artist* artist=(*iter);
        artist->remove(obj);
        if (artist->items.size()==0) {
            // Entry not used anymore, remove it
            deleteName (name);
        }
    }
}

ArtistManager::Artist* ArtistManager::storeName (const QString& name, ManagedObject* obj)
{
    Artist* artist=0l;

    QMap<QString,Artist*>::iterator iter = storage.find(name);
    if (iter!=storage.end()) {
        artist=(*iter);
    }
    else {
        artist=new Artist(name);
        storage.insert(name,artist);
        model->insert(name);
    }

    if (iter==storage.end() && !name.isEmpty()) {
        emit newArtist (QVariant::fromValue<Artist*>(artist), name);
    }

    if (obj!=0l) {
        artist->add(obj);
    }

    return artist;
}

void ArtistManager::reemitAll ()
{
    for (QMap<QString, Artist*>::iterator iter=storage.begin();iter!=storage.end();iter++) {
        iter.value()->setState(0);
        emit newArtist (QVariant::fromValue<Artist*>(iter.value()), iter.key());
    }
}

void ArtistManager::resetAll ()
{
    for (QMap<QString, Artist*>::iterator iter=storage.begin();iter!=storage.end();iter++) {
        iter.value()->setState(0);
    }
}

void ArtistManager::newArtistInformation (const QVariant& ref, QStringList proposals)
{
    Artist* artist=ref.value<Artist*>();
    if (artist!=0l) {
        unsigned state = artist->getState()|ResultAvailable;
        artist->proposals=proposals;

        if (proposals.size()>0) {
            if (artist->name==proposals[0]) {
                state|=IsCorrect;
            }
            else if (QString::compare(artist->name.trimmed(),proposals[0],Qt::CaseInsensitive)) {
                state|=MinorSpellErrors;
            }
        }

        if (proposals.size()>0) {
            state|=HasProposals;
        }

        artist->setState(state);
    }
}

}

#include "artistmanager.moc"
