/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "mastermanager.h"
#include "artistmanager.h"

#include <QObject>

#include "func/metatagservices/echonest/searchartist.h"
#include "func/metatagservices/musicbrainz/musicbrainz.h"
#include "options/preferences.h"

#include "util/loghook.h"

#include <ios>

#include "options/preferences.h"

#include <klocale.h>

namespace Texel {

ManagedObject::~ManagedObject()
{
}

MasterManager::MasterManager ()
: QObject(0l),
  m_ArtistManager(new ArtistManager()),
  m_BackgroundService(0l),
  m_BackgroundServiceState(NoBackgroundService)
{
}

MasterManager::~MasterManager ()
{
    delete m_ArtistManager;
    delete m_BackgroundService;
}

void MasterManager::createBackgroundService()
{
    installBackgroundService();
    QObject::connect (Preferences::instance(), SIGNAL(configChanged()), this, SLOT(installBackgroundService()));
}

void MasterManager::installBackgroundService()
{
    // Did the state change in comparsion to the current? If no, simply return
    if ((m_BackgroundServiceState==NoBackgroundService &&
        !Preferences::instance()->musicbrainz.m_Enabled &&
        !Preferences::instance()->echonest.m_Enabled) ||
        (m_BackgroundServiceState==MusicBrainzBackgroundService &&
         Preferences::instance()->musicbrainz.m_Enabled) ||
        (m_BackgroundServiceState==EchonestBackgroundService &&
         Preferences::instance()->echonest.m_Enabled))  return;

    CheckService::ArtistQuery* query=0;

#   ifdef TEXEL_MUSICBRAINZ_SUPPORT
    if (Preferences::instance()->musicbrainz.m_Enabled) {
        query=new CheckService::MusicBrainzQuery(0l);
        m_BackgroundServiceState=MusicBrainzBackgroundService;
    }
    else
#   endif
    if (Preferences::instance()->echonest.m_Enabled) {
        query=new EchoNest::SearchArtist(0l);
        m_BackgroundServiceState=EchonestBackgroundService;
    }
    else {
        m_BackgroundServiceState=NoBackgroundService;
    }

    // Deinstall old service
    if (m_BackgroundService!=0l) {
        delete m_BackgroundService;
        LogHook::instance()->message(i18n("Removed old background service"),0);
        m_BackgroundService=0l;
    }

    if (query!=0l)
    {
        try {
            m_BackgroundService = new CheckService::CheckArtist(query);
            // Set cache parameters before thread starts. Parallel access to preferences singleton
            // avoided
            m_BackgroundService->attach(getArtistManager(),
                                        SIGNAL(newArtist (QVariant, QString)),
                                        SLOT(newArtistInformation (QVariant, QStringList)));
            m_BackgroundService->start();
            getArtistManager()->reemitAll();
            LogHook::instance()->message(i18n("New background service installed"),0);
        }
        catch (std::ios_base::failure error) {
            LogHook::instance()->warning(error.what(),0l);
        }
    }
    else {
        getArtistManager()->resetAll();
    }
}

QPair<QStringList,QStringList> MasterManager::getCacheInformation () const
{
    if (m_BackgroundService!=0l) {
        return m_BackgroundService->getCacheInformation();
    }
    return QPair<QStringList,QStringList>();
}

}

