/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


*************************************************************************/

#include "model/tagextension.h"
#include "model/tagitem.h"
#include "model/texeltagmodel.h"
#include "options/preferences.h"

#include <QMutex>
#include <QMutexLocker>

#include "aid.h"

using namespace std;

namespace Texel {

    namespace {
        struct CacheGuard
        {
            CacheGuard (TagExtension* guarded)
                : m_Guarded(guarded)
            {
            }

            ~CacheGuard ()
            {
                if (!m_Guarded->modified()) {
                    if (m_Guarded->getItem()!=0l) {
                        m_Guarded->getItem()->resetExtension();
                        delete m_Guarded;
                    }
                }
            }

            TagExtension* m_Guarded;
        };

        class SynchronizedCache: private QCache<const TagItem*,CacheGuard>
        {
            typedef QCache<const TagItem*,CacheGuard> super;

        public:
            SynchronizedCache ()
                :super(0l),
                m_Mutex(QMutex::Recursive)
            {
            }

            virtual ~SynchronizedCache()
            {
            }

            void setMaxCost (int cost)
            {
                QMutexLocker locker(&m_Mutex);
                super::setMaxCost(cost);
            }

            bool remove (const TagItem* item)
            {
                QMutexLocker locker(&m_Mutex);
                return super::remove(item);
            }

            bool insert (const TagItem* item, CacheGuard* guard)
            {
                QMutexLocker locker(&m_Mutex);
                return super::insert(item,guard,1);
            }

        private:
            QMutex m_Mutex;
        };

        static SynchronizedCache s_LoadedExtensions;
    }

    CacheConfigGuard::CacheConfigGuard (QObject* parent)
        : QObject(parent)
    {
    }

    CacheConfigGuard::~CacheConfigGuard ()
    {
    }

    void CacheConfigGuard::configChanged ()
    {
        s_LoadedExtensions.setMaxCost(Preferences::instance()->extensions.m_CacheSize);
    }

    TagExtension::TagExtension (const TagItem* item,
                                const TagExtension::AllowedAttributeMap& allowedAttributes,
                                const QSet<QString>& allowedAttributesKeys,
                                const QSet<QString>& allowedImages)
        : m_Item(item),
        m_AllowedAttributs(allowedAttributes),
        m_AllowedAttributsKeys(allowedAttributesKeys),
        m_AllowedImages(allowedImages),
        m_Attribs(),
        m_Images(),
        m_AttribsToDelete(),
        m_ImagesToDelete(),
        m_AttribsUpdated(),
        m_ImagesUpdated()
    {
        if (m_Item!=0l) {
            s_LoadedExtensions.insert (m_Item, new CacheGuard(this));
        }
    }

    TagExtension::~TagExtension ()
    {
        s_LoadedExtensions.remove(getItem());
    }

    void TagExtension::addLyrics (const Lyrics& l)
    {
        m_Lyrics << l;
    }

    QList<TagExtension::Lyrics> TagExtension::getLyrics () const
    {
        return m_Lyrics;
    }

    const QImage* TagExtension::getImage (const QString& name)
    {
        if (m_Images.contains(name)) {
            return &m_Images[name];
        }
        return 0l;
    }

    bool TagExtension::initImage (const QString& name, const QImage& image)
    {
        if (m_AllowedImages.contains(name)) {
            m_Images.insert (name,image);
            return true;
        }
        return false;
    }

    bool TagExtension::setImage (const QString& name, const QImage& image)
    {
        if (m_AllowedImages.contains(name)) {
            const bool altered = modified();

            if (m_Images.contains(name) || m_ImagesToDelete.contains(name)) {
                m_ImagesUpdated.insert(name);
                m_ImagesToDelete.remove(name);
            }
            else {
                m_ImagesNew.insert(name);
            }
            m_Images.insert (name,image);

            if (!altered) {
                m_Item->getModel()->emitChangedRow(m_Item->getRow(),false);
            }

            m_Item->getModel()->emitAlteredExtension(m_Item->getRow(), 0, name, QVariant());
            return true;
        }
        return false;
    }

    bool TagExtension::rmImage (const QString& name)
    {
        if (m_Images.remove(name)>0) {
            const bool altered = modified();

            m_ImagesToDelete.insert(name);
            m_ImagesUpdated.remove(name);
            m_ImagesNew.remove(name);

            if (!altered) {
                m_Item->getModel()->emitChangedRow(m_Item->getRow(),false);
            }

            m_Item->getModel()->emitAlteredExtension(m_Item->getRow(), 0, name, QVariant());

            return true;
        }
        return false;
    }

    QSet<QString> TagExtension::getAvailableImageTypes () const
    {
        return m_AllowedImages;
    }

    bool TagExtension::initAttribute (const QString& name, const QString& image)
    {
        if (m_AllowedAttributs.contains(name)) {
            m_Attribs.insert(name,image);
            return true;
        }
        return false;
    }

    bool TagExtension::setAttribute (const QString& name, const QString& value)
    {
        if (m_AllowedAttributs.contains(name)) {
            const bool altered = modified();
            if (m_Attribs.contains(name)) {
                m_AttribsUpdated.insert(name);
                m_AttribsToDelete.remove(name);
            }
            else {
                m_AttribsNew.insert(name);
            }
            m_Attribs.insert (name,value);

            if (!altered) {
                m_Item->getModel()->emitChangedRow(m_Item->getRow(),false);
            }
            m_Item->getModel()->emitAlteredExtension(m_Item->getRow(),1,name, value);

            return true;
        }

        return false;
    }

    bool TagExtension::rmAttribute (const QString& name)
    {
        if (m_Attribs.remove(name)>0) {
            const bool altered = modified();
            m_AttribsToDelete.insert(name);
            m_AttribsUpdated.remove(name);
            m_AttribsNew.remove(name);
            if (!altered) {
                m_Item->getModel()->emitChangedRow(m_Item->getRow(),false);
            }
            m_Item->getModel()->emitAlteredExtension(m_Item->getRow(),2,name, "");
            return true;
        }
        return false;
    }

    QMap<QString, QString> TagExtension::getAttribs () const
    {
        return m_Attribs;
    }

    bool TagExtension::hasAttribute(const QString& name) const
    {
        return m_Attribs.contains(name);
    }

    QSet<QString> TagExtension::getAllowedAttribs () const
    {
        return m_AllowedAttributsKeys;
    }

    void TagExtension::resetChangeState ()
    {
        m_AttribsNew.clear();
        m_AttribsToDelete.clear();
        m_AttribsUpdated.clear();
        m_ImagesNew.clear();
        m_ImagesToDelete.clear();
        m_ImagesUpdated.clear();
    }

    TagExtension::AttributeMetaData TagExtension::getMetaData (const QString& name)
    {
        return m_AllowedAttributs[name];
    }

    void TagExtension::removeFromCache ()
    {
        const TagItem* tmp = m_Item;
        m_Item=0l;
        s_LoadedExtensions.remove(tmp);
    }

}
