/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#ifndef TAGEXTENSION_H
#define TAGEXTENSION_H

#include <QSet>
#include <QMap>
#include <QString>
#include <QImage>
#include <QCache>
#include <QMetaType>

namespace Texel {

    class TagItem;

    /**
      Little helper object which is just created once to listen to the signal configChanged
      from the preferences and adjust the static QCache tagextensions cache size
      */

    class CacheConfigGuard: public QObject
    {
        Q_OBJECT
    public:
        CacheConfigGuard (QObject* parent);
        virtual ~CacheConfigGuard ();

    public slots:
        void configChanged ();
    };

    class TagExtension
    {
    public:
        typedef QPair<QMetaType::Type, int> AttributeMetaData;
        typedef QMap<QString, AttributeMetaData> AllowedAttributeMap;

        struct Lyrics
        {
            Lyrics ()
            {
            }

            Lyrics (const QString& l)
                : lyrics(l)
            {
            }

            QString lyrics;
            QString description;
            QString lang;
        };

        TagExtension(const TagItem* item,
                     const TagExtension::AllowedAttributeMap& allowedAttributes,
                     const QSet<QString>& allowedAttributesKeys,
                     const QSet<QString>& allowedImages);
        ~TagExtension ();

        void addLyrics (const Lyrics& l);
        QList<Lyrics> getLyrics () const;

        const QImage* getImage (const QString& name);
        bool setImage (const QString& name, const QImage& image);
        bool initImage (const QString& name, const QImage& image);
        bool rmImage (const QString& name);
        QSet<QString> imagesToDelete() const;
        QSet<QString> imagesToUpdate() const;
        QSet<QString> imagesToAdd () const;
        QList<QString> images() const;

        QMap<QString, QString> getAttribs () const;
        QSet<QString> getAllowedAttribs () const;
        bool setAttribute(const QString& name, const QString& attr);
        bool initAttribute(const QString& name, const QString& attr);
        bool rmAttribute (const QString& name);
        bool hasAttribute(const QString& name) const;
        QString attrib (const QString name) const;
        QSet<QString> attribsToDelete() const;
        QSet<QString> attribsToUpdate() const;
        QSet<QString> attribsToAdd () const;
        TagExtension::AttributeMetaData getMetaData (const QString& name);

        QSet<QString> getAvailableImageTypes () const;

        void removeFromCache ();
        void resetChangeState ();

        bool modified () const;

        const TagItem* getItem() const;

    private:
        const TagItem* m_Item;

        const AllowedAttributeMap& m_AllowedAttributs;
        const QSet<QString>& m_AllowedAttributsKeys;
        const QSet<QString>& m_AllowedImages;

        QList<Lyrics> m_Lyrics;

        QMap<QString,QString> m_Attribs;
        QMap<QString,QImage> m_Images;

        QSet<QString> m_AttribsToDelete;
        QSet<QString> m_ImagesToDelete;
        QSet<QString> m_AttribsUpdated;
        QSet<QString> m_ImagesUpdated;
        QSet<QString> m_AttribsNew;
        QSet<QString> m_ImagesNew;
    };

    inline QSet<QString> TagExtension::imagesToDelete() const
    {
        return m_ImagesToDelete;
    }

    inline QSet<QString> TagExtension::imagesToUpdate() const
    {
        return m_ImagesUpdated;
    }

    inline QSet<QString> TagExtension::imagesToAdd() const
    {
        return m_ImagesNew;
    }

    inline QList<QString> TagExtension::images () const
    {
        return m_Images.keys();
    }

    inline QSet<QString> TagExtension::attribsToDelete() const
    {
        return m_AttribsToDelete;
    }

    inline QSet<QString> TagExtension::attribsToUpdate() const
    {
        return m_AttribsUpdated;
    }

    inline QSet<QString> TagExtension::attribsToAdd () const
    {
        return m_AttribsNew;
    }

    inline QString TagExtension::attrib (const QString name) const
    {
        return m_Attribs[name];
    }

    inline bool TagExtension::modified () const
    {
        return !m_AttribsToDelete.empty() ||
                !m_AttribsUpdated.empty() ||
                !m_AttribsNew.empty() ||
                !m_ImagesToDelete.empty() ||
                !m_ImagesUpdated.empty() ||
                !m_ImagesNew.empty();
    }

    inline const TagItem* TagExtension::getItem() const
    {
        return m_Item;
    }


}

#endif // TAGEXTENSION_H
