/************************************************************************
*************************************************************************/

#undef QT_NO_STL
#include <QFileInfo>
#include <QString>
#include <QCache>
#include <QFile>
#include <QRegExp>
#include <QBitArray>
#include <QMessageBox>
#include <QDir>

#include "tagitem.h"
#include "dirtree.h"
#include "model/problem.h"
#include "model/texeltagmodel.h"
#include "func/parserproblem.h"
#include "util/qtlevenshtein.h"
#include "options/preferences.h"
#include "model/columns.h"
#include "model/tagspecializations.h"
#include "view/messageview.h"
#include "util/errredirector.h"
#include "util/loghook.h"

#include "aid.h"

#include <taglib/taglib.h>
#include <taglib/tag.h>
#include <taglib/fileref.h>

#include <klocale.h>

using namespace std;

namespace Texel {

    inline quint32 TagItem::createUniqId ()
    {
        return uniqIdSequence++;
    }

    void TagItem::adjustUniqId (quint32 id)
    {
        if (id>uniqIdSequence) {
            uniqIdSequence=id+1;
        }
    }

    quint32 TagItem::uniqIdSequence=1;

    inline void TagItem::createColumns()
    {
        columns[ColumnManager::Pattern]=new Columns::Searchpattern();
        columns[ColumnManager::DestinationPath]=new Columns::DestinationPath();
        columns[ColumnManager::Path]=new Columns::Path();
        columns[ColumnManager::Filename]=new Columns::Filename();
        columns[ColumnManager::Artist]=new Columns::Artist();
        columns[ColumnManager::Title]=new Columns::Title();
        columns[ColumnManager::Album]=new Columns::Album();
        columns[ColumnManager::Comment]=new Columns::Comment();
        columns[ColumnManager::Track]=new Columns::Track();
        columns[ColumnManager::CD]=0l; // Only created with IDV2+MP4 tag
        columns[ColumnManager::Rating]=0l; // Only created with IDV2+MP4 tag
        columns[ColumnManager::Genre]=new Columns::Genre();
        columns[ColumnManager::Year]=new Columns::Year();
        columns[ColumnManager::Length]=new Columns::Length();
        columns[ColumnManager::Bitrate]=new Columns::Bitrate();
        columns[ColumnManager::Channels]=new Columns::Channels();
        columns[ColumnManager::SampleRate]=new Columns::SampleRate();
        columns[ColumnManager::FileSize]=new Columns::FileSize();
    }

    void TagItem::takeover (TagItem* other)
    {
        for (int i=0;i<18;i++) {
            delete columns[i];
            columns[i]=other->columns[i];
            other->columns[i]=0l;
        }
        m_Dependencies=other->m_Dependencies;

        for (int col=0; col<ColumnManager::instance()->getEditableColumns();col++) {
            if (m_Dependencies.testDependency(col,0)) {
                const int realcol = ColumnManager::instance()->mapFromEditableColumn(col);
                const QString f = data (realcol,Columns::Column::RolePersistence);
                setDataDirect(realcol,f,Columns::Column::RoleReevaluate);
            }
        }

        flags=other->flags;
    }

    void TagItem::merge (const TagItem* other)
    {
        for (int i=0;i<18;i++) {
            if (data(i,Columns::Column::RoleAsVariable)=="" &&
                other->data(i,Columns::Column::RoleAsVariable)!="") {
                // Use the long method via the core model as this provides undo/redo facilities
                getModel()->setData(getModel()->index(getRow(),i),other->data(i,Columns::Column::RoleAsVariable));
            }
        }
    }

    inline void TagItem::removeColumns()
    {
        for (int i=0;i<18;i++) delete columns[i];
    }

    TagItem::TagItem (TagModel* _model)
        : model(_model),
        thisRow(-1),
        uniqId(createUniqId()),
        relations(0l),
        flags (InCreation),
        m_NoOfWarnings(0),
        m_NoOfErrors(0),
        m_CachedSortingCol(-1),
        m_CachedSortingString()
    {
        createColumns();
    }

    TagItem::~TagItem ()
    {
        dynamic_cast<Columns::Path*>(columns[ColumnManager::Path])->removeFromDirTree(this);
        MasterManager::instance()->getArtistManager()->removeObject(data(ColumnManager::Artist,Columns::Column::RoleDisplay),this);
        removeColumns();
        //relations->onRemovalFromList(this);
        delete relations;
    }

    QString TagItem::data (int column, Columns::Column::Role role) const
    {
        if (column==ColumnManager::MarkForDeletion) {
            return QVariant(isToDelete()).toString();
        }

        if (column>=0 && column<ColumnManager::instance()->getColumnNumber() && columns[column]!=0l) {
            if (role==Columns::Column::RoleSorting) {
                if (m_CachedSortingCol!=column) {
                    m_CachedSortingCol=column;
                    m_CachedSortingString=columns[column]->data(role);
                }
                return m_CachedSortingString;
            }

            return columns[column]->data(role);
        }
        return QString();
    }

    bool TagItem::setDependency (const QString& name, bool org, const QVariant& context)
    {
        if (columns[ColumnManager::Pattern]!=0l &&
            dynamic_cast<Columns::Searchpattern*>(columns[ColumnManager::Pattern])->containsVar(name)) {
            m_Dependencies.setDependency(ColumnManager::instance()->mapToEditableColumn(context.toInt()),
                                         ColumnManager::instance()->mapToEditableColumn(0));
            return true;
        }

        const int col=ColumnManager::instance()->indexByName(name);
        if (col!=-1) {
            const int colEditableColumn = ColumnManager::instance()->mapToEditableColumn(col);
            if (!org) {
                const int contextCol = ColumnManager::instance()->mapToEditableColumn(context.toInt());
                if (colEditableColumn==contextCol) {
                    //throw new UnknownProblem (i18n("Expression contains self reference"));
                }
                else if (colEditableColumn!=-1) {
                    m_Dependencies.setDependency(contextCol,colEditableColumn);
                }
            }
            return true;
        }

        if (name=="g_COMMON_PATH" || name=="ext" || Preferences::instance()->m_GlobalVars.contains(name)) return true;

        return ::getenv(name.toAscii().data())!=0l;
    }

    QString TagItem::getVar (const QString& name, bool org, const QVariant&) const
    {
        if (columns[ColumnManager::Pattern]!=0l &&
            dynamic_cast<Columns::Searchpattern*>(columns[ColumnManager::Pattern])->containsVar(name)) {
            return dynamic_cast<Columns::Searchpattern*>(columns[ColumnManager::Pattern])->getVar(name);
        }
        else {
            const int col=ColumnManager::instance()->indexByName(name);
            if (col!=-1) {
                return data(col,(org==true)?Columns::Column::RoleAsRawVariable:Columns::Column::RoleAsVariable);
            }
        }

        if (name=="ext") {
            return suggestFilenameExtension();
        }

        if (name=="g_COMMON_PATH") {
            return getModel()->getDirCache()->getCommonPath();
        }

        if (Preferences::instance()->m_GlobalVars.contains(name)) {
            return Preferences::instance()->m_GlobalVars[name];
        }

        // Last resort: Do we have an environment variable?
        const char* var = ::getenv(name.toAscii().data());
        if (var!=0l) {
            return QString(var);
        }

        return "";
    }

    bool TagItem::setData (int column, const QVariant& v, Columns::Column::Role role)
    {
        if (isLockedForUpdate()) {
            return false;
        }
        QModelIndex idx = getModel()->index(getRow(),column);
        return getModel()->setData(idx,v,role);
    }

    bool TagItem::setDataDirect (int column, const QString& v, Columns::Column::Role role)
    {
        if (isLockedForUpdate()) {
            return false;
        }
        bool changed=false;

        if (column>=0 && column<ColumnManager::instance()->getColumnNumber() && columns[column]!=0l) {
            changed=columns[column]->setData(v,role,this);
            if (changed && role==Columns::Column::RoleReevaluate) {
                model->emitChangedCell(model->index(getRow(),column));
            }
        }

        if (changed && m_CachedSortingCol==column) {
            m_CachedSortingCol=-1;
            m_CachedSortingString="";
        }

        if (changed && role!=Columns::Column::RoleInitialization) {
            for (int col=0; col<ColumnManager::instance()->getEditableColumns();col++) {
                if (m_Dependencies.testDependency(col,column)) {
                    const int realCol = ColumnManager::instance()->mapFromEditableColumn(col);
                    const QString f = data (realCol,Columns::Column::RolePersistence);
                    cout << "Reevaluate: " << ColumnManager::instance()->getName(realCol) << ";" << f << endl;
                    setDataDirect(realCol,f,Columns::Column::RoleReevaluate);
                }
            }
        }

        return changed;
    }

    void TagItem::rename (const QFileInfo& o, const QFileInfo& n, bool forceRename) throw (IOError, RenameCollision)
    {
        // Check if the path changes. If so, gurantee that the new pah exists (create it when necessary)
        if (o.absolutePath()!=n.absolutePath()) {
            QDir dir=n.absoluteDir();
            if (!dir.exists()) {
                dir.mkpath(n.absolutePath());
            }
        }

        QFile mp3 (o.absoluteFilePath());
        const bool overwrite=forceRename || Preferences::instance()->io.m_OverwriteOnRename;
        if (!overwrite) {
            if (n.exists()) {
                throw RenameCollision(this,n.absoluteFilePath());
            }
        }

        if (n.exists()) {
            const QString absN=n.absoluteFilePath();

            if (!QFile::remove(absN)) {
                throw IOError(i18n("File %1 should be overwritten, but could not be delete",absN),this);
            }

            // Check if file is already in model and remove it from there -> it will be overwritten
            TagItem* toRemove = getModel()->getRowByOrgFilename(absN);
            if (toRemove!=0l) {
                if (toRemove!=this) {
                    getModel()->removeRow(toRemove);
                }
                else {
                    qFatal ("Can remove myself!");
                }
            }
        }

        if (!mp3.rename(n.absoluteFilePath())) {
            throw IOError (i18n("Could not rename file '%2'' to '%1'",n.canonicalFilePath(),o.canonicalFilePath()),this);
        }
   }

    QString* TagItem::update (bool forceRename, bool) throw (IOError, RenameCollision)
    {
        const QFileInfo info(getAbsolutFilename(Columns::Column::RoleAsRawVariable));
        bool nameWasAltered=false;

        if (isToDelete()) {
            if (Preferences::instance()->io.m_UseTrashCan) {
                QDir trashCan(Preferences::instance()->io.m_TrashCan);
                if (!trashCan.exists()) {
                    if (!trashCan.mkpath(".")) {
                        throw IOError (i18n("Trashcan %1 does not exist and creation failed",trashCan.absolutePath()),this);
                    }
                }

                const QString destFilename = trashCan.absoluteFilePath(info.fileName());
                const bool overwrite=Preferences::instance()->io.m_OverwriteInTrashCan;
                if (overwrite && QFile::exists(destFilename)) {
                    if (!QFile::remove(destFilename)) {
                        throw IOError (i18n("File does already exist in trashcan but could not be removed"),this);
                    }
                    // Check if file is already in model and remove it from there -> it will be overwritten
                    TagItem* toRemove = getModel()->getRowByOrgFilename(destFilename);
                    if (toRemove!=0l) {
                        if (toRemove!=this) {
                            getModel()->removeRow(toRemove);
                        }
                        else {
                            qFatal ("Can't remove myself!");
                        }
                    }
                }

                if (!QFile::rename(info.absoluteFilePath(),destFilename)) {
                    QString details = i18n("Unknown reason");
                    // Check why the deletion failed
                    if (QFile::exists(destFilename)) {
                        details=i18n("Filename %1 does already exist in trashcan",info.fileName());
                    }
                    else {
                        QFileInfo test(trashCan.absolutePath());
                        if (!test.isWritable()) {
                            details=i18n("Trashcan is not writeable");
                        }
                    }
                    throw IOError(i18n("Movement into configured trashcan (%1) failed: %2",trashCan.absolutePath(),details),this);
                }
            }
            else if (!QFile::remove(info.absoluteFilePath())) {
                throw IOError(i18n("Deletion of file failed"),this);
            }
            nameWasAltered=true;
        }
        else if (isAltered())
        {
            // First test if file is available, readable and writeable. This allows more precise
            // error messages

            if (!info.exists()) {
                throw IOError(i18n("File '%1' does not exist",info.absoluteFilePath()), this);
            }
            if (!info.isFile()) {
                throw IOError(i18n("Filename points not to a file!"), this);
            }
            if (!info.isWritable()) {
                QFile f(info.absoluteFilePath());
                if (Preferences::instance()->io.m_TryEnforceWrite) {
                    if (!f.setPermissions(f.permissions()|QFile::WriteOwner|QFile::WriteGroup|QFile::WriteOther)) {
                        throw IOError(i18n("File write permissions could not be given"), this);
                    }
                    LogHook::instance()->warning(i18n("Write permissions automatically given"), this);
                }
                else {
                    throw IOError(i18n("Write permissions not given."), this);
                }
            }

            //
            // Update metadata on disc
            //
            if (isAltered()) {
                ErrRedirector redirector;
                redirector.setItem(this);
                saveMetaTag();
            }

            //
            // Execute renaming
            //
            if (dynamic_cast<Columns::Editable*>(columns[ColumnManager::Path])->isUpdated() ||
                dynamic_cast<Columns::Editable*>(columns[ColumnManager::Filename])->isUpdated())
            {
                // Check if the new file has a suffix, and add than from the original if not
                QFileInfo newName(getAbsolutFilename(Columns::Column::RoleDisplay));
                if (newName.suffix().isEmpty()) {
                    newName.setFile (newName.filePath()+"."+info.suffix());
                }

                try {
                    if (info!=newName) {
                        rename(info,newName,forceRename);
                        dynamic_cast<Columns::Editable*>(columns[ColumnManager::Path])->setUpdated(false,this);
                        dynamic_cast<Columns::Editable*>(columns[ColumnManager::Filename])->setUpdated(false,this);
                        nameWasAltered=true;
                    }
                }
                catch (RenameRejected)
                {
                    dynamic_cast<Columns::Editable*>(columns[ColumnManager::Path])->reset(this);
                    dynamic_cast<Columns::Editable*>(columns[ColumnManager::Filename])->reset(this);
                }

                DirTree::Name filename=DirTree::Name(getFilenameId(Columns::Column::RoleAsVariable));
                getModel()->alteredPersistentFilename(this,filename);
            }

            //
            // Reset item by setting current content as 'original'
            //
            for (int i=0;i<18;i++) {
                if (columns[i]!=0L && columns[i]->isEditable()) {
                    dynamic_cast<Columns::Editable*>(columns[i])->setUpdated(false,this);
                }
            }
            m_Dependencies.clear();
        }

        QString* emptyDirectoryToDelete=0l;

        if (nameWasAltered && Preferences::instance()->io.m_RemoveEmptyDir) {
            QDir dir(info.absolutePath());
            if (QFile::exists(info.absolutePath()) &&
                dir.entryList(QDir::NoDotAndDotDot|QDir::AllEntries).isEmpty()) {
                if (Preferences::instance()->io.m_AckDirRemove==true) {
                    emptyDirectoryToDelete=new QString(dir.absolutePath());
                }
                else dir.rmpath(info.absolutePath());
            }
        }

        model->emitChangedRow(getRow(),true);

        return emptyDirectoryToDelete;
    }

    bool TagItem::isAltered (int column) const
    {
        if (column==ColumnManager::MarkForDeletion) return isToDelete();

        if ((column>0 && column<ColumnManager::instance()->getColumnNumber()) && columns[column]!=NULL && columns[column]->isEditable()) {
            return dynamic_cast<Columns::Editable*>(columns[column])->isUpdated();
        }
        return false;
    }

    int TagItem::getIntValue (int column) const
    {
        if (columns[column]!=0l) {
            Columns::EditableInt* c = dynamic_cast<Columns::EditableInt*>(columns[column]);
            if (c!=0l) return c->getIntValue(Columns::Column::RoleDisplay);
            else {
                Columns::StaticInt* c = dynamic_cast<Columns::StaticInt*>(columns[column]);
                if (c!=0l) return c->getIntValue(Columns::Column::RoleDisplay);
            }
        }
        return -1;
    }

    void TagItem::updateArtist (int state)
    {
        // When the object is still in creation the seting can be skipped. Actually it must be
        // skipped because otherwise the creation thread hassles with the main thread on the model
        if (!isInCreation() && thisRow!=-1) {
            model->setData(model->index(thisRow,ColumnManager::Artist),QVariant(state),TagModel::CorrectnessCheckRole);
        }
    }

    void TagItem::changeArtistName (const QString& name)
    {
        model->setData(model->index(thisRow,ColumnManager::Artist),name);
    }

    const QStringList& TagItem::getProposals () const
    {
        return dynamic_cast<Columns::Artist*>(columns[ColumnManager::Artist])->getProposals (Columns::Column::RoleDisplay);
    }

    InputProblem* TagItem::getProblemPointer (int column) const
    {
        if (column==ColumnManager::MarkForDeletion) return 0l;

        if (column==ColumnManager::Pattern) {
            return dynamic_cast<Columns::Searchpattern*>(columns[column])->getProblem();
        }

        if (column>=0 && column<ColumnManager::instance()->getColumnNumber() && columns[column]!=0l && columns[column]->isEditable()) {
            return dynamic_cast<Columns::Editable*>(columns[column])->getProblem();
        }

        return 0l;
    }

    QString TagItem::getProblem (int column) const
    {
        const InputProblem* p=getProblemPointer(column);
        if (p!=0l) {
            return p->getText();
        }
        return "";
    }

    void TagItem::displayTooltip (int column, const QPoint& pos, const QRect& tooltipBox) const
    {
        InputProblem* p=getProblemPointer(column);
        if (p!=0l) {
            return p->showTooltip(pos, tooltipBox);
        }
    }


    QString TagItem::suggestVarName (const QString& name) const
    {
        return dynamic_cast<Columns::Searchpattern*>(columns[ColumnManager::Pattern])->suggestVarName(name);
    }

    void TagItem::setProblem (int col, InputProblem* ip)
    {
        if (col>=0 && col<ColumnManager::instance()->getColumnNumber() && columns[col]!=NULL && columns[col]->isEditable()) {
            dynamic_cast<Columns::Editable*>(columns[col])->setProblem(ip);
        }
    }

    DirTree::Name TagItem::getFilenameId (Columns::Column::Role role)
    {
        return DirTree::Name(dynamic_cast<Columns::Path*>(columns[ColumnManager::Path])->getPathPointer(role),
                             columns[ColumnManager::Filename]->data(role));
    }

    bool TagItem::isAltered () const
    {
        for (int i=0;i<ColumnManager::instance()->getColumnNumber();i++) {
            if (isAltered(i)) return true;
        }

        return false;
    }

    bool TagItem::hasPropsals () const
    {
        return dynamic_cast<Columns::Editable*>(columns[ColumnManager::Artist])->hasProposals();
    }

    void TagItem::setHasProposals (bool f)
    {
        dynamic_cast<Columns::Editable*>(columns[ColumnManager::Artist])->setHasProposals(f);
    }

    void TagItem::setIsCorrect (bool f)
    {
        dynamic_cast<Columns::Editable*>(columns[ColumnManager::Artist])->setCorrect(f);
    }

    bool TagItem::isCorrect () const
    {
        return dynamic_cast<Columns::Editable*>(columns[ColumnManager::Artist])->isCorrect();
    }

    void TagItem::setIsChecked (bool f)
    {
        dynamic_cast<Columns::Editable*>(columns[ColumnManager::Artist])->setChecked(f);
    }

    bool TagItem::isChecked () const
    {
        return dynamic_cast<Columns::Editable*>(columns[ColumnManager::Artist])->isChecked();
    }

    void TagItem::doSaveMetaTag (TagLib::File&, TagLib::Tag* tag)
    {
        tag->setTitle(columns[ColumnManager::Title]->data(Columns::Column::RoleAsVariable).toStdWString());
        tag->setArtist(columns[ColumnManager::Artist]->data(Columns::Column::RoleAsVariable).toStdWString());
        tag->setAlbum(columns[ColumnManager::Album]->data(Columns::Column::RoleAsVariable).toStdWString());
        tag->setGenre((columns[ColumnManager::Genre]->data(Columns::Column::RoleAsVariable).toStdWString()));
        tag->setYear(dynamic_cast<Columns::EditableInt*>(columns[ColumnManager::Year])->getIntValue(Columns::Column::RoleAsVariable));
        tag->setTrack(dynamic_cast<Columns::EditableInt*>(columns[ColumnManager::Track])->getIntValue(Columns::Column::RoleAsVariable));
        tag->setComment(columns[ColumnManager::Comment]->data(Columns::Column::RoleAsVariable).toStdWString());
    }

    void TagItem::doReadMetaTag (TagLib::FileRef& f, const QFileInfo& info)
    {
        columns[ColumnManager::Filename]->setData(info.fileName(),Columns::Column::RoleInitialization,this);
        columns[ColumnManager::Path]->setData(info.absolutePath(),Columns::Column::RoleInitialization,this);
        dynamic_cast<Columns::FileSize*>(columns[ColumnManager::FileSize])->setIntValue(info.size(),Columns::Column::RoleInitialization,this);

        if (!f.isNull()) {
            if (f.tag()!=0l) {
                columns[ColumnManager::Title]->setData(QString::fromStdWString(f.tag()->title().toWString()),Columns::Column::RoleInitialization,this);
                columns[ColumnManager::Artist]->setData(QString::fromStdWString(f.tag()->artist().toWString()),Columns::Column::RoleInitialization,this);
                columns[ColumnManager::Album]->setData(QString::fromStdWString(f.tag()->album().toWString()),Columns::Column::RoleInitialization,this);
                columns[ColumnManager::Genre]->setData(QString::fromStdWString(f.tag()->genre().toWString()),Columns::Column::RoleInitialization,this);
                dynamic_cast<Columns::EditableInt*>(columns[ColumnManager::Year])->setIntValue(f.tag()->year(),Columns::Column::RoleInitialization,this);
                dynamic_cast<Columns::EditableInt*>(columns[ColumnManager::Track])->setIntValue(f.tag()->track(),Columns::Column::RoleInitialization,this);
                columns[ColumnManager::Comment]->setData(QString::fromStdWString(f.tag()->comment().toWString()),Columns::Column::RoleInitialization,this);
            }

            if (f.audioProperties()) {
                dynamic_cast<Columns::StaticInt*>(columns[ColumnManager::Bitrate])->setIntValue(f.audioProperties()->bitrate(),Columns::Column::RoleInitialization,this);
                dynamic_cast<Columns::StaticInt*>(columns[ColumnManager::Length])->setIntValue(f.audioProperties()->length(),Columns::Column::RoleInitialization,this);
                dynamic_cast<Columns::StaticInt*>(columns[ColumnManager::Channels])->setIntValue(f.audioProperties()->channels(),Columns::Column::RoleInitialization,this);
                dynamic_cast<Columns::StaticInt*>(columns[ColumnManager::SampleRate])->setIntValue(f.audioProperties()->sampleRate(),Columns::Column::RoleInitialization,this);
            }
        }
        else {
            throw InvalidFormat(i18n("Unknown audio(?) format of file %1",(f.file()->name())));
        }
    }

    void TagItem::reevaluate ()
    {
        static_cast<Columns::Searchpattern*>(columns[ColumnManager::Pattern])->parse(false,this);
        for (int i=1;i<17;i++) {
            Columns::Editable* editable = dynamic_cast<Columns::Editable*>(columns[i]);
            if (editable!=0l) editable->evaluate(this);
        }
    }

    DirTree::Dir* TagItem::getPathPointer(Columns::Column::Role role) const
    {
        return dynamic_cast<Columns::Path*>(columns[ColumnManager::Path])->getPathPointer(role);
    }

    QStringList TagItem::availableVariables () const
    {
        QStringList variables;
        variables << dynamic_cast<Columns::Searchpattern*>(columns[ColumnManager::Pattern])->varNames();
        for (int i=1;i<ColumnManager::instance()->getColumnNumber();i++) {
            variables << ColumnManager::instance()->getName(i);
        }
        return variables;
    }

    TagExtension* TagItem::getExtensions () const
    {
        return 0l;
    }

    void TagItem::resetExtension () const
    {
    }

    bool TagItem::supportsCoverImages () const
    {
        return false;
    }

    TagRelations* TagItem::getRelations () const
    {
        if (relations==0l) relations=new TagRelations;
        return relations;
    }

    QImage TagItem::getDefaultCoverImage () const
    {
        return QImage();
    }

    QString TagItem::getDefaultImageName () const
    {
        return "";
    }

    void TagItem::setLockedForUpdate (bool update)
    {
        if (update!=isLockedForUpdate()) {
            flags^=LockedForUpdate;
            model->emitChangedHeader(getRow());
        }
    }

}
