/*************************************************************************
*************************************************************************/

#ifndef texel_TAGITEM_H
#define texel_TAGITEM_H

#include "columnmanager.h"
#include "model/structure/mastermanager.h"
#include "func/parser.h"
#include "model/dependencymap.h"
#include "model/dirtree.h"
#include "model/columnsbase.h"
#include "model/tagrelations.h"
#include "view/messageview.h"

#include <QList>
#include <QVariant>
#include <QSet>
#include <QDir>

#include <stdexcept>

class QFileInfo;
class QCompleter;

namespace TagLib
{
    class Tag;
    class File;
    class FileRef;
}

namespace Texel {

    class TagModel;
    class InputProblem;
    class TagExtension;

    namespace TagItemType
    {
        enum Type {
            MPEG=1,
            Ogg_Speex=2,
            Ogg_Flag=3,
            Ogg_Vorbis=4,
            Flac=5,
            MPC=6,
            TrueAudio=7,
            WavPack=8,
            MP4=9,
            ASF=10,
            RIFF_AIFF=11,
            RIFF_WAV=12,

            MusicBrainzTagSuggestion=99
        };
    }

    /** This is the core class of the model. It represents the tag information of a single file and stores
  * not only meta tag information but also filename, directory, internally defined variables for this row
  * and dirty information as well as duplicates.
  *
  * Two interfaces are implemented,the ParserContext interface which request a variable management
  * and a dependency management between these objects @see @class ParserContext.
  */
    class TagItem: public ManagedObject, public ParserContext
    {
    public:
        struct InvalidFormat: public std::runtime_error
        {
            InvalidFormat (const QString& what)
                : std::runtime_error (static_cast<const char*>(what.toLocal8Bit()))
            {
            }
        };

        struct IOError: public MessageModel::Item
        {
            IOError (const QString& what_ , TagItem* item_)
                : MessageModel::Item(Enums::MessageError,what_,item_)
            {
            }
        };

        struct RenameCollision: public QPair<TagItem*,QString>
        {
            RenameCollision (TagItem* i1, const QString& s2)
                :QPair<TagItem*,QString>(i1,s2)
            {
            }
        };

        typedef QList<TagItem*> List;
        typedef QSet<TagItem*>  Set;

        //
        // Contructors and destructors
        //
        virtual ~TagItem ();

        //
        // Generic data access interface
        //
        QString data (int column, Columns::Column::Role role=Columns::Column::RoleDisplay) const;
        bool setData (int column, const QVariant& v, Columns::Column::Role role=Columns::Column::RoleDisplay);
        bool setDataDirect (int column, const QString& v, Columns::Column::Role role);
        int getIntValue (int column) const;

        bool isAltered (int column) const;
        virtual bool isAltered () const;
        bool hasColumn (int column) const;

        QString* update (bool forceRename,bool) throw (IOError, RenameCollision);
        TagModel* getModel() const;

        QCompleter* getCompleter(int col) const;

        QString getAbsolutFilename (Columns::Column::Role role=Columns::Column::RolePersistence) const;
        DirTree::Name getFilenameId (Columns::Column::Role role);
        DirTree::Dir* getPathPointer(Columns::Column::Role role) const;

        //
        // Implementation of the ParserContext interface
        //
        virtual bool setDependency (const QString& name, bool org, const QVariant& context);
        virtual QString getVar (const QString& name, bool org, const QVariant& context) const;
        virtual QString suggestVarName (const QString& name) const;
        virtual QStringList availableVariables () const;

        bool hasPropsals () const;
        void setHasProposals (bool f);
        void setIsCorrect (bool f);
        bool isCorrect () const;
        void setIsChecked (bool f);
        bool isChecked () const;

        quint32 getUniqId () const;
        void setUniqId (quint32 id);

        void setRow (int row);
        int getRow () const;

        //
        // Functionality for dynamic checked artist list
        //
        virtual void updateArtist (int state);
        virtual void changeArtistName (const QString& name);
        const QStringList& getProposals () const;

        //
        // Functionality to maintain duplicates
        //
        TagRelations* getRelations () const;
        bool hasRelations () const;

        //
        // Functionality to deal with Inputproblems
        //
        bool hasInputProblem (int column) const;
        QString getProblem (int col) const;
        void setProblem (int col, InputProblem* ip);
        void displayTooltip (int col, const QPoint&, const QRect& tooltipBox) const;

        void takeover (TagItem* other);

        void reevaluate ();

        static TagItem* createTagItem (TagModel* model,const QString& filename) throw (TagItem::InvalidFormat);
        static TagItem* createTagItemByType (TagModel* model,const QString& type);
        static TagItem* createTagItemByType (TagModel* model,TagItemType::Type type);

        virtual QString typeName () const=0;
        virtual TagItemType::Type type () const=0;
        virtual QString suggestFilenameExtension () const=0;

        int getWarnings () const;
        int getErrors () const;
        void setWarnings (int w);
        void setErrors (int w);
        void setHasHighlights (bool has) const;
        bool getHasHighlights () const;

        virtual TagExtension* getExtensions () const;
        virtual void resetExtension() const;
        virtual bool supportsCoverImages () const;
        virtual QImage getDefaultCoverImage () const;
        virtual QString getDefaultImageName () const;

        bool setToDelete (bool toDelete);
        bool isToDelete () const;

        void setLockedForUpdate (bool update);
        bool isLockedForUpdate () const;

        void setIsCreated ();
        bool isInCreation ();

        enum Flag {
            HasHighlights   = (1<<0),
            HasImages       = (1<<1),
            ToDelete        = (1<<2),
            LockedForUpdate = (1<<3),
            InCreation      = (1<<4)    // Setafter creaton and until emiting to model
        };

        Q_DECLARE_FLAGS (Flags, Flag);

        /** Merges data cells into the current one when cells in the other TagItem are filled
        and not in the current one.
      */
        void merge (const TagItem* other);

        DependencyMap getDependencies ();

    protected:
        TagItem (TagModel* _model);

        virtual void doReadMetaTag (TagLib::FileRef& f, const QFileInfo& info);
        virtual void doSaveMetaTag (TagLib::File& f, TagLib::Tag* tag);
        virtual void saveMetaTag () throw (IOError)=0;

        Columns::Column* columns[18];

    private:
        TagItem& operator= (const TagItem& other);

        InputProblem* getProblemPointer (int column) const;

        struct RenameRejected {};
        void rename (const QFileInfo& o, const QFileInfo& n, bool forceRename) throw (IOError, RenameCollision);

        TagModel* model;
        qint32 thisRow;
        quint32 uniqId;
        DependencyMap m_Dependencies;
        mutable TagRelations* relations;

        mutable Flags flags;

        // message counter
        quint8 m_NoOfWarnings;
        quint8 m_NoOfErrors;

        mutable qint8 m_CachedSortingCol;
        mutable QString m_CachedSortingString;

        //
        // Column management
        //
        void createColumns();
        void removeColumns();
        friend class Columns::Column;

        static quint32 createUniqId ();
        static void adjustUniqId (quint32 id);

        static quint32 uniqIdSequence;
    };

    Q_DECLARE_OPERATORS_FOR_FLAGS (TagItem::Flags);

    inline TagModel* TagItem::getModel () const
    {
        return model;
    }

    inline QString TagItem::getAbsolutFilename(Columns::Column::Role role) const
    {
        return getPathPointer(role)->getName()+data(ColumnManager::Filename,role);
    }

    inline void TagItem::setRow (int row)
    {
        thisRow=row;
    }

    inline int TagItem::getRow () const
    {
        return thisRow;
    }

    inline bool TagItem::hasInputProblem (int column) const
    {
        return getProblemPointer(column)!=0l;
    }

    inline bool TagItem::hasColumn(int i) const
    {
        return columns[i]!=0l || i==ColumnManager::MarkForDeletion;
    }

    inline QCompleter* TagItem::getCompleter(int col) const
    {
        return columns[col]->getCompleter();
    }

    inline int TagItem::getWarnings () const
    {
        return m_NoOfWarnings;
    }

    inline int TagItem::getErrors () const
    {
        return m_NoOfErrors;
    }

    inline void TagItem::setWarnings (int w)
    {
        m_NoOfWarnings=w;
    }

    inline void TagItem::setErrors (int e)
    {
        m_NoOfErrors=e;
    }

    inline void TagItem::setHasHighlights (bool has) const
    {
        if (getHasHighlights() != has ) {
            flags^=HasHighlights;
        }
    }

    inline bool TagItem::getHasHighlights () const
    {
        return flags.testFlag(HasHighlights);
    }

    inline bool TagItem::isToDelete () const
    {
        return flags.testFlag(ToDelete);
    }

    inline bool TagItem::setToDelete (bool toDelete)
    {
        if (toDelete!=isToDelete()) {
            flags^=ToDelete;
            return true;
        }
        return false;
    }

    inline bool TagItem::isLockedForUpdate () const
    {
        return flags.testFlag(LockedForUpdate);
    }

    inline bool TagItem::hasRelations () const
    {
        return relations!=0l;
    }

    inline DependencyMap TagItem::getDependencies ()
    {
        return m_Dependencies;
    }

    inline void TagItem::setIsCreated ()
    {
        flags &= (~InCreation);
    }

    inline bool TagItem::isInCreation ()
    {
        return flags.testFlag(InCreation);
    }

    inline quint32 TagItem::getUniqId () const
    {
        return uniqId;
    }

    inline void TagItem::setUniqId (quint32 id)
    {
        uniqId=id;
        adjustUniqId(id);
    }

}

Q_DECLARE_METATYPE(Texel::TagItem::List)
Q_DECLARE_METATYPE(Texel::TagItem*)

#endif
