#include "model/tagmodelproxy.h"
#include "model/tagextension.h"
#include "columns.h"
#include "options/preferences.h"
#include "func/coversearch/toolbox.h"

#include <QItemSelectionModel>
#include <QtAlgorithms>
#include <QMimeData>
#include <QPair>
#include <QtConcurrentFilter>
#include <QFile>
#include <QMutex>

#include <kdebug.h>
#include <kundostack.h>
#include <klocale.h>

#include <auto_ptr.h>
#include <algorithm>

#include "aid.h"

using namespace std;

namespace Texel {

    struct Mark
    {
        Mark()
            : from(0),
              to(0)
        {
        }

        Mark(short from_, short to_)
            : from(from_), to(to_)
        {
        }

        short from;
        short to;
    };

    struct HighlightIndex
    {
        HighlightIndex (const TagItem* tag_, int column_=-1)
            : tag(tag_), column(column_)
        {
        }

        bool operator< (const HighlightIndex& other) const
        {
            if (tag==other.tag) {
                if (other.column==-1 || column==-1) {
                    return false;
                }
                return column<other.column;
            }

            return tag<other.tag;
        }

        const TagItem* tag;
        int column;
    };

    typedef QPair<int /* Column */, Mark> RowHighlight;

    class HeaderFactory
    {
    public:
        typedef QPair<int,int> Sector;

        int createHeadline (const QString& title, int row)
        {
            m_HeaderRows << row;
            m_Headers << title;
            return -m_Headers.size();
        }

        void clear ()
        {
            m_HeaderRows.clear();
            m_Headers.clear();
        }

        QVector<Sector> getSectors () const
        {
            QVector<Sector> sectors;
            for (int i=0;i<m_HeaderRows.size()-1;i++) {
                if (m_HeaderRows[i]+1<m_HeaderRows[i+1]) {
                    sectors << Sector(m_HeaderRows[i]+1,m_HeaderRows[i+1]);
                }
            }
            sectors << Sector(m_HeaderRows.last(),m_HeaderRows.size());
            return sectors;
        }

        void setHeaderRow (int row, int value)
        {
            m_HeaderRows[row]=value;
        }

        QString getHeaderText (int i) const
        {
            return m_Headers[i];
        }

        QList<int> getAllHeaderRows () const
        {
            return m_HeaderRows;
        }

    private:
        QList<int> m_HeaderRows;
        QList<QString> m_Headers;
    };

    namespace Source
    {
        class Base
        {
        public:
            virtual const QVector<int> source() const=0;
            virtual ~Base() {}

            virtual bool isDirect() const
            {
                return false;
            }
        };

        class FromSelection: public Base
        {
        public:
            FromSelection (const QVector<int>& source)
                : m_Source(source)
            {
            }

            virtual ~FromSelection() {}

            virtual const QVector<int> source() const
            {
                return m_Source;
            }

        private:
            const QVector<int> m_Source;
        };

        class Direct: public Base
        {
        public:
            Direct (TagModelProxy* model)
                : m_Model(model)
            {
            }

            virtual ~Direct () {}

            virtual const QVector<int> source() const
            {
                const int size = m_Model->rowCount();
                QVector<int> result(size);
                int row=0;
                for (QVector<int>::iterator iter=result.begin();iter!=result.end();++iter) {
                    *iter=row++;
                }
                return result;
            }

            virtual bool isDirect() const
            {
                return true;
            }

        private:
            TagModelProxy* m_Model;
        };
    }

    namespace TagItemFilter
    {
        struct Keeper
        {
            Keeper (const int row_=None)
                : row(row_)
            {
            }

            Keeper (const int row_, const QVector<RowHighlight>& highlights_)
                : row(row_),
                  highlights(highlights_)
            {
            }

            static const int None;

            int row;
            QVector<RowHighlight> highlights;
        };

        const int Keeper::None=0x7fffffff;

        class Base
        {
        public:
            virtual Keeper keep (const TagItem* item) const=0;

            virtual ~Base () {}

        protected:
            Base ()
            {
            }
        };

        class Functor
        {
            Base* m_Filter;
            TagModel* m_Model;
        public:
            typedef Keeper result_type;

            Functor (Base* filter, TagModel* model)
                : m_Filter(filter),
                  m_Model (model)
            {
            }

            virtual ~Functor () {}

            Keeper operator() (int row) const
            {
                // headers are always accepted
                if (row<0) return Keeper(row);
                return m_Filter->keep(m_Model->getRow(row));
            }
        };

        static void reduce (QVector<Keeper>& result, const Keeper& tmp)
        {
            if (tmp.row!=Keeper::None) result << tmp;
        }

        static QVector<Keeper> filter (const QVector<int>& data, Base* filter, TagModelProxy* model)
        {
            Functor filterFunctor(filter, dynamic_cast<TagModel*>(model->sourceModel()));
            QVector<Keeper> temp = QtConcurrent::blockingMappedReduced (data, filterFunctor, reduce);
            return temp;
        }

        class FilterGroup: public Base
        {
        public:
            FilterGroup (const QList<Base*>& filters)
                : m_Filters(filters)
            {
            }

            virtual ~FilterGroup () {}

            virtual Keeper keep (const TagItem *item) const
            {
                Keeper result(item->getRow());

                foreach (Base* filter, m_Filters) {
                    Keeper k = filter->keep(item);
                    if (k.row==0l) return Keeper();
                }

                return result;
            }

        private:
            QList<Base*> m_Filters;
        };

        class Cover: public Base
        {
        public:
            Cover (bool includeCovers)
                : m_IncludeCovers(includeCovers)
            {
            }

            virtual ~Cover () {}

            virtual Keeper keep (const TagItem *item) const
            {
                if (item->supportsCoverImages()) {
                    TagExtension* extension=item->getExtensions();
                    if (extension!=0l && (extension->images().size()==0)!=m_IncludeCovers) {
                        return Keeper(item->getRow());
                    }
                }
                return Keeper();
            }

        private:
            bool m_IncludeCovers;
        };

        class CoverFormats: public Base
        {
        public:
            CoverFormats (bool includeCovers)
                : m_IncludeCovers(includeCovers)
            {
            }

            virtual ~CoverFormats () {}

            virtual Keeper keep (const TagItem *item) const
            {
                return (item->supportsCoverImages()==m_IncludeCovers)?Keeper(item->getRow()):Keeper();
            }

        private:
            bool m_IncludeCovers;
        };

        class ExpressionBase: public Base
        {
        public:
            ExpressionBase (TagModelProxy::ExpressionFilterFlags flags)
                : m_Flags(flags)
            {
            }

            virtual ~ExpressionBase () {}

        protected:
            QVector<Mark> contains (const QString& text, const QString& pattern) const
            {
                QVector<Mark> marks;

                const Qt::CaseSensitivity sensitivity = (m_Flags&TagModelProxy::CaseSensitive)?Qt::CaseSensitive:Qt::CaseInsensitive;

                int start=0;
                if (m_Flags&TagModelProxy::WholeWords) {
                    QRegExp exp("\b"+pattern+"\b",sensitivity);
                    while (true)
                    {
                        int pos=text.indexOf(exp,start);
                        if (pos==-1) break;

                        marks << Mark(pos,pos+exp.matchedLength());
                        start=pos+exp.matchedLength()+1;
                    }
                }
                else {
                    while (true)
                    {
                        int pos=text.indexOf(pattern,start,sensitivity);
                        if (pos==-1) break;

                        marks << Mark(pos,pos+pattern.length());
                        start=pos+pattern.length()+1;
                    }
                }
                return marks;
            }

        private:
            TagModelProxy::ExpressionFilterFlags m_Flags;
        };

        class ExpressionAsString : public ExpressionBase
        {
        public:
            ExpressionAsString (const QString& expression, TagModelProxy::ExpressionFilterFlags flags)
                : ExpressionBase (flags),
                m_Expression(expression)
            {
            }

            virtual ~ExpressionAsString ()
            {
            }

            virtual Keeper keep (const TagItem* item) const
            {
                QVector<RowHighlight> highlights;
                for (int j=0;j<ColumnManager::instance()->getColumnNumber();j++) {
                    if (ColumnManager::instance()->isSearchable(j)) {
                        foreach (Mark m, contains(item->data(j,Columns::Column::RoleDisplay),m_Expression)) {
                            highlights << RowHighlight(j,m);
                        }
                    }
                }

                if (highlights.isEmpty()) return Keeper();
                return Keeper(item->getRow(),highlights);
            }

        private:
            const QString m_Expression;
        };

        class ExpressionSingleWord : public ExpressionBase
        {
        public:
            ExpressionSingleWord (const QString& expression, TagModelProxy::ExpressionFilterFlags flags)
                : ExpressionBase(flags),
                m_ExpressionList(expression.split(' ',QString::SkipEmptyParts))
            {
            }

            ~ExpressionSingleWord ()
            {
            }

            virtual Keeper keep (const TagItem* item) const
            {
                QVector<RowHighlight> highlights;
                for (int j=0;j<ColumnManager::instance()->getColumnNumber();j++) {
                    if (ColumnManager::instance()->isSearchable(j)) {
                        foreach (Mark m, containsAtLeastOne (item->data(j,Columns::Column::RoleDisplay))) {
                            highlights << RowHighlight(j,m);
                        }
                    }
                }

                if (highlights.isEmpty()) return Keeper();
                return Keeper(item->getRow(),highlights);
            }

        private:
            QVector<Mark> containsAtLeastOne (const QString& text) const
            {
                QVector<Mark> marks;
                foreach (QString word,m_ExpressionList) {
                    marks << contains(text,word);
                }
                return marks;
            }

            const QStringList m_ExpressionList;
        };

        class ExpressionAllWords : public ExpressionBase
        {
        public:
            ExpressionAllWords (const QString& expression, TagModelProxy::ExpressionFilterFlags flags)
                : ExpressionBase(flags),
                m_ExpressionList(expression.split(' ',QString::SkipEmptyParts))
            {
            }

            ~ExpressionAllWords ()
            {
            }

            virtual Keeper keep (const TagItem* item) const
            {
                QVector<RowHighlight> highlights;

                foreach (QString word,m_ExpressionList) {
                    bool found=false;
                    for (int j=0;j<ColumnManager::instance()->getColumnNumber();j++) {
                        if (ColumnManager::instance()->isSearchable(j)) {
                            QVector<Mark> marks=contains(item->data(j,Columns::Column::RoleDisplay),word);
                            if (!marks.isEmpty()) {
                                found=true;
                                foreach (Mark m, marks) {
                                    highlights << RowHighlight(j,m);
                                }
                            }
                        }
                    }
                    if (!found) return Keeper();
                }

                return Keeper(item->getRow(),highlights);
            }

        private:
            const QStringList m_ExpressionList;
        };

        class Duplicates : public Base
        {
        public:
            Duplicates ()
                : Base()
            {
            }

            virtual Keeper keep (const TagItem* item) const
            {
                const bool r=item->hasRelations() && item->getRelations()->sizeOfDuplicates()>0;
                return r?Keeper(item->getRow()):Keeper();
            }
        };

        class IncompleteTags : public Base
        {
        public:
            IncompleteTags ()
                : Base()
            {
            }

            virtual Keeper keep (const TagItem* item) const
            {
                const bool r=item->data(ColumnManager::Artist,Columns::Column::RoleDisplay).isEmpty() ||
                             item->data(ColumnManager::Title,Columns::Column::RoleDisplay).isEmpty();
                return r?Keeper(item->getRow()):Keeper();
            }
        };

    }

    namespace Sorting
    {
        class Base
        {
        public:
            virtual void sort (const QVector<int>::iterator& start, const QVector<int>::iterator& end) const=0;

            virtual ~Base () {}
        };

        class Default: public Base
        {
            TagModel* m_Model;
        public:
            virtual void sort (const QVector<int>::iterator& start, const QVector<int>::iterator& end) const
            {
                qStableSort (start,end,*this);
            }

            Default (TagModel* model)
                : m_Model(model)
            {
            }

            virtual ~Default () {}

            inline bool operator() (int s1,int s2) const
            {
                return m_Model->getRow(s1)->getRow()<m_Model->getRow(s2)->getRow();
            }
        };

        class Standard: public Base
        {
        public:
            Standard (TagModel* model, int column, Qt::SortOrder order=Qt::AscendingOrder)
                : m_Model(model),
                m_Column(column),
                m_Lesser(order==Qt::AscendingOrder)
            {
                if (ColumnManager::instance()->getType(m_Column)==QMetaType::Int) {
                    compare=&Standard::compare_int;
                }
                else {
                    compare=&Standard::compare_string;
                }
            }

            virtual ~Standard () {}

            virtual void sort (const QVector<int>::iterator& start, const QVector<int>::iterator& end) const
            {
                qStableSort (start,end,*this) ;
            }

            inline bool operator() (int s1,int s2) const
            {
                return (this->*compare)(m_Model->getRow(s1),m_Model->getRow(s2));
            }

        private:
            bool compare_int (const TagItem* i1, const TagItem* i2) const
            {
                return m_Lesser==(i1->getIntValue(m_Column)<i2->getIntValue(m_Column));
            }

            bool compare_string (const TagItem* i1, const TagItem* i2) const
            {
                return m_Lesser==(QString::localeAwareCompare(i2->data(m_Column,Columns::Column::RoleSorting),
                                                              i1->data(m_Column,Columns::Column::RoleSorting))>0);
            }

            TagModel* m_Model;
            int m_Column;
            bool m_Lesser;
            bool (Standard::*compare)(const TagItem*, const TagItem*) const;
        };

    }

    namespace Grouping
    {
        class Base
        {
        public:
            Base (TagModelProxy* model, HeaderFactory* headerFactory)
                : m_Model(dynamic_cast<TagModel*>(model->sourceModel())),
                  m_HeaderFactory(headerFactory)
            {
            }

            virtual ~Base () {}

            QVector<int> group (QVector<int>& data, Sorting::Base* sorter)
            {
                QVector<int> result;

                order (data);

                foreach (int row, data) {
                    if (row>=0) {
                        add (m_Model->getRow(row),result);
                    }
                }

                if (sorter!=0l) {
                    foreach (HeaderFactory::Sector sector, m_HeaderFactory->getSectors()) {
                        sorter->sort (result.begin()+sector.first, result.begin()+sector.second);
                    }
                }

                return result;
            }

            virtual int getColumn () const
            {
                return -1;
            }

        protected:
            TagModel* m_Model;
            HeaderFactory* m_HeaderFactory;

        private:
            virtual void order (QVector<int>& input) const=0;
            virtual void add (const TagItem* item, QVector<int>& result)=0;
        };

        class ByDecade : public Base
        {
        public:
            ByDecade (TagModelProxy* model, HeaderFactory* headerFactory)
                : Base(model,headerFactory),
                  m_CurrentDecade (-1)
            {
            }

            virtual ~ByDecade () {}

            virtual int getColumn () const
            {
                return ColumnManager::Year;
            }

        private:
            virtual void order (QVector<int>& input) const
            {
                Sorting::Standard sorter(m_Model, ColumnManager::Year);
                sorter.sort(input.begin(),input.end());
            }

            virtual void add (const TagItem* item, QVector<int>& result)
            {
                const int year=getDecade(item);
                if (m_CurrentDecade<year) {
                    m_CurrentDecade=year;
                    QString title;
                    if (year==0) {
                        title="Unknown decade";
                    }
                    else title=QString::number(m_CurrentDecade)+"'s";
                    result << m_HeaderFactory->createHeadline(title,item->getRow());
                }
                result << item->getRow();
            }

            int m_CurrentDecade;

            int getDecade (const TagItem* item) const
            {
                int year=item->getIntValue(ColumnManager::Year);
                if (year>39 && year<100) {
                    year+=1900;
                }
                return year-(year%10);
            }
        };

        class ByDuplicate : public Base
        {
        public:
            ByDuplicate (TagModelProxy* model, HeaderFactory* headerFactory)
                : Base(model,headerFactory)
            {
            }

            virtual ~ByDuplicate() {}

        private:
            virtual void order (QVector<int>&) const
            {
                // No sorting necessary
            }

            virtual void add (const TagItem* item, QVector<int>& result)
            {
                if (item->hasRelations() && item->getRelations()->sizeOfDuplicates()>0) {
                    result << m_HeaderFactory->createHeadline("Duplicates of "+item->getAbsolutFilename(Columns::Column::RoleDisplay),result.size());
                }
                result << item->getRow();
                if (item->hasRelations()) {
                    foreach (TagItem* item, item->getRelations()->getDuplicates()) {
                        result << item->getRow();
                    }
                }
            }
        };

        class ByColumn : public Base
        {
        public:
            ByColumn (TagModelProxy* model, HeaderFactory* headerFactory, int column)
                : Base(model,headerFactory),
                  m_Column (column),
                  m_Current (QVariant::Invalid)
            {
            }

            virtual ~ByColumn() {}

            virtual int getColumn () const
            {
                return m_Column;
            }

        private:
            virtual void order (QVector<int>& input) const
            {
                Sorting::Standard sorter(m_Model, m_Column);
                sorter.sort(input.begin(),input.end());
            }

            virtual void add (const TagItem* item, QVector<int>& result)
            {
                const QVariant v=item->data(m_Column,Columns::Column::RoleDisplay);
                if (!m_Current.isValid() || v!=m_Current) {
                    m_Current=v;
                    result << m_HeaderFactory->createHeadline(v.toString(),result.size());
                }
                result << item->getRow();
            }

            int m_Column;
            QVariant m_Current;
        };

    }

    const int TagModelProxy::s_Marked = -((1<<(8*sizeof(int)-2))-1);

    // For lazy evaluation of the backmap
    struct BackMap:private QVector<QVector<int> >
    {
        typedef QVector<QVector<int> > super;

        BackMap (const QVector<int>& s, const TagModelProxy* m)
            : super(),
              dirty(0),
              source(s),
              model(m)
        {
        }

        inline void append (const QVector<int>& v)
        {
            super::append(v);
        }

        inline QVector<int>& operator[] (int i)
        {
            return get(i);
        }

        inline const QVector<int>& operator[] (int i) const
        {
            return const_cast<BackMap*>(this)->get(i);
        }

        inline QVector<int>& get (int i)
        {
            if (dirty>0) {
                createBackmap();
            }
            return super::operator[](i);
        }

        inline void clear ()
        {
            super::clear();
        }

        inline void resetBackmap ()
        {
            ++dirty;
        }

    private:
        int dirty;
        const QVector<int>& source;
        const TagModelProxy* model;

        void createBackmap ()
        {
            // Reset proxy mapper
            fill(QVector<int>(),model->sourceModel()->rowCount());

            // rebuild
            for (int i=0;i<source.size();i++) {
                if (source[i]>=0) {
                    super::operator[](source[i]).append(i);
                }
            }
            dirty=0;
        }
    };

    struct TagModelProxy::Private
    {
        Private(TagModelProxy* model)
            : m_MapToProxy(m_MapToSource,model),
              m_DisplayRole(Qt::DisplayRole),
              m_FilterFlags(TagModelProxy::AsString),
              m_HeaderFactory(new HeaderFactory),
              m_Source (new Source::Direct(model)),
              m_Parent(model)
        {
        }

        ~Private ()
        {
            removeFilters();
        }

        inline bool isDirect() const
        {
            return m_MapToSource.isEmpty();
        }

        inline bool hasGrouping () const
        {
            return m_Grouper.get()!=0l;
        }

        QVector<int> m_MapToSource;
        BackMap m_MapToProxy;
        // Hack: Remember the last 20 row chunks which where removed to reinsert the data correctly
        // in case of redo
        typedef QVector<int> RedoMap;
        QList<QPair<int, RedoMap> > m_RedoStack;
        int m_DisplayRole;

        ExpressionFilterFlags m_FilterFlags;

        std::auto_ptr<HeaderFactory> m_HeaderFactory;
        QList<TagItemFilter::Base*> m_Filters;
        std::auto_ptr<Sorting::Base> m_Sorter;
        std::auto_ptr<Grouping::Base> m_Grouper;
        std::auto_ptr<Source::Base> m_Source;
        TagModelProxy* m_Parent;

        void removeFilters ()
        {
            foreach (TagItemFilter::Base* filter, m_Filters) delete filter;
            m_Filters.clear();
        }

        int mapToSource (int proxyRow) const
        {
            if (isDirect()) return proxyRow;
            return m_MapToSource.value(proxyRow,0);
        }

        QVector<int> mapFromSource (int sourceRow) const
        {
            if (isDirect()) return QVector<int>(1,sourceRow);
            return m_MapToProxy[sourceRow];
        }

        QVector<TagItemFilter::Keeper> applyFilters (const QVector<int>& map) throw (NoFilter)
        {
            if (!m_Filters.isEmpty()) {
                TagItemFilter::FilterGroup group(m_Filters);
                return TagItemFilter::filter(map,&group,m_Parent);
            }
            throw NoFilter();
        }

        // Highlight sorted by <row,column> and then <from,to>
        QMultiMap<HighlightIndex,Mark> m_Highlights;
    };

    TagModelProxy::TagModelProxy(QObject* parent)
        : super(parent),
        p(new Private(this))
    {
    }

    TagModelProxy::~TagModelProxy()
    {
        delete p;
    }

    inline TagModel* TagModelProxy::superAsTagModel () const
    {
        return static_cast<TagModel*>(sourceModel());
    }

    TagItem* TagModelProxy::getRow (int row) const
    {
        const int mappedRow=p->mapToSource(row);
        if (mappedRow<0) {
            return 0L;
        }
        return superAsTagModel()->getRow(mappedRow);
    }

    QVector<int> TagModelProxy::getRow (const TagItem *row) const
    {
        if (row!=0l) {
            return p->mapFromSource(row->getRow());
        }
        return QVector<int>();
    }

    QString TagModelProxy::getFileName (int row) const
    {
        const int mappedRow=p->mapToSource(row);
        if (mappedRow<0) {
            return "";
        }
        return superAsTagModel()->getFileName(mappedRow);
    }

    bool TagModelProxy::isHeader (int row) const
    {
        if (row>=0 && row<p->m_MapToSource.size()) {
            return !p->isDirect() && p->m_MapToSource[row]<0;
        }
        return false;
    }

    void TagModelProxy::fillDownMoreThanOneRow(const QItemSelectionModel& sel)
    {
        foreach (QItemSelectionRange range, sel.selection()) {
            if (range.top()<range.bottom()) {
                for (int col=range.left();col<=range.right();col++) {
                    QList<SlimIndex> indexes;
                    for (int i=range.top()+1;i<=range.bottom();i++) indexes << SlimIndex(p->mapToSource(i),col);
                    dynamic_cast<TagModel*>(sourceModel())->fillDown(data(index(range.top(),col),TagModel::RawModeRole),indexes);
                }
            }
        }
    }

    void TagModelProxy::setDataBlock (const QVariant& ref, const QModelIndexList& indexes)
    {
        QList<SlimIndex> transList;
        foreach (QModelIndex ind, indexes) {
            transList << SlimIndex(mapToSource(ind));
        }
        superAsTagModel()->setDataBlock (ref,transList);
    }

    void TagModelProxy::paste (const QModelIndex& from, const QModelIndex& to, const QVariantList& values, const QString& description)
    {
        QList<SlimIndex> indexes;
        for (int x=from.row();x<=to.row();x++) {
            const int mappedRow = p->mapToSource(x);
            for (int y=from.column();y<=to.column();y++) {
                indexes << SlimIndex(mappedRow, y);
            }
        }
        superAsTagModel()->paste(indexes,values,description);
    }

    // TODO: replace with line_iterator (saves time & memory as no additional list must be created
    TagItem::List TagModelProxy::getTagItemList () const
    {
        TagItem::List list;
        for (int i=0;i<rowCount();i++) list << getRow(i);
        return list;
    }

    void TagModelProxy::onClear ()
    {
        p->m_FilterFlags=AsString;
        p->m_Sorter.reset();
        p->m_Grouper.reset();
        p->removeFilters();
        p->m_Source.reset(new Source::Direct(this));
        clearMarks();
    }

    void TagModelProxy::sort (int column, Qt::SortOrder order)
    {
        emit modelAboutToBeFiltered();
        if (p->isDirect()) {
            p->m_MapToSource=p->m_Source->source();
        }

        // Choose sorting alogorithm
        Sorting::Base* sorter=0l;
        if (column!=-1) {
            sorter=new Sorting::Standard(dynamic_cast<TagModel*>(sourceModel()),column,order);
            p->m_Sorter.reset(sorter);
        }
        else {
            sorter=new Sorting::Default(dynamic_cast<TagModel*>(sourceModel()));
            p->m_Sorter.reset();
        }

        // Sort
        if (!p->hasGrouping()) {
            sorter->sort (p->m_MapToSource.begin(),p->m_MapToSource.end());
        }
        else {
            // Sort every group internally
            foreach (HeaderFactory::Sector sector, p->m_HeaderFactory->getSectors()) {
                sorter->sort (p->m_MapToSource.begin()+sector.first, p->m_MapToSource.begin()+sector.second);
            }
        }

        emit hasFilter(true);
        createBackMap();
        reset ();
    }

    void TagModelProxy::showFormulars(bool show)
    {
        int oldRole=p->m_DisplayRole;
        if (show) p->m_DisplayRole=TagModel::RoleDisplayFormularFirst;
        else p->m_DisplayRole=Qt::DisplayRole;
        if (p->m_DisplayRole!=oldRole) reset();
    }

    int TagModelProxy::getGroupingColumn () const
    {
        return (p->m_Grouper.get()==0l)?-1:p->m_Grouper->getColumn();
    }

    TagItem::List TagModelProxy::getRowsForHeader (int headerrow) const
    {
        QPair<int,int> range=getRangeForHeader(headerrow);
        TagItem::List  rows;
        if (range!=QPair<int,int>(-1,-1)) {
            for (int i=range.first;i<=range.second;i++) {
                rows << getRow(i);
            }
        }
        return rows;
    }

    QPair<int,int> TagModelProxy::getRangeForHeader (int headerrow) const
    {
        if (!isHeader(headerrow) || isHeader(headerrow+1)) {
            return QPair<int,int>(-1,-1);
        }

        const int start=headerrow+1;
        int end=start;
        while (end<rowCount() && !isHeader(end)) {
            ++end;
        }

        return QPair<int,int>(start,end);
    }

    void TagModelProxy::removeFiles (const TagItem::List& list)
    {
        foreach (TagItem* item, list) {
            item->setToDelete(true);
        }

        TagItem::Set set=list.toSet();
        dynamic_cast<TagModel*>(sourceModel())->updateWhenNecessary(set,false);
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////
    //
    // Reimplementations from QAbstractItemModel
    //
    //////////////////////////////////////////////////////////////////////////////////////////////////////

    QModelIndex TagModelProxy::index(int row, int column, const QModelIndex &) const
    {
        return createIndex(row,column);
    }

    QModelIndex TagModelProxy::parent(const QModelIndex &) const
    {
        return QModelIndex();
    }

    int TagModelProxy::rowCount(const QModelIndex &parent) const
    {
        if (p->isDirect()) return sourceModel()->rowCount(parent);
        return p->m_MapToSource.size();
    }

    int TagModelProxy::columnCount(const QModelIndex &parent) const
    {
        return sourceModel()->columnCount(parent);
    }

    QVariant TagModelProxy::headerData (int section, Qt::Orientation orientation, int role) const
    {
        if (orientation==Qt::Vertical) {
            const int row = p->mapToSource(section);
            if (row<0) {
                if (role==Qt::BackgroundColorRole) return Preferences::instance()->appearance.m_headerBackground;
                return QVariant();
            }
            return sourceModel()->headerData(row,orientation,role);
        }

        if (role==Qt::BackgroundColorRole && section==getGroupingColumn()) {
            return Preferences::instance()->appearance.m_headerBackground;
        }
        return sourceModel()->headerData(section,orientation,role);
    }

    QVariant TagModelProxy::data (const QModelIndex& index, int role) const
    {
        if (!index.isValid()) return QVariant();

        const int row = p->mapToSource(index.row());
        if (row==s_Marked) return QVariant();
        if (row<0) {
            if (index.column()==0) {
                switch (role) {
                case Qt::DisplayRole: return p->m_HeaderFactory->getHeaderText(-row-1);
                case Qt::BackgroundRole: return Preferences::instance()->appearance.m_headerBackground;
                case Qt::ForegroundRole: return Preferences::instance()->appearance.m_headerForeground;
                case Qt::FontRole: return Preferences::instance()->appearance.m_headerFont;
                }
            }
            return QVariant();
        }

        if (role==Enums::RoleHighlights) {
            QList<QVariant> highlights;
            const TagItem* tag = dynamic_cast<TagModel*>(sourceModel())->getRow(row);
            if (tag->getHasHighlights()) {
                QMap<HighlightIndex, Mark>::iterator iter = p->m_Highlights.lowerBound(HighlightIndex(tag,index.column()));
                while (iter!=p->m_Highlights.upperBound(HighlightIndex(tag,index.column()))) {
                    highlights << QPoint((*iter).from,(*iter).to);
                    ++iter;
                }
            }
            return highlights;
        }

        if (role==Qt::DisplayRole) {
            role=p->m_DisplayRole;
        }

        return sourceModel()->data(sourceModel()->index(row,index.column()),role);
    }

    bool TagModelProxy::setData(const QModelIndex &index, const QVariant &value, int role)
    {
        const TagItem* tag = dynamic_cast<TagModel*>(sourceModel())->getRow(index.row());
        if (tag->getHasHighlights()) {
            p->m_Highlights.remove(HighlightIndex(tag,index.column()));
        }
        return QAbstractProxyModel::setData(index,value,role);
    }

    Qt::ItemFlags TagModelProxy::flags (const QModelIndex & index) const
    {
        if (!index.isValid()) {
            return Qt::ItemIsDropEnabled;
        }

        const int row = p->mapToSource(index.row());
        if (row<0) {
            return 0;
        }

        return sourceModel()->flags(sourceModel()->index(row,index.column()));
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////
    //
    // Reimplementations from QAbstractProxyModel
    //
    //////////////////////////////////////////////////////////////////////////////////////////////////////

    void TagModelProxy::setSourceModel (QAbstractItemModel* sourceModel)
    {
        super::setSourceModel(sourceModel);

        connect(sourceModel, SIGNAL(dataChanged(QModelIndex,QModelIndex)),this, SLOT(fromSourceDataChanged(QModelIndex,QModelIndex)));
        connect(sourceModel, SIGNAL(headerDataChanged(Qt::Orientation,int,int)), this, SLOT(fromSourceHeaderDataChanged(Qt::Orientation,int,int)));
        connect(sourceModel, SIGNAL(rowsAboutToBeInserted(QModelIndex,int,int)),this, SLOT(fromSourceRowsAboutToBeInserted(QModelIndex,int,int)));
        connect(sourceModel, SIGNAL(rowsInserted(QModelIndex,int,int)),this, SLOT(fromSourceRowsInserted(QModelIndex,int,int)));
        connect(sourceModel, SIGNAL(columnsAboutToBeInserted(QModelIndex,int,int)),this, SLOT(fromSourceColumnsAboutToBeInserted(QModelIndex,int,int)));
        connect(sourceModel, SIGNAL(columnsInserted(QModelIndex,int,int)),this, SLOT(fromSourceColumnsInserted(QModelIndex,int,int)));
        connect(sourceModel, SIGNAL(rowsAboutToBeRemoved(QModelIndex,int,int)),this, SLOT(fromSourceRowsAboutToBeRemoved(QModelIndex,int,int)));
        connect(sourceModel, SIGNAL(rowsRemoved(QModelIndex,int,int)),this, SLOT(fromSourceRowsRemoved(QModelIndex,int,int)));
        connect(sourceModel, SIGNAL(columnsAboutToBeRemoved(QModelIndex,int,int)),this, SLOT(fromSourceColumnsAboutToBeRemoved(QModelIndex,int,int)));
        connect(sourceModel, SIGNAL(columnsRemoved(QModelIndex,int,int)),this, SLOT(fromSourceColumnsRemoved(QModelIndex,int,int)));
        connect(sourceModel, SIGNAL(layoutAboutToBeChanged()),this, SLOT(fromSourceLayoutAboutToBeChanged()));
        connect(sourceModel, SIGNAL(layoutChanged()),this, SLOT(fromSourceLayoutChanged()));
        connect(sourceModel, SIGNAL(modelReset()), this, SLOT(fromSourceModelReset()));

        TagModel* tagmodel = superAsTagModel();
        connect(tagmodel, SIGNAL(alteredExtension(int,int,QString,QVariant)), this, SLOT(onAlteredExtension(int,int,QString,QVariant)));
        connect(tagmodel, SIGNAL(isEmpty(bool)), this, SLOT(onClear()));
    }

    QModelIndex TagModelProxy::mapToSource (const QModelIndex& proxyIndex) const
    {
        if (!proxyIndex.isValid()) return QModelIndex();
        if (p->isDirect()) return sourceModel()->index(proxyIndex.row(),proxyIndex.column());

        return sourceModel()->index(p->mapToSource(proxyIndex.row()),proxyIndex.column());
    }

    QModelIndex TagModelProxy::mapFromSource (const QModelIndex& sourceIndex) const
    {
        if (!sourceIndex.isValid()) return QModelIndex();
        if (p->isDirect()) return index(sourceIndex.row(),sourceIndex.column());

        QVector<int> proxyRows = p->mapFromSource(sourceIndex.row());
        if (proxyRows.isEmpty()) return QModelIndex();

        return index(proxyRows[0],sourceIndex.column());
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////
    //
    // Utilities for proxy functionalities
    //
    //////////////////////////////////////////////////////////////////////////////////////////////////////

    void TagModelProxy::createBackMap (int)
    {
        p->m_MapToProxy.resetBackmap();
    }

    void TagModelProxy::setNewMap (const QVector<int>& map)
    {
        const int oldSize = rowCount();
        const int newSize = map.size();

        // Signals must be blocked as redraw is inconsistent before installing new backmap
        blockSignals(true);
        if (newSize>oldSize) {
            beginInsertRows(QModelIndex(),0,newSize-oldSize-1);
            p->m_MapToSource=map;
            endInsertRows();
        }
        else if (newSize<oldSize) {
            beginRemoveRows(QModelIndex(),0,oldSize-newSize-1);
            p->m_MapToSource=map;
            endRemoveRows();
        }
        else p->m_MapToSource=map;
        createBackMap();
        blockSignals(false);

        if (!p->m_HeaderFactory->getAllHeaderRows().isEmpty()) {
            emit addHeader (p->m_HeaderFactory->getAllHeaderRows());
        }
    }

    void TagModelProxy::setSource (Source::Base* source)
    {
        p->m_Source.reset(source);

        if (source->isDirect() && p->m_Filters.empty() && p->m_Grouper.get()==0l && p->m_Sorter.get()==0l) {
            beginInsertRows(QModelIndex(),0,sourceModel()->rowCount()-rowCount());
            p->m_MapToSource.clear();
            p->m_MapToProxy.clear();
            endInsertRows();
            reset();
            emit hasFilter(false);
            return;
        }

        QVector<int> map=source->source();
        if (!p->m_Filters.isEmpty()) {
            try {
                QVector<TagItemFilter::Keeper> keep=p->applyFilters(map);
                map.clear();
                foreach (TagItemFilter::Keeper k, keep) {
                    map << k.row;
                    foreach (RowHighlight h, k.highlights) {
                        setMark(dynamic_cast<TagModel*>(sourceModel())->getRow(k.row), h.first, h.second.from, h.second.to);
                    }
                }
            }
            catch (NoFilter)
            {
            }
        }

        if (p->m_Grouper.get()!=0l) {
            map=setGrouping(p->m_Grouper.get(),map);
        }
        else if (p->m_Sorter.get()!=0l) {
            p->m_Sorter->sort(map.begin(),map.end());
        }

        emit hasFilter(true);
        setNewMap(map);
        reset();
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////
    //
    // Filter support
    //
    //////////////////////////////////////////////////////////////////////////////////////////////////////

    bool TagModelProxy::hasFilter() const
    {
        return !p->m_Filters.isEmpty();
    }

    void TagModelProxy::resetFilter ()
    {
        emit modelAboutToBeFiltered();

        if (!p->m_Filters.isEmpty() || !p->m_Source->isDirect()) {
            p->removeFilters();
            setSource(new Source::Direct(this));
            clearMarks();
        }
    }

    void TagModelProxy::filterCoverOnly ()
    {
        filterCover (true);
    }

    void TagModelProxy::filterNoCoversOnly ()
    {
        filterCover (false);
    }

    void TagModelProxy::filterCover (bool includeCovers)
    {
        TagItemFilter::Base* filter = new TagItemFilter::Cover(includeCovers);
        addFilter(filter);
    }

    void TagModelProxy::filterCoverSupportingFormats ()
    {
        TagItemFilter::Base* filter=new TagItemFilter::CoverFormats(true);
        addFilter (filter);
    }

    void TagModelProxy::filterDuplicates ()
    {
        TagItemFilter::Base* filter = new TagItemFilter::Duplicates();
        addFilter(filter);
    }

    void TagModelProxy::filterIncomplete ()
    {
        TagItemFilter::Base* filter = new TagItemFilter::IncompleteTags();
        addFilter(filter);
    }

    bool TagModelProxy::filter (const QString& expression)
    {
        TagItemFilter::Base* filter = 0l;
        switch (p->m_FilterFlags&3) {
        case AsString: filter = new TagItemFilter::ExpressionAsString(expression, p->m_FilterFlags); break;
        case SingleWord: filter = new TagItemFilter::ExpressionSingleWord(expression, p->m_FilterFlags); break;
        case AllWords: filter = new TagItemFilter::ExpressionAllWords(expression, p->m_FilterFlags); break;
        }
        if (filter!=0l) {
            addFilter(filter);
            return true;
        }
        return false;
    }

    void TagModelProxy::addFilter (TagItemFilter::Base* filter)
    {
        const bool isDirect = p->isDirect();
        if (isDirect) {
            p->m_MapToSource=p->m_Source->source();
        }
        QVector<TagItemFilter::Keeper> result=TagItemFilter::filter(p->m_MapToSource,filter,this);

        if (!result.isEmpty()) {
            p->m_Filters.append(filter);

            emit modelAboutToBeFiltered();
            clearMarks();
            QVector<int> newMap;
            int i=0;
            foreach (TagItemFilter::Keeper k, result) {
                newMap << k.row;
                foreach (RowHighlight h, k.highlights) {
                    setMark(dynamic_cast<TagModel*>(sourceModel())->getRow(k.row), h.first, h.second.from, h.second.to);
                }
                i++;
            }

            setNewMap (newMap);
            emit hasFilter(true);
            reset();
        }
        else {
            if (isDirect) {
                p->m_MapToSource.clear();
            }
            emit message(i18n("Filter contains no data. Ignored!"),2000);
        }
    }

    void TagModelProxy::setExpressionModeFlags (ExpressionFilterFlags mode)
    {
        p->m_FilterFlags=mode;
    }

    TagModelProxy::ExpressionFilterFlags TagModelProxy::getExpressionModeFlags() const
    {
        return p->m_FilterFlags;
    }

    void TagModelProxy::applyExternalFilter (QList<TagItem*> items)
    {
        if (!items.empty()) {
            QVector<int> map;
            map.reserve(items.size());
            foreach (TagItem* item, items) {
                map << item->getRow();
            }
            Source::Base* source = new Source::FromSelection (map);
            setSource(source);
        }
        else {
            emit message(i18n("Filter contains no data. Ignored!"),2000);
        }
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////
    //
    // Grouping support
    //
    //////////////////////////////////////////////////////////////////////////////////////////////////////

    QVector<int> TagModelProxy::setGrouping (Grouping::Base* grouper, QVector<int>& map)
    {
        emit clearHeaders ();
        p->m_HeaderFactory->clear();
        QVector<int> result=grouper->group(map,p->m_Sorter.get());
        return result;
    }

    void TagModelProxy::removeGrouping ()
    {
        if (p->hasGrouping()) {
            // At least duplicates may also contain duplicates this is taken into account here.
            // Simply use the (invalid anyway) m_MapToSource array to store which values
            // are already included in the rebuild m_MapToSource
            const int oldSize = rowCount();
            int l=0;
            for (int i=0;i<p->m_MapToSource.size();i++) {
                if (p->m_MapToSource[i]>=0 && p->m_MapToProxy[p->m_MapToSource[i]].size()>0) {
                    p->m_MapToSource[l++]=p->m_MapToSource[i];
                    p->m_MapToProxy[p->m_MapToSource[i]].clear();
                }
            }
            p->m_MapToSource.resize(l);

            p->m_HeaderFactory->clear();
            emit clearHeaders();
            p->m_Grouper.reset();

            beginRemoveRows(QModelIndex(),0,oldSize-rowCount());
            createBackMap();
            endRemoveRows();

            reset();
        }
    }

    void TagModelProxy::groupBy ()
    {
        QObject* object=sender();
        if (object==0l) {
            kWarning() << "can only be called via a signal. Aborting method." << endl;
        }

        QVariant groupBy=object->property("groupby");
        if (groupBy.isNull() || groupBy.type()!=QVariant::String) {
            QString warning;
            if (groupBy.isNull()) {
                warning="Undefined";
            }
            else {
                warning=QString("Invalid (Type ")+groupBy.typeName()+")";
            }
            kWarning() << warning << " property \"groupby\" send from action " << sender()->objectName() << endl;
            return;
        }

        if (groupBy=="Duplicates") {
            p->m_Grouper.reset(new Grouping::ByDuplicate (this, p->m_HeaderFactory.get()));
        }
        else if (groupBy=="Decade") {
            p->m_Grouper.reset(new Grouping::ByDecade (this, p->m_HeaderFactory.get()));
        }
        else if (groupBy=="Last") {
            // Do nothing, keep current state
            if (!p->hasGrouping()) {
                emit message (i18n("No preselected group by. Please do a group by first."),2000);
                return;
            }
        }
        else {
            const int col = ColumnManager::instance()->indexByName(groupBy.toString());
            if (col==-1) return;
            p->m_Grouper.reset(new Grouping::ByColumn (this, p->m_HeaderFactory.get(), col));
        }

        if (p->isDirect()) {
            p->m_MapToSource=p->m_Source->source();
        }
        QVector<int> map=setGrouping(p->m_Grouper.get(),p->m_MapToSource);
        setNewMap(map);
        emit hasFilter(true);
        reset();
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////
    //
    // Signal handling from source model
    //
    //////////////////////////////////////////////////////////////////////////////////////////////////////

    void TagModelProxy::onAlteredExtension (int row, int type, const QString& name, const QVariant& value)
    {
        foreach (int r, p->mapFromSource(row)) {
            emit alteredExtension(r,type,name,value);
        }
    }

    void TagModelProxy::fromSourceColumnsAboutToBeInserted (const QModelIndex& parent, int start, int end)
    {
        beginInsertColumns(mapFromSource(parent),start,end);
    }

    void TagModelProxy::fromSourceColumnsAboutToBeRemoved (const QModelIndex& parent, int start, int end )
    {
        beginRemoveColumns(mapFromSource(parent),start,end);
    }

    void TagModelProxy::fromSourceColumnsInserted (const QModelIndex &, int , int)
    {
        endInsertColumns();
    }

    void TagModelProxy::fromSourceColumnsRemoved (const QModelIndex&, int , int)
    {
        endRemoveColumns();
    }

    void TagModelProxy::fromSourceDataChanged ( const QModelIndex & topLeft, const QModelIndex & bottomRight )
    {
        emit dataChanged(mapFromSource(topLeft),mapFromSource(bottomRight));
    }

    void TagModelProxy::fromSourceHeaderDataChanged ( Qt::Orientation orientation, int first, int last )
    {
        emit headerDataChanged(orientation,first,last);
    }

    void TagModelProxy::fromSourceLayoutAboutToBeChanged ()
    {
        emit layoutAboutToBeChanged();
    }

    void TagModelProxy::fromSourceLayoutChanged ()
    {
        emit layoutChanged();
    }

    void TagModelProxy::fromSourceModelAboutToBeReset ()
    {
    }

    void TagModelProxy::fromSourceModelReset ()
    {
        reset();
    }

    void TagModelProxy::fromSourceRowsAboutToBeInserted (const QModelIndex& parent, int start, int end)
    {
        beginInsertRows(mapFromSource(parent),start,end);
    }

    void TagModelProxy::fromSourceRowsAboutToBeRemoved (const QModelIndex& parent, int start, int end)
    {
        if (p->isDirect()) {
            beginRemoveRows(mapFromSource(parent),start,end);
            return;
        }

        emit clearHeaders();

        /*
        for (int i=start;i<=end;i++) {
            for (int j=0;j<p->m_MapToProxy[i].size();j++) {
                p->m_MapToSource[p->m_MapToProxy[i][j]]=s_Marked;
            }
        }
        */

        Private::RedoMap redoMap;
        for (int i=0;i<p->m_MapToSource.size();i++) {
            if (p->m_MapToSource[i]>=start && p->m_MapToSource[i]<=end) {
                p->m_MapToSource[i]=s_Marked;
                // Remember mapping for redo operation (and restrict stacksize to 20)
                redoMap << i;
            }
        }
        p->m_RedoStack.append (QPair<int,Private::RedoMap>(start,redoMap));
        if (p->m_RedoStack.size()>20) p->m_RedoStack.takeFirst();

        // sweep over it again and remove orphaned headers
        int first=-1;
        for (int i=0;i<p->m_MapToSource.size();i++) {
            if (first==-1 && p->m_MapToSource[i]==s_Marked) {
                first=i;
            }
            else if (first>=0) {
                if (p->m_MapToSource[i]<0) {
                    if (first>0 && p->m_MapToSource[first-1]<0) p->m_MapToSource[first-1]=s_Marked;
                    first=i;
                }
                else if (p->m_MapToSource[i]!=s_Marked) {
                    first=-1;
                }
            }
        }

    }

    void TagModelProxy::fromSourceRowsInserted (const QModelIndex& parent, int start, int end)
    {
        if (p->isDirect()) {
            endInsertRows();
            return;
        }

        const int count = end-start+1;
        // Heuristic: Test if this a redo operation, and insert at the right place
        if (!p->m_RedoStack.isEmpty() && start==p->m_RedoStack.last().first) {
            Private::RedoMap redomap = p->m_RedoStack.takeLast().second;
            Private::RedoMap sorted=redomap;
            qSort(sorted);
            // Can be simply inserted as we assume that the model has the state before the
            // items where deleted
            QList<Chunk > chunks;
            int i=p->m_MapToSource.size()-1;
            p->m_MapToSource.resize(p->m_MapToSource.size()+count);
            for (int x=p->m_MapToSource.size()-1;i>=0;i--) {
                if (x==sorted.last()) {
                    int pos=-1;
                    for (int j=0;j<redomap.size();j++) {
                        if (redomap[j]==x) {
                            pos=j;
                            break;
                        }
                    }
                    chunks << Chunk(x,x);
                    p->m_MapToSource[x--]=start+pos;
                }

                int v=p->m_MapToSource[i];
                if (v>=start) {
                    v+=count;
                }
                p->m_MapToSource[x--]=v;

                //if (x>=0) cout << x << "," << p->m_MapToSource[x] << endl;
            }

            createBackMap();

            for (int i=0;i<chunks.size();i++) {
                beginInsertRows (QModelIndex(),chunks[i].first,chunks[i].second);
                endInsertRows();
            }

            /*        cout << "Size map: " << p->m_MapToSource.size() << endl
                << "Row count:"  << rowCount() << endl
                << "Source row count " << sourceModel()->rowCount() << endl;
*/
            return;
        }

        // append only
        const int size=p->m_MapToSource.size();
        beginInsertRows(mapFromSource(parent),size,size+count-1);
        for (int i=0;i<count;i++) {
            p->m_MapToSource << start++;
            p->m_MapToProxy.append(QVector<int>(1,size+i));
        }
        endInsertRows();
    }

    void TagModelProxy::fromSourceRowsRemoved (const QModelIndex&, int _start, int end)
    {
        if (p->isDirect()) {
            endRemoveRows();
            return;
        }

        const int count=end-_start+1;

        // copy back and shrink: Previouly marked map indices (set to s_Marked) are removed.
        // This is done in a single large copy loop. All subsequent marked rows are combined to a chunk
        // after the copy operation finished these chunks are reoved from the model by using
        // begin/endRemoveRows. To stay consistent all the time two tricks are used. First the
        // removement begins from the back so that all (absolute) row indexes stay valid even when
        // row are removed. Second keep a dummy field up to the old size at the end (which is marked
        // and will be ignored by the mapping in the data routines). By that a step by step reduction
        // of the mapping array works smothless.
        QList<Chunk> chunks;
        int x=0;
        int start=-1;
        int headerNo=0;
        for (int i=0;i<p->m_MapToSource.size();i++) {
            if (p->m_MapToSource[i]!=s_Marked) {
                int v=p->m_MapToSource[i];
                if (v<0) {
                    p->m_HeaderFactory->setHeaderRow(headerNo++,x);
                }
                else if (v>=_start) v-=count;
                p->m_MapToSource[x++]=v;
                if (start!=-1) {
                    chunks << Chunk(start,i-1);
                    start=-1;
                }
            }
            else {
                if (start==-1) start=i;
            }
        }
        if (start!=-1) {
            chunks << Chunk(start,p->m_MapToSource.size()-1);
        }

        if (sourceModel()->rowCount()==0) {
            // Model cleared, reset all filters
            if (!p->m_Source->isDirect()) {
                p->m_Source.reset(new Source::Direct(this));
            }
            p->m_Grouper.reset();
            p->m_Sorter.reset();
            emit hasFilter(false);
        }

        createBackMap(x);

        // Mark leftover rows as used so that these are not mapped when accessed
        // (source model doesn't have the rows anymore and reduction takes place
        // in parts) -> data() will no try to map and fetch from sourceModel
        for (int i=x;i<p->m_MapToSource.size();i++) p->m_MapToSource[i]=s_Marked;

        // Remove chunks backwards so that the rows do not change
        for (int i=chunks.size()-1;i>=0;i--) {
            beginRemoveRows (QModelIndex(),chunks[i].first,chunks[i].second);
            endRemoveRows();
        }
        p->m_MapToSource.resize(x);

        emit addHeader(p->m_HeaderFactory->getAllHeaderRows());
    }

    bool TagModelProxy::removeRows (int row, int count, const QModelIndex& parent)
    {
        if (row+count>rowCount() || row<0) return false;

        // Remove marks if any set
        if (!p->m_Highlights.isEmpty()) {
            for (int i=row;i<row+count;i++) {
                const TagItem* tag = getRow(i);
                QMap<HighlightIndex, Mark >::iterator iter = p->m_Highlights.find(HighlightIndex(tag));
                while (iter!=p->m_Highlights.end()) {
                    if (iter.key().tag==tag) iter=p->m_Highlights.erase(iter);
                    else break;
                }
            }
        }

        const QModelIndex sourceParent = mapToSource(parent);
        if (p->isDirect()) {
            sourceModel()->removeRows(row,count,sourceParent);
        }
        else {
            QVector<int> sourceIndexes;
            sourceIndexes.resize(count);
            for (int i=0;i<count;i++) {
                sourceIndexes[i]=p->m_MapToSource[i+row];
            }
            qSort(sourceIndexes.begin(),sourceIndexes.end(),qGreater<int>());

            int last=sourceIndexes[0];
            int first=last;
            for (int i=1;i<sourceIndexes.size();i++) {
                if (sourceIndexes[i]==first-1 || sourceIndexes[i]==first) {
                    --first;
                }
                else {
                    sourceModel()->removeRows(first,last-first+1,sourceParent);
                    first=last=sourceIndexes[i];
                }
            }
            sourceModel()->removeRows(first,last-first+1,sourceParent);
        }

        return true;
    }

    bool TagModelProxy::insertRows (int row, int count, const QModelIndex& parent)
    {
        if (p->isDirect()) return sourceModel()->insertRows(row,count,mapToSource(parent));
        return sourceModel()->insertRows(p->m_MapToSource[row],count,mapToSource(parent));
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////
    //
    // Mark and highligh support
    //
    //////////////////////////////////////////////////////////////////////////////////////////////////////

    void TagModelProxy::clearMarks ()
    {
        foreach (HighlightIndex index,p->m_Highlights.keys()) {
            index.tag->setHasHighlights(false);
            const QModelIndex idx=sourceModel()->index(index.tag->getRow(),index.column);
            dynamic_cast<TagModel*>(sourceModel())->emitChangedCell(idx);
        }

        p->m_Highlights.clear();
    }

    void TagModelProxy::setMark (const TagItem* row, int col, int from, int to)
    {
        if (row->getHasHighlights()==true) {
            for (QMultiMap<HighlightIndex,Mark>::Iterator iter=p->m_Highlights.lowerBound(HighlightIndex(row,col));
                 iter!=p->m_Highlights.upperBound(HighlightIndex(row,col));) {
                if (iter.value().from>=from and iter.value().to<=to) {
                    // item completly contained in new highlight mark: remove old one
                    iter=p->m_Highlights.erase(iter);
                    continue;
                }
                if (iter.value().from<=to && iter.value().to>=to) {
                    // item starts in the current highlight but moves over the right edge
                    from=std::min(iter.value().from,(short)from);
                    to=iter.value().to;
                    iter=p->m_Highlights.erase(iter);
                    continue;
                }
                if (iter.value().from<from && iter.value().to<=to) {
                    // item starts in the current highlight but moves over the left edge
                    from=iter.value().from;
                    to=std::min(iter.value().to,(short)to);
                    iter=p->m_Highlights.erase(iter);
                    continue;
                }
                if (iter.value().from<from && iter.value().to>to) {
                    // item is larger than the new highlight, nothing to insert
                    return;
                }
                iter++;
            }
        }

        p->m_Highlights.insertMulti(HighlightIndex(row,col),Mark(from,to));
        row->setHasHighlights(true);

        const QModelIndex idx=sourceModel()->index(row->getRow(),col);
        dynamic_cast<TagModel*>(sourceModel())->emitChangedCell(idx);
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////
    //
    // Drag and drop support
    //
    //////////////////////////////////////////////////////////////////////////////////////////////////////

    QStringList TagModelProxy::mimeTypes () const
    {
        QStringList mimes;
        mimes << "text/uri-list";
        return mimes;
    }

    Qt::DropActions TagModelProxy::supportedDropActions () const
    {
        return Qt::MoveAction | super::supportedDropActions();
    }

    bool TagModelProxy::dropMimeData (const QMimeData* data, Qt::DropAction action, int, int, const QModelIndex& )
    {
        if (action==Qt::IgnoreAction) return true;
        if (data==0l) return false;

        if (!data->hasFormat("text/uri-list")) {
            kWarning() << "Currently only \"text/uri.list\" MIME drops are accepted. The available types are: "
                    << data->formats() << endl;
            return false;
        }

        dynamic_cast<TagModel*>(sourceModel())->load (data->urls());

        return true;
    }

    //
    // iterator support
    //

    TagModelProxy::iterator::iterator ()
        : searchableOnly(false),
        wrapOver(false),
        getRawData(false),
        selection(0l),
        current(QModelIndex()),
        father(0l)
    {
    }

    TagModelProxy::iterator::iterator (QAbstractItemModel* _father, const QModelIndex& index)
        : searchableOnly(false),
        wrapOver(false),
        getRawData(false),
        selection(0l),
        selectedIndexes(),
        current(index),
        start(index),
        father(_father)
    {
    }

    TagModelProxy::iterator::iterator (const iterator& other)
        : searchableOnly(other.searchableOnly),
        wrapOver(other.wrapOver),
        getRawData(false),
        selection(other.selection),
        selectedIndexes(other.selectedIndexes),
        current(other.current),
        start(other.start),
        father(other.father)
    {
    }

    TagModelProxy::iterator::~iterator()
    {
    }

    TagModelProxy::iterator& TagModelProxy::iterator::operator= (const iterator& other)
    {
        searchableOnly=other.searchableOnly;
        wrapOver=other.wrapOver;
        getRawData=other.getRawData;
        selection=other.selection;
        selectedIndexes=other.selectedIndexes;
        current=other.current;
        father=other.father;
        start=other.start;
        return *this;
    }

    void TagModelProxy::iterator::setSearchableOnly (bool so)
    {
        searchableOnly=so;
        if (!isValidIterator(current)) current=fforward(current).qmodelindex(father);
    }

    void TagModelProxy::iterator::setSelectionOnly (const QItemSelectionModel* so)
    {
        if (so==0l) {
            selection=0l;
            selectedIndexes.clear();
        }
        else {
            selection=so;
            selectedIndexes=so->selectedIndexes();
            currentSelectedItem=selectedIndexes.begin();
            current=*currentSelectedItem;
        }

        if (!isValidIterator(current)) {
            current=fforward (current).qmodelindex(father);
        }
    }

    void TagModelProxy::iterator::setWrapOver (bool wo)
    {
        wrapOver=wo;
    }

    bool TagModelProxy::iterator::operator== (const iterator& other) const
    {
        return current==other.current;
    }

    bool TagModelProxy::iterator::operator!= (const iterator& other) const
    {
        return current!=other.current;
    }

    QVariant TagModelProxy::iterator::operator* () const
    {
        return father->data(current,(getRawData)?(int)TagModel::RawModeRole:(int)Qt::DisplayRole);
    }

    bool TagModelProxy::iterator::isValidIterator (const SlimIndex& idx) const
    {
        return !idx.isValid() ||
               (!searchableOnly || ColumnManager::instance()->isSearchable(idx.column()));
    }

    SlimIndex TagModelProxy::iterator::fforward (const SlimIndex& idx) const
    {
        if (selection!=0l) {
            while (currentSelectedItem!=selectedIndexes.end())
            {
                ++currentSelectedItem;
                if (currentSelectedItem!=selectedIndexes.end() && isValidIterator(*currentSelectedItem)) {
                    return *currentSelectedItem;
                }
            }
            return SlimIndex();
        }
        else {
            SlimIndex idx2(idx);
            do {
                if (idx2.column()+1>=ColumnManager::instance()->getColumnNumber()) {
                    if (idx2.row()+1==father->rowCount()) {
                        return SlimIndex();
                    }
                    idx2.incRow();
                    idx2.setCol(0);
                }
                else {
                    idx2.incCol();
                }
            } while (!isValidIterator(idx2));
            return idx2;
        }
    }

    SlimIndex TagModelProxy::iterator::fbackward (const SlimIndex& idx) const
    {
        if (selection!=0l) {
            do {
                if (currentSelectedItem==selectedIndexes.begin()) {
                    return SlimIndex();
                }
                --currentSelectedItem;
            } while (!isValidIterator(*currentSelectedItem));
            return *currentSelectedItem;
        }
        else {
            SlimIndex idx2(idx);
            do {
                if (idx2.column()==0) {
                    if (idx2.row()==0) {
                        return SlimIndex();
                    }
                    idx2=SlimIndex(idx2.row()-1,ColumnManager::instance()->getColumnNumber()-1);
                }
                else {
                    idx2.decCol();
                }
            } while (!isValidIterator(idx2));
            return idx2;
        }
    }

    TagModelProxy::iterator& TagModelProxy::iterator::operator++ (int)
    {
        current=fforward (current).qmodelindex(father);

        if (!current.isValid() && wrapOver) {
            // Superseeded last row. In case of no wrapover set to end
            current=fforward(SlimIndex(0,0)).qmodelindex(father);
        }
        if (current==start) {
            current=QModelIndex();
        }

        return *this;
    }

    TagModelProxy::iterator& TagModelProxy::iterator::operator-- (int)
    {
        current=fbackward(current).qmodelindex(father);
        if (current==start) {
            current==QModelIndex();
        }
        else {
            if (!current.isValid() && wrapOver) {
                current=father->index(father->rowCount()-1,ColumnManager::instance()->getColumnNumber()-1);
            }
        }
        return *this;
    }

    TagModelProxy::iterator TagModelProxy::begin(const QModelIndex& index)
    {
        if (index.isValid()) {
            return iterator(this, index);
        }
        return iterator (this, createIndex(0,0));
    }

    TagModelProxy::iterator TagModelProxy::end()
    {
        return iterator (this, QModelIndex());
    }

}

#include "tagmodelproxy.moc"
