#ifndef TAGMODELPROXY_H
#define TAGMODELPROXY_H

#include <QAbstractProxyModel>
#include <QList>
#include <QObject>

#include "model/texeltagmodel.h"

class QCompleter;
class QItemSelectionModel;

namespace Texel {

    namespace Grouping {
        class Base;
    }

    namespace TagItemFilter {
        class Base;
    }

    namespace Source {
        class Base;
    }

    class TagItem;

    /**
    */
    class TagModelProxy: public QAbstractProxyModel
    {
        Q_OBJECT

        typedef QAbstractProxyModel super;

    public:
        enum ExpressionFilterFlag
        {
            AsString=1,
            SingleWord=2,
            AllWords=3,
            CaseSensitive=4,
            WholeWords=8
         };

        Q_DECLARE_FLAGS(ExpressionFilterFlags, ExpressionFilterFlag)

        /**
        @brief
        Provides an configurable iterator over all (or a defined fraction) cells of the model.

        This iterator can be used to iterate over all cells regardless of their (column)type. ts
        useful e.g. for searching. The iterator is configurable to limit the set of columns to visit,
        to wrap around at the last cell. etc.

        The iterator allows to set its iteration properties at every time. That means tha you can
        change the mode after already iterating over some cells. But it is recommended that you
        create the iterator, set its properties and start iterating then.
        */
        class iterator
        {
        public:
            /** creates a new invalid iterator. It is not attached to a model and does not point to a cell.
        */
            iterator ();

            /** copy constructor.
        */
            iterator (const iterator& other);
            ~iterator();

            iterator& operator++ (int);
            iterator& operator-- (int);

            /** Dereferencing operator.
        @returns the content of a cell. This is a QVariant as this is de defined interface.
        */
            QVariant operator* () const;

            /** Checks to iterators for equality. Two iterators are equal when the current index of them
        is equal. It is not compared if the iterators belong to the same model, or have the
        same selection or direction criteria.
        @returns true, when this and the otheriterator point to the same index, false otherwise.
        */
            bool operator== (const iterator& other) const;

            /// Opposite of operator==(const iterator& other) const. Does also only compare the current
            /// index. @returns true, when the iterators are different.
            bool operator!= (const iterator& other) const;
            iterator& operator= (const iterator& other);

            /**
        When set to true the iterator does only visit cells which are marked in the ColumnManager
        as searchable.
        */
            void setSearchableOnly (bool);

            /** When set to true the iterator will iterate until all cells are visited, even when the start
        cell was somewhere in the middle. Otherwise the iterator will stop at the end of the model.
        */
            void setWrapOver (bool);

            /**
        Will only iterate over the cells marked by the provided selection model.
        */
            void setSelectionOnly (const QItemSelectionModel* selection);

            void setGetRawData (bool rw)
            {
                getRawData=rw;
            }

            /** Returns the current cell index the iterator points to. This is different from
        dereferencing the iterator which gets the value of the cell.
        @returns the current index
        */
            const QModelIndex& index () const
            {
                return current;
            }

            void pin ()
            {
                start=current;
            }

        private:

            /** @brief "Real" constructor of the iterator. Is a private element as it will be called
        by the ouer class begin() method. It attached the iterator to a model and makes it by that
        dereferencable.
        */
            explicit iterator (QAbstractItemModel* father, const QModelIndex& index);

            bool isValidIterator (const SlimIndex& idx) const;
            SlimIndex fforward (const SlimIndex& idx) const;
            SlimIndex fbackward (const SlimIndex& idx) const;

            bool searchableOnly;
            bool wrapOver;
            bool getRawData;

            const QItemSelectionModel* selection;
            QModelIndexList selectedIndexes;
            mutable QModelIndexList::iterator currentSelectedItem;

            QModelIndex current;
            QModelIndex start;
            QAbstractItemModel* father;

            friend class TagModelProxy;
        };

        enum Role { RoleGroupingRow=1000 };

        /**
    @param index Sarting cell of the iterator. By default this is the first cell in the model.
    @returns a new iterator object which ponts to the current model to the cell descibed by index.
    */
        iterator begin(const QModelIndex& index=QModelIndex());
        iterator end();

        TagModelProxy(QObject* parent=0l);
        virtual ~TagModelProxy();

        virtual QModelIndex mapToSource (const QModelIndex& proxyIndex) const;
        virtual QModelIndex mapFromSource (const QModelIndex& sourceIndex) const;

        virtual bool removeRows (int row, int count, const QModelIndex& parent = QModelIndex());
        virtual bool insertRows (int row, int count, const QModelIndex& parent = QModelIndex());

        TagItem* getRow (int row) const;
        QVector<int> getRow (const TagItem* row) const;

        QCompleter* getCompleter(const QModelIndex& index) const;

        void paste (const QModelIndex& from, const QModelIndex& to, const QVariantList& values, const QString& description);

        virtual bool dropMimeData (const QMimeData* data, Qt::DropAction action, int row, int column, const QModelIndex& parent);
        virtual QStringList mimeTypes () const;
        virtual Qt::DropActions supportedDropActions () const;
        /**
     Checks the current selection to make a smart guess about the best way to fill down. Starting at the
     current cursor position the function checks if a rectangle was selected. If so, all adjecent
     selected rows are used for filling down - but just within the selected range.
     Otherwise only the current column and only the adjecent selected rows are used for filldown.
     */
        void fillDownMoreThanOneRow(const QItemSelectionModel& sel);

        void clearMarks ();
        void setMark (const TagItem* row, int col, int from, int to);

        /** Provides the possibility to map the tags stored in the model (based on @class TagItem)
     to an arbitrary result set. The mapping is done parallized by usig QtConcurrent
     functionality.
     Just provide a @class MapFunctor which @fn operator() takes a @class TagItem pointer as
     input. Its result must be able to be stored in the sequence @class T.
     @return Sequence @class T with all members of the model mapped via the functor @class MapFunctor
     */
        template <typename T, typename MapFunctor>
                QFuture<T> mapTo (MapFunctor mapper) const
        {
            TagItem::List tags = getTagItemList();
            return QtConcurrent::mapped<T,TagItem::List::const_iterator,MapFunctor> (tags.begin(), tags.end(), mapper);
        }

        template <typename T, typename MapFunctor, typename ReduceFunctor>
                T mapAndReduce (MapFunctor mapper, ReduceFunctor reducer)
        {
            TagItem::List tags = getTagItemList();
            return QtConcurrent::blockingMappedReduced<T,TagItem::List::const_iterator,MapFunctor,ReduceFunctor> (tags.begin(), tags.end(), mapper, reducer);
        }

        virtual QVariant headerData ( int section, Qt::Orientation orientation, int role = Qt::DisplayRole ) const;
        virtual QVariant data (const QModelIndex& index, int role = Qt::DisplayRole) const;
        virtual bool setData(const QModelIndex &index, const QVariant &value, int role=Qt::EditRole);

        virtual QModelIndex index(int row, int column,
                                  const QModelIndex &parent = QModelIndex()) const;
        virtual QModelIndex parent(const QModelIndex &child) const;
        virtual int rowCount(const QModelIndex &parent = QModelIndex()) const;
        virtual int columnCount(const QModelIndex &parent = QModelIndex()) const;

        virtual void setSourceModel (QAbstractItemModel* sourceModel);

        virtual void sort (int column, Qt::SortOrder order = Qt::AscendingOrder);

        virtual Qt::ItemFlags flags (const QModelIndex& index) const;

        /**
           Filter as the items define. Can be used to define external filters of tracks.
           Just the items which are in the items list are displayed. All other filters
           are reseted.
          */
        void applyExternalFilter (QList<TagItem*> items);

        /**
           This function changes the view of formulars. Usually the result of a formular is displayed,
           but when this methods is called with show set to true, the formular itself is displayed.
           @param show The view displays formulars instead of the resulting string, when set to true.
        */
        void showFormulars (bool show);

        bool isHeader (int row) const;

        QString getFileName (int row) const;

        int getGroupingColumn () const;

        /**
          Returns a list of rows which are referred by this header. If the given row is not a header
          the result list is empty.
          */
        TagItem::List getRowsForHeader (int headerrow) const;

        QPair<int,int> getRangeForHeader (int headerrow) const;

        void setDataBlock (const QVariant& ref, const QModelIndexList& indexes);

        bool hasFilter() const;

        TagItem::List getTagItemList () const;

    signals:
        // The following signals are necessary as the view manages spans and a header is column 0
        // which does span all columns. The current layout is signaled by these method to the view.
        void clearHeaders ();
        //void addHeaders (const QVector<int>& rows);
        void addHeader(const QList<int>& rows);
        // --- end of spanning signals

        void message (const QString& message, int milliseconds);

        /** Emitted when extended attributes of a track are changed. To identify the altered attribute a generic
            mechanism is used.
            @param row Row whichs extension is altered
            @param type Type of the changed extension. Currently 0=Image, 1=lyrics and 2=string attribute is defined
            @param name Name of the attribute, image or lyric part changed
          */

        void alteredExtension (int row, int type, const QString& name, const QVariant& value);

        void hasFilter (bool filter);

        void modelAboutToBeFiltered ();

    public slots:
        /** Limits the currently visible rows to such which have duplicates attached to. If
       no duplicates are available this function does nothing (should be changed to some
       behaviour, as this doesn't show anything to the user).
     */
        void filterDuplicates ();
        void filterIncomplete ();
        void filterCoverOnly ();
        void filterNoCoversOnly ();
        void filterCoverSupportingFormats ();
        void resetFilter ();
        bool filter (const QString& expression);
        void groupBy ();
        void removeGrouping ();

        void setExpressionModeFlags (ExpressionFilterFlags mode);
        ExpressionFilterFlags getExpressionModeFlags () const;

        void removeFiles (const TagItem::List& list);

    private slots:
        /* The following list of functions are slots which capture the changes of the underlying
      model and execute necessary changes on the proxy.
     */
        void fromSourceColumnsAboutToBeInserted ( const QModelIndex & parent, int start, int end );
        void fromSourceColumnsAboutToBeRemoved ( const QModelIndex & parent, int start, int end );
        void fromSourceColumnsInserted ( const QModelIndex & parent, int start, int end );
        void fromSourceColumnsRemoved ( const QModelIndex & parent, int start, int end );
        void fromSourceDataChanged ( const QModelIndex & topLeft, const QModelIndex & bottomRight );
        void fromSourceHeaderDataChanged ( Qt::Orientation orientation, int first, int last );
        void fromSourceLayoutAboutToBeChanged ();
        void fromSourceLayoutChanged ();
        void fromSourceModelAboutToBeReset ();
        void fromSourceModelReset ();
        void fromSourceRowsAboutToBeInserted ( const QModelIndex & parent, int start, int end );
        void fromSourceRowsAboutToBeRemoved ( const QModelIndex & parent, int start, int end );
        void fromSourceRowsInserted ( const QModelIndex & parent, int start, int end );
        void fromSourceRowsRemoved ( const QModelIndex & parent, int start, int end );

        void onClear ();
        void onAlteredExtension (int row, int type, const QString& name, const QVariant& value);

    private:
        struct NoFilter {};

        void setNewMap (const QVector<int>& map);
        void addFilter (TagItemFilter::Base* filter);
        QVector<int> setGrouping (Grouping::Base* grouper, QVector<int>& map);
        void setSource (Source::Base* source);

        void filterCover (bool includeCovers);

        static const int s_Marked;
        typedef QPair<int,int> Chunk;

        TagModel* superAsTagModel () const;

        void createBackMap (int size=-1);

        struct Private;
        Private* p;
    };

    inline QCompleter* TagModelProxy::getCompleter(const QModelIndex& index) const
    {
        return getRow(index.row())->getCompleter(index.column());
    }

    Q_DECLARE_OPERATORS_FOR_FLAGS(TagModelProxy::ExpressionFilterFlags)

    }

#endif // TAGMODELPROXY_H
