/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

/// This file contains the different extensions for the different supported file formats

#undef QT_NO_STL
#include <QString>
#include <QFileInfo>
#include <QBuffer>
#include <QMutex>
#include <QMutexLocker>

#include "model/tagspecializations.h"
#include "model/texeltagmodel.h"
#include "model/columns.h"
#include "model/tagextension.h"
#include "util/errredirector.h"
#include "options/preferences.h"

#include <taglib/taglib.h>
#include <taglib/tag.h>
#include <taglib/fileref.h>
#include <taglib/mpegfile.h>
#include <taglib/id3v2tag.h>
#include <taglib/id3v1tag.h>
#include <taglib/vorbisfile.h>
#include <taglib/speexfile.h>
#include <taglib/oggflacfile.h>
#include <taglib/mpcfile.h>
#include <taglib/flacfile.h>
#include <taglib/wavpackfile.h>
#include <taglib/trueaudiofile.h>
#include <taglib/xiphcomment.h>
#include <taglib/textidentificationframe.h>
#include <taglib/attachedpictureframe.h>
#include <taglib/unsynchronizedlyricsframe.h>
#include <taglib/popularimeterframe.h>

#if (TAGLIB_MAJOR_VERSION==1) && (TAGLIB_MINOR_VERSION>5)
#include <taglib/mp4tag.h>
#include <taglib/mp4file.h>
#include <taglib/asffile.h>
#include <taglib/aifffile.h>
#include <taglib/wavfile.h>
#endif

#include "aid.h"

using namespace std;

namespace Texel {

    namespace {
        inline TagLib::String asTagLibString (const QString& s)
        {
            return TagLib::String(s.toUtf8().data(),TagLib::String::UTF8);
        }

        inline QString asQString (const TagLib::String& s)
        {
            return QString::fromStdWString(s.toWString());
        }

        // The taglib is neither threadsave nor reentrant, therefore a roughgrain lock
        // to allow only one file to be loaded at a time
        static QMutex s_TagLibMutex;
    }

    // Factory implementation of baseclass to keep all
    // special taglib related stuff in one file (makes it easier to support
    // both 1.5 and 1.6

    TagItem* TagItem::createTagItemByType(TagModel* model, const QString& type)
    {
        if (type=="MPEG") {
            return new TagMPEG(model);
        }
        else if (type=="Ogg::Speex") {
            return new TagOggSpeex(model);
        }
        else if (type=="Ogg::FLAC") {
            return new TagOggFLAC(model);
        }
        else if (type=="Ogg::Vorbis") {
            return new TagOggVorbis(model);
        }
        else if (type=="FLAC") {
            return new TagFLAC(model);
        }
        else if (type=="MPC") {
            return new TagMPC(model);
        }
        else if (type=="TrueAudio") {
            return new TagTrueAudio(model);
        }
        else if (type=="WavPack") {
            return new TagWavPack(model);
        }
#if (TAGLIB_MAJOR_VERSION==1) && (TAGLIB_MINOR_VERSION>5)
        else if (type=="MP4") {
            return new TagMP4(model);
        }
        else if (type=="ASF") {
            return new TagASF(model);
        }
        else if (type=="RIFF::AIFF") {
            return new TagRIFFAIFF(model);
        }
        else if (type=="RIFF::WAV") {
            return new TagRIFFWAV(model);
        }
#endif
        return 0l;
    }

    TagItem* TagItem::createTagItemByType(TagModel* model, TagItemType::Type type)
    {
        if (type==TagItemType::MPEG) {
            return new TagMPEG(model);
        }
        else if (type==TagItemType::Ogg_Speex) {
            return new TagOggSpeex(model);
        }
        else if (type==TagItemType::Ogg_Flag) {
            return new TagOggFLAC(model);
        }
        else if (type==TagItemType::Ogg_Vorbis) {
            return new TagOggVorbis(model);
        }
        else if (type==TagItemType::Flac) {
            return new TagFLAC(model);
        }
        else if (type==TagItemType::MPC) {
            return new TagMPC(model);
        }
        else if (type==TagItemType::TrueAudio) {
            return new TagTrueAudio(model);
        }
        else if (type==TagItemType::WavPack) {
            return new TagWavPack(model);
        }
#if (TAGLIB_MAJOR_VERSION==1) && (TAGLIB_MINOR_VERSION>5)
        else if (type==TagItemType::MP4) {
            return new TagMP4(model);
        }
        else if (type==TagItemType::ASF) {
            return new TagASF(model);
        }
        else if (type==TagItemType::RIFF_AIFF) {
            return new TagRIFFAIFF(model);
        }
        else if (type==TagItemType::RIFF_WAV) {
            return new TagRIFFWAV(model);
        }
#endif
        return 0l;
    }

    TagItem* TagItem::createTagItem(TagModel* model, const QString& file) throw (TagItem::InvalidFormat)
    {
        using namespace TagLib;

        TagItem* item=0l;
        QFileInfo info(file);
        ErrRedirector redirector;

        QMutexLocker locker(&s_TagLibMutex);
        FileRef f((const char*)file.toLocal8Bit(),true,AudioProperties::Accurate);

        if (dynamic_cast<MPEG::File*>(f.file())!=0l) {
            item=new TagMPEG(model);
        }
        else if (dynamic_cast<Ogg::Speex::File*>(f.file())!=0l) {
            item=new TagOggSpeex(model);
        }
        else if (dynamic_cast<Ogg::FLAC::File*>(f.file())!=0l) {
            item=new TagOggFLAC(model);
        }
        else if (dynamic_cast<Ogg::Vorbis::File*>(f.file())!=0l) {
            item=new TagOggVorbis(model);
        }
        else if (dynamic_cast<FLAC::File*>(f.file())!=0l) {
            item=new TagFLAC(model);
        }
        else if (dynamic_cast<MPC::File*>(f.file())!=0l) {
            item=new TagMPC(model);
        }
        else if (dynamic_cast<TrueAudio::File*>(f.file())!=0l) {
            item=new TagTrueAudio(model);
        }
#if (TAGLIB_MAJOR_VERSION==1) && (TAGLIB_MINOR_VERSION>5)
        else if (dynamic_cast<ASF::File*>(f.file())!=0l) {
            item=new TagASF(model);
        }
        else if (dynamic_cast<MP4::File*>(f.file())!=0l) {
            item=new TagMP4(model);
        }
        else if (dynamic_cast<RIFF::AIFF::File*>(f.file())!=0l) {
            item=new TagRIFFAIFF(model);
        }
        else if (dynamic_cast<RIFF::WAV::File*>(f.file())!=0l) {
            item=new TagRIFFWAV(model);
        }
#endif
        else if (dynamic_cast<WavPack::File*>(f.file())!=0l) {
            item=new TagWavPack(model);
        }

        if (item!=0l) {
            item->doReadMetaTag(f,info);
            redirector.setItem(item);
        }
        else {
            throw TagItem::InvalidFormat(i18n("Unknown music file format for %1",file));
        }

        return item;
    }

    ExtensionAwareItem::ExtensionAwareItem (TagModel* _model)
        : TagItem (_model),
        m_Ext(0l)
    {
    }

    ExtensionAwareItem::~ExtensionAwareItem ()
    {
        if (m_Ext!=0l) {
            if (m_Ext->modified()) {
                delete m_Ext;
            }
            else {
                m_Ext->removeFromCache();
            }
        }
    }

    TagExtension* ExtensionAwareItem::getExtensions () const
    {
        if (m_Ext!=0l) {
            return m_Ext;
        }

        QMutexLocker locker(&s_TagLibMutex);
        TagLib::FileRef f((const char*)getAbsolutFilename().toLocal8Bit(),true,TagLib::AudioProperties::Fast);
        if (!f.isNull()) {
            m_Ext=createExtension(f);
        }

        return m_Ext;
    }

    bool ExtensionAwareItem::isAltered () const
    {
        const bool _isAltered = TagItem::isAltered();

        if (!_isAltered && m_Ext!=0l) {
            return m_Ext->modified();
        }

        return _isAltered;
    }

    void ExtensionAwareItem::resetExtension () const
    {
        m_Ext=0l;
    }

    QByteArray ExtensionAwareItem::convertImage(const QImage &image)
    {
        QByteArray ba;
        QBuffer buffer(&ba);
        buffer.open(QIODevice::WriteOnly);

        QString format = Preferences::instance()->coverimages.m_SaveAsJPG?"JPG":"PNG";
        if (Preferences::instance()->coverimages.m_KeepJPGAndPNG &&
            (image.text("Format")=="JPG" || image.text("Format")=="PNG")) {
            format=image.text("Format");
        }

        image.save(&buffer, format.toLocal8Bit());
        return ba;
    }

    // MPEG Specification http://www.id3.org/id3v2.3.0

    TagMPEG::TagMPEG(TagModel* _model)
        : ExtensionAwareItem(_model)
    {
        columns[ColumnManager::CD] = new Columns::CDNumber();
        columns[ColumnManager::Rating] = new Columns::Rating();
    }

    TagMPEG::~TagMPEG ()
    {
    }

    void TagMPEG::doReadMetaTag (TagLib::FileRef& f, const QFileInfo& info)
    {
        TagItem::doReadMetaTag (f,info);

        TagLib::ID3v2::Tag* tag=dynamic_cast<TagLib::MPEG::File*>(f.file())->ID3v2Tag();
        if (tag!=0l) {
            // Read CD# if available
            TagLib::ID3v2::FrameList l = tag->frameListMap()["TPOS"];
            if (!l.isEmpty() && dynamic_cast<TagLib::ID3v2::TextIdentificationFrame*>(l.front())!=0l) {
                QString cdn=QString::fromStdWString(l.front()->toString().toWString());
                cdn=cdn.split('/').first();
                const int cdnumber = cdn.toInt();
                dynamic_cast<Columns::CDNumber*>(columns[ColumnManager::CD])->setIntValue(cdnumber,Columns::Column::RoleInitialization,this);
            }

            // Popularity tag (rating) POPM
            TagLib::ID3v2::FrameList l1 = tag->frameListMap()["POPM"];
            if (!l1.isEmpty()) {
                TagLib::ID3v2::PopularimeterFrame* pop = dynamic_cast<TagLib::ID3v2::PopularimeterFrame*>(l1.front());
                if (pop!=0l) {
                    const int rating = pop->rating();
                    int starRating=0;
                    if (rating>0) {
                        // 1 star: 53
                        // 2 stars: 104
                        // 3 stars: 154
                        // 4 stars: 205
                        // 5 stars: 255
                        if (rating<=53) starRating=1;
                        else if (rating<=104) starRating=2;
                        else if (rating<=154) starRating=3;
                        else if (rating<=205) starRating=4;
                        else starRating=5;
                    }
                    dynamic_cast<Columns::Rating*>(columns[ColumnManager::Rating])->setIntValue(starRating,Columns::Column::RoleInitialization,this);
                }
            }

        }
    }

    void TagMPEG::doSaveMetaTag (TagLib::File& f, TagLib::Tag* tag_)
    {
        TagItem::doSaveMetaTag (f,tag_);
        TagLib::ID3v2::Tag* tag=dynamic_cast<TagLib::ID3v2::Tag*>(tag_);

        // CD Number
        {
            TagLib::ID3v2::FrameList l = tag->frameListMap()["TPOS"];
            TagLib::ID3v2::TextIdentificationFrame* frame=0l;
            if (l.isEmpty()) {
                frame=new TagLib::ID3v2::TextIdentificationFrame("TPOS",TagLib::String::UTF8);
                tag->addFrame(frame);
            }
            else {
                frame=dynamic_cast<TagLib::ID3v2::TextIdentificationFrame*>(l.front());
            }
            frame->setText(data(ColumnManager::CD,Columns::Column::RoleAsVariable).toStdWString());
        }

        // Rating
        {
            TagLib::ID3v2::FrameList l = tag->frameListMap()["POPM"];
            TagLib::ID3v2::PopularimeterFrame* frame=0l;
            if (l.isEmpty()) {
                frame=new TagLib::ID3v2::PopularimeterFrame("POPM");
                tag->addFrame(frame);
            }
            else {
                frame=dynamic_cast<TagLib::ID3v2::PopularimeterFrame*>(l.front());
            }

            int rating=0;
            switch (getIntValue(ColumnManager::Rating)) {
                case 1: rating=53; break;
                case 2: rating=104; break;
                case 3: rating=154; break;
                case 4: rating=204; break;
                default: rating=255;
            }

            frame->setRating(rating);
        }
    }

    void TagMPEG::saveMetaTag () throw (TagItem::IOError)
    {
        TagLib::MPEG::File f((const char*)(getAbsolutFilename(Columns::Column::RoleAsRawVariable).toLocal8Bit()), false);
        TagLib::ID3v2::Tag* tag = f.ID3v2Tag(true);
        doSaveMetaTag (f,tag);
        if (m_Ext!=0l) {
            writeExtensions(m_Ext,tag);
        }
        if (!f.save(TagLib::MPEG::File::ID3v2,true)) {
            throw IOError (i18n("Could not update meta data."),this);
        }
    }

    QString TagMPEG::typeName () const
    {
        return "MPEG";
    }

    TagItemType::Type TagMPEG::type () const
    {
        return TagItemType::MPEG;
    }

    QString TagMPEG::suggestFilenameExtension () const
    {
        return "mp3";
    }

    QMap<int,QString> TagMPEG::s_AllowedImagesMap;
    QMap<QString,QString> TagMPEG::s_AllowedAttributesMap;
    QSet<QString> TagMPEG::s_AllowedAttributesKey;
    QSet<QString> TagMPEG::s_AllowedImages;
    TagExtension::AllowedAttributeMap TagMPEG::s_AllowedAttributes;

    QImage TagMPEG::getDefaultCoverImage() const
    {
        getExtensions();
        if (m_Ext!=0l) {
            const QImage* img = m_Ext->getImage("Front Cover");
            if (img!=0l) return *img;
            img = m_Ext->getImage("Other Image");
            if (img!=0l) return *img;
        }
        return QImage();
    }

    QString TagMPEG::getDefaultImageName () const
    {
        return "Front Cover";
    }

    TagExtension* TagMPEG::createExtension (TagLib::FileRef& f) const
    {
        using TagLib::ID3v2::AttachedPictureFrame;
        using namespace TagLib;

        if (s_AllowedImagesMap.isEmpty()) {
            s_AllowedImagesMap.insert(AttachedPictureFrame::Other,"Other Image");
            s_AllowedImagesMap.insert(AttachedPictureFrame::FileIcon,"File Icon");
            s_AllowedImagesMap.insert(AttachedPictureFrame::OtherFileIcon, "Alternate file icon");
            s_AllowedImagesMap.insert(AttachedPictureFrame::FrontCover, "Front Cover");
            s_AllowedImagesMap.insert(AttachedPictureFrame::BackCover, "Back Cover");
            s_AllowedImagesMap.insert(AttachedPictureFrame::LeafletPage, "Leaflet page");
            s_AllowedImagesMap.insert(AttachedPictureFrame::Media,"Media");
            s_AllowedImagesMap.insert(AttachedPictureFrame::LeadArtist, "Lead Artist");
            s_AllowedImagesMap.insert(AttachedPictureFrame::Artist, "Artist");
            s_AllowedImagesMap.insert(AttachedPictureFrame::Conductor, "Conductor");
            s_AllowedImagesMap.insert(AttachedPictureFrame::Band, "Band");
            s_AllowedImagesMap.insert(AttachedPictureFrame::Composer, "Composer");
            s_AllowedImagesMap.insert(AttachedPictureFrame::Lyricist, "Lyricist");
            s_AllowedImagesMap.insert(AttachedPictureFrame::RecordingLocation, "Recording location");
            s_AllowedImagesMap.insert(AttachedPictureFrame::DuringRecording, "During recording");
            s_AllowedImagesMap.insert(AttachedPictureFrame::DuringPerformance, "During Performance");
            s_AllowedImagesMap.insert(AttachedPictureFrame::MovieScreenCapture, "Movie screen capture");
            s_AllowedImagesMap.insert(AttachedPictureFrame::ColouredFish, "A bright coloured fish");
            s_AllowedImagesMap.insert(AttachedPictureFrame::Illustration, "Illustration");
            s_AllowedImagesMap.insert(AttachedPictureFrame::BandLogo, "Band logo");
            s_AllowedImagesMap.insert(AttachedPictureFrame::PublisherLogo, "Publisher logo");

            //s_AllowedAttributesMap.insert("TALB","Album title");
            s_AllowedAttributesMap.insert("TBPM","Beat per minute");
            s_AllowedAttributesMap.insert("TCOM","Composer");
            //s_AllowedAttributesMap.insert("TCON","Content type");
            s_AllowedAttributesMap.insert("TCOP","Copyright message");
            s_AllowedAttributesMap.insert("TDEN","Encoding time");
            s_AllowedAttributesMap.insert("TDLY","Playlist delay");
            s_AllowedAttributesMap.insert("TDOR","Original release time");
            //s_AllowedAttributesMap.insert("TDRC","Recording time");
            s_AllowedAttributesMap.insert("TDRL","Release time");
            s_AllowedAttributesMap.insert("TDTG","Tagging time");
            s_AllowedAttributesMap.insert("TENC","Encoded by");
            s_AllowedAttributesMap.insert("TEXT","Lyricist/Text writer");
            s_AllowedAttributesMap.insert("TFLT","File type");
            s_AllowedAttributesMap.insert("TIPL","Involved people list");
            s_AllowedAttributesMap.insert("TIT1","Content group description");
            //s_AllowedAttributesMap.insert("TIT2","Title/songname/content description");
            s_AllowedAttributesMap.insert("TIT3","Subtitle/Description refinement");
            s_AllowedAttributesMap.insert("TKEY","Initial key");
            s_AllowedAttributesMap.insert("TLAN","Language(s)");
            //s_AllowedAttributesMap.insert("TLEN","Length");
            s_AllowedAttributesMap.insert("TMCL","Musician credits list");
            s_AllowedAttributesMap.insert("TMED","Media type");
            s_AllowedAttributesMap.insert("TMOO","Mood");
            s_AllowedAttributesMap.insert("TOAL","Original album/movie/show title");
            s_AllowedAttributesMap.insert("TOFN","Original filename");
            s_AllowedAttributesMap.insert("TOLY","Original lyricist(s)/text writer(s)");
            s_AllowedAttributesMap.insert("TOPE","Original artist(s)/performer(s)");
            s_AllowedAttributesMap.insert("TOWN","File owner/licensee");
            //s_AllowedAttributesMap.insert("TPE1","Lead performer(s)/Soloist(s)");
            s_AllowedAttributesMap.insert("TPE2","Band/orchestra/accompaniment");
            s_AllowedAttributesMap.insert("TPE3","Conductor/performer refinement");
            s_AllowedAttributesMap.insert("TPE4","Interpreted, remixed, or otherwise modified by");
            //s_AllowedAttributesMap.insert("TPOS","Part of a set");
            s_AllowedAttributesMap.insert("TPRO","Produced notice");
            s_AllowedAttributesMap.insert("TPUB","Publisher");
            //s_AllowedAttributesMap.insert("TRCK","Track number/Position in set");
            s_AllowedAttributesMap.insert("TRSN","Internet radio station name");
            s_AllowedAttributesMap.insert("TRSO","Internet radio station owner");
            s_AllowedAttributesMap.insert("TSOA","Album sort order");
            s_AllowedAttributesMap.insert("TSOP","Performer sort order");
            s_AllowedAttributesMap.insert("TSOT","Title sort order");
            s_AllowedAttributesMap.insert("TSRC","ISRC (international standard recording code)");
            s_AllowedAttributesMap.insert("TSSE","Software/Hardware and settings used for encoding");
            s_AllowedAttributesMap.insert("TSST","Set subtitle");

            foreach (QString s, s_AllowedAttributesMap.values())
            {
                s_AllowedAttributes.insert (s, TagExtension::AttributeMetaData(QMetaType::QString,-1));
            }

            s_AllowedAttributesKey = s_AllowedAttributesMap.values().toSet();
            s_AllowedImages = s_AllowedImagesMap.values().toSet();
        }

        TagExtension* ext=new TagExtension (this, s_AllowedAttributes, s_AllowedAttributesKey, s_AllowedImages);

        ID3v2::Tag* tag=dynamic_cast<MPEG::File*>(f.file())->ID3v2Tag();
        if (tag!=0l) {
            foreach (QString key, s_AllowedAttributesMap.keys()) {
                ID3v2::FrameList l = tag->frameListMap()[key.toLatin1().data()];
                if (!l.isEmpty() && dynamic_cast<ID3v2::TextIdentificationFrame*>(l.front())!=0l) {
                    ext->initAttribute(s_AllowedAttributesMap[key],QString::fromStdWString(l.front()->toString().toWString()));
                }
            }

            {
                ID3v2::FrameList l = tag->frameListMap()["USLT"];
                if (!l.isEmpty() && dynamic_cast<ID3v2::UnsynchronizedLyricsFrame*>(l.front())!=0l) {
                    ID3v2::UnsynchronizedLyricsFrame* frame = dynamic_cast<ID3v2::UnsynchronizedLyricsFrame*>(l.front());
                    ext->addLyrics(QString::fromStdWString(frame->text().toWString()));
                }
            }

            {
                ID3v2::FrameList l = tag->frameListMap()["APIC"];
                for (ID3v2::FrameList::Iterator iter= l.begin();iter!=l.end();++iter) {
                    if (dynamic_cast<ID3v2::AttachedPictureFrame*>(*iter)!=0l) {
                        ID3v2::AttachedPictureFrame* frame = dynamic_cast<ID3v2::AttachedPictureFrame*>(*iter);
                        const int type = frame->type();
                        ByteVector data = frame->picture();
                        QImage img = QImage::fromData ((const unsigned char*)data.data(), (unsigned int)data.size(), 0l);
                        const QString desc = QString::fromStdWString(frame->description().toWString());
                        img.setText("",desc);
                        if (!img.isNull() && s_AllowedImagesMap.contains(type)) {
                            ext->initImage(s_AllowedImagesMap[type],img);
                        }
                    }
                }
            }
        }

        return ext;
    }

    void TagMPEG::writeExtensions (TagExtension* ext, TagLib::ID3v2::Tag* tag)
    {
        if (tag!=0l) {
            //
            // Images
            //

            // Delete images (remove also images to update)
            TagLib::ID3v2::FrameList l = tag->frameListMap()["APIC"];
            QSet<QString> toDelete=ext->imagesToDelete()+ext->imagesToUpdate();
            for (TagLib::ID3v2::FrameList::Iterator iter= l.begin();iter!=l.end();++iter) {
                if (dynamic_cast<TagLib::ID3v2::AttachedPictureFrame*>(*iter)!=0l) {
                    TagLib::ID3v2::AttachedPictureFrame* frame = dynamic_cast<TagLib::ID3v2::AttachedPictureFrame*>(*iter);
                    const int type = frame->type();
                    if (toDelete.contains(s_AllowedImagesMap[type])) {
                        tag->removeFrame(frame);
                    }
                }
            }

            // insert images (also the update only images)
            foreach (QString imgName, ext->imagesToAdd()+ext->imagesToUpdate()) {
                TagLib::ID3v2::AttachedPictureFrame* imgFrame = new TagLib::ID3v2::AttachedPictureFrame();
                const QImage* image = ext->getImage(imgName);
                QByteArray ba=convertImage(*image);

                imgFrame->setPicture(TagLib::ByteVector(ba.data(),ba.length()));
                if (image->text()!="") {
                    imgFrame->setDescription(image->text("Desc").toStdWString());
                }
                imgFrame->setType((TagLib::ID3v2::AttachedPictureFrame::Type)s_AllowedImagesMap.key(imgName));

                tag->addFrame(imgFrame);
            }

            //
            // Attributs
            //
            const TagLib::ID3v2::FrameListMap& frames = tag->frameListMap();
            foreach (QString value, ext->attribsToDelete()+ext->attribsToUpdate()) {
                if (value!="") {
                    TagLib::ID3v2::FrameList toDelete=frames[(const char*)s_AllowedAttributesMap.key(value).toAscii()];
                    for (TagLib::ID3v2::FrameList::Iterator iter= toDelete.begin();iter!=toDelete.end();++iter) {
                        tag->removeFrame(*iter);
                    }
                }
            }

            foreach (QString value, ext->attribsToAdd()+ext->attribsToUpdate()) {
                if (value!="") {
                    const QString revValue = s_AllowedAttributesMap.key(value);
                    if (value!="") {
                        TagLib::ID3v2::TextIdentificationFrame* newFrame = new TagLib::ID3v2::TextIdentificationFrame((const char*)revValue.toAscii());
                        newFrame->setText(ext->attrib(value).toStdWString());
                        tag->addFrame(newFrame);
                    }
                }
            }

            ext->resetChangeState();
        }
    }

    void TagMPEG::writeExtensions (TagExtension* ext)
    {
        TagLib::MPEG::File f((const char*)getAbsolutFilename().toLocal8Bit(),true,TagLib::AudioProperties::Fast);
        writeExtensions(ext,f.ID3v2Tag());
        f.save(TagLib::MPEG::File::ID3v2,true);
    }

    bool TagMPEG::supportsCoverImages() const
    {
        return true;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    //
    // MP4 Implementation
    //
    // List of MP4 attributes (taken from http://atomicparsley.sourceforge.net/mpeg-4files.html)
    //
    //    4char code      Name            Class/Flag                 Appearance     Comment
    //    ©alb            Album           1           text           iTunes 4.0     default
    //    ©art            Artist          1           text           iTunes 4.0     default
    //    aART            Album Artist    1           text           ??
    //    ©cmt            Comment         1           text           iTunes 4.0     default
    //    day             Year            1           text           iTunes 4.0     default
    //    nam             Title           1           text           iTunes 4.0     default
    //    gen | gnre      Genre           1 | 0 1     text | uint8   iTunes 4.0     default
    //    trkn            Track number    0           uint8          iTunes 4.0     default
    //    disk            Disk number     0           uint8          iTunes 4.0     default2
    //    wrt             Composer        1           text           iTunes 4.0     extended
    //    too             Encoder         1           text           iTunes 4.0     extended
    //    tmpo            BPM             21          uint8          iTunes 4.0     extended
    //    cprt            Copyright       1           text?          iTunes 4.0     extended
    //    cpil            Compilation     21          uint8          iTunes 4.0     extended
    //    covr            Artwork         13 | 14 2   jpeg | png     iTunes 4.0     extended
    //    rtng            Rating/Advisory 21          uint8          iTunes 4.0
    //    grp             Grouping        1           text           iTunes 4.2     extended
    //    stik            ?? (stik)       21          uint8          ??
    //    pcst            Podcast         21          uint8          iTunes 4.9
    //    catg            Category        1           text           iTunes 4.9     extended
    //    keyw            Keyword         1           text           iTunes 4.9     extended
    //    purl            Podcast URL     21 | 0 4    uint8          iTunes 4.9
    //    egid            Episode GUID    21 | 0 4    uint8          iTunes 4.9
    //    desc            Description     1           text           iTunes 5.0     extended
    //    lyr             Lyrics          1 3         text           iTunes 5.0
    //    tvnn            TV Network Name 1           text           iTunes 6.0     extended
    //    tvsh            TV Show Name    1           text           iTunes 6.0     extended
    //    tven            TV Episode No   1           text           iTunes 6.0     extended
    //    tvsn            TV Season       21          uint8          iTunes 6.0     extended
    //    tves            TV Episode      21          uint8          iTunes 6.0     extended
    //    purd            Purchase Date   1           text           iTunes 6.0.2   extended
    //    pgap            Gapless Playback 21         uint8          iTunes 7.0     extended
    //
    ///////////////////////////////////////////////////////////////////////////////////////////////

#if (TAGLIB_MAJOR_VERSION>1) || ((TAGLIB_MAJOR_VERSION==1) && (TAGLIB_MINOR_VERSION>5))

    QSet<QString> TagMP4::s_AllowedImages;
    TagExtension::AllowedAttributeMap TagMP4::s_AllowedAttributes;
    QSet<QString> TagMP4::s_AllowedAttributesKey;
    QMap<QString,QString> TagMP4::s_AllowedAttributesMap;

    TagExtension* TagMP4::createExtension (TagLib::FileRef& f) const
    {
        if (s_AllowedImages.isEmpty()) {
            s_AllowedImages << "Front Cover";

            QMap<QString,QString> temp;

            temp.insert("\251wrt","Composer");
            temp.insert("\251too","Encoder");
            temp.insert("cprt","Copyright");
            temp.insert("\251grp","Grouping");
            temp.insert("catg","Category");
            temp.insert("keyw","Keyword");
            temp.insert("desc","Description");
            temp.insert("tvnn","TV Network Name");
            temp.insert("tvsh","TV Show Name");
            temp.insert("tven","TV Episode Number");
            temp.insert("purd","Purchase Date");
            s_AllowedAttributesMap.unite(temp);

            foreach (QString s, temp.values())
            {
                s_AllowedAttributes.insert (s,TagExtension::AttributeMetaData(QMetaType::QString,255));
            }

            s_AllowedAttributesMap.insert ("tmpo","BPM");
            s_AllowedAttributes.insert ("BPM",TagExtension::AttributeMetaData(QMetaType::Int,255));
            s_AllowedAttributesMap.insert ("pgap","Gapless Playback");
            s_AllowedAttributes.insert ("Gapless Playback",TagExtension::AttributeMetaData(QMetaType::Int,1));
            s_AllowedAttributesMap.insert ("tvsn","TV Season");
            s_AllowedAttributes.insert ("TV Season",TagExtension::AttributeMetaData(QMetaType::Int,255));
            s_AllowedAttributesMap.insert ("tves","TV Episode");
            s_AllowedAttributes.insert ("TV Episode",TagExtension::AttributeMetaData(QMetaType::Int,255));
            s_AllowedAttributesMap.insert ("cpil","Compilation");
            s_AllowedAttributes.insert ("Compilation",TagExtension::AttributeMetaData(QMetaType::Int,255));

            s_AllowedAttributesKey = s_AllowedAttributesMap.values().toSet();
        }

        TagExtension* ext=new TagExtension (this, s_AllowedAttributes,s_AllowedAttributesKey,s_AllowedImages);

        TagLib::MP4::Tag* tag=dynamic_cast<TagLib::MP4::Tag*>(f.file()->tag());
        if (tag!=0l) {
            TagLib::MP4::ItemListMap items = tag->itemListMap();

            // Load extended attributes
            foreach (const QString key, s_AllowedAttributesMap.keys()) {
                TagLib::String _k = asTagLibString(key);
                if (items.contains(_k)) {
                    ext->initAttribute(s_AllowedAttributesMap[key], asQString(items[_k].toStringList().toString(",")));
                }
            }

            // Image support was broken before 1.6.1
#if (TAGLIB_PATCH_VERSION>0)
            // Load Cover images
            if (items.contains("covr")) {
                if (items["covr"].toCoverArtList().size()>0) {
                    TagLib::ByteVector data = items["covr"].toCoverArtList().front().data();
                    QByteArray byteArray(data.data(),data.size());
                    QImage image = QImage::fromData(byteArray);
                    if (!image.isNull()) {
                        ext->initImage ("Front Cover",image);
                    }
                }
            }
#endif
        }

        return ext;
    }


#if (TAGLIB_PATCH_VERSION>0)
    QImage TagMP4::getDefaultCoverImage () const
    {
        getExtensions();
        if (m_Ext!=0l) {
            return *m_Ext->getImage("Front Cover");
        }
        return QImage();
    }

    bool TagMP4::supportsCoverImages() const
    {
        return true;
    }

    QString TagMPEG::getDefaultImageName () const
    {
        return "Front Cover";
    }

#endif

    TagMP4::TagMP4(TagModel* _model)
        : ExtensionAwareItem(_model)
    {
        columns[ColumnManager::CD] = new Columns::CDNumber();
        columns[ColumnManager::Rating] = new Columns::Rating();
    }

    TagMP4::~TagMP4()
    {
    }

    void TagMP4::doReadMetaTag (TagLib::FileRef& f, const QFileInfo& info)
    {
        using namespace Columns;
        using namespace TagLib;

        TagItem::doReadMetaTag (f,info);
        MP4::Tag* tag = dynamic_cast<MP4::Tag*>(f.tag());
        if (tag!=0l) {
            MP4::ItemListMap& items = tag->itemListMap();
            if (items.contains("disk")) {
                dynamic_cast<CDNumber*>(columns[ColumnManager::CD])->setIntValue(items["disk"].toInt(),
                                                                                 Column::RoleInitialization,
                                                                                 this);
            }
            if (items.contains("rtng")) {
                dynamic_cast<Rating*>(columns[ColumnManager::Rating])->setIntValue(items["rtng"].toInt(),
                                                                                   Column::RoleInitialization,
                                                                                   this);
            }
        }
    }

    void TagMP4::doSaveMetaTag (TagLib::File& f, TagLib::Tag* tag_)
    {
        using namespace Columns;
        using namespace TagLib;

        TagItem::doSaveMetaTag (f,tag_);
        MP4::Tag* tag = dynamic_cast<MP4::Tag*>(tag_);
        if (tag!=0l) {
            MP4::ItemListMap& items = tag->itemListMap();
            items["disk"]=MP4::Item(dynamic_cast<CDNumber*>(columns[ColumnManager::CD])->getIntValue(Column::RoleDisplay),0);
            const int rating = dynamic_cast<Rating*>(columns[ColumnManager::Rating])->getIntValue(Column::RoleDisplay);
            items["rtng"]=MP4::Item(rating,0);
        }
    }

    void TagMP4::saveMetaTag () throw (TagItem::IOError)
    {
        TagLib::MP4::File f((const char*)(getAbsolutFilename(Columns::Column::RoleAsRawVariable).toLocal8Bit()), false);
        if (!f.isValid()) {
            throw TagItem::IOError (i18n("Could not open file %1 to update tag information",getAbsolutFilename()), this);
        }

        TagLib::MP4::Tag* tag = f.tag();

        if (m_Ext!=0l) {
            writeExtensions(m_Ext,tag);
        }
        doSaveMetaTag (f,tag);

        if (!f.save()) {
            throw IOError (i18n("Could not update meta data."),this);
        }
    }

    QString TagMP4::typeName () const
    {
        return "MP4";
    }

    TagItemType::Type TagMP4::type () const
    {
        return TagItemType::MP4;
    }

    QString TagMP4::suggestFilenameExtension () const
    {
        return "acc";
    }

    void TagMP4::writeExtensions (TagExtension* ext, TagLib::MP4::Tag* tag)
    {
        if (tag!=0l) {
            TagLib::MP4::ItemListMap& items = tag->itemListMap();

            // Care for extended attributes
            foreach (QString attrib, ext->attribsToDelete()) {
                items.erase(asTagLibString(attrib));
            }
            foreach (QString attrib, ext->attribsToUpdate()+ext->attribsToAdd()) {
                TagLib::StringList list;
                list.append(asTagLibString(ext->attrib(attrib)));
                TagLib::String _k = asTagLibString(s_AllowedAttributesMap.key(attrib));
                items.insert(_k,TagLib::MP4::Item(list));
            }

            // Image support was broken before 1.6.1
#if (TAGLIB_PATCH_VERSION>0)
            // Care for images
            if (!ext->imagesToDelete().isEmpty()) {
                items.erase("covr");
            }
            if (!ext->imagesToAdd().isEmpty() || !ext->imagesToUpdate().isEmpty()) {
                TagLib::MP4::CoverArtList covers;
                const QImage* img = m_Ext->getImage("Front Cover");
                QByteArray ba=convertImage(*img);
                TagLib::MP4::CoverArt::Format f=(img->text("Format")=="JPG")?TagLib::MP4::CoverArt::JPEG:TagLib::MP4::CoverArt::PNG;
                TagLib::ByteVector bv(ba.data(),ba.size());
                covers.append (TagLib::MP4::CoverArt(f,bv));
                items.insert("covr",TagLib::MP4::Item(covers));
            }
#endif
            ext->resetChangeState();
        }
    }

    void TagMP4::writeExtensions (TagExtension* ext)
    {
        TagLib::MP4::File f((const char*)getAbsolutFilename().toLocal8Bit(),true,TagLib::AudioProperties::Fast);
        writeExtensions(ext,f.tag());
        f.save();
    }

#endif

    ///////////////////////////////////////////////////////////////////////////////////////////////
    //
    // Ogg::Speex Implementation
    //
    ///////////////////////////////////////////////////////////////////////////////////////////////

    TagOggSpeex::TagOggSpeex(TagModel* _model)
        : TagItem(_model)
    {
    }

    TagOggSpeex::~TagOggSpeex()
    {
    }

    void TagOggSpeex::saveMetaTag () throw (TagItem::IOError)
    {
        TagLib::Ogg::Speex::File f((const char*)(getAbsolutFilename(Columns::Column::RoleAsRawVariable).toLocal8Bit()), false);
        TagLib::Ogg::XiphComment* tag = f.tag();
        if (tag==0l) {
            throw IOError (i18n("Can't create tag information of file type %1. Ignoring update.",typeName()),this);
        }
        doSaveMetaTag (f,tag);
        if (!f.save()) {
            throw IOError (i18n("Could not update meta data."),this);
        }
    }

    QString TagOggSpeex::typeName () const
    {
        return "Ogg::Speex";
    }

    TagItemType::Type TagOggSpeex::type () const
    {
        return TagItemType::Ogg_Speex;
    }

    QString TagOggSpeex::suggestFilenameExtension () const
    {
        return "ogg";
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    //
    // Ogg::FLAC Implementation
    //
    ///////////////////////////////////////////////////////////////////////////////////////////////

    TagOggFLAC::TagOggFLAC(TagModel* _model)
        : TagItem(_model)
    {
    }

    TagOggFLAC::~TagOggFLAC()
    {
    }

    void TagOggFLAC::saveMetaTag () throw (TagItem::IOError)
    {
        TagLib::Ogg::FLAC::File f((const char*)(getAbsolutFilename(Columns::Column::RoleAsRawVariable).toLocal8Bit()), false);
        TagLib::Ogg::XiphComment* tag = f.tag();
        if (tag==0l) {
            throw IOError (i18n("Can't create tag information of file type %1. Ignoring update.",typeName()),this);
        }
        doSaveMetaTag (f,tag);
        if (!f.save()) {
            throw IOError (i18n("Could not update meta data."),this);
        }
    }

    QString TagOggFLAC::typeName () const
    {
        return "Ogg::FLAC";
    }

    TagItemType::Type TagOggFLAC::type () const
    {
        return TagItemType::Ogg_Flag;
    }

    QString TagOggFLAC::suggestFilenameExtension () const
    {
        return "ogg";
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    //
    // OggVorbis Implementation
    //
    ///////////////////////////////////////////////////////////////////////////////////////////////

    TagOggVorbis::TagOggVorbis(TagModel* _model)
        : TagItem(_model)
    {
    }

    TagOggVorbis::~TagOggVorbis()
    {
    }

    void TagOggVorbis::saveMetaTag () throw (TagItem::IOError)
    {
        TagLib::Ogg::Vorbis::File f((const char*)(getAbsolutFilename(Columns::Column::RoleAsRawVariable).toLocal8Bit()), false);
        TagLib::Ogg::XiphComment* tag = f.tag();
        if (tag==0l) {
            throw IOError (i18n("Can't create tag information of file type %1. Ignoring update.",typeName()),this);
        }
        doSaveMetaTag (f,tag);
        if (!f.save()) {
            throw IOError (i18n("Could not update meta data."),this);
        }
    }

    QString TagOggVorbis::typeName () const
    {
        return "Ogg::Vorbis";
    }

    TagItemType::Type TagOggVorbis::type () const
    {
        return TagItemType::Ogg_Vorbis;
    }

    QString TagOggVorbis::suggestFilenameExtension () const
    {
        return "ogg";
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    //
    // FLAC Implementation
    //
    ///////////////////////////////////////////////////////////////////////////////////////////////

    TagFLAC::TagFLAC(TagModel* _model)
        : TagItem(_model)
    {
    }

    TagFLAC::~TagFLAC()
    {
    }

    void TagFLAC::saveMetaTag () throw (TagItem::IOError)
    {
        TagLib::FLAC::File f((const char*)(getAbsolutFilename(Columns::Column::RoleAsRawVariable).toLocal8Bit()), false);
        TagLib::Tag* tag = f.tag();
        if (tag==0l) {
            throw IOError (i18n("Can't create tag information of file type %1. Ignoring update.",typeName()),this);
        }
        doSaveMetaTag (f,tag);
        if (!f.save()) {
            throw IOError (i18n("Could not update meta data."),this);
        }
    }

    QString TagFLAC::typeName () const
    {
        return "FLAC";
    }

    TagItemType::Type TagFLAC::type () const
    {
        return TagItemType::Flac;
    }

    QString TagFLAC::suggestFilenameExtension () const
    {
        return "flac";
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    //
    // MPC Implementation
    //
    ///////////////////////////////////////////////////////////////////////////////////////////////

    TagMPC::TagMPC(TagModel* _model)
        : TagItem(_model)
    {
    }

    TagMPC::~TagMPC()
    {
    }

    void TagMPC::saveMetaTag () throw (TagItem::IOError)
    {
        TagLib::MPC::File f((const char*)(getAbsolutFilename(Columns::Column::RoleAsRawVariable).toLocal8Bit()), false);
        TagLib::Tag* tag = f.tag();
        if (tag==0l) {
            throw IOError (i18n("Can't create tag information of file type %1. Ignoring update.",typeName()),this);
        }
        doSaveMetaTag (f,tag);
        if (!f.save()) {
            throw IOError (i18n("Could not update meta data."),this);
        }
    }

    QString TagMPC::typeName () const
    {
        return "MPC";
    }

    TagItemType::Type TagMPC::type () const
    {
        return TagItemType::MPC;
    }

    QString TagMPC::suggestFilenameExtension () const
    {
        return "mpc";
    }


    ///////////////////////////////////////////////////////////////////////////////////////////////
    //
    // ASF Implementation
    //
    ///////////////////////////////////////////////////////////////////////////////////////////////

#if (TAGLIB_MAJOR_VERSION==1) && (TAGLIB_MINOR_VERSION>5)

    TagASF::TagASF(TagModel* _model)
        : TagItem(_model)
    {
    }

    TagASF::~TagASF()
    {
    }

    void TagASF::saveMetaTag () throw (TagItem::IOError)
    {
        TagLib::ASF::File f((const char*)(getAbsolutFilename(Columns::Column::RoleAsRawVariable).toLocal8Bit()), false);
        TagLib::ASF::Tag* tag = f.tag();
        if (tag==0l) {
            throw IOError (i18n("Can't create tag information of file type %1. Ignoring update.",typeName()),this);
        }
        doSaveMetaTag (f,tag);
        if (!f.save()) {
            throw IOError (i18n("Could not update meta data."),this);
        }
    }

    QString TagASF::typeName () const
    {
        return "ASF";
    }

    TagItemType::Type TagASF::type () const
    {
        return TagItemType::ASF;
    }

    QString TagASF::suggestFilenameExtension () const
    {
        return "asf";
    }

#endif

    ///////////////////////////////////////////////////////////////////////////////////////////////
    //
    // WavPack Implementation
    //
    ///////////////////////////////////////////////////////////////////////////////////////////////

    TagWavPack::TagWavPack(TagModel* _model)
        : TagItem(_model)
    {
    }

    TagWavPack::~TagWavPack()
    {
    }

    void TagWavPack::saveMetaTag () throw (TagItem::IOError)
    {
        TagLib::WavPack::File f((const char*)(getAbsolutFilename(Columns::Column::RoleAsRawVariable).toLocal8Bit()), false);
        TagLib::Tag* tag = f.tag();
        if (tag==0l) {
            throw IOError (i18n("Can't create tag information of file type %1. Ignoring update.",typeName()),this);
        }
        doSaveMetaTag (f,tag);
        if (!f.save()) {
            throw IOError (i18n("Could not update meta data."),this);
        }
    }

    QString TagWavPack::typeName () const
    {
        return "WavPack";
    }

    TagItemType::Type TagWavPack::type () const
    {
        return TagItemType::WavPack;
    }

    QString TagWavPack::suggestFilenameExtension () const
    {
        return "wv";
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    //
    // TrueAudio Implementation
    //
    ///////////////////////////////////////////////////////////////////////////////////////////////

    // TODO: Allow same support as MPEG files
    TagTrueAudio::TagTrueAudio(TagModel* _model)
        : TagItem(_model)
    {
    }

    TagTrueAudio::~TagTrueAudio()
    {
    }

    void TagTrueAudio::saveMetaTag () throw (TagItem::IOError)
    {
        TagLib::TrueAudio::File f((const char*)(getAbsolutFilename(Columns::Column::RoleAsRawVariable).toLocal8Bit()), false);
        TagLib::ID3v2::Tag* tag = f.ID3v2Tag(true);
        if (tag==0l) {
            throw IOError (i18n("Can't create tag information of file type %1. Ignoring update.",typeName()),this);
        }
        doSaveMetaTag (f,tag);
        if (!f.save()) {
            throw IOError (i18n("Could not update meta data."),this);
        }
    }

    QString TagTrueAudio::typeName () const
    {
        return "TrueAudio";
    }

    TagItemType::Type TagTrueAudio::type () const
    {
        return TagItemType::TrueAudio;
    }

    QString TagTrueAudio::suggestFilenameExtension () const
    {
        return "tta";
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    //
    // RIFF::AIFF Implementation
    //
    ///////////////////////////////////////////////////////////////////////////////////////////////

#if (TAGLIB_MAJOR_VERSION==1) && (TAGLIB_MINOR_VERSION>5)

    TagRIFFAIFF::TagRIFFAIFF(TagModel* _model)
        : TagItem(_model)
    {
    }

    TagRIFFAIFF::~TagRIFFAIFF()
    {
    }

    void TagRIFFAIFF::saveMetaTag () throw (TagItem::IOError)
    {
        TagLib::RIFF::AIFF::File f((const char*)(getAbsolutFilename(Columns::Column::RoleAsRawVariable).toLocal8Bit()), false);
        TagLib::Tag* tag = f.tag();
        if (tag==0l) {
            throw IOError (i18n("Can't create tag information of file type %1. Ignoring update.",typeName()),this);
        }
        doSaveMetaTag (f,tag);
        if (!f.save()) {
            throw IOError (i18n("Could not update meta data."),this);
        }
    }

    QString TagRIFFAIFF::typeName () const
    {
        return "RIFF::AIFF";
    }

    TagItemType::Type TagRIFFAIFF::type () const
    {
        return TagItemType::RIFF_AIFF;
    }

    QString TagRIFFAIFF::suggestFilenameExtension () const
    {
        return "riff";
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    //
    // RIFF::WAV Implementation
    //
    ///////////////////////////////////////////////////////////////////////////////////////////////

    TagRIFFWAV::TagRIFFWAV(TagModel* _model)
        : TagItem(_model)
    {
    }

    TagRIFFWAV::~TagRIFFWAV()
    {
    }

    void TagRIFFWAV::saveMetaTag () throw (TagItem::IOError)
    {
        TagLib::RIFF::WAV::File f((const char*)(getAbsolutFilename(Columns::Column::RoleAsRawVariable).toLocal8Bit()), false);
        TagLib::Tag* tag = f.tag();
        if (tag==0l) {
            throw IOError (i18n("Can't create tag information of file type %1. Ignoring update.",typeName()),this);
        }
        doSaveMetaTag (f,tag);
        if (!f.save()) {
            throw IOError (i18n("Could not update meta data."),this);
        }
    }

    QString TagRIFFWAV::typeName () const
    {
        return "RIFF::WAV";
    }

    TagItemType::Type TagRIFFWAV::type () const
    {
        return TagItemType::RIFF_WAV;
    }

    QString TagRIFFWAV::suggestFilenameExtension () const
    {
        return "riff";
    }

}

#endif
