/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2011 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "texeltagmodel.h"
#include "columnmanager.h"
#include "tagitem.h"
#include "model/problem.h"
#include "options/preferences.h"
#include "func/undocommands.h"
#include "options/enums.h"
#include "util/duplicatemap.h"
#include "util/chunkify.h"
#include "util/loghook.h"
#include "func/m3u/playlistmodel.h"
#include "view/tagcomparer/tagcomparer.h"
#include "model/jobs/modelloader.h"
#include "model/jobs/updatetagsjob.h"
#include "model/structure/artistmanager.h"
#include "model/mimesupport.h"
#include "view/tagcomparer/decisiondialog.h"
#include "func/metatagservices/freedb/cdentry.h"

#include <iostream>

#include <QFile>
#include <QDir>
#include <QPair>
#include <QMultiMap>
#include <QStringList>
#include <QUrl>
#include <QtAlgorithms>
#include <QMessageBox>
#include <QMultiMap>
#include <QTextDocument>
#include <QTextCursor>
#include <QApplication>
#include <QTextList>
#include <QTextBlock>
#include <QTextListFormat>
#include <QTextTable>
#include <QReadLocker>
#include <QReadWriteLock>
#include <QWriteLocker>

#include <kstringhandler.h>
#include <kundostack.h>
#include <kdebug.h>
#include <klocale.h>

using namespace Texel;
using namespace std;

Q_DECLARE_METATYPE ( QTextDocument* )

#define THREAD_CHECK p->check_thread (__PRETTY_FUNCTION__);

#if 0
#define READ_LOCK QReadLocker __locker(&p->m_Lock);
#else
#define READ_LOCK QWriteLocker __locker(&p->m_Lock);
#endif

#define WRITE_LOCK QWriteLocker __locker(&p->m_Lock);


struct NotifyFileDupicate
{
    void operator() (TagItem* item, bool duplicate, int count)
    {
        if (duplicate) {
            QString msg=i18n("Filename %1 appears %2x in the list. This will lead to clashes");
            msg=msg.arg(item->getAbsolutFilename(Columns::Column::RoleDisplay),QString::number(count));
            item->setProblem(ColumnManager::Filename, new UnknownProblem(msg));
        }
        else {
            item->setProblem(ColumnManager::Filename, 0l);
        }
    }
};

class FileList
{
public:
   FileList ()
   {
   }

   ~FileList ()
   {
       foreach (TagItem* item, OrderedList) delete item;
       clear();
   }

   int size () const
   {
      return OrderedList.size();
   }

   void clear ()
   {
       OrderedList.clear();
       filenameDuplicateDetector.clear();
   }

   void setRow (int row, TagItem* item)
   {
       OrderedList[row]=item;
       OrderedList[row]->setRow(row);
   }

   void insert (int start, const TagItem::List& rows)
   {
       // Resize list
       for (int i=0;i<rows.size();i++) OrderedList.append(0l);

       // Make available space of size 'rows.size()' at position start
       for (int i=OrderedList.size()-rows.size()-1;i>=start;--i) {
           setRow(i+rows.size(),OrderedList[i]);
       }

       // copy rows into the free space
       for (int i=0;i<rows.size();i++) {
           setRow(start+i,rows[i]);
       }

       // add filenames
       foreach (TagItem* item, rows) {
           const DirTree::Name id(item->getFilenameId(Columns::Column::RoleAsVariable));
           filenameDuplicateDetector.alter (item,id,NotifyFileDupicate());
       }
   }

   bool append (TagItem* item)
   {
       bool app=false;
       if (item!=0l) {
           const int row=item->getRow();
           if (row==-1 || row>=OrderedList.size()) {
               OrderedList.append (item);
               const int row = OrderedList.size()-1;
               if (item->getRow()!=-1 && item->getRow()!=row) {
                   kWarning() << "Appended a tag (row " << row
                              << ") which would have caused a gap in sequence. Desired was row "
                              << item->getRow() << endl;
               }
               item->setRow(row);
               app=true;
           }
           else {
               OrderedList[item->getRow()]->takeover(item);
               item=OrderedList[item->getRow()];
           }
           const DirTree::Name id(item->getFilenameId(Columns::Column::RoleAsVariable));
           filenameDuplicateDetector.alter(item, id, NotifyFileDupicate());
       }
       return app;
   }

   void alteredPersistentFilename (TagItem* item, const DirTree::Name& oldPersistantName)
   {
       filenameDuplicateDetector.alter (item, oldPersistantName, NotifyFileDupicate());
    }

   int append (const TagItem::List& list)
   {
       int size=0;
       foreach (TagItem* item, list) {
           if (append (item)) ++size;
       }
       return size;
   }

   TagItem* operator[] (int i)
   {
       if (i<0 || i>=OrderedList.size()) {
           kError() << "Invalid index " << i << endl;
           return 0L;
       }
       return OrderedList[i];
   }

   void remove (int start, int count)
   {
      if (start<0 || start+count>OrderedList.size()) return;

      // Remove deletion entries from index
      for (int i=0;i<count;i++) {
         TagItem* item=OrderedList[i+start];
         filenameDuplicateDetector.remove(item, NotifyFileDupicate());
         // no deletion of the items, this is done in the redo command!
      }

      // copy manually the entries in the list and adjust the row entries
      const int end=OrderedList.size();
      for (int i=start;i<end-count;i++) {
          setRow(i,OrderedList[i+count]);
      }

      // remove unneeded entries at the end of the list
      for (int i=0;i<count;i++) OrderedList.removeLast();
   }

   const TagItem::List& getOrderedList() const
   {
       return OrderedList;
   }

   TagItem::List lookupFilename (const DirTree::Name& filename) const
   {
       return filenameDuplicateDetector.getAll(filename);
   }

private:
   TagItem::List OrderedList;
   DuplicateMap::Detector<DirTree::Name, TagItem> filenameDuplicateDetector;
};

// ----------------------------------------------------------------------------
// model class
// ----------------------------------------------------------------------------

struct TagModel::Private
{
    Private (QObject* parent)
     : m_Owner(QThread::currentThread()),
       m_Playlists(new PlaylistModel(dynamic_cast<TagModel*>(parent),parent)),
       m_Lock(QReadWriteLock::Recursive)
    {
    }

    ~Private ()
    {
        m_Playlists->deleteAll();
        delete m_Playlists;
        delete m_UndoStack;
        // Be aware: m_DirCache must be destructed after filelist!
    }

    QThread* m_Owner;

    void check_thread (const char* func)
    {
        if (QThread::currentThread()!=m_Owner) {
            std::cerr << "Calling " << func << " from thread " << QThread::currentThread()->objectName() << std::endl;
        }
    }

   DirTree m_DirCache;
   FileList filelist;
   QString m_ListFilename;
   
   //
   TagItem::Set m_ChangedSinceLastSaveOfList;
   TagItem::Set m_ChangedSinceLastUpdateOfTags;

   int m_ChangedSinceLastFullWrite;
   bool m_AlreadyWritten;
   bool m_UnsavedDuplicateList;

   /// To ensure that in case of deleted rows the next save is a complete one and
   /// not only a partial one
   QSet<quint32> m_RemovedItems;

   int duplicateCounter;
   KUndoStack* m_UndoStack;

   MessageModel* m_MessageModel;

   PlaylistModel* m_Playlists;

   QReadWriteLock m_Lock;
};

TagModel::TagModel (QObject* parent)
: super(parent),
  p(new TagModel::Private(this))
{
   p->m_ChangedSinceLastFullWrite=0;
   p->m_AlreadyWritten=false;
   p->duplicateCounter=0;
   p->m_UndoStack=new KUndoStack(this);
   p->m_UndoStack->setUndoLimit(100);
   p->m_MessageModel=0l;

   static bool _init = false;
   if (!_init) {
       // Its not registered by default. If not registered the signal changedHeader seems to be forced
       // for execution in the current thread instead of the main thread
       qRegisterMetaType<Qt::Orientation>("Qt::Orientation");
       _init=true;
   }
}

TagModel::~TagModel ()
{
    delete p;
}

bool TagModel::load (const QList<QUrl>& dirs)
{
    Texel::ModelLoader* loader=new Texel::ModelLoader(this);

    bool result=loader->addFileNames(dirs);

    connect (loader, SIGNAL(loadedItems(const TagItem::List)), this, SLOT(appendList(const TagItem::List)));
    connect (loader, SIGNAL(loadPlaylist(QString)), this, SLOT(loadPlaylist(QString)));
    connect (loader, SIGNAL(loadAdditionalFile(QString,DirTree::AdditionalFileFormats)),
               this, SLOT(addAdditionalInfoFile(QString,DirTree::AdditionalFileFormats)));

    emit newBackgroundJob(loader);

    loader->start();

    setWrittenToList(false);
    return result;
}

void TagModel::addAdditionalInfoFile (const QString& file,DirTree::AdditionalFileFormats format)
{
    QList<TagItem*> items;
    p->m_Lock.lockForRead();
    DirTree::Dir* dir=p->m_DirCache.addAdditionalInformationFile(file, format);
    if (dir!=0l) {
        items=p->m_DirCache.getTagItems(dir->getName(),false);
    }
    p->m_Lock.unlock();
    if (dir!=0l) {
        foreach (TagItem* item, items) {
            emit headerDataChanged (Qt::Vertical,item->getRow(),item->getRow());
        }
    }
}

void TagModel::setListFilename (const QString& name)
{
    p->m_ListFilename=name;
}

const QString& TagModel::getListFilename () const
{    
   return p->m_ListFilename;
}

QString TagModel::getFileName (int row) const
{
   if (row<0 || p->filelist.size()<row) return QString();
   READ_LOCK;
   return p->filelist[row]->getAbsolutFilename();
}

int TagModel::rowCount ( const QModelIndex& parent) const
{
    READ_LOCK;
    if (parent.isValid()) {
        if (parent.column()!=0 || !p->filelist[parent.row()]->hasRelations()) return 0;
        return p->filelist[parent.row()]->getRelations()->sizeOfDuplicates();
    }
    return p->filelist.size();
}

int TagModel::columnCount (const QModelIndex& ) const
{
   return ColumnManager::instance()->getColumnNumber();
}

QList<quint32> TagModel::getRemovedIds () const
{
    return p->m_RemovedItems.toList();
}

void TagModel::removeRowsRedo (int first, int last)
{
    bool emitDirty=false;
    bool emitEmpty=false;

    {
        WRITE_LOCK;

        beginRemoveRows(QModelIndex(), first, last);
        for (int i=first;i<=last;i++) {
            p->m_ChangedSinceLastSaveOfList.remove(p->filelist[i]);
            p->m_ChangedSinceLastUpdateOfTags.remove(p->filelist[i]);
            emitDirty=p->m_ChangedSinceLastUpdateOfTags.isEmpty();
            p->m_RemovedItems.insert(p->filelist[i]->getUniqId());
        }
        p->filelist.remove(first,last-first+1);
        endRemoveRows();

        emitEmpty=(p->filelist.size()==0);
    }

    if (emitDirty) emit isDirty(false);
    if (emitEmpty) emit isEmpty(true);
}

void TagModel::insertRowsRedo (int start, const TagItem::List& rows)
{
    WRITE_LOCK;

    const bool wasEmpty = p->filelist.size()==0;

    beginInsertRows (QModelIndex(),start,start+rows.size()-1);
    p->filelist.insert(start,rows);
    foreach (TagItem* item, rows) {
         if (item->isAltered()) {
             p->m_ChangedSinceLastSaveOfList.insert(item);
             p->m_ChangedSinceLastUpdateOfTags.insert(item);
             if (p->m_ChangedSinceLastUpdateOfTags.size()==1) emit isDirty(true);
         }
         p->m_RemovedItems.remove(item->getUniqId());
    }
    endInsertRows();

    if (wasEmpty && p->filelist.size()>0) emit isEmpty(false);
}

void TagModel::clearRedo ()
{
    {
        WRITE_LOCK;
        beginRemoveRows(QModelIndex(), 0, p->filelist.size()-1);
        p->m_ChangedSinceLastUpdateOfTags.clear();
        p->m_ChangedSinceLastSaveOfList.clear();
        p->filelist.clear();
        endRemoveRows();
    }
    emit isEmpty(true);
    emit isDirty(false);
}

bool TagModel::removeRows (int row, int count, const QModelIndex&)
{
    if (count>0) {
        READ_LOCK;

        TagItem::List rows;
        for (int i=0;i<count;i++) {
            if (p->filelist[i+row]!=0l) {
                rows << p->filelist[i+row];
            }
        }

        if (row==0 && count==rowCount()) {
            p->m_UndoStack->push(new Clear(rows));
        }
        else {
            p->m_UndoStack->push(new RemoveRows(rows));
        }
    }

    return true;
}

void TagModel::clear ()
{
    int size=0;
    bool unsaved=false;

    {
        READ_LOCK;
        size=p->m_ChangedSinceLastUpdateOfTags.size();
        unsaved = !p->m_AlreadyWritten && size>0;
    }

    if (unsaved) {
        QMessageBox box(QMessageBox::Warning,
                        i18nc("Headline","Clear"),
                        i18n("Document has %1 unsaved changed tags",size),
                        QMessageBox::Save|QMessageBox::Cancel|QMessageBox::Ignore);
        const int result=box.exec();
        switch (result) {
        case QMessageBox::Save:
            updateChanged(false);
            break;
        case QMessageBox::Cancel:
            return;
        }
    }

   WRITE_LOCK;
   removeRows(0,rowCount());
   p->m_Playlists->deleteAll();
}

QVariant TagModel::displayData (const QModelIndex& index) const
{
    READ_LOCK;
    if (!index.isValid() || index.row()>p->filelist.size()) return QVariant();
    return p->filelist[index.row()]->data(index.column(),Columns::Column::RoleDisplay);
}

QVariant TagModel::tooltipData (const QModelIndex& index) const
{
   QString result;
   if (index.row()>p->filelist.size()) return result;

   READ_LOCK;
   if (index.column()==ColumnManager::Path) {
       result+="["+p->m_DirCache.getCommonPath()+"]"+p->filelist[index.row()]->data(index.column(),Columns::Column::RoleDisplay);
   }

   if (index.column()==ColumnManager::Filename) {
       result+=p->filelist[index.row()]->getAbsolutFilename(Columns::Column::RoleDisplay)+"\n";
   }

   if (p->filelist[index.row()]->isAltered() &&
       p->filelist[index.row()]->data(index.column(),Columns::Column::RoleAsVariable)!=
       p->filelist[index.row()]->data(index.column(),Columns::Column::RoleAsRawVariable)) {
      result+="Modified from: " + p->filelist[index.row()]->data(index.column(),Columns::Column::RoleDisplay);
   }

   return QVariant(result);
}

QModelIndex TagModel::index (int row, int column, const QModelIndex& parent) const
{
    READ_LOCK;
    if (parent.isValid()) {
        if (parent.column()!=0) return QModelIndex();
        return createIndex (row, column, p->filelist[parent.row()]);
    }
    TagItem* tag=0l;
    if (row>=0 && row<p->filelist.size()) tag=p->filelist[row];
    return createIndex(row,column,tag);
}

QVariant TagModel::data (const QModelIndex& index, int role) const
{
    READ_LOCK;

    if (index.row()>p->filelist.size()) return QVariant();
    const TagItem* tagitem = static_cast<TagItem*>(index.internalPointer());
    if (tagitem==0l) return QVariant();

    switch (role) {
    case Qt::CheckStateRole:
        if (index.column()==ColumnManager::MarkForDeletion) {
            return QVariant(tagitem->isToDelete()?Qt::Checked:Qt::Unchecked);
        }
        return QVariant();
    case Qt::DisplayRole:
        return displayData(index);
    case Qt::ToolTipRole:
        return tooltipData(index);
    case Qt::BackgroundRole:
        if (tagitem->isAltered(index.column())) return Preferences::instance()->appearance.m_alteredBackground;
        return Preferences::instance()->appearance.m_normalBackground;
    case Qt::ForegroundRole:
        if (tagitem->isToDelete()) {
            return Preferences::instance()->appearance.m_markedForDeletionForeground;
        }
        return Preferences::instance()->appearance.m_normalForeground;
    case Qt::FontRole:
        return Preferences::instance()->appearance.m_normalFont;
    case RawModeRole:
        return tagitem->data(index.column(),Columns::Column::RolePersistence);
    case ExtendedDisplayMode:
        return tagitem->data(index.column(),Columns::Column::RolePersistence);
    case RoleDisplayFormularFirst:
        return tagitem->data(index.column(),Columns::Column::RoleDisplayFormularFirst);
    case Enums::RoleProblem:
        if (tagitem->hasInputProblem(index.column())) return QVariant(Enums::ProblemError);
      if (index.column()==ColumnManager::Artist  && p->filelist[index.row()]->isChecked()) {
            if (tagitem->isCorrect()) return QVariant(Enums::ProblemNone);
            if (tagitem->hasPropsals()) return QVariant(Enums::ProblemWarning);
            return QVariant(3);
      }
      return QVariant(Enums::ProblemUnknown);
   case RoleFormular:
      return tagitem->data(index.column(),Columns::Column::RoleFormular);
   }

   return QVariant();
}

bool TagModel::setDataInternally (int row, int column, const QVariant& value, Columns::Column::Role role)
{
    WRITE_LOCK;

    TagItem* item=getRow(row);
    const bool rowWasChanged = item->isAltered();
    bool changed;
    if (column==ColumnManager::MarkForDeletion) {
        changed = item->setToDelete(value.toBool());
    }
    else {
        changed = item->setDataDirect(column,value.toString(),role);
    }

    if (changed) {
        TagItem::Set items;
        items << item;
        if (item->isAltered()) {
            setAsDirty(items);
        }
        else {
            setAsClean(items);
        }
        if (!rowWasChanged) emit changeDirtyState(row,row);
    }

    return changed;
}

bool TagModel::setDataRedo (const QModelIndex& index, const QVariant& value, Columns::Column::Role role)
{
    const bool changed=setDataInternally (index.row(),index.column(),value,role);

    if (changed) {
        emit dataChanged (index,index);
        emit headerDataChanged (Qt::Vertical,index.row(),index.row());
    }

    return changed;
}

bool TagModel::setDataRangeRedo (const QList<SlimIndex>& indexes, const QVariant& value, Columns::Column::Role role)
{
    for (int i=0;i<indexes.size();i++) {
        setDataInternally(indexes[i].row(), indexes[i].column(), value, role);
        if (i>0 && indexes[i-1].row()!=indexes[i].row()) {
            emit headerDataChanged (Qt::Vertical,indexes[i-1].row(),indexes[i-1].row());
        }
        const QModelIndex index=indexes[i].qmodelindex(this);
        emit dataChanged (index,index);
    }

    return true;
}

bool TagModel::setDataListRedo (const QList<UpdateList::Entry>& entries, UpdateList::Entry::ValueTypes type, Columns::Column::Role role)
{
    int lastRow=-1;
    foreach (UpdateList::Entry entry, entries)
    {
        setDataInternally(entry.m_Index.row(), entry.m_Index.column(), entry.value(type), role);
        if (lastRow!=entry.m_Index.row() && lastRow!=-1) {
            emit headerDataChanged (Qt::Vertical,lastRow,lastRow);
        }
        lastRow=entry.m_Index.row();
        const QModelIndex index=entry.m_Index.qmodelindex(this);
        emit dataChanged (index,index);
    }
    if (lastRow!=-1) {
        emit headerDataChanged (Qt::Vertical,lastRow,lastRow);
    }

    return true;
}

bool TagModel::setData (const QModelIndex& index, const QVariant& value, int role)
{
    if (!index.isValid() ||
        index.row()<0 || index.row()>=p->filelist.size() ||
        value==data(index,role) ||
        index.column()<0 || index.column()>=ColumnManager::instance()->getColumnNumber() ||
        index==QModelIndex()) {
        return false;
    }

    if (role==Qt::CheckStateRole) {
        p->m_UndoStack->push(new ToggleBool(index,value.toBool()));
    }
    else if (role==TagModel::CorrectnessCheckRole) {
         const int state = value.toInt();
         {
             WRITE_LOCK;
             p->filelist[index.row()]->setIsCorrect((state&ArtistManager::IsCorrect)!=0);
             p->filelist[index.row()]->setHasProposals((state&ArtistManager::HasProposals)!=0);
             p->filelist[index.row()]->setIsChecked(state!=0);
         }
         emit dataChanged(index,index);
    }
    else {
        READ_LOCK;
        const QString oldValue = p->filelist[index.row()]->data(index.column(),Columns::Column::RolePersistence);
        p->m_UndoStack->push(new UpdateList(this, SlimIndex(index),oldValue,value,Columns::Column::RoleDisplay));
    }

    return true;
}

Qt::ItemFlags TagModel::flags (const QModelIndex& index) const
{
    READ_LOCK;
    if (index.isValid()) {
        Qt::ItemFlags flags=0;
        TagItem* currentTag = getRow(index.row());
        if (currentTag->hasColumn(index.column())) {
            flags|=Qt::ItemFlags(Qt::ItemIsSelectable|Qt::ItemIsDropEnabled|Qt::ItemIsEnabled|Qt::ItemIsDragEnabled);
            if (ColumnManager::instance()->isEditable(index.column()) && !currentTag->isLockedForUpdate()) {
                flags|=Qt::ItemIsEditable;
            }
            if (index.column()==ColumnManager::MarkForDeletion) {
                flags|=Qt::ItemIsUserCheckable;
            }
        }
        return flags;
    }
    return Qt::ItemIsDropEnabled;
}

QVariant TagModel::headerData (int section, Qt::Orientation orientation, int role) const
{
    if (section<0) return QVariant();

   if (orientation==Qt::Horizontal) {
      if (role==Qt::DisplayRole) {
         return ColumnManager::instance()->getName(section);
      }
      else if (role==Qt::ToolTipRole) {
          return ColumnManager::instance()->getDescription (section);
      }
   }
   if (orientation==Qt::Vertical) {
       if (role==ChangedRole) {
           return QVariant(section<p->filelist.size() && p->filelist[section]->isAltered());
       } else if (role==Enums::RoleIcons) {
           QList<QVariant> icons;
           READ_LOCK;

           TagItem* currentTag = getRow(section);
           // Playlists
           if (currentTag->hasRelations() &&
               currentTag->getRelations()->numberOfPlaylists()>0) {
               icons << QIcon(KIconLoader::global()->loadIcon("playlist-marker",KIconLoader::Small));
           }
           // Duplicate detection
           if (currentTag->hasRelations() &&
               currentTag->getRelations()->sizeOfDuplicates()>0) {
               icons << QIcon(KIconLoader::global()->loadIcon("detect-duplicates",KIconLoader::Small));
           }
           // Messages
           if (currentTag->getErrors()>0) {
               icons << QIcon(MessageModel::getIcon(Enums::MessageError));
           }
           else if (currentTag->getWarnings()>0) {
               icons << QIcon(MessageModel::getIcon(Enums::MessageWarning));
           }
           // Currently in save?
           if (currentTag->isLockedForUpdate()) {
               icons << KIconLoader::global()->loadIcon("object-locked",KIconLoader::Toolbar);
           }
           // Additional metadata
           if (currentTag->getPathPointer(Columns::Column::RoleDisplay)->hasAdditionalInformationFiles()) {
               icons << KIcon(KIconLoader::global()->loadIcon("metadata-available",KIconLoader::Small));
           }

           return icons;
       } else if (role==Qt::ToolTipRole) {
           READ_LOCK;

           QTextFrame* previousFrame=0l;

           QTextDocument* doc=new QTextDocument;
           QTextCursor cursor(doc);
           QTextCharFormat format;
           format.setFont(QApplication::font());
           format.setFontPointSize(8.0);

           QTextCharFormat headerFormat=format;
           headerFormat.setFontPointSize(10.0);
           headerFormat.setFontItalic(true);
           headerFormat.setFontUnderline(true);

           QTextListFormat listFormat;
           listFormat.setIndent(1);
           listFormat.setStyle(QTextListFormat::ListSquare);

           QTextFrameFormat frameFormat = doc->rootFrame()->frameFormat();
           frameFormat.setMargin(0.0);
           QTextFrameFormat lastFrameFormat=frameFormat;
           frameFormat.setBottomMargin(5.0);

           QTextFrame* mainFrame = cursor.currentFrame();

           if (p->m_MessageModel!=0l &&
               (getRow(section)->getWarnings()>0 || getRow(section)->getErrors()>0)) {

               cursor=mainFrame->lastCursorPosition();
               cursor.insertText(i18n("Messages for file:"),headerFormat);
               cursor.insertFrame(lastFrameFormat);
               previousFrame=cursor.currentFrame();

               cursor.setCharFormat(format);
               foreach (MessageModel::Item* item, p->m_MessageModel->getMessagesForItem(getRow(section))) {
                   cursor.insertImage(p->m_MessageModel->data(item,0,Qt::DecorationRole).value<QPixmap>().toImage());
                   cursor.insertText(QString("  ")+p->m_MessageModel->data(item,0,Qt::DisplayRole).toString()+"\n");
               }
           }

           if (p->filelist[section]->hasRelations() &&
               p->filelist[section]->getRelations()->sizeOfDuplicates()>0) {

               if (previousFrame!=0l) {
                  previousFrame->setFrameFormat(frameFormat);
               }

               cursor=mainFrame->lastCursorPosition();
               cursor.insertText(i18n("Potential duplicates"),headerFormat);
               cursor.insertFrame(lastFrameFormat);
               previousFrame=cursor.currentFrame();
               QTextCursor localCursor (cursor.currentFrame());

               localCursor.setCharFormat(format);
               localCursor.insertList(listFormat);
               int n=0;
               foreach (TagItem* dup, p->filelist[section]->getRelations()->getDuplicates()) {
                   if (n>0) localCursor.insertBlock();
                   localCursor.insertText(dup->getAbsolutFilename());
                   if (++n>Preferences::instance()->m_DD_maxDuplicateDisplay) break;
               }
           }

           if (p->filelist[section]->hasRelations() &&
               p->filelist[section]->getRelations()->numberOfPlaylists()>0) {

               if (previousFrame!=0l) {
                  previousFrame->setFrameFormat(frameFormat);
               }

               cursor=mainFrame->lastCursorPosition();
               cursor.insertText(i18n("Playlists"),headerFormat);

               cursor.insertFrame(lastFrameFormat);
               cursor.setCharFormat(format);
               cursor.insertList(listFormat);
               int n=0;
               foreach (M3U::List list, p->filelist[section]->getRelations()->getPlaylists()) {
                   if (n>0) cursor.insertBlock();
                   cursor.insertText(list.getFileName());
                   if (++n>Preferences::instance()->m_DD_maxDuplicateDisplay) break;
               }

               previousFrame=cursor.currentFrame();
           }

           if (getRow(section)->getPathPointer(Columns::Column::RoleDisplay)->hasAdditionalInformationFiles())
           {
               if (previousFrame!=0l) {
                  previousFrame->setFrameFormat(frameFormat);
               }

               cursor=mainFrame->lastCursorPosition();
               cursor.insertText(i18n("Additional metadata available"),headerFormat);
               cursor.insertFrame(lastFrameFormat);
               cursor.setCharFormat(format);
               const int pos=cursor.position();
               QTextTable* t=cursor.insertTable(getRow(section)->getPathPointer(Columns::Column::RoleDisplay)->getMetaData().size(),2);
               QTextTableFormat tf = t->format();
               tf.setBorderStyle(QTextFrameFormat::BorderStyle_None);
               t->setFormat(tf);

               int row=0;
               foreach (DirTree::Dir::MetaData m, getRow(section)->getPathPointer(Columns::Column::RoleDisplay)->getMetaData())
               {
                   cursor.setPosition(pos);
                   if (m.first==DirTree::CDDB) {
                       t->mergeCells(row,0,1,2);
                       cursor.insertText(i18n("CDDB information in file \"%1\"",m.second));
                   }
                   else if (m.first==DirTree::Image) {
                       QImage img;
                       img.load(getRow(section)->getPathPointer(Columns::Column::RoleDisplay)->getName()+"/"+m.second);
                       QTextBlockFormat format=cursor.blockFormat();
                       format.setAlignment(Qt::Alignment(Qt::AlignVCenter|Qt::AlignLeft));
                       QTextCursor cellCursor=t->cellAt(row,0).firstCursorPosition();
                       cellCursor.insertBlock(format);
                       cellCursor.insertText(i18n("Image from file \"%1\"\n(%2 x %3)",m.second,img.width(),img.height()));

                       t->cellAt(row,1).firstCursorPosition().insertImage(img.scaled(64,64,Qt::KeepAspectRatio));
                   }
                   ++row;
               }
           }

           if (doc->isEmpty()) return QVariant();
           return QVariant::fromValue(doc);
       }
   }

   return QVariant(); //super::headerData (section, orientation, role);
} 

void TagModel::removeRows (const TagItem::List& toRemove)
{
    READ_LOCK;
    if (toRemove.size()>0) {
        QList<int> rows;
        foreach (TagItem* item, toRemove) {
            rows << item->getRow();
        }
        Chunkify::ChunkList chunks = Chunkify::chunkify(rows);

        for (int i=chunks.size();i>0;i--) {
            removeRows(chunks[i-1].first,chunks[i-1].second);
        }
    }
}

void TagModel::removeRowDirect (TagItem* item)
{
    WRITE_LOCK;
    removeRowsRedo (item->getRow(),item->getRow());
}

void TagModel::removeRow (TagItem* item)
{
    WRITE_LOCK;
    removeRows(item->getRow(),item->getRow());
}

void TagModel::setAsDirty (const TagItem::Set& items)
{
    bool emitDirty=false;
    int unsaved=0;

    {
        WRITE_LOCK;
        const int s = p->m_ChangedSinceLastUpdateOfTags.size();
        p->m_ChangedSinceLastUpdateOfTags+=items;
        p->m_ChangedSinceLastSaveOfList+=items;
        if (s==0 && !items.isEmpty()) emitDirty=true;
        unsaved=p->m_ChangedSinceLastUpdateOfTags.size()-s;
    }

    if (emitDirty) emit isDirty(true);
    if (unsaved!=0) emit setUnsaved(unsaved);
}

void TagModel::setAsClean (const TagItem::Set& items)
{
    bool emitDirty=false;
    int  emitUnsaved=0;

    {
        WRITE_LOCK;

        const int s = p->m_ChangedSinceLastUpdateOfTags.size();
        p->m_ChangedSinceLastUpdateOfTags-=items;
        p->m_ChangedSinceLastSaveOfList-=items;
        emitDirty = (s!=0 && p->m_ChangedSinceLastUpdateOfTags.empty());
        emitUnsaved=(p->m_ChangedSinceLastUpdateOfTags.size()-s);
    }

    if (emitDirty) emit isDirty(false);
    if (emitUnsaved!=0) emit setUnsaved(emitUnsaved);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Small class to define action on accept the rename dialog.

namespace {
    class OnAcceptRenames: public TagFunctor
    {
        TagModel* m_Model;

    public:
        OnAcceptRenames (TagModel* model)
            : m_Model(model)
        {
        }

        virtual ~OnAcceptRenames ()
        {
        }

        virtual void accept (TagItem::Set, TagItem::Set unchecked)
        {
            m_Model->updateWhenNecessary(unchecked,true);
        }

        virtual void reject (TagItem::Set rejected)
        {
            m_Model->setAsDirty(rejected);
        }

    };
}

void TagModel::updateWhenNecessary (const TagItem::Set& items, bool enforceRename)
{
    READ_LOCK;
    TagItem::Set toUpdate=items&p->m_ChangedSinceLastUpdateOfTags;
    UpdateTagsJob* job = createUpdateJob(toUpdate, enforceRename);
    connect (job,SIGNAL(result(KJob*)),this,SLOT(onUpdateChangedFinished (KJob*)));
    job->start();
}

void TagModel::updateChanged (bool removeUpdatedFromList)
{
    READ_LOCK;
    UpdateTagsJob* job = createUpdateJob(p->m_ChangedSinceLastUpdateOfTags, false);
    connect (job,SIGNAL(result(KJob*)),this,SLOT(onUpdateChangedFinished (KJob*)));
    if (removeUpdatedFromList) {
        connect (job,SIGNAL(tagUpdated(TagItem*)),this,SLOT(removeRow(TagItem*)));
    }
    job->start();
}

void TagModel::onUpdateChangedFinished (KJob* job)
{
    UpdateTagsJob* updateJob = dynamic_cast<UpdateTagsJob*>(job);
    if (updateJob!=0) {
        postProcessUpdateChanged (updateJob);
        WRITE_LOCK;
        p->m_Playlists->updateChanged();
    }
}

void TagModel::postProcessUpdateChanged (UpdateTagsJob*)
{
}

void TagModel::onUpdatedTag (TagItem* item)
{
    item->setLockedForUpdate(false);
    const int idx = item->getRow();

    LogHook::instance()->message (i18n("Updated row"), item);

    emit headerDataChanged (Qt::Vertical,idx,idx);
    emit dataChanged (createIndex(idx,0),
                      createIndex(idx,ColumnManager::instance()->getColumnNumber()));

    TagItem::Set items;
    items << item;
    setAsClean(items);
}

UpdateTagsJob* TagModel::createUpdateJob (TagItem::Set toUpdate, bool forceRename)
{
    UpdateTagsJob* job = new UpdateTagsJob(this);
    job->setEnforceRename(forceRename);
    job->setItemsForUpdate(toUpdate);
    connect (job, SIGNAL(updatedTag(TagItem*)), this, SLOT(onUpdatedTag(TagItem*)), Qt::QueuedConnection);
    connect (job, SIGNAL(removedTag(TagItem*)), this, SLOT(removeRowDirect(TagItem*)), Qt::QueuedConnection);
    connect (job, SIGNAL(changeUnsavedItems(int)), this, SIGNAL(setUnsaved(int)));

    DescisionDialog* remover = new DescisionDialog(this,0l);
    connect (job, SIGNAL(finished(KJob*)), remover, SLOT(endDirectorySearchJob()));
    connect (job, SIGNAL(directoryToDelete(QString)), remover, SLOT(emptyDirectoryToDelete(QString)));
    connect (job, SIGNAL(renameCollision(TagItem*,QString)), remover, SLOT(addCollision(TagItem*,QString)));

    emit newBackgroundJob(job);
    return job;
}

void TagModel::fillDown (const QVariant& ref, const QList<SlimIndex>& indexes)
{
    bool filled=false;

    if (indexes[0].column()==ColumnManager::Track) {
        bool isNumber;
        int track = ref.toInt(&isNumber);
        if (isNumber) {
            QList<UpdateList::Entry> entries;
            foreach (SlimIndex idx, indexes) {
                entries << UpdateList::Entry(idx,data(idx.qmodelindex(this),RawModeRole),++track);
            }

            p->m_UndoStack->push (new UpdateList(this, entries,
                                                 Columns::Column::RoleDisplay,
                                                 i18nc("redo comment","Track enumaration")
                                                 ));
            filled=true;
        }
    }

    if (!filled) {
        setDataBlock (ref,indexes);
    }
}

void TagModel::setDataBlock (const QVariant& ref, const QList<SlimIndex>& indexes)
{
    QList<UpdateList::Entry> entries;
    foreach (SlimIndex idx, indexes) {
        entries << UpdateList::Entry(idx,data(idx.qmodelindex(this),RawModeRole),ref);
    }

    p->m_UndoStack->push (new UpdateList ( this,
                                           entries,
                                           Columns::Column::RoleDisplay,
                                           i18nc ("redo comment","Fill block")
                                           ));
}

void TagModel::paste (const QList<SlimIndex>& indexes, const QVariantList& values,
                      const QString& description)
{
    QList<UpdateList::Entry> entries;
    for (int i=0;i<indexes.size();i++) {
        entries << UpdateList::Entry(indexes[i],
                                     data(indexes[i].qmodelindex(this),RawModeRole),
                                     values[i]);
    }

    p->m_UndoStack->push(new UpdateList(this, entries,
                                         Columns::Column::RoleDisplay,
                                         description));
}

void TagModel::appendList (const TagItem::List list)
{
    foreach (TagItem* item, list) {
        item->setIsCreated();
    }

    int size;
    int toInsert;
    {
        WRITE_LOCK;
        size = p->filelist.size();
        toInsert=p->filelist.append(list);

        beginInsertRows (QModelIndex(),size, size+toInsert-1);
        endInsertRows ();
    }
    if (size==0 && toInsert>0) emit isEmpty(false);
}

void TagModel::addDuplicateRelation (int row, const QList<int>& dups)
{
    WRITE_LOCK;
    TagItem* tag = getRow(row);
    if (tag!=0l) {
        foreach (int d, dups) {
            TagItem* dup = getRow(d);
            if (dup!=0l) {
                tag->getRelations()->addDuplicate(dup);
            }
        }
    }
}

void TagModel::updateSaveListStatistics (const QString& filename, bool saved, int delta)
{
    WRITE_LOCK;
    p->m_ListFilename=filename;
    if (saved==true) {
        if (delta==-1) {
            p->m_ChangedSinceLastFullWrite=0;
            p->m_RemovedItems.clear();
        }
        else {
            p->m_ChangedSinceLastFullWrite+=p->m_ChangedSinceLastSaveOfList.size();
        }
        p->m_AlreadyWritten=true;
        p->m_UnsavedDuplicateList=false;
    }
    else {
        p->m_ChangedSinceLastFullWrite=delta;
        p->m_RemovedItems.clear();
    }
    p->m_ChangedSinceLastSaveOfList.clear();
}

bool TagModel::recommendCompleteSave () const
{
    READ_LOCK;
    return !p->m_AlreadyWritten;
}

TagItem* TagModel::getRow (int i) const
{
    READ_LOCK;
    return p->filelist[i];
}

DirTree* TagModel::getDirCache ()
{
   return &p->m_DirCache;
}

void TagModel::changeArtistName (const QString& oldName, const QString& newName)
{
    WRITE_LOCK;
    MasterManager::instance()->getArtistManager()->changeName(oldName,newName);
}

void TagModel::alteredPersistentFilename (TagItem* item, const DirTree::Name& oldPersistantName)
{
    WRITE_LOCK;
    p->filelist.alteredPersistentFilename(item, oldPersistantName);
}

bool TagModel::hasDuplicates() const
{
    READ_LOCK;
    return p->duplicateCounter>0;
}

void TagModel::incDuplicateCounter(int n)
{
    WRITE_LOCK;
    if (p->duplicateCounter==0 && n>0) emit hasDuplicates(true);
    p->duplicateCounter+=n;
    if (p->duplicateCounter<0) p->duplicateCounter=0;
    if (p->duplicateCounter==0) emit hasDuplicates(false);
}

int TagModel::numberOfDuplicates () const
{
    READ_LOCK;
    return p->duplicateCounter;
}

TagItem::List TagModel::getTagItems() const
{
    READ_LOCK;
    return p->filelist.getOrderedList();
}

TagItem::List TagModel::getTagItemsToSave () const
{
    READ_LOCK;
    return p->m_ChangedSinceLastSaveOfList.toList();
}

void TagModel::emitChangedCell (const QModelIndex& index)
{
    emit dataChanged(index,index);
}

void TagModel::emitChangedRow (int row, bool saved)
{
    READ_LOCK;
    TagItem::Set items;
    items << getRow(row);
    if (saved) {
        setAsClean(items);
    }
    else {
        setAsDirty (items);
    }

    emitChangedHeader(row);
}

void TagModel::emitChangedHeader (int row)
{
    emit headerDataChanged (Qt::Vertical,row,row);
}

KUndoStack* TagModel::undoStack() const
{
    return p->m_UndoStack;
}

void TagModel::emitCurrentState()
{
    bool empty=false;
    bool dirty=false;

    {
        READ_LOCK;
        empty=(rowCount()==0);
        dirty=(p->m_ChangedSinceLastUpdateOfTags.size()>0);
    }

    emit isEmpty(empty);
    emit isDirty(dirty);
}

void TagModel::setMessageModel (MessageModel* messages)
{
    WRITE_LOCK;
    p->m_MessageModel=messages;
}

bool TagModel::isDirty () const
{
    READ_LOCK;
    return !p->m_ChangedSinceLastUpdateOfTags.isEmpty();
}

bool TagModel::hasUnsavedDuplicateList () const
{
    READ_LOCK;
    return p->m_UnsavedDuplicateList;
}

void TagModel::setNewDuplicateList ()
{
    WRITE_LOCK;
    p->m_UnsavedDuplicateList=true;
}

void TagModel::addPlaylist (M3U::List playlist)
{
    WRITE_LOCK;
    p->m_Playlists->addPlaylist(playlist);
}

QAbstractItemModel* TagModel::getPlayListModel () const
{
    READ_LOCK;
    return p->m_Playlists;
}

TagItem* TagModel::getRowByOrgFilename (const QString& name) const
{
    READ_LOCK;

    const DirTree::Name key = p->m_DirCache.getFileName(name);
    const QDir d(name);

    foreach (TagItem* t, p->filelist.lookupFilename(key))
    {
        // Check which is the original file and removes it
        const QDir f(t->getAbsolutFilename(Columns::Column::RoleAsRawVariable));
        if (f==d) {
            return t;
        }
    }

    return 0l;
}

void TagModel::loadPlaylist (const QString& playlistName)
{
    M3U::List playlist(this);
    playlist.load(playlistName);
    addPlaylist(playlist);
}

void TagModel::scanner (DirTree::Dir* start, unsigned& processed)
{
    const QString dirName = start->getName();

    TagItem::List toRemove;
    QFileInfo info;

    info.setFile(dirName);
    if (info.exists()) {
        // Check Subdirectories
        foreach (DirTree::Dir* i,start->getChildren()) {
            scanner (i,processed);
        }
        foreach (TagItem* tag, p->m_DirCache.getTagItems(dirName,false)) {
            info.setFile(tag->getAbsolutFilename());
            if (!info.exists()) {
                toRemove << tag;
            }
            ++processed;
        }
    }
    else {
        toRemove=p->m_DirCache.getTagItems(dirName,true);
        processed+=toRemove.size();
    }

    if (!toRemove.isEmpty()) {
        removeRows (toRemove);
    }

    emit numberAlreadyProcessed(processed);
}

void TagModel::removeStaleEntries ()
{
    WRITE_LOCK;
    p->m_DirCache.lock();
    DirTree::Dir* root= p->m_DirCache.getRoot();
    unsigned processed=0;
    scanner(root,processed);
    p->m_DirCache.unlock();
}

void TagModel::scannerForCDDBCompletion (DirTree::Dir* start, unsigned& processed)
{
    foreach (DirTree::Dir* i,start->getChildren()) {
        scannerForCDDBCompletion (i,processed);
    }

    foreach (DirTree::Dir::MetaData md, start->getMetaData()) {
        if (md.first==DirTree::CDDB) {
            // Found CDDB enries. Now check if unfilled elements exist
            bool unfilled=false;
            TagItem::List items = p->m_DirCache.getTagItems(start->getName(),false);
            foreach (TagItem* tag, items) {
                unfilled|=tag->data(ColumnManager::Artist).isEmpty()
                          |tag->data(ColumnManager::Album).isEmpty()
                          |tag->data(ColumnManager::Title).isEmpty();
            }
            if (unfilled) {
                QFile cddbEntry (md.second);
                cddbEntry.open(QIODevice::ReadOnly);
                CDEntry* entry = CDEntry::parse(&cddbEntry);
                cddbEntry.close();
                entry->mapContent(items,true);
                delete entry;
            }
            break;
        }
    }
}

void TagModel::setWrittenToList (bool state)
{
    WRITE_LOCK;
    p->m_AlreadyWritten=state;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Drag and drop support
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

QMimeData* TagModel::mimeData (const QModelIndexList& indexes) const
{
    MimeSupport* mime = new MimeSupport();
    mime->setItems (indexes, this);
    return mime;
}

QStringList TagModel::mimeTypes () const
{
    QStringList l;
    l << MimeSupport::s_TagItemSubFormat;
    return l;
}


#include "texeltagmodel.moc"

