/*************************************************************************
*************************************************************************/

#ifndef texel_TexelTagModel_H
#define texel_TexelTagModel_H

#include <QAbstractTableModel>
#include <QItemSelectionModel>
#include <QSet>
#include <QtConcurrentMap>

#include "model/tagitem.h"
#include "model/columnsbase.h"
#include "util/slimindex.h"
#include "func/undocommands.h"
#include "view/messageview.h"

#include "aid.h"

class QXmlStreamReader;
class QCompleter;
class KUndoStack;
class KJob;

namespace Texel
{

    class ModelLoader;
    class UpdateTagsJob;

    /**
  @brief
  This is the main model of Texel. Its a primary data layer which provides data for all cells.
  It does also maintain several indexes for fast access.

  As the model is quite complex as many suporting functions as possible are transfered to other
  classes. TagItem containes the logic from the meta data of a single audio file. The loading
  and storing facilities of an XML based list is done in TagModelXmlStreamer. Loading the audio
  file from disc and parsing the meta data is done in ModelLoader. Metadata of the different
  columns is maintained in ColumnManager. As this is a meta data class Its based on a singleton.

  The management of the data altering commands (for undo/redo capabilities) is done in the file
  @ref undocommands.h and @ref undocommands.cpp
  */
    class TagModel : public QAbstractTableModel
    {
        Q_OBJECT;

        typedef QAbstractTableModel super;

    public:
        enum StatisticsSaveMode { CompleteSave, PartialSave, Loaded };

        enum CustomDisplayRoles { RawModeRole=Qt::UserRole,
                                  DuplicatesRole,
                                  ChangedRole,
                                  ExtendedDisplayMode,
                                  CorrectnessCheckRole,
                                  RoleFormular,
                                  // This role is used to have a normal Display but when a formular
                                  // is used display that instead of the computed value
                                  RoleDisplayFormularFirst
                              };

        TagItem::List getTagItems () const;
        TagItem::List getTagItemsToSave () const;
        QList<quint32> getRemovedIds () const;

        TagModel (QObject* parent);
        virtual ~TagModel ();

        virtual int rowCount (const QModelIndex & parent = QModelIndex()) const;
        virtual int columnCount (const QModelIndex & parent = QModelIndex()) const;
        virtual QModelIndex index ( int row, int column, const QModelIndex & parent = QModelIndex()) const;

        virtual QVariant data (const QModelIndex& index, int role = Qt::DisplayRole) const;
        virtual QVariant headerData (int section, Qt::Orientation orientation, int role = Qt::DisplayRole) const;
        virtual bool setData (const QModelIndex& index, const QVariant& value, int role = Qt::EditRole);
        QString getFileName (int row) const;

        virtual bool removeRows ( int row, int count, const QModelIndex& parent=QModelIndex());

        void paste (const QList<SlimIndex>& indexes, const QVariantList& values, const QString& description);

        KUndoStack* undoStack() const;

        /** Do the downfiling undo save for the given column in the range from top to bottom. Please note that
        just a single undo command is created to undo the complete downfill.
        @param column Column to do the down filling for
        @param top Row from which the value is taken for downfilling. Cells below until bottom are replace
        @param cbottom last row to replace. This is including!
        */
        void fillDown (const QVariant& ref, const QList<SlimIndex>& indexes);

        /// The model contains a filename for a list to save to. When this name is set a subsequent
        /// call can easiliy be made to the same file. The filename can be retrieved with getListFilename()
        /// @param name Filename which should be stored in the model. The name is simply considered
        /// as a string and not further parsed.
        void setListFilename (const QString& name);

        /// @returns the list filename stored in the model or an empty string when no filename was set.
        /// Filenames are set with the function setListFilename (const QString& name).
        const QString& getListFilename () const;

        void setAsDirty (const TagItem::Set& items);
        void setAsClean (const TagItem::Set& items);

        /** The model maintains a list of changed items since the last save and the last complete
     sabve. When it seems cheaper to save trhe complete model freash this function recommends
     to do so.

     @returns true when the model believes the saving thecomplete model again is cheaper
     than just consider the changed tags.
     */
        bool recommendCompleteSave () const;

        void setWrittenToList (bool);

        TagItem* getRow (int row) const;
        // Returns all rows which a currently marked with the given filename.
        // This might be more than one as several files may point to one filename (which
        // will lead to collisions upon save)
        TagItem* getRowByOrgFilename (const QString& name) const;

        DirTree* getDirCache ();
        /*bool load (const QUrl& dir);*/
        bool load (const QList<QUrl>& dirs);

        void alteredSuggestedFilename (TagItem* item, const DirTree::Name* oldSuggestedFilename);
        void alteredPersistentFilename (TagItem* item, const DirTree::Name& oldPersistantName);

        void insert (int row, TagItem* item);

        void emitAlteredExtension (int row, int type, const QString& name, const QVariant& value);
        void setMessageModel (MessageModel* messages);

        bool isDirty () const;

        bool hasUnsavedDuplicateList () const;
        bool hasDuplicates () const;
        void setNewDuplicateList ();
        int numberOfDuplicates () const;

        void setDataBlock (const QVariant& ref, const QList<SlimIndex>& indexes);

        void emitChangedCell (const QModelIndex& index);
        void emitChangedRow (int row, bool saved);
        void emitChangedHeader (int row);
        void emitRemoveRow (TagItem* row);

        void incDuplicateCounter (int n);

        QAbstractItemModel* getPlayListModel () const;

        void removeRows (const TagItem::List& rows);

        /// Updates the items in the list when necessary (when they are dirty).
        void updateWhenNecessary (const TagItem::Set& items, bool enforceRename);

        QMimeData* mimeData (const QModelIndexList& indexes) const;
        QStringList mimeTypes () const;

    signals:
        void newBackgroundJob (KJob* job);

        // needed to display current status of long lasting operations
        void numberOfFilesToAdd (int);
        void displayProgress (bool);
        void details (const QString&);
        void numberAlreadyProcessed (int);

        // tells a little more about internals
        /** This signal is emitted when the document gets either empty (no items included) or
     * when it was empty and got filled. The boolean is set accordingly.
     * @param empty is true when the document is empty after signal is emitted
     */
        void isEmpty (bool empy);

        /** This signal is emitted when either a dirty model is persisted or when it is first changed
     * after a consistent persistent state.
     */
        void isDirty (bool dirty);

        void setUnsaved (int num);

        void hasDuplicates(bool);
        void changeDirtyState (int start, int end);

        /** Emitted when extended attributes of a track are changed. To identify the altered attribute a generic
       mechanism is used.
       @param row Row whichs extension is altered
       @param type Type of the changed extension. Currently 0=Image, 1=lyrics and 2=string attribute is defined
       @param name Name of the attribute, image or lyric part changed
     */
        void alteredExtension (int row, int type, const QString& name, const QVariant& value);

    public slots:
        void changeArtistName (const QString& oldName, const QString& newName);
        void appendList (const TagItem::List list);
        void clear ();
        void updateChanged (bool removeUpdatedFromList);
        void removeRow (TagItem* item);
        // Removes a row without offering undo - useful when a file is deleted permanently on disk
        void removeRowDirect (TagItem* item);
        void emitCurrentState ();

        void loadPlaylist (const QString& playlist);
        void addPlaylist (M3U::List playlist);
        void updateSaveListStatistics (const QString& filename, bool saved, int delta);
        void addAdditionalInfoFile (const QString& file,DirTree::AdditionalFileFormats format);

        void removeStaleEntries ();

    private slots:
        void onUpdatedTag (TagItem* item);
        void onUpdateChangedFinished (KJob* job);
        void addDuplicateRelation (int rows, const QList<int>& dups);

    protected:
        /** Provides the properties of the cells. The implementation relies on the meta information
     provided by the ColumnManager. Additional the actual cell at the index is considered. If
     it doesn't exist (which might happend because the complete column is globally disabled or
     because the provided feature fro this column is not supported by the current meta data)
     the flags limit access to the cell.
     @returns the flags combination for the cell.
     */
        virtual Qt::ItemFlags flags (const QModelIndex& index) const;

        virtual Qt::DropActions supportedDropActions () const
        {
            return Qt::DropActions(Qt::CopyAction|Qt::MoveAction);
        }

    private:
        // Needed for recursive checking of all directories, see slot removeStaleEntries
        void scanner (DirTree::Dir* start, unsigned& processed);

        void scannerForCDDBCompletion (DirTree::Dir* start, unsigned& processed);

        UpdateTagsJob* createUpdateJob (TagItem::Set toUpdate, bool forceRename);
        void postProcessUpdateChanged (UpdateTagsJob* updateJob);

        // -------------------------------------------------------------------------------------------
        // Used to provide undo/redo commands. The general idea is to generate a QUndoCommand
        // derived object which does the action itself. Butit acts simply as a delegate for
        // an model internal function. Therefore the external undo command classes are friends
        // to call private functions
        //

        /** Copies a list of @var values into the model starting at @from. The @var role
     * is used for all data settings. This methods takes care of efficient signal
     * emitting for changed cells.
     * @param from Start index in the mode from where the list is copied to
     * @param values Values to be copied into the list. Does also define the number of altered
     *             rows.
     */
        bool setDataListRedo (const QList<UpdateList::Entry>& entries, UpdateList::Entry::ValueTypes type, Columns::Column::Role role);

        /** Copies a single value into a range of cells.
     @param from Top,left index of the range (inclusive)
     @param to Bottom, right corner of the range (inclusive)
     @param value Value which is copied into all of the cells.
     @param role Role which is used with setData
     */
        bool setDataRangeRedo (const QList<SlimIndex>& indexes, const QVariant& value, Columns::Column::Role role);

        /** Sets a value in a certain cell. This function does set the value of a single cell. It does
     * also take care of signals for immediate display of the new content.
     * @param index index of the cell to update
     * @param value Value to set.
     * @param role Role to use when setting the data
     */
        bool setDataRedo (const QModelIndex& index, const QVariant& value, Columns::Column::Role role);

        /** Data setter which maintaines all internal search indexes. As this is a prvate function no
     * additional validity checks on the input parameters is done. Invalid rows and columns may
     * crash the application.
     * @param row Row to set
     * @param column Column to set
     * @param value Value to set
     * @param role Role which is used set set the data in the columns.
     */
        bool setDataInternally (int row, int column, const QVariant& value, Columns::Column::Role role);

        void removeRowsRedo (int first, int last);
        void insertRowsRedo (int start, const TagItem::List& rows);
        void clearRedo ();

        friend class UpdateList;
        friend class Clear;
        friend class RemoveRows;
        friend class ToggleBool;

        //
        // End redo
        // -------------------------------------------------------------------------------------------

        QVariant displayData (const QModelIndex & index) const;
        QVariant tooltipData (const QModelIndex & index) const;

        void loadTags(QXmlStreamReader& reader);

        struct Private;
        Private* p;
    };

    inline void TagModel::emitAlteredExtension (int row, int type, const QString& name, const QVariant& value)
    {
        emit alteredExtension (row,type,name,value);
    }
}

#endif // TexelTagModel
