/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "preferences.h"

#include <QApplication>
#include <QPalette>
#include <QFileInfo>
#include <QDir>

#include <kstandarddirs.h>
#include <kmessagebox.h>
#include <klocale.h>

#include "aid.h"

using namespace std;

namespace Texel {

    Preferences::Preferences()
        : KConfigSkeleton()
    {
        setCurrentGroup("Search & Replace Defaults");
        addItemBool("defCaseSensitive",search.m_CaseSensitive,"CaseSensitive");
        addItemBool("defWholeWords",search.m_WholeWords,false,"WholeWordsOnly");
        addItemBool("defRegExp",search.m_RegExp,false,"FindBackwards");
        addItemBool("defSelection",search.m_Selection,false,"SelectedTextOnly");

        setCurrentGroup("Search & Replace");
        addItemBool ("persist",search.m_Persist,false,"persistdata");
        addItemInt ("historysize",search.m_Historysize,10,"listsize");

        setCurrentGroup("IO Options");
        addItemString("TrashCan",io.m_TrashCan,"/tmp/","TrashcanDirectory");
        addItemBool("ackFileRemove",io.m_AckFileRemove,true,"AcknowledgeFileRemoval");
        addItemBool("ackDirRemove",io.m_AckDirRemove,false,"AcknowledgeEmptyDirectoryRemoval");
        addItemBool("RemoveEmptyDir",io.m_RemoveEmptyDir,true,"RemoveEmptyDirectory");
        addItemBool("OverwriteOnRename",io.m_OverwriteOnRename,false,"OverwriteFilesOnRename");
        addItemBool("UseTrashCan",io.m_UseTrashCan,false,"UseOwnTrashcanOnRemove");
        addItemBool("IOOverwriteinTrashCan",io.m_OverwriteInTrashCan,true,"OverwriteinTrashCan");
        addItemBool("IOTryEnforceWrite",io.m_TryEnforceWrite, true, "TryEnforceWrite");
        addItemString ("TrackPatterns",io.m_TrackPatterns, "", "TrackFilePatterns");
        addItemString ("PlaylistPatterns",io.m_PlaylistPatterns, "m3u;M3U", "PlaylistFilePatterns");

        setCurrentGroup("Duplicate Detection");
        addItemInt("maxErrorsInARow",m_DD_maxErrorInARow,2);
        addItemInt("maxDuplicateDisplay",m_DD_maxDuplicateDisplay,10);

        setCurrentGroup("File locations");
        addItemString ("replacePatternFile",filelocs.m_ReplacePatternsFile,KStandardDirs::locateLocal("appdata", "patterns.xml"));
        addItemString ("phrasesFile",filelocs.m_PhrasesFile,KStandardDirs::locateLocal("appdata", "phrases.xml"));
        addItemString ("defaultDirectory",filelocs.m_DefaultDirecory,QDir::homePath());
        addItemString ("lastAccessDir",filelocs.m_LastAccessedDirectory,"");
        addItemBool ("useLastDirectory",filelocs.m_UseLastDirectory,true);
        addItemString ("defaultPlaylistName",filelocs.m_DefaultPlaylistName,QDir::homePath()+QDir::separator()+i18nc("default name for playlists","Untitled",QDir::separator()));

        setCurrentGroup("Debug");
        addItemBool ("trace_dragging", debug.m_TraceDragging, false);

        setCurrentGroup("Appearance");
        addItemFont ("fontNormal", appearance.m_normalFont,QApplication::font());
        addItemFont ("fontHeader", appearance.m_headerFont,QApplication::font());
        addItemColor ("colorNormal", appearance.m_normalForeground,QApplication::palette().text().color());
        addItemColor ("colorDeleted", appearance.m_markedForDeletionForeground,QApplication::palette().color(QPalette::Disabled,QPalette::WindowText));
        addItemColor ("colorHeader", appearance.m_headerForeground,QApplication::palette().text().color());
        addItemColor ("colorBackgroundNormal", appearance.m_normalBackground,QApplication::palette().base().color());
        addItemColor ("colorBackgroundHeader", appearance.m_headerBackground,QApplication::palette().alternateBase().color());
        addItemColor ("colorAlteredBackground", appearance.m_alteredBackground,QApplication::palette().midlight().color());
        addItemBool ("showHelpLinks", appearance.m_ShowHelpLinks, true);

        setCurrentGroup("Musicbrainz");
        addItemBool ("MBenabled", musicbrainz.m_Enabled, false, "enabled");
        addItemString ("MBmusicBrainzURL",musicbrainz.m_URL, "http://musicbrainz.org","url");

        setCurrentGroup("PersistentCache");
        addItemString ("MBcacheFile", cache.m_CacheFile, KStandardDirs::locateLocal("cache","texel.cache"),"cacheFile");
        addItemInt ("MBmaxCacheFileSize", cache.m_MaxFileSize, 100, "maxCacheFileSize");
        addItemInt ("PCOldestEntriesInDays", cache.m_OldestEntriesInDays, 360, "oldestEntriesInDays");

        setCurrentGroup("FreeDB");
        addItemString ("FDBUrl", freedb.m_Url, "http://freedb.org:80","url");
        addItemString ("indexFilePath",freedb.m_IndexFile,KStandardDirs::locateLocal("appdata","reversefreedb.index"));
        addItemString ("FDBindexDownloadURL",freedb.m_IndexDownloadURL,"http://downloads.sourceforge.net/project/texel/reverse lookup indexes v2/reversefreedb.index.gz");
        readConfig();

        setCurrentGroup("Echonest");
        addItemBool ("ENenabled", echonest.m_Enabled, false, "enabled");
        addItemString ("ENechonestURL",echonest.m_URL, "http://developer.echonest.com/api/v4/artist/search","url");

        setCurrentGroup("DuplicateHelper");
        addItemInt ("IncludeInFilter",(int&)duplicatehelper.m_Properties,Enums::FilterFormatMPEG|Enums::FilterFormatMP4|Enums::FilterFormatOgg|Enums::FilterFormatFLAC|
                    Enums::FilterQualityStereo|Enums::FilterQualityHigh|Enums::FilterQualityMedium);
        addItemStringList ("FilterIncludeDirectories",duplicatehelper.m_IncludeDirectories);

        setCurrentGroup ("Texel view settings");
        addItemBool ("RestoreDocks",behaviour.m_RestoreDocksOnStartup,true,"restoreDocks");
        addItemBool ("ShowWelcomeScreen",behaviour.m_ShowWelcomeScreen,true,"showWelcomeScreen");
        addItemBool ("CheckDirty",behaviour.m_CheckDirty,true,"checkDirty");
        addItemUInt ("UndoStackSize",behaviour.m_UndoStackSize,100,"undoStackSize");
        addItemUInt ("MaxLogEntries",behaviour.m_MaxLogEntries,1000,"maxLogEntries");

        setCurrentGroup ("TipOfDay");
        addItemBool ("ShowTOD",behaviour.m_ShowTOD,true,"RunOnStart");

        setCurrentGroup ("Cover images");
        addItemBool ("SaveFileNameInDescription",coverimages.m_SaveFileNameInDescription,false);
        addItemBool ("KeepJPGAndPNG",coverimages.m_KeepJPGAndPNG,true);
        addItemBool ("SaveAsJPG",coverimages.m_SaveAsJPG,true);
        addItemBool ("ShrinkImage",coverimages.m_ShrinkImage,true);
        addItemInt ("ShrinkX",coverimages.m_ShrinkX,300);
        addItemInt ("ShrinkY",coverimages.m_ShrinkY,300);

        setCurrentGroup ("Cover search");
        addItemBool ("SquareImagesOnly",coverimages.m_SquareImagesOnly,false);
        addItemInt ("MaximumItemsToQuery",coverimages.m_MaximumItemsToQuery,20);

        setCurrentGroup ("Tag Extensions");
        addItemInt ("CacheSize",extensions.m_CacheSize,1000);

        setCurrentGroup ("HTML Grabber");
        addItemStringList ("covergrabbers", htmlgrabber.m_ActiveCoverGrabbers);

        readGlobalVars();

        KGlobal::dirs()->addResourceType("grabber","appdata","grabber");
    }

    bool Preferences::HtmlGrabber::isCoverGrabberActive (const QString& grabber) const
    {
        return m_ActiveCoverGrabbers.contains(grabber);
    }

    void Preferences::HtmlGrabber::setGrabberState (const QString& grabber, bool active)
    {
        if (active && !isCoverGrabberActive(grabber)) {
            m_ActiveCoverGrabbers << grabber;
        }
        else if (!active) {
            m_ActiveCoverGrabbers.removeAll(grabber);
        }
    }

    QString Preferences::getReadNamePatternsFile () const
    {
        const QFileInfo info(filelocs.m_ReplacePatternsFile);
        if (!info.exists()) {
            return KStandardDirs::locate("appdata", "patterns.xml");
        }
        return filelocs.m_ReplacePatternsFile;
    }

    QString Preferences::getWriteNamePatternsFile () const
    {
        const QString local = KStandardDirs::locateLocal("appdata", "patterns.xml");
        if (filelocs.m_ReplacePatternsFile==local) {
            const QString global = KStandardDirs::locate("appdata", "patterns.xml");
            if (global!=local) {
                QFileInfo info(global);
                if (info.isWritable()) {
                    int result=KMessageBox::questionYesNo  (NULL,
                                                            i18n("You changed patterns and they are about to be saved. The global "
                                                                 "patterns file (%2) is writeable for you. Do you like to overwrite that "
                                                                 "instead of the local file \"%1\"?",local,global));
                    if (result==KMessageBox::Yes) {
                        return global;
                    }
                }
            }
        }
        return filelocs.m_ReplacePatternsFile;
    }

    QString Preferences::getReadNamePhrasesFile () const
    {
        const QFileInfo info(filelocs.m_PhrasesFile);
        if (!info.exists()) {
            return KStandardDirs::locate("appdata", "phrases.xml");
        }
        return filelocs.m_PhrasesFile;
    }

    QString Preferences::getWriteNamePhrasesFile () const
    {
        const QString local = KStandardDirs::locateLocal("appdata", "phrases.xml");
        if (filelocs.m_PhrasesFile==local) {
            const QString global = KStandardDirs::locate("appdata", "phrases.xml");
            if (global!=local) {
                QFileInfo info(global);
                if (info.isWritable()) {
                    int result=KMessageBox::questionYesNo  (0l,
                                                            i18n("You changed phrases and they are about to be saved. The global "
                                                                 "phrases file (%2) is writeable for you. Do you like to overwrite that "
                                                                 "instead of the local file \"%1\"?",local,global));
                    if (result==KMessageBox::Yes) {
                        return global;
                    }
                }
            }
        }
        return filelocs.m_PhrasesFile;
    }

    void Preferences::readGlobalVars ()
    {
        KConfigGroup group = config()->group("Global Variables");
        m_GlobalVars.clear();
        // Do not use group.entryMap() to allow otential filtering here
        foreach (QString name, group.keyList()) {
            m_GlobalVars.insert(name,group.readEntry(name));
        }
    }

    void Preferences::writeGlobalVars ()
    {
        KConfigGroup group = config()->group("Global Variables");

        // Delete entries from group which are not in the global vars map
        foreach (QString name, group.keyList()) {
            if (!m_GlobalVars.contains(name)) {
                group.deleteEntry(name);
            }
        }

        // Set existing entries
        foreach (QString name, m_GlobalVars.keys()) {
            group.writeEntry(name,m_GlobalVars[name]);
        }
    }

    Preferences::~Preferences()
    {
        writeGlobalVars();
        writeConfig();
    }

}
