/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "texel.h"
#include "model/texeltagmodel.h"
#include "model/tagspecializations.h"
#include "view/messageview.h"
#include "model/jobs/tagmodelxmlstreamer.h"
#include "model/columnmanager.h"
#include "model/tagmodelproxy.h"
#include "model/jobs/jobtracker.h"
#include "model/structure/artistmanager.h"
#include "model/structure/mastermanager.h"

#include "util/filedialog.h"
#include "util/slowclicktoolbutton.h"
#include "util/completionlineeditor.h"
#include "util/loghook.h"

#include "view/formulareditor/formulareditor.h"
#include "view/docks/trackdetails.h"
#include "view/docks/duplicatehelper.h"
#include "view/docks/directorydock.h"
#include "view/docks/playlistdock.h"
#include "view/docks/playerdock.h"
#include "view/docks/logdock.h"
#include "view/credits.h"
#include "view/layout/tablelayout.h"

#include "func/duplicatedetection.h"
#include "func/phrases/phraseview.h"
#include "func/phrases/phrasemodel.h"
#include "func/phrases/phraseitem.h"
#include "func/metatagservices/freedb/freedbaccess.h"

#include "options/preferences.h"
#include "options/IOOptions.h"
#include "options/rsoption.h"
#include "options/filteroption.h"
#include "options/fileoptions.h"
#include "options/appearanceoption.h"
#include "options/externalresources.h"
#include "options/bhoptions.h"
#include "options/globalvars.h"
#include "options/coveroptions.h"

#include <kfiledialog.h>
#include <kmenu.h>
#include <krecentfilesaction.h>
#include <kconfigdialog.h>
#include <kpastetextaction.h>
#include <kstandardaction.h>
#include <kactioncollection.h>
#include <kundostack.h>
#include <kstandarddirs.h>
#include <kdebug.h>
#include <kmessagebox.h>
#include <ktoolbar.h>
#include <kaboutapplicationdialog.h>
#include <kprogressdialog.h>
#include <kdialog.h>

#include <iostream>

#include <QClipboard>
#include <QFile>
#include <QFileInfo>
#include <QXmlStreamReader>
#include <QXmlStreamWriter>
#include <QMimeData>
#include <QSet>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QWidgetAction>
#include <QUndoView>
#include <QItemSelectionRange>
#include <QFont>
#include <QCloseEvent>
#include <QTextDocument>
#include <QTextCursor>
#include <QTextEdit>
#include <QHBoxLayout>

using namespace std;

#undef QT_NO_DEBUG

namespace Texel {
QString s_EmptyString;

texel::texel(QWidget* parent)
    : KMainWindow(parent),
    m_RecentDirs(0l),
    m_RecentLists(0l),
    m_player(0l),
    m_WelcomeScreen(0l),
    m_ActionCollection(new KActionCollection(this)),
    m_PhraseView(0),
    m_PhraseModel(0),
    showProgress(false),
    progress(0),
    modelDirty(false),
    numModelDirty(0),
    m_MessageModel(new MessageModel(this)),
    m_CoreModel(new Texel::TagModel(this)),
    m_TrackDetailsDock(0l),
    m_AboutData(0l)
{
    setupUi(this);
    tableWidget->setDragDropMode(QAbstractItemView::DragOnly);
    tableWidget->setDragEnabled(true);

    KToolBar::setToolBarsLocked(false);

    // Add icons
    actionSaveListAs->setIcon(KIconLoader::global()->loadIcon("document-save-as",KIconLoader::Toolbar));
    actionSaveList->setIcon(KIconLoader::global()->loadIcon("document-save",KIconLoader::Toolbar));
    actionLoadList->setIcon(KIconLoader::global()->loadIcon("document-open",KIconLoader::Toolbar,KIconLoader::SizeHuge));
    actionLastList->setIcon(KIconLoader::global()->loadIcon("document-open-recent",KIconLoader::Toolbar,KIconLoader::SizeHuge));
    actionFitToScreen->setIcon(KIconLoader::global()->loadIcon("zoom-fit-best",KIconLoader::Toolbar));
    actionClearList->setIcon(KIconLoader::global()->loadIcon("edit-clear-list",KIconLoader::Toolbar));
    actionAddCoverToSelected->setIcon(KIconLoader::global()->loadIcon("search-cover",KIconLoader::Toolbar));
    actionFilter->setIcon(KIconLoader::global()->loadIcon("view-filter",KIconLoader::Toolbar));
    actionRemoveFilter->setIcon(KIconLoader::global()->loadIcon("kt-remove-filters",KIconLoader::Toolbar));
    actionDuplicateDetection->setIcon(KIconLoader::global()->loadIcon("detect-duplicates",KIconLoader::Toolbar));
    actionOpen->setIcon(KIconLoader::global()->loadIcon("document-open-folder",KIconLoader::Toolbar,KIconLoader::SizeHuge));
    actionSave->setIcon(KIconLoader::global()->loadIcon("document-save-all",KIconLoader::Toolbar));
    actionFillDown->setIcon(KIconLoader::global()->loadIcon("arrow-down-double",KIconLoader::Toolbar));
    actionGroupBy->setIcon(KIconLoader::global()->loadIcon("group-by",KIconLoader::Toolbar));
    actionLastDir->setIcon(KIconLoader::global()->loadIcon("folder-sound",KIconLoader::Toolbar,KIconLoader::SizeHuge));

    menuMetadata->setIcon(KIconLoader::global()->loadIcon("metadata-available",KIconLoader::Toolbar));
    menuFile_operatons->setIcon(KIconLoader::global()->loadIcon("drive-harddisk",KIconLoader::Toolbar));
    menuFilter->setIcon(KIconLoader::global()->loadIcon("view-filter",KIconLoader::Toolbar));
    menuGroupBy->setIcon(KIconLoader::global()->loadIcon("group-by",KIconLoader::User));
    menuCover_Images->setIcon(KIconLoader::global()->loadIcon("search-cover",KIconLoader::Toolbar));
    menuSortBy->setIcon(KIconLoader::global()->loadIcon("view-sort-ascending",KIconLoader::Toolbar));

    updateCaption();
    Texel::TagModelProxy* tagmodel=new Texel::TagModelProxy(this);
    tagmodel->setSourceModel(m_CoreModel);
    tableWidget->setModel (tagmodel);
    tableWidget->setControler(this);
    connect (m_CoreModel, SIGNAL(hasDuplicates(bool)), tableWidget->verticalHeader(), SLOT(onDuplicates(bool)));
    connect (m_CoreModel, SIGNAL(changeDirtyState(int,int)), tableWidget->verticalHeader(), SLOT(headerDataChanged(int,int)));
    connect (m_CoreModel, SIGNAL(headerDataChanged(Qt::Orientation,int,int)), tableWidget->verticalHeader(), SLOT(headerDataChanged(Qt::Orientation,int,int)));
    connect (m_CoreModel, SIGNAL(newBackgroundJob(KJob*)), this, SLOT(addBackgroundJob(KJob*)));
    connect (tagmodel, SIGNAL(message(QString,int)), statusBar(), SLOT(showMessage(QString,int)));
    connect (functionEditorButton, SIGNAL(clicked()), this, SLOT(editFormular()));

    KConfigGroup texelConfig(KGlobal::config(),"Texel view settings");
    // Read hidden/shown columns
    // create default
    QStringList def;
    for (int i=0;i<Texel::ColumnManager::instance()->getColumnNumber();i++) {
        if (Texel::ColumnManager::instance()->isHidden(i)) { def << QString::number(i); }
    }
    QStringList hiddenColumns = texelConfig.readEntry("hidden columns",def);
    foreach (const QString& i, hiddenColumns) {
        bool conv=false;
        int col = i.toInt(&conv);
        if (conv) tableWidget->hideColumn(col);
    }

    //
    // Small widget to display the current tag type
    //

    {
        m_TagTypeDisplay = new QLabel(statusBar());
        m_TagTypeDisplay->setToolTip(i18n("Tag type from tag of current cell"));
        QFontMetrics m(m_TagTypeDisplay->font());
        m_TagTypeDisplay->setMinimumSize(m.size(0,"Ogg::Vorbis"));
        m_TagTypeDisplay->setAlignment(Qt::AlignCenter);
        statusBar()->addPermanentWidget(m_TagTypeDisplay);
        m_TagTypeDisplay->setFrameStyle(QFrame::Panel|QFrame::Sunken);
        connect (tableWidget, SIGNAL(currentIndexChanged(QModelIndex,QModelIndex)), this, SLOT(updateType(QModelIndex,QModelIndex)));
    }

    //
    // Small widget to display the current row number
    //

    {
        QFrame* frame = new QFrame(statusBar());
        QLabel* icon = new QLabel(frame);
        icon->setPixmap(KIconLoader::global()->loadIcon("flag-red",KIconLoader::Small));
        icon->setMargin(0);
        m_TagCurrentRowDisplay = new QLabel(statusBar());
        frame->setToolTip(i18n("Shows the current row number"));
        QFontMetrics m(m_TagCurrentRowDisplay->font());
        m_TagCurrentRowDisplay->setMinimumSize(m.size(0,"00000"));
        m_TagCurrentRowDisplay->setAlignment((Qt::Alignment)Qt::AlignRight|(Qt::Alignment)Qt::AlignVCenter);
        QLayout* layout=new QHBoxLayout(frame);
        frame->setLayout(layout);
        layout->setContentsMargins(1,0,1,0);
        layout->addWidget(icon);
        layout->addWidget(m_TagCurrentRowDisplay);
        statusBar()->addPermanentWidget(frame);
        frame->setFrameStyle(QFrame::Panel|QFrame::Sunken);
        connect (tableWidget, SIGNAL(currentIndexChanged(QModelIndex,QModelIndex)), this, SLOT(updateCurrentRow(QModelIndex,QModelIndex)));
    }

    //
    // Widget for currently unsaved rows
    //

    {
        QFrame* dirtyTagsDisplayFrame = new QFrame(statusBar());
        QLabel* dirtyTagsDisplayIcon = new QLabel(dirtyTagsDisplayFrame);
        dirtyTagsDisplayIcon->setPixmap(KIconLoader::global()->loadIcon("document-save",KIconLoader::Small));
        dirtyTagsDisplayIcon->setMargin(0);
        m_TagCurrentDirtyDisplay = new QLabel(dirtyTagsDisplayFrame);
        m_TagCurrentDirtyDisplay->setMargin(0);
        QFontMetrics m(dirtyTagsDisplayFrame->font());
        m_TagCurrentDirtyDisplay->setMinimumSize(m.size(0,"0000"));
        m_TagCurrentDirtyDisplay->setAlignment((Qt::Alignment)Qt::AlignRight|(Qt::Alignment)Qt::AlignVCenter);
        dirtyTagsDisplayFrame->setToolTip(i18n("Number of currently unsaved elements"));
        QLayout* layout=new QHBoxLayout(dirtyTagsDisplayFrame);
        dirtyTagsDisplayFrame->setLayout(layout);
        layout->setContentsMargins(1,0,1,0);
        layout->addWidget(dirtyTagsDisplayIcon);
        layout->addWidget(m_TagCurrentDirtyDisplay);
        statusBar()->addPermanentWidget(dirtyTagsDisplayFrame);
        dirtyTagsDisplayFrame->setFrameStyle(QFrame::Panel|QFrame::Sunken);
        connect (m_CoreModel, SIGNAL(setUnsaved(int)), this, SLOT(updateUnsavedNumber(int)));
    }

    //
    // Define Edit menu & toolbar actions
    //
    QAction* undoaction = m_CoreModel->undoStack()->createUndoAction(m_ActionCollection);
    QAction* redoaction = m_CoreModel->undoStack()->createRedoAction(m_ActionCollection);
    m_CoreModel->undoStack()->setUndoLimit(Preferences::instance()->behaviour.m_UndoStackSize);

    QList<QAction*> list;

    list << undoaction << redoaction;
    menuEdit->addActions(list);
    list.clear();
    menuEdit->addSeparator();

    list << KStandardAction::paste(this, SLOT(paste()), centralWidget())
            << KStandardAction::copy(this, SLOT(copy()), centralWidget())
            << KStandardAction::cut(this, SLOT(cut()), centralWidget());
    foreach (QAction* action, list) { action->setShortcutContext(Qt::WidgetWithChildrenShortcut); }
    centralwidget->addActions(list);
    menuEdit->addActions(list);

    editToolBar->addActions(list);
    editToolBar->addAction(redoaction);
    editToolBar->addAction(undoaction);
    // Use the popmenu feature of an toolbutton to show an undoview: Create an QWidgetAction
    // set the QUndoView as its default and store the QWidgetAction in a QMenu wich in turn is
    // attached to the undoaction (und than automatically to the toolbutton)
    QWidgetAction* undowidget = new QWidgetAction(editToolBar);
    undowidget->setDefaultWidget(new QUndoView(m_CoreModel->undoStack()));
    QMenu* tmpmenu=new QMenu(editToolBar);
    tmpmenu->addAction(undowidget);
    undoaction->setMenu(tmpmenu);

    searchWidget->hide();
    searchWidget->setModel(tagmodel);
    searchWidget->setTableWidget(tableWidget);
    stdActionSearch = KStandardAction::find(this,SLOT(search()),centralWidget());
    stdActionSearch->setShortcutContext(Qt::WidgetWithChildrenShortcut);
    centralWidget()->addAction(stdActionSearch);
    stdActionReplace = KStandardAction::replace(this,SLOT(search()),centralWidget());
    stdActionReplace->setShortcutContext(Qt::WidgetWithChildrenShortcut);
    centralWidget()->addAction(stdActionReplace);

    menuEdit->addSeparator();
    menuEdit->addAction(stdActionSearch);
    menuEdit->addAction(stdActionReplace);

    stdActionOptions = KStandardAction::preferences(this,SLOT(showOptions()),this);
    menuEdit->addSeparator();
    menuEdit->addAction(stdActionOptions);

    connect (actionOpen, SIGNAL(triggered()), this, SLOT(openByDir()));
    connect (actionSave, SIGNAL(triggered()), this, SLOT(save()));
    connect (actionSaveList, SIGNAL(triggered()), this, SLOT(writeList()));
    connect (actionSaveListAs, SIGNAL(triggered()), this, SLOT(writeListAs()));
    connect (actionLoadList, SIGNAL(triggered()), this, SLOT(loadList()));
    stdActionQuit = KStandardAction::quit(this, SLOT(close()), this);
    menuFile->insertAction(0l,stdActionQuit);

    connect (actionFillDown, SIGNAL(triggered()), this, SLOT(fillDown()));

    connect (actionRemoveFiles, SIGNAL(triggered()), this, SLOT(removeSelectedFiles()));
    connect (actionRemoveFromList, SIGNAL(triggered()), this, SLOT(removeSelectedRows()));
    connect (actionSaveAndRemove, SIGNAL(triggered()), this, SLOT (saveAndRemove()));
    connect (actionClearList, SIGNAL(triggered()), m_CoreModel, SLOT(clear()));
    connect (actionDuplicateDetection, SIGNAL(triggered()), this, SLOT(duplicatedetection()));
    connect (actionLastDir, SIGNAL(triggered()), this, SLOT(loadMostRecentDir()));
    connect (actionLastList, SIGNAL(triggered()), this, SLOT(loadMostRecentList()));

    // Create Message Logger and helpers
    qRegisterMetaType<Enums::MessageTypes>("Enums::MessageTypes");

    connect (LogHook::instance()->getLoggerObject(), SIGNAL(log(Enums::MessageTypes,QString,TagItem*)), m_MessageModel, SLOT(message(Enums::MessageTypes,QString,TagItem*)));
    connect (m_MessageModel, SIGNAL(newMaxFaultLevel(int)), this, SLOT(maxFaultLevel(int)));
    m_MessageButton = new QPushButton(statusbar);
    m_MessageButton->setFlat(true);
    m_MessageButton->setToolTip(i18n("Get texel activity log."));
    connect (m_MessageButton, SIGNAL(clicked()), aLoggerDock, SLOT(toggle()));
    statusbar->addPermanentWidget (m_MessageButton);
    m_MessageButton->show();

    m_CoreModel->setMessageModel(m_MessageModel);

    connect (tableWidget, SIGNAL(currentText(const QString&)), currentItemEdit, SLOT(setText(const QString&)));
    tableWidget->setEditBuddy(currentItemEdit);
    connect (currentItemEdit, SIGNAL(editingFinished()), tableWidget, SLOT(currentEditingFinished()));
    connect (m_CoreModel, SIGNAL(rowsInserted (const QModelIndex &, int, int )), this, SLOT(onModelRowChange(const QModelIndex &, int, int )));
    connect (m_CoreModel, SIGNAL(rowsRemoved (const QModelIndex &, int, int )), this, SLOT(onModelRowChange(const QModelIndex &, int, int )));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), this, SLOT(onEmptyModel(bool)));

    //
    // Extend QT Designer generated UI with KDE standard elements
    //

    // KDE Help Menu
    KMenu* help = customHelpMenu( "Help" );
    menuBar()->addMenu(help);

    // Recent files menus
    m_RecentDirs = new KRecentFilesAction("Recent accessed dirs ...",menuFile);
    m_RecentLists = new KRecentFilesAction("Recent accessed lists ...",menuFile);
    m_RecentDirs->loadEntries(KConfigGroup(KGlobal::config(),"Recent Dirs"));
    m_RecentLists->loadEntries(KConfigGroup(KGlobal::config(),"Recent Lists"));
    connect (m_RecentLists, SIGNAL(urlSelected(const KUrl&)), this, SLOT(loadList(const KUrl&)));
    connect (m_RecentDirs, SIGNAL(urlSelected(const KUrl&)), this, SLOT(addDir(const KUrl&)));
    menuFile->insertAction (actionSave, m_RecentDirs);
    menuFile->insertAction (actionSaveList, m_RecentLists);

    //
    // Action Menu
    //

    connect (actionRemoveStaleTags, SIGNAL(triggered()), this, SLOT(removeStaleTags()));

    // Phonon audio replay facilities
    m_player = Texel::Player::Player::instance();
    m_player->setItemView(tableWidget);
    m_player->initToolBar (playerToolBar);
    m_player->initStatusBar (statusbar);

    connect (m_player, SIGNAL(error(QString)), dynamic_cast<MessageModel*>(m_MessageModel), SLOT(error(QString)));
    connect (m_player, SIGNAL(openPlayDock(bool)), aPlayerDock, SLOT(setChecked(bool)));

    QMenu* playerMenu=m_player->createMenu();
    menuAction->insertMenu(actionFillDown, playerMenu);

    connect (actionAddCoverToSelected, SIGNAL(triggered()), tableWidget, SLOT(addCoverImagesToSelected()));
    connect (actionAddCoverToAllWithout, SIGNAL(triggered()), tableWidget, SLOT(addCoverImagesWhenNonExists()));
    connect (actionAddCoverToAll, SIGNAL(triggered()), tableWidget, SLOT(addCoverImagesToAll()));

    //
    // Phrases menu
    //

    m_PhraseModel=new Texel::PhraseModel (this);
    m_PhraseView=new Texel::PhraseView (this, m_PhraseModel, menuPhrases);
    connect (m_PhraseModel, SIGNAL(phraseTriggered(PhraseItem*)), this, SLOT(phraseTriggered(PhraseItem*)));

    setAutoSaveSettings();
    tableWidget->setFocus();

    //
    // View menu
    //

    // filter menu
    connect (actionWithDuplicates, SIGNAL(triggered()), tagmodel, SLOT(filterDuplicates()));
    connect (actionFilterIncomplete, SIGNAL(triggered()), tagmodel, SLOT(filterIncomplete()));
    connect (actionRemoveFilter, SIGNAL(triggered()), tagmodel, SLOT(resetFilter()));
    connect (actionFilterWithCoverImages,SIGNAL(triggered()), tagmodel, SLOT(filterCoverOnly()));
    connect (actionFilterWithoutCoverImages,SIGNAL(triggered()), tagmodel, SLOT(filterNoCoversOnly()));
    connect (actionFilterFormatsWithCoverSupport,SIGNAL(triggered()), tagmodel, SLOT(filterCoverSupportingFormats()));
    // connect group by
    connect (actionRemoveGrouping, SIGNAL(triggered()), tagmodel, SLOT(removeGrouping()));
    foreach (QAction* action, menuGroupBy->actions()) {
        if (action!=actionRemoveGrouping) {
            connect (action, SIGNAL(triggered()), tagmodel, SLOT(groupBy()));
        }
    }
    // sort by menu
    for (int i=0;i<ColumnManager::instance()->getColumnNumber();i++) {
        QAction* action = menuSortBy->addAction (ColumnManager::instance()->getName(i),this, SLOT(sortBy()));
        action->setProperty("column",QVariant(i));
    }
    connect (tableWidget->horizontalHeader(), SIGNAL(sortIndicatorChanged(int,Qt::SortOrder)), this, SLOT(sortedBy(int,Qt::SortOrder)));
    connect (actionShowFormulars, SIGNAL(triggered(bool)), this, SLOT(showFormulars(bool)));
    connect (actionRemoveSorting, SIGNAL(triggered()), this, SLOT(sortBy()));
    actionRemoveSorting->setProperty ("column",QVariant(-1));

    //
    // Filter toolbar
    //

    createFilterToolbar (tagmodel);

    // add submenu
    QMenu* filterButtonMenu = new QMenu(filterToolBar);
    filterButtonMenu->addAction(actionFilterFlagCase);
    filterButtonMenu->addSeparator();
    filterButtonMenu->addAction(actionFilterFlagAllWords);
    filterButtonMenu->addAction(actionFilterFlagSingleWord);
    filterButtonMenu->addAction(actionFilterFlagString);
    QActionGroup* group = new QActionGroup(filterButtonMenu);
    group->addAction(actionFilterFlagAllWords);
    group->addAction(actionFilterFlagSingleWord);
    group->addAction(actionFilterFlagString);
    actionFilter->setMenu(filterButtonMenu);
    actionFilterFlagAllWords->setData(TagModelProxy::AllWords);
    actionFilterFlagSingleWord->setData(TagModelProxy::SingleWord);
    actionFilterFlagString->setData(TagModelProxy::AsString);
    connect (actionFilterFlagAllWords, SIGNAL(triggered()), this, SLOT(setFilterType()));
    connect (actionFilterFlagSingleWord, SIGNAL(triggered()), this, SLOT(setFilterType()));
    connect (actionFilterFlagString, SIGNAL(triggered()), this, SLOT(setFilterType()));

    // Disable all action which are useless in certain situations
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), actionDuplicateDetection, SLOT(setDisabled(bool)));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), actionGroupBy, SLOT(setDisabled(bool)));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), actionFilter, SLOT(setDisabled(bool)));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), actionClearList, SLOT(setDisabled(bool)));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), actionFillDown, SLOT(setDisabled(bool)));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), stdActionSearch, SLOT(setDisabled(bool)));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), stdActionReplace, SLOT(setDisabled(bool)));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), actionRemoveFiles, SLOT(setDisabled(bool)));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), actionRemoveFromList, SLOT(setDisabled(bool)));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), actionAddCoverToSelected, SLOT(setDisabled(bool)));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), actionAddCoverToAllWithout, SLOT(setDisabled(bool)));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), actionAddCoverToAll, SLOT(setDisabled(bool)));

    connect (m_CoreModel, SIGNAL(isEmpty(bool)), menuFilter, SLOT(setDisabled(bool)));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), menuGroupBy, SLOT(setDisabled(bool)));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), menuSortBy, SLOT(setDisabled(bool)));
    connect (tagmodel, SIGNAL(hasFilter(bool)), actionRemoveFilter, SLOT(setEnabled(bool)));
    actionRemoveFilter->setEnabled(false);
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), actionRemoveGrouping, SLOT(setDisabled(bool)));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), actionRemoveSorting, SLOT(setDisabled(bool)));

    connect (m_CoreModel, SIGNAL(isDirty(bool)), actionSave, SLOT(setEnabled(bool)));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), actionSaveList, SLOT(setDisabled(bool)));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), actionSaveListAs, SLOT(setDisabled(bool)));
    connect (m_CoreModel, SIGNAL(isDirty(bool)), actionSaveAndRemove, SLOT(setEnabled(bool)));

    connect (m_CoreModel, SIGNAL(isEmpty(bool)), actionRemoveStaleTags, SLOT(setDisabled(bool)));
    connect (m_CoreModel, SIGNAL(isEmpty(bool)), actionFillFromCDDB, SLOT(setDisabled(bool)));

    // status bar widget to display currently loaded items.
    m_NoOfItems = new QLabel(statusBar());
    m_NoOfItems->setToolTip(i18n("Displays the currently available tags in Texel (<in selection>/<in memory>)"));
    statusBar()->addPermanentWidget(m_NoOfItems);
    updateNumberOfItems();
    connect (m_CoreModel, SIGNAL(rowsInserted(QModelIndex, int, int)), this, SLOT(updateNumberOfItems()));
    connect (m_CoreModel, SIGNAL(rowsRemoved(QModelIndex, int, int)), this, SLOT(updateNumberOfItems()));
    connect (tableWidget->model(), SIGNAL(rowsInserted(QModelIndex, int, int)), this, SLOT(updateNumberOfItems()));
    connect (tableWidget->model(), SIGNAL(rowsRemoved(QModelIndex, int, int)), this, SLOT(updateNumberOfItems()));

    //
    // Window menu
    //

    connect (aTrackDetails, SIGNAL(toggled(bool)), this, SLOT(addDetailsDock(bool)));
    connect (aDuplicateHelper, SIGNAL(toggled(bool)), this, SLOT(addDuplicateDock(bool)));
    connect (aDirectoryView, SIGNAL(toggled(bool)), this, SLOT(addDirectoryDock(bool)));
    connect (aPlaylistDock, SIGNAL(toggled(bool)), this, SLOT(addPlaylistDock(bool)));
    connect (aPlayerDock, SIGNAL(toggled(bool)), this, SLOT(addPlayerDock(bool)));
    connect (aLoggerDock, SIGNAL(toggled(bool)), this, SLOT(addLogDock(bool)));

    //
    // Arbitrary actions
    //

    connect (aSelectCurrentDirectory, SIGNAL(triggered()), tableWidget, SLOT(selectAllInSameDirectory()));
    connect (aSelectCurrentDirectoryRec, SIGNAL(triggered()), tableWidget, SLOT(selectAllInSameDirectoryRec()));
    connect (aSetDeferedDeletionTag, SIGNAL(triggered()), tableWidget, SLOT(setDeferedDeletionTag()));
    tableWidget->addAction(aSetDeferedDeletionTag);
    connect (aUnsetDeferedDeletionTag, SIGNAL(triggered()), tableWidget, SLOT(unsetDeferedDeletionTag()));
    tableWidget->addAction(aUnsetDeferedDeletionTag);
    connect (actionRotateDetails, SIGNAL(triggered()), this, SLOT(rotateDetailsDock()));
    tableWidget->addAction(actionRotateDetails);

    //
    // JobTracker
    //

    m_JobTracker = new JobTracker();
    statusBar()->insertWidget (0,m_JobTracker->widget(0l));
    connect (m_JobTracker, SIGNAL(percentExecuted(unsigned)), this, SLOT(captionProgress(unsigned)));
    connect (m_JobTracker, SIGNAL(done()), this, SLOT(removeCaptionProgress()));

    //
    // Backgroundtask: Lookup at musicbrainz
    //

    MasterManager::instance()->createBackgroundService();

    //
    // Create monitor to alter extension cache size when changed in preferences
    //

    CacheConfigGuard* guard = new CacheConfigGuard(this);
    connect (Preferences::instance(), SIGNAL(configChanged()), guard, SLOT(configChanged()));
    guard->configChanged();

    //
    // Setup welcome screen
    //

    QPalette p = page_2->palette();
    p.setColor(QPalette::Background,QColor(255,255,255));
    page_2->setPalette(p);
    page_2->setAutoFillBackground(true);
    WelcomeScreen->setCoreModel(m_CoreModel);

    WelcomeScreen->setPalette(QApplication::palette());

    WelcomeScreen->toolButton_1->setDefaultAction(actionOpen);
    WelcomeScreen->toolButton_2->setDefaultAction(actionLastDir);
    WelcomeScreen->toolButton_3->setDefaultAction(actionLoadList);
    WelcomeScreen->toolButton_4->setDefaultAction(actionLastList);
    QAction* mr=mostRecent(m_RecentDirs);
    if (mr!=0L) {
        WelcomeScreen->toolButton_2->setToolTip(i18n("Read from <br/>%1",mr->text()));
    }
    else {
        WelcomeScreen->toolButton_2->setEnabled(false);
    }
    mr=mostRecent(m_RecentLists);
    if (mr!=0l) {
        WelcomeScreen->toolButton_4->setToolTip(i18n("Read from %1",mr->text()));
    }
    else {
        WelcomeScreen->toolButton_4->setEnabled(false);
    }
    connect (WelcomeScreen, SIGNAL(loadDir(KUrl)), this, SLOT(addDir(KUrl)));
    connect (WelcomeScreen, SIGNAL(loadList(KUrl)), this, SLOT(loadList(KUrl)));

    searchWidget->setup();

    restoreDocks();
    m_CoreModel->emitCurrentState();
}

void texel::createFilterToolbar (Texel::TagModelProxy* tagmodel)
{
    // define sub menu for group by action
    QMenu* groupByMenu=new QMenu(filterToolBar);
    QWidgetAction* menuDescriptor = new QWidgetAction(groupByMenu);
    QLabel* label = new QLabel(i18n("Group by:"),groupByMenu);
    QFont font = label->font();
    font.setItalic(true);
    label->setFont(font);
    menuDescriptor->setDefaultWidget(label);
    groupByMenu->addAction(menuDescriptor);
    foreach (QAction* action, menuGroupBy->actions()) {
        if (action!=actionRemoveGrouping) {
            groupByMenu->addAction(action);
        }
    }
    actionGroupBy->setMenu(groupByMenu);
    connect (actionGroupBy, SIGNAL(triggered()), tagmodel, SLOT(groupBy()));

    // Add a new widget for filtering into the toolbar
    m_FilterEdit = new Texel::CompletionLineEditor(filterToolBar,30);
    m_FilterEdit->setClearButtonShown(true);
    m_FilterEdit->setClickMessage(i18n("Filter pattern"));
    QWidgetAction* wa = new QWidgetAction(this);
    wa->setDefaultWidget(m_FilterEdit);
    filterToolBar->addAction(wa);
    connect (m_FilterEdit, SIGNAL(returnPressed()), this, SLOT(filterSearchText()));
    connect (m_FilterEdit, SIGNAL(returnPressed()), tableWidget, SLOT(setFocus()));
    connect (actionFilter, SIGNAL(triggered()), this, SLOT(filterSearchText()));
    connect (actionRemoveFilter, SIGNAL(triggered()), m_FilterEdit, SLOT(clear()));
}

void texel::showFormulars (bool show)
{
    dynamic_cast<TagModelProxy*>(tableWidget->model())->showFormulars(show);
}

void texel::sortBy ()
{
    QAction* action = dynamic_cast<QAction*>(sender());
    QVariant v = action->property("column");
    if (v.isValid()) {
        int column = v.toInt();
        Qt::SortOrder order = Qt::AscendingOrder;
        if (column!=-1 && tableWidget->verticalHeader()->sortIndicatorSection()==column) {
            order = (tableWidget->verticalHeader()->sortIndicatorOrder()==Qt::AscendingOrder)?Qt::DescendingOrder:Qt::AscendingOrder;
        }
        if (column==-1) {
            // implementation of QTableView suppressed currently sorting of column -1 (unsorting)
            tableWidget->horizontalHeader()->setSortIndicator(-1,Qt::AscendingOrder);
            tableWidget->model()->sort(column,order);
        }
        else tableWidget->sortByColumn(column,order);
    }
}

void texel::sortedBy (int column, Qt::SortOrder order)
{
    QList<QAction*> actions=menuSortBy->actions();
    for (int i=0;i<actions.size();i++) {
        const QVariant v = actions[i]->property("column");
        if (v.isValid()) {
            const int c = v.toInt();
            if (c==column) {
                actions[i]->setIconVisibleInMenu(true);
                if (order==Qt::AscendingOrder) {
                    actions[i]->setIcon (KIconLoader::global()->loadIcon("view-sort-ascending",KIconLoader::Small));
                }
                else {
                    actions[i]->setIcon (KIconLoader::global()->loadIcon("view-sort-descending",KIconLoader::Small));
                }
            }
            else {
                if (actions[i]->isIconVisibleInMenu()) {
                    actions[i]->setIconVisibleInMenu(false);
                    actions[i]->setIcon(QIcon());
                }
            }
        }
    }
}

void texel::updateCaption ()
{
    QString progresstext;
    if (showProgress) progresstext=QString::number(progress)+"%";
    setCaption (progresstext,modelDirty);
}

void texel::changedDirtyState(bool dirty)
{
    if (dirty!=modelDirty) {
        modelDirty=dirty;
        updateCaption();
    }
}

void texel::captionProgress(unsigned percent)
{
    progress=percent;
    updateCaption();
}

void texel::removeCaptionProgress ()
{
    showProgress=false;
    updateCaption();
}

texel::~texel()
{
    // Remove job tracker first to stop background tasks
    delete m_JobTracker;

    KConfigGroup texelConfig(KGlobal::config(),"Texel view settings");
    // Read hidden/shown columns
    // create default
    QStringList hiddenCols;
    for (int i=0;i<Texel::ColumnManager::instance()->getColumnNumber();i++) {
        if (tableWidget->isColumnHidden(i)) { hiddenCols << QString::number(i); }
    }
    texelConfig.writeEntry ("hidden columns", hiddenCols);

    m_RecentDirs->saveEntries(KConfigGroup(KGlobal::config(),"Recent Dirs"));
    m_RecentLists->saveEntries(KConfigGroup(KGlobal::config(),"Recent Lists"));
    saveDocks();

    KGlobal::config()->sync();
}

void texel::showOptions ()
{
    if (KConfigDialog::showDialog("settings")) return;

    KConfigDialog *dialog = new KConfigDialog(this, "settings", Texel::Preferences::instance());
    dialog->setFaceType(KPageDialog::List);
    dialog->addPage(new IOOptions(dialog), i18n("File Operations") )->setIcon(KIcon(KIconLoader::global()->loadIcon("document-save",KIconLoader::Panel)));
    dialog->addPage(new RSOption(dialog), i18n("Search & Replace") )->setIcon(KIcon(KIconLoader::global()->loadIcon("system-search",KIconLoader::Panel)));
    Texel::FilterOption* fo=new Texel::FilterOption(dialog);
    dialog->addPage(fo, i18n("String filter"))->setIcon(KIcon(KIconLoader::global()->loadIcon("view-filter",KIconLoader::Panel)));
    dialog->addPage(new FileOptions(dialog), i18n("File locations"))->setIcon(KIcon(KIconLoader::global()->loadIcon("drive-harddisk",KIconLoader::Panel)));
    dialog->addPage(new AppearanceOption(dialog), i18n("Appearance"))->setIcon(KIcon(KIconLoader::global()->loadIcon("fill-color",KIconLoader::Panel)));
    dialog->addPage(new ExternalResources(dialog), i18n("External databases"))->setIcon(KIcon(KIconLoader::global()->loadIcon("server-database",KIconLoader::Panel)));
    dialog->addPage(new BHOptions(dialog), i18n("Behaviour"))->setIcon(KIcon(KIconLoader::global()->loadIcon("start-here",KIconLoader::Panel)));
    dialog->addPage(new GlobalVars(dialog), i18n("Global variables"))->setIcon(KIcon(KIconLoader::global()->loadIcon("code-variable",KIconLoader::Panel)));
    dialog->addPage(new CoverOptions(dialog), i18n("Covers"))->setIcon(KIcon(KIconLoader::global()->loadIcon("search-cover",KIconLoader::Panel)));

    dialog->show();
}

void texel::onModelRowChange (const QModelIndex& , int , int )
{
    currentItemEdit->setEnabled(tableWidget->model()->rowCount()!=0);
}

void texel::closeEvent (QCloseEvent* event)
{
    if (m_CoreModel->isDirty() && Preferences::instance()->behaviour.m_CheckDirty) {
        int result=KMessageBox::questionYesNo(0L, i18n("There are unsaved changes. Do you really want to quit?"), i18n("Really quit?"));
        if (result==KMessageBox::No) {
            event->ignore();
            return;
        }
    }
    event->accept();
}

void texel::addBackgroundJob(KJob* job)
{
    m_JobTracker->registerJob(job);
    showProgress=true;
}

void texel::addDir (const KUrl& dir)
{
    QFileInfo info(dir.path());
    if (info.isDir()) {
        QList<QUrl> dirs;
        dirs << dir;
        if (m_CoreModel->load(dirs)) {
            m_RecentDirs->addUrl(dir);
        }
    }
}

void texel::save ()
{
    m_CoreModel->updateChanged(false);
}

void texel::saveAndRemove ()
{
    m_CoreModel->updateChanged(true);
}

void texel::removeSelectedFiles ()
{
    TagItem::List list=tableWidget->getSelectedRowsSorted();
    TagModelProxy* _m = dynamic_cast<TagModelProxy*>(tableWidget->model());
    if (list.size()>0 && _m!=0l) {
        QModelIndex current = tableWidget->currentIndex();
        _m->removeFiles(list);
        int currentRow = current.row();
        if (currentRow<_m->rowCount()) {
            currentRow=_m->rowCount()-1;
        }
        if (_m->isHeader(currentRow)) {
            currentRow++;
        }
        tableWidget->setCurrentIndex(_m->index(currentRow,current.column()));
    }
}

void texel::removeSelectedRows ()
{
    TagItem::List toRemove=tableWidget->getSelectedRowsSorted(true);

    if (toRemove.size()>0) {
        TagModelProxy* _m = dynamic_cast<TagModelProxy*>(tableWidget->model());
        QModelIndex current=tableWidget->currentIndex();
        int currentRow = current.row();
        m_CoreModel->removeRows(toRemove);
        if (currentRow>=_m->rowCount()) {
            currentRow=_m->rowCount()-1;
        }
        if (_m->isHeader(currentRow)) {
            ++currentRow;
        }
        tableWidget->setCurrentIndex(_m->index(currentRow,current.column()));
    }
    else {
        tableWidget->clearSelectedCells();
    }
}

void texel::openByDir ()
{
    QString dir=FileDialog::getExistingDirectory();
    if (!dir.isEmpty()) {
        addDir (dir);
    }
}

void texel::fillDown ()
{
    QItemSelectionModel* sel = tableWidget->selectionModel();
    if (sel->hasSelection()) {
        dynamic_cast<TagModelProxy*>(tableWidget->model())->fillDownMoreThanOneRow(*sel);
    }
}

void texel::copy ()
{
    QClipboard* clipboard = QApplication::clipboard();
    const QItemSelectionModel* sel = tableWidget->selectionModel();
    const QAbstractItemModel* model = tableWidget->model();

    if (sel->hasSelection()) {
        //
        // internal
        QByteArray rawBytes;
        QDataStream serializer(&rawBytes, QIODevice::WriteOnly);

        // Start with offset
        int mostLeft=model->columnCount();
        int mostTop=model->rowCount();
        foreach (QItemSelectionRange range, sel->selection()) {
            if (mostLeft>range.left()) mostLeft=range.left();
            if (mostTop>range.top()) mostTop=range.top();
        }
        serializer << mostLeft << mostTop;

        // All blocks
        foreach (QItemSelectionRange range, sel->selection()) {
            serializer << range.left() << range.top() << range.right() << range.bottom();
            for (int x=range.top();x<=range.bottom();x++) {
                for (int y=range.left();y<=range.right();y++) {
                    serializer << model->data(model->index(x,y));
                }
            }
        }

        //
        // text/plain: First block, columns tab separated
        // text/csv: First block, comma separated
        QString plainItems;
        QString csvItems;
        bool hasContent=false;
        QItemSelectionRange range = sel->selection().first();

        for (int x=range.top();x<=range.bottom();x++) {
            for (int y=range.left();y<=range.right();y++) {
                const QString s = model->data(model->index(x,y)).toString();
                hasContent |= !s.isEmpty(); // hasContent is true, when at least one cell has non empty content
                plainItems+=s;
                csvItems+=s;
                if (y!=range.right()) {
                    plainItems+="\t";
                    csvItems+=",";
                }
            }
            plainItems+="\n";
            csvItems+="\r\n";
        }

        if (hasContent) {
            QMimeData* mime=new QMimeData ();
            mime->setData ("text/csv",csvItems.toUtf8());
            mime->setData ("text/plain",plainItems.toUtf8());
            mime->setData ("application/texel-cells",rawBytes);
            clipboard->setMimeData(mime);
        }
    }
    else {
        const QString s = model->data(tableWidget->currentIndex()).toString();
        if (!s.isEmpty()) {
            clipboard->setText(s);
        }
    }
}

void texel::paste ()
{
    const QClipboard *clipboard = QApplication::clipboard();
    const QMimeData* mime=clipboard->mimeData();
    const QModelIndex& current = tableWidget->currentIndex();
    TagModelProxy* model=dynamic_cast<TagModelProxy*>(tableWidget->model());

    if (!current.isValid()) {
        // give up when not clear where to insert
        return;
    }

    // Define the range which should be pasted
    QModelIndex from;
    QModelIndex to;
    QVariantList values;
    const QString description = i18nc("Redo command","Paste");

    if (mime->hasFormat("application/texel-cells")) {
        const QByteArray rawBytes = mime->data("application/texel-cells");
        QDataStream deserializer(rawBytes);
        int offsetLeft, offsetTop;
        deserializer >> offsetLeft >> offsetTop;
        offsetLeft=current.column()-offsetLeft;
        offsetTop=current.row()-offsetTop;
        while (!deserializer.atEnd()) {
            int left,right,top,bottom;
            deserializer >> left >> top >> right >> bottom;
            values.clear();
            for (int x=top;x<=bottom;x++) {
                for (int y=left;y<=right;y++) {
                    QVariant data;
                    deserializer >> data;
                    values << data;
                }
            }
            from=model->index(top+offsetTop,left+offsetLeft);
            to=model->index(bottom+offsetTop,right+offsetLeft);
            model->paste (from,to,values,description);
        }
    }
    else if (mime->hasFormat("text/csv")) {
        from=tableWidget->currentIndex();
        QStringList lines = QString::fromUtf8(mime->data("text/csv")).split(QRegExp("\r?\n"));

        // first detect range size
        int max=0;
        foreach (QString line, lines) {
            const int c=line.count(',');
            if (c>max) max=c;
        }
        // max is the number of columns -1 as only separators were counted
        to=dynamic_cast<TagModelProxy*>(tableWidget->model())->index(from.row()+lines.size()-1,from.column()+max);

        for (int x=0;x<lines.size();x++) {
            if (!lines[x].isEmpty()) {
                QStringList cells = lines[x].split(',');
                for (int y=0;y<cells.size();y++) {
                    values << QVariant(cells[y]);
                }
                // fill up range if different column range are used
                for (int y=cells.size();y<=max;y++) values << QVariant();
            }
        }

        model->paste(from,to,values,description);
    }
    // currently more or less a copy from text/csv - I'm not yet sure if this will last
    else if (mime->hasFormat("text/plain")) {
        from=tableWidget->currentIndex();
        QString all=QString::fromUtf8(mime->data("text/plain"));
        QStringList lines = all.split(QRegExp("\r?\n"));
        while (lines.last().isEmpty()) lines.removeLast();

        // autodetect separator. Idea is that these signs are typically not used in
        // meta tags, therefore when at least one has the same number of occurences as
        // number of lines. Otherwise no sperator is defined.
        QList<QChar> possibleSeparators;
        possibleSeparators << ',' << ':' << ';' << '\t';
        QChar sep=0;
        int count=lines.size();
        foreach (QChar ch, possibleSeparators ) {
            int c=all.count(ch);
            if (c>=count) {
                count=c;
                sep=ch;
            }
        }

        // detect range size
        int max=0;
        foreach (QString line, lines) {
            const int c=line.count(sep);
            if (c>max) max=c;
        }
        // max is the number of columns -1 as only separators were counted
        to=model->index(from.row()+lines.size()-1,from.column()+max);

        for (int x=0;x<lines.size();x++) {
            QStringList cells = lines[x].split('\t');
            for (int y=0;y<cells.size();y++) {
                values << QVariant(cells[y]);
            }
            // fill up range if different column range are used
            for (int y=cells.size();y<=max;y++) values << QVariant();
        }

        model->paste(from,to,values,description);
    }
    else {
        model->paste (tableWidget->currentIndex(), tableWidget->currentIndex(),
                      values  << clipboard->text(), description);
        kWarning() << "Unknown MIME format from paste: " << endl;
        foreach (QString f, mime->formats()) kWarning() << "MIME: " << f << endl;
    }
}

void texel::cut ()
{
    QClipboard *clipboard = QApplication::clipboard();
    clipboard->setText(tableWidget->model()->data(tableWidget->currentIndex()).toString());
    dynamic_cast<TagModelProxy*>(tableWidget->model())->setData(tableWidget->currentIndex(),QString(""));
}

void texel::search ()
{
    searchWidget->show();
}

void texel::writeList ()
{
    if (m_CoreModel->getListFilename().isEmpty()) {
        const QString filename = FileDialog::getSaveFileName("*.txl",i18n("Write list"));
        writeList (filename);
    }

    writeList(m_CoreModel->getListFilename());
}

void texel::writeList (const QString& filename)
{
    if (filename.isEmpty()) return;

    //m_CoreModel->setListFilename (filename);
    m_RecentLists->addUrl (KUrl::fromPath(filename));

    ArchiveSave* job = new ArchiveSave(filename, *m_CoreModel);
    addBackgroundJob(job);
    connect (job, SIGNAL(errorOccured(QString,QString)), this, SLOT(onSaveError(QString,QString)));
    job->exec();
}

void texel::writeListAs ()
{
    const QString filename = FileDialog::getSaveFileName("*.txl",i18n("Write list"));
    m_CoreModel->setWrittenToList(false); // Was not written to this file
    writeList (filename);
}

QAction* texel::mostRecent (KRecentFilesAction* recent)
{
    QList<QAction*> list=recent->actions();
    if (list.size()>0) {
        return list.back();
    }
    return 0L;
}

void texel::loadMostRecentDir ()
{
    QAction* action = mostRecent (m_RecentDirs);
    if (action!=0L) action->trigger();
}

void texel::loadMostRecentList ()
{
    QAction* action = mostRecent (m_RecentLists);
    if (action!=0L) action->trigger();
}

void texel::onSaveError (const QString& filename, const QString& what)
{
    const QString errorText = i18n("Could not write %1.",filename);
    KMessageBox::detailedError(this, errorText, what, i18n("Couldn't save into %1",filename));
}

void texel::onLoadError (const QString& filename, const QString& what)
{
    const QString errorText = i18n("File <%1> has invalid contents. Could not load it.",filename);
    KMessageBox::detailedError(this, errorText, what, i18n("Couldn't load %1",filename));
}

void texel::loadList (const KUrl& url)
{
    const bool empty = m_CoreModel->rowCount()==0;

    ArchiveLoad* job = new ArchiveLoad(url.path(),*m_CoreModel);
    connect (job, SIGNAL(saveListStatisticHint(QString,bool,int)), m_CoreModel, SLOT(updateSaveListStatistics(QString,bool,int)));
    connect (job, SIGNAL(newTags(TagItem::List)), m_CoreModel, SLOT(appendList(TagItem::List)));
    connect (job, SIGNAL(newPlaylist(M3U::List)), m_CoreModel, SLOT(addPlaylist(M3U::List)));
    connect (job, SIGNAL(errorOccured(QString,QString)), this, SLOT(onLoadError(QString,QString)));
    connect (job, SIGNAL(addDuplicateRelation(int,QList<int>)), m_CoreModel, SLOT(addDuplicateRelation(int,QList<int>)));
    addBackgroundJob(job);
    job->start();
    m_RecentLists->addUrl (url);

    if (empty) m_CoreModel->setWrittenToList(true);
}

void texel::onEmptyModel (bool empty)
{
    if (empty) {
        if (Preferences::instance()->behaviour.m_ShowWelcomeScreen) {
            stackedWidget->setCurrentIndex(1);
        }
        // Reset number of open items to zero
        updateUnsavedNumber(-numModelDirty);
    }
    else {
        stackedWidget->setCurrentIndex(0);
    }
}

void texel::checkEmptyModel ()
{
    if (tableWidget->model()->rowCount()==0) {
        onEmptyModel(true);
    }
}

void texel::loadList ()
{
    const QString filename = FileDialog::getLoadFileName("*.txl",i18n("Load list"));
    if (!filename.isEmpty()) {
        //m_CoreModel->setListFilename (filename);
        loadList (KUrl::fromPath(filename));
    }
}

void texel::duplicatedetection ()
{
    TagModelProxy* m=dynamic_cast<TagModelProxy*>(tableWidget->model());
    QList<TagItem*> selection;
    if (tableWidget->selectionModel()->hasSelection()) {
        QModelIndexList sel = tableWidget->selectionModel()->selectedRows();
        foreach (QModelIndex i,sel) {
            selection << m->getRow(i.row());
        }
    }
    KDialog* d=new Texel::DuplicateDetection(m,selection);
    d->exec();
}

void texel::addToPhrases()
{
    const QItemSelectionModel* sel = tableWidget->selectionModel();
    if (sel->hasSelection()) {
        const QModelIndexList list = sel->selectedIndexes();
        m_PhraseView->model()->addRow(tableWidget->model(),list);
        m_PhraseView->show();
    }
}

void texel::phraseTriggered (Texel::PhraseItem* phrase)
{
    TagItem::List list=tableWidget->getSelectedRowsSorted();
    // When the list is empty no cmplete rows are selected. However, the phrase can still
    // be executed on the current row
    if (list.isEmpty()) {
        list << dynamic_cast<TagModelProxy*>(tableWidget->model())->getRow(tableWidget->currentIndex().row());
    }
    foreach (TagItem* item, list) {
        phrase->execute(item);
    }
}

void texel::filterSearchText ()
{
    const QString expression = m_FilterEdit->text();
    if (expression.isEmpty()) {
        static_cast<Texel::TagModelProxy*>(tableWidget->model())->resetFilter();
    }
    else {
        static_cast<Texel::TagModelProxy*>(tableWidget->model())->filter(expression);
        m_FilterEdit->addCompletionString(expression);
    }
}

void texel::setFilterType ()
{
    QAction* action = dynamic_cast<QAction*>(sender());
    if (action==0l) {
        kWarning() << "Unallowed call. Sender was not a QAction: " << sender()->metaObject()->className() << endl;
    }
    const Texel::TagModelProxy::ExpressionFilterFlag num = (Texel::TagModelProxy::ExpressionFilterFlag)action->data().toInt();
    const Texel::TagModelProxy::ExpressionFilterFlags mode = static_cast<Texel::TagModelProxy*>(tableWidget->model())->getExpressionModeFlags();
    static_cast<Texel::TagModelProxy*>(tableWidget->model())->setExpressionModeFlags(mode|num);
}

void texel::updateNumberOfItems () const
{
    const int s1 = m_CoreModel->rowCount();
    const int s2 = tableWidget->model()->rowCount();
    m_NoOfItems->setText(QString::number(s2)+"/"+QString::number(s1));
}

void texel::updateUnsavedNumber (int delta)
{
    numModelDirty+=delta;
    if (numModelDirty<0) numModelDirty=0;
    m_TagCurrentDirtyDisplay->setText(QString::number(numModelDirty));
}

void texel::updateType (const QModelIndex& current, const QModelIndex& previous)
{
    if (current.isValid() && current.row()!=previous.row()) {
        const int row = current.row();
        TagModelProxy* m = dynamic_cast<TagModelProxy*>(tableWidget->model());
        if (m!=0l) {
            TagItem* item=m->getRow(row);
            if (item!=0l) {
                m_TagTypeDisplay->setText(item->typeName());
            }
        }
    }
}

void texel::updateCurrentRow (const QModelIndex& current, const QModelIndex& previous)
{
    if (current.isValid() && current.row()!=previous.row()) {
        const int row = current.row();
        m_TagCurrentRowDisplay->setText(QString::number(row));
    }
}

void texel::editFormular ()
{
    TagModelProxy* m = dynamic_cast<TagModelProxy*>(tableWidget->model());
    const int row = tableWidget->currentIndex().row();
    FormularEditor* editor = new FormularEditor(m->getRow(row));
    editor->show();
}

void texel::saveDocks ()
{
    QStringList docks;
    foreach (QAction* a,menuDocks->actions()) {
        if (a->isChecked()) {
            docks << a->objectName();
        }
    }
    KConfigGroup texelConfig(KGlobal::config(),"Texel view settings");
    texelConfig.writeEntry("visibleDocks",docks);
}

void texel::restoreDocks ()
{
    if (Preferences::instance()->behaviour.m_RestoreDocksOnStartup) {
        const KConfigGroup texelConfig(KGlobal::config(),"Texel view settings");
        const QStringList docks=texelConfig.readEntry("visibleDocks",QStringList());
        foreach (QAction* a, menuDocks->actions()) {
            if (docks.contains(a->objectName()) && !a->isChecked()) {
                a->setChecked(true);
            }
        }
    }
}

void texel::rotateDetailsDock ()
{
    if (!aTrackDetails->isChecked()) {
        aTrackDetails->toggle();
    }

    if (m_TrackDetailsDock!=0l) {
        m_TrackDetailsDock->rotatePane();
    }
}

void texel::addDetailsDock (bool _show)
{
    if (_show) {
        m_TrackDetailsDock = new TrackDetails(this);

        if (!restoreDockWidget (m_TrackDetailsDock)) {
            addDockWidget(Qt::RightDockWidgetArea,m_TrackDetailsDock);
        }

        connect (tableWidget, SIGNAL(currentIndexChanged(QModelIndex,QModelIndex)),
                 m_TrackDetailsDock, SLOT(entered(QModelIndex)));
        connect (tableWidget->model(), SIGNAL(dataChanged(QModelIndex,QModelIndex)),
                 m_TrackDetailsDock, SLOT(changed(QModelIndex,QModelIndex)));
        connect (dynamic_cast<TagModelProxy*>(tableWidget->model()), SIGNAL(alteredExtension(int,int,QString,QVariant)),
                 m_TrackDetailsDock, SLOT(onAlteredExtension(int,int,QString,QVariant)));

        QAction* s = dynamic_cast<QAction*>(sender());
        if (s!=0l) {
            connect (m_TrackDetailsDock, SIGNAL(close()), s, SLOT(toggle()));
            connect (s, SIGNAL(toggled(bool)), m_TrackDetailsDock, SLOT(deleteThisOnFalse(bool)));
        }
        else {
            kWarning() << "Unallowed call. Sender was not a QAction: " << sender()->metaObject()->className() << endl;
        }

        m_TrackDetailsDock->entered (tableWidget->currentIndex());
        m_TrackDetailsDock->show();
    }
    else {
        m_TrackDetailsDock=0l;
    }
}

void texel::addPlayerDock (bool _show)
{
    if (_show) {
        PlayerDock* dock = new PlayerDock(this);
        if (!restoreDockWidget (dock)) {
            addDockWidget(Qt::LeftDockWidgetArea,dock);
        }

        dock->show();

        connect (dock, SIGNAL(close()), aPlayerDock, SLOT(toggle()));
        connect (aPlayerDock, SIGNAL(toggled(bool)), dock, SLOT(deleteThisOnFalse(bool)));
    }
}

void texel::addLogDock (bool _show)
{
    if (_show) {
        LogDock* dock = new LogDock(this, m_MessageModel);
        if (!restoreDockWidget (dock)) {
            //addDockWidget(Qt::NoDockWidgetArea,dock);
            //dock->setFloating(true);
        }
        dock->show();

        connect (dock, SIGNAL(close()), aLoggerDock, SLOT(toggle()));
        connect (aLoggerDock, SIGNAL(toggled(bool)), dock, SLOT(deleteThisOnFalse(bool)));
    }
}

void texel::addDuplicateDock(bool _show)
{
    if (_show) {
        DuplicateHelper* dock = new DuplicateHelper(this);
        if (!restoreDockWidget (dock)) {
            addDockWidget(Qt::RightDockWidgetArea,dock);
        }

        connect (tableWidget, SIGNAL(currentIndexChanged(QModelIndex,QModelIndex)), dock, SLOT(entered(QModelIndex,QModelIndex)));
        connect (tableWidget->model(), SIGNAL(dataChanged(QModelIndex,QModelIndex)), dock, SLOT(changed(QModelIndex,QModelIndex)));
        connect (dock, SIGNAL(playSong(QString,QString)), m_player, SLOT(playSong(QString,QString)));

        QAction* s = dynamic_cast<QAction*>(sender());
        if (s!=0l) {
            connect (dock, SIGNAL(close()), s, SLOT(toggle()));
            connect (s, SIGNAL(toggled(bool)), dock, SLOT(deleteThisOnFalse(bool)));
        }
        else {
            kWarning() << "Unallowed call. Sender was not a QAction: " << sender()->metaObject()->className() << endl;
        }

        dock->entered(tableWidget->currentIndex(),QModelIndex());
        dock->show();
    }
}

void texel::addDirectoryDock (bool _show)
{
    if (_show) {
        DirectoryDock* dock = new DirectoryDock(this);
        if (!restoreDockWidget (dock)) {
            addDockWidget(Qt::LeftDockWidgetArea,dock);
        }

        dock->setDirTree (m_CoreModel->getDirCache());
        dock->setProxyModel(dynamic_cast<TagModelProxy*>(tableWidget->model()));
        connect (dock, SIGNAL (loadDir(KUrl)), this, SLOT(addDir(KUrl)));

        QAction* s = dynamic_cast<QAction*>(sender());
        if (s!=0l) {
            connect (dock, SIGNAL(close()), aDirectoryView, SLOT(toggle()));
            connect (aDirectoryView, SIGNAL(toggled(bool)), dock, SLOT(deleteThisOnFalse(bool)));
        }
        else {
            kWarning() << "Unallowed call. Sender was not a QAction: " << sender()->metaObject()->className() << endl;
        }
        dock->show();
    }
}

void texel::addPlaylistDock (bool _show)
{
    if (_show) {
        PlaylistDock* dock = new PlaylistDock(m_CoreModel->getPlayListModel(),tableWidget,this);
        dock->setModel(dynamic_cast<TagModelProxy*>(tableWidget->model()));

        if (!restoreDockWidget (dock)) {
            addDockWidget(Qt::LeftDockWidgetArea,dock);
        }

        QAction* s = dynamic_cast<QAction*>(sender());
        if (s!=0l) {
            connect (dock, SIGNAL(close()), s, SLOT(toggle()));
            connect (s, SIGNAL(toggled(bool)), dock, SLOT(deleteThisOnFalse(bool)));
        }
        else {
            kWarning() << "Unallowed call. Sender was not a QAction: " << sender()->metaObject()->className() << endl;
        }
        dock->show();
    }
}

void texel::removeStaleTags()
{
    KProgressDialog* d = new KProgressDialog(this,i18n("Remove stale tag entries"));
    d->setAllowCancel(false);
    d->setAutoClose(true);
    d->setMinimumDuration(500);
    d->progressBar()->setMinimum(0);
    d->progressBar()->setMaximum(m_CoreModel->rowCount());
    d->show();
    connect (m_CoreModel, SIGNAL(numberAlreadyProcessed(int)),d->progressBar(), SLOT(setValue(int)));
    m_CoreModel->removeStaleEntries();
    d->hide();
    d->deleteLater();
}

void texel::showAboutApplication ()
{
    KAboutApplicationDialog* dialog = new KAboutApplicationDialog(m_AboutData);
    dialog->setDetailsWidget(new Credits());
    dialog->setDetailsWidgetVisible(true);
    dialog->show();
}

void texel::maxFaultLevel (int level)
{
    m_MessageButton->setIcon(Texel::MessageModel::getIcon(level));
}


}

#include "texel.moc"
