/*************************************************************************
*************************************************************************/

#ifndef texel_texel_H
#define texel_texel_H

#include <KMainWindow>

#include "ui_main.h"
#include "func/player/player.h"
#include "view/welcome.h"
#include "model/texeltagmodel.h"
#include "view/messageview.h"
#include "view/texeltablewidget.h"

class KRecentFilesAction;
class KUrl;
class KAction;
class KActionCollection;
class KAboutData;
class KLineEdit;

class QLabel;
class QCloseEvent;

namespace Texel
{
    class MessageModel;
    class PhraseView;
    class PhraseModel;
    class PhraseItem;
    class JobTracker;
    class CompletionLineEditor;
    class TrackDetails;
    class TagModelProxy;
    class TagModel;
    class Welcome;

    namespace Player {
    class Player;
    }

    /** Core object which manages the complete functionality of Texel. It cares for all overarching
  * commands and actions which are defined in the menus and toolbars. In most cases the implementation
  * in this class acts as a thin wrapper to delegate the requested command to the TagModel model
  * or specialized classes.
  */
    class texel : public KMainWindow, private Ui::TexelMainWindow
    {
        Q_OBJECT

    public:
        texel(QWidget* parent=0);
        virtual ~texel();

        void checkEmptyModel ();

        void setAboutData (KAboutData* data);

    protected:
        void closeEvent (QCloseEvent* event);

    signals:
        void test (int,const QString& artist);

    public slots:
        void onModelRowChange (const QModelIndex & parent, int start, int end);
        void loadList (const KUrl& url);
        void addDir (const KUrl& dir);

        void addBackgroundJob (KJob* job);

    protected slots:
        virtual void showAboutApplication ();

    private slots:
        void onLoadError (const QString& archive, const QString& what);
        void onSaveError (const QString& filename, const QString& what);

        void editFormular ();

        void updateType (const QModelIndex& current, const QModelIndex& previous);
        void updateCurrentRow (const QModelIndex& current, const QModelIndex& previous);
        void updateUnsavedNumber (int delta);

        //
        // Commands
        //

        /** Called when the fill down command is called. This takes the most upper
      * cell of the selected column and copies it into the other selected cells.
      */
        void fillDown ();

        void copy ();
        void cut ();
        void paste ();
        void search ();
        void writeList ();
        void writeListAs ();
        void loadList ();
        void addDetailsDock (bool show);
        void addDuplicateDock (bool show);
        void addDirectoryDock  (bool show);
        void addPlaylistDock (bool show);
        void addPlayerDock (bool show);
        void addLogDock (bool show=true);

        void addToPhrases ();
        void phraseTriggered (PhraseItem* phrase);

        void filterSearchText();
        void setFilterType ();
        void maxFaultLevel (int level);

        /** Removes all selected files from disc.
      * The selection is reduced to just rows (and all rows are belonging to the selection
      * for which at least one cell is selected). When no cells are selected this command
      * does nothing.
      */
        void removeSelectedFiles();

        /** Remove all selected rows from the model. Files on disc will not be touched.
      * The selection is reduced to just rows (and all rows are belonging to the selection
      * for which at least one cell is selected). When no cells are selected this command
      * does nothing.
      */
        void removeSelectedRows();

        /** Stores all altered files to disc and remove them from the mode.
      */
        void saveAndRemove();
        void save ();

        void showOptions();
        void openByDir ();
        void loadMostRecentDir();
        void loadMostRecentList();
        void duplicatedetection ();

        void removeStaleTags ();

        //
        // State changes
        //
        void captionProgress (unsigned progress);

        /** Is called when the core @class TagModel model contains an unsaved change or after the
      * complete model change was saved. Important to show a modify flag in the caption
      */
        void changedDirtyState (bool dirty);

        /** Is called when the show/hide state of the progress bar toggles. Important to be able to
      * show the progress also in the caption of the window. If the hide would not be capture than
      * the caption would show a 100% copletion until an update due to another change would be done.
      */
        void removeCaptionProgress ();
        void onEmptyModel(bool);

        void sortBy ();
        void sortedBy (int column, Qt::SortOrder order);
        void showFormulars (bool show);

        void updateNumberOfItems () const;

        void rotateDetailsDock ();

    private:
        void createFilterToolbar (Texel::TagModelProxy* tagmodel);

        Texel::TagModel* tableModel () const;

        void saveDocks ();
        void restoreDocks ();

        /** Updates the caption of the main window. Takes care of model modification and show the
      * percentage of completion for background processes. To do that the internal progressbar
      * is mirrored to the caption when its visible.
      */
        void updateCaption();

        void writeList (const QString& filename);
        QAction* mostRecent (KRecentFilesAction* recent);

        KRecentFilesAction* m_RecentDirs;
        KRecentFilesAction* m_RecentLists;
        Texel::Player::Player* m_player;

        /** When no data is loaded texel displayes a welcome screen instead of an empty spread sheet
        The welcome screen allows to quickly access the latest directories worked with.
      */
        Texel::Welcome* m_WelcomeScreen;
        KActionCollection* m_ActionCollection;
        Texel::PhraseView* m_PhraseView;
        Texel::PhraseModel* m_PhraseModel;

        KAction* stdActionQuit;
        KAction* stdActionOptions;
        KAction* stdActionSearch;
        KAction* stdActionReplace;

        bool showProgress;
        unsigned progress;
        bool modelDirty;
        int numModelDirty;

        Texel::MessageModel* m_MessageModel;
        Texel::TagModel* m_CoreModel;
        Texel::CompletionLineEditor* m_FilterEdit;
        Texel::TrackDetails* m_TrackDetailsDock;

        QLabel* m_NoOfItems;
        QLabel* m_TagTypeDisplay;
        QLabel* m_TagCurrentRowDisplay;
        QLabel* m_TagCurrentDirtyDisplay;

        QPushButton* m_MessageButton;

        KAboutData* m_AboutData;

        JobTracker* m_JobTracker;
    };

    inline Texel::TagModel* texel::tableModel() const
    {
        return dynamic_cast<Texel::TagModel*>(tableWidget->model());
    }

    inline void texel::setAboutData (KAboutData* data)
    {
        m_AboutData=data;
    }

}

#endif // texel_texel_H
