/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#ifndef DUPLICATEMAP_H
#define DUPLICATEMAP_H

#include <boost/bimap/bimap.hpp>
#include <boost/bimap/list_of.hpp>
#include <boost/bimap/multiset_of.hpp>
#include <boost/bimap/set_of.hpp>

#include "aid.h"

namespace Texel {

namespace DuplicateMap {

/** @brief Class to manage duplicates in a certain set efficiently. For a set of items
  duplicates of a certain attribute are searched. In case of duplicates the corresponding
  items are notified by using the functot provided in the methods.

  @param Attribute attribute which will be searched for duplicates.
  @param Item items to store (class just uses pointer and does not take ownership
  of any item).
  */
template <class Attribute, class Item>
class Detector
{
    typedef boost::bimaps::bimap<boost::bimaps::multiset_of<Attribute>,
                                 boost::bimaps::set_of<Item*>,
                                 boost::bimaps::list_of_relation> Bimap;

    typedef typename Bimap::right_const_iterator item_const_iterator;
    typedef typename Bimap::left_const_iterator attrib_const_iterator;
    typedef typename Bimap::value_type value_type;

    typedef typename Bimap::right_map item_map;
    typedef typename Bimap::left_map attrib_map;

    // Using a member (instead of inheritance) seems much clearer with the typedefs
    Bimap data;

protected:
    inline item_map byItem () const
    {
        return data.right;
    }

    inline attrib_map byAttrib () const
    {
        return data.left;
    }

public:

    /** Removes the item from the duplicate list. When this removes a duplicate situation
      the left over item is cleared by callig the functor notify. The item itself is not
      deleted.
      @param item to remove from the search structure. When it is not padrt of the structure it
      is ignored.
      @param notify Functor which is called when the duplicate state of some item changes (will not
      be called for the item to remove). The functor must provide the
      operator() (Item* item, bool duplicate) where the boolean parameter is true when a duplicate
      situation exists.
      */
    template <typename Functor>
    inline void remove (Item* item, Functor notify)
    {
        item_const_iterator toErase=data.right.find(item);
        if (toErase!=byItem().end()) {
            const Attribute oldAtt = toErase->second;
            byItem().erase(item);
            const bool problem=byAttrib().count(oldAtt)>1;
            attrib_const_iterator is = byAttrib().find(oldAtt);
            const int count = byAttrib().count(oldAtt);
            while (is!=byAttrib().end() && is->first==oldAtt) {
                notify(is->second,problem,count);
                ++is;
            }
        }
    }

    /**
      @brief alters the attribute of item to newAtt. In case of changing duplicate states the
      functor notify is called.
      @param item item for which the attribute shall be altered.
      @param newAtt new attribute after altering
      @param notify Functor which is called when the duplicate state of some item changes
      The functor must provide the operator() (Item* item, bool duplicate) where the boolean
      parameter is true when a duplicate situation exists.
      */
    template <typename Functor>
    inline void alter (Item* item, const Attribute& newAtt, Functor notify)
    {
        remove (item, notify);
        data.push_back(value_type(newAtt,item));

        attrib_const_iterator is = byAttrib().find(newAtt);
        const int count=byAttrib().count(newAtt);
        if (is!=byAttrib().end()) {
            while (is!=byAttrib().end() && is->first==newAtt) {
                notify(is->second,count>1,count);
                ++is;
            }
        }
    }

    inline void clear ()
    {
        data.clear();
    }

    inline bool contains (const Attribute& att) const
    {
        return byAttrib().count(att)>0;
    }

    inline Item* operator[] (const Attribute& att) const
    {
        attrib_const_iterator iter=byAttrib().find(att);
        if (iter!=byAttrib().end()) return iter->second;
        return 0l;
    }

    inline QList<Item*> getAll (const Attribute& att) const
    {
        QList<Item*> r;
        attrib_const_iterator iter=byAttrib().find(att);
        while (iter->first==att) {
            r << iter->second;
            iter++;
        }
        return r;
    }
};

}

}
#endif // DUPLICATEMAP_H
