/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "persistentcache.h"
#include "options/preferences.h"
#include "util/loghook.h"

#include <QMutexLocker>

#include <klocale.h>
#include <kdebug.h>

#include <auto_ptr.h>

#include "aid.h"

namespace Texel {

void PersistentCache::insert (int type, const QString& key, const QByteArray& content) throw (SqLite::Exception)
{
    QMutexLocker lock(&m_Mutex);

    m_MemoryCache.insert(Key(type,key),new QByteArray(content),content.size());

    m_CacheDB->beginTransaction();
    m_InsertStatement->bind()(type)(key)(content.size())(content).execute();
    ++m_Inserts;
    if (m_Inserts%100) {
        while (getCachedSize()>Preferences::instance()->cache.m_MaxFileSize*(2<<20) /* in MB */) {
            m_DeleteStatement->execute();
        }
    }
    m_CacheDB->commit();
}

QByteArray PersistentCache::lookup (int type, const QString& key) throw (SqLite::Exception)
{
    QMutexLocker lock(&m_Mutex);

    if (m_MemoryCache.contains(Key(type,key))) {
        return *m_MemoryCache.object(Key(type,key));
    }

    m_CacheDB->beginTransaction();
    m_LookupStatement->bind()(type)(key).execute(false);

    QByteArray result;
    if (m_LookupStatement->hasRow()) {
        result=m_LookupStatement->getBlob(0);
        m_UpdateLastAccessStatement->bind()(type)(key).execute();
    }

    m_LookupStatement->reset();
    m_CacheDB->commit();
    return result;
}

int PersistentCache::getCachedSize() throw (SqLite::Exception)
{
    m_CalcSize->execute();
    if (m_CalcSize->hasRow()) {
        return m_CalcSize->getInt(0);
    }
    return 0;
}

PersistentCache::PersistentCache () throw (SqLite::Exception)
    : m_Inserts(0),
      m_CacheDB(new SqLite::Connection(Preferences::instance()->cache.m_CacheFile)),
      m_MemoryCache(5*(1<<20) /* Max size: 5 MB */),
      m_Mutex()
{
    QMutexLocker lock(&m_Mutex);
    LogHook::instance()->message (i18n("Opened cache file %1",Preferences::instance()->cache.m_CacheFile),0l);
    createStructure();
    m_InsertStatement=m_CacheDB->prepare("REPLACE INTO entries VALUES (?,?,datetime('now'),?,?);");
    m_LookupStatement=m_CacheDB->prepare("SELECT content FROM entries WHERE type=? and key=?;");
    m_DeleteStatement=m_CacheDB->prepare("DELETE FROM entries WHERE last_access=(SELECT min(last_access) FROM entries);");
    m_UpdateLastAccessStatement=m_CacheDB->prepare("UPDATE entries SET last_access=datetime('now') WHERE type=? AND key=?;");
    m_CalcSize=m_CacheDB->prepare("SELECT total(size) FROM entries;");
}

void PersistentCache::compress() throw (SqLite::Exception)
{
    QMutexLocker lock(&m_Mutex);
    m_CacheDB->beginTransaction();
    std::auto_ptr<SqLite::Statement> clean(m_CacheDB->prepare("DELETE FROM entries WHERE last_access<datetime('now',?);"));
    clean->bind(1,QString("-%1 days").arg(Preferences::instance()->cache.m_OldestEntriesInDays));
    clean->execute();
    m_CacheDB->commit();
    m_CacheDB->prepare("VACUUM;")->executeOnce();
}

void PersistentCache::createStructure () throw (SqLite::Exception)
{
    m_CacheDB->beginTransaction();
    std::auto_ptr<SqLite::Statement> header(m_CacheDB->prepare("SELECT 1 FROM sqlite_master WHERE type='table' AND name='entries';")->execute());
    if (!header->hasRow()) {
        m_CacheDB->prepare ("CREATE TABLE entries ("
                          "type INTEGER,"
                           "key TEXT,"
                           "last_access DATETIME,"
                           "size INT,"
                           "content BLOB"
                           ");"
                           )->executeOnce();

        m_CacheDB->prepare ("CREATE UNIQUE INDEX key_index ON entries(type,key);")->executeOnce();
        m_CacheDB->prepare ("CREATE INDEX last_access_index ON entries(last_access);")->executeOnce();
    }
    m_CacheDB->commit();
}

PersistentCache::~PersistentCache()
{
}

}

