/*************************************************************************
*************************************************************************/

#include "qtlevenshtein.h"

#include <QChar>
#include <QVector>
#include <QSet>
#include <QMap>

#include <stdlib.h>

#include <boost/tuple/tuple.hpp>
#include <set>

#include "aid.h"

using namespace std;

namespace Texel
{

    namespace Levenshtein
    {
        typedef QPair<QChar,Node*> BoundedNode;

        bool CharLess(const BoundedNode& a1, const BoundedNode& a2)
        {
            return a1.first<a2.first;
        }

        struct Node
        {
            QVector<BoundedNode> m_Childs;
            QVector<int> destinations;
            int n;

            Node()
                : m_Childs(),
                destinations()
            {
                n=nodes++;
            }

            ~Node ()
            {
                for (int i=0;i<m_Childs.size();i++) delete m_Childs[i].second;
                nodes--;
            }

            Node* insert (const QChar& ch)
            {
                QVector<BoundedNode>::const_iterator found=find(ch);
                if (m_Childs.isEmpty() || m_Childs.end()==found) {
                    Node* node = new Node();
                    m_Childs.append(BoundedNode(ch,node));
                    qSort(m_Childs.begin(),m_Childs.end(),CharLess);
                    return node;
                }
                return (*found).second;
            }

            bool contains (const QChar& ch) const
            {
                return find(ch)!=m_Childs.end();
            }

            QVector<BoundedNode>::const_iterator find (const QChar& ch) const
            {
                if (m_Childs.isEmpty()) return m_Childs.end();
                return qBinaryFind(m_Childs.begin(),m_Childs.end(),BoundedNode(ch,NULL),CharLess);
            }

            Node* operator[] (const QChar& ch) const
            {
                QVector<BoundedNode>::const_iterator iter=find(ch);
                return (*iter).second;
            }

            void insertDestination (int n1)
            {
                destinations.append(n1);
            }

            static int nodes;
        };

        int Node::nodes=0;

        struct State
        {
            State ()
                : distance(0), i(0), errorsInARow(0), node(0)
            {
            }

            State(int d, const Node* n, int i_)
                : distance(d),
                i(i_),
                errorsInARow(0),
                node(n)
            {
            }

            int distance;
            int i;
            int errorsInARow;
            const Node* node;
        };

        struct StateStack
        {
            QVector<State> stack;
            int p; // pointer to first empty cell

            StateStack ()
                : stack(500),
                p(0)
            {
            }

            bool isEmpty () const
            {
                return size()==0;
            }

            int size () const
            {
                return p;
            }

            State& pop ()
            {
                return stack[--p];
            }

            void push (int d, const Node* n, int i, int errs)
            {
                State& s = push();
                s.distance=d;
                s.node=n;
                s.i=i;
                s.errorsInARow=errs;
            }

            State& push ()
            {
                if (p>=stack.size()) stack.resize(stack.size()+200);
                return stack[p++];
            }
        };
    }

    using namespace Levenshtein;

    static int min (int a, int b)
    {
        return (a<b)?a:b;
    }

    int distance(const QString& source, const QString& target, int max)
    {
        // Step 1
        const int n = source.size();
        const int m = target.size();

        if (n == 0) {
            return m;
        }
        if (m == 0) {
            return n;
        }
        if (abs(n-m)>max) return max+1;

        QVector<QVector<int> > matrix;

        matrix.resize(n+1);
        for (int i = 0; i <= n; i++) {
            matrix[i].resize(m+1);
        }

        // Step 2
        for (int i = 0; i <= n; i++) {
            matrix[i][0]=i;
        }

        for (int j = 0; j <= m; j++) {
            matrix[0][j]=j;
        }

        // Step 3

        for (int i = 1; i <= n; i++) {
            // Step 4
            bool below=false;
            for (int j = 1; j <= m; j++) {
                // Step 5
                const int cost=(source[i-1]==target[j-1])?0:1;

                // Step 6
                const int above = matrix[i-1][j];
                const int left = matrix[i][j-1];
                const int diag = matrix[i-1][j-1];
                int cell = min( above + 1, min(left + 1, diag + cost));

                // Step 6A: Cover transposition, in addition to deletion,
                // insertion and substitution. This step is taken from:
                // Berghel, Hal ; Roach, David : "An Extension of Ukkonen's
                // Enhanced Dynamic Programming ASM Algorithm"
                // (http://www.acm.org/~hlb/publications/asm/asm.html)

                int trans=0;
                if (i>2 && j>2) {
                    trans=matrix[i-2][j-2]+1;
                    if (source[i-2]!=target[j-1]) trans++;
                    if (source[i-1]!=target[j-2]) trans++;
                    if (cell>trans) cell=trans;
                }

                matrix[i][j]=cell;
                if (cell<=max) below=true;
            }
            if (!below) return max+1;
        }

        // Step 7
        return matrix[n][m];
    }

    void LevenstheinDistanceTree::Item::out (std::ostream& o) const
    {
        for (int i=0;i<size();i++) if ((*this)[i]!=0) o<<(*this)[i]; else o<<"#";
    }

    LevenstheinDistanceTree::LevenstheinDistanceTree ()
        : root(new Node())
    {
    }

    LevenstheinDistanceTree::~LevenstheinDistanceTree ()
    {
        delete root;
    }

    int LevenstheinDistanceTree::nodes ()
    {
        return Node::nodes;
    }

    void LevenstheinDistanceTree::insert (int n, const Item& name)
    {
        Node* current=root;

        for (int i=0;i<name.size();i++) {
            current=current->insert(name[i]);
        }

        current->insertDestination(n);
    }

    namespace
    {
        typedef boost::tuple<const Node*,quint16,quint16,quint16> UniqueItem;

        struct Tless
        {
            bool operator() (const UniqueItem& ui1, const UniqueItem& ui2) const
            {
                if (ui1.get<0>()==ui2.get<0>()) {
                    if (ui1.get<1>()==ui2.get<1>()) {
                        if (ui1.get<2>()==ui2.get<2>()) {
                            return ui1.get<3>()<ui2.get<3>();
                        }
                        return ui1.get<2>()<ui2.get<2>();
                    }
                    return ui1.get<1>()<ui2.get<1>();
                }
                return ui1.get<0>()<ui2.get<0>();
            }
        };
    }


    QList<int> LevenstheinDistanceTree::findMinimalDistanceItems (int n, const Item& name, int max, int maxErrorsInARow) const
    {
        int maxStackSize=0;
        if (maxErrorsInARow<=0) maxErrorsInARow=max;

        StateStack stack;
        QSet<int> results;
        std::set<UniqueItem,Tless> visited;

        stack.push(0,root,0,0);

        while (!stack.isEmpty()) {
            if (stack.size()>maxStackSize) maxStackSize=stack.size();

            const State& current=stack.pop();
            const int i=current.i;
            const Node* node=current.node;
            const int penalty=current.distance;
            const int errorsInARow=current.errorsInARow;

            // Memorize which nodes where already visited on the current position
            // and skip if aleady examined (the path below is always the same)
            if (visited.count(UniqueItem(node,penalty,i,errorsInARow))==1) {
                continue;
            }
            visited.insert(UniqueItem(node,penalty,i,errorsInARow));

            if (i==name.size()) {
                if ((node->destinations.size()>1 || !node->destinations.contains(n)) && penalty<=max) {
                    foreach (int dest,node->destinations) {
                        if (dest!=n) { results << dest; }
                    }
                }
            }
            else {
                if (penalty<max && errorsInARow<maxErrorsInARow) {
                    for (QVector<BoundedNode>::const_iterator iter=node->m_Childs.begin();
                    iter!=node->m_Childs.end();iter++) {
                        const QChar entry=(*iter).first;
                        const Node* value=(*iter).second;

                        if (entry!=QChar(0) && entry!=name[i]) {
                            // alter current character
                            stack.push(penalty+1, value,i+1, errorsInARow+1);
                            // skip current character
                            stack.push(penalty+1, value,i, errorsInARow+1);
                            // insert additional character
                            stack.push(penalty+1, node,i+1, errorsInARow+1);
                        }
                        else if (entry==name[i]) {
                            stack.push(penalty,value,i+1,0);
                        }
                    }

                    // Switch two characters
                    if (i+1<name.size() && node->contains(name[i+1]) && (*node)[name[i+1]]->contains(name[i])) {
                        stack.push(penalty+1,(*(*node)[name[i+1]])[name[i]],i+2,errorsInARow+1);
                    }
                }
                else {
                    QVector<BoundedNode>::const_iterator iter = node->find(name[i]);
                    if (iter!=node->m_Childs.end()) {
                        stack.push(penalty, (*iter).second,i+1,0);
                    }
                }
            }
        }

        return results.toList();
    }

}
