#include "sqlitewrapper.h"

namespace Texel {
namespace SqLite {

void Exception::throw_exception (sqlite3* db, int errnum, const QString& desc, const QString& details_)
{
    Exception e;
    e.errno=errnum;
    if (!desc.isEmpty()) {
        e.msg=desc+"\n";
    }
    e.msg+=QString(sqlite3_errmsg(db));
    e.details=details_;
    sqlite3_exec(db,"rollback;",0l,0l,0l);
    throw e;
}

Statement::~Statement ()
{
    sqlite3_finalize(stmt);
}

void Statement::reset () throw (Exception)
{
    const int result=sqlite3_reset(stmt);
    const int result2=sqlite3_clear_bindings(stmt);
    if (result!=SQLITE_OK || result2!=SQLITE_OK) {
        Exception::throw_exception(db,result==SQLITE_OK?result2:result,"reset",sql);
    }
    m_Result=SQLITE_DONE;
}

void Statement::bind (int index, int value) throw (Exception)
{
    const int result=sqlite3_bind_int(stmt, index, value);
    if (result!= SQLITE_OK) {
        Exception::throw_exception(db,result,"bind",sql);
    }
}

void Statement::bind (int index, const QString& value) throw (Exception)
{
    const int result=sqlite3_bind_text(stmt, index, value.toUtf8(), -1, SQLITE_TRANSIENT);
    if (result!= SQLITE_OK) {
        Exception::throw_exception(db,result,"bind",sql);
    }
}

void Statement::bind (int index, const QByteArray& value) throw (Exception)
{
    const int result=sqlite3_bind_blob(stmt, index, value.data(), value.size(), SQLITE_TRANSIENT);
    if (result!= SQLITE_OK) {
        Exception::throw_exception(db,result,"bind",sql);
    }
}

Binder& Statement::bind()
{
    binder.reset();
    return binder;
}

void Statement::bind (int index) throw (Exception)
{
    const int result=sqlite3_bind_null(stmt,index);
    if (result!= SQLITE_OK) {
        Exception::throw_exception(db,result,"bind",sql);
    }
}

void Statement::executeOnce () throw (Exception)
{
    const int result = sqlite3_step(stmt);
    if (result!=SQLITE_DONE) {
        Exception::throw_exception(db,result,"executeOnce",sql);
    }
    delete this;
}

Statement* Statement::execute () throw (Exception)
{
    const int result = sqlite3_step(stmt);
    if (result!=SQLITE_DONE && result!=SQLITE_ROW) {
        Exception::throw_exception(db,result,"execute",sql);
    }
    m_Result=result;
    return this;
}

bool Statement::hasRow () const
{
    return m_Result==SQLITE_ROW;
}

int Statement::getInt (int col) const throw (Exception)
{
    if (!hasRow()) {
        Exception::throw_exception(db,SQLITE_ERROR,QString("No (more) rows returned"));
    }
    return sqlite3_column_int (stmt, col);
}

QString Statement::getString (int col) const throw (Exception)
{
    if (!hasRow()) {
        Exception::throw_exception(db,SQLITE_ERROR,QString("No (more) rows returned"));
    }
    return QString::fromUtf8(reinterpret_cast<const char*>(sqlite3_column_text(stmt,col)));
}

QByteArray Statement::getBlob (int col) const throw (Exception)
{
    if (!hasRow()) {
        Exception::throw_exception(db,SQLITE_ERROR,QString("No (more) rows returned"));
    }
    const void* p = sqlite3_column_text(stmt,col);
    const int size = sqlite3_column_bytes(stmt,col);

    if (p==0l) return QByteArray();
    return QByteArray(reinterpret_cast<const char*>(p),size);
}

bool Statement::isNull (int col) const
{
    return sqlite3_column_type(stmt,col)==SQLITE_NULL;
}

Statement::Statement (sqlite3* connection, const char* sql_)
    : db(connection),
      m_Result(SQLITE_DONE),
      binder(this),
      sql(sql_)
{
    const int result=sqlite3_prepare_v2(connection, sql_, strlen(sql_)+1, &stmt, 0l);
    if (result!=SQLITE_OK) {
        Exception::throw_exception (db,result,QString("While preparation of ")+sql_,sql);
    }
}

Binder::Binder(Statement* stmt_)
    : stmt(stmt_),
      col(0)
{
}

Binder& Binder::operator() (const QString& val)
{
    stmt->bind(++col,val);
    return *this;
}

Binder& Binder::operator() (const QByteArray& val)
{
    stmt->bind(++col,val);
    return *this;
}

Binder& Binder::operator() (int val)
{
    stmt->bind(++col,val);
    return *this;
}

void Binder::reset()
{
    col=0;
}

void Binder::execute (bool reset)
{
    stmt->execute();
    if (reset) {
        stmt->reset();
    }
}

Connection::Connection (const QString& filename) throw (Exception)
{
    const int result=sqlite3_open (filename.toUtf8(), &db);
    if (result!=SQLITE_OK) {
        Exception::throw_exception (db,result);
    }
    m_Commit = prepare("COMMIT;");
    m_Rollback = prepare("ROLLBACK;");
    m_BeginTransaction = prepare ("BEGIN;");
}

Connection::~Connection ()
{
    delete m_Commit;
    delete m_Rollback;
    sqlite3_close (db);
}

Statement* Connection::prepare (const char* sql) throw (Exception)
{
    return new Statement (db,sql);
}

void Connection::commit () throw (Exception)
{
    m_Commit->execute();
}

void Connection::rollback () throw (Exception)
{
    m_Rollback->execute();
}

void Connection::beginTransaction () throw (Exception)
{
    m_BeginTransaction->execute();
}

}
}
