#include "tooltip.h"

#include <QWidget>
#include <QStylePainter>
#include <QStyleOptionFrame>
#include <QTimerEvent>
#include <QHBoxLayout>
#include <QApplication>
#include <QMouseEvent>
#include <QDesktopWidget>
#include <QCoreApplication>
#include <QKeyEvent>
#include <QToolTip>

#include <iostream>
#include <algorithm>

using std::cout;
using std::endl;
using std::min;

namespace Texel
{
    Tooltip* Tooltip::s_Instance=0l;

    static inline int screen (const QPoint& pos, QWidget* widget)
    {
        if (QApplication::desktop()->isVirtualDesktop() || widget==0l) {
            return QApplication::desktop()->screenNumber(pos);
        }
        return QApplication::desktop()->screenNumber(widget);
    }

    bool Tooltip::hasTooltip (const QWidget* owner)
    {
        if (s_Instance==0l || owner==0l) {
            return s_Instance!=0l;
        }
        return s_Instance->m_Owner==owner;
    }


    void Tooltip::show (QWidget* tip, const QPoint& pos, const QRect& validityBox, QWidget* owner)
    {
        if (hasTooltip(owner)) return;

        QRect box = validityBox;
        if (box.isNull()) {
            box=owner->rect().translated(owner->mapToGlobal(QPoint(0,0)));
        }

        new Tooltip (tip, pos, box, owner);
    }

    Tooltip::Tooltip (QWidget* tip, const QPoint& pos, const QRect& validityBox, QWidget* parent=0l) :
            QWidget(QApplication::desktop()->screen(screen(pos,parent)), Qt::ToolTip),
            m_Tooltip(tip),
            m_Screen(screen(pos,parent)),
            m_ToolTipBox(validityBox),
            m_Persist(false)
    {
        cout << "BOX: " << validityBox.x() << "," << validityBox.width() << ";" << validityBox.y() << "," << validityBox.height() << endl;
        cout << "POS: " << pos.x() << "," << pos.y() << endl;

        delete s_Instance;
        s_Instance=this;

        tip->setParent(this);
        tip->setPalette(QToolTip::palette());
        tip->setForegroundRole(QPalette::ToolTipText);
        tip->setBackgroundRole(QPalette::ToolTipBase);

        setWindowOpacity(style()->styleHint(QStyle::SH_ToolTipLabel_Opacity, 0, this) / 255.0);
        setSizePolicy(QSizePolicy(QSizePolicy::Preferred,QSizePolicy::Preferred));

        QLayout* _layout = new QHBoxLayout(this);
        _layout->setMargin(1 + style()->pixelMetric(QStyle::PM_ToolTipLabelFrameWidth, 0, this));
        setLayout(_layout);
        _layout->addWidget(tip);

        if (parent!=0l) {
            connect (parent, SIGNAL(destroyed()), this, SLOT(deleteLater()));
        }
        connect (tip,SIGNAL(destroyed()), this, SLOT(deleteLater()));

        QCoreApplication::instance()->installEventFilter(this);

        resize(m_ToolTipBox.size());
        placeTip(pos);
        showWidget ();
    }

    Tooltip::~Tooltip ()
    {
        s_Instance=0l;
    }

    void Tooltip::placeTip (const QPoint& pos)
    {
        // Copied from qtooltip.cpp to gain same layout

#     ifdef Q_WS_MAC
        const QRect screen = QApplication::desktop()->availableGeometry(m_Screen);
#     else
        const QRect screen = QApplication::desktop()->screenGeometry(m_Screen);
#     endif

        QPoint p = pos;

        static const int cursorHeight =
#     ifdef Q_WS_WIN
                21;
#     else
        16;
#     endif

        // Move position of the tooltip to the lower end of the cursor but do
        // not leave the valid rect where the tooltip is displayed
        p.setX(min(p.x()+2,            m_ToolTipBox.x()+m_ToolTipBox.width()));
        p.setY(min(p.y()+cursorHeight, m_ToolTipBox.y()+m_ToolTipBox.height()));

        if (p.x() + this->width() > screen.x() + screen.width())
            p.rx() -= 4 + this->width();
        if (p.y() + this->height() > screen.y() + screen.height())
            p.ry() -= 24 + this->height();

        if (p.y() < screen.y())
            p.setY(screen.y());
        if (p.x() + this->width() > screen.x() + screen.width())
            p.setX(screen.x() + screen.width() - this->width());
        if (p.x() < screen.x())
            p.setX(screen.x());
        if (p.y() + this->height() > screen.y() + screen.height())
            p.setY(screen.y() + screen.height() - this->height());

        move(p);
    }

    void Tooltip::showWidget ()
    {
        QWidget::show();
        m_TimerId=startTimer(3000);
    }

    void Tooltip::hideWidget ()
    {
        if (!m_Persist) {
            hide();
            if (m_TimerId!=0) {
                killTimer(m_TimerId);
                m_TimerId=0;
            }
            deleteLater();
        }
    }

    void Tooltip::paintEvent(QPaintEvent* ev)
    {
        QStylePainter p(this);
        QStyleOptionFrame opt;
        opt.init(this);
        p.drawPrimitive(QStyle::PE_PanelTipLabel, opt);
        p.end();

        QWidget::paintEvent(ev);
    }

    void Tooltip::timerEvent (QTimerEvent* event)
    {
        if (event->timerId()!=m_TimerId) return;

        if (!rect().contains(mapFromGlobal(QCursor::pos()))) {
            hideWidget();
        }
    }

    bool Tooltip::eventFilter(QObject *o, QEvent *e)
    {
        switch (e->type()) {

        case QEvent::MouseMove:
            {
                const QPoint pos = dynamic_cast<QMouseEvent*>(e)->globalPos();
                if (!m_ToolTipBox.contains(pos) && !rect().contains(mapFromGlobal(pos))) {
                    hideWidget();
                }
            }
            break;

        case QEvent::KeyPress:
            {
                const int pressedKey = dynamic_cast<QKeyEvent*>(e)->key();
                if (pressedKey==Qt::Key_Escape) {
                    hideWidget();
                }
                else if (pressedKey==Qt::Key_AltGr) {
                    m_Persist=true;
                }
            }
            break;

        case QEvent::KeyRelease:
            {
                const int pressedKey = dynamic_cast<QKeyEvent*>(e)->key();
                if (pressedKey==Qt::Key_AltGr && m_Persist) {
                    m_Persist=false;
                    hideWidget();
                }
            }
            break;

        case QEvent::Move:
        case QEvent::Hide:
        case QEvent::Resize:
        case QEvent::WindowActivate:
        case QEvent::WindowDeactivate:
            if (o==m_Owner) {
                hideWidget();
            }
            break;

        default:
            break;
        }

        return false;
    }

}
