/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "xmlhighlighter.h"

#include <boost/spirit/include/classic.hpp>
#include <boost/spirit/include/classic_position_iterator.hpp>

#include "aid.h"

using namespace BOOST_SPIRIT_CLASSIC_NS;

static char operator- (QChar c, char n)
{
    return c.toAscii()-n;
}

static char operator+ (QChar c, char n)
{
    return c.toAscii()+n;
}

QTextCharFormat operator+ (Qt::GlobalColor c, QFont::Weight m)
{
    QTextCharFormat f;
    f.setForeground(c);
    f.setFontWeight(m);

    return f;
}

QTextCharFormat& operator+ (QTextCharFormat& f, QFont::Weight m)
{
    f.setFontWeight(m);
    return f;
}

QTextCharFormat operator+ (Qt::GlobalColor c, QFont::Style s)
{
    QTextCharFormat f;
    f.setForeground(c);
    if (s==QFont::StyleItalic) {
        f.setFontItalic(true);
    }

    return f;
}

QTextCharFormat& operator+ (QTextCharFormat& f, QFont::Style s)
{
    if (s==QFont::StyleItalic) {
        f.setFontItalic(true);
    }
    return f;
}

struct XQueryParser: public grammar<XQueryParser>
{
    typedef QString::const_iterator posIterator;

    XQueryParser (XmlHighlighter* h)
        : highlighter(h)
    {
    }

    mutable XmlHighlighter* highlighter;
    QString block;

    void setFormat (posIterator start, int count, const QTextCharFormat& format) const
    {
        highlighter->setFormat(start-block.begin(), count, format);
    }

    void setState (int state) const
    {
        highlighter->setCurrentBlockState(state);
    }

    int getPrevState () const
    {
        return highlighter->previousBlockState();
    }

    struct do_SaveState
    {
        do_SaveState (const XQueryParser& h, int s)
            : parser(h),
              state(s)
        {
        }

        void operator() (posIterator, posIterator ) const
        {
            parser.setState(state);
        }

        const XQueryParser& parser;
        int state;
    };

    friend struct do_SaveState;

    struct do_Highlight
    {
        do_Highlight (const XQueryParser& h, const Qt::GlobalColor c, int o=0)
            : format(),
              parser(h),
              offset(o)
        {
            format.setForeground(c);
        }

        do_Highlight (const XQueryParser& h, const QTextCharFormat& f, int o=0)
            : format(f),
              parser(h),
              offset(o)
        {
        }

        void operator() (posIterator start, posIterator end) const
        {
            parser.setFormat (start-offset,end-start+offset, format);
        }

        QTextCharFormat format;
        const XQueryParser& parser;
        int offset;
    };

    template <typename ScannerT> struct definition
    {
        const XQueryParser& s;

        definition (XQueryParser const& self)
            : s(self)
        {
            keywords =
                    (str_p("declare")|str_p("function")|str_p("namespace")|str_p("default")|str_p("schema")|
                     str_p("if")|str_p("then")|str_p("else")|str_p(":=")|str_p("gt")|str_p("lt")|
                    str_p("returns")|str_p("sortby")|str_p("where")|str_p("for")|str_p("let")|
                    str_p("ascending")|str_p("descending")|str_p("before")|str_p("after")|str_p("and")|str_p("or")|
                    str_p("some")|str_p("in")|str_p("every")|str_p("satisfies")|
                    str_p("typeswitch")|str_p("as")|str_p("return")|str_p("case")|
                     str_p("instanceof")|str_p("only")|str_p("eq")|str_p("div")|
                    str_p("union")|str_p("intersect")|str_p("intercept")|
                     str_p("comment")|str_p("text")|str_p("node")|str_p("processing-instruction"))[do_Highlight(self,Qt::blue+QFont::Bold)];

            in_xq_comment = confix_p(eps_p(anychar_p),*anychar_p,str_p(":)")[do_SaveState(self,0)]|end_p[do_SaveState(self,1)])[do_Highlight(self,Qt::green)] >> document;

            in_xml_comment = confix_p(eps_p(anychar_p),*anychar_p,(str_p("-->")[do_SaveState(self,0)]|end_p[do_SaveState(self,2)]))[do_Highlight(self,Qt::green)] >> document;

            document = *blank_p >> *(tag | xquery) >> *blank_p;

            tag = ch_p('<') >> ( confix_p("!--", *anychar_p ,(str_p("-->")[do_SaveState(self,0)]|end_p[do_SaveState(self,2)]))[do_Highlight(self,Qt::green,1)]
                          | (*blank_p >> (qname[do_Highlight(self,Qt::blue)] >> *blank_p >> !(attributes >> *blank_p) >> (str_p("/>") | ch_p('>')))
                          | (ch_p('/') >> *blank_p >> qname[do_Highlight(self,Qt::blue)] >> *blank_p >> ch_p('>'))));

            qname = name >> !(':' >> name);

            name = startchar >> *( startchar | '.' | '-' | chset<QChar>("0-9"));

            startchar = chset<QChar>("a-zA-Z_");

            attributes = attribute % +blank_p;

            attribute = name >> *blank_p >> '=' >> *blank_p >> stringliteral;

            stringliteral = (confix_p('\'',*anychar_p,'\'')|confix_p('\"',*anychar_p,'\"'))[do_Highlight(self,Qt::darkGreen)];

            tagcontent = *xml_comment >> characters;

            characters = *(anychar_p-ch_p('<'));

            xquery = +(*blank_p >> (tag | fragments));

            fragments = keywords | stringliteral | var | expr | comment_or_bracket | attrib_ref | brackets;

            comment_or_bracket = longest_d[confix_p("(:",*anychar_p,str_p(":)")[do_SaveState(self,0)]|end_p[do_SaveState(self,1)])[do_Highlight(self,Qt::green)]|'('];

            brackets = (+chset<QChar>("{}"))[do_Highlight(self,Qt::black+QFont::Bold)];

            var = (ch_p('$') >> qname)[do_Highlight(self,Qt::darkRed)] >> (+blank_p|eps_p(chset<QChar>(")/,=")));

            attrib_ref = (ch_p('@') >> qname)[do_Highlight(self,Qt::black+QFont::StyleItalic)] >> (+blank_p|eps_p(chset<QChar>(")/,=")));

            //expr = +chset<QChar>(anychar_p-chset<QChar>("<$(\'\"@{}"));

            expr = +(anychar_p-ch_p('<')-ch_p('$')-ch_p('(')-ch_p('\'')-ch_p('\"')-ch_p('@')-ch_p('{')-ch_p('}')-blank_p);
        }

        rule<ScannerT> document, tag, qname, name, startchar, attributes, attribute, stringliteral, attrib_ref,
        xml_comment, tag_or_comment, characters, tagcontent, keywords, xq_comment, comment_or_bracket, brackets,
        in_xq_comment, in_xml_comment;

        rule<ScannerT> xquery, var, expr, fragments;

        rule<ScannerT> const& start() const {
            int state=s.getPrevState();
            if (state==1) {
                return in_xq_comment;
            }
            else if (state==2) {
                return in_xml_comment;
            }
            return document;
        }
    };

    void highlight (const QString& input)
    {
        block=input;
        boost::spirit::classic::parse(input.begin(), input.end(), *this/*, space_p*/);
    }
};

XmlHighlighter::XmlHighlighter(QObject* parent)
: QSyntaxHighlighter(parent),
  p(new XQueryParser(this))
{
}

XmlHighlighter::XmlHighlighter(QTextDocument* parent)
: QSyntaxHighlighter(parent),
  p(new XQueryParser(this))
{
}

XmlHighlighter::XmlHighlighter(QTextEdit* parent)
: QSyntaxHighlighter(parent),
  p(new XQueryParser(this))
{
}

XmlHighlighter::~XmlHighlighter()
{
    delete p;
}

void XmlHighlighter::highlightBlock (const QString& text)
{
    setCurrentBlockState(previousBlockState());
    p->highlight(text);
}
