/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include <func/undocommands.h>
#include <model/texeltagmodel.h>
#include "detailedattributetable.h"

#include <QStandardItem>
#include <QMenu>
#include <QLineEdit>
#include <QIntValidator>
#include <QTableView>
#include <QPushButton>

#include <klocale.h>
#include <kundostack.h>

#include "aid.h"

using namespace std;

namespace Texel {

    StandardItemUndoModel::StandardItemUndoModel (TagExtension* ext, QWidget* parent)
        : QStandardItemModel(parent),
          m_Extension(ext)
    {
    }

    StandardItemUndoModel::~StandardItemUndoModel ()
    {
    }

    bool StandardItemUndoModel::setData (const QModelIndex& index, const QVariant& value, int role)
    {
        if (index.column()==1 && role==Qt::EditRole) {
            // Change of data
            QVariant old = data(index);
            const QModelIndex idx=this->index(index.row(),0);
            emit attributeChanged (data(idx).toString(), old, value);
            return true;
        }
        return QStandardItemModel::setData(index,value,role);
    }

    void StandardItemUndoModel::reset (const QString& name, const QVariant& value)
    {
        for (int i=0;i<rowCount();i++) {
            if (item(i)->data(Qt::DisplayRole).toString()==name) {
                const QModelIndex idx=index(i,1);
                QStandardItemModel::setData(idx,value,Qt::EditRole);
                return;
            }
        }

        // Row not found, add it
        QList<QStandardItem*> row = createRow(name,value.toString());
        appendRow(row);
    }

    void StandardItemUndoModel::remove (const QString& name)
    {
        for (int i=0;i<rowCount();i++) {
            if (item(i)->data(Qt::DisplayRole).toString()==name) {
                removeRow(i);
                return;
            }
        }
    }

    QList<QStandardItem*> StandardItemUndoModel::createRow (const QString& key_, const QString& name)
    {
        QList<QStandardItem*> row;
        QStandardItem* key = new QStandardItem(key_);
        key->setEditable(false);
        QStandardItem* value=new QStandardItem(name);
        TagExtension::AttributeMetaData meta = m_Extension->getMetaData(key_);
        value->setData(QVariant(meta.first),Qt::UserRole+1);
        value->setData(QVariant(meta.second),Qt::UserRole+2);
        row << key << value;
        return row;
    }

    DetailedAttributeTable::DetailedAttributeTable(QWidget *parent) :
            TableEditor(parent)
    {
        init();
        table()->setItemDelegate(new Internal::AttributeDelegate(table()));
        StandardItemUndoModel* model = dynamic_cast<StandardItemUndoModel*>(table()->model());
        model->setColumnCount(2);
        model->setHorizontalHeaderItem(0,new QStandardItem(i18n("Property")));
        model->setHorizontalHeaderItem(1,new QStandardItem(i18n("Value")));
        table()->setModel(model);
        connect (model, SIGNAL(attributeChanged(QString,QVariant,QVariant)), this, SLOT(onChangedData(QString, QVariant, QVariant)));
    }

    QAbstractItemModel* DetailedAttributeTable::doInitModel (QWidget* parent) const
    {
        return new StandardItemUndoModel(m_Extension,parent);
    }

    void DetailedAttributeTable::onChangedData(const QString& attrib, const QVariant& old, const QVariant& upd)
    {
        UpdateExtensionAttrib* cmd = new UpdateExtensionAttrib(m_Extension->getItem(),attrib,old.toString(),upd.toString());
        m_Extension->getItem()->getModel()->undoStack()->push(cmd);
    }

    void DetailedAttributeTable::setExtension (TagExtension* extension)
    {
        m_Extension=extension;

        QMap<QString,QString> attribs = m_Extension->getAttribs();
        StandardItemUndoModel* model = dynamic_cast<StandardItemUndoModel*>(table()->model());
        model->setExtension(extension);

        model->removeRows(0,model->rowCount());
        for (QMap<QString,QString>::iterator iter=attribs.begin();iter!=attribs.end();++iter) {
            QList<QStandardItem*> row = model->createRow (iter.key(),iter.value());
            model->appendRow(row);
        }

        createAddMenu();
    }

    void DetailedAttributeTable::appendRow ()
    {
        const QString name =  dynamic_cast<QAction*>(sender())->data().toString();
        InsertExtensionAttrib* cmd = new InsertExtensionAttrib(m_Extension->getItem(),name,"",false);
        m_Extension->getItem()->getModel()->undoStack()->push(cmd);
        createAddMenu();
    }

    int DetailedAttributeTable::doAppend ()
    {
        return 0;
    }

    void DetailedAttributeTable::doRemove (int row)
    {
        StandardItemUndoModel* model = dynamic_cast<StandardItemUndoModel*>(table()->model());

        const QString name = model->data(model->index(row,0)).toString();
        const QString value = model->data(model->index(row,1)).toString();

        InsertExtensionAttrib* cmd = new InsertExtensionAttrib(m_Extension->getItem(),name,value,true);
        m_Extension->getItem()->getModel()->undoStack()->push(cmd);
    }

    void DetailedAttributeTable::createAddMenu ()
    {
        QPushButton* b = dynamic_cast<QPushButton*>(add_Button);

        if (b->menu()!=0l) {
            b->menu()->deleteLater();
        }

        QMenu* chooseType = new QMenu(this);
        foreach (QString n, m_Extension->getAllowedAttribs()) {
            if (!m_Extension->hasAttribute(n)) {
                chooseType->addAction(n,this,SLOT(appendRow()))->setData(n);
            }
        }

        b->setMenu (chooseType);
    }

    namespace Internal
    {
        AttributeDelegate::AttributeDelegate (QObject* parent)
            : QStyledItemDelegate(parent)
        {
        }

        AttributeDelegate::~AttributeDelegate ()
        {
        }

        QWidget* AttributeDelegate::createEditor (QWidget* parent,
                                                  const QStyleOptionViewItem& option,
                                                  const QModelIndex& index ) const
        {
            QWidget* editor=QStyledItemDelegate::createEditor (parent,option,index);
            QLineEdit* lineEdit = dynamic_cast<QLineEdit*>(editor);
            if (lineEdit!=0) {
                QVariant type = index.data(Qt::UserRole+1);
                QVariant length = index.data(Qt::UserRole+2);
                if (type.isValid()) {
                    switch (type.toInt()) {
                    case QMetaType::Int:
                        lineEdit->setValidator(new QIntValidator(0,length.toInt(),lineEdit));
                        break;
                    case QMetaType::QString:
                        if (length.toInt()!=-1) {
                            lineEdit->setMaxLength(length.toInt());
                        }
                        break;
                    }
                }
            }
            return editor;
        }

    }

}
