/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "dirmodel.h"
#include "dirview.h"
#include "model/tagitem.h"
#include "model/texeltagmodel.h"

#include <QWidget>
#include <QMenu>
#include <QContextMenuEvent>

#include <klocale.h>
#include <kio/deletejob.h>
#include <kio/copyjob.h>
#include <kio/chmodjob.h>
#include <kmessagebox.h>

namespace Texel {

    DirView::DirView(QWidget* parent)
        :KFileTreeView(parent)
    {
    }

    void DirView::contextMenuEvent (QContextMenuEvent* event)
    {
        QMenu* popup = new QMenu(this);
        DirModel* m = dynamic_cast<DirModel*>(model());
        if (m!=0l) {
            QModelIndex idx = indexAt(event->pos());
            const QString& path = m->itemForIndex(idx).localPath();
            PopupActions* actions = new PopupActions(popup,m,path);

            if (m->isLoaded(idx)) {
                popup->addAction (i18n("Reload directory"),actions, SLOT(onReloadDirectory()));
                popup->addAction (i18n("Save tracks in directory"), actions, SLOT(onWriteDirectory()));
                popup->addAction (i18n("Remove content from model"), actions, SLOT(onRemoveDirectoryFromModel()));
            }

            popup->addSeparator();
            popup->addAction (i18n("Erase directory from disk"), actions, SLOT(eraseFromDisk()));
            popup->addAction (i18n("Move directory to trash"), actions, SLOT(moveToTrash()));
            popup->addAction (i18n("Set write permission for user"), actions, SLOT(setWritePermission()));

            popup->popup(mapToGlobal(event->pos()));
        }
    }

    PopupActions::PopupActions (QObject* parent, DirModel* model, const QString& path)
        :QObject(parent),
        m_Model(model),
        m_Path(path)
    {
    }

    PopupActions::~PopupActions ()
    {
    }

    void PopupActions::setWritePermission ()
    {
        KFileItemList list;
        list << KFileItem(m_Path,"",S_IFDIR);
        KIO::chmod (list,0200,0200,"","",true);
    }

    void PopupActions::eraseFromDisk ()
    {
        int result=KMessageBox::questionYesNo(0l,
                                              i18n("Delete directory %1",m_Path),
                                              i18n("Do you want to delete %1?",m_Path)
                                              );
        if (result==KMessageBox::Yes) {
            KIO::del(KUrl::fromLocalFile(m_Path));
        }
    }

    void PopupActions::moveToTrash ()
    {
        KIO::trash(KUrl::fromLocalFile(m_Path));
    }

    void PopupActions::onWriteDirectory ()
    {
        TagItem::List items = m_Model->getDirTree()->getTagItems(m_Path,false);
        if (!items.empty()) {
            items.first()->getModel()->updateWhenNecessary(items.toSet(),false);
        }
    }

    void PopupActions::onRemoveDirectoryFromModel ()
    {
        TagItem::List items = m_Model->getDirTree()->getTagItems(m_Path,false);
        if (!items.empty()) {
            items.first()->getModel()->removeRows(items);
            m_Model->onPathStateChange(m_Path);
        }
    }

    void PopupActions::onReloadDirectory()
    {
        TagItem::List items = m_Model->getDirTree()->getTagItems(m_Path,false);
        if (!items.empty()) {
            items.first()->getModel()->removeRows(items);
            QList<QUrl> dirs;
            dirs << QUrl(m_Path);
            items.first()->getModel()->load(dirs);
        }
    }
}

#include "dirview.moc"
