/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "duplicatehelper.h"
#include "ui_duplicatehelper.h"

#include "model/tagmodelproxy.h"
#include "options/preferences.h"
#include "util/filedialog.h"

#include <QModelIndex>
#include <QTextDocument>
#include <QTextTable>
#include <QTextCursor>
#include <QAbstractTextDocumentLayout>
#include <QPainter>
#include <QMenuBar>
#include <QTextBrowser>
#include <QPushButton>
#include <QCheckBox>

#include "aid.h"

using namespace std;

static uint qHash (const QStringList& list)
{
    uint hash=list.size();
    foreach (QString s,list) {
        uint h2=qHash(s);
        hash=((hash << 16) | (hash >> 16)) ^ h2;
    }
    return hash;
}

namespace Texel {

    ////////////////////////////////////////////////////////////////////////////////////////////////////
    //
    // Delegate
    //
    ////////////////////////////////////////////////////////////////////////////////////////////////////

    RichTextDelegate::RichTextDelegate (QObject* parent)
        :QItemDelegate(parent),
        m_Orientation(Qt::Vertical)
    {
    }

    RichTextDelegate::~RichTextDelegate ()
    {
    }

    void RichTextDelegate::paint (QPainter* painter, const QStyleOptionViewItem& option, const QModelIndex& index) const
    {
        QVariant data = index.data(Qt::DisplayRole);
        if (data.type()==QVariant::StringList && option.rect.width()>0) {
            QTextDocument* doc=formattedData (data.toStringList(), option, index);
            painter->save();
            painter->translate(option.rect.topLeft()+QPoint(0,35));
            QRect r=option.rect;
            r.moveTo(0,0);
            drawBackground(painter,option,index);
            doc->drawContents(painter,r);
            painter->restore();
        }
        else {
            QItemDelegate::paint(painter,option,index);
        }
    }

    QSize RichTextDelegate::sizeHint (const QStyleOptionViewItem& option, const QModelIndex& index) const
    {
        QVariant data = index.data(Qt::DisplayRole);
        if (data.type()==QVariant::StringList) {
            const QTextDocument* doc=formattedData (data.toStringList(), option, index);
            return doc->documentLayout()->documentSize().toSize()+QSize(0,35);
        }
        else return QItemDelegate::sizeHint(option,index);
    }

    QTextDocument* RichTextDelegate::formattedData (const QStringList& item, const QStyleOptionViewItem& option, const QModelIndex& index) const
    {
        int cols = option.rect.width()/300;
        cols = (cols==0)?1:cols;
        const int width = (m_Orientation==Qt::Horizontal)?300:(option.rect.width()/cols);
        const bool highlight = index.data(Qt::ForegroundRole).toBool();

        const CacheKey key(DocProperties(width,highlight),item);
        if (m_DocCache.contains(key)) {
            return m_DocCache[key];
        }

        QPalette p = option.palette;
        if (!highlight) {
            p.setCurrentColorGroup(QPalette::Disabled);
        }

        QTextDocument* doc=new QTextDocument();
        doc->setTextWidth(width);

        QTextCursor cursor(doc);
        QTextCharFormat format;
        QTextBlockFormat blockFormat;
        cursor.setBlockFormat(blockFormat);
        format.setFontPointSize(12.0);
        format.setForeground(p.text().color());
        format.setFontItalic(true);
        cursor.insertText(item[0]+"\n"+item[1]+"\n",format);

        format.setFontItalic(false);
        format.setFontPointSize(10.0);
        cursor.insertText(item[2]+"\n",format);

        QTextTableFormat tableFormat;
        tableFormat.setCellPadding(0.0);
        tableFormat.setCellSpacing(0.0);
        tableFormat.setWidth(QTextLength(QTextLength::FixedLength,width-5));
        tableFormat.setBorderStyle(QTextFrameFormat::BorderStyle_None);

        QTextTable* table = cursor.insertTable(1,2,tableFormat);
        QTextCursor c=table->cellAt(0,0).firstCursorPosition();
        blockFormat.setAlignment(Qt::AlignLeft);
        c.setBlockFormat(blockFormat);
        c.insertText("Year:"+item[10],format);

        c=table->cellAt(0,1).firstCursorPosition();
        blockFormat.setAlignment(Qt::AlignRight);
        c.setBlockFormat(blockFormat);
        c.insertText("Genre:"+item[11],format);

        cursor.movePosition(QTextCursor::End);
        blockFormat.setAlignment(Qt::AlignLeft);
        blockFormat.setNonBreakableLines(false);
        cursor.setBlockFormat(blockFormat);
        QTextCharFormat courierFormat;
        courierFormat.setFontPointSize(8.0);
        courierFormat.setFontFamily("Courier");
        courierFormat.setForeground(p.text().color());
        cursor.insertText(item[3],courierFormat);

        format.setFontPointSize(9.0);

        cursor.movePosition(QTextCursor::End);
        table = cursor.insertTable(1,3,tableFormat);
        c=table->cellAt(0,0).firstCursorPosition();
        blockFormat.setAlignment(Qt::AlignLeft);
        c.setBlockFormat(blockFormat);
        c.insertText(item[4],format);

        c=table->cellAt(0,1).firstCursorPosition();
        blockFormat.setAlignment(Qt::AlignCenter);
        c.setBlockFormat(blockFormat);
        c.insertText(item[5],format);

        c=table->cellAt(0,2).firstCursorPosition();
        blockFormat.setAlignment(Qt::AlignRight);
        c.setBlockFormat(blockFormat);
        c.insertText(item[6],format);

        cursor.movePosition(QTextCursor::End);
        table = cursor.insertTable(1,3,tableFormat);
        c=table->cellAt(0,0).firstCursorPosition();
        blockFormat.setAlignment(Qt::AlignLeft);
        c.insertText(item[7]+" bps",format);

        c=table->cellAt(0,1).firstCursorPosition();
        blockFormat.setAlignment(Qt::AlignCenter);
        c.setBlockFormat(blockFormat);
        c.insertText(item[8]+i18n(" 1/s"),format);

        c=table->cellAt(0,2).firstCursorPosition();
        blockFormat.setAlignment(Qt::AlignRight);
        c.setBlockFormat(blockFormat);
        c.insertText("CD#"+item[9],format);

        m_DocCache.insert(key,doc);

        return doc;
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////
    //
    // Model class
    //
    ////////////////////////////////////////////////////////////////////////////////////////////////////

    DuplicatesDetailsModel::DuplicatesDetailsModel (QObject* parent)
        : QAbstractItemModel(parent),
        m_Properties(Preferences::instance()->duplicatehelper.m_Properties)
    {
    }

    DuplicatesDetailsModel::~DuplicatesDetailsModel ()
    {
    }

    QModelIndex DuplicatesDetailsModel::index (int row, int column, const QModelIndex&) const
    {
        return createIndex(row,column,0);
    }

    QModelIndex DuplicatesDetailsModel::parent (const QModelIndex&) const
    {
        return QModelIndex();
    }

    int DuplicatesDetailsModel::rowCount (const QModelIndex&) const
    {
        return m_Duplicates.size();
    }

    int DuplicatesDetailsModel::columnCount (const QModelIndex&) const
    {
        return 1;
    }

    Qt::ItemFlags DuplicatesDetailsModel::flags (const QModelIndex& index) const
    {
        if (index.isValid()) {
            return (Qt::ItemFlags)(Qt::ItemIsEnabled);
        }
        return Qt::NoItemFlags;
    }

    QVariant DuplicatesDetailsModel::data (const QModelIndex& index, int role) const
    {
        if (!index.isValid() || index.row()<0 || index.row()>rowCount()) {
            return QVariant();
        }

        if (role==Qt::ForegroundRole) {
            return QVariant(m_Duplicates[index.row()].filtered);
        }

        if (role==Qt::DisplayRole) {
            QStringList items;
            TagItem* row = m_Duplicates[index.row()].item;
            items << row->data(ColumnManager::Artist,Columns::Column::RoleDisplay)
                    << row->data(ColumnManager::Title,Columns::Column::RoleDisplay)
                    << row->data(ColumnManager::Album,Columns::Column::RoleDisplay)
                    << row->getAbsolutFilename(Columns::Column::RoleDisplay)
                    << row->data(ColumnManager::Length,Columns::Column::RoleDisplay)
                    << row->data(ColumnManager::Channels,Columns::Column::RoleDisplay)
                    << row->data(ColumnManager::FileSize,Columns::Column::RoleDisplay)
                    << row->data(ColumnManager::Bitrate,Columns::Column::RoleDisplay)
                    << row->data(ColumnManager::SampleRate,Columns::Column::RoleDisplay)
                    << row->data(ColumnManager::CD,Columns::Column::RoleDisplay)
                    << row->data(ColumnManager::Year,Columns::Column::RoleDisplay)
                    << row->data(ColumnManager::Genre,Columns::Column::RoleDisplay);
            return QVariant(items);
        }

        return QVariant();
    }

    void DuplicatesDetailsModel::setDuplicateList (const QList<TagItem*> duplicates)
    {
        const int oldSize = m_Duplicates.size();

        m_Duplicates.clear();
        foreach (TagItem* item, duplicates) {
            m_Duplicates << Duplicates(item,inFilter(item));
        }

        if (oldSize<m_Duplicates.size()) {
            beginInsertRows(QModelIndex(),oldSize+1,m_Duplicates.size());
            endInsertRows();
        }
        else if (oldSize>m_Duplicates.size()) {
            beginRemoveRows(QModelIndex(),m_Duplicates.size()+1,oldSize);
            endRemoveRows();
        }
        reset();
    }

    void DuplicatesDetailsModel::setKillButton (int row, QCheckBox* button)
    {
        m_Duplicates[row].button=button;
    }

    void DuplicatesDetailsModel::reevaluateFilter ()
    {
        for (int i=0;i<m_Duplicates.size();i++) {
            m_Duplicates[i].filtered=inFilter(m_Duplicates[i].item);
        }
        reset();
    }

    bool DuplicatesDetailsModel::inFilter (TagItem* item) const
    {
        // Stereo only?
        if (m_Properties.testFlag(Enums::FilterQualityStereo) && item->getIntValue(ColumnManager::Channels)==1) {
            return false;
        }

        // File format
        Enums::FilterProperties type=0;
        if (item->typeName()=="MPEG") type=Enums::FilterFormatMPEG;
        else if (item->typeName().startsWith("Ogg")) type=Enums::FilterFormatOgg;
        else if (item->typeName()=="MP4") type=Enums::FilterFormatMP4;
        else if (item->typeName()=="FLAC") type=Enums::FilterFormatFLAC;
        else if (item->typeName()=="TrueAudio") type=Enums::FilterFormatTrueAudio;
        else if (item->typeName()=="WavPack") type=Enums::FilterFormatWavPack;
        else if (item->typeName()=="ASF") type=Enums::FilterFormatASF;
        else if (item->typeName().startsWith("RIFF")) type=Enums::FilterFormatRIFF;
        if ((m_Properties&type)==0) {
            return false;
        }

        // Quality (more difficult as the rest as quality is defined relative and may be based on
        // the file format)
        if ((type==Enums::FilterFormatFLAC || type==Enums::FilterFormatWavPack) &&
            !m_Properties.testFlag(Enums::FilterQualityHigh)) {
            return false;
        }

        // The samplerate is only used when no low/medium quality songs are wished
        const int samplerate = item->getIntValue(ColumnManager::SampleRate);
        if (samplerate<22000 && !m_Properties.testFlag(Enums::FilterQualityLow)) return false;
        if (samplerate<44000 && !m_Properties.testFlag(Enums::FilterQualityMedium)) return false;

        // More quality information at
        // http://en.wikipedia.org/wiki/Comparison_of_audio_codecs
        const int bps = item->getIntValue(ColumnManager::Bitrate);
        if ((type&(Enums::FilterFormatMPEG|Enums::FilterFormatMP4|Enums::FilterFormatOgg))>0) {
            if (bps<128 && !m_Properties.testFlag(Enums::FilterQualityLow)) return false;
            if (bps>192 && !m_Properties.testFlag(Enums::FilterQualityHigh)) return false;
            if (!m_Properties.testFlag(Enums::FilterQualityMedium)) return false;
        }

        //
        // Directory checks
        //

        const QString dir = item->data(ColumnManager::Path,Columns::Column::RoleAsRawVariable);
        QStringList& allowedDirs = Preferences::instance()->duplicatehelper.m_IncludeDirectories;
        if (!allowedDirs.isEmpty()) {
            foreach (QString d, allowedDirs) {
                cout << "Check " << d << ":" << dir << endl;
                if (dir.startsWith(d)) {
                    return true;
                }
            }
            return false;
        }

        return true;
    }

    void DuplicatesDetailsModel::adjustFilter(bool checked)
    {
        QAction* s = dynamic_cast<QAction*>(sender());
        if (s!=0l) {
            Enums::FilterProperty property = Enums::FilterProperty(s->data().toInt());
            if (m_Properties.testFlag(property)!=checked) {
                m_Properties^=property;
            }
        }
        reevaluateFilter();
    }

    void DuplicatesDetailsModel::update (const QModelIndex& idx)
    {
        for (int i=0;i<m_Duplicates.size();++i) {
            if (m_Duplicates[i].item->getRow()==idx.row()) {
                m_Duplicates[i].filtered=inFilter(m_Duplicates[i].item);
                const QModelIndex local = index(i,0);
                emit dataChanged(local,local);
                m_Duplicates[i].button->setChecked(m_Duplicates[i].item->isToDelete());
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////
    //
    // Dock class
    //
    ////////////////////////////////////////////////////////////////////////////////////////////////////

    DuplicateHelper::DuplicateHelper(QWidget *parent) :
            BaseDock(parent),
            ui(new Ui::DuplicateHelper),
            m_Row(0l)
    {
        ui->setupUi(this);

        m_Model = new DuplicatesDetailsModel(this);
        ui->overview->setModel(m_Model);
        ui->overview->setItemDelegate(new RichTextDelegate(this));
        ui->overview->setSelectionMode(QAbstractItemView::NoSelection);

        connect (this, SIGNAL(dockLocationChanged(Qt::DockWidgetArea)), this, SLOT(onDockLocationChanged(Qt::DockWidgetArea)));

        // Install filter menu to highlight certain elements
        QMenu* filterMenu = new QMenu(ui->menuButton);
        m_Directories = new QMenu(i18n("Directories"),ui->menuButton);
        m_Directories ->addAction(i18n("Add directory"),this, SLOT(addDirectory()));
        m_RemoveDir = new QMenu(i18n("Remove"),ui->menuButton);
        m_Directories ->addMenu(m_RemoveDir);
        m_Directories ->addSeparator();
        foreach (QString dir, Preferences::instance()->duplicatehelper.m_IncludeDirectories) {
            m_RemoveDir->addAction(dir,this,SLOT(removeDirectory()))->setData(dir);
            m_Directories ->addAction(dir)->setData(dir);
        }

        QMenu* fileFormats = new QMenu(i18n("Formats"),ui->menuButton);
        fileFormats->addAction(checkableAction(i18n("MPEG"),
                                               ui->menuButton,
                                               Enums::FilterFormatMPEG));
        fileFormats->addAction(checkableAction(i18n("MP4"),
                                               ui->menuButton,
                                               Enums::FilterFormatMP4));
        fileFormats->addAction(checkableAction(i18n("Ogg"),
                                               ui->menuButton,
                                               Enums::FilterFormatOgg));
        fileFormats->addAction(checkableAction(i18n("TrueAudio"),
                                               ui->menuButton,
                                               Enums::FilterFormatTrueAudio));
        fileFormats->addAction(checkableAction(i18n("Flac"),
                                               ui->menuButton,
                                               Enums::FilterFormatFLAC));
        fileFormats->addAction(checkableAction(i18n("ASF"),
                                               ui->menuButton,
                                               Enums::FilterFormatASF));
        fileFormats->addAction(checkableAction(i18n("WavPack"),
                                               ui->menuButton,
                                               Enums::FilterFormatWavPack));
        fileFormats->addAction(checkableAction(i18n("RIFF"),
                                               ui->menuButton,
                                               Enums::FilterFormatRIFF));

        QMenu* quality = new QMenu(i18n("Quality"),ui->menuButton);
        quality->addAction(checkableAction(i18n("Stereo only"),
                                           ui->menuButton,
                                           Enums::FilterQualityStereo));
        quality->addSeparator();
        quality->addAction(checkableAction(i18n("low quality"),
                                           ui->menuButton,
                                           Enums::FilterQualityLow));
        quality->addAction(checkableAction(i18n("medium quality"),
                                           ui->menuButton,
                                           Enums::FilterQualityMedium));
        quality->addAction(checkableAction(i18n("high quality"),
                                           ui->menuButton,
                                           Enums::FilterQualityHigh));

        quality->addSeparator();
        filterMenu->addAction (i18n("Clear"));
        filterMenu->addSeparator();
        filterMenu->addMenu (m_Directories);
        filterMenu->addMenu (fileFormats);
        filterMenu->addMenu (quality);
        ui->menuButton->setMenu (filterMenu);
    }

    DuplicateHelper::~DuplicateHelper()
    {
        delete ui;
        delete m_Model;
    }

    void DuplicateHelper::addDirectory()
    {
        QString dir=FileDialog::getExistingDirectory();
        if (dir!="") {
            QStringList& list=Preferences::instance()->duplicatehelper.m_IncludeDirectories;
            if (!list.contains(dir)) {
                list << dir;
                m_Directories->addAction(dir)->setData(dir);
                m_RemoveDir->addAction(dir,this,SLOT(removeDirectory()))->setData(dir);
            }
            m_Model->reevaluateFilter();
        }
    }

    void DuplicateHelper::removeDirectory()
    {
        QAction* a = dynamic_cast<QAction*>(sender());
        if (a!=0l) {
            const QString dir = a->data().toString();
            QStringList& list=Preferences::instance()->duplicatehelper.m_IncludeDirectories;
            if (list.contains(dir)) {
                m_RemoveDir->removeAction(a);
                // For directory is more difficult as a text matc must be done
                foreach (QAction* act, m_Directories->actions()) {
                    if (act->data().toString()==dir) {
                        m_Directories->removeAction(act);
                    }
                }
                list.removeAll(dir);
            }
            m_Model->reevaluateFilter();
        }
    }

    QAction* DuplicateHelper::checkableAction (const QString& text, QWidget* parent, Enums::FilterProperty prop)
    {
        QAction* a = new QAction(text,parent);
        connect (a, SIGNAL(toggled(bool)), m_Model, SLOT(adjustFilter(bool)));
        a->setCheckable(true);
        a->setChecked(m_Model->m_Properties.testFlag(prop));
        a->setData(QVariant(prop));
        return a;
    }

    void DuplicateHelper::changeEvent(QEvent *e)
    {
        QDockWidget::changeEvent(e);
        switch (e->type()) {
        case QEvent::LanguageChange:
            ui->retranslateUi(this);
            break;
        default:
            break;
        }
    }

    void DuplicateHelper::entered (const QModelIndex& current, const QModelIndex& previous)
    {
        if ((current.isValid()!=previous.isValid()) || (current.row()!=previous.row())) {
            checkDuplicates(current);
        }
    }

    void DuplicateHelper::changed (const QModelIndex& from, const QModelIndex& to)
    {
        for (int r=from.row();r<=to.row();++r) {
            for (int c=from.column();c<=to.column();++c) {
                m_Model->update(from.model()->index(r,c));
            }
        }
    }

    void DuplicateHelper::onKillClick (bool toDelete)
    {
        QVariant row=sender()->property("row");
        if (!row.isNull() && m_Row!=0l) {
            const int r = row.toInt();
            m_Row->getModel()->setData(m_Row->getModel()->index(r,ColumnManager::MarkForDeletion),toDelete);
        }
    }

    void DuplicateHelper::onPlayClick ()
    {
        QVariant row=sender()->property("row");
        if (!row.isNull()) {
            TagItem* t=m_Model->m_Duplicates[row.toInt()].item;
            emit playSong(t->getAbsolutFilename(Columns::Column::RoleAsRawVariable),
                          t->data(ColumnManager::Filename,Columns::Column::RoleDisplay));
        }
    }

    void DuplicateHelper::onMergeClick ()
    {
        QVariant row=sender()->property("row");
        if (!row.isNull() && m_Row!=0l) {
            m_Row->merge (m_Model->m_Duplicates[row.toInt()].item);
        }
    }

    void DuplicateHelper::onDockLocationChanged (Qt::DockWidgetArea area)
    {
        RichTextDelegate* delegate=dynamic_cast<RichTextDelegate*>(ui->overview->itemDelegate());

        switch (area) {
        case Qt::LeftDockWidgetArea:
        case Qt::RightDockWidgetArea:
            ui->overview->setFlow(QListView::TopToBottom);
            if (delegate!=0l) delegate->setOrienation(Qt::Vertical);
            break;
        case Qt::BottomDockWidgetArea:
        case Qt::TopDockWidgetArea:
            ui->overview->setFlow(QListView::LeftToRight);
            if (delegate!=0l) delegate->setOrienation(Qt::Horizontal);
            break;
        default:
            ;
        }
    }

    void DuplicateHelper::checkDuplicates (const QModelIndex& index)
    {
        if (index.isValid()) {
            const TagModelProxy* model=dynamic_cast<const TagModelProxy*>(index.model());
            if (model!=0l) {
                if (model->getRow(index.row())==m_Row) return;

                m_Row=model->getRow(index.row());

                QList<TagItem*> duplicates;
                if (m_Row->hasRelations()) {
                    duplicates=m_Row->getRelations()->getDuplicates();
                    if (!duplicates.isEmpty()) {
                        duplicates.prepend(m_Row);
                    }
                }

                m_Model->setDuplicateList(duplicates);
                // Kind of a hack: Add a button bar in the upper part of the item.
                // The paiting knows about the toolbar and keeps the area empty
                for (int i=0;i<duplicates.size();i++) {
                    QFrame* bar = new QFrame(0l);
                    bar->setContentsMargins(5,5,5,5);
                    bar->setFrameStyle(QFrame::StyledPanel);
                    QBoxLayout* layout = new QHBoxLayout(bar);
                    layout->setMargin(2);
                    QCheckBox* toDelete = new QCheckBox(i18n("Deferred deletion"),bar);
                    toDelete->setToolTip(i18n("If checked, the entry is marked for deletion upon the next save"));
                    toDelete->setProperty("row",duplicates[i]->getRow());
                    m_Model->setKillButton(i,toDelete);
                    connect (toDelete, SIGNAL(clicked(bool)), this, SLOT(onKillClick(bool)));

                    QToolButton* play = new QToolButton(bar);
                    play->setAutoRaise(true);
                    play->setToolTip(i18n("Prepends song to current play queue and plays it"));
                    play->setIcon(KIconLoader::global()->loadIcon("media-playback-start",KIconLoader::Toolbar));
                    play->setProperty("row",i);
                    connect (play, SIGNAL(clicked()), this, SLOT(onPlayClick()));

                    QToolButton* merge = new QToolButton(bar);
                    merge->setAutoRaise(true);
                    merge->setToolTip(i18n("Copy records to current when empty"));
                    merge->setIcon(KIconLoader::global()->loadIcon("merge",KIconLoader::Toolbar));
                    merge->setProperty("row",i);
                    connect (merge, SIGNAL(clicked()), this, SLOT(onMergeClick()));

                    layout->addWidget(toDelete);
                    layout->addStretch();
                    layout->addWidget(play);
                    layout->addWidget(merge);
                    bar->setLayout(layout);
                    bar->setFixedHeight(35);
                    ui->overview->setIndexWidget(m_Model->index(i,0),bar);
                }

                ui->overview->setCurrentIndex(m_Model->index(0,0));
            }            
        }
        else {
            m_Row=0l;
            m_Model->setDuplicateList(QList<TagItem*>());
        }

        if (m_Model->rowCount()==0) {
            ui->duplicateNumbers->setText("");
        }
        else {
            ui->duplicateNumbers->setText(i18n("Potential duplicates: %1",m_Model->rowCount()));
        }
    }

}

#include "duplicatehelper.moc"
