/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

/*

  This file cumulates all necessary classes to provide a dock which displayes all
  possible (calculated) duplicates of a file.

  */

#ifndef DUPLICATEHELPER_H
#define DUPLICATEHELPER_H

#include "options/enums.h"
#include "basedock.h"

#include <QAbstractItemModel>
#include <QList>
#include <QItemDelegate>
#include <QCache>
#include <QListView>

#include <iostream>

class QModelIndex;
class QTextDocument;
class QMenu;
class QCheckBox;

namespace Ui {
    class DuplicateHelper;
}

namespace Texel {

    class TagItem;

    class RichTextDelegate: public QItemDelegate
    {
    public:
        RichTextDelegate (QObject* parent);
        virtual ~RichTextDelegate ();

        virtual void paint (QPainter* painter, const QStyleOptionViewItem& option, const QModelIndex& index ) const;
        virtual QSize sizeHint (const QStyleOptionViewItem& option, const QModelIndex& index) const;

        void setOrienation(Qt::Orientation o);

    private:
        QTextDocument* formattedData (const QStringList& item, const QStyleOptionViewItem& option, const QModelIndex& index) const;

        typedef QPair<int,bool> DocProperties;
        typedef QPair<DocProperties,QStringList/*Data*/> CacheKey;
        mutable QCache<CacheKey, QTextDocument> m_DocCache;

        // Necessary to determine the width of an item. In case of a vertical orientation
        // the complete available width is taken, otherwise a predefined 300 pixels are used.
        Qt::Orientation m_Orientation;
    };

    inline void RichTextDelegate::setOrienation(Qt::Orientation o)
    {
        m_Orientation=o;
    }

    /**
      This model stores only pointers to all possible duplicate items. The data method delivers
      a list of al displayed items as strings. Furthermore the model provides the filter
      functionality and deliveres the result as the foreground color of the data method.
      */
    class DuplicatesDetailsModel: public QAbstractItemModel
    {
        Q_OBJECT
    public:
        DuplicatesDetailsModel (QObject* parent);
        virtual ~DuplicatesDetailsModel ();

        virtual QModelIndex index (int row, int column, const QModelIndex& parent = QModelIndex()) const;
        virtual QModelIndex parent (const QModelIndex& index) const;
        virtual int rowCount (const QModelIndex& parent = QModelIndex()) const;
        virtual int columnCount (const QModelIndex& parent = QModelIndex()) const;
        virtual QVariant data (const QModelIndex& index, int role = Qt::DisplayRole) const;
        virtual Qt::ItemFlags flags (const QModelIndex& index) const;

        void setDuplicateList (const QList<TagItem*> duplicates);
        void update (const QModelIndex& row);
        void reevaluateFilter ();
        void setKillButton (int row, QCheckBox* button);

    public slots:
        void adjustFilter (bool checked);

    private:
        bool inFilter (TagItem* item) const;

        struct Duplicates
        {
            Duplicates (TagItem* i, bool f): item(i),filtered(f),button(0l) {}

            TagItem* item;
            bool     filtered;
            QCheckBox* button;
        };

        QList<Duplicates> m_Duplicates;
        Enums::FilterProperties& m_Properties;

        friend class DuplicateHelper;
    };

    class DuplicateHelper : public BaseDock {
        Q_OBJECT
    public:
        DuplicateHelper(QWidget *parent = 0);
        ~DuplicateHelper();

    signals:
        void playSong (const QString& filename, const QString& name);

    public slots:
        void entered (const QModelIndex& current, const QModelIndex& previous);
        void changed (const QModelIndex& from , const QModelIndex& to);

    protected slots:
        void addDirectory ();
        void removeDirectory ();

    private slots:
        void onKillClick(bool);
        void onPlayClick ();
        void onMergeClick();

        void onDockLocationChanged (Qt::DockWidgetArea area);

    protected:
        void changeEvent(QEvent *e);

    private:
        void checkDuplicates(const QModelIndex& index);
        QAction* checkableAction (const QString& text, QWidget* parent, Enums::FilterProperty prop);

        Ui::DuplicateHelper *ui;
        TagItem* m_Row;
        DuplicatesDetailsModel* m_Model;

        QMenu* m_Directories;
        QMenu* m_RemoveDir;
    };

}

#endif // DUPLICATEHELPER_H
