#include "logdock.h"
#include "ui_logdock.h"

#include <QTableView>
#include <QStyledItemDelegate>
#include <QResizeEvent>
#include <QFontMetrics>
#include <QScrollBar>

#include <algorithm>

#include <iostream>

namespace Texel {

namespace {
struct Delegate: public QStyledItemDelegate
{
    Delegate (QObject* parent)
        :QStyledItemDelegate (parent)
    {
    }

    virtual ~Delegate ()
    {
    }

    virtual QSize sizeHint (const QStyleOptionViewItem & option, const QModelIndex& index) const
    {
        QStyleOptionViewItemV4 optionV4 = option;
        initStyleOption(&optionV4, index);

        QFontMetrics metrics(option.font);
        QRect textBoundary=metrics.boundingRect(0,0,option.rect.width()-option.decorationSize.width(),10,
                                                Qt::TextWordWrap, index.data().toString());
        int width=textBoundary.width();
        int height=textBoundary.height();

        if (option.showDecorationSelected && optionV4.features&QStyleOptionViewItemV3::HasDecoration) {
            width+=option.decorationSize.width();
            height=std::max(height,option.decorationSize.height());
        }

        return QSize(width,height);
    }
};
}

LogDock::LogDock (QWidget *parent, MessageModel* model)
    : BaseDock(parent),
      ui(new Ui::logdock),
      m_AttachedAtEnd(false)
{
    ui->setupUi(this);
    ui->messageTableView->installEventFilter(this);
    ui->messageTableView->setItemDelegate (new Delegate(this));

    ui->filterLabel->setPixmap(KIconLoader::global()->loadIcon("view-filter",KIconLoader::Toolbar));

    ui->messageTableView->setModel (model);
    connect (ui->closeButton, SIGNAL(clicked()), this, SLOT(close()));
    connect (ui->clearButton, SIGNAL(clicked()), model, SLOT(clear()));
    connect (ui->maxLevelSlider, SIGNAL(valueChanged(int)), model, SLOT(setCurrentMaxLevel(int)));

    connect (ui->messageTableView, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(onDoubleClicked(QModelIndex)));
    connect (ui->filterText, SIGNAL(textChanged(QString)), this, SLOT(filter(QString)));

    connect (ui->messageTableView->verticalScrollBar(), SIGNAL(sliderReleased()), this, SLOT(onSliderValue()));

    model->setCurrentMaxLevel(ui->maxLevelSlider->value());

    setWindowTitle(i18n("View log messages"));
}

LogDock::~LogDock ()
{
    delete ui;
}

void LogDock::onSliderValue()
{
    const int value = ui->messageTableView->verticalScrollBar()->value();

    if (m_AttachedAtEnd && value<ui->messageTableView->verticalScrollBar()->maximum()-5) {
        m_AttachedAtEnd=false;
        disconnect(this, SLOT(ensureVisibilityOfLastRow()));
    }
    else if (value==ui->messageTableView->verticalScrollBar()->maximum() && !m_AttachedAtEnd) {
        m_AttachedAtEnd=true;
        connect (ui->messageTableView->model(), SIGNAL(rowsInserted(QModelIndex,int,int)), this, SLOT(ensureVisibilityOfLastRow()));
        connect (ui->messageTableView->model(), SIGNAL(rowsRemoved(QModelIndex,int,int)), this, SLOT(ensureVisibilityOfLastRow()));
    }
}

void LogDock::ensureVisibilityOfLastRow()
{
    ui->messageTableView->verticalScrollBar()->setValue(ui->messageTableView->verticalScrollBar()->maximum());
}

void LogDock::onDoubleClicked (const QModelIndex& idx)
{
    ui->messageTableView->resizeRowToContents (idx.row());
}

void LogDock::filter (const QString& text)
{
    if (text.isEmpty()) {
        for (int i=0;i<model()->rowCount();i++) {
            ui->messageTableView->setRowHidden(i,false);
        }
    }
    else {
        for (int i=0;i<model()->rowCount();i++) {
            ui->messageTableView->setRowHidden(i,!model()->data(model()->index(i,0)).toString().contains(text) &&
                                                     !model()->data(model()->index(i,1)).toString().contains(text));
        }
    }
}

bool LogDock::eventFilter(QObject *watched, QEvent *event)
{
    if (watched==ui->messageTableView && event->type()==QEvent::Resize) {
        QResizeEvent* e = static_cast<QResizeEvent*>(event);
        const int w = e->size().width()-32;
        ui->messageTableView->setColumnWidth(0,w/2);
        ui->messageTableView->setColumnWidth(1,w/2);
        return true;
    }
    else {
        return QObject::eventFilter(watched, event);
    }
}

MessageModel* LogDock::model() const
{
    return dynamic_cast<MessageModel*>(ui->messageTableView->model());
}

void LogDock::displayMessages ()
{
    show();
}

}
