/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "playlistdock.h"
#include "ui_playlistdock.h"

#include "func/m3u/playlistmodel.h"
#include "model/tagmodelproxy.h"
#include "view/texeltablewidget.h"

#include <QMenu>
#include <QFileInfo>
#include <QClipboard>
#include <QBuffer>

#include <kmessagebox.h>

namespace Texel {

    PlaylistDock::PlaylistDock(QAbstractItemModel* playlistModel, TableWidget* view, QWidget *parent) :
            BaseDock(parent),
            ui(new Ui::PlaylistDock),
            m_ProxyModel(0l),
            m_View(view)
    {
        ui->setupUi(this);
        ui->playlistTree->setModel(playlistModel);
        ui->playlistTree->viewport()->setAcceptDrops(true);
        ui->playlistTree->setAcceptDrops(true);

        connect (ui->playlistTree, SIGNAL(customContextMenuRequested(QPoint)), this, SLOT(onContextMenu(QPoint)));
    }

    PlaylistDock::~PlaylistDock()
    {
        delete ui;
    }

    void PlaylistDock::setModel (TagModelProxy* model)
    {
        m_ProxyModel=model;
        connect (ui->playlistTree, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(onDoubleClick(QModelIndex)) );
    }

    void PlaylistDock::changeEvent(QEvent *e)
    {
        QDockWidget::changeEvent(e);
        switch (e->type()) {
        case QEvent::LanguageChange:
            ui->retranslateUi(this);
            break;
        default:
            break;
        }
    }

    void PlaylistDock::onDoubleClick (const QModelIndex& idx)
    {
        if (m_ProxyModel!=0l) {
            if (!idx.parent().isValid()) {
                const PlaylistModel* m = dynamic_cast<const PlaylistModel*>(idx.model());
                QList<TagItem*> items = m->getTracks(idx);
                m_ProxyModel->applyExternalFilter(items);
            }
            else {
                const PlaylistModel* m = dynamic_cast<const PlaylistModel*>(idx.model());
                const int row=m->getPlaylist(idx.parent().row()).getItem(idx.row())->getRow();
                QModelIndex index=m_ProxyModel->index(row,0);
                m_View->selectionModel()->select(index,QItemSelectionModel::Rows|QItemSelectionModel::ClearAndSelect);
                m_View->scrollTo(index,QAbstractItemView::PositionAtCenter);
            }
        }
    }

    void PlaylistDock::onContextMenu (const QPoint& pos)
    {
        QModelIndex idx = ui->playlistTree->indexAt(pos);
        if (idx.isValid()) {
            QMenu* popup = new QMenu(ui->playlistTree);
            PlaylistModel* playlistModel = dynamic_cast<PlaylistModel*>(ui->playlistTree->model());

            QModelIndexList selectedIndexed=ui->playlistTree->selectionModel()->selectedIndexes();
            PlaylistModel::sortIndexList (selectedIndexed);

            PlayListDockPopupMenuActions* executor =
                    new PlayListDockPopupMenuActions(idx,
                                                     playlistModel,
                                                     ui->playlistTree,
                                                     popup,
                                                     selectedIndexed);
            QModelIndex parent=idx.parent();
            if (!parent.isValid()) {
                // Popupmenu for playlist
                popup->addAction(i18n("Copy playlist order to track data"),executor, SLOT(onCopyPlaylistOrder()));
                popup->addAction(i18n("Remove playlist"), executor, SLOT(removePlaylist()));
                popup->addAction(i18n("Remove playlist from disc"), executor, SLOT(removePlaylistFromDisc()));
                if (playlistModel->getPlaylist(idx.row()).isDirty()) {
                    popup->addAction(i18n("Save"), executor, SLOT(save()));
                }
                popup->addAction(i18n("Rename ..."), executor, SLOT(rename()));

                QMenu* clipboardMenu = new QMenu(i18n("Clipboard"),popup);
                clipboardMenu->addAction(i18n("Copy playlist filename"),executor, SLOT(copyNameToClipboard()));
                clipboardMenu->addAction(i18n("Copy playlist"),executor, SLOT(copyContenttoClipboard()));
                popup->addMenu(clipboardMenu);
            }
            else {
                // Popupmenu for playlist item
                popup->addAction(i18n("Remove track from playlist"), executor, SLOT(removeTrack()));
            }
            popup->addAction(i18n("Remove selected items from playlist"),executor,SLOT(removeItems()));

            if (popup->actions().size()>0) {
                popup->popup(ui->playlistTree->mapToGlobal(pos));
            }
        }
    }


    PlayListDockPopupMenuActions::PlayListDockPopupMenuActions (const QModelIndex& idx,
                                                                PlaylistModel* playlistModel,
                                                                QTreeView* view,
                                                                QObject* parent,
                                                                const QModelIndexList& selected)
        : QObject(parent),
        m_Index(idx),
        m_PlaylistModel(playlistModel),
        m_SelectedIndexes(selected),
        m_View(view)
    {
    }

    void PlayListDockPopupMenuActions::onCopyPlaylistOrder ()
    {
        TagItem::List list = m_PlaylistModel->getTracks(m_Index);
        for (int i=0;i<list.size();i++) {
            if (list[i]!=0l) {
                const QModelIndex idx = list[i]->getModel()->index(list[i]->getRow(),ColumnManager::Track);
                list[i]->getModel()->setData (idx,QVariant(i+1));
            }
        }
    }

    void PlayListDockPopupMenuActions::removePlaylist ()
    {
        m_PlaylistModel->removePlaylist(m_Index.row(), false);
    }

    void PlayListDockPopupMenuActions::removePlaylistFromDisc ()
    {
        const QString filename = m_PlaylistModel->getFilename(m_Index.row());
        QFileInfo info(filename);
        if (info.exists() && info.isWritable()) {
            if (QFile::remove(filename)) {
                removePlaylist();
                return;
            }
        }
        KMessageBox::error(0l,i18n("Couldn't not delete file %1",filename));
    }

    void PlayListDockPopupMenuActions::removeTrack ()
    {
        m_PlaylistModel->removeTrack(m_Index);
    }

    void PlayListDockPopupMenuActions::removeItems ()
    {
        foreach (QModelIndex i, m_SelectedIndexes) {
            if (i.parent().isValid()) {
                m_PlaylistModel->removeTrack(i);
            }
            else {
                m_PlaylistModel->removePlaylist(i.row(),true);
            }
        }
    }

    void PlayListDockPopupMenuActions::save ()
    {
        m_PlaylistModel->getPlaylist(m_Index.row()).save();
    }

    void PlayListDockPopupMenuActions::copyNameToClipboard ()
    {
        QClipboard* clipboard = QApplication::clipboard();
        clipboard->setText (m_PlaylistModel->getPlaylist(m_Index.row()).getFileName());
    }

    void PlayListDockPopupMenuActions::copyContenttoClipboard ()
    {
        QBuffer buffer;
        buffer.open(QIODevice::WriteOnly);
        m_PlaylistModel->getPlaylist(m_Index.row()).save(&buffer,true);
        buffer.close();
        QClipboard* clipboard = QApplication::clipboard();
        clipboard->setText (buffer.buffer());
    }

    void PlayListDockPopupMenuActions::rename ()
    {
        m_View->edit (m_Index);
    }
}
