/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "trackdetails.h"
#include "ui_trackdetails.h"

#include "model/columnmanager.h"
#include "model/tagitem.h"
#include "model/tagmodelproxy.h"
#include "model/tagextension.h"
#include "util/filedialog.h"
#include "func/coversearch/cover.h"
#include "func/coversearch/simpleselector.h"

#include <QLabel>
#include <QToolButton>
#include <QStandardItemModel>
#include <QModelIndex>
#include <QModelIndexList>
#include <QPoint>
#include <QMenu>
#include <QClipboard>
#include <QTableView>

#include <klineedit.h>
#include <kactioncollection.h>
#include <kaction.h>
#include <kstandardaction.h>

#include "aid.h"

using namespace std;

namespace Texel {

    TrackDetails::TrackDetails(QWidget *parent) :
            BaseDock(parent),
            ui(new Ui::TrackDetails),
            m_Row(0l),
            m_Extension(0l),
            m_ImageCount(0),
            m_StandardInitialized(false),
            m_ArtWorkInitialized(false),
            m_ExtendedAttribsInitialized(false),
            m_TimerId(0)
    {
        ui->setupUi(this);
        setFloating(false);
        ui->toolBox->setEnabled(false);
        ui->ratingWidget->setHalfStepsEnabled(false);
        ui->ratingWidget->setMaxRating(5);
        setColumnMap();

        // Create image action menu
        ui->coverMenu->setIcon(KIconLoader::global()->loadIcon("favorites",KIconLoader::Toolbar));
        QMenu* imageMenu = new QMenu(ui->imageBox);

        KAction* find=KStandardAction::find(this, SLOT(searchForCover()), ui->imageBox);
        find->setShortcutContext(Qt::WidgetWithChildrenShortcut);
        imageMenu->addAction(find);
        ui->imageBox->addAction(find);

        KAction* add=new KAction(i18n("Add image from filesystem"),ui->imageBox);
        add->setShortcut(i18n("Ctrl+A"));
        add->setShortcutContext(Qt::WidgetWithChildrenShortcut);
        connect (add, SIGNAL(triggered()), this, SLOT(addImage()));
        ui->imageBox->addAction(add);
        imageMenu->addAction(add);

        KAction* remove=new KAction(i18n("Remove current image"), ui->imageBox);
        remove->setShortcut(i18n("DEL"));
        remove->setShortcutContext(Qt::WidgetWithChildrenShortcut);
        connect (remove, SIGNAL(triggered()), this, SLOT(removeImage()));
        connect (this, SIGNAL(displaysImage(bool)), remove, SLOT(setEnabled(bool)));
        ui->imageBox->addAction(remove);
        imageMenu->addAction(remove);

        KAction* paste=KStandardAction::paste(this, SLOT(pasteImage()), ui->imageBox);
        paste->setShortcutContext(Qt::WidgetWithChildrenShortcut);
        imageMenu->addAction(paste);
        ui->imageBox->addAction(paste);

        KAction* copy=KStandardAction::find(this, SLOT(copyImage()), ui->imageBox);
        copy->setShortcutContext(Qt::WidgetWithChildrenShortcut);
        imageMenu->addAction(copy);
        ui->imageBox->addAction(copy);
        connect (this, SIGNAL(displaysImage(bool)), copy, SLOT(setEnabled(bool)));

        KAction* cut=KStandardAction::cut(this, SLOT(cutImage()), ui->imageBox);
        cut->setShortcutContext(Qt::WidgetWithChildrenShortcut);
        imageMenu->addAction(cut);
        ui->imageBox->addAction(cut);
        connect (this, SIGNAL(displaysImage(bool)), cut, SLOT(setEnabled(bool)));

        ui->toolBox->setCurrentIndex(0);
        ui->coverMenu->setMenu(imageMenu);

        connect (ui->toolBox, SIGNAL(currentChanged(int)), this, SLOT(onPageChanged(int)));
        connect (ui->ratingWidget, SIGNAL(ratingChanged(int)), this, SLOT(alteredRating(int)));
        connect (ui->nextAvailableImage, SIGNAL(clicked(bool)), this, SLOT(nextAvailableImage()));
        connect (ui->prevAvailableImage, SIGNAL(clicked(bool)), this, SLOT(prevAvailableImage()));
        connect (ui->imageDescription, SIGNAL(editingFinished()), this, SLOT(imageDescriptionChanged()));

        connect (ui->imageType, SIGNAL(activated(int)), this, SLOT(setImageType(int)));

        connect (ui->image, SIGNAL(urlsToLoad(QList<QUrl>)), this, SLOT(droppedUrlList(QList<QUrl>)));
        connect (ui->image, SIGNAL(imageToLoad(QImage)), this, SLOT(droppedImage(QImage)));
        connect (ui->image, SIGNAL(customContextMenuRequested(QPoint)), this, SLOT(imagePopupMenu(QPoint)));
    }

    void TrackDetails::setColumnMap ()
    {
        for (int i=1;i<ColumnManager::instance()->getColumnNumber();i++) {
            QString widgetName = "col_"+ColumnManager::instance()->getXmlName(i);
            QWidget* widget = findChild<QWidget*>(widgetName);

            if (ColumnManager::instance()->isEditable(i)) {
                KLineEdit* edit = dynamic_cast<KLineEdit*>(widget);
                if (edit!=0l) {
                    m_ColumnMapper.insert(i,edit);
                } else {
                    KTextEdit* textedit = dynamic_cast<KTextEdit*>(widget);
                    if (textedit!=0l) {
                        m_ColumnMapper.insert(i,textedit);
                    }
                }
            }
            else {
                QLabel* label = dynamic_cast<QLabel*>(widget);
                if (label!=0l) {
                    m_ColumnMapper.insert(i,label);
                }
            }
        }
    }

    TrackDetails::~TrackDetails()
    {
        delete ui;
    }

    void TrackDetails::changeEvent(QEvent *e)
    {
        QDockWidget::changeEvent(e);
        switch (e->type()) {
        case QEvent::LanguageChange:
            ui->retranslateUi(this);
            break;
        default:
            break;
        }
    }

    void TrackDetails::timerEvent (QTimerEvent* event)
    {
        killTimer(event->timerId());
        ui->toolBox->setEnabled(true);
        updateExtension();
    }


    void TrackDetails::entered (const QModelIndex& index)
    {
        if (m_TimerId!=0) killTimer(m_TimerId);

        if (!index.isValid()) {
            ui->toolBox->setDisabled(true);
            return;
        }

        const TagModelProxy* model=dynamic_cast<const TagModelProxy*>(index.model());
        if (model!=0l) {
            if (model->getRow(index.row())==m_Row) {
                return;
            }

            m_Row=model->getRow(index.row());
            m_TimerId = startTimer(150);
        }
    }

    void TrackDetails::changed (const QModelIndex& from, const QModelIndex& to)
    {
        if (!from.isValid() && to.isValid()) {
            ui->toolBox->setEnabled(true);
        }
        else if (from.isValid() && !to.isValid()) {
            ui->toolBox->setEnabled(false);
            return;
        }

        if (m_Row!=0l) {
            if (m_Row->getRow()<=to.row() && from.row()>=m_Row->getRow()) {
                for (int col=from.column();col<=to.column();col++) {
                    const Columns::Column::Role role=
                            (col==ColumnManager::Path)?Columns::Column::RolePersistence:
                                                       Columns::Column::RoleDisplay;
                    const QString text=m_Row->data(col,role);

                    QWidget* widget=m_ColumnMapper[col];
                    if (dynamic_cast<KLineEdit*>(widget)!=0l) dynamic_cast<KLineEdit*>(widget)->setText(text);
                    else if (dynamic_cast<QLabel*>(widget)!=0l) dynamic_cast<QLabel*>(widget)->setText(text);
                    else if (dynamic_cast<KTextEdit*>(widget)!=0l) dynamic_cast<KTextEdit*>(widget)->setText(text);
                }
            }
        }
    }

    void TrackDetails::onAlteredExtension (int row, int type, const QString& name, const QVariant& value)
    {
        if (row==m_Row->getRow()) {
            if (type==0) {
                // Image changed - value is ignored
                if (ui->toolBox->currentIndex()==1 && (
                        *m_CurrentImage==name || m_ImageCount==0)) {
                    // When no image was yet set change the current image to the new added one
                    if (m_ImageCount==0) nextAvailableImage();
                    else showImage();
                }
            }
            else if (type==1 && m_ExtendedAttribsInitialized) {
                // Attribute changed
                static_cast<StandardItemUndoModel*>(ui->extraPropertiesTable->model())->reset(name,value);
            }
            else if (type==2 && m_ExtendedAttribsInitialized) {
                // Attribute removed
                static_cast<StandardItemUndoModel*>(ui->extraPropertiesTable->model())->remove(name);
            }
        }
    }

    void TrackDetails::initOverview ()
    {
        for (QMap<int,QWidget*>::iterator i=m_ColumnMapper.begin();i!=m_ColumnMapper.end();i++) {
            const Columns::Column::Role role=
                    (i.key()==ColumnManager::Path)?Columns::Column::RolePersistence:
                    Columns::Column::RoleDisplay;
            const QString text=m_Row->data(i.key(),role);

            if (dynamic_cast<KLineEdit*>(i.value())!=0l) dynamic_cast<KLineEdit*>(i.value())->setText(text);
            else if (dynamic_cast<QLabel*>(i.value())!=0l) dynamic_cast<QLabel*>(i.value())->setText(text);
            else if (dynamic_cast<KTextEdit*>(i.value())!=0l) dynamic_cast<KTextEdit*>(i.value())->setText(text);
        }

        if (m_Row->hasColumn(ColumnManager::Rating)) {
            ui->ratingWidget->setEnabled(true);
            ui->ratingWidget->setRating(m_Row->getIntValue(ColumnManager::Rating));
        }
        else {
            ui->ratingWidget->setDisabled(true);
            ui->ratingWidget->setRating(0);
        }
    }

    void TrackDetails::alteredRating (int rating)
    {
        m_Row->setData(ColumnManager::Rating,QVariant(rating));
    }

    void TrackDetails::rotatePane ()
    {
        int index=ui->toolBox->indexOf(ui->toolBox->currentWidget());
        ++index;
        if (index>=ui->toolBox->count()) {
            index=0;
        }
        ui->toolBox->setCurrentIndex(index);
    }

    ///////////////////////////////////////////////////////////////////////////////////

    void TrackDetails::onPageChanged (int index)
    {
        if (m_Extension!=0l && m_Row!=0l) {
            initCurrent(index);
        }
    }

    void TrackDetails::updateExtension ()
    {
        if (m_Row!=0l) {
            m_Extension=m_Row->getExtensions();

            m_StandardInitialized=m_ArtWorkInitialized=m_ExtendedAttribsInitialized=false;
            if (m_Extension!=0l) {
                initCurrent(ui->toolBox->currentIndex());
                ui->toolBox->setEnabled(true);
            }
            else {
                ui->toolBox->setDisabled(true);
            }
        }
    }

    void TrackDetails::initCurrent (int index)
    {
        switch (index) {
        case 0:
            if (!m_StandardInitialized) initOverview();
            m_StandardInitialized=true;
            break;
        case 1:
            if (!m_ArtWorkInitialized) {
                initLyrics();
                initImages();
            }
            m_ArtWorkInitialized=true;
            break;
        case 2:
            if (!m_ExtendedAttribsInitialized) initExtra();
            m_ExtendedAttribsInitialized=true;
            break;
        }
    }

    ///////////////////////////////////////////////////////////////////////////////////
    //
    // Lyrics support
    //
    ///////////////////////////////////////////////////////////////////////////////////

    void TrackDetails::initLyrics ()
    {
        m_Lyrics = m_Extension->getLyrics();
        if (m_Lyrics.size()==0) {
            m_Lyrics.append(TagExtension::Lyrics());
        }
        m_CurrentLyrics=m_Lyrics.begin();
        showLyrics();
    }

    void TrackDetails::showLyrics ()
    {
        ui->lyrics->setText((*m_CurrentLyrics).lyrics);
        ui->lyricsDescription->setText((*m_CurrentLyrics).description);
    }

    void TrackDetails::nextLyric ()
    {
        ++m_CurrentLyrics;
        if (m_CurrentLyrics==m_Lyrics.end()) m_CurrentLyrics=m_Lyrics.begin();
        showLyrics();
    }

    void TrackDetails::prevLyric ()
    {
        if (m_CurrentLyrics==m_Lyrics.begin()) m_CurrentLyrics=m_Lyrics.end();
        --m_CurrentLyrics;
        showLyrics();
    }

    ///////////////////////////////////////////////////////////////////////////////////
    //
    // Image support
    //
    ///////////////////////////////////////////////////////////////////////////////////

    void TrackDetails::initImages ()
    {
        m_ImageTypes = m_Extension->getAvailableImageTypes().toList();
        ui->imageType->clear();
        ui->imageType->addItems (m_ImageTypes);
        m_CurrentImage = m_ImageTypes.begin();
        m_ImageCount=checkMoveButtons();
        if (m_ImageCount>0) {
            nextAvailableImage();
        }
        else {
            showImage();
        }
    }

    int TrackDetails::checkMoveButtons ()
    {
        int count=0;
        if (m_ImageTypes.size()>1) {
            foreach (QString n, m_ImageTypes) {
                if (m_Extension->getImage(n)!=0l) count++;
            }
        }
        ui->prevAvailableImage->setEnabled(count>1);
        ui->nextAvailableImage->setEnabled(count>1);
        return count;
    }

    void TrackDetails::setImageType (int type)
    {
        m_CurrentImage=m_ImageTypes.begin()+type;
        showImage();
    }

    void TrackDetails::nextImage ()
    {
        ++m_CurrentImage;
        if (m_CurrentImage==m_ImageTypes.end()) m_CurrentImage=m_ImageTypes.begin();
        showImage();
    }

    void TrackDetails::nextAvailableImage ()
    {
        if (m_ImageTypes.isEmpty()) return;
        if (m_ImageTypes.size()==1) {
            m_CurrentImage=m_ImageTypes.begin();
        }
        else {
            bool found=false;
            while (!found) {
                ++m_CurrentImage;
                if (m_CurrentImage==m_ImageTypes.end()) m_CurrentImage=m_ImageTypes.begin();
                found=(m_Extension->getImage(*m_CurrentImage)!=0l);
            }
        }
        showImage();
    }

    void TrackDetails::prevAvailableImage ()
    {
        if (m_ImageTypes.isEmpty()) return;
        if (m_ImageTypes.size()==1) {
            m_CurrentImage=m_ImageTypes.begin();
        }
        else {
            bool found=false;
            while (!found) {
                if (m_CurrentImage==m_ImageTypes.begin()) m_CurrentImage=m_ImageTypes.end();
                --m_CurrentImage;
                found=m_Extension->getImage(*m_CurrentImage)!=0l;
            }
        }
        showImage();
    }

    void TrackDetails::prevImage ()
    {
        if (m_CurrentImage==m_ImageTypes.begin()) m_CurrentImage=m_ImageTypes.end();
        --m_CurrentImage;
        showImage();
    }

    void TrackDetails::showImage ()
    {
        ui->imageType->setCurrentIndex (m_CurrentImage-m_ImageTypes.begin());
        const QImage* img=m_Extension->getImage(*m_CurrentImage);
        if (img!=0l) {
            QPixmap pix;
            const int w =ui->image->lineWidth()+ui->image->midLineWidth()+2;
            const QSize size=ui->image->size()-QSize(w,w);
            if (img->width()>size.width() || img->height()>size.height()) {
                pix=QPixmap::fromImage(img->scaled(size,Qt::KeepAspectRatio));
            }
            else {
                pix=QPixmap::fromImage(*img);
            }
            ui->image->setPixmap (pix);
            const QSize imgSize = img->size();
            QString sizeText = QString::number(imgSize.width())+"x"+QString::number(imgSize.height());
            ui->sizeLabel->setText (sizeText);
            ui->imageDescription->setText(img->text("Desc"));
            ui->imageDescription->setEnabled(true);
            emit displaysImage(true);
        }
        else {
            ui->image->setText (i18nc("Displayed when no image was found for a certain type","No image"));
            ui->sizeLabel->setText ("");
            ui->imageDescription->setEnabled(false);
            ui->imageDescription->setText("");
            emit displaysImage(false);
        }
    }

    void TrackDetails::removeImage ()
    {
        m_Extension->rmImage(*m_CurrentImage);
    }

    void TrackDetails::addImage ()
    {
        QString file = FileDialog::getLoadFileName("*.png *.jpg *.gif *.bmp",i18n("Load image"));
        if (file!="") {
            QImage img = CoverSearch::Cover::loadImage(file);
            m_Extension->setImage(*m_CurrentImage,img);
            checkMoveButtons();
        }
    }

    void TrackDetails::imageDescriptionChanged ()
    {
        QImage img = *m_Extension->getImage(*m_CurrentImage);
        img.setText("Desc",ui->imageDescription->text());
        m_Extension->setImage(*m_CurrentImage,img);
    }

    void TrackDetails::searchForCover ()
    {
        if (m_Row!=0l) {
            TagItem::List list;
            list << m_Row;
            CoverSearch::SimpleSelector::search(list,
                                                this,
                                                SLOT(onCoverSelected(QString, QString, QImage)));
        }
    }

    void TrackDetails::onCoverSelected (const QString& , const QString& , const QImage& cover)
    {
        m_Extension->setImage(*m_CurrentImage,cover);
        checkMoveButtons();
        showImage();
    }

    void TrackDetails::droppedUrlList (const QList<QUrl>& urls)
    {
        QImage img = CoverSearch::Cover::loadImage(urls.first());
        m_Extension->setImage(*m_CurrentImage,img);
        checkMoveButtons();
        showImage();
    }

    void TrackDetails::droppedImage (const QImage& image)
    {
        m_Extension->setImage(*m_CurrentImage,image);
        checkMoveButtons();
    }

    void TrackDetails::cutImage ()
    {
        copyImage();
        removeImage();
    }

    void TrackDetails::copyImage ()
    {
        QClipboard* clipboard = QApplication::clipboard();
        clipboard->setImage(*m_Extension->getImage(*m_CurrentImage));
    }

    void TrackDetails::pasteImage ()
    {
        QClipboard* clipboard = QApplication::clipboard();
        m_Extension->setImage(*m_CurrentImage, clipboard->image());
    }

    void TrackDetails::imagePopupMenu (const QPoint& pos)
    {
        const QImage* img=m_Extension->getImage(*m_CurrentImage);
        if (img!=0l && !img->isNull()) {
            ui->coverMenu->menu()->popup(ui->coverMenu->mapToGlobal(pos));
        }
    }

    ///////////////////////////////////////////////////////////////////////////////////
    //
    // Generic attributes
    //
    ///////////////////////////////////////////////////////////////////////////////////

    void TrackDetails::initExtra ()
    {
        ui->extraPropertiesTable->setExtension(m_Extension);
    }


}

#include "trackdetails.moc"
