/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include <boost/bimap.hpp>
#include <boost/bimap/vector_of.hpp>
#include <boost/bimap/multiset_of.hpp>

#include "messageview.h"

#include "model/tagitem.h"
#include "ui_messageview.h"
#include "options/preferences.h"

#include "aid.h"

using namespace std;

using namespace boost::bimaps;

namespace Texel {

    MessageModel::Item::Item (int _level, const QString& _message, Texel::TagItem* _tag)
        :level(_level),
        message(_message),
        tag(_tag)
    {
        if (tag!=0l) {
            if (level==Enums::MessageWarning) {
                tag->setWarnings(tag->getWarnings()+1);
            }
            else if (level==Enums::MessageError) {
                tag->setErrors(tag->getErrors()+1);
            }
        }
    }

    MessageModel::Item::~Item ()
    {
        if (tag!=0l) {
            if (level==Enums::MessageWarning) {
                tag->setWarnings(tag->getWarnings()-1);
            }
            else if (level==Enums::MessageError) {
                tag->setErrors(tag->getErrors()-1);
            }
        }
    }

    struct MessageModel::Private
    {
        int size (Enums::MessageTypes level)
        {
            switch (level)
            {
            case Enums::MessageError: return m_Errors.size();
            case Enums::MessageWarning: return m_Warnings.size();
            case Enums::MessageInfo: return m_AllLogs.size();
            default:
                return 0;
            }
        }

        Item* get (int row, Enums::MessageTypes level) const
        {
            switch (level)
            {
            case Enums::MessageError:   return m_Errors[row];
            case Enums::MessageWarning: return m_Warnings[row];
            case Enums::MessageInfo:    return m_AllLogs.at(row).get<LogEntry>();
            default: return 0l;
            }
        }

        void clear ()
        {
            m_Warnings.clear();
            m_Errors.clear();
            for (LogEntries::map_by<LogEntry>::iterator iter=m_AllLogs.by<LogEntry>().begin();
                                      iter!=m_AllLogs.by<LogEntry>().end();
                                      iter++) {
                delete iter->get<LogEntry>();
            }
            m_AllLogs.clear();
        }

        Enums::MessageTypes pop_first ()
        {
            Item* logentry=m_AllLogs.by<LogEntry>().front().get<LogEntry>();
            m_AllLogs.by<LogEntry>().pop_front();

            Enums::MessageTypes newMaxLevel=Enums::MessageNoChange;
            if (!m_Warnings.isEmpty() && m_Warnings.first()==logentry) {
                m_Warnings.pop_front();
                if (!m_Errors.isEmpty() && m_Errors.first()==logentry) {
                    m_Errors.pop_front();
                    if (m_Errors.size()==0 && m_Warnings.size()>0) {
                        newMaxLevel=Enums::MessageWarning;
                    }
                }
                else if (m_Warnings.size()==0) {
                    newMaxLevel=Enums::MessageInfo;
                }
            }
            else if (m_AllLogs.size()==0) {
                newMaxLevel=Enums::MessageNone;
            }

            delete logentry;
            return newMaxLevel;
        }

        Enums::MessageTypes push_last (Item *item)
        {
            Enums::MessageTypes newMaxLevel=Enums::MessageNoChange;

            m_AllLogs.push_back (LogEntries::value_type(item,item->tag));
            if (item->level!=Enums::MessageInfo) {
                m_Warnings << item;
                if (item->level==Enums::MessageError) {
                    m_Errors << item;
                    if (m_Errors.size()==1) {
                       newMaxLevel=Enums::MessageError;
                    }
                }
                else if (m_Errors.size()==0 && m_Warnings.size()==1) {
                    newMaxLevel=Enums::MessageWarning;
                }
            }
            else if (m_AllLogs.size()==1) {
                newMaxLevel=Enums::MessageInfo;
            }

            return newMaxLevel;
        }

        Item::List getMessagesForItem (Texel::TagItem* item) const
        {
            Item::List result;
            for (LogEntries::map_by<TagItemEntry>::const_iterator iter=m_AllLogs.by<TagItemEntry>().lower_bound(item);
                 iter!=m_AllLogs.by<TagItemEntry>().upper_bound(item);iter++)
            {
                result << iter->get<LogEntry>();
            }
            return result;
        }

        struct LogEntry;
        struct TagItemEntry;

        typedef bimap<vector_of< tagged<Item*, LogEntry> >,
                      multiset_of< tagged<Texel::TagItem*, TagItemEntry> >
                      > LogEntries;

        LogEntries m_AllLogs;

        Item::List m_Warnings;
        Item::List m_Errors;
    };

    MessageModel::MessageModel (QWidget* parent)
        : QAbstractListModel (parent),
          p(new Private),
         m_DisplayedLevel(Enums::MessageInfo),
         m_CurrentMaxLevel(Enums::MessageNone)
    {
        connect (Preferences::instance(), SIGNAL(configChanged()), this, SLOT(checkConfig()));
    }

    MessageModel::~MessageModel ()
    {
        delete p;
    }

    void MessageModel::checkConfig ()
    {
        Enums::MessageTypes oldCurrentLevel=m_CurrentMaxLevel;

        while (p->m_AllLogs.size()>Preferences::instance()->behaviour.m_MaxLogEntries) {
            const bool visible=p->get(0,Enums::MessageInfo)->level<=m_DisplayedLevel;
            if (visible) beginRemoveRows (QModelIndex(), 0, 0);
            Enums::MessageTypes r=p->pop_first();
            if (r!=Enums::MessageNoChange) m_CurrentMaxLevel=r;
            if (visible) endRemoveRows();
        }

        if (m_CurrentMaxLevel!=oldCurrentLevel) {
            emit newMaxFaultLevel(m_CurrentMaxLevel);
        }
    }

    int MessageModel::rowCount (const QModelIndex& ) const
    {
        return p->size(m_DisplayedLevel);
    }

    int MessageModel::columnCount (const QModelIndex& ) const
    {
        return 2;
    }

    void MessageModel::clear ()
    {
        beginRemoveRows (QModelIndex(), 0, p->size(Enums::MessageInfo));
        p->clear();
        endRemoveRows();
        emit newMaxFaultLevel(Enums::MessageNone);
    }

    MessageModel::Item::List MessageModel::getMessagesForItem (Texel::TagItem* item) const
    {
        return p->getMessagesForItem(item);
    }

    QVariant MessageModel::data ( const QModelIndex & index, int role) const
    {
        if (index.isValid()) {
            return data(p->get(index.row(),m_DisplayedLevel),index.column(),role);
        }
        return QVariant();
    }

    QVariant MessageModel::data (Item* item, int col, int role) const
    {
        if (col==0) {
            if (role==Qt::DisplayRole) return QVariant(item->message);
            if (role==Qt::DecorationRole) return QVariant(getIcon(item->level));
            if (role==Qt::ToolTipRole) return QVariant(item->message);
        }
        else if (col==1) {
            if (role==Qt::DisplayRole && item->tag!=0L) {
                return QVariant(item->tag->data(ColumnManager::Filename,Columns::Column::RoleDisplay));
            }
            if (role==Qt::ToolTipRole && item->tag!=0L) {
                return QVariant(item->tag->getAbsolutFilename());
            }
        }
        return QVariant();
    }

    void MessageModel::setCurrentMaxLevel (int current)
    {
        beginResetModel();
        m_DisplayedLevel=(Enums::MessageTypes)current;
        endResetModel();
    }

    void MessageModel::message (Enums::MessageTypes level, const QString& message, TagItem* item)
    {
        Item* i=new Item(level, message, item);
        Enums::MessageTypes oldCurrentLevel=m_CurrentMaxLevel;

        const bool visible = level<=m_DisplayedLevel;
        if (visible) beginInsertRows(QModelIndex(),p->size(m_DisplayedLevel),p->size(m_DisplayedLevel));
        Enums::MessageTypes r=p->push_last(i);
        if (r!=Enums::MessageNoChange) m_CurrentMaxLevel=r;
        if (visible) endInsertRows();

        while (p->m_AllLogs.size()>Preferences::instance()->behaviour.m_MaxLogEntries) {
            const bool visible=p->get(0,Enums::MessageInfo)->level<=m_DisplayedLevel;
            if (visible) beginRemoveRows (QModelIndex(), 0, 0);
            Enums::MessageTypes r=p->pop_first();
            if (r!=Enums::MessageNoChange) m_CurrentMaxLevel=r;
            if (visible) endRemoveRows();
        }

        if (m_CurrentMaxLevel!=oldCurrentLevel) {
            emit newMaxFaultLevel(m_CurrentMaxLevel);
        }
    }

    void MessageModel::error(const QString& m)
    {
        message (Enums::MessageError,m,0L);
    }

    QPixmap MessageModel::getIcon(int level)
    {
        switch (level) {
        case Enums::MessageInfo: return KIconLoader::global()->loadIcon("dialog-information",KIconLoader::Small);
        case Enums::MessageWarning: return KIconLoader::global()->loadIcon("warning",KIconLoader::Small);
        case Enums::MessageError: return KIconLoader::global()->loadIcon("alert",KIconLoader::Small);
        }
        return QPixmap();
    }

}

#include "messageview.moc"
