/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "problemawaredelegate.h"
#include "options/enums.h"
#include "model/columnmanager.h"
#include "model/texeltagmodel.h"

#include <QLineEdit>
#include <QPoint>
#include <QPainter>
#include <QApplication>

#include <algorithm>

#include <kicon.h>

#include "aid.h"

// #define MEASURE_LATENCY

#ifdef MEASURE_LATENCY
#include <sys/time.h>
#endif

using namespace std;

namespace Texel {

    ProblemAwareDelegate::ProblemAwareDelegate (QWidget* parent)
        : QItemDelegate(parent),
          time(0), count(0)
    {
    }

    ProblemAwareDelegate::~ProblemAwareDelegate ()
    {
    }

    void ProblemAwareDelegate::myDrawDisplay (QPainter *painter, const QStyleOptionViewItem& option, const QRect&, const QModelIndex& index) const
    {
        const QStyleOptionViewItemV4* o=qstyleoption_cast<const QStyleOptionViewItemV4*>(&option);

        textOpt.setWrapMode(QTextOption::WrapAtWordBoundaryOrAnywhere);

        QTextLayout layout;
        layout.setCacheEnabled(true);
        layout.setFont(index.data(Qt::FontRole).value<QFont>());
        layout.setTextOption(textOpt);

        // Format
        format.setBackground(o->backgroundBrush);
        format.setForeground(index.data(Qt::ForegroundRole).value<QBrush>().color());
        format.setFont(index.data(Qt::FontRole).value<QFont>());

        const QVariant value = index.data(Enums::RoleProblem);
        Enums::ProblemCategories problem=Enums::ProblemUnknown;
        if (!value.isNull()) {
            problem = static_cast<Enums::ProblemCategories>((value).toInt());
            switch (problem) {
            case Enums::ProblemError:
                format.setUnderlineStyle(QTextCharFormat::WaveUnderline);
                format.setUnderlineColor(QColor(200,0,0));
                break;
            case Enums::ProblemWarning:
                format.setUnderlineStyle(QTextCharFormat::WaveUnderline);
                format.setUnderlineColor(QColor(200,200,0));
                break;
            default:
                ;// just to keep compiler happy
            }
        }

        layout.setText(index.data(Qt::DisplayRole).toString());

        layout.beginLayout();
        qreal height=0.0;
        while (true) {
            QTextLine line=layout.createLine();
            if (!line.isValid())
                break;
            line.setLineWidth(option.rect.width());
            line.setPosition(QPointF(0, height));
            height+=line.height();
            if (height+line.height()>option.rect.height()) break;
        }
        layout.endLayout();

        QVector<QTextLayout::FormatRange> formats;
        QTextLayout::FormatRange r;
        r.format=format;
        r.start=0;
        r.length=layout.text().length();
        formats << r;

        QVariant highlights = index.data(Enums::RoleHighlights);
        if (highlights.isValid()) {
            format.setBackground(QColor(0,240,0));
            QList<QVariant> highlightRanges = highlights.toList();
            foreach (QVariant r,highlightRanges) {
                QTextLayout::FormatRange formatRange;
                QPoint range=r.value<QPoint>();
                formatRange.format=format;
                formatRange.start=range.x();
                formatRange.length=range.y()-range.x();
                formats << formatRange;
            }
        }

        // Top margin to center text in the cell
        const int topMargin = (option.rect.height()-height)/2;
        layout.draw(painter,option.rect.topLeft()+QPoint(0,topMargin),formats,option.rect);
    }

    void ProblemAwareDelegate::paint (QPainter* painter, const QStyleOptionViewItem& option, const QModelIndex& index) const
    {
#ifdef MEASURE_LATENCY
        timespec ts;
        clock_gettime(CLOCK_REALTIME, &ts);
#endif

        QRect r = option.rect;

        drawBackground (painter, option, index);
        drawFocus(painter, option, r);
        if (index.model()->flags(index).testFlag(Qt::ItemIsUserCheckable)) {
            QRect r1 = QApplication::style()->subElementRect(QStyle::SE_CheckBoxIndicator,&option);
            r1.moveCenter(r.center());
            drawCheck (painter, option, r1, (Qt::CheckState)index.data(Qt::CheckStateRole).toInt());
        }
        else {
            myDrawDisplay(painter, option, r, index);
        }

        const QVariant value = index.data(Enums::RoleProblem);
        if (!value.isNull()) {
            Enums::ProblemCategories problem = static_cast<Enums::ProblemCategories>((value).toInt());
            if (problem==Enums::ProblemNone) {
                const QRect rect(option.rect.x()+option.rect.width()-11,option.rect.y()+option.rect.height()-11,10,10);
                painter->drawPixmap(rect,KIconLoader::global()->loadIcon("dialog-ok",KIconLoader::Small));
            }
        }

#ifdef MEASURE_LATENCY
        timespec ts1;
        clock_gettime(CLOCK_REALTIME, &ts1); // Works on Linux

        time+=(ts1.tv_sec-ts.tv_sec)*1000000000+ts1.tv_nsec-ts.tv_nsec;
        count++;
        if ((count%1000)==0) {
            cout << "Average time :" << time/1000000000.0 << "ms" << endl;
            time=count=0;
        }
#endif
    }

    QSize ProblemAwareDelegate::sizeHint (const QStyleOptionViewItem&, const QModelIndex&) const
    {
        return QSize(100,80);
    }

}

#ifdef MEASURE_LATENCY
#undef MEASURE_LATENCY
#endif

#include "problemawaredelegate.moc"

