/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "tagcomparer.h"
#include "ui_tagcomparer.h"

#include <QStandardItemModel>
#include <QMenu>

#include "model/tagitem.h"
#include "model/texeltagmodel.h"

namespace Texel {

    TagComparer::TagComparer(TagModel* model_, QWidget *parent) :
        QWidget(parent),
        ui(new Ui::TagComparer),
        m_Model(model_),
        m_Executor(0l),
        m_FilterChecked(false)
    {
        ui->setupUi(this);

        setWindowIcon(KIconLoader::global()->loadIcon("document-save",KIconLoader::Toolbar));

        QStandardItemModel* model = new QStandardItemModel(ui->comparisonList);
        connect (model, SIGNAL(dataChanged(QModelIndex,QModelIndex)), this, SLOT(onChangedCheckState(QModelIndex,QModelIndex)));

        model->setColumnCount(3);
        QStringList headers;
        headers << i18n("Keep\noriginal") << i18n("New file") << i18n("Original");
        model->setHorizontalHeaderLabels(headers);
        ui->comparisonList->setModel (model);
        ui->comparisonList->horizontalHeader()->setResizeMode(0,QHeaderView::ResizeToContents);
        ui->comparisonList->horizontalHeader()->setResizeMode(1,QHeaderView::Stretch);
        ui->comparisonList->horizontalHeader()->setResizeMode(2,QHeaderView::Stretch);

        m_LocalModel=new TagModel(this);

        connect (ui->comparisonList, SIGNAL(clicked(QModelIndex)), this, SLOT(onComparePairClick(QModelIndex)));
        connect (ui->comparisonList->selectionModel(), SIGNAL(currentChanged(QModelIndex,QModelIndex)),
                                                 this, SLOT(onComparePairClick(QModelIndex)));

        QMenu* selectionMenu = new QMenu(ui->selectionMenu);
        ui->selectionMenu->setMenu(selectionMenu);
        selectionMenu->addAction (i18n("Unset all"),this, SLOT(unsetAll()));
        selectionMenu->addAction (i18n("Keep originals"), this, SLOT(keepOriginals()));
        selectionMenu->addAction (i18n("Overwrite original"), this, SLOT(overwriteOriginals()));
        selectionMenu->addAction (i18n("Keep oldest file"), this, SLOT(keepOlderFiles()));
        selectionMenu->addAction (i18n("Keep newest file"), this, SLOT(keepNewerFiles()));
        selectionMenu->addAction (i18n("Keep highest quality"));

        connect (ui->filterButton, SIGNAL(toggled(bool)), this, SLOT(filterRows(bool)));
        connect (ui->filterEdit, SIGNAL(returnPressed()), this, SLOT(filterByText()));
    }

    TagComparer::~TagComparer()
    {
        delete ui;
        delete m_LocalModel;
        delete m_Executor;
    }

    void TagComparer::onAccept ()
    {
        if (m_Executor!=0l) {
            TagItem::Set checked;
            TagItem::Set unchecked;

            QStandardItemModel* model = dynamic_cast<QStandardItemModel*>(ui->comparisonList->model());
            for (int i=0;i<model->rowCount();i++) {
                TagItem* item=model->item(i,1)->data(Qt::UserRole+1).value<TagItem*>();
                if (model->item(i,0)->checkState()==Qt::Checked) {
                    checked.insert(item);
                }
                else {
                    unchecked.insert(item);
                }
            }

            m_Executor->accept(checked, unchecked);
        }
    }

    void TagComparer::onReject ()
    {
        if (m_Executor!=0l) {
            TagItem::Set items;
            QStandardItemModel* model = dynamic_cast<QStandardItemModel*>(ui->comparisonList->model());
            for (int i=0;i<model->rowCount();i++) {
                items.insert(model->item(i,1)->data(Qt::UserRole+1).value<TagItem*>());
            }
            m_Executor->reject (items);
        }
    }

    void TagComparer::changeEvent(QEvent *e)
    {
        QWidget::changeEvent(e);
        switch (e->type()) {
        case QEvent::LanguageChange:
            ui->retranslateUi(this);
            break;
        default:
            break;
        }
    }

    void TagComparer::add (TagItem* first, const QString& filename)
    {
        QStandardItemModel* model = dynamic_cast<QStandardItemModel*>(ui->comparisonList->model());

        TagItem* second = m_Model->getRowByOrgFilename(filename);
        if (second==0l) {
            second=TagItem::createTagItem(m_LocalModel,filename);
        }

        const bool emptyModel = model->rowCount()==0;

        QList<QStandardItem*> row;
        row << new QStandardItem();
        row.last()->setCheckable(true);
        row.last()->setTristate(true);
        row.last()->setCheckState(Qt::PartiallyChecked);
        row.last()->setTextAlignment(Qt::AlignHCenter);
        row << new QStandardItem(first->getAbsolutFilename(Columns::Column::RoleDisplay));
        row.last()->setData(QVariant::fromValue(first),Qt::UserRole+1);
        row.last()->setToolTip(first->getAbsolutFilename(Columns::Column::RoleDisplay));
        row << new QStandardItem(filename);
        row.last()->setData(QVariant::fromValue(second),Qt::UserRole+1);
        row.last()->setToolTip(second->getAbsolutFilename());
        model->appendRow(row);

        if (emptyModel) {
            ui->comparisonList->selectRow(0);
        }
    }

    void TagComparer::onComparePairClick (const QModelIndex& idx)
    {
        const int row = idx.row();
        TagItem* item=idx.model()->data(idx.model()->index(row,1),Qt::UserRole+1).value<TagItem*>();
        if (item!=0l) ui->tags->setTagItem(1,item);
        item=idx.model()->data(idx.model()->index(row,2),Qt::UserRole+1).value<TagItem*>();
        if (item!=0l) ui->tags->setTagItem(2,item);
    }

    bool TagComparer::isEmpty() const
    {
        return ui->comparisonList->model()->rowCount()==0;
    }

    void TagComparer::unsetAll ()
    {
        setCheckStateToAll(Qt::PartiallyChecked);
    }

    void TagComparer::keepOriginals ()
    {
        setCheckStateToAll(Qt::Checked);
    }

    void TagComparer::overwriteOriginals ()
    {
        setCheckStateToAll(Qt::Unchecked);
    }

    void TagComparer::setCheckStateToAll (Qt::CheckState state)
    {
        QStandardItemModel* model = dynamic_cast<QStandardItemModel*>(ui->comparisonList->model());

        for (int i=0;i<model->rowCount();i++) {
            model->item(i)->setCheckState(state);
        }
    }

    void TagComparer::keepOlderFiles ()
    {
        QStandardItemModel* model = dynamic_cast<QStandardItemModel*>(ui->comparisonList->model());

        for (int i=0;i<model->rowCount();i++) {
            if (isOlderFile(model->item(i,1),model->item(i,2))) {
                model->item(i)->setCheckState(Qt::Checked);
            }
            else {
                model->item(i)->setCheckState(Qt::Unchecked);
            }
        }
    }

    void TagComparer::keepNewerFiles ()
    {
        QStandardItemModel* model = dynamic_cast<QStandardItemModel*>(ui->comparisonList->model());

        for (int i=0;i<model->rowCount();i++) {
            if (isOlderFile(model->item(i,1),model->item(i,2))) {
                model->item(i)->setCheckState(Qt::Unchecked);
            }
            else {
                model->item(i)->setCheckState(Qt::Checked);
            }
        }
    }

    bool TagComparer::isOlderFile (QStandardItem* item1, QStandardItem* item2)
    {
        QFileInfo info1(item1->text());
        QFileInfo info2(item2->text());

        return info1.created()<info2.created();
    }

    void TagComparer::filterByText ()
    {
        QString text=ui->filterEdit->text();

        if (text.isEmpty()) {
            showAll();
        }
        else {
            ui->filterButton->setChecked(false);
            QStandardItemModel* model = dynamic_cast<QStandardItemModel*>(ui->comparisonList->model());
            for (int i=0;i<model->rowCount();i++) {
                if (!model->item(i,1)->text().contains(text) && !model->item(i,2)->text().contains(text)) {
                    ui->comparisonList->hideRow(i);
                }
            }
        }
    }

    void TagComparer::showAll ()
    {
        QStandardItemModel* model = dynamic_cast<QStandardItemModel*>(ui->comparisonList->model());
        for (int i=0;i<model->rowCount();i++) {
            ui->comparisonList->showRow(i);
        }
    }

    void TagComparer::filterRows (bool hide)
    {
        QStandardItemModel* model = dynamic_cast<QStandardItemModel*>(ui->comparisonList->model());

        if (hide) {
            ui->filterEdit->setText("");
            for (int i=0;i<model->rowCount();i++) {
                if (model->item(i)->checkState()!=Qt::PartiallyChecked) {
                    ui->comparisonList->hideRow(i);
                }
            }
        }
        else {
            showAll();
        }

        m_FilterChecked=hide;
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////
    //
    // Care for toggled check states when filter is applied. Rows are hidden / shown accordingly
    //
    ////////////////////////////////////////////////////////////////////////////////////////////////////////

    void TagComparer::onChangedCheckState (const QModelIndex& from, const QModelIndex& to)
    {
        if (m_FilterChecked) {
            QStandardItemModel* model = dynamic_cast<QStandardItemModel*>(ui->comparisonList->model());
            if (from.column()==0) {
                for (int i=from.row();i<=to.row();i++) {
                    if (model->item(i)->checkState()!=Qt::PartiallyChecked) {
                        ui->comparisonList->hideRow(i);
                    }
                    else {
                        ui->comparisonList->showRow(i);
                    }
                }
            }
        }
    }
}


#include "tagcomparer.moc"
