/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "texelheaderview.h"
#include "model/texeltagmodel.h"
#include "model/tagmodelproxy.h"
#include "options/enums.h"
#include "util/widgets/tooltip.h"
#include "texeltablewidget.h"

#include <QTextEdit>
#include <QTextDocument>
#include <QPainter>
#include <QHelpEvent>
#include <QTableView>

#include "aid.h"

using namespace std;

Q_DECLARE_METATYPE (QTextDocument*)

namespace Texel {

TexelHeaderView::TexelHeaderView (QWidget* parent)
: super(Qt::Vertical,parent),
  x(0)
{
    setClickable(true);
    TableWidget* table=dynamic_cast<TableWidget*>(parent);
    if (table!=0l) {
        connect (this, SIGNAL(sectionDoubleClicked(int)),table,SLOT(selectAllInSameDirectory(int)));
    }
}

TexelHeaderView::~TexelHeaderView()
{
}

void TexelHeaderView::onDuplicates (bool dup)
{
    x=dup?10:0;
    emit geometriesChanged ();
}

void TexelHeaderView::paintSection (QPainter* painter, const QRect& rect, int logicalIndex) const
{
    QStyleOptionFrameV3 option;
    option.initFrom(this);
    option.rect=rect;
    option.state|=QStyle::State_Raised;
    option.lineWidth=2;
    option.midLineWidth=1;
    option.features=QStyleOptionFrameV2::Flat;
    option.frameShape=QFrame::StyledPanel;
    style()->drawPrimitive(QStyle::PE_Frame, &option, painter);

    painter->save();
    QVariant icons = model()->headerData(logicalIndex, Qt::Vertical, Enums::RoleIcons);
    if (icons.isValid() && icons.type()==QVariant::List) {
        QList<QVariant> iconList=icons.toList();

        // Determine icon sizes. With three icons the icons are compressed
        int w = style()->pixelMetric(QStyle::PM_ListViewIconSize);
        if (iconList.size()==3) {
            w=(2*w)/3;
        }

        for (int i=0;i<iconList.size();i++) {
            const QIcon icon = iconList[i].value<QIcon>();
            icon.paint(painter,rect.left()+2+i*(w+1),
                       rect.top()+(sectionSize(logicalIndex)-style()->pixelMetric(QStyle::PM_ListViewIconSize))/2,
                       w,
                       style()->pixelMetric(QStyle::PM_ListViewIconSize));
        }
    }
    painter->restore();

    // takes care of a marker for modified rows
    if (model()->headerData(logicalIndex, Qt::Vertical, TagModel::ChangedRole).toBool()==true)
    {
        QRect marker(rect.topRight()-QPoint(3,-1),rect.bottomRight()-QPoint(1,1));
        painter->fillRect (marker,QColor(200,10,10));
    }
}

void TexelHeaderView::headerDataChanged (int first, int last)
{
    super::headerDataChanged(Qt::Vertical,first,last);
}

QSize TexelHeaderView::sizeHint () const
{
    return QSize(style()->pixelMetric(QStyle::PM_ListViewIconSize)*2+4,defaultSectionSize());
}

bool TexelHeaderView::viewportEvent (QEvent* event)
{
    if (event->type()==QEvent::ToolTip)
    {
        QHelpEvent* e = dynamic_cast<QHelpEvent*>(event);
        QTextDocument* doc = model()->headerData(logicalIndexAt(e->pos()),Qt::Vertical,Qt::ToolTipRole).value<QTextDocument*>();
        if (doc!=0l) {
            doc->adjustSize();
            QTextEdit* tooltip=new QTextEdit(this);
            tooltip->setGeometry(0,0,doc->size().width()+5,doc->size().height()+5);

            doc->setParent(tooltip);

            tooltip->setDocument(doc);
            tooltip->setReadOnly(true);
            tooltip->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
            tooltip->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
            QRect box(static_cast<QHelpEvent*>(event)->globalPos(), doc->size().toSize());
            Tooltip::show (tooltip, static_cast<QHelpEvent*>(event)->globalPos(), box, this);
        }
    }

    return super::viewportEvent (event);
}

void TexelHeaderView::onDoubleClick(int row)
{
    QTableView* table=dynamic_cast<QTableView*>(parent());
    if (table!=0l) {
        //table->selectionModel()->clearSelection();
        TagModelProxy* proxy = static_cast<TagModelProxy*>(model());
        DirTree::Dir* dir = proxy->getRow(row)->getPathPointer(Columns::Column::RoleDisplay);
        foreach (TagItem* item, dir->getTagItems()) {
            QVector<int> aRow=proxy->getRow(item);
            if (!aRow.isEmpty()) {
                table->selectionModel()->select(model()->index(aRow.first(),0),QItemSelectionModel::Select|QItemSelectionModel::Rows);
            }
        }
    }
}

}

#include "texelheaderview.moc"
