/*************************************************************************

    Texel - tagging like excel ; MP3 Tagger
    Copyright (C) 2009-2013 Jens Ketterer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************************/

#include "texeltablewidget.h"

#include "view/texelheaderview.h"
#include "view/popupmenucontroller.h"
#include "texel.h"

#include "model/completermodel.h"
#include "model/tagmodelproxy.h"
#include "model/columnmanager.h"
#include "model/texeltagmodel.h"

#include "options/preferences.h"

#include "func/metatagservices/freedb/querymenu.h"
#include "func/addcoverimages.h"
#include "func/coversearch/toolbox.h"
#include "func/metatagservices/musicbrainz/puid.h"

#include <QAction>
#include <QKeySequence>
#include <QKeyEvent>
#include <QIntValidator>
#include <QHeaderView>
#include <QMenu>
#include <QVarLengthArray>
#include <QPixmap>
#include <QImage>
#include <QPainter>
#include <QLineEdit>
#include <QHeaderView>
#include <QCompleter>

#include <KUrl>
#include <kdebug.h>
#include <kmessagebox.h>
#include <kstandardaction.h>
#include <kaction.h>

#include <set>
#include <auto_ptr.h>
#include <iostream>

using namespace std;

namespace Texel {

    /** Allows to either enter a number or a formula (beginning with a '=').
  * This class is better than using a regex, as the number range can be better
  * specified.
  */

    class FormularIntValidator: public QIntValidator
    {
    public:
        FormularIntValidator (QObject* parent, int upper)
            : QIntValidator(0,upper, parent)
        {
        }

        virtual ~FormularIntValidator ()
        {
        }

        virtual QValidator::State validate (QString& input, int& pos ) const
        {
            if (input[0]=='=') {
                return QValidator::Acceptable;
            }
            else return QIntValidator::validate(input,pos);
        }
    };

    Del::Del (QWidget* parent)
        : ProblemAwareDelegate(parent)
    {
    }

    Del::~Del ()
    {
    }

    QWidget* Del::createEditor (QWidget * parent, const QStyleOptionViewItem & option, const QModelIndex & index) const
    {
        QWidget* editor = QItemDelegate::createEditor(parent,option,index);

        lineEditor = dynamic_cast<QLineEdit*>(editor);
        QCompleter* completer=dynamic_cast<const TagModelProxy*>(index.model())->getCompleter(index);
        if (lineEditor!=0L && completer!=0L) {
            lineEditor->setCompleter(completer);
            connect (lineEditor, SIGNAL(textChanged(QString)), this, SLOT(checkForFormular(QString)));

            QAction* nextCompletion = new QAction(lineEditor);
            nextCompletion->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_Down));
            nextCompletion->setShortcutContext(Qt::WidgetWithChildrenShortcut);
            connect (nextCompletion, SIGNAL(triggered()), this, SLOT(nextCompletion()));
            lineEditor->addAction(nextCompletion);

            QAction* prevCompletion = new QAction(lineEditor);
            prevCompletion->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_Up));
            prevCompletion->setShortcutContext(Qt::WidgetWithChildrenShortcut);
            connect (prevCompletion, SIGNAL(triggered()), this, SLOT(prevCompletion()));
            lineEditor->addAction(prevCompletion);

            QAction* setAround = new QAction(lineEditor);
            setAround->setShortcut(QKeySequence(Qt::ALT + Qt::Key_Return));
            prevCompletion->setShortcutContext(Qt::WidgetWithChildrenShortcut);
            connect (setAround, SIGNAL(triggered()), this, SLOT(setAround()));
            lineEditor->addAction(setAround);
            m_Index=index;
        }

        return editor;
    }

    void Del::checkForFormular (const QString& text)
    {
        CompleterModel* model = dynamic_cast<CompleterModel*>(lineEditor->completer()->model());
        if (model!=0L) {
            if (model->getActiveSet()==1 && (text.isEmpty() || text[0]!='=')) {
                model->setActiveSet(0);
                lineEditor->completer()->complete();
            }
            else if (model->getActiveSet()==0 && !text.isEmpty() && text[0]=='=') {
                model->setActiveSet(1);
                lineEditor->completer()->complete();
            }
        }
    }

    void Del::iterCompletion (int inc)
    {
        QCompleter* completer=lineEditor->completer();

        if (completer!=0L) {
            int row=completer->currentRow()+inc;
            if (row<0) {
                row=completer->completionCount()-1;
            }
            if (completer->setCurrentRow(row)==false) {
                completer->setCurrentRow(0);
            }
            completer->complete();
        }
    }


    void Del::nextCompletion()
    {
        iterCompletion(1);
    }

    void Del::prevCompletion()
    {
        iterCompletion(-1);
    }

    void Del::setAround ()
    {
        const QVariant value = lineEditor->text();
        const QVariant oldValue = m_Index.data();
        QAbstractItemModel* model=const_cast<QAbstractItemModel*>(m_Index.model());
        int row = m_Index.row();

        for (int i=row+1;i<model->rowCount();i++) {
            QModelIndex idx=model->index(i,m_Index.column());
            if (model->data(idx)==oldValue) {
                model->setData(idx,value);
            }
            else break;
        }

        for (int i=row-1;i>=0;i--) {
            QModelIndex idx=model->index(i,m_Index.column());
            if (model->data(idx)==oldValue) {
                model->setData(idx,value);
            }
            else break;
        }

        lineEditor->parentWidget()->setFocus();
    }

    void Del::setEditorData (QWidget * editor, const QModelIndex& index) const
    {
        QLineEdit* e = dynamic_cast<QLineEdit*>(editor);
        if (e!=0L) {
            e->setText (index.model()->data(index).toString());
            TableWidget* w = dynamic_cast<TableWidget*>(parent());
            if (w!=0L && w->getEditBuddy()!=0L) {
                connect (editor,SIGNAL(textChanged(const QString&)),w->getEditBuddy(), SLOT(setText(const QString&)));
            }

            // TODO: Better in editor creation, not in data settings! Actually might be worth to create a new factory
            if (ColumnManager::instance()->getType(index.column())==QMetaType::Int) {
                const int threshold = (index.column()==ColumnManager::Year)?2100:(index.column()==ColumnManager::Rating)?5:255;
                e->setValidator(new FormularIntValidator(const_cast<Del*>(this), threshold));
            }
        }
        else QItemDelegate::setEditorData (editor, index);
    }

    namespace {
        inline TagModel* model_cast (QAbstractItemModel* model)
        {
            const TagModelProxy* proxy=dynamic_cast<TagModelProxy*>(model);
            return dynamic_cast<TagModel*>(proxy->sourceModel());
        }
    }

    TableWidget::TableWidget (QWidget* w)
        : super(w),
        m_EditBuddy(0L),
        m_popup(0L),
        m_Controler(0L),
        m_OldPosition(0l)
    {
        setVerticalHeader(new TexelHeaderView(this));

        viewport()->setAcceptDrops(true);
        setItemDelegate (new Del(this));
        QHeaderView* header = horizontalHeader();

        setContextMenuPolicy (Qt::CustomContextMenu);
        connect (this, SIGNAL(customContextMenuRequested(const QPoint&)), this, SLOT(contextMenu (const QPoint&)));
        header->setContextMenuPolicy (Qt::CustomContextMenu);
        connect (header, SIGNAL(customContextMenuRequested(const QPoint&)), this, SLOT(headerContextMenu (const QPoint&)));
        connect (header, SIGNAL(sectionDoubleClicked(int)), this, SLOT(selectColumn(int)));

        setSortingEnabled (true);
        header->setMovable(true);
    }

    void TableWidget::setModel (QAbstractItemModel* model)
    {
        super::setModel(model);
        TagModelProxy* m=dynamic_cast<TagModelProxy*>(model);
        if (m!=0L) {
            connect (m, SIGNAL(clearHeaders()), this, SLOT(clearHeaders()));
            connect (m, SIGNAL(addHeader(QList<int>)), this, SLOT(addHeader(QList<int>)));
            connect (m, SIGNAL(modelAboutToBeFiltered()), this, SLOT(onModelFilterStart()));
        }

        // Hack:
        // Move last row to the first position. Due to (array) internals of tagitem and columnmanager
        // this should be done on the virtual level
        QHeaderView* header = horizontalHeader();
        header->moveSection(ColumnManager::MarkForDeletion,0);
        header->resizeSection(ColumnManager::MarkForDeletion,20); //style()->sizeFromContents(QStyle::CT_CheckBox,&opt,QSize(24,24)).width());
    }

    void TableWidget::dataChanged(const QModelIndex &topLeft, const QModelIndex &bottomRight)
    {
        // Use QRect to simplify containment calculation
        const QRect rect(QPoint(topLeft.row(),topLeft.column()), QPoint(bottomRight.row(),bottomRight.column()));
        if (rect.contains(currentIndex().row(),currentIndex().column(),false)) {
            emit currentText(model()->data(currentIndex(),TagModel::ExtendedDisplayMode).toString());
        }
        super::dataChanged(topLeft,bottomRight);
    }

    void TableWidget::currentChanged (const QModelIndex& current, const QModelIndex& previous)
    {
        if (current.isValid()) {
            emit currentText(model()->data(current,TagModel::ExtendedDisplayMode).toString());
        }
        emit currentIndexChanged (current,previous);
        super::currentChanged (current,previous);
    }

    void TableWidget::reset ()
    {
        super::reset();
        if (m_OldPosition.get()!=0l) {
            if (model()->rowCount()>=m_OldPosition->m_Index.row()) {
                setCurrentIndex (model()->index(m_OldPosition->m_Index.row(),
                                                m_OldPosition->m_Index.column()));
            }
            else {
                setCurrentIndex (model()->index(model()->rowCount()-1,m_OldPosition->m_Index.column()));
            }
            m_OldPosition.reset();
        }
    }

    void TableWidget::currentEditingFinished()
    {
        const QLineEdit* e = dynamic_cast<const QLineEdit*>(sender());
        if (e!=0L) {
            model()->setData(currentIndex(),e->text());
        }
        setFocus(Qt::OtherFocusReason);
    }

    namespace {
        struct LessThan
        {
            bool operator() (const TagItem* a, const TagItem* b) const
            {
                return a->getRow()<b->getRow();
            }
        };
    }

    TagItem::List TableWidget::getSelectedRowsSorted (bool completeLinesOnly) const
    {
        TagModelProxy* proxy = dynamic_cast<TagModelProxy*>(model());

        TagItem::List result;
        if (completeLinesOnly) {
            foreach (QModelIndex index, selectionModel()->selectedRows(0)) {
                result << proxy->getRow(index.row());
            }
        }

        if (!completeLinesOnly && result.size()==0 && selectionModel()->hasSelection()) {
            // Not complete Rows are selected. Calculate by myself
            std::set<TagItem*> rows;
            foreach (QModelIndex ind, selectionModel()->selectedIndexes()) {
                rows.insert(proxy->getRow(ind.row()));
            }
            // Copy set to selectedRows list
            for (std::set<TagItem*>::iterator iter=rows.begin();iter!=rows.end();iter++) {
                result << *iter;
            }
        }

        qSort(result.begin(),result.end(),LessThan());
        return result;
    }

    void TableWidget::contextMenu (const QPoint& pos)
    {
        const QModelIndex index=indexAt(pos);
        if (index!=QModelIndex()) {
            delete m_popup;
            m_popup = new QMenu(this);
            PopupMenuController* controller = new PopupMenuController(m_popup);
            controller->m_Index=index;
            controller->m_Widget=this;

            TagModelProxy* proxymodel = static_cast<TagModelProxy*>(model());

            if (proxymodel->isHeader(index.row())) {
                if (proxymodel->getGroupingColumn()==ColumnManager::Album) {
                    const TagItem::List rows = proxymodel->getRowsForHeader(index.row());
                    m_popup->addMenu(new FreeDB::QueryMenu(proxymodel, rows, m_popup));
                }
                m_popup->popup (mapToGlobal(pos));
                return;
            }

            if (index.column()==ColumnManager::Artist) {
                QStringList proposals=proxymodel->getRow(index.row())->getProposals();
                if (proposals.size()>0) {
                    QMenu* m1=new QMenu(i18nc("Change an artist or title","Change this name to ..."),m_popup);
                    QMenu* m2=new QMenu(i18n("Change artist name to ..."),m_popup);
                    foreach (QString proposal, proposals) {
                        m1->addAction(proposal,controller,SLOT(setModifiedContentForOne()))->setData(proposal);
                        m2->addAction(proposal,controller,SLOT(setModifiedContentForAll()))->setData(proposal);
                    }
                    m_popup->addMenu(m1);
                    m_popup->addMenu(m2);
                }
            }

            TagItem::List selectedRows=getSelectedRowsSorted();
            controller->m_SelectedRows=selectedRows;

            if (selectionModel()->hasSelection()) {
                m_popup->addSeparator();

                m_popup->addAction (KStandardAction::copy(m_Controler, SLOT(copy()), m_popup));
                m_popup->addAction (KStandardAction::paste(m_Controler, SLOT(paste()), m_popup));

                if (selectedRows.size()>0) {
                    // Playlist menu
                    m_popup->addAction (i18n("Write to playlist"),this, SLOT(writeSelectionToM3U()));
                    // FreeDB query
                    if (selectedRows.size()<40) {
                        m_popup->addMenu(new FreeDB::QueryMenu(proxymodel, selectedRows, m_popup));
                    }
                }

                // check for phrase insertion (can't use selectedRows as not complete selected rows are
                // considered
                bool oneRow=true;
                const QModelIndexList selection=selectionModel()->selectedIndexes();
                for (int i=1;i<selection.size();i++) {
                    if (selection[i-1].row()!=selection[i].row()) {
                        oneRow=false;
                        break;
                    }
                }

                if (oneRow) {
                    m_popup->addAction(i18n("Add as phrase"), m_Controler, SLOT(addToPhrases()));
#ifdef TEXEL_OFA_DETECTION
                    m_popup->addAction("Show PUID", this, SLOT(showPUID()))->setData(
                                QVariant::fromValue<TagItem*>
                                (dynamic_cast<const TagModelProxy*>(index.model())->getRow(index.row())));
#endif
                }
            }

            // Check if all selected rows are in the same directory
            DirTree::Dir* dir=0l;
            if (selectedRows.size()>0) {
                foreach (TagItem* item, selectedRows) {
                    if (dir==0l) dir=item->getPathPointer(Columns::Column::RoleDisplay);
                    else {
                        if (dir!=item->getPathPointer(Columns::Column::RoleDisplay)) {
                            dir=0l;
                            break;
                        }
                    }
                }
            }
            else {
                dir=proxymodel->getRow(index.row())->getPathPointer(Columns::Column::RoleDisplay);
            }

            if (dir!=0l) {
                // All items are from the same directory
                if (dir->hasAdditionalInformationFiles()) {
                    QMenu* metaDataMenu=new QMenu(i18n("Metadata"),m_popup);
                    foreach (DirTree::Dir::MetaData data,dir->getMetaData()) {
                        if (data.first==DirTree::CDDB) {
                            metaDataMenu->addAction(i18n("Take CDDB data from file %1",data.second),
                                                    controller,
                                                    SLOT(parseCDDBFile())
                                                   )->setData(dir->getName()+QDir::separator()+data.second);
                        }
                        else if (data.first==DirTree::Image) {
                            const QString imgName=dir->getName()+QDir::separator()+data.second;
                            QPixmap img;
                            img.load(imgName);
                            QIcon icon(img);
                            metaDataMenu->addAction(icon,
                                                    i18n("Add image %1 as cover",data.second),
                                                    controller,
                                                    SLOT(addLocalImageAsCover())
                                                   )->setData(imgName);
                        }
                    }
                    m_popup->addMenu(metaDataMenu);
                }
            }

            if (selectedRows.size()>0) {
                m_popup->addAction(i18n("Add cover image"), this, SLOT(addCoverImagesToSelected()));

                bool hasIsMarked=false;
                bool hasIsNotMarked=false;
                foreach (TagItem* item, selectedRows) {
                    hasIsMarked|=item->isToDelete();
                    hasIsNotMarked|=(!item->isToDelete());
                }

                if (hasIsMarked) {
                    m_popup->addAction(i18n("Remove deferred deletion tag"),controller,SLOT(alterDeferedDeletionMark()))->setData(QVariant(false));
                }
                if (hasIsNotMarked) {
                    m_popup->addAction(i18n("Set deferred deletion tag"),controller,SLOT(alterDeferedDeletionMark()))->setData(QVariant(true));
                }
            }

            m_popup->popup (mapToGlobal(pos));
        }
    }

#ifdef USE_MUSICBRAINZ_SUPPORT
    void TableWidget::showPUID ()
    {
        PUID* puid = new PUID();
        puid->getPUID(dynamic_cast<QAction*>(sender())->data().value<TagItem*>());
    }
#endif

    void TableWidget::addCoverImages (const TagItem::List& rows)
    {
        if (rows.empty()) return;

        TagModelProxy* m=dynamic_cast<TagModelProxy*>(model());
        AddCoverImages* worker = new AddCoverImages(m,rows);
        worker->search(rows);
    }

    void TableWidget::addCoverImagesToSelected()
    {
        TagItem::List selection = getSelectedRowsSorted();

        if (selection.isEmpty()) {
            if (!currentIndex().isValid()) {
                return;
            }
            selection << dynamic_cast<TagModelProxy*>(model())->getRow(currentIndex().row());
        }
        addCoverImages (selection);
    }

    void TableWidget::addCoverImagesToAll ()
    {
        TagItem::List allRows;
        for (int i=0;i<model()->rowCount();i++) {
            TagItem* item = dynamic_cast<TagModelProxy*>(model())->getRow(i);
            if (item!=0l) allRows << item;
        }
        addCoverImages(allRows);
    }

    void TableWidget::addCoverImagesWhenNonExists ()
    {
        CoverSearch::ToolBox toolbox;
        CoverSearch::FilterForCoverJob* job=toolbox.filterForCover(dynamic_cast<TagModelProxy*>(model()), false);
        m_Controler->addBackgroundJob(job);
        connect (job, SIGNAL(result(KJob*)), this, SLOT(onFinishedCoverFilterJob1(KJob*)));
        job->start();
    }

    void TableWidget::onFinishedCoverFilterJob1 (KJob* job)
    {
        CoverSearch::FilterForCoverJob* filterJob = dynamic_cast<CoverSearch::FilterForCoverJob*>(job);
        if (job->error()==0) {
            addCoverImages(filterJob->getFoundCovers());
        }
        filterJob->deleteLater();
    }

    void TableWidget::writeSelectionToM3U ()
    {
        TagModelProxy* m=dynamic_cast<TagModelProxy*>(model());
        TagModel* sourceModel = dynamic_cast<TagModel*>(m->sourceModel());

        TagItem::List tracks=getSelectedRowsSorted();

        M3U::List playlist = M3U::List::createPlaylist(tracks);
        playlist.save();
        sourceModel->addPlaylist(playlist);
    }

    void TableWidget::headerContextMenu (const QPoint& pos)
    {
        const int p=horizontalHeader()->logicalIndexAt(pos);

        delete m_popup;
        m_popup = new QMenu(this);

        if (!selectionModel()->isColumnSelected(p,QModelIndex())) {
            m_popup->addAction (i18nc("Verb","Select"), this, SLOT(triggerSelectColumn()))->setData(p);
        }
        m_popup->addAction (i18n("Optimize column width"), this, SLOT(triggerResizeColumn()))->setData(p);
        m_popup->addAction (i18n("Optimize all width"), this, SLOT(resizeColumnsToContents()));
        m_popup->addAction (i18n("Fit columns to screen"), this, SLOT(triggerFitToScreen()));

        //
        // add hide/unhide menu items
        //
        m_popup->addSeparator();
        m_popup->addAction (i18nc("Verb","Hide"), this, SLOT(triggerHideColumn()))->setData (-p);
        if (p!=0 && isColumnHidden(p-1)) m_popup->addAction (i18nc("Refers to colum","Unhide left"), this, SLOT(triggerHideColumn()))->setData(p-1);
        if (p!=ColumnManager::instance()->getColumnNumber()-1 && isColumnHidden(p+1)) m_popup->addAction ("Unhide right", this, SLOT(triggerHideColumn()))->setData(p+1);

        // General hide/unhide
        QMenu* hideMenu = new QMenu (i18nc("Verb","Hide ..."), m_popup);
        QMenu* unHideMenu = new QMenu (i18nc("Verb","Unhide ..."), m_popup);
        for (int i=0;i<ColumnManager::instance()->getColumnNumber();i++) {
            if (i==p) continue;
            if (isColumnHidden(i)) {
                QAction* a=unHideMenu->addAction (ColumnManager::instance()->getName(i),this, SLOT(triggerHideColumn()));
                a->setData(i);
            }
            else {
                QAction* a=hideMenu->addAction (ColumnManager::instance()->getName(i),this, SLOT(triggerHideColumn()));
                a->setData(-i);
            }
        }
        m_popup->addMenu(hideMenu);
        if (horizontalHeader()->sectionsHidden()) m_popup->addMenu(unHideMenu);

        // show
        m_popup->popup (mapToGlobal(pos));
    }

    void TableWidget::triggerHideColumn ()
    {
        const int col = dynamic_cast<QAction*>(sender())->data().toInt();
        if (col>0) {
            showColumn(col);
        }
        else {
            hideColumn(-col);
        }
    }

    void TableWidget::triggerSelectColumn ()
    {
        const int col = dynamic_cast<QAction*>(sender())->data().toInt();
        selectColumn (col);
    }

    void TableWidget::triggerResizeColumn ()
    {
        const int col = dynamic_cast<QAction*>(sender())->data().toInt();
        resizeColumnToContents (col);
    }

    void TableWidget::triggerFitToScreen ()
    {
        const int w=width()-verticalHeader()->width()-1;
        const int n=ColumnManager::instance()->getColumnNumber();
        QVarLengthArray<int> sized(n);
        int sumSize=0;
        for (int i=0;i<n;i++) {
            if (!isColumnHidden(i)) {
                sized[i]=sizeHintForColumn(i);
                sumSize+=sized[i];
            }
        }

        if (sumSize>w) {
            float fraction=0.0;
            int hidden=0;
            do {
                fraction=(float)w/(float)sumSize;
                if (fraction<0.7) {
                    const ColumnManager::ColumnNames name = ColumnManager::instance()->getLeastPriority(hidden);
                    if (name==ColumnManager::Invalid) break;
                    sumSize-=sized[(int)name];
                    sized[(int)name]=0;
                    hideColumn((int)name);
                }
            } while (fraction<0.7);

            for (int i=0;i<n;i++) if (sized[i]!=0) setColumnWidth(i,sized[i]*fraction-1);
        }
    }

    QModelIndex TableWidget::moveCursor (CursorAction cursorAction, Qt::KeyboardModifiers modifiers)
    {
        QModelIndex idx = QTableView::moveCursor(cursorAction,modifiers);
        TagModelProxy* m=dynamic_cast<TagModelProxy*>(model());
        if (cursorAction==QAbstractItemView::MovePageDown) {
            int row=idx.row();
            while (row<model()->rowCount() && m->isHeader(row)) ++row;
            if (row==model()->rowCount()) {
                // last lines are all headers, move back to the last unheaders row
                row=model()->rowCount()-1;
                while (row>0 && m->isHeader(row)) --row;
            }
            idx=model()->index(row,idx.column());
        }
        else if (cursorAction==QAbstractItemView::MovePageUp)  {
            int row=idx.row();
            while (row>=0 && m->isHeader(row)) --row;
            if (row<0) {
                // first lines are all headers, move back to the last unheaders row
                row=0;
                while (row<model()->rowCount() && m->isHeader(row)) ++row;
            }
            idx=model()->index(row,idx.column());
        }
        return idx;
    }

    void TableWidget::keyReleaseEvent (QKeyEvent* event)
    {
        if (event->key()==Qt::Key_Delete) {
            model()->setData(currentIndex(),QVariant(""));
        }
        else QTableView::keyReleaseEvent(event);
    }

    bool TableWidget::viewportEvent (QEvent* event)
    {
        if (event->type()==QEvent::ToolTip)
        {
            QHelpEvent* e = dynamic_cast<QHelpEvent*>(event);
            QModelIndex index=indexAt(e->pos());
            TagModel* m=model_cast(model());
            if (index.isValid() && index.row()<model()->rowCount()) {
                if (m->getRow(index.row())!=0L &&
                    m->getRow(index.row())->hasInputProblem(index.column())) {
                    m->getRow(index.row())->displayTooltip(index.column(), e->globalPos(),
                                                           // visual rect translated to global coordinates
                                                           visualRect(index).translated(viewport()->mapToGlobal(QPoint(0,0))));
                    event->accept();
                    return true;
                }
            }
        }

        return super::viewportEvent (event);
    }

    void TableWidget::commitData (QWidget* editor)
    {
        super::commitData(editor);

        if (getEditBuddy()!=0L) {
            getEditBuddy()->setText(model()->data(currentIndex(),TagModel::RawModeRole).toString());
        }
    }

    void TableWidget::dragEnterEvent(QDragEnterEvent *event)
    {
        if (Preferences::instance()->debug.m_TraceDragging) {
            const QMimeData* data = event->mimeData();
            QString formats=data->formats().join(",");
            kDebug() << "Detected formats: " << formats << endl;
        }
        super::dragEnterEvent (event);
    }

    void TableWidget::clearHeaders ()
    {
        clearSpans();
    }

    void TableWidget::addHeader (const QList<int>& rows)
    {
        foreach (int row, rows) {
            setSpan(row,0,1,ColumnManager::instance()->getColumnNumber());
        }
    }

    void TableWidget::mouseDoubleClickEvent (QMouseEvent* event)
    {
        QModelIndex idx=indexAt(event->pos());
        // In case of double click on a header mark the belonging rows
        TagModelProxy* m=static_cast<TagModelProxy*>(model());
        if (m->isHeader(idx.row())) {
            TagItem::List toMark = m->getRowsForHeader(idx.row());
            QItemSelectionModel::SelectionFlags command = QItemSelectionModel::Select;
            if (!(event->modifiers()&Qt::ControlModifier)) {
                command|=QItemSelectionModel::Clear;
            }
            selectionModel()->select(QItemSelection(model()->index(toMark.first()->getRow(),0),
                                                    model()->index(toMark.last()->getRow(),
                                                    ColumnManager::instance()->getColumnNumber()-1)),
                                     command);
            event->accept();
        }
        else {
            super::mouseDoubleClickEvent(event);
        }
    }

    void TableWidget::clearSelectedCells ()
    {
        foreach (QModelIndex index, selectedIndexes()) {
            if (ColumnManager::instance()->getType(index.column())==QMetaType::Int) {
                model()->setData(index,"0",Qt::EditRole);
            }
            else {
                model()->setData(index,"",Qt::EditRole);
            }
        }
    }

    void TableWidget::onModelFilterStart ()
    {
        m_OldPosition=auto_ptr<OldPosition>(new OldPosition);
        m_OldPosition->m_Index = currentIndex();
        TagModelProxy* m=dynamic_cast<TagModelProxy*>(model());
        if (m!=0L) {
            m_OldPosition->m_SourceIndex = m->mapToSource(model()->index(m_OldPosition->m_Index.row(),
                                                                         m_OldPosition->m_Index.column()));
        }
    }

    void TableWidget::selectAllInSameDirectory(int row)
    {
        TagModelProxy* proxy = static_cast<TagModelProxy*>(model());
        DirTree::Dir* dir = proxy->getRow(row)->getPathPointer(Columns::Column::RoleDisplay);
        foreach (TagItem* item, dir->getTagItems()) {
            QVector<int> aRow=proxy->getRow(item);
            if (!aRow.isEmpty()) {
                selectionModel()->select(model()->index(aRow.first(),0),QItemSelectionModel::Select|QItemSelectionModel::Rows);
            }
        }
    }

    void TableWidget::selectAllInSameDirectory()
    {
        selectAllInSameDirectory(currentIndex().row());
    }

    void TableWidget::selectAllInSameDirectoryRec ()
    {
        TagModelProxy* proxy = static_cast<TagModelProxy*>(model());
        DirTree::Dir* dir = proxy->getRow(currentIndex().row())->getPathPointer(Columns::Column::RoleDisplay);
        selectAllInSameDirectoryRec(dir);
    }

    void TableWidget::selectAllInSameDirectoryRec (DirTree::Dir* dir)
    {
        TagModelProxy* proxy = static_cast<TagModelProxy*>(model());
        foreach (TagItem* item, dir->getTagItems()) {
            QVector<int> aRow=proxy->getRow(item);
            if (!aRow.isEmpty()) {
                selectionModel()->select(model()->index(aRow.first(),0),QItemSelectionModel::Select|QItemSelectionModel::Rows);
            }
        }

        foreach (DirTree::Dir* d, dir->getChildren()) {
            selectAllInSameDirectoryRec (d);
        }
    }

    void TableWidget::deferedDeletionTag (bool set)
    {
        std::auto_ptr<PopupMenuController> controller (new PopupMenuController(m_popup));
        controller->m_Index=currentIndex();
        controller->m_Widget=this;
        TagItem::List selectedRows=getSelectedRowsSorted();
        controller->m_SelectedRows=selectedRows;
        controller->alterDeferedDeletionMark(set);
    }

    void TableWidget::setDeferedDeletionTag ()
    {
        deferedDeletionTag(true);
    }

    void TableWidget::unsetDeferedDeletionTag ()
    {
        deferedDeletionTag(false);
    }


}

#include "texeltablewidget.moc"
