/***********************************************************************
*************************************************************************/

#include "welcome.h"
#include "aid.h"
#include "util/filedialog.h"
#include "model/texeltagmodel.h"
#include "options/preferences.h"

#include <QMenu>
#include <QFileInfo>
#include <QDragEnterEvent>
#include <QDropEvent>

#include <kmessagebox.h>
#include <kdebug.h>

using namespace std;

namespace {
    const QString s_Group("Welcome");
    const QString toolButton("toolButton_");
}

namespace Texel {

    Welcome::Welcome(QWidget *parent) :
            QWidget(parent),
            m_CoreModel(0l)
    {
        setupUi(this);

        // Configuration of the first four buttons is done externally as the buttons are
        // in the public interface. This helps to let the class as dumb as possible, as
        // it must nothing know about most recently used directories.

        KConfigGroup config(KGlobal::config(),s_Group);
        for (int i=5;i<9;i++) {
            QString name=toolButton+QString::number(i);
            if (config.hasKey(name)) {
                KUrl dir=config.readEntry(name);
                QFileInfo info(dir.path());
                QToolButton* button=findChild<QToolButton*>(name);
                if (info.exists() && info.isReadable()) {
                    if (info.isDir()) {
                        configButtonForDir (button, dir);
                    }
                    else if (info.isFile()) {
                        configButtonForList(button, dir);
                    }
                }
                else {
                    // Could no access file. Add a chance for a error dialog
                    button->setText(i18n("Could not access URL %1 [click for more information]",dir.url()));
                    connect (button,SIGNAL(clicked()),this,SLOT(showUnaccessibleError()));
                }
            }
            else {
                QToolButton* button=findChild<QToolButton*>(name);
                resetButtonForDir (button);
            }
        }
    }

    void Welcome::showUnaccessibleError()
    {
        QToolButton* button=dynamic_cast<QToolButton*>(sender());
        QString name=button->objectName();
        KConfigGroup config(KGlobal::config(),s_Group);
        KUrl dir=config.readEntry(name);

        QFileInfo info(dir.path());
        QString message;
        if (!info.exists()) {
            message=i18n("%1 does not exist",dir.path());
        }
        else if (!info.isReadable()) {
            message=i18n("Access rights to access %1 missing!",dir.path());
        }
        else if (!info.isFile()) {
            message=i18n("%1 is not a file!",dir.path());
        }

        int result=KMessageBox::messageBox(NULL,
                                           KMessageBox::WarningYesNo,
                                           message,
                                           i18n("Could not open directory"),
                                           KStandardGuiItem::ok(),
                                           KStandardGuiItem::reset());

        if (result==KMessageBox::No) {
            resetButtonForDir(button);
        }
    }

    void Welcome::changeEvent(QEvent *e)
    {
        QWidget::changeEvent(e);
        switch (e->type()) {
        case QEvent::LanguageChange:
            retranslateUi(this);
            break;
        default:
            break;
        }
    }

    /**
  * Selects a directory and bounds it to the button.
  */

    void Welcome::addFixedDir ()
    {
        const QString name = sender()->objectName();
        const KUrl dir=FileDialog::getExistingDirectory(i18n("Select directory for quick access"));

        if (dir.isValid()) {
            KConfigGroup config(KGlobal::config(),s_Group);
            config.writeEntry(name,dir);
            config.sync();

            QToolButton* button=dynamic_cast<QToolButton*>(sender());
            if (button==NULL) {
                // If the signal does not come from the button but from a action use
                // the actions user data and retrieve the meant button from there
                QAction* action=dynamic_cast<QAction*>(sender());
                QString name = action->data().toString();
                button=findChild<QToolButton*>(name);
            }
            if (button!=NULL) {
                configButtonForDir (button,dir);
            }
        }
    }

    /**
  * Selects a file and bounds it to the button.
  */

    void Welcome::addFixedList ()
    {
        QToolButton* button=dynamic_cast<QToolButton*>(sender());
        if (button==0l) {
            // If the signal does not come from the button but from a action use
            // the actions user data and retrieve the meant button from there
            QAction* action=dynamic_cast<QAction*>(sender());
            QString name = action->data().toString();
            button=findChild<QToolButton*>(name);
        }
        // No button did trigger this. Ignore it.
        if (button==0l) return;

        const QString name = button->objectName();
        const KUrl file=FileDialog::getLoadFileName("*.txl",i18n("Select list for quick access"));

        if (file.isValid()) {
            KConfigGroup config(KGlobal::config(),s_Group);
            config.writeEntry(name,file);
            config.sync();

            configButtonForList (button,file);
        }
    }

    void Welcome::resetButton (QToolButton* button)
    {
        button->disconnect (SIGNAL(clicked()));
        if (button->menu()!=0l) {
            button->menu()->deleteLater();
        }
        button->setMenu(0l);
    }

    void Welcome::configButtonForList (QToolButton* button, const KUrl& list)
    {
        resetButton(button);

        button->setText(i18nc("Command on button to load a certain list","Load list ")+list.path());
        QMenu* menu=new QMenu(button);
        menu->addAction(i18nc("Command to reset a button command to default","Reset"), this, SLOT(resetButtonForDir()))->setData(button->objectName());
        button->setMenu(menu);
        connect (button, SIGNAL(clicked()), SLOT(loadItem()));
    }

    void Welcome::configButtonForDir (QToolButton* button, const KUrl& dir)
    {
        resetButton (button);

        button->setText(i18nc("Command on button to load a certain dir","Load dir ")+dir.path());
        QMenu* menu=new QMenu(button);
        menu->addAction(i18nc("Command to reset a button command to default","Reset"), this, SLOT(resetButtonForDir()))->setData(button->objectName());
        button->setMenu(menu);
        connect (button, SIGNAL(clicked()), SLOT(loadItem()));
    }

    void Welcome::loadItem ()
    {
        QString name = sender()->objectName();
        KConfigGroup config(KGlobal::config(),s_Group);
        KUrl dir=config.readEntry(name);
        QFileInfo info(dir.path());
        if (info.exists() && info.isReadable()) {
            if (info.isDir()) {
                emit loadDir (dir);
            }
            else if (info.isFile()) {
                emit loadList (dir);
            }
        }
    }

    void Welcome::resetButtonForDir (QToolButton* button)
    {
        if (button!=0l) {
            resetButton(button);
            button->setText(i18n("Click to add a fixed directory"));

            QMenu* menu=new QMenu(button);
            menu->addAction(i18nc("List refers to saves tag list","Add a fixed list entry"), this, SLOT(addFixedList()))->setData(button->objectName());
            button->setMenu(menu);

            KConfigGroup config(KGlobal::config(),s_Group);
            config.deleteEntry(button->objectName());
            config.sync();

            connect (button, SIGNAL(clicked()), this, SLOT(addFixedDir()));
        }
        else {
            kDebug() << "Error: button was NULL" << endl;
        }
    }

    void Welcome::resetButtonForDir ()
    {
        QToolButton* button=findChild<QToolButton*>(dynamic_cast<QAction*>(sender())->data().toString());
        resetButtonForDir(button);
    }

    void Welcome::dragEnterEvent(QDragEnterEvent* event)
    {
        if (Preferences::instance()->debug.m_TraceDragging) {
            const QMimeData* data = event->mimeData();
            const QString formats=data->formats().join(",");
            kDebug() << "Detected formats: " << formats << endl;
        }

        event->acceptProposedAction();
    }

    void Welcome::dropEvent(QDropEvent* event)
    {
        event->acceptProposedAction();
        m_CoreModel->load (event->mimeData()->urls());
    }

}

#include "welcome.moc"
